#!/bin/sh

# Martin Peters, Jan 2009

##################################################
# Name: Usage
# Desc: Prints a usage message and then exits
# Args:
##################################################
usage () {
  echo "                                                                     "
  echo " Usage:                                                              "
  echo "  genMetalFF.sh -n Name                                              "
  echo "                -h help                                              "
  echo "                                                                     "
  echo " Description: This script helps build a metal FF                     "
  echo "                                                                     "
  echo " Arguments: Name name of the original pdb file                       "
  echo "            Output is the log file                                   "
  echo "                                                                     "
  return
}

##################################################
# Name: readCommandLine
# Desc: Read Command Line Arguments
# Args: $@
##################################################
readCommandLine () {
  version=1.0
  startdir=$PWD
  name="-"

  bName=0
  bOutput=0

  while [ $# -gt 0 ]
  do
     case "$1" in
        -n | --name ) name=$2 ; bName=1 ; shift 2 ;;
        -h | --help ) usage ; exit ;;
        -ver* ) echo "genMetalFF, version $version" ; exit 1 ;;
        * ) echo "Unrecognised flag $arg" ; usage ; exit 1 ;;
     esac
  done
}


##################################################
# Name: genScripts
# Desc: 
# Args: $1  -> name
##################################################
genScripts () {
  if [ $# -lt 1 ] ; then
     usage
     return 1
  fi

  if [ $bName -eq 0 ]
  then
     usage
     echo " Please provide a name "
     return 1
  fi

  readme=$1.README
  rm -rf $readme
  echo " #### genMetalFF ####" >> $readme
  echo " #### Ensure that the environment variable: " >> $readme
  echo " ####   AMBERHOME is set correctly " >> $readme
  echo "" >> $readme
  echo "1)" >> $readme
  echo " Downloaded $1.pdb" >> $readme
  echo "" >> $readme
  echo "2)" >> $readme
  echo "Removed headings and connect info from pdb --> $1_fixed.pdb" >> $readme
  echo "" >> $readme
  echo "3)" >> $readme
  echo "MCPB -i $1_addHs.bcl -l $1_addHs.bcl.log" >> $readme
  echo "   Adds Hs to $1_fixed.pdb --> $1_fixed_H.pdb" >> $readme
  echo "" >> $readme
  echo "4)" >> $readme
  echo "Reduced the size of $1.pdb --> $1_reduced.pdb" >> $readme
  echo "   Reduce size of system while we generate parameters, speeds up " >> $readme
  echo "   debugging if required" >> $readme
  echo "" >> $readme
  echo "5)" >> $readme
  echo "Create settings file $1_settings.bcl" >> $readme
  echo "   Reduces the size of the bcl files and avoid retyping the same info" >> $readme
  echo "   The variable 'NAME' is set" >> $readme
  echo "" >> $readme
  echo "6)" >> $readme
  echo "MCPB -i $1_sidechain.bcl -l $1_sidechain.bcl.log" >> $readme
  echo "   Creates the smallest possible model of the metal center to generate bond and" >> $readme
  echo "   angle force constants" >> $readme
  echo "   The 'NAME'.xml and 'NAME'_sidechain.com files are created" >> $readme
  echo "" >> $readme
  echo "7)" >> $readme
  echo "MCPB -i $1_addStdMol.bcl -l $1_addStdMol.bcl.log" >> $readme
  echo "   The 'NAME'_stdMol.xml file is created" >> $readme
  echo "" >> $readme
  echo "8)" >> $readme
  echo "MCPB -i $1_large.bcl -l $1_large.bcl.log" >> $readme
  echo "   Creates the large model of the metal center to generate charges" >> $readme
  echo "   The 'NAME'_large.com is created" >> $readme
  echo "" >> $readme
  echo "9)" >> $readme 
  echo "./createRespAdditions.sh " >> $readme
  echo "  bb Definitions:" >> $readme
  echo "    - 0 No restraints" >> $readme
  echo "    - 1 Heavy Atoms in Backbone (bb_heavy, [ca, n, c, o]) set to parm94 values" >> $readme
  echo "    - 2 Atoms in Backbone (bb, [ca, h, ha, n, nh, c, o]) set to parm94 values" >> $readme
  echo "    - 3 Atoms in Backbone plus CB (bbb, [ca, h, ha, n, nh, c, o, cb]) set to parm94 values" >> $readme
  echo "   The 'NAME'_large_respAdds0, 'NAME'_large_respAdds1, 'NAME'_large_respAdds2, and 'NAME'_large_respAdds3 files are created" >> $readme
  echo "" >> $readme
  echo "10)" >> $readme
  echo "./getCharges.sh 0; ./getCharges.sh 1; ./getCharges.sh 2; ./getCharges.sh 3" >> $readme
  echo "    Runs respgen, espgen, and resp" >> $readme
  echo "    Requires 'Name'_large_mk.log ">> $readme
  echo "" >> $readme
  echo "11)" >> $readme
  echo "./createLibraries.sh" >> $readme
  echo "    The 'NAME'_chg0.xml, 'NAME'_chg1.xml, 'NAME'_chg2.xml, and 'NAME'_chg3.xml files are created" >> $readme
  echo "" >> $readme
  echo "12)" >> $readme
  echo "Create force field modification files " >> $readme
  echo " a) MCPB -i $1_sidechain_fc_sem.bcl -l $1_sidechain_fc_sem.bcl.log" >> $readme
  echo "    The 'NAME'_params_fc_sem.xml file is created" >> $readme
  echo "    Requires 'Name'_sidechain_opt_md.fchk ">> $readme
  echo " b) MCPB -i $1_sidechain_fc_md.bcl -l $1_sidechain_fc_md.bcl.log" >> $readme
  echo "    The 'NAME'_params_fc_md.xml file is created" >> $readme
  echo "    Requires 'Name'_sidechain_opt_md.log ">> $readme
  echo " c,d,e) grep \"Bond: |\" $1_sidechain_fc_sem.bcl.log > $1_sem.txt;" >> $readme
  echo "    grep \"Angle: |\" $1_sidechain_fc_sem.bcl.log >> $1_sem.txt;" >> $readme
  echo "    grep \"Bond: |\" $1_sidechain_fc_md.bcl.log > $1_md.txt;" >> $readme
  echo "    grep \"Angle: |\" $1_sidechain_fc_md.bcl.log >> $1_md.txt;" >> $readme
  echo "    diff $1_sem.txt $1_md.txt " >> $readme
  echo "" >> $readme
  echo "13)" >> $readme
  echo "Create AMBER prep/frcmod files " >> $readme
  echo " a) MCPB -i $1_toAmberFormats_sem.bcl -l $1_toAmberFormats_sem.bcl.log" >> $readme
  echo "    The 'NAME'_sem.frcmod and 'NAME'_chg2.prep files are created" >> $readme
  echo " b) MCPB -i $1_toAmberFormats_md.bcl -l $1_toAmberFormats_md.bcl.log" >> $readme
  echo "    The 'NAME'_md.frcmod and 'NAME'_chg2.prep files are created" >> $readme
  echo "" >> $readme
  echo "14)" >> $readme
  echo "MCPB -i $1_resNames.bcl -l $1_resNames.bcl.log" >> $readme
  echo "    The 'NAME'.leaprc, 'NAME'.pml, and 'NAME'_mcpb.pdb files are created" >> $readme


  f1=$1_addHs.bcl
  rm -rf $f1
  echo "# Load settings" >> $f1
  echo "source $1_settings.bcl" >> $f1
#  echo "# Load AMBER Parameters" >> $f1
#  echo "loadParam $AMBERHOME/dat/mtkpp/parm94.xml" >> $f1
#  echo "loadParam $AMBERHOME/dat/mtkpp/parm_gaff.xml" >> $f1
#  echo "loadParam $AMBERHOME/dat/mtkpp/metals/metalParm.xml" >> $f1
#  echo "" >> $f1
#  echo "# Load AMBER libraries" >> $f1
#  echo "loadLib $AMBERHOME/dat/mtkpp/amino94.xml" >> $f1
#  echo "loadLib $AMBERHOME/dat/mtkpp/aminont94.xml" >> $f1
#  echo "loadLib $AMBERHOME/dat/mtkpp/aminoct94.xml" >> $f1
#  echo "loadLib $AMBERHOME/dat/mtkpp/fragLib/terminal.xml" >> $f1
#  echo "loadLib $AMBERHOME/dat/mtkpp/metals/metals.xml" >> $f1
  echo "" >> $f1
  echo "# Open PDB file" >> $f1
  echo "readPdb $1 $1_fixed.pdb" >> $f1
  echo "" >> $f1
  echo "# Assign S-S bonds" >> $f1
  echo "assignDisulfideBonds" >> $f1
  echo "" >> $f1
  echo "# Atom type" >> $f1
  echo "atomType" >> $f1
  echo "" >> $f1
  echo "# Assign bonds, angles, and torsions" >> $f1
  echo "assignConnectivity" >> $f1
  echo "" >> $f1
  echo "# Add hydrogens" >> $f1
  echo "addHs /$1/" >> $f1
  echo "" >> $f1
  echo "renumber" >> $f1
  echo "" >> $f1
  echo "# Write pdb/mol files" >> $f1
  echo "writePdb $1_fixed_H.pdb" >> $f1
  echo "" >> $f1
  echo "# Exit MCPB" >> $f1
  echo "quit" >> $f1


  # Settings file
  f2=$1_settings.bcl
  rm -rf $f2
  echo "set NAME $1" >> $f2
  echo "" >> $f2
  echo "# Load AMBER Parameters" >> $f2
  echo "loadParam $AMBERHOME/dat/mtkpp/parm94.xml" >> $f2
  echo "loadParam $AMBERHOME/dat/mtkpp/parm_gaff.xml" >> $f2
  echo "loadParam $AMBERHOME/dat/mtkpp/metals/metalParm.xml" >> $f2
  echo "" >> $f2
  echo "# Load AMBER libraries" >> $f2
  echo "loadLib $AMBERHOME/dat/mtkpp/amino94.xml" >> $f2
  echo "loadLib $AMBERHOME/dat/mtkpp/aminont94.xml" >> $f2
  echo "loadLib $AMBERHOME/dat/mtkpp/aminoct94.xml" >> $f2
  echo "loadLib $AMBERHOME/dat/mtkpp/fragLib/terminal.xml" >> $f2
  echo "loadLib $AMBERHOME/dat/mtkpp/metals/metals.xml" >> $f2


  # 
  f3=getCharges.sh
  rm -rf $f3
  echo "#!/bin/sh" >> $f3
  echo "" >> $f3
  echo "# Set name " >> $f3
  echo "name=$1_large" >> $f3
  echo "" >> $f3
  echo "# Create resp1/resp2 files" >> $f3
  echo "./run2_respgen \$name" >> $f3
  echo "" >> $f3
  echo "# Remove 2 blank lines from resp1/resp2" >> $f3
  echo "./removeLast2Lines.sh \$name" >> $f3
  echo "cat \$name""\_respAdds\$1 >> \$name.resp1" >> $f3
  echo "cat \$name""\_respAdds\$1 >> \$name.resp2" >> $f3
  echo "mv \$name.resp1 \$name\_mk.resp1" >> $f3
  echo "mv \$name.resp2 \$name\_mk.resp2" >> $f3
  echo "" >> $f3
  echo "# Create esp file from MK g03 log file" >> $f3
  echo "./run3_espgen \$name\_mk" >> $f3
  echo "" >> $f3
  echo "# Run RESP and get charges " >> $f3
  echo "./run4_resp \$name\_mk" >> $f3
  echo "" >> $f3
  echo "# Outputed charges are in name_mkX.resp2.chg" >> $f3
  echo "mv \$name\_mk.resp2.chg \$name\_mk\$1.resp2.chg " >> $f3
  chmod +x getCharges.sh


  # 
  f3=createRespAdditions.sh
  rm -rf $f3
  echo "#!/bin/sh" >> $f3
  echo "" >> $f3
  echo "chgTypes=\"0 1 2 3\"" >> $f3
  echo "for c in \${chgTypes}" >> $f3
  echo "do" >> $f3
  echo "  $AMBERHOME/bin/MCPB -i $1_large_mk\$c.bcl -l $1_large_mk\$c.bcl.log" >> $f3
  echo "  mv $1_large_respAdds $1_large_respAdds\$c" >> $f3
  echo "done" >> $f3
  chmod +x createRespAdditions.sh


  # 
  f3=createLibraries.sh
  rm -rf $f3
  echo "#!/bin/sh" >> $f3
  echo "" >> $f3
  echo "chgTypes=\"0 1 2 3\"" >> $f3
  echo "for c in \${chgTypes}" >> $f3
  echo "do" >> $f3
  echo "  $AMBERHOME/bin/MCPB -i $1_large_chg\$c.bcl -l $1_large_chg\$c.bcl.log" >> $f3
  echo "done" >> $f3
  chmod +x createLibraries.sh


  chgTypes="0 1 2 3"
  for c in ${chgTypes}
  do
    # 
    f4=$1_large_mk$c.bcl
    rm -rf $f4
    echo "# Load AMBER Parameters" >> $f4
    echo "loadParam $1_params.xml" >> $f4
    echo "loadParam $AMBERHOME/dat/mtkpp/parm94.xml" >> $f4
    echo "loadParam $AMBERHOME/dat/mtkpp/parm_gaff.xml" >> $f4
    echo "loadParam $AMBERHOME/dat/mtkpp/metals/metalParm.xml" >> $f4
    echo "" >> $f4
    echo "# Load AMBER libraries" >> $f4
    echo "loadLib $AMBERHOME/dat/mtkpp/amino94.xml" >> $f4
    echo "loadLib $AMBERHOME/dat/mtkpp/aminont94.xml" >> $f4
    echo "loadLib $AMBERHOME/dat/mtkpp/aminoct94.xml" >> $f4
    echo "loadLib $AMBERHOME/dat/mtkpp/fragLib/terminal.xml" >> $f4
    echo "loadLib $AMBERHOME/dat/mtkpp/metals/metals.xml" >> $f4
    echo "loadLib $1.xml" >> $f4
    echo "" >> $f4
    echo "# Read PDB FILE" >> $f4
    echo "readPdb $1 $1_large.pdb" >> $f4
    echo "" >> $f4
    echo "# Atom type" >> $f4
    echo "atomType" >> $f4
    echo "" >> $f4
    echo "# Assign bonds, angles, and torsions" >> $f4
    echo "assignConnectivity" >> $f4
    echo "" >> $f4
    echo "# Create respgen additions" >> $f4
    echo "respgenAdditions $1 $1_large $c" >> $f4
    echo "" >> $f4
    echo "# Exit MCPB" >> $f4
    echo "quit" >> $f4


    # 
    f5=$1_large_chg$c.bcl
    rm -rf $f5
    echo "# Load settings" >> $f5
    echo "source $1_settings.bcl" >> $f5
    echo "" >> $f5
    echo "# Load AMBER parameters" >> $f5
    echo "loadParam NAME_params.xml" >> $f5
    echo "" >> $f5
    echo "# Load AMBER libraries" >> $f5
    echo "loadLib NAME_stdMol.xml" >> $f5
    echo "" >> $f5
    echo "# Read PDB file" >> $f5
    echo "readPdb NAME NAME_large.pdb" >> $f5
    echo "" >> $f5
    echo "# Atom type" >> $f5
    echo "atomType" >> $f5
    echo "" >> $f5
    echo "# Add bonds, angles, torsions, and impropers" >> $f5
    echo "assignConnectivity" >> $f5
    echo "" >> $f5
    echo "# Read charge file" >> $f5
    echo "readRespCharges /NAME/NAME_large NAME_large_mk$c.resp2.chg" >> $f5
    echo "" >> $f5
    echo "# Update charges in lib file, mol, groupName" >> $f5
    echo "updateRespCharges /NAME/NAME_large NAME" >> $f5
    echo "" >> $f5
    echo "# Write standard library" >> $f5
    echo "writeLib NAME NAME_chg$c.xml" >> $f5

    echo "" >> $f5
    echo "# Exit MCPB" >> $f5
    echo "quit" >> $f5
  done


  # 
  f6=$1_sidechain_fc_sem.bcl
  rm -rf $f6
  echo "# Load settings" >> $f6
  echo "source $1_settings.bcl" >> $f6
  echo "" >> $f6
  echo "# Load AMBER Parameters" >> $f6
  echo "loadParam NAME_params.xml" >> $f6
  echo "" >> $f6
  echo "# Load AMBER libraries" >> $f6
  echo "loadLib NAME.xml" >> $f6
  echo "" >> $f6
  echo "# Open sd file" >> $f6
  echo "readSdf NAME NAME_sidechain.sdf" >> $f6
  echo "" >> $f6
  echo "# Read G03 Formatted Checkpoint file" >> $f6
  echo "readFormattedChkPtFile NAME_sidechain_opt_md.fchk" >> $f6
  echo "" >> $f6
#  echo "# Update force constants, mol, groupName" >> $f6
#  echo "updateForceConstants /NAME/NAME NAME" >> $f6
  echo "# Update force constants, mol, groupName X Y" >> $f6
  echo "#  X Values: " >> $f6
  echo "#   - 0 Do not update bonds and angle equilibrium values" >> $f6
  echo "#   - 1 Do update bonds and angle equilibrium values" >> $f6
  echo "#  Y Values:" >> $f6
  echo "#   - 0 Seminario Method" >> $f6
  echo "#   - 1 Z-matrix Method" >> $f6
  echo "updateForceConstants /$1/$1 $1 1 0" >> $f6
  echo "" >> $f6
  echo "# Write FRCMOD file" >> $f6
  echo "writeParams NAME NAME_params_fc_sem.xml" >> $f6
  echo "" >> $f6
  echo "# Exit buildCluster" >> $f6
  echo "quit" >> $f6


  # 
  f6a=$1_sidechain_fc_md.bcl
  rm -rf $f6a
  echo "# Load AMBER Parameters" >> $f6a
  echo "loadParam $1_params.xml" >> $f6a
  echo "loadParam $AMBERHOME/dat/mtkpp/parm94.xml" >> $f6a
  echo "loadParam $AMBERHOME/dat/mtkpp/parm_gaff.xml" >> $f6a
  echo "loadParam $AMBERHOME/dat/mtkpp/metals/metalParm.xml" >> $f6a
  echo "" >> $f6a
  echo "# Load AMBER libraries" >> $f6a
  echo "loadLib $AMBERHOME/dat/mtkpp/amino94.xml" >> $f6a
  echo "loadLib $AMBERHOME/dat/mtkpp/aminont94.xml" >> $f6a
  echo "loadLib $AMBERHOME/dat/mtkpp/aminoct94.xml" >> $f6a
  echo "loadLib $AMBERHOME/dat/mtkpp/fragLib/terminal.xml" >> $f6a
  echo "loadLib $AMBERHOME/dat/mtkpp/metals/metals.xml" >> $f6a
  echo "loadLib $1.xml" >> $f6a
  echo "" >> $f6a
  echo "# Open SDF file, col and mol are named $1" >> $f6a
  echo "readSdf $1 $1_sidechain.sdf" >> $f6a
  echo "" >> $f6a
  echo "# Read G03 file" >> $f6a
  echo "readG03Output $1_sidechain_fc_md.log" >> $f6a
  echo "" >> $f6a
  echo "# Update force constants, mol, groupName X Y" >> $f6a
  echo "   X Values: " >> $f6a
  echo "    - 0 Do not update bonds and angle equilibrium values" >> $f6a
  echo "    - 1 Do update bonds and angle equilibrium values" >> $f6a
  echo "   Y Values:" >> $f6a
  echo "    - 0 Seminario Method" >> $f6a
  echo "    - 1 Z-matrix Method" >> $f6a
  echo "updateForceConstants /$1/$1 $1 1 1" >> $f6a
  echo "" >> $f6a
  echo "# Write FRCMOD file" >> $f6a
  echo "writeParams $1 $1_params_fc_md.xml" >> $f6a
  echo "" >> $f6a
  echo "# Exit MCPB" >> $f6a
  echo "quit" >> $f6a


  # 
  f7=$1_toAmberFormats_sem.bcl
  rm -rf $f7
  echo "loadParam $AMBERHOME/dat/mtkpp/parm94.xml" >> $f7
  echo "loadParam $AMBERHOME/dat/mtkpp/parm_gaff.xml" >> $f7
  echo "loadParam $AMBERHOME/dat/mtkpp/metals/metalParm.xml" >> $f7
  echo "" >> $f7
  echo "# " >> $f7
  echo "writeFrcmodFile metals.frcmod metals" >> $f7
  echo "" >> $f7
  echo "loadParam $1_params_fc_sem.xml" >> $f7
  echo "writeFrcmodFile $1_sem.frcmod $1" >> $f7
  echo "" >> $f7
  echo "# " >> $f7
  echo "loadLib $1_chg2.xml" >> $f7
  echo "writePrepFile $1_chg2.prep $1" >> $f7
  echo "" >> $f7
  echo "quit" >> $f7


  # 
  f7=$1_toAmberFormats_md.bcl
  rm -rf $f7
  echo "loadParam $AMBERHOME/dat/mtkpp/parm94.xml" >> $f7
  echo "loadParam $AMBERHOME/dat/mtkpp/parm_gaff.xml" >> $f7
  echo "loadParam $AMBERHOME/dat/mtkpp/metals/metalParm.xml" >> $f7
  echo "" >> $f7
  echo "# " >> $f7
  echo "writeFrcmodFile metals.frcmod metals" >> $f7
  echo "" >> $f7
  echo "loadParam $1_params_fc_md.xml" >> $f7
  echo "writeFrcmodFile $1_md.frcmod $1" >> $f7
  echo "" >> $f7
  echo "# " >> $f7
  echo "loadLib $1_chg2.xml" >> $f7
  echo "writePrepFile $1_chg2.prep $1" >> $f7
  echo "" >> $f7
  echo "quit" >> $f7


  # 
  f8=$1_resNames.bcl
  rm -rf $f8
  echo "source $1_settings.bcl" >> $f8
  echo "" >> $f8
  echo "# Load AMBER Parameters" >> $f8
  echo "loadParam NAME_params_fc_md.xml" >> $f8
  echo "" >> $f8
  echo "# Load AMBER libraries" >> $f8
  echo "# loadLib NAME_chg2.xml" >> $f8
  echo "loadLib NAME_stdMol.xml" >> $f8
  echo "" >> $f8
  echo "# Read pdb " >> $f8
  echo "readPdb NAME NAME_fixed_H.pdb" >> $f8
  echo "" >> $f8
  echo "# Assign S-S bonds" >> $f8
  echo "assignDisulfideBonds" >> $f8
  echo "" >> $f8
  echo "# Atom type" >> $f8
  echo "atomType" >> $f8
  echo "" >> $f8
  echo "# Assign bonds, angles, and torsions" >> $f8
  echo "assignConnectivity" >> $f8
  echo "" >> $f8
  echo "# Find metal centers" >> $f8
  echo "findMetalCenters" >> $f8
  echo "# print /" >> $f8
  echo "" >> $f8
  echo "# Write leaprc file" >> $f8
  echo "writeLeap NAME NAME_mcpb.pdb" >> $f8
  echo "" >> $f8
  echo "# Write pdb with new residue names" >> $f8
  echo "writePdb NAME_mcpb.pdb" >> $f8
  echo "" >> $f8
  echo "# Exit MCPB" >> $f8
  echo "quit" >> $f8


  return
}

#
# MAIN PROGRAM
#
bName=0

readCommandLine $@ # $@ are the arguments from the command line

genScripts $name

#
#
#
