#!/bin/csh 

###############################################################
#                                                             #
#      Charmm Lipid Builder to AMBER Lipid 10 Convertor       #
#                         v1.0                                #
#                          by                                 #
#                    Ross Walker (SDSC)                       #
#                         and                                 #
#                    Age Skjevik (UiB)                        #
#                                                             #
###############################################################

#This script is designed to convert pdb output from the Charmm
#Lipid builder to a pdb that is readable by AMBER's Leap using
#the Lipid10 force field.
#
#Charmm Lipid Builder: http://www.charmm-gui.org/?doc=input/membrane
#
#Revision History
#----------------
#Jun 17 2010 - Version 1.0
#


#Usage:
#./Convert_Charmm_Lipid_Builder_PDB_to_AMBER.x foo_input.pdb foo_output.pdb
#

if ($#argv != 2) then
   echo "Incorrect number of arguments."
   echo "usage:  ./Convert_Charmm_Lipid_Builder_PDB_to_AMBER.x foo_input.pdb foo_output.pdb"
   exit(1)
endif

set input_filename = $1
set output_filename = $2

if (! -e $input_filename) then
  echo "Input file: $input_filename does not exist."
  exit(1)
endif

echo "Charmm Lipid Builder to AMBER Lipid 10 Convertor"
echo "------------------------------------------------"
echo "                                                "
echo "                    v1.0                        "
echo "                     by                         "
echo "               Ross Walker (SDSC)               "
echo "                    and                         "
echo "               Age Skjevik (UiB)                "
echo "                                                "
echo "------------------------------------------------"
echo " "
echo "Input PDB file: $input_filename"
echo "Output PDB file: $output_filename"
echo " "

echo " *** STAGE 1 : STRIP REMARKS AND END STATEMENTS *** "
echo " "
grep -v "REMARK" $input_filename >lipid_tmp
grep -v "END" lipid_tmp >lipid_tmp2
mv lipid_tmp2 lipid_tmp


echo " *** STAGE 2 : STRIP AND SAVE WATER *** "
echo " "
echo " Currently supported water: TIP3"
echo " "

#We need to strip off the water and save it in AMBER format. We will
#re-append it later to the output file.

rm -f water_tmp
grep " TIP3" lipid_tmp >/dev/null
if ( $status == 0 ) then
  echo " Found TIP3 Water"
  grep " TIP3" lipid_tmp >water_tmp
  grep -v " TIP3"  lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
  set water_found = 1
else
  echo " Water was not found."
  set water_found = 0
endif
echo " "

if ( $water_found == 1 ) then
  sed 's/ TIP3/ WAT /' water_tmp > water_tmp2
  sed 's/OH2/O  /' water_tmp2 >water_tmp
  rm -f water_tmp2
endif

echo " *** STAGE 3 : STRIP AND SAVE IONS *** "
echo " "
echo " Currently supported ions: CLA, POT, SOD"
echo " "

#Process Ions
#First check for ions in the file.

rm -f ions_tmp
touch ions_tmp

grep "CLA CLA" lipid_tmp >/dev/null
if ( $status == 0 ) then
  echo " Found Chlorine Ions"
  grep "CLA CLA" lipid_tmp >>ions_tmp
  grep -v " CLA" lipid_tmp > lipid_tmp2
  mv lipid_tmp2 lipid_tmp
endif

grep "SOD SOD" lipid_tmp >/dev/null
if ( $status == 0 ) then
  echo " Found Sodium Ions"
  grep "SOD SOD" lipid_tmp >>ions_tmp
  grep -v " SOD" lipid_tmp > lipid_tmp2
  mv lipid_tmp2 lipid_tmp
endif

grep "POT POT" lipid_tmp >/dev/null
if ( $status == 0 ) then
  echo " Found Potassium Ions"
  grep "SOD SOD" lipid_tmp >>ions_tmp
  grep -v " POT" lipid_tmp2 > lipid_tmp2
  mv lipid_tmp2 lipid_tmp
endif


if (-z ions_tmp) then
  echo " No ions found."
  set ions_found = 0
  rm -f ions_tmp
else 
  echo " Ions were found."
  set ions_found = 1

  sed 's/CLA CLA/Cl- Cl-/' ions_tmp >ions_tmp2
  sed 's/SOD SOD/Na+ Na+/' ions_tmp2 >ions_tmp
  sed 's/POT POT/K+  K+ /' ions_tmp >ions_tmp2
  mv ions_tmp2 ions_tmp
endif
echo " "

echo " *** STAGE 4 : STRIP INDIVIDUAL LIPIDS TO SEPARATE FILES *** "
echo " "
echo " Currently supported lipids: DPPA DPPC DPPE DPPS DPPG DOPA DOPC DOPE DOPS DOPG"
echo "                             POPA POPC POPE POPS POPG SAPC SDPC SOPC DAPC"
echo " "

set lipids_found = 0

#DPPA
grep " DPPA" lipid_tmp >/dev/null
if ( $status == 0 ) then
  echo " Found DPPA"
  set dppa_found = 1
  set lipids_found = 1
  grep " DPPA" lipid_tmp >DPPA_tmp
  grep -v " DPPA" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dppa_found = 0
endif

#DPPC
grep " DPPC" lipid_tmp >/dev/null
if ( $status == 0 ) then
  echo " Found DPPC"
  set dppc_found = 1
  set lipids_found = 1
  grep " DPPC" lipid_tmp >DPPC_tmp
  grep -v " DPPC" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dppc_found = 0
endif

#DPPE
grep " DPPE" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set dppe_found = 1
  set lipids_found = 1
  grep " DPPE" lipid_tmp >DPPS_tmp
  grep -v " DPPE" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dppe_found = 0
endif

#DPPS
grep " DPPS" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set dpps_found = 1
  set lipids_found = 1
  grep " DPPS" lipid_tmp >DPPS_tmp
  grep -v " DPPS" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dpps_found = 0
endif

#DPPG
grep " DPPG" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set dppg_found = 1
  set lipids_found = 1
  grep " DPPG" lipid_tmp >DPPG_tmp
  grep -v " DPPG" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dppg_found = 0
endif

#DOPA
grep " DOPA" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set dopa_found = 1
  set lipids_found = 1
  grep " DOPA" lipid_tmp >DOPA_tmp
  grep -v " DOPA" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dopa_found = 0
endif

#DOPC
grep " DOPC" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set dopc_found = 1
  set lipids_found = 1
  grep " DOPC" lipid_tmp >DOPC_tmp
  grep -v " DOPC" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dopc_found = 0
endif

#DOPE
grep " DOPE" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set dope_found = 1
  set lipids_found = 1
  grep " DOPE" lipid_tmp >DOPE_tmp
  grep -v " DOPE" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dope_found = 0
endif

#DOPS
grep " DOPS" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set dops_found = 1
  set lipids_found = 1
  grep " DOPS" lipid_tmp >DOPS_tmp
  grep -v " DOPS" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dops_found = 0
endif

#DOPG
grep " DOPG" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set dopg_found = 1
  set lipids_found = 1
  grep " DOPG" lipid_tmp >DOPG_tmp
  grep -v " DOPG" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dopg_found = 0
endif

#POPA
grep " POPA" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set popa_found = 1
  set lipids_found = 1
  grep " POPA" lipid_tmp >POPA_tmp
  grep -v " POPA" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set popa_found = 0
endif

#POPC
grep " POPC" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set popc_found = 1
  set lipids_found = 1
  grep " POPC" lipid_tmp >POPC_tmp
  grep -v " POPC" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set popc_found = 0
endif

#POPE
grep " POPE" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set pope_found = 1
  set lipids_found = 1
  grep " POPE" lipid_tmp >POPE_tmp
  grep -v " POPE" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set pope_found = 0
endif

#POPS
grep " POPS" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set pops_found = 1
  set lipids_found = 1
  echo " Found POPS"
  grep " POPS" lipid_tmp >POPS_tmp
  grep -v " POPS" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set pops_found = 0
endif

#POPG
grep " POPG" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set popg_found = 1
  set lipids_found = 1
  grep " POPG" lipid_tmp >POPG_tmp
  grep -v " POPG" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set popg_found = 0
endif

#SAPC
grep " SAPC" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set sapc_found = 1
  set lipids_found = 1
  grep " SAPC" lipid_tmp >SAPC_tmp
  grep -v " SAPC" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set sapc_found = 0
endif

#SDPC
grep " SDPC" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set sdpc_found = 1
  set lipids_found = 1
  grep " SDPC" lipid_tmp >SDPC_tmp
  grep -v " SDPC" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set sdpc_found = 0
endif

#SOPC
grep " SOPC" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set sopc_found = 1
  set lipids_found = 1
  grep " SOPC" lipid_tmp >SOPC_tmp
  grep -v " SOPC" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set sopc_found = 0
endif

#DAPC
grep " DAPC" lipid_tmp >/dev/null
if ( $status == 0 ) then
  set dapc_found = 1
  set lipids_found = 1
  grep " DAPC" lipid_tmp >DAPC_tmp
  grep -v " DAPC" lipid_tmp >lipid_tmp2
  mv lipid_tmp2 lipid_tmp
else
  set dapc_found = 0
endif

echo " "

#Lipid tmp file should now just contain protein.
if ( -z lipid_tmp ) then
  echo " No protein atoms detected in input pdb file."
  rm -f lipid_tmp
  set proteins_found = 0
else
  echo " Protein atoms were detected in the input pdb file."
  echo -n " Line count for protein section of pdb = "
  wc -l lipid_tmp | awk '{print $1}'
  mv lipid_tmp protein_tmp
  set proteins_found = 1
endif
echo " "

if ( $lipids_found == 0 ) then
  echo " ERROR - No supported lipids found in pdb file."
  exit (1)
endif

echo " *** STAGE 5 : CONVERT LIPID FILES TO AMBER LIPID 10 FORMAT *** "
echo " "
echo " Currently supported lipids: POPC POPE POPS DPPC"
echo " "

#The next stage is to take the individual residules making up the various lipid files and rearrange
#things / modify the lipid atom name to give what AMBER lipid10 expects.

touch ../lipid_tmp_final

#POPC
if ( $popc_found == 1 ) then
  #POPC is 134 atoms per lipid. 
  set atoms_per_lipid = 134

  echo " Processing: POPC"

  #Check how many we have and see if it ads up correctly
  set line_count=`wc -l POPC_tmp | awk '{print$1}'`
  echo "  POPC Line count = $line_count"
  @ lipid_remainder = $line_count % $atoms_per_lipid
  if ( $lipid_remainder != 0 ) then
    echo " ERROR: Extracted POPC Lipids do NOT give a multiple of $atoms_per_lipid atoms."
    echo "                       Line Count = $line_count"
    echo "        Division by $atoms_per_lipid remainder = $lipid_remainder"
  endif
  @ lipid_count = $line_count / $atoms_per_lipid
  echo " POPC Lipid count = $lipid_count" 

  mkdir split_tmp
  cd split_tmp
  split -l $atoms_per_lipid ../POPC_tmp
  rm -f ../POPC_tmp 

  #Step 1 - rearrange atoms to be in the correct order for the Lipid10 force field.
  #SN1 tail goes first.
  set current_processing = 0
  echo -n " Processing POPC Lipid: "
  foreach i ( * )
    @ current_processing++
    echo -n " $current_processing "

    #Extract the head and two tails.
    grep -A2 "C32 POPC" $i > tail1
    grep -A42 "C33 POPC" $i >> tail1

    grep -A2 "C22 POPC" $i > tail2
    grep -A46 "C23 POPC" $i >> tail2

    head -n32 $i > head
    grep -A5 "C3  POPC" $i >> head

    #Modify residue and atom names
    #Tail 1
    sed 's/ C32 POPC/ C12 PA  /' tail1 >tail1_tmp
    sed 's/ C33 POPC/ C13 PA  /' tail1_tmp >tail1
    sed 's/ C34 POPC/ C14 PA  /' tail1 >tail1_tmp
    sed 's/ C35 POPC/ C15 PA  /' tail1_tmp >tail1
    sed 's/ C36 POPC/ C16 PA  /' tail1 >tail1_tmp
    sed 's/ C37 POPC/ C17 PA  /' tail1_tmp >tail1
    sed 's/ C38 POPC/ C18 PA  /' tail1 >tail1_tmp
    sed 's/ C39 POPC/ C19 PA  /' tail1_tmp >tail1
    sed 's/C310 POPC/C110 PA  /' tail1 >tail1_tmp
    sed 's/C311 POPC/C111 PA  /' tail1_tmp >tail1
    sed 's/C312 POPC/C112 PA  /' tail1 >tail1_tmp
    sed 's/C313 POPC/C113 PA  /' tail1_tmp >tail1
    sed 's/C314 POPC/C114 PA  /' tail1 >tail1_tmp
    sed 's/C315 POPC/C115 PA  /' tail1_tmp >tail1
    sed 's/C316 POPC/C116 PA  /' tail1 >tail1_tmp
    sed 's/C317 POPC/C117 PA  /' tail1_tmp >tail1
    sed 's/C318 POPC/C118 PA  /' tail1 >tail1_tmp
    sed 's/ H2X POPC/ H2R PA  /' tail1_tmp >tail1
    sed 's/ H2Y POPC/ H2S PA  /' tail1 >tail1_tmp
    sed 's/ H3X POPC/ H3R PA  /' tail1_tmp >tail1
    sed 's/ H3Y POPC/ H3S PA  /' tail1 >tail1_tmp
    sed 's/ H4X POPC/ H4R PA  /' tail1_tmp >tail1
    sed 's/ H4Y POPC/ H4S PA  /' tail1 >tail1_tmp
    sed 's/ H5X POPC/ H5R PA  /' tail1_tmp >tail1
    sed 's/ H5Y POPC/ H5S PA  /' tail1 >tail1_tmp
    sed 's/ H6X POPC/ H6R PA  /' tail1_tmp >tail1
    sed 's/ H6Y POPC/ H6S PA  /' tail1 >tail1_tmp
    sed 's/ H7X POPC/ H7R PA  /' tail1_tmp >tail1
    sed 's/ H7Y POPC/ H7S PA  /' tail1 >tail1_tmp
    sed 's/ H8X POPC/ H8R PA  /' tail1_tmp >tail1
    sed 's/ H8Y POPC/ H8S PA  /' tail1 >tail1_tmp
    sed 's/ H9X POPC/ H9R PA  /' tail1_tmp >tail1
    sed 's/ H9Y POPC/ H9S PA  /' tail1 >tail1_tmp
    sed 's/H10X POPC/H10R PA  /' tail1_tmp >tail1
    sed 's/H10Y POPC/H10S PA  /' tail1 >tail1_tmp
    sed 's/H11X POPC/H11R PA  /' tail1_tmp >tail1
    sed 's/H11Y POPC/H11S PA  /' tail1 >tail1_tmp
    sed 's/H12X POPC/H12R PA  /' tail1_tmp >tail1
    sed 's/H12Y POPC/H12S PA  /' tail1 >tail1_tmp
    sed 's/H13X POPC/H13R PA  /' tail1_tmp >tail1
    sed 's/H13Y POPC/H13S PA  /' tail1 >tail1_tmp
    sed 's/H14X POPC/H14R PA  /' tail1_tmp >tail1
    sed 's/H14Y POPC/H14S PA  /' tail1 >tail1_tmp
    sed 's/H15X POPC/H15R PA  /' tail1_tmp >tail1
    sed 's/H15Y POPC/H15S PA  /' tail1 >tail1_tmp
    sed 's/H16X POPC/H16R PA  /' tail1_tmp >tail1
    sed 's/H16Y POPC/H16S PA  /' tail1 >tail1_tmp
    sed 's/H16Z POPC/H16T PA  /' tail1_tmp >tail1
    rm -f tail1_tmp

    #Tail 2
    sed 's/ C22 POPC/ C12 OL  /' tail2 >tail2_tmp
    sed 's/ C23 POPC/ C13 OL  /' tail2_tmp >tail2
    sed 's/ C24 POPC/ C14 OL  /' tail2 >tail2_tmp
    sed 's/ C25 POPC/ C15 OL  /' tail2_tmp >tail2
    sed 's/ C26 POPC/ C16 OL  /' tail2 >tail2_tmp
    sed 's/ C27 POPC/ C17 OL  /' tail2_tmp >tail2
    sed 's/ C28 POPC/ C18 OL  /' tail2 >tail2_tmp
    sed 's/ C29 POPC/ C19 OL  /' tail2_tmp >tail2
    sed 's/C210 POPC/C110 OL  /' tail2 >tail2_tmp
    sed 's/C211 POPC/C111 OL  /' tail2_tmp >tail2
    sed 's/C212 POPC/C112 OL  /' tail2 >tail2_tmp
    sed 's/C213 POPC/C113 OL  /' tail2_tmp >tail2
    sed 's/C214 POPC/C114 OL  /' tail2 >tail2_tmp
    sed 's/C215 POPC/C115 OL  /' tail2_tmp >tail2
    sed 's/C216 POPC/C116 OL  /' tail2 >tail2_tmp
    sed 's/C217 POPC/C117 OL  /' tail2_tmp >tail2
    sed 's/C218 POPC/C118 OL  /' tail2 >tail2_tmp
    sed 's/ H2R POPC/ H2R OL  /' tail2_tmp >tail2
    sed 's/ H2S POPC/ H2S OL  /' tail2 >tail2_tmp
    sed 's/ H3R POPC/ H3R OL  /' tail2_tmp >tail2
    sed 's/ H3S POPC/ H3S OL  /' tail2 >tail2_tmp
    sed 's/ H4R POPC/ H4R OL  /' tail2_tmp >tail2
    sed 's/ H4S POPC/ H4S OL  /' tail2 >tail2_tmp
    sed 's/ H5R POPC/ H5R OL  /' tail2_tmp >tail2
    sed 's/ H5S POPC/ H5S OL  /' tail2 >tail2_tmp
    sed 's/ H6R POPC/ H6R OL  /' tail2_tmp >tail2
    sed 's/ H6S POPC/ H6S OL  /' tail2 >tail2_tmp
    sed 's/ H7R POPC/ H7R OL  /' tail2_tmp >tail2
    sed 's/ H7S POPC/ H7S OL  /' tail2 >tail2_tmp
    sed 's/ H8R POPC/ H8R OL  /' tail2_tmp >tail2
    sed 's/ H8S POPC/ H8S OL  /' tail2 >tail2_tmp
    sed 's/ H91 POPC/ H9R OL  /' tail2_tmp >tail2
    sed 's/H101 POPC/H10R OL  /' tail2 >tail2_tmp
    sed 's/H11R POPC/H11R OL  /' tail2_tmp >tail2
    sed 's/H11S POPC/H11S OL  /' tail2 >tail2_tmp
    sed 's/H12R POPC/H12R OL  /' tail2_tmp >tail2
    sed 's/H12S POPC/H12S OL  /' tail2 >tail2_tmp
    sed 's/H13R POPC/H13R OL  /' tail2_tmp >tail2
    sed 's/H13S POPC/H13S OL  /' tail2 >tail2_tmp
    sed 's/H14R POPC/H14R OL  /' tail2_tmp >tail2
    sed 's/H14S POPC/H14S OL  /' tail2 >tail2_tmp
    sed 's/H15R POPC/H15R OL  /' tail2_tmp >tail2
    sed 's/H15S POPC/H15S OL  /' tail2 >tail2_tmp
    sed 's/H16R POPC/H16R OL  /' tail2_tmp >tail2
    sed 's/H16S POPC/H16S OL  /' tail2 >tail2_tmp
    sed 's/H17R POPC/H17R OL  /' tail2_tmp >tail2
    sed 's/H17S POPC/H17S OL  /' tail2 >tail2_tmp
    sed 's/H18R POPC/H18R OL  /' tail2_tmp >tail2
    sed 's/H18S POPC/H18S OL  /' tail2 >tail2_tmp
    sed 's/H18T POPC/H18T OL  /' tail2_tmp >tail2
    rm -f tail2_tmp

    #Head
    sed 's/ N   POPC/ N31 PC  /' head > head_tmp
    sed 's/ C12 POPC/ C33 PC  /' head_tmp > head
    sed 's/ C13 POPC/ C34 PC  /' head > head_tmp
    sed 's/ C14 POPC/ C35 PC  /' head_tmp > head
    sed 's/ C11 POPC/ C32 PC  /' head > head_tmp
    sed 's/ C15 POPC/ C31 PC  /' head_tmp > head
    sed 's/ O1  POPC/ O32 PC  /' head > head_tmp
    sed 's/ P1  POPC/ P31 PC  /' head_tmp > head
    sed 's/ O2  POPC/ O31 PC  /' head > head_tmp
    sed 's/ O3  POPC/ O33 PC  /' head_tmp > head
    sed 's/ O4  POPC/ O34 PC  /' head > head_tmp
    sed 's/ C1  POPC/ C3  PC  /' head_tmp > head
    sed 's/ C2  POPC/ C2  PC  /' head > head_tmp
    sed 's/ C3  POPC/ C1  PC  /' head_tmp > head
    sed 's/ O31 POPC/ O11 PC  /' head > head_tmp
    sed 's/ C31 POPC/ C11 PC  /' head_tmp > head
    sed 's/ O32 POPC/ O12 PC  /' head > head_tmp
    sed 's/ O21 POPC/ O21 PC  /' head_tmp > head
    sed 's/ C21 POPC/ C21 PC  /' head > head_tmp
    sed 's/ O22 POPC/ O22 PC  /' head_tmp > head
    sed 's/ HX  POPC/ HR  PC  /' head > head_tmp
    sed 's/ HY  POPC/ HS  PC  /' head_tmp > head
    sed 's/ HS  POPC/ HX  PC  /' head > head_tmp
    sed 's/ HA  POPC/ HA  PC  /' head_tmp > head
    sed 's/ HB  POPC/ HB  PC  /' head > head_tmp
    sed 's/ H51 POPC/ H1A PC  /' head_tmp > head
    sed 's/ H52 POPC/ H1B PC  /' head > head_tmp
    sed 's/ H11 POPC/ H2A PC  /' head_tmp > head
    sed 's/ H12 POPC/ H2B PC  /' head > head_tmp
    sed 's/ H21 POPC/ H3A PC  /' head_tmp > head
    sed 's/ H22 POPC/ H3B PC  /' head > head_tmp
    sed 's/ H23 POPC/ H3C PC  /' head_tmp > head
    sed 's/ H31 POPC/ H4A PC  /' head > head_tmp
    sed 's/ H32 POPC/ H4B PC  /' head_tmp > head
    sed 's/ H33 POPC/ H4C PC  /' head > head_tmp
    sed 's/ H41 POPC/ H5A PC  /' head_tmp > head
    sed 's/ H42 POPC/ H5B PC  /' head > head_tmp
    sed 's/ H43 POPC/ H5C PC  /' head_tmp > head
    rm -f head_tmp

    #Append the modified lipid to the lipid_tmp_final file.
    cat tail1 >> ../lipid_tmp_final
    cat head >> ../lipid_tmp_final
    cat tail2 >> ../lipid_tmp_final
    echo "TER " >> ../lipid_tmp_final 

  end  

  cd ../
  rm -rf split_tmp
  echo " "
endif


#POPE
if ( $pope_found == 1 ) then
  #POPE is 125 atoms per lipid. 
  set atoms_per_lipid = 125

  echo " Processing: POPE"

  #Check how many we have and see if it ads up correctly
  set line_count=`wc -l POPE_tmp | awk '{print$1}'`
  echo "  POPE Line count = $line_count"
  @ lipid_remainder = $line_count % $atoms_per_lipid
  if ( $lipid_remainder != 0 ) then
    echo " ERROR: Extracted POPE Lipids do NOT give a multiple of $atoms_per_lipid atoms."
    echo "                       Line Count = $line_count"
    echo "        Division by $atoms_per_lipid remainder = $lipid_remainder"
  endif
  @ lipid_count = $line_count / $atoms_per_lipid
  echo " POPE Lipid count = $lipid_count" 

  mkdir split_tmp
  cd split_tmp
  split -l $atoms_per_lipid ../POPE_tmp
  rm -f ../POPE_tmp 

  #Step 1 - rearrange atoms to be in the correct order for the Lipid10 force field.
  #SN1 tail goes first.
  set current_processing = 0
  echo -n " Processing POPE Lipid: "
  foreach i ( * )
    @ current_processing++
    echo -n " $current_processing "

    #Extract the head and two tails.
    grep -A2 "C32 POPE" $i > tail1
    grep -A42 "C33 POPE" $i >> tail1

    grep -A2 "C22 POPE" $i > tail2
    grep -A46 "C23 POPE" $i >> tail2

    head -n23 $i > head
    grep -A5 "C3  POPE" $i >> head

    #Modify residue and atom names
    #Tail 1
    sed 's/ C32 POPE/ C12 PA  /' tail1 >tail1_tmp
    sed 's/ C33 POPE/ C13 PA  /' tail1_tmp >tail1
    sed 's/ C34 POPE/ C14 PA  /' tail1 >tail1_tmp
    sed 's/ C35 POPE/ C15 PA  /' tail1_tmp >tail1
    sed 's/ C36 POPE/ C16 PA  /' tail1 >tail1_tmp
    sed 's/ C37 POPE/ C17 PA  /' tail1_tmp >tail1
    sed 's/ C38 POPE/ C18 PA  /' tail1 >tail1_tmp
    sed 's/ C39 POPE/ C19 PA  /' tail1_tmp >tail1
    sed 's/C310 POPE/C110 PA  /' tail1 >tail1_tmp
    sed 's/C311 POPE/C111 PA  /' tail1_tmp >tail1
    sed 's/C312 POPE/C112 PA  /' tail1 >tail1_tmp
    sed 's/C313 POPE/C113 PA  /' tail1_tmp >tail1
    sed 's/C314 POPE/C114 PA  /' tail1 >tail1_tmp
    sed 's/C315 POPE/C115 PA  /' tail1_tmp >tail1
    sed 's/C316 POPE/C116 PA  /' tail1 >tail1_tmp
    sed 's/ H2X POPE/ H2R PA  /' tail1_tmp >tail1
    sed 's/ H2Y POPE/ H2S PA  /' tail1 >tail1_tmp
    sed 's/ H3X POPE/ H3R PA  /' tail1_tmp >tail1
    sed 's/ H3Y POPE/ H3S PA  /' tail1 >tail1_tmp
    sed 's/ H4X POPE/ H4R PA  /' tail1_tmp >tail1
    sed 's/ H4Y POPE/ H4S PA  /' tail1 >tail1_tmp
    sed 's/ H5X POPE/ H5R PA  /' tail1_tmp >tail1
    sed 's/ H5Y POPE/ H5S PA  /' tail1 >tail1_tmp
    sed 's/ H6X POPE/ H6R PA  /' tail1_tmp >tail1
    sed 's/ H6Y POPE/ H6S PA  /' tail1 >tail1_tmp
    sed 's/ H7X POPE/ H7R PA  /' tail1_tmp >tail1
    sed 's/ H7Y POPE/ H7S PA  /' tail1 >tail1_tmp
    sed 's/ H8X POPE/ H8R PA  /' tail1_tmp >tail1
    sed 's/ H8Y POPE/ H8S PA  /' tail1 >tail1_tmp
    sed 's/ H9X POPE/ H9R PA  /' tail1_tmp >tail1
    sed 's/ H9Y POPE/ H9S PA  /' tail1 >tail1_tmp
    sed 's/H10X POPE/H10R PA  /' tail1_tmp >tail1
    sed 's/H10Y POPE/H10S PA  /' tail1 >tail1_tmp
    sed 's/H11X POPE/H11R PA  /' tail1_tmp >tail1
    sed 's/H11Y POPE/H11S PA  /' tail1 >tail1_tmp
    sed 's/H12X POPE/H12R PA  /' tail1_tmp >tail1
    sed 's/H12Y POPE/H12S PA  /' tail1 >tail1_tmp
    sed 's/H13X POPE/H13R PA  /' tail1_tmp >tail1
    sed 's/H13Y POPE/H13S PA  /' tail1 >tail1_tmp
    sed 's/H14X POPE/H14R PA  /' tail1_tmp >tail1
    sed 's/H14Y POPE/H14S PA  /' tail1 >tail1_tmp
    sed 's/H15X POPE/H15R PA  /' tail1_tmp >tail1
    sed 's/H15Y POPE/H15S PA  /' tail1 >tail1_tmp
    sed 's/H16X POPE/H16R PA  /' tail1_tmp >tail1
    sed 's/H16Y POPE/H16S PA  /' tail1 >tail1_tmp
    sed 's/H16Z POPE/H16T PA  /' tail1_tmp >tail1
    rm -f tail1_tmp

    #Tail 2
    sed 's/ C22 POPE/ C12 OL  /' tail2 >tail2_tmp
    sed 's/ C23 POPE/ C13 OL  /' tail2_tmp >tail2
    sed 's/ C24 POPE/ C14 OL  /' tail2 >tail2_tmp
    sed 's/ C25 POPE/ C15 OL  /' tail2_tmp >tail2
    sed 's/ C26 POPE/ C16 OL  /' tail2 >tail2_tmp
    sed 's/ C27 POPE/ C17 OL  /' tail2_tmp >tail2
    sed 's/ C28 POPE/ C18 OL  /' tail2 >tail2_tmp
    sed 's/ C29 POPE/ C19 OL  /' tail2_tmp >tail2
    sed 's/C210 POPE/C110 OL  /' tail2 >tail2_tmp
    sed 's/C211 POPE/C111 OL  /' tail2_tmp >tail2
    sed 's/C212 POPE/C112 OL  /' tail2 >tail2_tmp
    sed 's/C213 POPE/C113 OL  /' tail2_tmp >tail2
    sed 's/C214 POPE/C114 OL  /' tail2 >tail2_tmp
    sed 's/C215 POPE/C115 OL  /' tail2_tmp >tail2
    sed 's/C216 POPE/C116 OL  /' tail2 >tail2_tmp
    sed 's/C217 POPE/C117 OL  /' tail2_tmp >tail2
    sed 's/C218 POPE/C118 OL  /' tail2 >tail2_tmp
    sed 's/ H2R POPE/ H2R OL  /' tail2_tmp >tail2
    sed 's/ H2S POPE/ H2S OL  /' tail2 >tail2_tmp
    sed 's/ H3R POPE/ H3R OL  /' tail2_tmp >tail2
    sed 's/ H3S POPE/ H3S OL  /' tail2 >tail2_tmp
    sed 's/ H4R POPE/ H4R OL  /' tail2_tmp >tail2
    sed 's/ H4S POPE/ H4S OL  /' tail2 >tail2_tmp
    sed 's/ H5R POPE/ H5R OL  /' tail2_tmp >tail2
    sed 's/ H5S POPE/ H5S OL  /' tail2 >tail2_tmp
    sed 's/ H6R POPE/ H6R OL  /' tail2_tmp >tail2
    sed 's/ H6S POPE/ H6S OL  /' tail2 >tail2_tmp
    sed 's/ H7R POPE/ H7R OL  /' tail2_tmp >tail2
    sed 's/ H7S POPE/ H7S OL  /' tail2 >tail2_tmp
    sed 's/ H8R POPE/ H8R OL  /' tail2_tmp >tail2
    sed 's/ H8S POPE/ H8S OL  /' tail2 >tail2_tmp
    sed 's/ H91 POPE/ H9R OL  /' tail2_tmp >tail2
    sed 's/H101 POPE/H10R OL  /' tail2 >tail2_tmp
    sed 's/H11R POPE/H11R OL  /' tail2_tmp >tail2
    sed 's/H11S POPE/H11S OL  /' tail2 >tail2_tmp
    sed 's/H12R POPE/H12R OL  /' tail2_tmp >tail2
    sed 's/H12S POPE/H12S OL  /' tail2 >tail2_tmp
    sed 's/H13R POPE/H13R OL  /' tail2_tmp >tail2
    sed 's/H13S POPE/H13S OL  /' tail2 >tail2_tmp
    sed 's/H14R POPE/H14R OL  /' tail2_tmp >tail2
    sed 's/H14S POPE/H14S OL  /' tail2 >tail2_tmp
    sed 's/H15R POPE/H15R OL  /' tail2_tmp >tail2
    sed 's/H15S POPE/H15S OL  /' tail2 >tail2_tmp
    sed 's/H16R POPE/H16R OL  /' tail2_tmp >tail2
    sed 's/H16S POPE/H16S OL  /' tail2 >tail2_tmp
    sed 's/H17R POPE/H17R OL  /' tail2_tmp >tail2
    sed 's/H17S POPE/H17S OL  /' tail2 >tail2_tmp
    sed 's/H18R POPE/H18R OL  /' tail2_tmp >tail2
    sed 's/H18S POPE/H18S OL  /' tail2 >tail2_tmp
    sed 's/H18T POPE/H18T OL  /' tail2_tmp >tail2
    rm -f tail2_tmp

    #Head
    sed 's/ N   POPE/ N31 PE  /' head > head_tmp
    sed 's/ C12 POPE/ C32 PE  /' head_tmp > head
    sed 's/ C11 POPE/ C31 PE  /' head > head_tmp
    sed 's/ O12 POPE/ O32 PE  /' head_tmp > head
    sed 's/ P   POPE/ P31 PE  /' head > head_tmp
    sed 's/ O11 POPE/ O31 PE  /' head_tmp > head
    sed 's/ O13 POPE/ O33 PE  /' head > head_tmp
    sed 's/ O14 POPE/ O34 PE  /' head_tmp > head
    sed 's/ C1  POPE/ C3  PE  /' head > head_tmp
    sed 's/ C2  POPE/ C2  PE  /' head_tmp > head
    sed 's/ C3  POPE/ C1  PE  /' head > head_tmp
    sed 's/ O31 POPE/ O11 PE  /' head_tmp > head
    sed 's/ C31 POPE/ C11 PE  /' head > head_tmp
    sed 's/ O32 POPE/ O12 PE  /' head_tmp > head
    sed 's/ O21 POPE/ O21 PE  /' head > head_tmp
    sed 's/ C21 POPE/ C21 PE  /' head_tmp > head
    sed 's/ O22 POPE/ O22 PE  /' head > head_tmp
    sed 's/ HX  POPE/ HR  PE  /' head_tmp > head
    sed 's/ HY  POPE/ HS  PE  /' head > head_tmp
    sed 's/ HS  POPE/ HX  PE  /' head_tmp > head
    sed 's/ HA  POPE/ HA  PE  /' head > head_tmp
    sed 's/ HB  POPE/ HB  PE  /' head_tmp > head
    sed 's/H11A POPE/ H1A PE  /' head > head_tmp
    sed 's/H11B POPE/ H1B PE  /' head_tmp > head
    sed 's/H12A POPE/ H2A PE  /' head > head_tmp
    sed 's/H12B POPE/ H2B PE  /' head_tmp > head
    sed 's/ HN1 POPE/HN1A PE  /' head > head_tmp
    sed 's/ HN2 POPE/HN1B PE  /' head_tmp > head
    sed 's/ HN3 POPE/HN1C PE  /' head > head_tmp
    mv head_tmp head  
    rm -f head_tmp

    #Append the modified lipid to the lipid_tmp_final file.
    cat tail1 >> ../lipid_tmp_final
    cat head >> ../lipid_tmp_final
    cat tail2 >> ../lipid_tmp_final
    echo "TER " >> ../lipid_tmp_final 

  end  

  cd ../
  rm -rf split_tmp
  echo " "
endif


#POPS
if ( $pops_found == 1 ) then
  #POPS is 127 atoms per lipid. 
  set atoms_per_lipid = 127

  echo " Processing: POPS"

  #Check how many we have and see if it ads up correctly
  set line_count=`wc -l POPS_tmp | awk '{print$1}'`
  echo "  POPS Line count = $line_count"
  @ lipid_remainder = $line_count % $atoms_per_lipid
  if ( $lipid_remainder != 0 ) then
    echo " ERROR: Extracted POPS Lipids do NOT give a multiple of $atoms_per_lipid atoms."
    echo "                       Line Count = $line_count"
    echo "        Division by $atoms_per_lipid remainder = $lipid_remainder"
  endif
  @ lipid_count = $line_count / $atoms_per_lipid
  echo " POPS Lipid count = $lipid_count" 

  mkdir split_tmp
  cd split_tmp
  split -l $atoms_per_lipid ../POPS_tmp
  rm -f ../POPS_tmp 

  #Step 1 - rearrange atoms to be in the correct order for the Lipid10 force field.
  #SN1 tail goes first.
  set current_processing = 0
  echo -n " Processing POPS Lipid: "
  foreach i ( * )
    @ current_processing++
    echo -n " $current_processing "

    #Extract the head and two tails.
    grep -A2 "C32 POPS" $i > tail1
    grep -A42 "C33 POPS" $i >> tail1

    grep -A2 "C22 POPS" $i > tail2
    grep -A46 "C23 POPS" $i >> tail2

    head -n25 $i > head
    grep -A5 "C3  POPS" $i >> head

    #Modify residue and atom names
    #Tail 1
    sed 's/ C32 POPS/ C12 PA  /' tail1 >tail1_tmp
    sed 's/ C33 POPS/ C13 PA  /' tail1_tmp >tail1
    sed 's/ C34 POPS/ C14 PA  /' tail1 >tail1_tmp
    sed 's/ C35 POPS/ C15 PA  /' tail1_tmp >tail1
    sed 's/ C36 POPS/ C16 PA  /' tail1 >tail1_tmp
    sed 's/ C37 POPS/ C17 PA  /' tail1_tmp >tail1
    sed 's/ C38 POPS/ C18 PA  /' tail1 >tail1_tmp
    sed 's/ C39 POPS/ C19 PA  /' tail1_tmp >tail1
    sed 's/C310 POPS/C110 PA  /' tail1 >tail1_tmp
    sed 's/C311 POPS/C111 PA  /' tail1_tmp >tail1
    sed 's/C312 POPS/C112 PA  /' tail1 >tail1_tmp
    sed 's/C313 POPS/C113 PA  /' tail1_tmp >tail1
    sed 's/C314 POPS/C114 PA  /' tail1 >tail1_tmp
    sed 's/C315 POPS/C115 PA  /' tail1_tmp >tail1
    sed 's/C316 POPS/C116 PA  /' tail1 >tail1_tmp
    sed 's/C317 POPS/C117 PA  /' tail1_tmp >tail1
    sed 's/C318 POPS/C118 PA  /' tail1 >tail1_tmp
    sed 's/ H2X POPS/ H2R PA  /' tail1_tmp >tail1
    sed 's/ H2Y POPS/ H2S PA  /' tail1 >tail1_tmp
    sed 's/ H3X POPS/ H3R PA  /' tail1_tmp >tail1
    sed 's/ H3Y POPS/ H3S PA  /' tail1 >tail1_tmp
    sed 's/ H4X POPS/ H4R PA  /' tail1_tmp >tail1
    sed 's/ H4Y POPS/ H4S PA  /' tail1 >tail1_tmp
    sed 's/ H5X POPS/ H5R PA  /' tail1_tmp >tail1
    sed 's/ H5Y POPS/ H5S PA  /' tail1 >tail1_tmp
    sed 's/ H6X POPS/ H6R PA  /' tail1_tmp >tail1
    sed 's/ H6Y POPS/ H6S PA  /' tail1 >tail1_tmp
    sed 's/ H7X POPS/ H7R PA  /' tail1_tmp >tail1
    sed 's/ H7Y POPS/ H7S PA  /' tail1 >tail1_tmp
    sed 's/ H8X POPS/ H8R PA  /' tail1_tmp >tail1
    sed 's/ H8Y POPS/ H8S PA  /' tail1 >tail1_tmp
    sed 's/ H9X POPS/ H9R PA  /' tail1_tmp >tail1
    sed 's/ H9Y POPS/ H9S PA  /' tail1 >tail1_tmp
    sed 's/H10X POPS/H10R PA  /' tail1_tmp >tail1
    sed 's/H10Y POPS/H10S PA  /' tail1 >tail1_tmp
    sed 's/H11X POPS/H11R PA  /' tail1_tmp >tail1
    sed 's/H11Y POPS/H11S PA  /' tail1 >tail1_tmp
    sed 's/H12X POPS/H12R PA  /' tail1_tmp >tail1
    sed 's/H12Y POPS/H12S PA  /' tail1 >tail1_tmp
    sed 's/H13X POPS/H13R PA  /' tail1_tmp >tail1
    sed 's/H13Y POPS/H13S PA  /' tail1 >tail1_tmp
    sed 's/H14X POPS/H14R PA  /' tail1_tmp >tail1
    sed 's/H14Y POPS/H14S PA  /' tail1 >tail1_tmp
    sed 's/H15X POPS/H15R PA  /' tail1_tmp >tail1
    sed 's/H15Y POPS/H15S PA  /' tail1 >tail1_tmp
    sed 's/H16X POPS/H16R PA  /' tail1_tmp >tail1
    sed 's/H16Y POPS/H16S PA  /' tail1 >tail1_tmp
    sed 's/H16Z POPS/H16T PA  /' tail1_tmp >tail1
    rm -f tail1_tmp

    #Tail 2
    sed 's/ C22 POPS/ C12 OL  /' tail2 >tail2_tmp
    sed 's/ C23 POPS/ C13 OL  /' tail2_tmp >tail2
    sed 's/ C24 POPS/ C14 OL  /' tail2 >tail2_tmp
    sed 's/ C25 POPS/ C15 OL  /' tail2_tmp >tail2
    sed 's/ C26 POPS/ C16 OL  /' tail2 >tail2_tmp
    sed 's/ C27 POPS/ C17 OL  /' tail2_tmp >tail2
    sed 's/ C28 POPS/ C18 OL  /' tail2 >tail2_tmp
    sed 's/ C29 POPS/ C19 OL  /' tail2_tmp >tail2
    sed 's/C210 POPS/C110 OL  /' tail2 >tail2_tmp
    sed 's/C211 POPS/C111 OL  /' tail2_tmp >tail2
    sed 's/C212 POPS/C112 OL  /' tail2 >tail2_tmp
    sed 's/C213 POPS/C113 OL  /' tail2_tmp >tail2
    sed 's/C214 POPS/C114 OL  /' tail2 >tail2_tmp
    sed 's/C215 POPS/C115 OL  /' tail2_tmp >tail2
    sed 's/C216 POPS/C116 OL  /' tail2 >tail2_tmp
    sed 's/C217 POPS/C117 OL  /' tail2_tmp >tail2
    sed 's/C218 POPS/C118 OL  /' tail2 >tail2_tmp
    sed 's/ H2R POPS/ H2R OL  /' tail2_tmp >tail2
    sed 's/ H2S POPS/ H2S OL  /' tail2 >tail2_tmp
    sed 's/ H3R POPS/ H3R OL  /' tail2_tmp >tail2
    sed 's/ H3S POPS/ H3S OL  /' tail2 >tail2_tmp
    sed 's/ H4R POPS/ H4R OL  /' tail2_tmp >tail2
    sed 's/ H4S POPS/ H4S OL  /' tail2 >tail2_tmp
    sed 's/ H5R POPS/ H5R OL  /' tail2_tmp >tail2
    sed 's/ H5S POPS/ H5S OL  /' tail2 >tail2_tmp
    sed 's/ H6R POPS/ H6R OL  /' tail2_tmp >tail2
    sed 's/ H6S POPS/ H6S OL  /' tail2 >tail2_tmp
    sed 's/ H7R POPS/ H7R OL  /' tail2_tmp >tail2
    sed 's/ H7S POPS/ H7S OL  /' tail2 >tail2_tmp
    sed 's/ H8R POPS/ H8R OL  /' tail2_tmp >tail2
    sed 's/ H8S POPS/ H8S OL  /' tail2 >tail2_tmp
    sed 's/ H91 POPS/ H9R OL  /' tail2_tmp >tail2
    sed 's/H101 POPS/H10R OL  /' tail2 >tail2_tmp
    sed 's/H11R POPS/H11R OL  /' tail2_tmp >tail2
    sed 's/H11S POPS/H11S OL  /' tail2 >tail2_tmp
    sed 's/H12R POPS/H12R OL  /' tail2_tmp >tail2
    sed 's/H12S POPS/H12S OL  /' tail2 >tail2_tmp
    sed 's/H13R POPS/H13R OL  /' tail2_tmp >tail2
    sed 's/H13S POPS/H13S OL  /' tail2 >tail2_tmp
    sed 's/H14R POPS/H14R OL  /' tail2_tmp >tail2
    sed 's/H14S POPS/H14S OL  /' tail2 >tail2_tmp
    sed 's/H15R POPS/H15R OL  /' tail2_tmp >tail2
    sed 's/H15S POPS/H15S OL  /' tail2 >tail2_tmp
    sed 's/H16R POPS/H16R OL  /' tail2_tmp >tail2
    sed 's/H16S POPS/H16S OL  /' tail2 >tail2_tmp
    sed 's/H17R POPS/H17R OL  /' tail2_tmp >tail2
    sed 's/H17S POPS/H17S OL  /' tail2 >tail2_tmp
    sed 's/H18R POPS/H18R OL  /' tail2_tmp >tail2
    sed 's/H18S POPS/H18S OL  /' tail2 >tail2_tmp
    sed 's/H18T POPS/H18T OL  /' tail2_tmp >tail2
    rm -f tail2_tmp

    #Head
    sed 's/ N   POPS/ N31 PS  /' head > head_tmp
    sed 's/ C12 POPS/ C32 PS  /' head_tmp > head
    sed 's/ C13 POPS/ C33 PS  /' head > head_tmp
    sed 's/O13A POPS/ O35 PS  /' head_tmp > head
    sed 's/O13B POPS/ O36 PS  /' head > head_tmp
    sed 's/ O12 POPS/ O32 PS  /' head_tmp > head
    sed 's/ P   POPS/ P31 PS  /' head > head_tmp
    sed 's/ O11 POPS/ O31 PS  /' head_tmp > head
    sed 's/ O13 POPS/ O33 PS  /' head > head_tmp
    sed 's/ C11 POPS/ C31 PS  /' head_tmp > head
    sed 's/ O14 POPS/ O34 PS  /' head > head_tmp
    sed 's/ C1  POPS/ C3  PS  /' head_tmp > head
    sed 's/ C2  POPS/ C2  PS  /' head > head_tmp
    sed 's/ C3  POPS/ C1  PS  /' head_tmp > head
    sed 's/ O31 POPS/ O11 PS  /' head > head_tmp
    sed 's/ C31 POPS/ C11 PS  /' head_tmp > head
    sed 's/ O32 POPS/ O12 PS  /' head > head_tmp
    sed 's/ O21 POPS/ O21 PS  /' head_tmp > head
    sed 's/ C21 POPS/ C21 PS  /' head > head_tmp
    sed 's/ O22 POPS/ O22 PS  /' head_tmp > head
    sed 's/ HX  POPS/ HR  PS  /' head > head_tmp
    sed 's/ HY  POPS/ HS  PS  /' head_tmp > head
    sed 's/ HS  POPS/ HX  PS  /' head > head_tmp
    sed 's/ HA  POPS/ HA  PS  /' head_tmp > head
    sed 's/ HB  POPS/ HB  PS  /' head > head_tmp
    sed 's/H11A POPS/ H1A PS  /' head_tmp > head
    sed 's/H11B POPS/ H1B PS  /' head > head_tmp
    sed 's/H12A POPS/ H2A PS  /' head_tmp > head
    sed 's/ HN1 POPS/HN1A PS  /' head > head_tmp
    sed 's/ HN2 POPS/HN1B PS  /' head_tmp > head
    sed 's/ HN3 POPS/HN1C PS  /' head > head_tmp
    mv head_tmp head

    #Append the modified lipid to the lipid_tmp_final file.
    cat tail1 >> ../lipid_tmp_final
    cat head >> ../lipid_tmp_final
    cat tail2 >> ../lipid_tmp_final
    echo "TER " >> ../lipid_tmp_final 

  end  

  cd ../
  rm -rf split_tmp
  echo " "
endif


#DPPC
if ( $dppc_found == 1 ) then
  #DPPC is 130 atoms per lipid. 
  set atoms_per_lipid = 130

  echo " Processing: DPPC"

  #Check how many we have and see if it ads up correctly
  set line_count=`wc -l DPPC_tmp | awk '{print$1}'`
  echo "  DPPC Line count = $line_count"
  @ lipid_remainder = $line_count % $atoms_per_lipid
  if ( $lipid_remainder != 0 ) then
    echo " ERROR: Extracted DPPC Lipids do NOT give a multiple of $atoms_per_lipid atoms."
    echo "                       Line Count = $line_count"
    echo "        Division by $atoms_per_lipid remainder = $lipid_remainder"
  endif
  @ lipid_count = $line_count / $atoms_per_lipid
  echo " DPPC Lipid count = $lipid_count" 

  mkdir split_tmp
  cd split_tmp
  split -l $atoms_per_lipid ../DPPC_tmp
  rm -f ../DPPC_tmp 

  #Step 1 - rearrange atoms to be in the correct order for the Lipid10 force field.
  #SN1 tail goes first.
  set current_processing = 0
  echo -n " Processing DPPC Lipid: "
  foreach i ( * )
    @ current_processing++
    echo -n " $current_processing "

    #Extract the head and two tails.
    grep -A2 "C32 DPPC" $i > tail1
    grep -A42 "C33 DPPC" $i >> tail1

    grep -A2 "C22 DPPC" $i > tail2
    grep -A42 "C23 DPPC" $i >> tail2

    head -n32 $i > head
    grep -A5 "C3  DPPC" $i >> head

    #Modify residue and atom names
    #Tail 1
    sed 's/ C32 DPPC/ C12 PA  /' tail1 >tail1_tmp
    sed 's/ C33 DPPC/ C13 PA  /' tail1_tmp >tail1
    sed 's/ C34 DPPC/ C14 PA  /' tail1 >tail1_tmp
    sed 's/ C35 DPPC/ C15 PA  /' tail1_tmp >tail1
    sed 's/ C36 DPPC/ C16 PA  /' tail1 >tail1_tmp
    sed 's/ C37 DPPC/ C17 PA  /' tail1_tmp >tail1
    sed 's/ C38 DPPC/ C18 PA  /' tail1 >tail1_tmp
    sed 's/ C39 DPPC/ C19 PA  /' tail1_tmp >tail1
    sed 's/C310 DPPC/C110 PA  /' tail1 >tail1_tmp
    sed 's/C311 DPPC/C111 PA  /' tail1_tmp >tail1
    sed 's/C312 DPPC/C112 PA  /' tail1 >tail1_tmp
    sed 's/C313 DPPC/C113 PA  /' tail1_tmp >tail1
    sed 's/C314 DPPC/C114 PA  /' tail1 >tail1_tmp
    sed 's/C315 DPPC/C115 PA  /' tail1_tmp >tail1
    sed 's/C316 DPPC/C116 PA  /' tail1 >tail1_tmp
    sed 's/ H2X DPPC/ H2R PA  /' tail1_tmp >tail1
    sed 's/ H2Y DPPC/ H2S PA  /' tail1 >tail1_tmp
    sed 's/ H3X DPPC/ H3R PA  /' tail1_tmp >tail1
    sed 's/ H3Y DPPC/ H3S PA  /' tail1 >tail1_tmp
    sed 's/ H4X DPPC/ H4R PA  /' tail1_tmp >tail1
    sed 's/ H4Y DPPC/ H4S PA  /' tail1 >tail1_tmp
    sed 's/ H5X DPPC/ H5R PA  /' tail1_tmp >tail1
    sed 's/ H5Y DPPC/ H5S PA  /' tail1 >tail1_tmp
    sed 's/ H6X DPPC/ H6R PA  /' tail1_tmp >tail1
    sed 's/ H6Y DPPC/ H6S PA  /' tail1 >tail1_tmp
    sed 's/ H7X DPPC/ H7R PA  /' tail1_tmp >tail1
    sed 's/ H7Y DPPC/ H7S PA  /' tail1 >tail1_tmp
    sed 's/ H8X DPPC/ H8R PA  /' tail1_tmp >tail1
    sed 's/ H8Y DPPC/ H8S PA  /' tail1 >tail1_tmp
    sed 's/ H9X DPPC/ H9R PA  /' tail1_tmp >tail1
    sed 's/ H9Y DPPC/ H9S PA  /' tail1 >tail1_tmp
    sed 's/H10X DPPC/H10R PA  /' tail1_tmp >tail1
    sed 's/H10Y DPPC/H10S PA  /' tail1 >tail1_tmp
    sed 's/H11X DPPC/H11R PA  /' tail1_tmp >tail1
    sed 's/H11Y DPPC/H11S PA  /' tail1 >tail1_tmp
    sed 's/H12X DPPC/H12R PA  /' tail1_tmp >tail1
    sed 's/H12Y DPPC/H12S PA  /' tail1 >tail1_tmp
    sed 's/H13X DPPC/H13R PA  /' tail1_tmp >tail1
    sed 's/H13Y DPPC/H13S PA  /' tail1 >tail1_tmp
    sed 's/H14X DPPC/H14R PA  /' tail1_tmp >tail1
    sed 's/H14Y DPPC/H14S PA  /' tail1 >tail1_tmp
    sed 's/H15X DPPC/H15R PA  /' tail1_tmp >tail1
    sed 's/H15Y DPPC/H15S PA  /' tail1 >tail1_tmp
    sed 's/H16X DPPC/H16R PA  /' tail1_tmp >tail1
    sed 's/H16Y DPPC/H16S PA  /' tail1 >tail1_tmp
    sed 's/H16Z DPPC/H16T PA  /' tail1_tmp >tail1
    rm -f tail1_tmp

    #Tail 2
    sed 's/ C22 DPPC/ C12 PA  /' tail2 >tail2_tmp
    sed 's/ C23 DPPC/ C13 PA  /' tail2_tmp >tail2
    sed 's/ C24 DPPC/ C14 PA  /' tail2 >tail2_tmp
    sed 's/ C25 DPPC/ C15 PA  /' tail2_tmp >tail2
    sed 's/ C26 DPPC/ C16 PA  /' tail2 >tail2_tmp
    sed 's/ C27 DPPC/ C17 PA  /' tail2_tmp >tail2
    sed 's/ C28 DPPC/ C18 PA  /' tail2 >tail2_tmp
    sed 's/ C29 DPPC/ C19 PA  /' tail2_tmp >tail2
    sed 's/C210 DPPC/C110 PA  /' tail2 >tail2_tmp
    sed 's/C211 DPPC/C111 PA  /' tail2_tmp >tail2
    sed 's/C212 DPPC/C112 PA  /' tail2 >tail2_tmp
    sed 's/C213 DPPC/C113 PA  /' tail2_tmp >tail2
    sed 's/C214 DPPC/C114 PA  /' tail2 >tail2_tmp
    sed 's/C215 DPPC/C115 PA  /' tail2_tmp >tail2
    sed 's/C216 DPPC/C116 PA  /' tail2 >tail2_tmp
    sed 's/ H2R DPPC/ H2R PA  /' tail2_tmp >tail2
    sed 's/ H2S DPPC/ H2S PA  /' tail2 >tail2_tmp
    sed 's/ H3R DPPC/ H3R PA  /' tail2_tmp >tail2
    sed 's/ H3S DPPC/ H3S PA  /' tail2 >tail2_tmp
    sed 's/ H4R DPPC/ H4R PA  /' tail2_tmp >tail2
    sed 's/ H4S DPPC/ H4S PA  /' tail2 >tail2_tmp
    sed 's/ H5R DPPC/ H5R PA  /' tail2_tmp >tail2
    sed 's/ H5S DPPC/ H5S PA  /' tail2 >tail2_tmp
    sed 's/ H6R DPPC/ H6R PA  /' tail2_tmp >tail2
    sed 's/ H6S DPPC/ H6S PA  /' tail2 >tail2_tmp
    sed 's/ H7R DPPC/ H7R PA  /' tail2_tmp >tail2
    sed 's/ H7S DPPC/ H7S PA  /' tail2 >tail2_tmp
    sed 's/ H8R DPPC/ H8R PA  /' tail2_tmp >tail2
    sed 's/ H8S DPPC/ H8S PA  /' tail2 >tail2_tmp
    sed 's/ H9R DPPC/ H9R PA  /' tail2_tmp >tail2
    sed 's/ H9S DPPC/ H9S PA  /' tail2 >tail2_tmp
    sed 's/H10R DPPC/H10R PA  /' tail2_tmp >tail2
    sed 's/H10S DPPC/H10S PA  /' tail2 >tail2_tmp
    sed 's/H11R DPPC/H11R PA  /' tail2_tmp >tail2
    sed 's/H11S DPPC/H11S PA  /' tail2 >tail2_tmp
    sed 's/H12R DPPC/H12R PA  /' tail2_tmp >tail2
    sed 's/H12S DPPC/H12S PA  /' tail2 >tail2_tmp
    sed 's/H13R DPPC/H13R PA  /' tail2_tmp >tail2
    sed 's/H13S DPPC/H13S PA  /' tail2 >tail2_tmp
    sed 's/H14R DPPC/H14R PA  /' tail2_tmp >tail2
    sed 's/H14S DPPC/H14S PA  /' tail2 >tail2_tmp
    sed 's/H15R DPPC/H15R PA  /' tail2_tmp >tail2
    sed 's/H15S DPPC/H15S PA  /' tail2 >tail2_tmp
    sed 's/H16R DPPC/H16R PA  /' tail2_tmp >tail2
    sed 's/H16S DPPC/H16S PA  /' tail2 >tail2_tmp
    sed 's/H16T DPPC/H16T PA  /' tail2_tmp >tail2
    rm -f tail2_tmp

    #Head
    sed 's/ N   DPPC/ N31 PC  /' head > head_tmp
    sed 's/ C13 DPPC/ C33 PC  /' head_tmp > head
    sed 's/ C14 DPPC/ C34 PC  /' head > head_tmp
    sed 's/ C15 DPPC/ C35 PC  /' head_tmp > head
    sed 's/ C12 DPPC/ C32 PC  /' head > head_tmp
    sed 's/ C11 DPPC/ C31 PC  /' head_tmp > head
    sed 's/ O12 DPPC/ O32 PC  /' head > head_tmp
    sed 's/ P   DPPC/ P31 PC  /' head_tmp > head
    sed 's/ O11 DPPC/ O31 PC  /' head > head_tmp
    sed 's/ O13 DPPC/ O33 PC  /' head_tmp > head
    sed 's/ O14 DPPC/ O34 PC  /' head > head_tmp
    sed 's/ C1  DPPC/ C3  PC  /' head_tmp > head
    sed 's/ C2  DPPC/ C2  PC  /' head > head_tmp
    sed 's/ C3  DPPC/ C1  PC  /' head_tmp > head
    sed 's/ O31 DPPC/ O11 PC  /' head > head_tmp
    sed 's/ C31 DPPC/ C11 PC  /' head_tmp > head
    sed 's/ O32 DPPC/ O12 PC  /' head > head_tmp
    sed 's/ O21 DPPC/ O21 PC  /' head_tmp > head
    sed 's/ C21 DPPC/ C21 PC  /' head > head_tmp
    sed 's/ O22 DPPC/ O22 PC  /' head_tmp > head
    sed 's/ HX  DPPC/ HR  PC  /' head > head_tmp
    sed 's/ HY  DPPC/ HS  PC  /' head_tmp > head
    sed 's/ HS  DPPC/ HX  PC  /' head > head_tmp
    sed 's/ HA  DPPC/ HA  PC  /' head_tmp > head
    sed 's/ HB  DPPC/ HB  PC  /' head > head_tmp
    sed 's/H11A DPPC/ H1A PC  /' head_tmp > head
    sed 's/H11B DPPC/ H1B PC  /' head > head_tmp
    sed 's/H12A DPPC/ H2A PC  /' head_tmp > head
    sed 's/H12B DPPC/ H2B PC  /' head > head_tmp
    sed 's/H13A DPPC/ H3A PC  /' head_tmp > head
    sed 's/H13B DPPC/ H3B PC  /' head > head_tmp
    sed 's/H13C DPPC/ H3C PC  /' head_tmp > head
    sed 's/H14A DPPC/ H4A PC  /' head > head_tmp
    sed 's/H14B DPPC/ H4B PC  /' head_tmp > head
    sed 's/H14C DPPC/ H4C PC  /' head > head_tmp
    sed 's/H15A DPPC/ H5A PC  /' head_tmp > head
    sed 's/H15B DPPC/ H5B PC  /' head > head_tmp
    sed 's/H15C DPPC/ H5C PC  /' head_tmp > head
    rm -f head_tmp

    #Append the modified lipid to the lipid_tmp_final file.
    cat tail1 >> ../lipid_tmp_final
    cat head >> ../lipid_tmp_final
    cat tail2 >> ../lipid_tmp_final
    echo "TER " >> ../lipid_tmp_final 

  end  

  cd ../
  rm -rf split_tmp
  echo " "
endif

#DPPE
if ( $dppe_found == 1 ) then
  echo " ERROR: DPPE not yet supported."
endif

echo " *** STAGE 6 : COMBINE PROTEIN, LIPID, ION AND WATER FILES INTO OUTPUT PDB ***"
echo " "

if ( $proteins_found == 1 ) then
  mv protein_tmp $output_filename
  echo "TER " >>$output_filename
else
  touch $output_filename
endif

cat lipid_tmp_final >> $output_filename
rm -f lipid_tmp_final

if ( $ions_found == 1 ) then
  cat ions_tmp >> $output_filename
  rm -f ions_tmp
  echo "TER " >> $output_filename
endif

if ( $water_found == 1 ) then
  cat water_tmp >> $output_filename
  rm -f water_tmp
  echo "TER " >> $output_filename
endif

echo "END " >> $output_filename

echo " ******* CONVERSION COMPLETE ******* "
echo " "

exit(0)

