#include <object.hpp>
#include "check.hpp"

namespace amber
{
    void check_charge( const molecule_t& m )
    {
        std::cout << "Checking charge: ";
        double chg = charge( m );
        double dif = std::abs( chg - int(chg) );
        if( dif > 0.01 && dif < 0.99 )
        {
            std::cout << std::endl;
            std::cout << "Warning: non-integral total charge: " << chg << std::endl;
        }
        else
        {
            std::cout << chg << ", OK!" << std::endl;
        }
    }

    void check_bond_length( const molecule_t& m, double lowcutoff, double hghcutoff )
    {
        std::cout << "Checking bond lengths: ";
        bool ok = true;
        bonditer_t bi = m.bond_begin();
        for( ; bi != m.bond_end(); ++bi )
        {
            atom_t a1 = atom_1st( *bi );
            atom_t a2 = atom_2nd( *bi );
            double ln = dist( a1, a2 );
            if( ln < lowcutoff || ln > hghcutoff )
            {
                resd_t r1 = a1.resd();
                resd_t r2 = a2.resd();
                ok = false;
                std::cout << std::endl;
                std::cout << format("Warning: There is a bond of %4.2f angstroms between: ") % ln;
                std::cout << r1.get_s(NAME) << "." << a1.get_s(NAME) << " and ";
                std::cout << r2.get_s(NAME) << "." << a2.get_s(NAME);
            }
        }

        if( ok )
        {
            std::cout << "OK!" << std::endl;
        }
        else
        {
            std::cout << std::endl;
        }
    }

    void check_atom_distance( const molecule_t& m, double cutoff )
    {
        std::cout << "Checking nonbonded atom contacts: ";
        bool ok = true;
        atomiter_t ai = m.atom_begin();
        for( ; ai != m.atom_end(); ++ai )
        {
            atomiter_t aj = ai + 1;
            for( ; aj != m.atom_end(); ++aj )
            {
                if( bond_t::has(*ai, *aj) )
                {
                    continue;
                }

                double d = dist( *ai, *aj );
                if( d < cutoff )
                {
                    ok = false;
                    resd_t r1 = ai->resd();
                    resd_t r2 = aj->resd();
                    std::cout << std::endl;
                    std::cout << format( "Warning: Close contact of %4.2f Angstroms between atoms ") % d;
                    std::cout << r1.get_s(NAME) << "." << ai->name() << " and ";
                    std::cout << r2.get_s(NAME) << "." << aj->name();
                }
            }
        }

        if( ok )
        {
            std::cout << "OK!" << std::endl;
        }
        else
        {
            std::cout << std::endl;
        }
    }

    void check_atom_params( const molecule_t& m, const molecule_t& ff )
    {
        std::cout << "Checking atom parameters: ";
        bool ok = true;
        atomiter_t ai = m.atom_begin();
        for( ; ai != m.atom_end(); ++ai )
        {
            string t = ai->type();
            if( !atom_t::has(ff, t) )
            {
                ok = false;
                std::cout << "Warning: could not find parameters for atom: " << t << std::endl;
            }
        }

        if( ok )
        {
            std::cout << "OK!" << std::endl;
        }
        else
        {
            std::cout << std::endl;
        }
    }

    void check_bond_params( const molecule_t& m, const molecule_t& ff )
    {
        std::cout << "Checking bond parameters: ";
        bool ok = true;
        bonditer_t bi = m.bond_begin();
        for( ; bi != m.bond_end(); ++bi )
        {
            atom_t a1 = atom_1st( *bi );
            atom_t a2 = atom_2nd( *bi );
            string t1 = a1.type();
            string t2 = a2.type();

            atom_t pa1(ff, -1), pa2(ff, -1);

            if( atom_t::get(ff, t1, pa1) && atom_t::get(ff, t2, pa2) )
            {
                if( bond_t::has(pa1, pa2) )
                {
                    continue;
                }
            }

            ok = false;
            std::cout << std::endl;
            std::cout << "Warning: could not find parameters for bond " << t1 << "-" << t2 << std::endl;
        }

        if( ok )
        {
            std::cout << "OK!" << std::endl;
        }
        else
        {
            std::cout << std::endl;
        }
    }

    void check_angl_params( const molecule_t& m, const molecule_t& ff )
    {
        std::cout << "Checking angle parameters: ";
        bool ok = true;
        atomiter_t ai = m.atom_begin();
        for( ; ai != m.atom_end(); ++ai )
        {
            atomiter_t aj = ai->atom_begin();
            for( ; aj != ai->atom_end(); ++aj )
            {
                atomiter_t ak = aj + 1;
                for( ; ak != ai->atom_end(); ++ak )
                {
                    string t1 = ai->type();
                    string t2 = aj->type();
                    string t3 = ak->type();
                    atom_t pa1(ff, -1);
                    atom_t pa2(ff, -1);
                    atom_t pa3(ff, -1);
                    if( atom_t::get(ff, t1, pa1) && atom_t::get(ff, t2, pa2) && atom_t::get(ff, t3, pa3) )
                    {
                        if( angl_t::has(pa2, pa1, pa3) )
                        {
                            continue;
                        }
                    }

                    ok = false;
                    std::cout << std::endl;
                    std::cout << "Warning: could not find parameters for angle " << t2 << "-" << t1 << "-" << t3;
                }
            }
        }

        if( ok )
        {
            std::cout << "OK!" << std::endl;
        }
        else
        {
            std::cout << std::endl;
        }
    }

    check_command::check_command()
        : command_i( "check" )
    {
    }

    check_command::check_command( bool checkmolcharge,  bool checkatomparams,
                                  bool checkbondparams, bool checkangleparams,
                                  bool checkbondlength, bool checkatomclashes,
                                  const string& molnam )
        : m_checkmolcharge(checkmolcharge), m_checkatomparams(checkatomparams),
        m_checkbondparams(checkbondparams), m_checkangleparams(checkangleparams),
        m_checkbondlength(checkbondlength), m_checkatomclashes(checkatomclashes),
        m_molnam( molnam )
    {
    }

    check_command::~check_command( )
    {
    }

    bool check_command::exec( )
    {
        int ndot = std::count( m_molnam.begin(), m_molnam.end(), '.' );
        if( ndot != 0 )
        {
            throw std::runtime_error( "Error: Arguments, other than flags, must be units!" );
        }
        
        std::cout << "Checking unit \"" << m_molnam << "\"" << std::endl;

        molecule_ptr pmol = content().get_mol( m_molnam );
        molecule_ptr parm = content().get_mol( "_amberffp" );

	if (m_checkmolcharge == true)
	{
	    check_charge( *pmol );
	}
	
        if (m_checkatomparams == true)
        {
            check_atom_params( *pmol, *parm );
        }
        
        if (m_checkbondparams == true)
        {
            check_bond_params( *pmol, *parm );
        }
        
        if (m_checkangleparams == true)
        {
            check_angl_params( *pmol, *parm );
        }
        
        if (m_checkbondlength == true)
        {
            check_bond_length( *pmol, 0.5, 3.0 );
        }
        
        if (m_checkatomclashes == true)
        {
            check_atom_distance( *pmol, 1.5 );
        }
        
        return true;
    }

    void check_command::undo()
    {
    }

    const char* check_command::info() const
    {
        return "check unit";
    }

    command_ptr check_command::clone( const vector< string >& args ) const
    {
        bool checkall = false;
        bool checkmolcharge = false;
        bool checkatomparams = false;
        bool checkbondparams = false;
        bool checkangleparams = false;
        bool checkbondlength = false;
        bool checkatomclashes = false;
        bool nocheckatomclashes = false;
        
        std::string unit = "";
        
        int i;
        
        //if( args.size()!=2 )
        //{
        //    throw std::runtime_error( "Error: Wrong number of arguments!" );
        //}
        
        for (i = 1; i < args.size(); i++)
        {
            if (args[i] == "-charge")
            {
                checkmolcharge = true;
                std::cout << "charge check requested" << std::endl;
            }
            else if (args[i] == "-atomparams")
            {
                checkatomparams = true;
                std::cout << "atom parameter check requested" << std::endl;
            }
            else if (args[i] == "-bondparams")
            {
                checkbondparams = true;
                std::cout << "bond parameter check requested" << std::endl;
            }
            else if (args[i] == "-angleparams")
            {
                checkangleparams = true;
                std::cout << "angle parameter check requested" << std::endl;
            }
            else if (args[i] == "-bondlength")
            {
                checkbondlength = true;
                std::cout << "bond length check requested" << std::endl;
            }
            else if (args[i] == "-clashes")
            {
                if (nocheckatomclashes == true)
                {
                    throw std::runtime_error("Error: can't use -clashes and -noclashes together!");
                }
                
                checkatomclashes = true;
                std::cout << "atom proximity (clash) check requested" << std::endl;
            }
            else if (args[i] == "-noclashes")
            {
                if (checkatomclashes == true)
                {
                    throw std::runtime_error("Error: can't use -clashes and -noclashes together!");
                }
                if (checkall == true)
                {
                    throw std::runtime_error("Error: can't use -all and -noclashes together!");
                }
                
                nocheckatomclashes = true;
                // Set everything else, too
                checkmolcharge = true;
                checkatomparams = true;
                checkbondparams = true;
                checkangleparams = true;
                checkbondlength = true;
                checkatomclashes = false;
                checkall = false;
                std::cout << "all except atom proximity (clash) check requested" << std::endl;
            }
            else if (args[i] == "-all")
            {
                if (nocheckatomclashes == true)
                {
                    throw std::runtime_error("Error: can't use -all and -noclashes together!");
                }
                
                checkall = true;
            }
            else if (args[i].substr(0,1) == "-")
            {
                throw std::runtime_error("Error: in check, unrecognised flag: " + args[i] + "\n"
                                       + "       Allowed flags:"    + "\n"
                                       + "            -charge"      + "\n"
                                       + "            -atomparams"  + "\n"
                                       + "            -bondparams"  + "\n"
                                       + "            -angleparams" + "\n"
                                       + "            -bondlength"  + "\n"
                                       + "            -clashes"     + "\n"
                                       + "            -noclashes"   + "\n"
                                       + "            -all"            );
            }
            else if (unit == "")
            {
                unit = args[i];
            }
            else
            {
                throw std::runtime_error("Error: Too many arguments!");
            }
        }
        
        if (unit == "")
        {
            throw std::runtime_error("Error: A unit must be supplied as an argument!");
        }
        
        // If everything is false except possibly checkall (whose truth
        // or falsity is irrelevant), we want to check everything
        if (checkmolcharge == false && checkatomparams == false
         && checkbondparams == false && checkangleparams == false
         && checkbondlength == false && checkatomclashes == false
         && nocheckatomclashes == false)
        {
            checkmolcharge = true;
            checkatomparams = true;
            checkbondparams = true;
            checkangleparams = true;
            checkbondlength = true;
            checkatomclashes = true;
            std::cout << "all checks requested" << std::endl;
        }

        return shared_ptr< command_i >( new check_command(checkmolcharge, checkatomparams,
                                                          checkbondparams, checkangleparams,
                                                          checkbondlength, checkatomclashes,
                                                          unit ) );
    }

} // namespace amber

amber::check_command g_check_command;

