#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "BSpline.h"

#include "pmeRecipDS.h"
#include "pmeDirectDS.h"

/***=======================================================================***/
/*** BSpln: this is the basic function for B-Spline computation.  It's not ***/
/***        at all optimal, but it is easy to code using the recursive     ***/
/***        definition of B-Splines.                                       ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   x:    the value to compute on this spline                           ***/
/***   ordr: the order of the spline (the spline will have nonzero values  ***/
/***         for 0 <= x <= ordr)                                           ***/
/***=======================================================================***/
double BSpln(double x, int ordr)
{
  double om1, iom1, y;

  om1 = ordr-1.0;
  iom1 = 1.0/om1;
  if (ordr > 2) {
    y = (x*iom1)*BSpln(x,om1) + ((ordr-x)*iom1)*BSpln(x-1.0, ordr-1);
  }
  else if (ordr == 2) {
    if (x > 0.0 && x < 2.0) {
      y = 1.0 - fabs(x-1.0);
    }
    else {
      y = 0.0;
    }
  }
  else {
    printf("BSpln >> Error.  Order = %d.\n", ordr);
    exit(1);
  }

  return y;
}

/***=======================================================================***/
/*** dBSpln: this is a basic function for B-Spline derivative computation. ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   x:    the value to compute on this spline                           ***/
/***   ordr: the order of the spline (the spline will have nonzero values  ***/
/***         for 0 <= x <= ordr)                                           ***/
/***=======================================================================***/
double dBSpln(double x, int ordr)
{
  double y;

  if (ordr > 2) {
    y = BSpln(x, ordr-1) - BSpln(x-1.0, ordr-1);
  }
  else {
    printf("dBSpln >> Error.  Order = %d.\n", ordr);
    exit(1);
  }

  return y;
}

/***=======================================================================***/
/*** CreateBmap: create a data structure to hold all information necessary ***/
/***             to perform particle<-->mesh operations.  The structure is ***/
/***             filled by SplCoeff (see below).                           ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   rcinp:  the reciprocal space control data                           ***/
/***   natom:  the number of atoms in the system                           ***/
/***=======================================================================***/
bmap CreateBmap(reccon *rcinp, int natom)
{
  int maxatom;
  bmap pmmap;

  pmmap.natom = natom;
  maxatom = natom+32;
  pmmap.maxatom = maxatom;
  pmmap.xcof = (bcof*)malloc(maxatom*rcinp->ordr[0]*sizeof(bcof));
  pmmap.ycof = (bcof*)malloc(maxatom*rcinp->ordr[1]*sizeof(bcof));
  pmmap.zcof = (bcof*)malloc(maxatom*rcinp->ordr[2]*sizeof(bcof));
  pmmap.atmid = (int*)malloc(maxatom*sizeof(int));

  return pmmap;
}

/***=======================================================================***/
/*** DestroyBmap: destroy a bmap data structure.                           ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   A:  the B-spline map to destroy                                     ***/
/***=======================================================================***/
void DestroyBmap(bmap *A)
{
  free(A->atmid);
  free(A->xcof);
  free(A->ycof);
  free(A->zcof);
}

/***=======================================================================***/
/*** FillQMap: fill the charge map for this charge in this dimension.      ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   SP:    the array of B-spline coefficients and charge mesh indices   ***/
/***            in this dimension (the dimension of interest)              ***/
/***   uc:    the result of ceil() acting on the scaled fractional         ***/
/***            coordinate of the particle in the dimension of interest    ***/
/***   ng:    the number of mesh points in the dimension of interest       ***/
/***   ordr:  the interpolation order in the dimension of interest         ***/
/***=======================================================================***/
void FillQMap(bcof* SP, int uc, int ng, int ordr)
{
  int i;
  double invng;

  /*** Charge map ***/
  invng = 1.0/ng;
  for (i = 0; i < ordr; i++) {
    SP[i].m = uc + i;
    SP[i].m = SP[i].m - ng*floor(SP[i].m*invng);
  }
}

/***=======================================================================***/
/*** Extend4: extend third order B-spline coefficients to fourth order and ***/
/***          compute derivatives.                                         ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   SP:    the array of B-spline coefficients and charge mesh indices   ***/
/***            in this dimension (the dimension of interest)              ***/
/***   w:     one minus the non-integer part of the scaled fractional      ***/
/***            coordinate                                                 ***/
/***   uc:    the result of ceil() acting on the scaled fractional         ***/
/***            coordinate of the particle in the dimension of interest    ***/
/***   ng:    the number of mesh points in the dimension of interest       ***/
/***=======================================================================***/
void Extend4(bcof* SP, double w, int uc, int ng)
{
  /*** Fourth order spline derivatives ***/
  SP[0].d = -SP[0].s;
  SP[1].d = SP[0].s - SP[1].s;
  SP[2].d = SP[1].s - SP[2].s;
  SP[3].d = SP[2].s;

  /*** Fourth order splines ***/
  SP[3].s = w*SP[2].s/3.0;
  SP[2].s = ((w+1.0)*SP[1].s + (3.0-w)*SP[2].s)/3.0;
  SP[0].s = (1.0-w)*SP[0].s/3.0;
  SP[1].s = 1.0-SP[0].s-SP[2].s-SP[3].s;

  /*** Charge map ***/
  FillQMap(SP, uc-4, ng, 4);
}

/***=======================================================================***/
/*** Extend6: extend third order B-spline coefficients to sixth order and  ***/
/***          compute derivatives.                                         ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   See Extend4(), above.                                               ***/
/***=======================================================================***/
void Extend6(bcof* SP, double w, int uc, int ng)
{
  /*** Fourth order splines ***/
  SP[3].s = w*SP[2].s/3.0;
  SP[2].s = ((w+1.0)*SP[1].s + (3.0-w)*SP[2].s)/3.0;
  SP[0].s = (1.0-w)*SP[0].s/3.0;
  SP[1].s = 1.0-SP[0].s-SP[2].s-SP[3].s;

  /*** Fifth order splines ***/
  SP[4].s = 0.25*w*SP[3].s;
  SP[3].s = 0.25*((w+1.0)*SP[2].s + (4.0-w)*SP[3].s);
  SP[2].s = 0.25*((w+2.0)*SP[1].s + (3.0-w)*SP[2].s);
  SP[1].s = 0.25*((w+3.0)*SP[0].s + (2.0-w)*SP[1].s);
  SP[0].s = 0.25*(1.0-w)*SP[0].s;

  /*** Sixth order spline derivatives ***/
  SP[0].d = -SP[0].s;
  SP[1].d = SP[0].s - SP[1].s;
  SP[2].d = SP[1].s - SP[2].s;
  SP[3].d = SP[2].s - SP[3].s;
  SP[4].d = SP[3].s - SP[4].s;
  SP[5].d = SP[4].s;

  /*** Sixth order splines ***/
  SP[5].s = 0.2*w*SP[4].s;
  SP[4].s = 0.2*((w+1.0)*SP[3].s + (5.0-w)*SP[4].s);
  SP[3].s = 0.2*((w+2.0)*SP[2].s + (4.0-w)*SP[3].s);
  SP[2].s = 0.2*((w+3.0)*SP[1].s + (3.0-w)*SP[2].s);
  SP[1].s = 0.2*((w+4.0)*SP[0].s + (2.0-w)*SP[1].s);
  SP[0].s = 0.2*(1.0-w)*SP[0].s;

  /*** Charge map ***/
  FillQMap(SP, uc-6, ng, 6);
}

/***=======================================================================***/
/*** ExtendN: extend (N-1)th order B-spline coefficients to Nth order.     ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   SP:    the array of B-spline coefficients and charge mesh indices   ***/
/***            in this dimension (the dimension of interest)              ***/
/***   w:     one minus the non-integer part of the scaled fractional      ***/
/***            coordinate                                                 ***/
/***   ordr:  the interpolation order in the dimension of interest         ***/
/***=======================================================================***/
void ExtendN(bcof* SP, double w, int ordr)
{
  int i, om1;
  double dv;

  om1 = ordr-1;
  dv = 1.0/om1;
  SP[om1].s = dv*w*SP[ordr-2].s;
  for (i = 1; i <= ordr-2; i++) {
    SP[om1-i].s = dv*((w+i)*SP[ordr-i-2].s + (ordr-i-w)*SP[om1-i].s);
  }
  SP[0].s = dv*(1.0-w)*SP[0].s;
}

/***=======================================================================***/
/*** DiffN: obtain the derivatives of an N-th order B-Spline from (N-1)th  ***/
/***        order B-Spline coefficients.                                   ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   SP:    the array of B-spline coefficients and charge mesh indices   ***/
/***            in this dimension (the dimension of interest)              ***/
/***   ordr:  the interpolation order in the dimension of interest         ***/
/***=======================================================================***/
void DiffN(bcof* SP, int ordr)
{
  int i;

  SP[0].d = -SP[0].s;
  for (i = 1; i <= ordr-2; i++) {
    SP[i].d = SP[i-1].s - SP[i].s;
  }
  SP[ordr-1].d = SP[ordr-2].s;
}

/***=======================================================================***/
/*** SplCoeff: this is an advanced function for optimal computation of     ***/
/***           B-Spline coefficients.                                      ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:    the system coordinates                                      ***/
/***   pmmap:  the particle <--> mesh map                                  ***/
/***   rcinp:  the reciprocal space control data                           ***/
/***=======================================================================***/
void SplCoeff(coord *crd, bmap *pmmap, reccon *rcinp)
{
  int i, j, all4, a644, a446, all6, t0i, t1i, t2i, ngx, ngy, ngz;
  int *tordr;
  double ufacx, ufacy, ufacz, wx, wy, wz, ux, uy, uz, ucx, ucy, ucz;
  double *tloc;
  bcof *tSPx, *tSPy, *tSPz;

  /*** Shortcuts ***/
  ngx = rcinp->ng[0];
  ngy = rcinp->ng[1];
  ngz = rcinp->ng[2];
  ufacx = ngx/crd->gdim[0];
  ufacy = ngy/crd->gdim[1];
  ufacz = ngz/crd->gdim[2];
  tloc = crd->loc;
  tordr = rcinp->ordr;
  all4 = (tordr[0] == 4 && tordr[1] == 4 && tordr[2] == 4) ? 1 : 0;
  a644 = (tordr[0] == 6 && tordr[1] == 4 && tordr[2] == 4) ? 1 : 0;
  a446 = (tordr[0] == 4 && tordr[1] == 4 && tordr[2] == 6) ? 1 : 0;
  all6 = (tordr[0] == 6 && tordr[1] == 6 && tordr[2] == 6) ? 1 : 0;

  /*** Loop over all atoms ***/
  for (i = 0; i < crd->natom; i++) {

    /*** Atom identification number ***/
    pmmap->atmid[i] = crd->atmid[i];

    /*** Pointers! ***/
    t0i = tordr[0]*i;
    t1i = tordr[1]*i;
    t2i = tordr[2]*i;
    tSPx = &pmmap->xcof[t0i];
    tSPy = &pmmap->ycof[t1i];
    tSPz = &pmmap->zcof[t2i];

    /*** Convert to scaled fractional coordinates ***/
    ux = ufacx*tloc[3*i];
    uy = ufacy*tloc[3*i+1];
    uz = ufacz*tloc[3*i+2];
    ucx = ceil(ux);
    ucy = ceil(uy);
    ucz = ceil(uz);

    /*** One pass to order 3 ***/
    wx = ux - ucx + 1.0;
    wy = uy - ucy + 1.0;
    wz = uz - ucz + 1.0;

    /*** Compute third order splines ***/
    tSPx[2].s = 0.5*wx*wx;
    tSPy[2].s = 0.5*wy*wy;
    tSPz[2].s = 0.5*wz*wz;
    tSPx[0].s = 0.5*(1.0-wx)*(1.0-wx);
    tSPy[0].s = 0.5*(1.0-wy)*(1.0-wy);
    tSPz[0].s = 0.5*(1.0-wz)*(1.0-wz);
    tSPx[1].s = 1.0 - tSPx[0].s - tSPx[2].s;
    tSPy[1].s = 1.0 - tSPy[0].s - tSPy[2].s;
    tSPz[1].s = 1.0 - tSPz[0].s - tSPz[2].s;

    /*** Special case for isotropic 4th order interpolation ***/
    if (all4 == 1) {
      Extend4(tSPx, wx, ucx, ngx);
      Extend4(tSPy, wy, ucy, ngy);
      Extend4(tSPz, wz, ucz, ngz);
    }

    /*** Special case for anisotropic 6-4-4 interpolation ***/ 
    else if (a644 == 1) {
      Extend6(tSPx, wx, ucx, ngx);
      Extend4(tSPy, wy, ucy, ngy);
      Extend4(tSPz, wz, ucz, ngz);
    }

    /*** Special case for anisotropic 4-4-6 interpolation ***/ 
    else if (a446 == 1) {
      Extend4(tSPx, wx, ucx, ngx);
      Extend4(tSPy, wy, ucy, ngy);
      Extend6(tSPz, wz, ucz, ngz);
    }

    /*** Special case for isotropic 6th order interpolation ***/
    else if (all6 == 1) {
      Extend6(tSPx, wx, ucx, ngx);
      Extend6(tSPy, wy, ucy, ngy);
      Extend6(tSPz, wz, ucz, ngz);
    }

    /*** General case ***/
    else {

      /*** Things in the X dimension ***/
      if (tordr[0] >= 4) {
	for (j = 4; j < tordr[0]; j++) {
	  ExtendN(tSPx, wx, j);
	}
	DiffN(tSPx, tordr[0]);
	ExtendN(tSPx, wx, tordr[0]);
      }
      else {
	tSPx[0].d = wx - 1.0;
	tSPx[1].d = 1.0 - wx - wx;
	tSPx[2].d = wx;
      }
      FillQMap(tSPx, ucx-tordr[0], ngx, tordr[0]);

      /*** Things in the Y dimension ***/
      if (tordr[1] >= 4) {
	for (j = 4; j < tordr[1]; j++) {
	  ExtendN(tSPy, wy, j);
	}
	DiffN(tSPy, tordr[1]);
	ExtendN(tSPy, wy, tordr[1]);
      }
      else {
	tSPy[0].d = wy - 1.0;
	tSPy[1].d = 1.0 - wy - wy;
	tSPy[2].d = wy;
      }
      FillQMap(tSPy, ucy-tordr[1], ngy, tordr[1]);

      /*** Things in the Z dimension ***/
      if (tordr[2] >= 4) {
	for (j = 4; j < tordr[2]; j++) {
	  ExtendN(tSPz, wz, j);
	}
	DiffN(tSPz, tordr[2]);
	ExtendN(tSPz, wz, tordr[2]);
      }
      else {
	tSPz[0].d = wz - 1.0;
	tSPz[1].d = 1.0 - wz - wz;
	tSPz[2].d = wz;
      }
      FillQMap(tSPz, ucz-tordr[2], ngz, tordr[2]);
    }
  }
}
