#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "Random.h"
#include "mdgxVector.h"
#include "Thermostats.h"
#include "CellManip.h"
#include "pmeRecip.h"
#include "CrdManip.h"
#include "Integrator.h"
#include "Macros.h"
#include "Trajectory.h"

#include "TopologyDS.h"
#include "CrdManipDS.h"
#include "pmeDirectDS.h"

/***=======================================================================***/
/*** ScaleCoordinates: this routine adjusts all system coordinates, taking ***/
/***                   into account rigidly constrained groups (rigidly    ***/
/***                   constrained groups are moved so that their centers  ***/
/***                   of mass are rescaled).                              ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:    the system coordinates                                      ***/
/***   tp:     the system topology                                         ***/
/***   chi:    the system rescaling factor                                 ***/
/***=======================================================================***/
static void ScaleCoordinates(coord *crd, prmtop *tp, double* chi,
			     dircon *dcinp, reccon *rcinp, bckit *PPk,
			     cellgrid *CG)
{
  int i, j, h1, h2, neednewgrid;
  int cdim[3];
  double tmass, cofm, scofm;
  double xpos[24];

  /*** If the unit cell is not orthorhombic, we must ***/
  /*** work in terms of the fractional coordinates   ***/
  if (crd->isortho == 0) {
    printf("ScaleCoordinates >> Error.  Not yet ready to scale "
	   "non-rectangular cells.\n");
    exit(1);
  }

  /*** Otherwise, an orthorhombic unit cell is easy to work with ***/
  else {

    /*** Update positions of every atom not in a constraint group ***/
    for (i = 0; i < tp->natom; i++) {
      if (tp->SHL[i].exe == 0) {
        for (j = 3*i; j < 3*i+3; j++) {
	  crd->loc[j] *= chi[j-3*i];
        }
      }
    }

    /*** Update the positions of all constraint groups ***/
    for (i = 0; i < tp->natom; i++) {
      if (tp->SHL[i].exe == 1) {

	/*** Re-image the constraint group ***/
        h1 = tp->SHL[i].blist[0];
        h2 = tp->SHL[i].blist[1];
	for (j = 0; j < 3; j++) {
	  xpos[j] = crd->loc[3*i+j];
	  xpos[3+j] = crd->loc[3*h1+j];
	  xpos[6+j] = crd->loc[3*h2+j];
	}
	ReimageConstraintGroup(xpos, crd, 3);
	for (j = 0; j < 3; j++) {
	  crd->loc[3*i+j] = xpos[j];
	  crd->loc[3*h1+j] = xpos[3+j];
	  crd->loc[3*h2+j] = xpos[6+j];
	}

        /*** Compute current center of mass ***/
        tmass = tp->Masses[i] + tp->Masses[h1] + tp->Masses[h2];
        for (j = 0; j < 3; j++) {
	  cofm = (tp->Masses[i]*crd->loc[3*i+j] +
	          tp->Masses[h1]*crd->loc[3*h1+j] +
	          tp->Masses[h2]*crd->loc[3*h2+j])/tmass;
	  scofm = (chi[j] - 1.0)*cofm;
	  crd->loc[3*i+j] += scofm;
          crd->loc[3*h1+j] += scofm;
          crd->loc[3*h2+j] += scofm;
	}
      }
    }
  }

  /*** Update unit cell dimensions ***/
  for (i = 0; i < 3; i++) {
    crd->gdim[i] *= chi[i];
    crd->hgdim[i] = 0.5*crd->gdim[i];
  }
  CompXfrm(crd->gdim, crd->U, crd->invU);

  /*** Update the PME machinery to compensate for the new box size ***/
  UpdateBCKit(rcinp, crd, tp, PPk);

  /*** Check the cell grid dimensions ***/
  TakeCellGridDims(cdim, crd, dcinp);
  neednewgrid = 0;
  for (i = 0; i < 3; i++) {
    if (cdim[i] != CG->ng[i]) {
      neednewgrid = 1;
    }
  }
  if (neednewgrid == 1) {

    /*** Destroy the existing cell grid and make ***/
    /*** a new one with the correct dimensions   ***/
    DestroyCellGrid(CG);
    *CG = CreateCellGrid(crd, dcinp);
  }
}

/***=======================================================================***/
/*** PrepMCBarostat: prepare volume rescaling factors for a Monte-Carlo    ***/
/***                 barostat.                                             ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tj:     the trajectory control data                                 ***/
/***   crd:    the system coordinates                                      ***/
/***=======================================================================***/
void PrepMCBarostat(trajcon *tj, coord *crd)
{
  if (tj->MCBarostatFac[0] < 0.0) {
    printf("PrepMCBarostat >> Error.  Bad input to Monte Carlo barostat, "
	   "X-direction /\nPrepMCBarostat >> isotropic rescaling factor set "
	   "to %.6lf\n", tj->MCBarostatFac[0]);
    exit(1);
  }

  /*** Decide whether Monte-Carlo volume moves are to be isotropic ***/
  if (tj->MCBarostatFac[1] < 0.0 && tj->MCBarostatFac[2] < 0.0) {
    tj->MCBarostatIso = 1;
    tj->MCBarostatFac[1] = tj->MCBarostatFac[0];
    tj->MCBarostatFac[2] = tj->MCBarostatFac[0];
  }
  else {
    if (fabs(tj->MCBarostatFac[1] - tj->MCBarostatFac[0]) < 1.0e-8 &&
	fabs(tj->MCBarostatFac[2] - tj->MCBarostatFac[0]) < 1.0e-8) {
      tj->MCBarostatIso = 1;
    }
    else {
      tj->MCBarostatIso = 0;
    }
  }

  /*** Finally, if one of the Y or Z dimensions is unset, set it to zero ***/
  if (tj->MCBarostatFac[1] < 0.0) {
    tj->MCBarostatFac[1] = 0.0;
  }
  if (tj->MCBarostatFac[2] < 0.0) {
    tj->MCBarostatFac[2] = 0.0;
  }

  /*** Decide on the volume rescaling amount, an absolute ***/
  /*** volume based on the system's initial volume        ***/
  tj->mcdVmax = ((1.0 + tj->MCBarostatFac[0]) * (1.0 + tj->MCBarostatFac[1]) *
		 (1.0 + tj->MCBarostatFac[2]) - 1.0) *
                (crd->invU.map[0][0] * crd->invU.map[1][1] *
		 crd->invU.map[2][2]);
}

/***=======================================================================***/
/*** CurrentSystemPressure: compute the current pressure of the system.    ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   sysUV:  the system energy and virial data structure                 ***/
/***=======================================================================***/
double CurrentSystemPressure(Energy *sysUV, coord *crd)
{
  double virial, volume;

  virial = sysUV->Vir[0] + sysUV->Vir[4] + sysUV->Vir[8];
  volume = crd->invU.data[0]*crd->invU.data[4]*crd->invU.data[8];

  return (2.0*sysUV->kine - virial) / (3.0 * volume);
}

/***=======================================================================***/
/*** BerendsenBarostat: use a Berendsen piston to adjust the system volume ***/
/***                    in response to the instantaneous virial.           ***/
/***=======================================================================***/
void BerendsenBarostat(coord *crd, prmtop *tp, trajcon *tj, Energy *sysUV,
                       dircon *dcinp, reccon *rcinp, bckit *PPk, cellgrid *CG)
{
  double chi, bptc, beta, pressure;
  double chivec[3];

  /*** Compute chi, the volume rescale ***/
  /*** factor for isotropic rescaling  ***/
  beta = tj->BerendsenPCoupl;
  bptc = tj->BerendsenPTime;
  sysUV->kine = CurrentKineticEnergy(crd, tp);
  pressure = CurrentSystemPressure(sysUV, crd);
  chi = 1.0 - beta * (tj->dt / bptc) * (tj->Ptarget - pressure*PCONVFAC);
  if (chi < 0.95) {
    printf("BerendsenBarostat >> Warning.  Large rescaling value clamped at "
	   "95%% of\nBerendsenBarostat >> instantaneous box size.\n");
    chi = 0.95;
  }
  else if (chi > 1.05) {

    /*** Terminate the run if this is truly a problem; ***/
    /*** prevent mdgx from gobbling up system memory!  ***/ 
    if (chi > 1.10) {
      printf("BerendsenBarostat >> Rescaling value %.4lf.  This is far too "
	     "large and\nBerendsenBarostat >> indicates a problem with the "
	     "system.\n", chi);

      /*** Run some quick diagnostics to try ***/
      /*** and indicate what went wrong      ***/
      int i, maxatm, maxatm2;
      double maxfrc, maxfrc2, ifrc, sfx, sfy, sfz;
      maxfrc = 0.0;
      maxfrc2 = 0.0;
      maxatm = -1;
      maxatm2 = -1;
      sfx = 0.0;
      sfy = 0.0;
      sfz = 0.0;
      for (i = 0; i < crd->natom; i++) {
	ifrc = sqrt(crd->frc[3*i]*crd->frc[3*i] +
		    crd->frc[3*i+1]*crd->frc[3*i+1] +
		    crd->frc[3*i+2]*crd->frc[3*i+2]);
	if (ifrc > maxfrc) {
	  maxfrc2 = maxfrc;
	  maxatm2 = maxatm;
	  maxfrc = ifrc;
	  maxatm = i;
	}
	if (ifrc > maxfrc2 && ifrc < maxfrc) {
	  maxfrc2 = ifrc;
	  maxatm2 = i;
	}
	sfx += crd->frc[3*i];
	sfy += crd->frc[3*i+1];
	sfz += crd->frc[3*i+2];
      }
      printf("BerendsenBarostat >> High absolute force: %.4lf on atom %d.\n",
	     maxfrc, maxatm); 
      printf("BerendsenBarostat >> High absolute force: %.4lf on atom %d.\n",
	     maxfrc2, maxatm2);
      printf("BerendsenBarostat >> Total force on system:\n"
	     "Berendsenbarostat >> [ %16.8lf %16.8lf %16.8lf ];\n",
	     sfx, sfy, sfz);
      printf("All forces:\n");
      for (i = 0; i < crd->natom; i++) {
	printf("%16.8lf%16.8lf%16.8lf\n", crd->frc[3*i], crd->frc[3*i+1],
	       crd->frc[3*i+2]);
      }
      exit(1);
    }

    /*** If the rescaling is not too severe, continue ***/
    printf("BerendsenBarostat >> Warning.  Large rescaling value clamped at "
           "105%% of\nBerendsenBarostat >> instantaneous box size.\n");
    chi = 1.05;
  }

  /*** Rescale positions by chi ^ 1/3 ***/
  chi = pow(chi, 1.0 / 3.0);
  chivec[0] = chi;
  chivec[1] = chi;
  chivec[2] = chi;

  /*** Update positions ***/
  ScaleCoordinates(crd, tp, chivec, dcinp, rcinp, PPk, CG);
}

/***=======================================================================***/
/*** CalcPdV: compute (P dV) for a volume change of the unit cell based on ***/
/***          the initial coordinates and a set of rescaling factors.      ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:    the system coordinates                                      ***/
/***   chi:    the rescaling factors                                       ***/
/***   tj:     the trajectory control data                                 ***/
/***=======================================================================***/
static double CalcPdV(coord *crd, trajcon *tj, double *chi)
{
  int i;
  double Eorig, Enew, dE;
  double celldim[6];

  /*** Save the original unit cell parameters ***/
  for (i = 0; i < 6; i++) {
    celldim[i] = crd->gdim[i];
  }

  /*** Compute the original pressure-volume energy ***/
  Eorig = (crd->invU.map[0][0] * crd->invU.map[1][1] * crd->invU.map[2][2]) *
    tj->Ptarget;

  /*** Compute the new dimensions ***/
  for (i = 0; i < 3; i++) {
    crd->gdim[i] *= chi[i];
  }
  CompXfrm(crd->gdim, crd->U, crd->invU);

  /*** Compute the new pressure-volume energy ***/
  Enew = (crd->invU.map[0][0] * crd->invU.map[1][1] * crd->invU.map[2][2]) *
    tj->Ptarget;

  /*** Compute the difference in energies and restate it in kcal/mol ***/
  dE = (Enew - Eorig) * 6.0221367e-2 / 4184.0;

  /*** Reinstate the original cell dimensions to leave them unchanged ***/
  for (i = 0; i < 6; i++) {
    crd->gdim[i] = celldim[i];
  }
  CompXfrm(crd->gdim, crd->U, crd->invU);

  return dE;
}

/***=======================================================================***/
/*** MonteCarloBarostat: this routine implements a Monte-Carlo barostat;   ***/
/***                     while conceptually simple, requiring merely a     ***/
/***                     coordinate rescaling followed by an energy        ***/
/***                     recalculation and acceptance of the move based on ***/
/***                     the Metropolis criterion, the barostat may not be ***/
/***                     very efficient unless the new energy can be       ***/
/***                     easily computed.  Energy recalculations are easy  ***/
/***                     if there are no bond constraints and coordinate   ***/
/***                     rescaling is isotropic.                           ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:    the system coordinates                                      ***/
/***   tp:     the system topology                                         ***/
/***   tj:     the trajectory control data                                 ***/
/***=======================================================================***/
void MonteCarloBarostat(coord *crd, prmtop *tp, trajcon *tj, Energy *sysUV,
			dircon *dcinp, FrcTab *Etab, FrcTab *EHtab,
			reccon *rcinp, bmap *pmmap, bckit *PPk, cellgrid *CG,
			execon *etimers)
{
  int i;
  double cEtot, nEtot, dV;
  double invchi, invchi6, PVwork, beta, expfac, accprob;
  double chi[3], revchi[3];
  Energy trialUV;

  /*** Decide on the move to make ***/
  dV = tj->mcdVmax*(ran2(&tj->rndcon) - 0.5) / 
    (crd->invU.map[0][0]*crd->invU.map[1][1]*crd->invU.map[2][2]);
  if (tj->MCBarostatIso == 1) {
    chi[0] = pow(1.0+dV, 1.0/3.0);
    chi[1] = chi[0];
    chi[2] = chi[0];
  }
  else {
    printf("MonteCarloBarostat >> Error.  Not yet prepared for anisotropic "
	   "rescaling.\n");
    exit(1);
  }

  /*** Compute the pressure-volume work ***/
  PVwork = CalcPdV(crd, tj, chi);

  /*** Compute current system energy ***/
  CurrentSystemTemperature(crd, tp, sysUV, 1);
  SumTotalEnergy(sysUV);
  cEtot = sysUV->eptot;

  /*** With isotropic position rescaling and no constraints,   ***/
  /*** we can easily estimate the new energy without rescaling ***/
  if (tj->MCBarostatIso == 1 && tp->ncnst == 0) {
    invchi  = 1.0/chi[0];
    invchi6 = pow(invchi, 6.0);

    /*** Estimate the rescaled system energy ***/
    nEtot = sysUV->angl + sysUV->dihe + invchi*(sysUV->relec + sysUV->delec) +
      invchi6*invchi6*sysUV->vdw12 + invchi6*sysUV->vdw6;
    for (i = 0; i < tp->nBAH.nbond; i++) {
      nEtot += tp->BParam[i].K*(sysUV->BondUdc[3*i]*chi[0]*chi[0] -
				2.0*chi[0]*sysUV->BondUdc[3*i+1] +
				sysUV->BondUdc[3*i+2]);
    }
  }

  /*** If we don't have such a nice situation,   ***/
  /*** we must recompute the energy from scratch ***/
  else {

    /*** Buffer the current locations and forces ***/
    ReflectDVec(crd->bufloc, crd->loc, 3*crd->natom);
    ReflectDVec(crd->buffrc, crd->frc, 3*crd->natom);

    /*** Rescale the coordinates ***/
    ScaleCoordinates(crd, tp, chi, dcinp, rcinp, PPk, CG);

    /*** Recompute the energy, specifically setting the updateU ***/
    /*** field of the sysUV struct to -1 to signal that energy  ***/
    /*** alone, without a force calculation, is needed.         ***/
    InitializeEnergy(&trialUV, tj, tp, 1);
    trialUV.updateU = -1;
    AtomForces(crd, CG, tp, dcinp, Etab, EHtab, rcinp, pmmap, PPk, &trialUV,
	       etimers);
    SumTotalEnergy(&trialUV);
    nEtot = trialUV.eptot;
    DestroyEnergyTracker(&trialUV);
  }

  /*** Now, compare the new energy (nEtot + PVwork) with the ***/
  /*** old energy (cEtot) and decide whether to accept or    ***/
  /*** reject the move based on the Metropolis criterion.    ***/
  /*** The temperature used in the Metropolis criterion is   ***/
  /*** that used in the system thermostat.                   ***/
  beta = 1.0/(GASCNST*tj->Ttarget);
  expfac = nEtot - cEtot + PVwork - tp->nprtcl*log(chi[0]*chi[1]*chi[2])/beta;
  accprob = MIN(1.0, exp(-beta*expfac));
  double rnum = ran2(&tj->rndcon);
  if (rnum < accprob) {

    /*** Accept the move; rescale coordinates if rescaling has    ***/
    /*** not already occured and replace forces with their values ***/
    /*** from before the move if rescaling has already occurred.  ***/
    if (tj->MCBarostatIso == 1 && tp->ncnst == 0) {
      ScaleCoordinates(crd, tp, chi, dcinp, rcinp, PPk, CG);
    }
    else {
      ReflectDVec(crd->frc, crd->buffrc, 3*crd->natom);
    }
  }
  else {

    /*** Reject the move; return coordinates to their    ***/
    /*** original positions and forces to their original ***/
    /*** values if rescaling has already occured.        ***/
    if (tj->MCBarostatIso != 1 || tp->ncnst != 0) {
      for (i = 0; i < 3; i++) {
	revchi[i] = 1.0/chi[i];
      }
      ScaleCoordinates(crd, tp, revchi, dcinp, rcinp, PPk, CG);
      ReflectDVec(crd->frc, crd->buffrc, 3*crd->natom);
    }
  }
}
