/***=======================================================================***/
/*** Like the nonbonded loop, there are different versions of the bond,    ***/
/*** angle, and dihedral routines.  The different variants can calculate   ***/
/*** forces alone, energies alone, forces and energies, or forces,         ***/
/*** energies, and virial contributions.                                   ***/
/***=======================================================================***/
#if NEEDFORCE == 1
  #if NEEDENERGY == 1
    #if NEEDVIRIAL == 1 
      #define BONDCALC BondVir
      #define ANGLCALC AnglVir
      #define DIHECALC DiheVir
    #else
      #define BONDCALC BondFrcNrg
      #define ANGLCALC AnglFrcNrg
      #define DIHECALC DiheFrcNrg
    #endif
  #else
    #define BONDCALC BondFrc
    #define ANGLCALC AnglFrc
    #define DIHECALC DiheFrc
  #endif
#else
  #define BONDCALC BondNrg
  #define ANGLCALC AnglNrg
  #define DIHECALC DiheNrg
#endif

/***=======================================================================***/
/*** BondFrc: computes the force on two atoms due to a bonded interaction. ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   [ab]ptr:    the coordinates for atoms A or B                        ***/
/***   [ab]frc:    the forces for atoms A or B                             ***/
/***   bcom:       the bond command                                        ***/
/***   tp:         the topology                                            ***/
/***   Cfrc:       the "coarse" electrostatic force/energy spline table    ***/
/***   Hfrc:       the "fine" electrostatic force/energy spline table      ***/
/***   sysUV:      the system energy and virial (state information)        ***/
/***=======================================================================***/
#if NEEDFORCE == 1
#if NEEDENERGY == 1
void BONDCALC(double *aptr, double *bptr, double *afrc, double *bfrc,
	      bondcomm *bcom, prmtop *tp, coord *crd, FrcTab *Cfrc,
	      FrcTab *Hfrc, Energy *sysUV)
#else
void BONDCALC(double *aptr, double *bptr, double *afrc, double *bfrc,
              bondcomm *bcom, prmtop *tp, coord *crd, FrcTab *Cfrc,
              FrcTab *Hfrc)
#endif
#else
void BONDCALC(double *aptr, double *bptr, bondcomm *bcom, prmtop *tp,
              coord *crd, FrcTab *Cfrc, FrcTab *Hfrc, Energy *sysUV)
#endif
{
  int irc, irh;
  double r, r2, invr, invr2, dx, dy, dz, dl, qq;
#if NEEDFORCE == 1
  double fmag;
#endif
#if NEEDVIRIAL == 1
#ifdef IDEBUG
  double bondfmag, corrfmag;
#endif
#endif
  CSpln *CdSD, *HdSD, *CSD, *HSD;

  /*** Compute displacement ***/
  dx = bptr[0] - aptr[0];
  dy = bptr[1] - aptr[1];
  dz = bptr[2] - aptr[2];

  /*** Reimage ***/
  if (crd->isortho == 1) {
    OrthoReim(&dx, &dy, &dz, crd->hgdim);
  }
  else {
    NonOrthoReim(&dx, &dy, &dz, &crd->U, &crd->invU);
  }

  /*** Accumulate the bond force and energy ***/
  r2 = dx*dx + dy*dy + dz*dz;
  r = sqrt(r2);
  invr = 1.0/r;
  dl = tp->BParam[bcom->t].l0 - r;
  if (bcom->H == 0 || (bcom->H == 1 && tp->rattle == 0)) {
#if NEEDFORCE == 1
    fmag = -2.0*tp->BParam[bcom->t].K*dl*invr;
#endif
#if NEEDENERGY == 1
    sysUV->bond += tp->BParam[bcom->t].K*dl*dl;
    sysUV->BondUdc[3*bcom->t] += r*r;
    sysUV->BondUdc[3*bcom->t+1] += r*tp->BParam[bcom->t].l0;
    sysUV->BondUdc[3*bcom->t+2] += pow(tp->BParam[bcom->t].l0, 2.0);
#endif
#if NEEDVIRIAL == 1
#ifdef IDEBUG
    bondfmag = fmag;
#endif
#endif
  }
#if NEEDFORCE == 1
  else {
    fmag = 0.0;
  }
#endif

  /*** Correct 1:2 electrostatics ***/
  qq = tp->Charges[bcom->a]*tp->Charges[bcom->b];
  if (r2 < MINNB2) {
    irc = r2*Cfrc->ivdr;
    CdSD = Cfrc->dSD;
    CSD = Cfrc->SD;
#if NEEDFORCE == 1
    fmag -= qq*(((CdSD[irc].A*r2 + CdSD[irc].B)*r2 + CdSD[irc].C)*r2 +
		CdSD[irc].D);
#endif
#if NEEDENERGY == 1
    sysUV->delec -= qq*(((CSD[irc].A*r2 + CSD[irc].B)*r2 + CSD[irc].C)*r2 +
			CSD[irc].D);
#endif
    irh = r2*Hfrc->ivdr;
    HdSD = Hfrc->dSD;
    HSD = Hfrc->SD;
#if NEEDFORCE == 1
    fmag += qq*(((HdSD[irh].A*r2 + HdSD[irh].B)*r2 + HdSD[irh].C)*r2 +
		HdSD[irh].D);
#endif
#if NEEDENERGY == 1
    sysUV->delec += qq*(((HSD[irh].A*r2 + HSD[irh].B)*r2 + HSD[irh].C)*r2 +
			HSD[irh].D);
#endif
  }
  invr2 = invr*invr;
#if NEEDFORCE == 1
  fmag += BIOQ*qq*invr*invr2;
#endif
#if NEEDENERGY == 1
  sysUV->delec -= BIOQ*qq*invr;
#endif

  /*** Correct 1:2 van-der Waals interactions ***/
  if (r2 > MINNB2) {

    int ljA, ljB;
    double invr4;

    ljA = tp->LJIdx[bcom->a];
    ljB = tp->LJIdx[bcom->b];
    if (ljA >= 0 && ljB >= 0) {
      invr4 = invr2*invr2;
#if NEEDFORCE == 1
      fmag -= invr4*invr4*(tp->LJftab.map[ljA][2*ljB]*invr2*invr4 +
			   tp->LJftab.map[ljA][2*ljB+1]);
#endif
#if NEEDENERGY == 1
      sysUV->vdw12 -= invr4*invr2*invr4*invr2*tp->LJutab.map[ljA][2*ljB];
      sysUV->vdw6 -= invr4*invr2*tp->LJutab.map[ljA][2*ljB+1];
#endif
    }
  }

#if NEEDFORCE == 1
  /*** Accumulate the force ***/
  afrc[0] += fmag*dx;
  afrc[1] += fmag*dy;
  afrc[2] += fmag*dz;
  bfrc[0] -= fmag*dx;
  bfrc[1] -= fmag*dy;
  bfrc[2] -= fmag*dz;

#if NEEDVIRIAL == 1
#ifdef IDEBUG
  /*** Separate the virial contributions fron bonds and 1:2 exclusions ***/
  corrfmag = fmag - bondfmag;
  sysUV->Bond12Corr[0] += dx*corrfmag*dx;
  sysUV->Bond12Corr[1] += dx*corrfmag*dy;
  sysUV->Bond12Corr[2] += dx*corrfmag*dz;
  sysUV->Bond12Corr[3] += dy*corrfmag*dx;
  sysUV->Bond12Corr[4] += dy*corrfmag*dy;
  sysUV->Bond12Corr[5] += dy*corrfmag*dz;
  sysUV->Bond12Corr[6] += dz*corrfmag*dx;
  sysUV->Bond12Corr[7] += dz*corrfmag*dy;
  sysUV->Bond12Corr[8] += dz*corrfmag*dz;

  /*** Accumulate the stress tensor specific to bonds ***/
  sysUV->BondVir[0] += dx*bondfmag*dx;
  sysUV->BondVir[1] += dx*bondfmag*dy;
  sysUV->BondVir[2] += dx*bondfmag*dz;
  sysUV->BondVir[3] += dy*bondfmag*dx;
  sysUV->BondVir[4] += dy*bondfmag*dy;
  sysUV->BondVir[5] += dy*bondfmag*dz;
  sysUV->BondVir[6] += dz*bondfmag*dx;
  sysUV->BondVir[7] += dz*bondfmag*dy;
  sysUV->BondVir[8] += dz*bondfmag*dz;
#endif

  /*** Accumulate the stress tensor ***/
  sysUV->Vir[0] += dx*fmag*dx;
  sysUV->Vir[1] += dx*fmag*dy;
  sysUV->Vir[2] += dx*fmag*dz;
  sysUV->Vir[3] += dy*fmag*dx;
  sysUV->Vir[4] += dy*fmag*dy;
  sysUV->Vir[5] += dy*fmag*dz;
  sysUV->Vir[6] += dz*fmag*dx;
  sysUV->Vir[7] += dz*fmag*dy;
  sysUV->Vir[8] += dz*fmag*dz;
#endif
#endif
}

/***=======================================================================***/
/*** AnglFrc: computes the force on two atoms due to an angle interaction. ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   [abc]ptr:   the coordinates for atoms A, B, or C                    ***/
/***   [abc]frc:   the forces for atoms A, B, or C                         ***/
/***   acom:       the angle command                                       ***/
/***   tp:         the topology                                            ***/
/***   Cfrc:       the "coarse" electrostatic force/energy spline table    ***/
/***   Hfrc:       the "fine" electrostatic force/energy spline table      ***/
/***   sysUV:      the system energy and virial (state information)        ***/
/***=======================================================================***/
#if NEEDFORCE == 1
#if NEEDENERGY == 1
void ANGLCALC(double *aptr, double *bptr, double *cptr, double *afrc,
	     double *bfrc, double *cfrc, anglcomm *acom, prmtop *tp,
	     coord *crd, FrcTab *Cfrc, FrcTab *Hfrc, Energy *sysUV)
#else
void ANGLCALC(double *aptr, double *bptr, double *cptr, double *afrc,
	     double *bfrc, double *cfrc, anglcomm *acom, prmtop *tp,
	     coord *crd, FrcTab *Cfrc, FrcTab *Hfrc)
#endif
#else
void ANGLCALC(double *aptr, double *bptr, double *cptr, anglcomm *acom,
              prmtop *tp, coord *crd, FrcTab *Cfrc, FrcTab *Hfrc,
              Energy *sysUV)
#endif
{
  int i, irc, irh;
  double r2, invr, invr2, costheta, theta, dA, mgba, mgbc, mbabc;
#if NEEDFORCE == 1
  double fmag;
#endif
  double invbabc, sqba, sqbc, adf, cdf, qq, dtheta;
  double ac[3], ba[3], bc[3];
  CSpln *CdSD, *HdSD, *CSD, *HSD;

  /*** Compute displacements ***/
  for (i = 0; i < 3; i++) {
    ba[i] = aptr[i] - bptr[i];
    bc[i] = cptr[i] - bptr[i];
    ac[i] = cptr[i] - aptr[i];
  }

  /*** Reimage ***/
  if (crd->isortho == 1) {
    OrthoReim(&ba[0], &ba[1], &ba[2], crd->hgdim);
    OrthoReim(&bc[0], &bc[1], &bc[2], crd->hgdim);
    OrthoReim(&ac[0], &ac[1], &ac[2], crd->hgdim);
  }
  else {
    NonOrthoReim(&ba[0], &ba[1], &ba[2], &crd->U, &crd->invU);
    NonOrthoReim(&bc[0], &bc[1], &bc[2], &crd->U, &crd->invU);
    NonOrthoReim(&ac[0], &ac[1], &ac[2], &crd->U, &crd->invU);
  }

  /*** On to the angle force computation ***/
  mgba = ba[0]*ba[0] + ba[1]*ba[1] + ba[2]*ba[2];
  mgbc = bc[0]*bc[0] + bc[1]*bc[1] + bc[2]*bc[2];
  invbabc = 1.0/sqrt(mgba*mgbc);
  costheta = (ba[0]*bc[0] + ba[1]*bc[1] + ba[2]*bc[2]) * invbabc;
  costheta = (costheta < -1.0) ? -1.0 : (costheta > 1.0) ? 1.0 : costheta;
  theta = acos(costheta);
  dtheta = theta - tp->AParam[acom->t].th0;
  dA = -2.0*tp->AParam[acom->t].K*dtheta /
    sqrt(1.0 - costheta*costheta);
  sqba = dA/mgba;
  sqbc = dA/mgbc;
  mbabc = dA * invbabc;

#if NEEDFORCE == 1
  /*** Accumulate the angle forces and stress tensor ***/
  for (i = 0; i < 3; i++) {
    adf = bc[i]*mbabc - costheta*ba[i]*sqba;
    cdf = ba[i]*mbabc - costheta*bc[i]*sqbc;
    afrc[i] -= adf;
    bfrc[i] += adf + cdf;
    cfrc[i] -= cdf;
#if NEEDVIRIAL == 1
#ifdef IDEBUG
    /*** Accumulate the virial specific to angles ***/
    sysUV->AnglVir[0+i] += ba[0]*adf + bc[0]*cdf;
    sysUV->AnglVir[3+i] += ba[1]*adf + bc[1]*cdf;
    sysUV->AnglVir[6+i] += ba[2]*adf + bc[2]*cdf;
#endif
    sysUV->Vir[0+i] += ba[0]*adf + bc[0]*cdf;
    sysUV->Vir[3+i] += ba[1]*adf + bc[1]*cdf;
    sysUV->Vir[6+i] += ba[2]*adf + bc[2]*cdf;
#endif
  }
#endif

#if NEEDENERGY == 1
  /*** Accumulate the angle energy ***/
  sysUV->angl += tp->AParam[acom->t].K*dtheta*dtheta;
#endif

  /*** Bail out if this angle's A:C interactions do not need to be ***/
  /*** excluded (i.e. they were already excluded by some bond)     ***/
  if (acom->excl == 0) {
    return; 
  }

  /*** Correct 1:3 electrostatics ***/
  qq = tp->Charges[acom->a]*tp->Charges[acom->c];
  r2 = ac[0]*ac[0] + ac[1]*ac[1] + ac[2]*ac[2];
  invr = 1.0/sqrt(r2);
  invr2 = invr*invr;
#if NEEDFORCE == 1
  fmag = BIOQ*qq*invr*invr2;
#endif
#if NEEDENERGY == 1
  sysUV->delec -= BIOQ*qq*invr;
#endif
  if (r2 < MINNB2) {
    irc = r2*Cfrc->ivdr;
    CdSD = Cfrc->dSD;
    CSD = Cfrc->SD;
#if NEEDFORCE == 1
    fmag -= qq*(((CdSD[irc].A*r2 + CdSD[irc].B)*r2 + CdSD[irc].C)*r2 +
		CdSD[irc].D);
#endif
#if NEEDENERGY == 1
    sysUV->delec -= qq*(((CSD[irc].A*r2 + CSD[irc].B)*r2 + CSD[irc].C)*r2 +
			CSD[irc].D);
#endif
    irh = r2*Hfrc->ivdr;
    HdSD = Hfrc->dSD;
    HSD = Hfrc->SD;
#if NEEDFORCE == 1
    fmag += qq*(((HdSD[irh].A*r2 + HdSD[irh].B)*r2 + HdSD[irh].C)*r2 +
		HdSD[irh].D);
#endif
#if NEEDENERGY == 1
    sysUV->delec += qq*(((HSD[irh].A*r2 + HSD[irh].B)*r2 + HSD[irh].C)*r2 +
			HSD[irh].D);
#endif
  }

  /*** Correct 1:3 van-der Waals interactions ***/
  if (r2 > MINNB2) {

    int ljA, ljB;
    double invr4;

    ljA = tp->LJIdx[acom->a];
    ljB = tp->LJIdx[acom->c];
    if (ljA >= 0 && ljB >= 0) {
      invr4 = invr2*invr2;
#if NEEDFORCE == 1
      fmag -= invr4*invr4*(tp->LJftab.map[ljA][2*ljB]*invr2*invr4 +
			   tp->LJftab.map[ljA][2*ljB+1]);
#endif
#if NEEDENERGY == 1
      sysUV->vdw12 -= invr4*invr2*invr4*invr2*tp->LJutab.map[ljA][2*ljB];
      sysUV->vdw6 -= invr4*invr2*tp->LJutab.map[ljA][2*ljB+1];
#endif
    }
  }

#if NEEDFORCE == 1
  /*** Accumulate the force ***/
  afrc[0] += fmag*ac[0];
  afrc[1] += fmag*ac[1];
  afrc[2] += fmag*ac[2];
  cfrc[0] -= fmag*ac[0];
  cfrc[1] -= fmag*ac[1];
  cfrc[2] -= fmag*ac[2];

#if NEEDVIRIAL == 1
#ifdef IDEBUG
  /*** Accumulate the 1:3 correction specific to angles ***/
  sysUV->Angl13Corr[0] += ac[0]*fmag*ac[0];
  sysUV->Angl13Corr[1] += ac[0]*fmag*ac[1];
  sysUV->Angl13Corr[2] += ac[0]*fmag*ac[2];
  sysUV->Angl13Corr[3] += ac[1]*fmag*ac[0];
  sysUV->Angl13Corr[4] += ac[1]*fmag*ac[1];
  sysUV->Angl13Corr[5] += ac[1]*fmag*ac[2];
  sysUV->Angl13Corr[6] += ac[2]*fmag*ac[0];
  sysUV->Angl13Corr[7] += ac[2]*fmag*ac[1];
  sysUV->Angl13Corr[8] += ac[2]*fmag*ac[2];

  printf("Added %16.6lf to Angl13Corr[8].\n", ac[2]*fmag*ac[2]);
#endif

  /*** Accumulate the virial ***/
  sysUV->Vir[0] += ac[0]*fmag*ac[0];
  sysUV->Vir[1] += ac[0]*fmag*ac[1];
  sysUV->Vir[2] += ac[0]*fmag*ac[2];
  sysUV->Vir[3] += ac[1]*fmag*ac[0];
  sysUV->Vir[4] += ac[1]*fmag*ac[1];
  sysUV->Vir[5] += ac[1]*fmag*ac[2];
  sysUV->Vir[6] += ac[2]*fmag*ac[0];
  sysUV->Vir[7] += ac[2]*fmag*ac[1];
  sysUV->Vir[8] += ac[2]*fmag*ac[2];
#endif
#endif
}

#ifndef DIHE_CALC_FUNCS
#define DIHE_CALC_FUNCS
/***=======================================================================***/
/*** BracketAB: this function takes two vectors V and W and returns the    ***/
/***            scalar quantity defined as follows:                        ***/
/***                                                                       ***/
/***            [V,W]_a = \sum_b ( 1 - \delta_{a,b} ) V^b W^b              ***/
/***                                                                       ***/
/***            This routine is needed in order to compute the virial for  ***/
/***            dihedral interactions.                                     ***/
/***=======================================================================***/
static double BracketAB(double* V, double* W, int a)
{
  int b;
  double rslt;

  rslt = 0.0;
  for (b = 0; b < 3; b++) {
    rslt += V[b]*W[b];
  }
  rslt -= V[a]*W[a];

  return rslt;
}

/***=======================================================================***/
/*** DiheStressTensorAB: this routine computes the contribution of a       ***/
/***                     dihedral to the (a,b) component of the stress     ***/
/***                     tensor.                                           ***/
/***=======================================================================***/
double DiheStressTensorAB(double* rij, double* rjk, double* rkn, int a, int b,
			  double diheang)
{
  int i;
  double mrij, mrjk, mrkn, pib, pnb, pjkb, gib, gkb, hjb, hnb;
  double sigab, mcrijk, mcrjkn;
  double crijrjk[3], crjkrkn[3];

  /*** Compute cross products of consecutive links in the dihedral ***/
  CrossP(rij, rjk, crijrjk);
  CrossP(rjk, rkn, crjkrkn);

  /*** Compute the length of each link in the dihedral ***/
  mrij = 0.0;
  mrjk = 0.0;
  mrkn = 0.0;
  mcrijk = 0.0;
  mcrjkn = 0.0;
  for (i = 0; i < 3; i++) {
    mrij += rij[i]*rij[i];
    mrjk += rjk[i]*rjk[i];
    mrkn += rkn[i]*rkn[i];
    mcrijk += crijrjk[i]*crijrjk[i];
    mcrjkn += crjkrkn[i]*crjkrkn[i];
  }
  mrij = sqrt(mrij);
  mrjk = sqrt(mrjk);
  mrkn = sqrt(mrkn);
  mcrijk = sqrt(mcrijk);
  mcrjkn = sqrt(mcrjkn);

  pib = (rjk[b]*BracketAB(rjk, rkn, b) - rkn[b]*BracketAB(rjk, rjk, b)) /
    (mcrijk*mcrjkn);
  pnb = (rjk[b]*BracketAB(rij, rjk, b) - rij[b]*BracketAB(rjk, rjk, b)) /
    (mcrijk*mcrjkn);
  pjkb = (rij[b]*BracketAB(rjk, rkn, b) + rkn[b]*BracketAB(rij, rjk, b) -
	  2.0*rjk[b]*BracketAB(rij, rkn, b)) / (mcrijk*mcrjkn);
  gib = 2.0*(rij[b]*BracketAB(rjk, rjk, b) - rjk[b]*BracketAB(rij, rjk, b)) /
    (mcrijk*mcrijk);
  gkb = 2.0*(rjk[b]*BracketAB(rij, rij, b) - rij[b]*BracketAB(rij, rjk, b)) /
    (mcrijk*mcrijk);
  hjb = 2.0*(rjk[b]*BracketAB(rkn, rkn, b) - rkn[b]*BracketAB(rjk, rkn, b)) /
    (mcrjkn*mcrjkn);
  hnb = 2.0*(rkn[b]*BracketAB(rjk, rjk, b) - rjk[b]*BracketAB(rjk, rkn, b)) /
    (mcrjkn*mcrjkn);
  sigab = rij[a]*pib + rjk[a]*pjkb + rkn[a]*pnb -
    0.5*cos(diheang)*(rij[a]*gib + rjk[a]*gkb + rjk[a]*hjb + rkn[a]*hnb);

  return sigab;
}
#endif

/***=======================================================================***/
/*** DiheFrc: computes the force on four atoms due to a dihedral motion.   ***/
/***          This also applies to improper dihedrals.  Standard and       ***/
/***          improper dihedrals are treated using the following notation: ***/
/***                                                                       ***/
/***             Standard Dihedral          Improper Dihedral              ***/
/***                                                                       ***/
/***                       D                           D                   ***/
/***                      /                           /                    ***/
/***                  B--C                        B--C                     ***/
/***                 /                                \                    ***/
/***                A                                  A                   ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   [abcd]ptr:   the coordinates for atoms A, B, C, or D                ***/
/***   [abcd]frc:   the forces for atoms A, B, C, or D                     ***/
/***   hcom:        the dihedral command, giving details on all Fourier    ***/
/***                series terms and a flag to tell whether this is an     ***/
/***                improper or a standard dihedral                        ***/
/***   hdef:        the array of dihedral definitions                      ***/
/***=======================================================================***/
#if NEEDFORCE == 1
#if NEEDENERGY == 1
void DIHECALC(double *aptr, double *bptr, double *cptr, double *dptr,
	      double *afrc, double *bfrc, double *cfrc, double *dfrc,
	      dihecomm *hcom, prmtop *tp, coord *crd, FrcTab *Cfrc,
	      FrcTab *Hfrc, Energy *sysUV)
#else
void DIHECALC(double *aptr, double *bptr, double *cptr, double *dptr,
	      double *afrc, double *bfrc, double *cfrc, double *dfrc,
	      dihecomm *hcom, prmtop *tp, coord *crd, FrcTab *Cfrc,
	      FrcTab *Hfrc)
#endif
#else
void DIHECALC(double *aptr, double *bptr, double *cptr, double *dptr,
	      dihecomm *hcom, prmtop *tp, coord *crd, FrcTab *Cfrc,
	      FrcTab *Hfrc, Energy *sysUV)
#endif
{
  int i, idx;
  double theta, costheta, fa, fb1, fb2, fc1, fc2, fd, isinc2;
#if NEEDFORCE == 1
  double fr;
#endif
  double mgab, mgbc, mgcd, invab, invbc, invcd, cosb, cosc, isinb2;
  double invabc, invbcd, sangle;
  double ab[3], bc[3], cd[3], crabbc[3], crbccd[3], scr[3];
  dihedef *hdef;

  /*** Compue displacements ***/
  for (i = 0; i < 3; i++) {
    ab[i] = bptr[i] - aptr[i];
    bc[i] = cptr[i] - bptr[i];
    cd[i] = dptr[i] - cptr[i];
  }

  /*** Reimage ***/
  if (crd->isortho == 1) {
    OrthoReim(&ab[0], &ab[1], &ab[2], crd->hgdim);
    OrthoReim(&bc[0], &bc[1], &bc[2], crd->hgdim);
    OrthoReim(&cd[0], &cd[1], &cd[2], crd->hgdim);
  }
  else {
    NonOrthoReim(&ab[0], &ab[1], &ab[2], &crd->U, &crd->invU);
    NonOrthoReim(&bc[0], &bc[1], &bc[2], &crd->U, &crd->invU);
    NonOrthoReim(&cd[0], &cd[1], &cd[2], &crd->U, &crd->invU);
  }

  CrossP(ab, bc, crabbc);
  CrossP(bc, cd, crbccd);
  costheta = crabbc[0]*crbccd[0] + crabbc[1]*crbccd[1] + crabbc[2]*crbccd[2];
  costheta /=
    sqrt((crabbc[0]*crabbc[0] + crabbc[1]*crabbc[1] + crabbc[2]*crabbc[2]) *
	 (crbccd[0]*crbccd[0] + crbccd[1]*crbccd[1] + crbccd[2]*crbccd[2]));
  CrossP(crabbc, crbccd, scr);
  costheta = (costheta < -1.0) ? -1.0 : (costheta > 1.0) ? 1.0 : costheta;
  if (scr[0]*bc[0] + scr[1]*bc[1] + scr[2]*bc[2] > 0.0) {
    theta = acos(costheta);
  }
  else {
    theta = -acos(costheta);
  }

  /*** Compute the magnitude of the force and accumulate the energy ***/
#if NEEDFORCE == 1
  fr = 0.0;
#endif
  hdef = tp->HParam;
  for (i = 0; i < hcom->nt; i++) {
    idx = hcom->t[i];
    sangle = hdef[idx].N*theta - hdef[idx].Phi;
#if NEEDFORCE == 1
    fr += hdef[idx].K * hdef[idx].N * sin(sangle);
#endif
#if NEEDENERGY == 1
    sysUV->dihe += hdef[idx].K * (1.0 + cos(sangle));
#endif
  }

#if NEEDFORCE == 1
  /*** Other pre-computations ***/
  mgab = sqrt(ab[0]*ab[0] + ab[1]*ab[1] + ab[2]*ab[2]);
  invab = 1.0/mgab;
  mgbc = sqrt(bc[0]*bc[0] + bc[1]*bc[1] + bc[2]*bc[2]);
  invbc = 1.0/mgbc;
  mgcd = sqrt(cd[0]*cd[0] + cd[1]*cd[1] + cd[2]*cd[2]);
  invcd = 1.0/mgcd;
  cosb = -(ab[0]*bc[0] + ab[1]*bc[1] + ab[2]*bc[2])*invab*invbc;
  isinb2 = (cosb*cosb < 0.9999) ? 1.0/(1.0 - cosb*cosb) : 0.0;
  cosc = -(bc[0]*cd[0] + bc[1]*cd[1] + bc[2]*cd[2])*invbc*invcd;
  isinc2 = (cosc*cosc < 0.9999) ? 1.0/(1.0 - cosc*cosc) : 0.0;
  isinb2 *= fr;
  isinc2 *= fr;
  invabc = invab*invbc;
  invbcd = invbc*invcd;
  for (i = 0; i < 3; i++) {
    crabbc[i] *= invabc;
    crbccd[i] *= invbcd;
  }

  /*** Transform the dihedral forces to cartesian coordinates ***/
  fa = -invab * isinb2;
  fb1 = (mgbc - mgab*cosb) * invabc * isinb2;
  fb2 = cosc * invbc * isinc2;
  fc1 = (mgbc - mgcd*cosc) * invbcd * isinc2;
  fc2 = cosb * invbc * isinb2;
  fd = -invcd * isinc2;
  for (i = 0; i < 3; i++) {
    afrc[i] += crabbc[i] * fa;
    bfrc[i] += fb1 * crabbc[i] - fb2 * crbccd[i];
    cfrc[i] += -fc1 * crbccd[i] + fc2 * crabbc[i];
    dfrc[i] += -fd * crbccd[i];
  }
#endif

  /*** This segment is temporarily commented out because dihedrals make ***/
  /*** no contribution to the system virial, and because I'm not sure   ***/
  /*** exactly what the formulae for each component of the tensor are.  ***/
  /*** Accumulate the stress tensor ***/
  //  for (i = 0; i < 3; i++) {
  //    for (j = 0; j < 3; j++) {
  //      sysUV->Vir[3*i+j] = DiheStressTensorAB(ab, bc, cd, i, j, theta);
  //    }
  //  }

  /*** Evaluate 1-4 interactions ***/
  if (hcom->eval14 == 1) {

    int lja, ljb, irc, irh;
    double dx, dy, dz, r2, invr, invr2, invr4, qq, ljabA, ljabB;
#if NEEDFORCE == 1
    double fmag;
#endif
    CSpln *CdSD, *HdSD, *CSD, *HSD;

    /*** Compute the displacement ***/
    dx = dptr[0] - aptr[0];
    dy = dptr[1] - aptr[1];
    dz = dptr[2] - aptr[2];

    /*** Reimage ***/
    if (crd->isortho == 1) {
      OrthoReim(&dx, &dy, &dz, crd->hgdim);
    }
    else {
      NonOrthoReim(&dx, &dy, &dz, &crd->U, &crd->invU);
    }
    r2 = dx*dx + dy*dy + dz*dz;
    qq = tp->Charges[hcom->a]*tp->Charges[hcom->d];
    invr2 = 1.0/r2;
    if (fabs(qq) > 1.0e-8) {
      invr = sqrt(invr2);
#if NEEDFORCE == 1
      fmag = -tp->elec14fac*qq*invr2*invr;
#endif
#if NEEDENERGY == 1
      sysUV->delec += tp->elec14fac*qq*invr;
#endif
      if (r2 < MINNB2) {
	irc = r2*Cfrc->ivdr;
	CdSD = Cfrc->dSD;
	CSD = Cfrc->SD;
#if NEEDFORCE == 1
	fmag -= qq*(((CdSD[irc].A*r2 + CdSD[irc].B)*r2 + CdSD[irc].C)*r2 +
		    CdSD[irc].D);
#endif
#if NEEDENERGY == 1
	sysUV->delec -= qq*(((CSD[irc].A*r2 + CSD[irc].B)*r2 + CSD[irc].C)*r2 +
			    CSD[irc].D);
#endif
	irh = r2*Hfrc->ivdr;
	HdSD = Hfrc->dSD;
	HSD = Hfrc->SD;
#if NEEDFORCE == 1
	fmag += qq*(((HdSD[irh].A*r2 + HdSD[irh].B)*r2 + HdSD[irh].C)*r2 +
		    HdSD[irh].D);
#endif
#if NEEDENERGY == 1
	sysUV->delec += qq*(((HSD[irh].A*r2 + HSD[irh].B)*r2 + HSD[irh].C)*r2 +
			    HSD[irh].D);
#endif
      }
    }
#if NEEDFORCE == 1
    else {
      fmag = 0.0;
    }
#endif
    lja = tp->LJIdx[hcom->a];
    ljb = tp->LJIdx[hcom->d];
    if (lja >= 0 && ljb >= 0) {
      invr4 = invr2*invr2;
      ljabA = tp->LJftab.map[lja][2*ljb];
      ljabB = tp->LJftab.map[lja][2*ljb+1];
#if NEEDFORCE == 1
      fmag += tp->lj14fac*invr4*invr4*(ljabA*invr4*invr2 + ljabB);
#endif
      ljabA = tp->LJutab.map[lja][2*ljb];
      ljabB = tp->LJutab.map[lja][2*ljb+1];
#if NEEDENERGY == 1
      sysUV->vdw12 += tp->lj14fac*invr4*invr4*invr4*ljabA;
      sysUV->vdw6 += tp->lj14fac*invr4*invr2*ljabB;
#endif
    }
#if NEEDFORCE == 1
    dfrc[0] -= fmag*dx;
    dfrc[1] -= fmag*dy;
    dfrc[2] -= fmag*dz;
    afrc[0] += fmag*dx;
    afrc[1] += fmag*dy;
    afrc[2] += fmag*dz;

#if NEEDVIRIAL == 1
#ifdef IDEBUG
    /*** Accumulate the virial specific to 1-4 interactions ***/
    sysUV->OFVir[0] += dx*fmag*dx;
    sysUV->OFVir[1] += dx*fmag*dy;
    sysUV->OFVir[2] += dx*fmag*dz;
    sysUV->OFVir[3] += dy*fmag*dx;
    sysUV->OFVir[4] += dy*fmag*dy;
    sysUV->OFVir[5] += dy*fmag*dz;
    sysUV->OFVir[6] += dz*fmag*dx;
    sysUV->OFVir[7] += dz*fmag*dy;
    sysUV->OFVir[8] += dz*fmag*dz;
#endif

    /*** Accumulate the virial ***/
    sysUV->Vir[0] += dx*fmag*dx;
    sysUV->Vir[1] += dx*fmag*dy;
    sysUV->Vir[2] += dx*fmag*dz;
    sysUV->Vir[3] += dy*fmag*dx;
    sysUV->Vir[4] += dy*fmag*dy;
    sysUV->Vir[5] += dy*fmag*dz;
    sysUV->Vir[6] += dz*fmag*dx;
    sysUV->Vir[7] += dz*fmag*dy;
    sysUV->Vir[8] += dz*fmag*dz;
#endif
#endif
  }
}

#undef BONDCALC
#undef ANGLCALC
#undef DIHECALC
