#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "mdgxVector.h"
#include "Bonded.h"
#include "CrdManip.h"

#include "TopologyDS.h"
#include "TrajectoryDS.h"

/***=======================================================================***/
/*** ListBondedIntr: compute all bonded interactions for atoms in this     ***/
/***                 system.  The atoms are arranged in a single, large    ***/
/***                 list.                                                 ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:    the coordinates                                             ***/
/***   tp:     the topology                                                ***/
/***   Cfrc:   the "coarse" electrostatic force/energy spline table        ***/
/***   Hfrc:   the "fine" electrostatic force/energy spline table          ***/
/***   sysUV:  the system energy and virial (state information)            ***/
/***=======================================================================***/
void ListBondedIntr(coord *crd, prmtop *tp, FrcTab *Cfrc, FrcTab *Hfrc,
		    Energy *sysUV)
{
  int i, j;
  double *loctmp, *frctmp, *atmcrd, *atmfrc;
  bondlist *tblc;
  bondcomm *tbcom;
  angllist *talc;
  anglcomm *tacom;
  dihelist *thlc;
  dihecomm *thcom;

  loctmp = crd->loc;
  frctmp = crd->frc;
  for (i = 0; i < tp->natom; i++) {
    atmcrd = &loctmp[3*i];
    atmfrc = &frctmp[3*i];

    /*** Bonds ***/
    if (tp->BLC[i].nbond > 0) {
      tblc = &tp->BLC[i];
      for (j = 0; j < tblc->nbond; j++) {
	tbcom = &tblc->BC[j];
        if (sysUV->updateU == 2) {
	  BondVir(atmcrd, &loctmp[3*tbcom->b], atmfrc, &frctmp[3*tbcom->b],
		  tbcom, tp, crd, Cfrc, Hfrc, sysUV);
        }
	else if (sysUV->updateU == 1) {
	  BondFrcNrg(atmcrd, &loctmp[3*tbcom->b], atmfrc, &frctmp[3*tbcom->b],
		     tbcom, tp, crd, Cfrc, Hfrc, sysUV);
	}
	else if (sysUV->updateU == -1) {
	  BondNrg(atmcrd, &loctmp[3*tbcom->b], tbcom, tp, crd, Cfrc, Hfrc,
		  sysUV);
	}
	else {
	  BondFrc(atmcrd, &loctmp[3*tbcom->b], atmfrc, &frctmp[3*tbcom->b],
		  tbcom, tp, crd, Cfrc, Hfrc);
	}
      }
    }

    /*** Angles ***/
    if (tp->ALC[i].nangl > 0) {
      talc = &tp->ALC[i];
      for (j = 0; j < talc->nangl; j++) {
	tacom = &talc->AC[j];
	if (sysUV->updateU == 2) {
	  AnglVir(&loctmp[3*tacom->a], atmcrd, &loctmp[3*tacom->c],
		  &frctmp[3*tacom->a], atmfrc, &frctmp[3*tacom->c],
		  tacom, tp, crd, Cfrc, Hfrc, sysUV);
	}
	else if (sysUV->updateU == 1) {
	  AnglFrcNrg(&loctmp[3*tacom->a], atmcrd, &loctmp[3*tacom->c],
		     &frctmp[3*tacom->a], atmfrc, &frctmp[3*tacom->c],
		     tacom, tp, crd, Cfrc, Hfrc, sysUV);
	}
	else if (sysUV->updateU == -1) {
	  AnglNrg(&loctmp[3*tacom->a], atmcrd, &loctmp[3*tacom->c],
		  tacom, tp, crd, Cfrc, Hfrc, sysUV);
	}
	else {
	  AnglFrc(&loctmp[3*tacom->a], atmcrd, &loctmp[3*tacom->c],
		  &frctmp[3*tacom->a], atmfrc, &frctmp[3*tacom->c],
		  tacom, tp, crd, Cfrc, Hfrc);
	}
      }
    }

    /*** Dihedrals ***/
    if (tp->HLC[i].ndihe > 0) {
      thlc = &tp->HLC[i];
      for (j = 0; j < thlc->ndihe; j++) {
	thcom = &thlc->HC[j];
	if (sysUV->updateU == 2) {
	  DiheVir(&loctmp[3*thcom->a], atmcrd, &loctmp[3*thcom->c],
		  &loctmp[3*thcom->d], &frctmp[3*thcom->a], atmfrc,
		  &frctmp[3*thcom->c], &frctmp[3*thcom->d], thcom, tp, crd,
		  Cfrc, Hfrc, sysUV);
	}
	else if (sysUV->updateU == 1) {
	  DiheFrcNrg(&loctmp[3*thcom->a], atmcrd, &loctmp[3*thcom->c],
		     &loctmp[3*thcom->d], &frctmp[3*thcom->a], atmfrc,
		     &frctmp[3*thcom->c], &frctmp[3*thcom->d], thcom, tp, crd,
		     Cfrc, Hfrc, sysUV);
	}
	else if (sysUV->updateU == -1) {
	  DiheNrg(&loctmp[3*thcom->a], atmcrd, &loctmp[3*thcom->c],
		  &loctmp[3*thcom->d], thcom, tp, crd, Cfrc, Hfrc, sysUV);
	}
	else if (sysUV->updateU == 0) {
	  DiheFrc(&loctmp[3*thcom->a], atmcrd, &loctmp[3*thcom->c],
		  &loctmp[3*thcom->d], &frctmp[3*thcom->a], atmfrc,
		  &frctmp[3*thcom->c], &frctmp[3*thcom->d], thcom, tp, crd,
		  Cfrc, Hfrc);
	}
      }
    }
  }
}

/***=======================================================================***/
/*** CellBondedIntr: compute bonded interactions for a cell decomposition  ***/
/***                 of atom coordinates.                                  ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   C:      the cell of interest                                        ***/
/***   tp:     the topology                                                ***/
/***   Cfrc:   the "coarse" electrostatic force/energy spline table        ***/
/***   Hfrc:   the "fine" electrostatic force/energy spline table          ***/
/***   sysUV:  the system energy and virial (state information)            ***/
/***=======================================================================***/
void CellBondedIntr()
{
  
}

/*** BLOCKS 1 and 2: the forces are required ***/
#define NEEDFORCE 1

/*** BLOCK 1: energy (and virial) are not needed ***/
#define NEEDENERGY 0
#define NEEDVIRIAL 0
#include "BondContrib.c"
#undef NEEDVIRIAL
#undef NEEDENERGY

/*** BLOCK 2a: energy (but not virial) is needed, ***/
/***           in addition to forces              ***/
#define NEEDENERGY 1
#define NEEDVIRIAL 0
#include "BondContrib.c"
#undef NEEDVIRIAL

/*** BLOCK 2b: energy, virial, and forces are all needed ***/
#define NEEDVIRIAL 1
#include "BondContrib.c"
#undef NEEDVIRIAL
#undef NEEDENERGY

#undef NEEDFORCE
/*** END BLOCKS 1 and 2 ***/

/*** BLOCK 3: energy alone, but not forces or virials, are needed ***/
#define NEEDFORCE 0
#define NEEDENERGY 1
#define NEEDVIRIAL 0
#include "BondContrib.c"
#undef NEEDVIRIAL
#undef NEEDENERGY
#undef NEEDFORCE
/*** END BLOCK 3 ***/
