#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "CellManip.h"
#include "mdgxVector.h"
#include "Matrix.h"
#include "Nonbonded.h"

#include "CrdManipDS.h"
#include "pmeDirectDS.h"
#include "CellManipDS.h"
#include "TopologyDS.h"
#include "CompFrcDS.h"
#include "TrajectoryDS.h"

/***=======================================================================***/
/*** CreateCell: this routine creates a cell, the basic unit of input for  ***/
/***             the tower and plate or triple2 methods.                   ***/
/***=======================================================================***/
cell CreateCell(int maxatom)
{
  int i;
  cell C;

  C.maxatom = maxatom;
  C.nr = (int*)malloc(8*sizeof(int));
  C.nsr = (int*)malloc(8*sizeof(int));
  C.vel = (double*)malloc(3*maxatom*sizeof(double));
  C.ordr = CreateImat(8, maxatom);
  C.data = (atomc*)malloc(8*maxatom*sizeof(atomc));
  C.map = (atomc**)malloc(8*sizeof(atomc*));
  for (i = 0; i < 8; i++) {
    C.map[i] = &C.data[i*maxatom];
  }
  C.import = (atomb*)malloc(maxatom*sizeof(atomb));
  C.export = (atomb*)malloc(maxatom*sizeof(atomb));
  C.Vimport = (atombv*)malloc(maxatom*sizeof(atombv));
  C.Vexport = (atombv*)malloc(maxatom*sizeof(atombv));

  return C;
}

/***=======================================================================***/
/*** ExtendCell: this routine extends a cell, in the event that it must    ***/
/***             store more atoms than it was initially designed for.      ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   C:  the cell                                                        ***/
/***=======================================================================***/
void ExtendCell(cell *C, int maxatom)
{
  int i, j, omax;

  omax = C->maxatom;
  C->vel = (double*)realloc(C->vel, 3*maxatom*sizeof(double));
  C->data = (atomc*)realloc(C->data, 8*maxatom*sizeof(atomc));
  for (i = 7; i >= 0; i--) {
    for (j = 0; j < omax; j++) {
      C->data[i*maxatom+j] = C->data[i*omax+j];
    }
    C->map[i] = &C->data[i*maxatom];
  }
  ReallocImat(&C->ordr, 8, maxatom);
  C->import = (atomb*)realloc(C->import, maxatom*sizeof(atomb));
  C->export = (atomb*)realloc(C->export, maxatom*sizeof(atomb));
  C->Vimport = (atombv*)realloc(C->Vimport, maxatom*sizeof(atombv));
  C->Vexport = (atombv*)realloc(C->Vexport, maxatom*sizeof(atombv));
  C->maxatom = maxatom;
}

/***=======================================================================***/
/*** TakeCellGridDims: compute the cell grid dimensions based on direct    ***/
/***                   space cutoffs and the box dimensions.               ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   cdim:   the dimensions of the cell grid (a 3-element vector)        ***/
/***   crd:    the coordinates                                             ***/
/***   dcinp:  the direct space parameters                                 ***/
/***=======================================================================***/
void TakeCellGridDims(int* cdim, coord *crd, dircon *dcinp)
{
  int i;
  double xyz0[3], thx[3], thy[3], thz[3], ic1[3], ic2[3], ic3[3];
  dmat *invU;

  /*** Determine cell dimensions and the number of cells.  This uses the ***/
  /*** Hessian normal form to compute the thickness of each cell (the    ***/
  /*** distance between two cell faces), which works generally for       ***/
  /*** non-orthorhombic as well as orthorhombic simulation boxes.        ***/
  invU = &crd->invU;
  xyz0[0] = invU->data[0] + invU->data[1] + invU->data[2];
  xyz0[1] = invU->data[4] + invU->data[5];
  xyz0[2] = invU->data[8];
  for (i = 0; i < 3; i++) {
    ic1[i] = invU->map[i][0];
    ic2[i] = invU->map[i][1];
    ic3[i] = invU->map[i][2];
  }
  CrossP(ic2, ic3, thx);
  CrossP(ic1, ic3, thy);
  CrossP(ic1, ic2, thz);
  Normalize(thx, 3);
  Normalize(thy, 3);
  Normalize(thz, 3);
  cdim[0] = floor(fabs(DotP(thx, xyz0, 3)))/dcinp->Mcut;
  cdim[1] = floor(fabs(DotP(thy, xyz0, 3)))/dcinp->Mcut;
  cdim[2] = floor(fabs(DotP(thz, xyz0, 3)))/dcinp->Mcut;
}

/***=======================================================================***/
/*** CreateCellGrid: this routine creates a cell grid, for storing atoms   ***/
/***                 that will be involved in direct space interactions.   ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:    the coordinates                                             ***/
/***   dcinp:  the direct space parameters                                 ***/
/***   CG:     the cell grid                                               ***/
/***=======================================================================***/
cellgrid CreateCellGrid(coord *crd, dircon *dcinp)
{
  int i, j, k, maxatom;
  int cdim[3];
  cellgrid CG;

  /*** Compute the cell grid dimensions and allocate the cell grid ***/
  TakeCellGridDims(cdim, crd, dcinp);
  for (i = 0; i < 3; i++) {
    CG.ng[i] = cdim[i];
    CG.celldim[i] = crd->gdim[i]/CG.ng[i];
  }
  CG.data = (cell*)malloc(CG.ng[0]*CG.ng[1]*CG.ng[2]*sizeof(cell));

  /*** Determine a suitable number of atoms that each cell can be ***/
  /*** expected to hold.  Note that this number can be increased  ***/
  /*** later in the simulation if needed.                         ***/
  maxatom = 2*crd->natom/(CG.ng[0]*CG.ng[1]*CG.ng[2]);
  maxatom = 8*(maxatom/8 + 1);

  /*** Allocate each cell ***/
  CG.map = (cell***)malloc(CG.ng[0]*sizeof(cell**));
  for (i = 0; i < CG.ng[0]; i++) {
    CG.map[i] = (cell**)malloc(CG.ng[1]*sizeof(cell*));
    for (j = 0; j < CG.ng[1]; j++) {
      CG.map[i][j] = &CG.data[(i*CG.ng[1] + j)*CG.ng[2]];
      for (k = 0; k < CG.ng[2]; k++) {
	CG.map[i][j][k] = CreateCell(maxatom);
      }
    }
  }
  for (i = 0; i < CG.ng[0]*CG.ng[1]*CG.ng[2]; i++) {
    CG.data[i].CPUrank = 0;
  }

  return CG;
}

/***=======================================================================***/
/*** DestroyCell: this routine destroys a cell.                            ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   C:  the cell                                                        ***/
/***=======================================================================***/
void DestroyCell(cell *C)
{
  free(C->nr);
  free(C->nsr);
  free(C->vel);
  free(C->map);
  free(C->data);
  free(C->import);
  free(C->export);
  free(C->Vimport);
  free(C->Vexport);
  DestroyImat(&C->ordr);
}

/***=======================================================================***/
/*** DestroyCellGrid: this routine destroys a cell grid.                   ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   CG:  the cell grid                                                  ***/
/***=======================================================================***/
void DestroyCellGrid(cellgrid *CG)
{
  int i;

  for (i = 0; i < CG->ng[0]*CG->ng[1]*CG->ng[2]; i++) {
    DestroyCell(&CG->data[i]);
  }
  for (i = 0; i < CG->ng[0]; i++) {
    free(CG->map[i]);
  }
  free(CG->map);
  free(CG->data);
}

/***=======================================================================***/
/*** ComputeCellOrigins: compute the origin of each cell, to act as a      ***/
/***                     reference point during cell-cell transfers and    ***/
/***                     other operations.                                 ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   CG:     the cell grid                                               ***/
/***   crd:    the coordinates                                             ***/
/***=======================================================================***/
static void ComputeCellOrigins(cellgrid *CG, coord *crd)
{
  int i, j, k, nx, ny, nz;
  double mmx, mmy, mmz, invnx, invny, invnz;
  double *utmp;
  cell** ctm2p;
  cell* ctmp;

  utmp = crd->invU.data;
  nx = CG->ng[0];
  ny = CG->ng[1];
  nz = CG->ng[2];

  /*** The orthorhombic case ***/
  if (crd->isortho == 1) {
    for (i = 0; i < nx; i++) {
      ctm2p = CG->map[i];
      mmx = i*CG->celldim[0];
      for (j = 0; j < ny; j++) {
	ctmp = ctm2p[j];
	mmy = j*CG->celldim[1];
	for (k = 0; k < nz; k++) {
	  mmz = k*CG->celldim[2];
	  ctmp[k].orig[0] = mmx;
	  ctmp[k].orig[1] = mmy;
	  ctmp[k].orig[2] = mmz;
	}
      }
    }
    return;
  }

  /*** The non-orthorhombic case ***/
  invnx = 1.0/nx;
  invny = 1.0/ny;
  invnz = 1.0/nz;
  for (i = 0; i < nx; i++) {
    ctm2p = CG->map[i];
    mmx = i*invnx;
    for (j = 0; j < ny; j++) {
      ctmp = ctm2p[j];
      mmy = j*invny;
      for (k = 0; k < nz; k++) {
	mmz = k*invnz;
	ctmp[k].orig[0] = utmp[0]*mmx + utmp[1]*mmy + utmp[2]*mmz;
	ctmp[k].orig[1] = utmp[3]*mmx + utmp[4]*mmy + utmp[5]*mmz;
	ctmp[k].orig[2] = utmp[6]*mmx + utmp[7]*mmy + utmp[8]*mmz;
      }
    }
  }
}

/***=======================================================================***/
/*** AtomsToCellsOrtho: import of atoms into cells for orthorhombic boxes. ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:    the coordinates                                             ***/
/***   dcinp:  the direct space parameters                                 ***/
/***   CG:     the cell grid                                               ***/
/***=======================================================================***/
void AtomsToCellsOrtho(coord *crd, dircon *dcinp, cellgrid *CG, prmtop *tp)
{
  int h, i, j, nmaxatm;
  int nc[3];
  double celldim[3], gdim[3], invgdim[3], invcelldim[3];
  double *dtmp, *vtmp;
  cell *C;

  for (i = 0; i < 3; i++) {
    gdim[i] = crd->gdim[i];
    celldim[i] = gdim[i]/floor(gdim[i]*dcinp->invMcut);
    CG->celldim[i] = celldim[i];
    invcelldim[i] = 1.0/celldim[i];
    invgdim[i] = 1.0/gdim[i];
  }
  ComputeCellOrigins(CG, crd);

  /*** Set the number of atoms in all cells to zero ***/
  for (i = 0; i < CG->ng[0]*CG->ng[1]*CG->ng[2]; i++) {
    for (j = 0; j < 8; j++) {
      CG->data[i].nr[j] = 0;
    }
  }

  /*** Loop over all atoms and assign each to a cell ***/
  for (h = 0; h < crd->natom; h++) {
    dtmp = &crd->loc[3*h];
    vtmp = &crd->vel[3*h];
    for (i = 0; i < 3; i++) {
      dtmp[i] = dtmp[i]*invgdim[i];
      dtmp[i] = gdim[i]*(dtmp[i] - floor(dtmp[i]));
      nc[i] = dtmp[i]*invcelldim[i];
      if (nc[i] >= CG->ng[i] && dtmp[i]*invcelldim[i] - 1.0e-8 < CG->ng[i]) {
	nc[i] = dtmp[i]*invcelldim[i] - 1.0e-8;
      }
      if (nc[i] >= CG->ng[i] || nc[i] < 0) {
	printf("Oops!  Atom %d assigned to cell[%d][%d]!  Range %d -> %d.\n",
	       h, i, nc[i], 0, CG->ng[i]);
	printf("Atom coordinates: [ %9.4lf %9.4lf %9.4lf ]\n", dtmp[0],
	       dtmp[1], dtmp[2]);
      }
    }
    C = &CG->map[nc[0]][nc[1]][nc[2]];
    j = C->nr[0];
    C->map[0][j].id = h;
    C->map[0][j].lj = tp->LJIdx[h];
    C->map[0][j].q = tp->Charges[h];
    for (i = 0; i < 3; i++) {
      C->map[0][j].loc[i] = dtmp[i];
      C->vel[3*j+i] = vtmp[i];
      C->map[0][j].frc[i] = 0.0;
    }
    j++;
    C->nr[0] = j;
    if (j >= C->maxatom) {

      /*** Extend all cells to be on the safe side ***/
      nmaxatm = j + 8;
      for (i = 0; i < CG->ng[0]*CG->ng[1]*CG->ng[2]; i++) {
	ExtendCell(&CG->data[i], nmaxatm);
      }
    }
  }
}

/***=======================================================================***/
/*** CopyCellContents: move all relevant coordinates and information from  ***/
/***                   one cell's primary sector to another cell.  Before  ***/
/***                   the transfer, this routine tests each of the atoms  ***/
/***                   in the original cell's primary sector to see if it  ***/
/***                   will be useful to the destination cell.  Only those ***/
/***                   atoms which are useful will be transferred to save  ***/
/***                   bandwidth.                                          ***/
/***                                                                       ***/
/***                   To prepare for computing direct space interactions, ***/
/***                   this routine will also re-image atoms if the cells  ***/
/***                   lie on the periphery of the simulation box.         ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   CG:     the cell grid                                               ***/
/***   O[ijk]: the indices of the original cell in the cell grid           ***/
/***   N[ijk]: the indices of the destination cell in the cell grid        ***/
/***   m[xyz]: the move that is to be made (all these parameters should    ***/
/***           have values of -1, 0, or perhaps 1--we move atoms between   ***/
/***           neighboring cells)                                          ***/
/***   dreg:   the destination region within the destination cell (values  ***/
/***           between 1 and 7--0 is the cell's primary region, containing ***/
/***           atoms whose coordinates it will integrate)                  ***/
/***   Mcut:   the maximum direct space cutoff (electrostatic and van-der  ***/
/***           Waals forces)                                               ***/
/***   crd:    the coordinates                                             ***/
/***=======================================================================***/
void CopyCellContents(cellgrid *CG, int Oi, int Oj, int Ok, int Ni, int Nj,
		      int Nk, int mx, int my, int mz, int dreg, double Mcut,
		      coord *crd, prmtop *tp)
{
  int h, i, tcrb;
  double dx, dy, dz, reimx, reimy, reimz;
  cell *cO, *cN;
  atomc *tatm;

  /*** Determine if we need to re-image atoms as we transfer them ***/
  reimx = 0.0;
  reimy = 0.0;
  reimz = 0.0;
  if (crd->isortho == 1) {
    if (mx == -1 && Oi == 0) {
      reimx += crd->gdim[0];
    }
    else if (mx == 1 && Oi == CG->ng[0]-1) {
      reimx -= crd->gdim[0];
    }
    if (my == -1 && Oj == 0) {
      reimy += crd->gdim[1];
    }
    else if (my == 1 && Oj == CG->ng[1]-1) {
      reimy -= crd->gdim[1];
    }
    if (mz == -1 && Ok == 0) {
      reimz += crd->gdim[2];
    }
    else if (mz == 1 && Ok == CG->ng[2]-1) {
      reimz -= crd->gdim[2];
    }
  }
  else {
    printf("CopyCellContents >> Error.  Not ready for non-orthorhombic unit "
	   "cells!\n");
    exit(1);
  }

  /*** Pointers to the cells that will share atom data ***/
  cO = &CG->map[Oi][Oj][Ok];
  cN = &CG->map[Ni][Nj][Nk];

  /*** If both cells are owned by the same processor, just set a pointer ***/
  if (cO->CPUrank == cN->CPUrank &&
      (abs(Oi - Ni) <= 1 && CG->ng[0] > 2) &&
      (abs(Oj - Nj) <= 1 && CG->ng[1] > 2) &&
      (abs(Ok - Nk) <= 1 && CG->ng[2] > 2)) {
    cN->map[dreg] = cO->data;
    cN->nr[dreg] = cO->nr[0];
    return;
  }

  /*** Loop over all atoms in cO's primary sector and transfer ***/
  cO->nexp = 0;
  tatm = cO->map[0];
  for (h = 0; h < cO->nr[0]; h++) {

    /*** Determine the displacements from the plane of interest ***/
    dx = 0.0;
    dy = 0.0;
    dz = 0.0;
    if (crd->isortho == 1) {
      if (mx == -1) {
        dx = tatm[h].loc[0] - cO->orig[0];
	if (dx > Mcut) continue;
      }
      else if (mx == 1) {
        dx = cO->orig[0] + CG->celldim[0] - tatm[h].loc[0];
	if (dx > Mcut) continue;
      }
      if (my == -1) {
        dy = tatm[h].loc[1] - cO->orig[1];
	if (dy > Mcut) continue;
      }
      else if (my == 1) {
        dy = cO->orig[1] + CG->celldim[1] - tatm[h].loc[1];
	if (dy > Mcut) continue;
      }
      if (mz == -1) {
        dz = tatm[h].loc[2] - cO->orig[2];
	if (dz > Mcut) continue;
      }
      else if (mz == 1) {
        dz = cO->orig[2] + CG->celldim[2] - tatm[h].loc[2];
	if (dz > Mcut) continue;
      }
    }
    else {
      printf("CopyCellContents >> Error.  Not yet ready to do "
	     "non-orthorhombic boxes.\n");
      exit(1);
    }

    /*** If we're still here, this atom is a keeper.  Put it in the ***/
    /*** cell's buffer to ready it for transfer.                    ***/
    tcrb = cO->nexp;
    cO->export[tcrb].id = tatm[h].id;
    cO->export[tcrb].loc[0] = tatm[h].loc[0] + reimx;
    cO->export[tcrb].loc[1] = tatm[h].loc[1] + reimy;
    cO->export[tcrb].loc[2] = tatm[h].loc[2] + reimz;
    cO->nexp = tcrb+1;
  }

  /*** Copy the buffer over to the destination cell ***/
  for (h = 0; h < cO->nexp; h++) {
    cN->import[h] = cO->export[h];
  }
  cN->nimp = cO->nexp;

  /*** Add details of the newly imported atoms in the destination cell. ***/
  /*** At this point, work is done exclusively in cN, so the following  ***/
  /*** code could be made into its own function for efficient parallel  ***/
  /*** execution.                                                       ***/
  tatm = cN->map[dreg];
  for (h = 0; h < cN->nimp; h++) {
    tatm[h].id = cN->import[h].id;
    for (i = 0; i < 3; i++) {
      tatm[h].loc[i] = cN->import[h].loc[i];
      tatm[h].frc[i] = 0.0;
    }
    tatm[h].lj = tp->LJIdx[tatm[h].id];
    tatm[h].q = tp->Charges[tatm[h].id];
  }
  cN->nr[dreg] = cN->nimp;
}

/***=======================================================================***/
/*** ZeroCellForces: zero forces on atoms in the primary sectors of all    ***/
/***                 cells.                                                ***/
/***=======================================================================***/
static void ZeroCellForces(cellgrid *CG)
{
  int i, j;
  cell *C;

  for (i = 0; i < CG->ng[0]*CG->ng[1]*CG->ng[2]; i++) {
    C = &CG->data[i];
    for (j = 0; j < C->nr[0]; j++) {
      C->data[j].frc[0] = 0.0;
      C->data[j].frc[1] = 0.0;
      C->data[j].frc[2] = 0.0;
    }
  }
}

/***=======================================================================***/
/*** ShareTriple2: have all cells share their primary sector contents with ***/
/***               with neighboring cells.  This prepares for direct space ***/
/***               interactions via the triple2 method.                    ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   CG:   the cell grid                                                 ***/
/***   L:    the maximum direct space cutoff (electrostatic and van-der    ***/
/***         Waals forces)                                                 ***/
/***   crd:  the coordinates                                               ***/
/***=======================================================================***/
void ShareTriple2(cellgrid *CG, double L, coord *crd, prmtop *tp)
{
  int i, j, k, it, jt, kt;

  /*** Zero forces on all cells' primary sectors ***/
  ZeroCellForces(CG);

  /*** Loop over all cells ***/
  for (i = 0; i < CG->ng[0]; i++) {
    it = (i == 0) ? CG->ng[0]-1 : i-1;
    for (j = 0; j < CG->ng[1]; j++) {
      jt = (j == 0) ? CG->ng[1]-1 : j-1;
      for (k = 0; k < CG->ng[2]; k++) {
	kt = (k == 0) ? CG->ng[2]-1 : k-1;
	CopyCellContents(CG, i, j, k, it, j, k, -1, 0, 0, 1, L, crd, tp);
	CopyCellContents(CG, i, j, k, i, jt, k, 0, -1, 0, 2, L, crd, tp);
	CopyCellContents(CG, i, j, k, it, jt, k, -1, -1, 0, 3, L, crd, tp);
	CopyCellContents(CG, i, j, k, i, j, kt, 0, 0, -1, 4, L, crd, tp);
	CopyCellContents(CG, i, j, k, it, j, kt, -1, 0, -1, 5, L, crd, tp);
	CopyCellContents(CG, i, j, k, i, jt, kt, 0, -1, -1, 6, L, crd, tp);
	CopyCellContents(CG, i, j, k, it, jt, kt, -1, -1, -1, 7, L, crd, tp);
      }
    }
  }
}

/***=======================================================================***/
/*** CompactOrder: this function concatenates the orderings of atoms in    ***/
/***               one of four sectors with those in other sectors.        ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tA:    the number of atoms already accumulated in the list          ***/
/***   cordr: the padded list, containing gaps so that various sectors can ***/
/***          hold a potentially large number of atoms                     ***/
/***   cnsr:  the number of atoms in this segment                          ***/
/***   astr:  the starting point of the sector in cordr                    ***/
/***=======================================================================***/
static void CompactOrder(int* cordr, int astr, int bstr, int n)
{
  int i;

  for (i = 0; i < n; i++) {
    cordr[astr+i] = cordr[bstr+i];
  }
}

#define VGTE 1
#include "Ordering.c"
#undef VGTE

#define VGTE 0
#include "Ordering.c"
#undef VGTE

/***=======================================================================***/
/*** TranslateCell: translate atoms in a cell to satisfy re-imaging needs. ***/
/***=======================================================================***/
void TranslateCell(cellgrid *CG, int ipos, int jpos, int kpos, int itrg,
		   int jtrg, int ktrg, double frwbck)
{
  int h;
  double reimx, reimy, reimz;
  cell *C;

  /*** Decide on the translation ***/
  reimx = 0.0;
  reimy = 0.0;
  reimz = 0.0;
  if (ipos == CG->ng[0]-1 && itrg == 0) {
    reimx = -CG->ng[0]*CG->celldim[0];
  }
  else if (ipos == 0 && itrg == CG->ng[0]-1) {
    reimx = CG->ng[0]*CG->celldim[0];
  }
  if (jpos == CG->ng[1]-1 && jtrg == 0) {
    reimy = -CG->ng[1]*CG->celldim[1];
  }
  else if (jpos == 0 && jtrg == CG->ng[1]-1) {
    reimy = CG->ng[1]*CG->celldim[1];
  }
  if (kpos == CG->ng[2]-1 && ktrg == 0) {
    reimz = -CG->ng[2]*CG->celldim[2];
  }
  else if (kpos == 0 && ktrg == CG->ng[2]-1) {
    reimz = CG->ng[2]*CG->celldim[2];
  }

  /*** Translate ***/
  reimx *= frwbck;
  reimy *= frwbck;
  reimz *= frwbck;
  C = &CG->map[ipos][jpos][kpos];
  for (h = 0; h < C->nr[0]; h++) {
    C->data[h].loc[0] += reimx;
    C->data[h].loc[1] += reimy;
    C->data[h].loc[2] += reimz;
  }
}

/***=======================================================================***/
/*** ManageIntr: after partitioning atoms in each of two sectors into four ***/
/***             different sub-regions, we can exploit the known distances ***/
/***             between each sub-region to reduce the number of squared   ***/
/***             distances that must be evaluated.                         ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***=======================================================================***/
void ManageIntr(cell *C, int sA, int sB, FrcTab *EFrc, dircon *dcinp,
		prmtop *tp, Energy *sysUV)
{
  int i, cstr;
  int *cordrA, *cordrB, *cnsr;

  /*** Set pointers ***/
  cordrA = C->ordr.map[0];
  cordrB = C->ordr.map[4];
  cnsr = C->nsr;

  /*** A1:B(1,2,3,4), A2:B(1,2,3), A3:B(1,2), A4:B1 ***/
  for (i = 0; i < 4; i++) {
    cstr = (i == 0) ? 0 : cnsr[i-1];

    /*** Compute energies, virials, and forces ***/
    if (sysUV->updateU == 2) {
      if (dcinp->Vcut > dcinp->Ecut) {
	if (i*0.25*dcinp->Vcut >= MINNB) {
	  CompIntrVgtEnrgvir(C->map[sA], C->map[sB], &cordrA[cstr],
			     &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i],
			     EFrc, dcinp, tp, sysUV);
	}
	else {
	  CompIntrVgtECRnrgvir(C->map[sA], C->map[sB], &cordrA[cstr],
			       &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i],
			       EFrc, dcinp, tp, sysUV);
	}
      }
      else {
	if (i*0.25*dcinp->Vcut >= MINNB) {
	  CompIntrVeqEnrgvir(C->map[sA], C->map[sB], &cordrA[cstr],
			     &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i],
			     EFrc, dcinp, tp, sysUV);
	}
	else {
	  CompIntrVeqECRnrgvir(C->map[sA], C->map[sB], &cordrA[cstr],
			       &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i],
			       EFrc, dcinp, tp, sysUV);
	}
      }
      continue;
    }

    /*** Compute energies and forces ***/
    else if (sysUV->updateU == 1) {
      if (dcinp->Vcut > dcinp->Ecut) {
	if (i*0.25*dcinp->Vcut >= MINNB) {
	  CompIntrVgtEnrg(C->map[sA], C->map[sB], &cordrA[cstr],
			  &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i], EFrc,
			  dcinp, tp, sysUV);
	}
	else {
	  CompIntrVgtECRnrg(C->map[sA], C->map[sB], &cordrA[cstr],
			    &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i],
			    EFrc, dcinp, tp, sysUV);
	}
      }
      else {
	if (i*0.25*dcinp->Vcut >= MINNB) {
	  CompIntrVeqEnrg(C->map[sA], C->map[sB], &cordrA[cstr],
			  &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i], EFrc,
			  dcinp, tp, sysUV);
	}
	else {
	  CompIntrVeqECRnrg(C->map[sA], C->map[sB], &cordrA[cstr],
			    &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i],
			    EFrc, dcinp, tp, sysUV);
	}
      }
      continue;
    }

    /*** Compute energies alone ***/
    else if (sysUV->updateU == -1) {
      if (dcinp->Vcut > dcinp->Ecut) {
	if (i*0.25*dcinp->Vcut >= MINNB) {
	  CompIntrVgtEnrgx(C->map[sA], C->map[sB], &cordrA[cstr],
			   &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i],
			   EFrc, dcinp, tp, sysUV);
	}
	else {
	  CompIntrVgtECRnrgx(C->map[sA], C->map[sB], &cordrA[cstr],
			     &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i],
			     EFrc, dcinp, tp, sysUV);
	}
      }
      else {
	if (i*0.25*dcinp->Vcut >= MINNB) {
	  CompIntrVeqEnrgx(C->map[sA], C->map[sB], &cordrA[cstr],
			   &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i],
			   EFrc, dcinp, tp, sysUV);
	}
	else {
	  CompIntrVeqECRnrgx(C->map[sA], C->map[sB], &cordrA[cstr],
			     &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i],
			     EFrc, dcinp, tp, sysUV);
	}
      }
      continue;
    }

    /*** If we're still here, only forces are necessary ***/
    else {
      if (dcinp->Vcut > dcinp->Ecut) {
        if (i*0.25*dcinp->Vcut >= MINNB) {
	  CompIntrVgtE(C->map[sA], C->map[sB], &cordrA[cstr],
		       &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i], EFrc,
		       dcinp, tp);
	}
	else {
	  CompIntrVgtECR(C->map[sA], C->map[sB], &cordrA[cstr],
			 &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i], EFrc,
			 dcinp, tp);
	}
      }
      else {
	if (i*0.25*dcinp->Vcut >= MINNB) {
	  CompIntrVeqE(C->map[sA], C->map[sB], &cordrA[cstr],
		       &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i], EFrc,
		       dcinp, tp);
	}
	else {
	  CompIntrVeqECR(C->map[sA], C->map[sB], &cordrA[cstr],
			 &cordrB[i*C->maxatom], cnsr[i]-cstr, cnsr[7-i], EFrc,
			 dcinp, tp);
	}
      }
    }
  }
}

/***=======================================================================***/
/*** DirectTriple2: perform a direct space nonbonded calculation for all   ***/
/***                atoms in this cell using the "triple2," a.k.a. the     ***/
/***                "1/8th Shell" method.                                  ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   C:       the cell                                                   ***/
/***   celldim: the dimensions of the cell (taken from the overarching     ***/
/***            cellgrid structure)                                        ***/
/***   EFrc:    the electrostatic force spline table (also contains spline ***/
/***            coefficients for the potential, if desired)                ***/
/***   dcinp:   the direct space input parameters                          ***/
/***   tp:      the topology (passed in for rare cases in which we must    ***/
/***            refer to the nonbonded exclusion list to determine whether ***/
/***            an interaction should actually be computed)                ***/
/***=======================================================================***/
void DirectTriple2(cell *C, double* celldim, FrcTab *EFrc, dircon *dcinp,
		   prmtop *tp, Energy *sysUV)
{
  int i;

  /*** First, do interactions within the primary sector/region ***/
  for (i = 0; i < C->nr[0]; i++) {
    C->ordr.data[i] = i;
  }
  if (dcinp->Vcut > dcinp->Ecut) {
    if (sysUV->updateU == 2) {
      CompIntrVgtECRnrgvir(C->data, C->data, C->ordr.data, C->ordr.data,
			   C->nr[0], C->nr[0], EFrc, dcinp, tp, sysUV);
    }
    else if (sysUV->updateU == 1) {
      CompIntrVgtECRnrg(C->data, C->data, C->ordr.data, C->ordr.data,
			C->nr[0], C->nr[0], EFrc, dcinp, tp, sysUV);
    }
    else if (sysUV->updateU == -1) {
      CompIntrVgtECRnrgx(C->data, C->data, C->ordr.data, C->ordr.data,
			 C->nr[0], C->nr[0], EFrc, dcinp, tp, sysUV);
    }
    else {
      CompIntrVgtECR(C->data, C->data, C->ordr.data, C->ordr.data,
		     C->nr[0], C->nr[0], EFrc, dcinp, tp);
    }

    /*** Next, do interactions across faces (3) ***/
    OrderOrthoVgtE(C, celldim, 0, 0, dcinp, tp);
    ManageIntr(C, 0, 1, EFrc, dcinp, tp, sysUV);
    OrderOrthoVgtE(C, celldim, 1, 0, dcinp, tp);
    ManageIntr(C, 0, 2, EFrc, dcinp, tp, sysUV);
    OrderOrthoVgtE(C, celldim, 2, 0, dcinp, tp);
    ManageIntr(C, 0, 4, EFrc, dcinp, tp, sysUV);

    /*** Next, do interactions across edges (6) ***/
    OrderOrthoVgtE(C, celldim, 0, 1, dcinp, tp);
    ManageIntr(C, 0, 6, EFrc, dcinp, tp, sysUV);
    OrderOrthoVgtE(C, celldim, 1, 1, dcinp, tp);
    ManageIntr(C, 0, 5, EFrc, dcinp, tp, sysUV);
    OrderOrthoVgtE(C, celldim, 2, 1, dcinp, tp);
    ManageIntr(C, 0, 3, EFrc, dcinp, tp, sysUV);
    OrderOrthoVgtE(C, celldim, 0, 2, dcinp, tp);
    ManageIntr(C, 4, 2, EFrc, dcinp, tp, sysUV);
    OrderOrthoVgtE(C, celldim, 1, 2, dcinp, tp);
    ManageIntr(C, 4, 1, EFrc, dcinp, tp, sysUV);
    OrderOrthoVgtE(C, celldim, 2, 3, dcinp, tp);
    ManageIntr(C, 2, 1, EFrc, dcinp, tp, sysUV);

    /*** Finally, do interactions across the center point (4) ***/
    OrderOrthoVgtE(C, celldim, 0, 4, dcinp, tp);
    ManageIntr(C, 0, 7, EFrc, dcinp, tp, sysUV);
    OrderOrthoVgtE(C, celldim, 1, 4, dcinp, tp);
    ManageIntr(C, 1, 6, EFrc, dcinp, tp, sysUV);
    OrderOrthoVgtE(C, celldim, 2, 4, dcinp, tp);
    ManageIntr(C, 2, 5, EFrc, dcinp, tp, sysUV);
    OrderOrthoVgtE(C, celldim, 3, 4, dcinp, tp);
    ManageIntr(C, 3, 4, EFrc, dcinp, tp, sysUV);
  }
  else {
    if (sysUV->updateU == 2) {
      CompIntrVeqECRnrgvir(C->data, C->data, C->ordr.data, C->ordr.data,
			   C->nr[0], C->nr[0], EFrc, dcinp, tp, sysUV);
    }
    else if (sysUV->updateU == 1) {
      CompIntrVeqECRnrg(C->data, C->data, C->ordr.data, C->ordr.data,
			C->nr[0], C->nr[0], EFrc, dcinp, tp, sysUV);
    }
    else if (sysUV->updateU == -1) {
      CompIntrVeqECRnrgx(C->data, C->data, C->ordr.data, C->ordr.data,
			 C->nr[0], C->nr[0], EFrc, dcinp, tp, sysUV);
    }
    else {
      CompIntrVeqECR(C->data, C->data, C->ordr.data, C->ordr.data,
		     C->nr[0], C->nr[0], EFrc, dcinp, tp);
    }

    /*** Next, do interactions across faces (3) ***/
    OrderOrthoVeqE(C, celldim, 0, 0, dcinp, tp);
    ManageIntr(C, 0, 1, EFrc, dcinp, tp, sysUV);
    OrderOrthoVeqE(C, celldim, 1, 0, dcinp, tp);
    ManageIntr(C, 0, 2, EFrc, dcinp, tp, sysUV);
    OrderOrthoVeqE(C, celldim, 2, 0, dcinp, tp);
    ManageIntr(C, 0, 4, EFrc, dcinp, tp, sysUV);

    /*** Next, do interactions across edges (6) ***/
    OrderOrthoVeqE(C, celldim, 0, 1, dcinp, tp);
    ManageIntr(C, 0, 6, EFrc, dcinp, tp, sysUV);
    OrderOrthoVeqE(C, celldim, 1, 1, dcinp, tp);
    ManageIntr(C, 0, 5, EFrc, dcinp, tp, sysUV);
    OrderOrthoVeqE(C, celldim, 2, 1, dcinp, tp);
    ManageIntr(C, 0, 3, EFrc, dcinp, tp, sysUV);
    OrderOrthoVeqE(C, celldim, 0, 2, dcinp, tp);
    ManageIntr(C, 4, 2, EFrc, dcinp, tp, sysUV);
    OrderOrthoVeqE(C, celldim, 1, 2, dcinp, tp);
    ManageIntr(C, 4, 1, EFrc, dcinp, tp, sysUV);
    OrderOrthoVeqE(C, celldim, 2, 3, dcinp, tp);
    ManageIntr(C, 2, 1, EFrc, dcinp, tp, sysUV);

    /*** Finally, do interactions across the center point (4) ***/
    OrderOrthoVeqE(C, celldim, 0, 4, dcinp, tp);
    ManageIntr(C, 0, 7, EFrc, dcinp, tp, sysUV);
    OrderOrthoVeqE(C, celldim, 1, 4, dcinp, tp);
    ManageIntr(C, 1, 6, EFrc, dcinp, tp, sysUV);
    OrderOrthoVeqE(C, celldim, 2, 4, dcinp, tp);
    ManageIntr(C, 2, 5, EFrc, dcinp, tp, sysUV);
    OrderOrthoVeqE(C, celldim, 3, 4, dcinp, tp);
    ManageIntr(C, 3, 4, EFrc, dcinp, tp, sysUV);
  }
}

/***=======================================================================***/
/*** ImportCellForces: import forces from neighboring cells and accumulate ***/
/***                   the totals on atoms in each cell's primary sector.  ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   CG:     the cell grid                                               ***/
/***   O[ijk]: the indices of the original cell in the cell grid           ***/
/***   N[ijk]: the indices of the destination cell in the cell grid        ***/
/***   m[xyz]: the move that is to be made (all these parameters should    ***/
/***           have values of -1, 0, or perhaps 1--we move atoms between   ***/
/***           neighboring cells)                                          ***/
/***   dreg:   the destination region within the destination cell (values  ***/
/***           between 1 and 7--0 is the cell's primary region, containing ***/
/***           atoms whose coordinates it will integrate)                  ***/
/***=======================================================================***/
void ImportCellForces(cellgrid *CG, int Oi, int Oj, int Ok, int Ni, int Nj,
                      int Nk, int mx, int my, int mz, int dreg)
{
  int h, i, j;
  atomc *tatm;
  atomb *tbuf, *obuf;
  cell *cO, *cN;

  /*** Set pointers to the two cells of interest ***/
  cO = &CG->map[Oi][Oj][Ok];
  cN = &CG->map[Ni][Nj][Nk];
  if (cO->CPUrank == cN->CPUrank &&
      abs(Oi - Ni) <= 1 && abs(Oj - Nj) <= 1 && abs(Ok - Nk) <= 1 &&
      CG->ng[0] > 2 && CG->ng[1] > 2 && CG->ng[2] > 2) {
    return;
  }
  if (cN->nr[dreg] == 0) {
    return;
  }

  /*** Loop over atoms in cN's destination sector, prepare the buffer. ***/
  tatm = cN->map[dreg];
  tbuf = cN->export;
  for (h = 0; h < cN->nr[dreg]; h++) {
    tbuf[h].id = tatm[h].id;
    for (i = 0; i < 3; i++) {
      tbuf[h].loc[i] = tatm[h].frc[i];
    }
  }
  cN->nexp = cN->nr[dreg];

  /*** Transfer the buffer ***/
  obuf = cO->import;
  for (h = 0; h < cN->nr[dreg]; h++) {
    obuf[h] = tbuf[h];
  }
  cO->nimp = cN->nexp;

  /*** Accumulate forces on atoms in cO's primary sector ***/
  i = 0;
  tatm = cO->map[0];
  for (h = 0; h < cO->nr[0]; h++) {
    if (tatm[h].id == obuf[i].id) {
      for (j = 0; j < 3; j++) {
	tatm[h].frc[j] += obuf[i].loc[j];
      }
      i++;
      if (i == cO->nimp) {
	break;
      }
    }
  }

  /*** Check ***/
  if (i != cN->nr[dreg]) {
    printf("ImportCellForces >> Error.  Found %d of %d atoms in cO "
	   "sector 0.\n", i, cN->nr[dreg]);
  }
}

/***=======================================================================***/
/*** MergeTriple2: merge the results of a force calculation.  After this   ***/
/***               is done, the processors should be ready to integrate    ***/
/***               the equations of motion for particles in each cell's    ***/
/***               primary sector.                                         ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   CG:   the cell grid                                                 ***/
/***=======================================================================***/
void MergeTriple2(cellgrid *CG)
{
  int i, j, k, idt, jdt, kdt;

  for (i = 0; i < CG->ng[0]; i++) {
    idt = (i == 0) ? CG->ng[0]-1 : i-1;
    for (j = 0; j < CG->ng[1]; j++) {
      jdt = (j == 0) ? CG->ng[1]-1 : j-1;
      for (k = 0; k < CG->ng[2]; k++) {
        kdt = (k == 0) ? CG->ng[2]-1 : k-1;
        ImportCellForces(CG, i, j, k, idt, j, k, -1, 0, 0, 1);
        ImportCellForces(CG, i, j, k, i, jdt, k, 0, -1, 0, 2);
        ImportCellForces(CG, i, j, k, idt, jdt, k, -1, -1, 0, 3);
        ImportCellForces(CG, i, j, k, i, j, kdt, 0, 0, -1, 4);
        ImportCellForces(CG, i, j, k, idt, j, kdt, -1, 0, -1, 5);
        ImportCellForces(CG, i, j, k, i, jdt, kdt, 0, -1, -1, 6);
        ImportCellForces(CG, i, j, k, idt, jdt, kdt, -1, -1, -1, 7);
      }
    }
  }
}

/***=======================================================================***/
/*** MapCellsToAtoms: this function, which should not be called during     ***/
/***                  typical MD steps, takes forces from the atom lists   ***/
/***                  in cells and maps them back to the ordered list of   ***/
/***                  atoms.  Calling this function OVERWRITES coordinates ***/
/***                  and forces in the ordered list with the values from  ***/
/***                  the cell grid.                                       ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   CG:   the cell grid                                                 ***/
/***   crd:  the coordinates structure (the ordered list of atoms)         ***/
/***=======================================================================***/
void MapCellsToAtoms(cellgrid *CG, coord *crd)
{
  int i, j, k, aid;
  cell *C;

  for (i = 0; i < CG->ng[0]*CG->ng[1]*CG->ng[2]; i++) {
    C = &CG->data[i];
    for (j = 0; j < C->nr[0]; j++) {
      aid = 3*C->data[j].id;
      for (k = 0; k < 3; k++) {
	crd->loc[aid+k] = C->data[j].loc[k];
	crd->frc[aid+k] = C->data[j].frc[k];
      }
    }
  }
}

/***=======================================================================***/
/*** ReleaseAtoms: when atoms move out of a cell, they are released to     ***/
/***               other cells.  This could require that each cell         ***/
/***               communicate with twenty-six nearest neighbors, but by   ***/
/***               releasing atoms sequentially in X, Y, and Z there are   ***/
/***               only six messages per cell.                             ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   C:       the original cell                                          ***/
/***   vv:      the direction along which to release (+X = 1, +Y = 2,      ***/
/***              +Z = 3, -X = 1, -Y = 2, -Z = 3)                          ***/
/***   celldim: the dimensions of the cell                                 ***/
/***=======================================================================***/
static void ReleaseAtoms(cell *C, int vv, double* celldim)
{
  int i, j, nexp, mflag;
  double clim;
  double *vtmp;
  atomc *atmp;
  atombv *vexp;

  /*** Determine the cell limits ***/
  if (vv < 0) {
    vv *= -1;
    vv -= 1;
    mflag = -1;
    clim = C->orig[vv];
  }
  else {
    vv -= 1;
    mflag = 1;
    clim = C->orig[vv] + celldim[vv];
  }

  /*** Loop over all atoms in the cell: which ones have moved (-)? ***/
  atmp = C->data;
  vexp = C->Vexport;
  vtmp = C->vel;
  nexp = 0;
  for (i = 0; i < C->nr[0]; i++) {
    if ((mflag == -1 && atmp[i].loc[vv] < clim) ||
	(mflag == 1 && atmp[i].loc[vv] >= clim)) {
      vexp[nexp].id = atmp[i].id;
      for (j = 0; j < 3; j++) {
	vexp[nexp].loc[j] = atmp[i].loc[j];
	vexp[nexp].vel[j] = vtmp[3*i+j];
      }
      nexp++;

      /*** Mark this atom as having left ***/
      atmp[i].id = -1;
    }
  }
  C->nexp = nexp;
}

/***=======================================================================***/
/*** MigrateAtoms: this function performs the transfer of atoms between    ***/
/***               two cells.  It is called after ReleaseAtoms (above).    ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   CG:      the cell grid                                              ***/
/***   c[ijk]:  the location of the cell in the cell grid                  ***/
/***   vv:      the direction along which to release (+X = 1, +Y = 2,      ***/
/***              +Z = 3, -X = 1, -Y = 2, -Z = 3)                          ***/
/***   tp:      the topology                                               ***/
/***=======================================================================***/
static void MigrateAtoms(cellgrid *CG, int* cijk, int vv, prmtop *tp)
{
  int i, j, k, mflag, nplaced, maxfill;
  int dijk[3];
  int *ljtmp;
  cell *C, *D;
  double *qtmp, *vtmp;
  atomc *tatm;
  atombv *vimp, *vexp;

  C = &CG->map[cijk[0]][cijk[1]][cijk[2]];
  dijk[0] = cijk[0];
  dijk[1] = cijk[1];
  dijk[2] = cijk[2];
  if (vv < 0) {
    mflag = -1;
    vv *= -1;
  }
  else {
    mflag = 1;
  }
  vv -= 1;
  if (mflag == -1 && cijk[vv] == 0) {
    dijk[vv] = CG->ng[vv] - 1;
  }
  else if (mflag == 1 && cijk[vv] == CG->ng[vv] - 1) {
    dijk[vv] = 0;
  }
  else {
    dijk[vv] = cijk[vv] + mflag;
  }
  D = &CG->map[dijk[0]][dijk[1]][dijk[2]];

  /*** Transfer the buffer from C to D ***/
  vimp = D->Vimport;
  vexp = C->Vexport;
  for (i = 0; i < C->nexp; i++) {
    vimp[i] = vexp[i];
  }
  D->nimp = C->nexp;

  /*** Add atoms to D's regular buffer and add details from topology.  ***/
  /*** From this point onward, work is done exclusively in the D cell. ***/
  tatm = D->data;
  vtmp = D->vel;
  qtmp = tp->Charges;
  ljtmp = tp->LJIdx;

  /*** Overwrite atoms that have migrated out of D ***/
  nplaced = 0;
  maxfill = -1;
  for (i = 0; i < D->nr[0]; i++) {
    if (tatm[i].id == -1) {
      if (nplaced < D->nimp) {
	tatm[i].id = vimp[nplaced].id;
	tatm[i].q = qtmp[vimp[nplaced].id];
	tatm[i].lj = ljtmp[vimp[nplaced].id];
	for (j = 0; j < 3; j++) {
	  tatm[i].loc[j] = vimp[nplaced].loc[j];
	  vtmp[3*i+j] = vimp[nplaced].vel[j];
	}
	nplaced++;
      }
      else {
	break;
      }
    }
  }
  maxfill = i;

  /*** Add the remaining atoms ***/
  if (nplaced < D->nimp) {
    k = D->nr[0];
    for (i = nplaced; i < D->nimp; i++) {
      tatm[k].id = vimp[i].id;
      tatm[k].q = qtmp[vimp[i].id];
      tatm[k].lj = ljtmp[vimp[i].id];
      for (j = 0; j < 3; j++) {
	tatm[k].loc[j] = vimp[i].loc[j];
	vtmp[3*k+j] = vimp[i].vel[j];
      }
      k++;
    }
    D->nr[0] = k;
  }

  /*** Scour any -1 atom ID numbers ***/
  else {
    j = maxfill;
    for (i = maxfill+1; i < D->nr[0]; i++) {
      if (tatm[i].id >= 0) {
	tatm[j] = tatm[i];
	vtmp[3*j] = vtmp[3*i];
	vtmp[3*j+1] = vtmp[3*i+1];
	vtmp[3*j+2] = vtmp[3*i+2];
	j++;
      }
    }
    D->nr[0] = j;
  }
}

/***=======================================================================***/
/*** UpdateCells: this function wraps the ReleaseAtoms and MigrateAtoms    ***/
/***              functions (above).                                       ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   CG:      the cell grid                                              ***/
/***   tp:      the topology                                               ***/
/***=======================================================================***/
void UpdateCells(cellgrid *CG, prmtop *tp)
{
  int i, j, k, m;
  int cijk[3];

  for (m = 1; m < 4; m++) {
    for (i = 0; i < CG->ng[0]; i++) {
      cijk[0] = i;
      for (j = 0; j < CG->ng[1]; j++) {
	cijk[1] = j;
	for (k = 0; k < CG->ng[2]; k++) {
	  cijk[2] = k;
	  ReleaseAtoms(&CG->map[i][j][k], m, CG->celldim);
	  MigrateAtoms(CG, cijk, m, tp);
	  ReleaseAtoms(&CG->map[i][j][k], -m, CG->celldim);
	  MigrateAtoms(CG, cijk, -m, tp);
	}
      }
    }
  }
}
