#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "Grid.h"
#include "pmeRecip.h"
#include "mdgxVector.h"
#include "Constants.h"
#include "ChargeMap.h"
#include "Matrix.h"
#include "BSpline.h"

#include "BSplineDS.h"
#include "TopologyDS.h"

/***=======================================================================***/
/*** Charges2Book: map a set of atoms to a double-precision book for later ***/
/***               use in various Particle Mesh Ewald calculations.  Note  ***/
/***               that this routine will zero the book before mapping any ***/
/***               charges.                                                ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   pmmap:  the "particle<-->mesh map" created by SplCoeff (see library ***/
/***           BSpline)                                                    ***/
/***   tp:     the topology, used here to assign partial charges           ***/
/***   Q:      the charge mesh                                             ***/
/***=======================================================================***/
void Charges2Book(bmap *pmmap, reccon *rcinp, prmtop *tp, dbook *Q)
{
  int h, i, j, k, natom, ordrX, ordrY ,ordrZ;
  double spx, spxy, atmq;
  double *dtmp;
  double **dtm2p;
  bcof *xcf, *ycf, *zcf;

  /*** Shortcuts ***/
  natom = pmmap->natom;
  ordrX = rcinp->ordr[0];
  ordrY = rcinp->ordr[1];
  ordrZ = rcinp->ordr[2];

  /*** Zero all elements of the book ***/
  j = (Q->pfft == 1) ? 2*(Q->col/2+1) : Q->col;
  SetDVec(Q->data, Q->pag*Q->row*j, 0.0);

  /*** Map all the atoms ***/
  for (h = 0; h < natom; h++) {
    xcf = &pmmap->xcof[ordrX*h];
    ycf = &pmmap->ycof[ordrY*h];
    zcf = &pmmap->zcof[ordrZ*h];
    atmq = tp->Charges[h];

    for (i = 0; i < ordrX; i++) {
      dtm2p = Q->map[xcf[i].m];
      spx = atmq*xcf[i].s;
      for (j = 0; j < ordrY; j++) {
	dtmp = dtm2p[ycf[j].m];
	spxy = spx*ycf[j].s;
	for (k = 0; k < ordrZ; k++) {
	  dtmp[zcf[k].m] += spxy*zcf[k].s;
	}
      }
    }
  }
}

/***=======================================================================***/
/*** IntrpFrc: interpolate forces from the mesh to atoms.  Note that this  ***/
/***           routine will not zero forces, but only make contributions   ***/
/***           to whatever forces are already accumulated on each atom.    ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   pmmap:  the "particle<-->mesh map" created by SplCoeff (see library ***/
/***           BSpline)                                                    ***/
/***   crd:    the coordinates and forces structure (see library pmeDirect ***/
/***           and also cells <--> atoms mapping functions in library      ***/
/***           CellManip); also contains simulation box dimensions         ***/
/***   rcinp:  the reciprocal space input parameters                       ***/
/***   tp:     the topology, used here to assign partial charges           ***/
/***   U:      the potential mesh                                          ***/
/***=======================================================================***/
void IntrpFrc(bmap *pmmap, coord *crd, reccon *rcinp, prmtop *tp, dbook *U)
{
  int h, i, j, k, natom, ordrX, ordrY, ordrZ, isortho;
  double SPx, SPy, SPz, SPdx, SPxy, SPdxy, SPxdy, fx, fy, fz, uk, atmq;
  double invgx, invgy, invgz, ngx, ngy, ngz;
  double *utmp, *umd;
  double **utm2p;
  bcof *xcf, *ycf, *zcf;

  /*** Shortcuts ***/
  natom = crd->natom;
  ordrX = rcinp->ordr[0];
  ordrY = rcinp->ordr[1];
  ordrZ = rcinp->ordr[2];
  ngx = rcinp->ng[0];
  ngy = rcinp->ng[1];
  ngz = rcinp->ng[2];
  invgx = 1.0/crd->gdim[0];
  invgy = 1.0/crd->gdim[1];
  invgz = 1.0/crd->gdim[2];
  umd = crd->U.data;
  isortho = crd->isortho;

  /*** Loop over all atoms ***/
  for (h = 0; h < natom; h++) {
    xcf = &pmmap->xcof[ordrX*h];
    ycf = &pmmap->ycof[ordrY*h];
    zcf = &pmmap->zcof[ordrZ*h];
    atmq = tp->Charges[h];
    fx = 0.0;
    fy = 0.0;
    fz = 0.0;
    for (i = 0; i < ordrX; i++) {
      utm2p = U->map[xcf[i].m];
      SPx = atmq*xcf[i].s;
      SPdx = atmq*xcf[i].d;
      for (j = 0; j < ordrY; j++) {
	utmp = utm2p[ycf[j].m];
	SPy = ycf[j].s;
	SPxy = SPx*SPy;
	SPdxy = SPdx*SPy;
	SPxdy = SPx*ycf[j].d;
	for (k = 0; k < ordrZ; k++) {
	  uk = utmp[zcf[k].m];
	  SPz = zcf[k].s;
	  fx -= SPdxy*SPz*uk;
	  fy -= SPxdy*SPz*uk;
	  fz -= SPxy*zcf[k].d*uk;
	}
      }
    }

    /*** The total force must be normalized within the box ***/
    fx *= ngx;
    fy *= ngy;
    fz *= ngz;
    if (crd->isortho == 1) {
      crd->frc[3*h] += fx*invgx;
      crd->frc[3*h+1] += fy*invgy;
      crd->frc[3*h+2] += fz*invgz;
    }
    else {
      crd->frc[3*h] += umd[0]*fx + umd[3]*fy + umd[6]*fz;
      crd->frc[3*h+1] += umd[1]*fx + umd[4]*fy + umd[7]*fz;
      crd->frc[3*h+2] += umd[2]*fx + umd[5]*fy + umd[8]*fz;
    }
  }
}
