#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>
#include "Constants.h"
#include "Parse.h"
#include "Macros.h"
#include "Matrix.h"
#include "pmeDirect.h"
#include "Random.h"

#include "pmeDirectDS.h"
#include "pmeRecipDS.h"
#include "TrajectoryDS.h"
#include "TopologyDS.h"
#include "CrdManipDS.h"

/***=======================================================================***/
/*** GetDataFileNames: this function reads in the names of all necessary   ***/
/***                   data (input and output) for a molecular dynamics    ***/
/***                   run.  This routine also seeks "base" names and the  ***/
/***                   corresponding suffixes, so mdgx can write numerous  ***/
/***                   output files with the same base and suffix but      ***/
/***                   different numbers (i.e. trj101.crd, where the base  ***/
/***                   is "trj", the suffix is ".crd", and the number is   ***/
/***                   101).                                               ***/
/***                                                                       ***/
/*** Note that the initialization of file names is done in the             ***/
/*** CommandLineControl function.                                          ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:      the topology structure (stores its own source file)        ***/
/***   tj:      the trajectory structure (stores the restart file name, as ***/
/***              well as many other output file names)                    ***/
/***   inp:     the input file                                             ***/
/***=======================================================================***/
static void GetDataFileNames(prmtop *tp, trajcon *tj, FILE *inp)
{
  int collect, EXtp, EXeprules, EXinpcrd, EXforcedump, EXtrjbase, EXvelbase;
  int EXenebase, EXfrcbase, EXrstbase, EXoutbase;
  char line[MAXLINE];
  cmat lwords;

  /*** First, verify that certain variables which could have    ***/
  /*** been given on the command line are not already specified ***/
  EXtp = (strcmp(tp->source, "prmtop") != 0) ? 1 : 0;
  EXeprules = (tp->eprulesource[0] != '\0') ? 1 : 0;
  EXinpcrd = (strcmp(tj->ipcname, "inpcrd") != 0) ? 1 : 0;
  EXforcedump = (strcmp(tj->dumpname, "forcedump.dat") != 0) ? 1 : 0;
  EXtrjbase = (strcmp(tj->trjbase, "mdcrd") != 0) ? 1 : 0;
  EXvelbase = (strcmp(tj->velbase, "mdvel") != 0) ? 1 : 0;
  EXenebase = (strcmp(tj->enebase, "mden") != 0) ? 1 : 0;
  EXfrcbase = (strcmp(tj->frcbase, "mdfrc") != 0) ? 1 : 0;
  EXrstbase = (strcmp(tj->rstbase, "restrt") != 0) ? 1 : 0;
  EXoutbase = (strcmp(tj->outbase, "mdout") != 0) ? 1 : 0;

  collect = AdvanceToSegment(inp, "files");
  while (collect == 1) {

    /*** Read the next line ***/
    fgets(line, MAXLINE, inp);
    RemoveWhiteSpace(line, MAXLINE);
    RemoveComments(line);

    /*** Break if the line is "&end" ***/
    collect = DetectNamelistEnd(line, "GetDataFileNames");
    if (collect == 0) {
      continue;
    }

    /*** Eliminate and add spaces between special characters ***/
    /*** "=", "\n", and "," ***/
    NixCommaCarriage(line);
    EqualSpace(line);
    lwords = ParseWords(line);

    /*** List of arguments to search for ***/
    if (EXtp == 0) {
      SeekString(lwords, tp->source, "Topology", "-p");
    }
    if (EXeprules == 0) {
      SeekString(lwords, tp->eprulesource, "ExtraPointRules", "-xtp");
    }
    if (EXinpcrd == 0) {
      SeekString(lwords, tj->ipcname, "StartCoord", "-c");
    }
    if (EXforcedump == 0) {
      SeekString(lwords, tj->dumpname, "ForceDump", "-d");
    }
    if (EXtrjbase == 0) {
      SeekString(lwords, tj->trjbase, "CrdTrajBase", "-x");
    }
    if (EXvelbase == 0) {
      SeekString(lwords, tj->velbase, "VelTrajBase", "-v");
    }
    if (EXfrcbase == 0) {
      SeekString(lwords, tj->frcbase, "FrcTrajBase", "-f");
    }
    if (EXenebase == 0) {
      SeekString(lwords, tj->enebase, "EnergyBase", "-e");
    }
    if (EXrstbase == 0) {
      SeekString(lwords, tj->rstbase, "Restart", "-r");
    }
    if (EXrstbase == 0) {
      SeekString(lwords, tj->outbase, "OutputBase", "-o");
    }
    SeekString(lwords, tj->rsrptname, "ResReport", "-rrp");
    SeekString(lwords, tj->rstsuff, "RestartSuff", "-rsf");
    SeekString(lwords, tj->outsuff, "OutputSuff", "-osf");
    SeekString(lwords, tj->frcsuff, "FrcTrajSuff", "-fsf");
    SeekString(lwords, tj->enesuff, "EnergySuff", "-esf");
    SeekString(lwords, tj->velsuff, "VelTrajSuff", "-vsf");
    SeekString(lwords, tj->trjsuff, "CrdTrajSuff", "-xsf");

    /*** Free allocated memory ***/
    DestroyCmat(&lwords);
  }
}

/***=======================================================================***/
/*** GetTopolNamelist: this function reads in additional information about ***/
/***                   a topology that cannot be specified in the typical  ***/
/***                   AMBER topology format.  Essentially, information    ***/
/***                   provided in the &topol namelist alerts MDGX that    ***/
/***                   it's dealing with a non-standard topology, and to   ***/
/***                   look for certain things when it actually does read  ***/
/***                   in the topology file.                               ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:      the topology structure                                     ***/
/***=======================================================================***/
static void GetTopolNamelist(prmtop *tp, FILE *inp)
{
  int collect;
  char line[MAXLINE];
  cmat lwords;

  /*** Default settings ***/
  tp->ljbuck = 0;
  collect = AdvanceToSegment(inp, "cntrl");
  while (collect == 1) {

    /*** Read the next line ***/
    fgets(line, MAXLINE, inp);
    RemoveWhiteSpace(line, MAXLINE);
    RemoveComments(line);

    /*** Break if the line is "&end" ***/
    collect = DetectNamelistEnd(line, "GetCntrlNamelist");
    if (collect == 0) {
      continue;
    }

    /*** Eliminate and add spaces between special characters ***/
    /*** "=", "\n", and ","                                  ***/
    NixCommaCarriage(line);
    EqualSpace(line);
    lwords = ParseWords(line);

    /*** Information about the topology that cannot go in a ***/
    /*** standard AMBER prmtop file                         ***/
    SeekInt(lwords, &tp->ljbuck, "vdWstyle", "ljstyle");
    SeekInt(lwords, &tp->qshape, "ChargeShape", "qshape");
    SeekInt(lwords, &tp->ReadVirtualQ, "ReadVirtualQ", "virtualq");

    /*** Free allocated memory ***/
    DestroyCmat(&lwords);
  }
}

/***=======================================================================***/
/*** GetCntrlNamelist: this function reads in run parameters, as would be  ***/
/***                   found in the &cntrl namelist in PMEMD or SANDER     ***/
/***                   command input files.                                ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   dcinp:   direct space command information                           ***/
/***   rcinp:   reciprocal space command information                       ***/
/***   tp:      the topology structure                                     ***/
/***   tj:      the trajectory structure (stores the restart file name, as ***/
/***              well as many other output file names)                    ***/
/***   inp:     the input file                                             ***/
/***=======================================================================***/
static void GetCntrlNamelist(dircon *dcinp, reccon *rcinp, prmtop *tp,
			     trajcon *tj, FILE *inp)
{
  int collect;
  double EVcut;
  char line[MAXLINE];
  cmat lwords;

  /*** Default settings for dynamics / Hamiltonian ***/
  tp->lj14fac = 0.5;
  tp->elec14fac = 5.0/6.0;
  dcinp->Ecut = 8.0;
  dcinp->Vcut = 8.0;
  EVcut = -1.0;
  dcinp->LRvdw = 0;
  sprintf(tp->WaterName, "WAT ");
  tj->mode = 0;
  tj->starttime = 0.0;
  tj->dt = 0.001;
  tp->rattle = 0;
  tp->settle = 0;

  /*** Default settings for thermostat and barostat ***/
  tj->Ttarget = 298.0;
  tj->Tinit = -100.0;
  tj->Ptarget = 1.0;
  tp->rattletol = 1.0e-6;
  tj->RemoveMomentum = 0;
  tj->ioutfm = 0;
  tj->ntt = 0;
  tj->ntp = 0;
  tj->vrand = 1000;
  tj->BerendsenTCoupl = 0.4;
  tj->BerendsenPTime = 1.0;
  tj->BerendsenPCoupl = 44.6;
  tj->MCBarostatFac[0] = 2.0e-3;
  tj->MCBarostatFac[1] = -1.0;
  tj->MCBarostatFac[2] = -1.0;
  tj->MCBarostatFreq = 100;
  tj->npth.TauT = 1.0;
  tj->npth.TauP = 1.0;
  tj->igseed = 72177;

  /*** Default settings for trajectory control ***/
  tj->nstep = 1;
  tj->nfistep = 0;
  tj->currfi = 0;
  tj->ntwr = 0;
  tj->ntwx = 0;
  tj->ntwv = 0;
  tj->ntwf = 0;
  tj->ntwe = 0;
  tj->ntpr = 0;
  tj->irest = 0;
  tj->topchk = 1;

  /*** Initialize random number generator ***/
  tj->rndcon = -1;
  ran2(&tj->rndcon);

  collect = AdvanceToSegment(inp, "cntrl");
  while (collect == 1) {

    /*** Read the next line ***/
    fgets(line, MAXLINE, inp);
    RemoveWhiteSpace(line, MAXLINE);
    RemoveComments(line);

    /*** Break if the line is "&end" ***/
    collect = DetectNamelistEnd(line, "GetCntrlNamelist");
    if (collect == 0) {
      continue;
    }

    /*** Eliminate and add spaces between special characters ***/
    /*** "=", "\n", and ","                                  ***/
    NixCommaCarriage(line);
    EqualSpace(line);
    lwords = ParseWords(line);

    /*** Dynamics / Hamiltonian directives ***/
    SeekReal(lwords, &tp->lj14fac, "Vdw14Fac", "scnb");
    SeekReal(lwords, &tp->elec14fac, "Elec14Fac", "scee");
    SeekReal(lwords, &dcinp->Ecut, "ElecCut", "es_cutoff");
    SeekReal(lwords, &dcinp->Vcut, "VdwCut", "vdw_cutoff");
    SeekReal(lwords, &EVcut, "DirectCutoff", "cut");
    SeekReal(lwords, &tj->starttime, "StartTime", "t");
    SeekReal(lwords, &tj->dt, "TimeStep", "dt");
    SeekInt(lwords, &tj->irest, "RestartMD", "irest");
    SeekInt(lwords, &tp->rattle, "DoRATTLE", "rigidbond");
    SeekInt(lwords, &tp->settle, "DoSETTLE", "rigidwat");
    SeekInt(lwords, &tj->igseed, "RandomSeed", "ig");
    SeekInt(lwords, &tj->RemoveMomentum, "ZeroMomentum", "nscm");
    SeekInt(lwords, &tj->mode, "RunMode", "imin");

    /*** Thermostat and barostat directives ***/
    SeekInt(lwords, &tj->ntt, "Thermostat", "ntt");
    SeekInt(lwords, &tj->ntp, "Barostat", "ntp");
    SeekInt(lwords, &tj->vrand, "RandomReset", "vrand");
    SeekInt(lwords, &tj->MCBarostatFreq, "MCBarostatFrq", "mcbfrq");
    SeekReal(lwords, &tj->Ttarget, "Temperature", "temp0");
    SeekReal(lwords, &tj->Ptarget, "Pressure", "pres0");
    SeekReal(lwords, &tj->Tinit, "StartTemp", "tempi");
    SeekReal(lwords, &tj->BerendsenTCoupl, "BerendsenTC", "taup");
    SeekReal(lwords, &tj->BerendsenPCoupl, "BerendsenPC", "comp");
    SeekReal(lwords, &tj->npth.TauT, "HooverTC", "tauthv");
    SeekReal(lwords, &tj->npth.TauP, "HooverPC", "tauphv");
    SeekReal(lwords, &tj->MCBarostatFac[0], "MCBarostatPC", "mccomp");
    SeekReal(lwords, &tj->MCBarostatFac[0], "MCBarostatPCX", "mccompx");
    SeekReal(lwords, &tj->MCBarostatFac[1], "MCBarostatPCY", "mccompy");
    SeekReal(lwords, &tj->MCBarostatFac[2], "MCBarostatPCZ", "mccompz");

    /*** Output management directives ***/
    SeekLLInt(lwords, &tj->nstep, "StepCount", "nstlim");
    SeekLLInt(lwords, &tj->nfistep, "FileStepCount", "nfistep");
    SeekInt(lwords, &tj->ntpr, "WriteDiagnostics", "ntpr");
    SeekInt(lwords, &tj->ntwx, "WriteCoordinates", "ntwx");
    SeekInt(lwords, &tj->ntwv, "WriteVelocities", "ntwv");
    SeekInt(lwords, &tj->ntwf, "WriteForces", "ntwf");
    SeekInt(lwords, &tj->ntwe, "WriteEnergies", "ntwe");
    SeekInt(lwords, &tj->ntwr, "WriteRestart", "ntwr");
    SeekInt(lwords, &tj->topchk, "TopologyCheck", "tchk");
    SeekInt(lwords, &tj->ioutfm, "CoordFormat", "ioutfm");

    /*** Free allocated memory ***/
    DestroyCmat(&lwords);
  }

  /*** Restate control data in internal units/conventions ***/
  if (EVcut > 0.0) {
    dcinp->Ecut = EVcut;
    dcinp->Vcut = EVcut;
  }
  tp->lj14fac = tp->lj14fac - 1.0;
  tp->elec14fac = BIOQ*(tp->elec14fac - 1.0);
  dcinp->Mcut = MAX(dcinp->Ecut, dcinp->Vcut);
  dcinp->invMcut = 1.0/dcinp->Mcut;
  tj->rndcon = tj->igseed;
  tj->BerendsenPCoupl *= 1.0e-6;

  /*** Check input ***/
  if (dcinp->Vcut < dcinp->Ecut) {
    printf("GetCntrlNamelist >> Error.  Vdw cutoff (%9.4lf) must be >= "
	   "electrostatic\nGetCntrlNamelist >> cutoff (%9.4lf)\n.",
	   dcinp->Vcut, dcinp->Ecut);
    exit(1);
  }
  if (tp->rattle == 2) {
    printf("GetCntrlNamelist >> Error.  Constraining all bonds in not "
	   "supported.\nGetCntrlNamelist >> Set DoRATTLE/ntc = 0 or 1.\n");
    exit(1);
  }
  if (tj->nfistep > 0) {
    if (tj->nstep % tj->nfistep != 0) {
      printf("GetCntrlNamelist >> Warning.  nstlim (mod) nfistep is nonzero.\n"
	     "GetCntrlNamelist >> nfistep / FileStepCount set to zero.\n");
      tj->nfistep = 0;
    }
    if (tj->ntwr > tj->nfistep || tj->ntwe > tj->nfistep ||
	tj->ntwx > tj->nfistep || tj->ntwv > tj->nfistep ||
	tj->ntwf > tj->nfistep || tj->ntpr > tj->nfistep) {
      printf("GetCntrlNamelist >> Warning.  When the run output is split into "
	     "segments,\nGetCntrlNamelist >> ntwr, ntpr, ntwx, ntwv, ntwf, "
	     "and ntwe cannot exceed\nGetCntrlNamelist >> nfistep.\n");
      if (tj->ntwr > tj->nfistep) {
	tj->ntwr = tj->nfistep;
      }
      if (tj->ntwe > tj->nfistep) {
	tj->ntwe = tj->nfistep;
      }
      if (tj->ntwx > tj->nfistep) {
	tj->ntwx = tj->nfistep;
      }
      if (tj->ntwv > tj->nfistep) {
	tj->ntwv = tj->nfistep;
      }
      if (tj->ntwf > tj->nfistep) {
	tj->ntwf = tj->nfistep;
      }
      if (tj->ntpr > tj->nfistep) {
	tj->ntpr = tj->nfistep;
      }
    }
  }
}

/***=======================================================================***/
/*** GetEwaldNamelist: this function reads in Ewald parameters, as would   ***/
/***                   be found in the &ewald namelist in PMEMD or SANDER  ***/
/***                   command input files.                                ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   dcinp:   direct space command information                           ***/
/***   rcinp:   reciprocal space command information                       ***/
/***   inp:     the input file                                             ***/
/***=======================================================================***/
static void GetEwaldNamelist(dircon *dcinp, reccon *rcinp, FILE *inp)
{
  int i, collect, genordr;
  char line[MAXLINE];
  cmat lwords;

  /*** Default settings for SPME ***/
  rcinp->ng = (int*)malloc(3*sizeof(int));
  rcinp->ng[0] = -1;
  rcinp->ng[1] = -1;
  rcinp->ng[2] = -1;
  rcinp->ordr[0] = 4;
  rcinp->ordr[1] = 4;
  rcinp->ordr[2] = 4;
  dcinp->Dtol = 6.0e-6;
  dcinp->lkpspc = 0.0625;
  dcinp->ewcoeff = -1.0;
  rcinp->S = -1.0;
  genordr = -1;

  /*** Default settings for MLE ***/
  rcinp->ggordr = 8;
  rcinp->nlev = 1;
  rcinp->nslab = 1;
  rcinp->nstrip = 1;
  for (i = 0; i < 4; i++) {
    rcinp->cfac[i] = -1.0;
    rcinp->PadYZ[i] = -1;
  }
  rcinp->cfac[0] = 1.0;
  rcinp->PadYZ[0] = 1;

  /*** Default settings for MLE optimization ***/
  rcinp->OptMesh = 1;
  rcinp->ntrg = 4;
  rcinp->maxiter = 8;
  rcinp->MeshIterTol = 1.0e-5;
  rcinp->dg0 = 1.0e-3;
  rcinp->mindg = 1.0e-9;

  /*** Scan the input file ***/
  collect = AdvanceToSegment(inp, "ewald");
  while (collect == 1) {

    /*** Read the next line ***/
    fgets(line, MAXLINE, inp);
    RemoveWhiteSpace(line, MAXLINE);
    RemoveComments(line);

    /*** Break if the line is "&end" ***/
    collect = DetectNamelistEnd(line, "GetEwaldNamelist");
    if (collect == 0) {
      continue;
    }

    /*** Eliminate and add spaces between special characters ***/
    /*** "=", "\n", and ","                                  ***/
    NixCommaCarriage(line);
    EqualSpace(line);
    lwords = ParseWords(line);

    /*** PME directives ***/
    SeekReal(lwords, &dcinp->Dtol, "DSumTol", "dsum_tol");
    SeekReal(lwords, &dcinp->ewcoeff, "EwaldCof", "ew_coeff");
    SeekReal(lwords, &rcinp->S, "Sigma", "sigma");
    SeekReal(lwords, &dcinp->lkpspc, "SplnSpc", "eedtbdns");
    SeekInt(lwords, &rcinp->ng[0], "MeshDimX", "nfft1");
    SeekInt(lwords, &rcinp->ng[1], "MeshDimY", "nfft2");
    SeekInt(lwords, &rcinp->ng[2], "MeshDimZ", "nfft3");
    SeekInt(lwords, &rcinp->ordr[0], "OrderX", "ordr1");
    SeekInt(lwords, &rcinp->ordr[1], "OrderY", "ordr2");
    SeekInt(lwords, &rcinp->ordr[2], "OrderZ", "ordr3");
    SeekInt(lwords, &genordr, "Order", "order");

    /*** MLE directives ***/
    SeekInt(lwords, &rcinp->nlev, "EwaldLevels", "nlev");
    SeekInt(lwords, &rcinp->PadYZ[0], "Padding1", "lpad1");
    SeekInt(lwords, &rcinp->PadYZ[1], "Padding2", "lpad2");
    SeekInt(lwords, &rcinp->PadYZ[2], "Padding3", "lpad3");
    SeekReal(lwords, &rcinp->cfac[1], "Spread2", "cfac2");
    SeekReal(lwords, &rcinp->cfac[2], "Spread3", "cfac3");
    SeekReal(lwords, &rcinp->cfac[3], "Spread4", "cfac4");
    SeekInt(lwords, &rcinp->ggordr, "GridOrder", "ggordr");

    /*** MLE mesh optimization directives ***/
    SeekInt(lwords, &rcinp->OptMesh, "OptimizeMesh", "mopt");
    SeekInt(lwords, &rcinp->ntrg, "NTarget", "ntrg");
    SeekInt(lwords, &rcinp->maxiter, "Iterations", "iter");
    SeekReal(lwords, &rcinp->MeshIterTol, "OptTol", "mopt_tol");
    SeekReal(lwords, &rcinp->dg0, "OptStep0", "step0");
    SeekReal(lwords, &rcinp->mindg, "OptStepMin", "stepmin");

    /*** Free allocated memory ***/
    DestroyCmat(&lwords);
  }

  /*** If a general order is specified, apply it ***/
  if (genordr > 0) {
    rcinp->ordr[0] = genordr;
    rcinp->ordr[1] = genordr;
    rcinp->ordr[2] = genordr;
  }

  /*** Compute the Ewald coefficient and then the Gaussian spread ***/
  if (rcinp->S <= 0.0 && dcinp->ewcoeff <= 0.0) {
    dcinp->ewcoeff = EwaldCoefficient(dcinp->Ecut, dcinp->Dtol);
    rcinp->S = 0.5/dcinp->ewcoeff;
  }
  else {
    if (rcinp->S > 0.0) {
      dcinp->ewcoeff = 0.5/rcinp->S;
    }
    else {
      rcinp->S = 0.5/dcinp->ewcoeff;
    }
    dcinp->Dtol = (1.0 - erf(dcinp->ewcoeff*dcinp->Ecut))/dcinp->Ecut;
  }

  /*** Check input ***/
  if (dcinp->Dtol < 0.0) {
    printf("GetEwaldNamelist >> Error.  Direct sum tolerance must be a "
	   "positive real value.GetEwaldNamelist >> Value of %9.6lf is "
	   "unacceptable.\n", dcinp->Dtol);
    exit(1);
  }
  for (i = 0; i < 3; i++) {
    if (rcinp->ordr[i] < 3) {
      printf("GetEwaldNamelist >> Error.  Order must be at least 3.\n"
	     "GetEwaldNamelist >> Order was specified as %d.\n",
	     rcinp->ordr[i]);
      exit(1);
    }
  }
}

/***=======================================================================***/
/*** GetForceNamelist: this function reads the &force namelist, which has  ***/
/***                   no analog other than perhaps &debug in SANDER.      ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tj:      the trajectory control information (the force report file  ***/
/***              is considered a trajectory type of output)               ***/
/***   inp:     the input file                                             ***/
/***=======================================================================***/
void GetForceNamelist(trajcon *tj, FILE *inp)
{
  int collect;
  char line[MAXLINE];
  cmat lwords;

  /*** By default, all output is written to the variable "S" ***/
  /*** in the force report                                   ***/
  sprintf(tj->DMPvar, "S");
  tj->DMPcrd = 1;
  tj->DMPbond = 1;
  tj->DMPangl = 1;
  tj->DMPdihe = 1;
  tj->DMPrelec = 1;
  tj->DMPdelec = 1;
  tj->DMPvdw = 1;
  tj->DMPall = 1;

  collect = AdvanceToSegment(inp, "force");
  while (collect == 1) {

    /*** Read the next line ***/
    fgets(line, MAXLINE, inp);
    RemoveWhiteSpace(line, MAXLINE);
    RemoveComments(line);

    /*** Break if the line is "&end" ***/
    collect = DetectNamelistEnd(line, "GetCntrlNamelist");
    if (collect == 0) {
      continue;
    }

    /*** Eliminate and add spaces between special characters ***/
    /*** "=", "\n", and ","                                  ***/
    NixCommaCarriage(line);
    EqualSpace(line);
    lwords = ParseWords(line);

    /*** Directives for output organization ***/
    SeekString(lwords, tj->DMPvar, "VarName", "var");

    /*** Directives for output suppression ***/
    SeekInt(lwords, &tj->DMPcrd, "DumpCoord", "dumpcrd");
    SeekInt(lwords, &tj->DMPbond, "DumpBond", "dumpbond");
    SeekInt(lwords, &tj->DMPangl, "DumpAngl", "dumpangl");
    SeekInt(lwords, &tj->DMPdihe, "DumpDihe", "dumpdihe");
    SeekInt(lwords, &tj->DMPrelec, "DumpRElec", "dumprelec");
    SeekInt(lwords, &tj->DMPdelec, "DumpDElec", "dumpdelec");
    SeekInt(lwords, &tj->DMPvdw, "DumpVdw", "dumpvdw");
    SeekInt(lwords, &tj->DMPall, "DumpAll", "dumpall");

    /*** Free allocated memory ***/
    DestroyCmat(&lwords);
  }
}

/***=======================================================================***/
/*** ReadCommandFile: function for reading an mdgx command file.           ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   dcinp:   direct space command information                           ***/
/***   rcinp:   reciprocal space command information                       ***/
/***   tp:      the topology structure (stores its own source file)        ***/
/***   tj:      the trajectory structure (stores the restart file name, as ***/
/***              well as many other output file names)                    ***/
/***   source:  the command input file                                     ***/
/***=======================================================================***/
void ReadCommFile(dircon *dcinp, reccon *rcinp, prmtop *tp, trajcon *tj,
		  char* source)
{
  int i, maxlinelen;
  char line[MAXLINE];
  FILE *inp;

  /*** Open the input file ***/
  if ((inp = fopen(source, "r")) == NULL) {
    printf("ReadCommFile >> Error.  Command file %s not specified.\n",
	   source);
    exit(1);
  }

  /*** First, scan for file information ***/
  GetDataFileNames(tp, tj, inp);

  /*** Scan for any additional information about the topology. ***/
  /*** This is done here so that the topology reader can go on ***/
  /*** reading standard AMBER topology files, but can also be  ***/
  /*** queued to look for additional information that is not   ***/
  /*** found in a standard topology file.                      ***/
  GetTopolNamelist(tp, inp);

  /*** Next, scan for &cntrl namelist information ***/
  GetCntrlNamelist(dcinp, rcinp, tp, tj, inp);

  /*** Now, scan for &ewald namelist information ***/
  GetEwaldNamelist(dcinp, rcinp, inp);

  /*** Scan for &force namelist information ***/
  GetForceNamelist(tj, inp);

  /*** Finally, scan the file verbatim into memory ***/
  rewind(inp);
  i = 0;
  maxlinelen = 0;
  while (fgets(line, MAXLINE, inp) != NULL) {
    i++;
    if (strlen(line) > maxlinelen) {
      maxlinelen = strlen(line);
    }
  }
  rewind(inp);
  tj->inptext = CreateCmat(i, maxlinelen+1);
  rewind(inp);
  i = 0;
  while (fgets(line, MAXLINE, inp) != NULL) {
    strcpy(tj->inptext.map[i], line);
    i++;
  }
  fclose(inp);
}
