#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "mdgxVector.h"
#include "CrdManip.h"

#include "TopologyDS.h"
#include "CrdManipDS.h"
#include "TrajectoryDS.h"

/***=======================================================================***/
/*** Vec2PrimeC: function needed by settlec.  Essentially computes a       ***/
/***             matrix-vector multiplication, where n1, n2, and n0 are    ***/
/***             the rows of the matrix.                                   ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   V:      the vector to transform                                     ***/
/***   n[120]: the rows of the transformation matrix                       ***/
/***=======================================================================***/
static void Vec2PrimeC(double* V, double* n1, double* n2, double* n0)
{
  double Vp[3];

  Vp[0] = Dot3(n1, V);
  Vp[1] = Dot3(n2, V);
  Vp[2] = Dot3(n0, V);
  V[0] = Vp[0];
  V[1] = Vp[1];
  V[2] = Vp[2];
}

/***=======================================================================***/
/*** settlec: the first stage of SETTLE (analytic bond length constraints  ***/
/***          for rigid, three-point water molecules).                     ***/
/***                                                                       ***/
/*** This function is adapted from code developed by the Theoretical and   ***/
/*** Computational Biophysics Group in the Beckman Institute for Advanced  ***/
/*** Science and Technology at the University of Illinois at               ***/
/*** Urbana-Champaign.                                                     ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   ref:     the initial coordinates of all three atoms after the most  ***/
/***            recent application of constraints, in the order oxygen,    ***/
/***            hydrogen 1, hydrogen 2                                     ***/
/***   m[OH]:   the masses of the oxygen (O) and hydrogen (H) atoms        ***/
/***   pos:     on input, the perturbed coordinates (as after a position   ***/
/***            update, prompted by forces acting on the atoms); on        ***/
/***            output, the constrained coordinates                        ***/
/***   r[abc]:  the canonical positions of water atoms (see the diagram in ***/
/***            Cosntraints.h)                                             ***/
/***=======================================================================***/
void settlec(const double* ref, double* pos, settleparm *FWtab)
{
  int i;
  double sinphi, cosphi, sinpsi, cospsi, rbphi, tmp, tmp1, tmp2;
  double alpha, beta, gamma, a2b2, sintheta, costheta, mO, mH, ra, rb, rc;
  double b0[3], c0[3], d0[3], a1[3], b1[3], c1[3], a2[3], b2[3], c2[3];
  double a3[3], b3[3], c3[3], m0[3], m1[3], m2[3], n0[3], n1[3], n2[3];

  /*** Unpack SETTLE parameters ***/
  mO = FWtab->mO;
  mH = FWtab->mH;
  ra = FWtab->ra;
  rb = FWtab->rb;
  rc = FWtab->rc;

  for (i = 0; i < 3; i++) {

    /*** Vectors in the plane of the original positions ***/
    b0[i] = ref[3+i]-ref[i];
    c0[i] = ref[6+i]-ref[i];

    /*** Compute and store the new center of mass ***/
    d0[i] = (pos[i]*mO + pos[3+i]*mH + pos[6+i]*mH)/(mO+mH+mH);

    /*** Translate the system's center of mass to the origin ***/
    a1[i] = pos[i] - d0[i];
    b1[i] = pos[3+i] - d0[i];
    c1[i] = pos[6+i] - d0[i];
  }

  /*** Vectors describing transformation from the original coordinate ***/
  /*** system to the 'primed' coordinate system as in the diagram.    ***/
  CrossP(b0, c0, n0);
  CrossP(a1, n0, n1);
  CrossP(n0, n1, n2);
  UnitVector3(n0);
  UnitVector3(n1);
  UnitVector3(n2);
  Vec2PrimeC(b0, n1, n2, n0);
  Vec2PrimeC(c0, n1, n2, n0);
  Vec2PrimeC(a1, n1, n2, n0);
  Vec2PrimeC(b1, n1, n2, n0);
  Vec2PrimeC(c1, n1, n2, n0);

  /*** Compute positions of canonical water ***/
  sinphi = a1[2]/ra;
  tmp = 1.0-sinphi*sinphi;
  cosphi = sqrt(tmp);
  sinpsi = (b1[2] - c1[2])/(2.0*rc*cosphi);
  tmp = 1.0-sinpsi*sinpsi;
  cospsi = sqrt(tmp);

  rbphi = -rb*cosphi;
  tmp1 = rc*sinpsi*sinphi;
  tmp2 = rc*sinpsi*cosphi;
 
  a2[0] = 0.0;
  a2[1] = ra*cosphi;
  a2[2] = ra*sinphi;
  b2[0] = -rc*cospsi;
  b2[1] = rbphi - tmp1;
  b2[2] = -rb*sinphi + tmp2;
  c2[0] = rc*cosphi;
  c2[1] = rbphi+tmp1;
  c2[2] = -rb*sinphi - tmp2;

  /*** There are no a0 terms because a0 was subtracted off ***/
  /*** when b0 and c0 were first defined.                  ***/
  alpha = b2[0]*(b0[0] - c0[0]) + b0[1]*b2[1] + c0[1]*c2[1];
  beta  = b2[0]*(c0[1] - b0[1]) + b0[0]*b2[1] + c0[0]*c2[1];
  gamma  = b0[0]*b1[1] - b1[0]*b0[1] + c0[0]*c1[1] - c1[0]*c0[1];
 
  a2b2 = alpha*alpha + beta*beta;
  sintheta = (alpha*gamma - beta*sqrt(a2b2 - gamma*gamma))/a2b2;
  costheta = sqrt(1.0 - sintheta*sintheta);

  a3[0] = -a2[1]*sintheta;
  a3[1] = a2[1]*costheta;
  a3[2] = a1[2];
  b3[0] = b2[0]*costheta - b2[1]*sintheta;
  b3[1] = b2[0]*sintheta + b2[1]*costheta;
  b3[2] = b1[2];
  c3[0] = -b2[0]*costheta - c2[1]*sintheta;
  c3[1] = -b2[0]*sintheta + c2[1]*costheta;
  c3[2] = c1[2];

  /*** Undo the transformation; generate new normal vectors ***/
  /*** from the transpose.                                  ***/
  m1[0] = n1[0];
  m1[1] = n2[0];
  m1[2] = n0[0];
  m2[0] = n1[1];
  m2[1] = n2[1];
  m2[2] = n0[1];
  m0[0] = n1[2];
  m0[1] = n2[2];
  m0[2] = n0[2];
  Vec2PrimeC(a3, m1, m2, m0);
  Vec2PrimeC(b3, m1, m2, m0);
  Vec2PrimeC(c3, m1, m2, m0);

  /*** Translate the system's center of mass back to its proper position ***/
  for (i = 0; i < 3; i++) {
    pos[i] = a3[i] + d0[i];
    pos[3+i] = b3[i] + d0[i];
    pos[6+i] = c3[i] + d0[i];
  }
}

/***=======================================================================***/
/*** settlev: velocity constraints portion of the SETTLE algorithm.        ***/
/***                                                                       ***/
/*** This function is adapted from code developed by the Theoretical and   ***/
/*** Computational Biophysics Group in the Beckman Institute for Advanced  ***/
/*** Science and Technology at the University of Illinois at               ***/
/*** Urbana-Champaign.                                                     ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   pos:    the constrained positions of the three atoms (as output by  ***/
/***           settlec, above)                                             ***/
/***   m[OH]:  masses of oxygen (O) and hydrogen (H) atoms                 ***/
/***   vel:    on input, the unconstrained velocities, given in the same   ***/
/***           order as the atoms (O, H1, H2); on output, the constrained  ***/
/***           velocities                                                  ***/
/***=======================================================================***/
void settlev(const double* pos, double* vel, settleparm *FWtab)
{
  int i;
  double cosA, cosB, cosC, vab, vbc, vca, mOH, d, tab, tbc, tca, mO, mH;
  double AB[3], BC[3], CA[3], vtAB[3], vtBC[3], vtCA[3], ga[3], gb[3], gc[3];

  /*** Unpack SETTLE parameters ***/
  mO = FWtab->mO;
  mH = FWtab->mH;

  for (i = 0; i < 3; i++) {
    AB[i] = pos[3+i]-pos[i];
    BC[i] = pos[6+i]-pos[3+i];
    CA[i] = pos[i]-pos[6+i];
  }

  UnitVector3(AB);
  UnitVector3(BC);
  UnitVector3(CA);

  cosA = -Dot3(AB, CA);
  cosB = -Dot3(BC, AB);
  cosC = -Dot3(CA, BC);

  for (i = 0; i < 3; i++) {
    vtAB[i] = (vel[3+i]-vel[i]);
    vtBC[i] = (vel[6+i]-vel[3+i]);
    vtCA[i] = (vel[i]-vel[6+i]);
  }
  vab = Dot3(AB, vtAB);
  vbc = Dot3(BC, vtBC);
  vca = Dot3(CA, vtCA);

  mOH = mO+mH;
  d = mH/(mOH*mOH + mH*cosA*(mO*cosB*cosC - mH*cosA) -
	  0.5*mO*mOH*(cosB*cosB + cosC*cosC));
  tab = (vab*(2.0*mOH - mO*cosC*cosC) + 
	 vbc*(mH*cosC*cosA - mOH*cosB) +
	 vca*(mO*cosB*cosC - 2.0*mH*cosA))*mO*d;
  tbc = (vbc*(mOH*mOH - mH*mH*cosA*cosA) +
	 vca*mO*(mH*cosA*cosB - mOH*cosC) +
	 vab*mO*(mH*cosC*cosA - mOH*cosB))*d;
  tca = (vca*(2*mOH - mO*cosB*cosB) +
	 vab*(mO*cosB*cosC - 2*mH*cosA) +
	 vbc*(mH*cosA*cosB - mOH*cosC))*mO*d;

  mO = 0.5/mO;
  mH = 0.5/mH;
  for (i = 0; i < 3; i++) {
    ga[i] = tab*AB[i] - tca*CA[i];
    gb[i] = tbc*BC[i] - tab*AB[i];
    gc[i] = tca*CA[i] - tbc*BC[i];
    vel[i] += mO*ga[i];
    vel[3+i] += mH*gb[i];
    vel[6+i] += mH*gc[i];
  }
}

#if 0
/***=======================================================================***/
/*** rattlec: first stage (position constraints) of the RATTLE algorithm   ***/
/***          for rigid bonds in the context of Velocity-Verlet            ***/
/***          integration.                                                 ***/
/***                                                                       ***/
/*** This function is adapted from code in the tinker software package     ***/
/*** and the AMBER sff program.                                            ***/
/***                                                                       ***/
/***=======================================================================***/
void rattlec(double dt, double* x, double* xold, double* v, double* minv,
	     prmtop *tp, const double eptol, const int maxiter)
/*
double dt;
double *x, *xold, *v, *minv;
*/
{
  int i, niter, done, nratH, nrat, at1, at2, atyp;
  double xr, yr, zr, xo, yo, zo, xterm, yterm, zterm;
  double dist2, dot;
  double krat, term, delta, rma, rmb;

  niter = 0;
  done = 0;
  nratH = tp->withH.nbond;
  nrat = (irattle == 2) ? 0 : tp->woH.nbond;
  while (done == 0 && niter < maxiter) {
    niter++;
    done = 1;

    for (i = 0; i < nratH + nrat; i++) {
      if (i < nratH) {
	at1 = prm->BondHAt1[i];
	at2 = prm->BondHAt2[i];
	atyp = prm->BondHNum[i] - 1;
	krat = prm->Req[atyp];
      }
      else {
	at1 = prm->BondAt1[i - nratH];
	at2 = prm->BondAt2[i - nratH];
	atyp = prm->BondNum[i - nratH] - 1;
	krat = prm->Req[atyp];
      }

      xr = x[at2] - x[at1];
      yr = x[at2 + 1] - x[at1 + 1];
      zr = x[at2 + 2] - x[at1 + 2];
      dist2 = xr * xr + yr * yr + zr * zr;

      delta = krat * krat - dist2;

      if (fabs(delta) > eps) {
	done = 0;
	xo = xold[at2] - xold[at1];
	yo = xold[at2 + 1] - xold[at1 + 1];
	zo = xold[at2 + 2] - xold[at1 + 2];

	dot = xr * xo + yr * yo + zr * zo;
	rma = minv[at1];
	rmb = minv[at2];

#define FAC 1.2
	term = FAC * delta / (2.0 * (rma + rmb) * dot);

	xterm = xo * term;
	yterm = yo * term;
	zterm = zo * term;

	x[at1] -= xterm * rma;
	x[at1 + 1] -= yterm * rma;
	x[at1 + 2] -= zterm * rma;

	x[at2] += xterm * rmb;
	x[at2 + 1] += yterm * rmb;
	x[at2 + 2] += zterm * rmb;

	rma = rma / dt;
	rmb = rmb / dt;
	v[at1] -= xterm * rma;
	v[at1 + 1] -= yterm * rma;
	v[at1 + 2] -= zterm * rma;

	v[at2] += xterm * rmb;
	v[at2 + 1] += yterm * rmb;
	v[at2 + 2] += zterm * rmb;
      }
    }
  }

  if (niter >= maxiter - 1)
    fprintf(nabout, "ERROR in RATTLE\n");

  return (niter);
}

int rattle2(double dt, double * x, double * v, double * minv)
/*
double dt;
double *x, *v, *minv;
*/
{
  int maxiter, niter, done, i;
  int nratH, nrat;
  int at1, at2, atyp;
  double eps, xr, yr, zr, xv, yv, zv, xterm, yterm, zterm;
  double dot, rma, rmb, krat, term;

  eps = EPS / dt;
  maxiter = 1000;
  niter = 0;
  nratH = prm->Nbonh;
  nrat = prm->Mbona;
  if( irattle == 2 ) nrat=0;
  done = 0;

  while ((done == 0) && (niter < maxiter)) {
    niter++;
    done = 1;

    for (i = 0; i < nratH + nrat; i++) {
      if (i < nratH) {
	at1 = prm->BondHAt1[i];
	at2 = prm->BondHAt2[i];
	atyp = prm->BondHNum[i] - 1;
	krat = prm->Req[atyp];
      } else {
	at1 = prm->BondAt1[i - nratH];
	at2 = prm->BondAt2[i - nratH];
	atyp = prm->BondNum[i - nratH] - 1;
	krat = prm->Req[atyp];
      }

      xr = x[at2] - x[at1];
      yr = x[at2 + 1] - x[at1 + 1];
      zr = x[at2 + 2] - x[at1 + 2];

      xv = v[at2] - v[at1];
      yv = v[at2 + 1] - v[at1 + 1];
      zv = v[at2 + 2] - v[at1 + 2];

      dot = xr * xv + yr * yv + zr * zv;
      rma = minv[at1];
      rmb = minv[at2];

      term = -dot * FAC / ((rma + rmb) * krat * krat);

      if (fabs(term) > eps) {
	done = 0;
	xterm = xr * term;
	yterm = yr * term;
	zterm = zr * term;

	v[at1] -= xterm * rma;
	v[at1 + 1] -= yterm * rma;
	v[at1 + 2] -= zterm * rma;

	v[at2] += xterm * rmb;
	v[at2 + 1] += yterm * rmb;
	v[at2 + 2] += zterm * rmb;
      }

    }
  }
  if (niter >= maxiter - 1)
    fprintf(nabout, "ERROR in RATTLE\n");
  return (niter);
}
#endif

/***=======================================================================***/
/*** ReimageConstraintGroup: this makes sure that we don't try to SETTLE   ***/
/***                         or RATTLE molecules that are split across two ***/
/***                         sides of the simulation box.  The consequence ***/
/***                         of calling this function is that constraint   ***/
/***                         groups will be re-imaged relative to the      ***/
/***                         central constraint atom.                      ***/
/***                                                                       ***/
/***                         This function should not be used if           ***/
/***                         constraints are implemented from a cell-based ***/
/***                         list of coordinates (re-imaging is already    ***/
/***                         done for cells).                              ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   x:     the coordinates of atoms in this constraint group (the first ***/
/***            atom is taken to be the central constraint atom, used as   ***/
/***            the reference when re-imaging)                             ***/
/***   crd:   the system coordinates (needed for box dimensions and box    ***/
/***            transformation matrices)                                   ***/
/***   natom: the number of atoms in the constraint group, including the   ***/
/***            central atom                                               ***/
/***=======================================================================***/
void ReimageConstraintGroup(double* x, coord *crd, int natom)
{
  int i, i3;
  double cnx, cny, cnz, dx, dy, dz;

  /*** Loop over all atoms ***/
  cnx = x[0];
  cny = x[1];
  cnz = x[2];
  for (i = 1; i < natom; i++) {
    i3 = 3*i;

    /*** Compute differences ***/
    dx = x[i3] - cnx;
    dy = x[i3+1] - cny;
    dz = x[i3+2] - cnz;
    if (crd->isortho == 1) {
      OrthoReim(&dx, &dy, &dz, crd->hgdim);
    }
    else {
      NonOrthoReim(&dx, &dx, &dz, &crd->U, &crd->invU);
    }
    x[i3] = cnx + dx;
    x[i3+1] = cny + dy;
    x[i3+2] = cnz + dz;
  }
}

/***=======================================================================***/
/*** PositionConstraints: this is a wrapper for SETTLE, RATTLE, and other  ***/
/***                      constraint functions.  Like the CompBondedIntr   ***/
/***                      function in the Bonded library, this function    ***/
/***                      loops over all atoms and performs whatever       ***/
/***                      constraints are required in each case.  More     ***/
/***                      precisely, each constraint is "owned" by a       ***/
/***                      particular atom, and whatever cell owns the atom ***/
/***                      on a particular step is responsible for          ***/
/***                      computing its constraints.                       ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:  the coordinates                                               ***/
/***   tp:   the topology                                                  ***/
/***=======================================================================***/
void PositionConstraints(coord *crd, prmtop *tp, double invdt)
{
  int i, j, i3, nh13, nh23;
  double xpos[9], xposc[9], xref[9];
  cnstcomm *tSHL;

  tSHL = tp->SHL;

  /*** SETTLE ***/
  if (tp->settle == 1) {
    for (i = 0; i < tp->natom; i++) {
      if (tSHL[i].exe == 1) {
	i3 = 3*i;
	nh13 = 3*tSHL[i].blist[0];
	nh23 = 3*tSHL[i].blist[1];
	for (j = 0; j < 3; j++) {
	  xref[j] = crd->prvloc[i3+j];
	  xref[3+j] = crd->prvloc[nh13+j];
	  xref[6+j] = crd->prvloc[nh23+j];
	  xpos[j] = crd->loc[i3+j];
	  xpos[3+j] = crd->loc[nh13+j];
	  xpos[6+j] = crd->loc[nh23+j];
	}
	ReimageConstraintGroup(xref, crd, 3);
	ReimageConstraintGroup(xpos, crd, 3);
	for (j = 0; j < 9; j++) {
	  xposc[j] = xpos[j];
	}
	settlec(xref, xpos, &tp->FWtab);
	for (j = 0; j < 3; j++) {
	  crd->loc[i3+j] = xpos[j];
	  crd->loc[nh13+j] = xpos[3+j];
	  crd->loc[nh23+j] = xpos[6+j];
	  crd->vel[i3+j] += (xpos[j] - xposc[j])*invdt;
	  crd->vel[nh13+j] += (xpos[3+j] - xposc[3+j])*invdt;
	  crd->vel[nh23+j] += (xpos[6+j] - xposc[6+j])*invdt;
	}
      }
    }
  }

  /*** RATTLE ***/
}

/***=======================================================================***/
/*** VelocityConstraints: this is a wrapper for SETTLE, RATTLE, and other  ***/
/***                      constraint functions.  Like the CompBondedIntr   ***/
/***                      loops over all atoms and performs whatever       ***/
/***                      constraints are required in each case.  More     ***/
/***                      precisely, each constraint is "owned" by a       ***/
/***                      particular atom, and whatever cell owns the atom ***/
/***                      on a particular step is responsible for          ***/
/***                      computing its constraints.                       ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:  the coordinates                                               ***/
/***   tp:   the topology                                                  ***/
/***=======================================================================***/
void VelocityConstraints(coord *crd, prmtop *tp)
{
  int i, j, i3, nh13, nh23;
  double xpos[24], xvel[24];
  cnstcomm *tSHL;

  tSHL = tp->SHL;

  /*** SETTLE ***/
  if (tp->settle == 1) {
    for (i = 0; i < tp->natom; i++) {
      if (tSHL[i].exe == 1) {
	i3 = 3*i;
	nh13 = 3*tSHL[i].blist[0];
	nh23 = 3*tSHL[i].blist[1];
	for (j = 0; j < 3; j++) {
	  xpos[j] = crd->loc[i3+j];
	  xpos[3+j] = crd->loc[nh13+j];
	  xpos[6+j] = crd->loc[nh23+j];
	  xvel[j] = crd->vel[i3+j];
	  xvel[3+j] = crd->vel[nh13+j];
	  xvel[6+j] = crd->vel[nh23+j];
	}
	ReimageConstraintGroup(xpos, crd, 3);
	settlev(xpos, xvel, &tp->FWtab);

	for (j = 0; j < 3; j++) {
	  crd->vel[i3+j] = xvel[j];
	  crd->vel[nh13+j] = xvel[3+j];
	  crd->vel[nh23+j] = xvel[6+j];
	}
      }
    }
  }

  /*** RATTLE ***/
}

/***=======================================================================***/
/*** VVConstraintVirial: compute the virial contributions from the SETTLE  ***/
/***                     bond constraints in either the first or second    ***/
/***                     Velocity-Verlet integration step.                 ***/
/*** Arguments:                                                            ***/
/***   crd:     the coordinates (for velocity information)                 ***/
/***   tp:      the topology (for masses)                                  ***/
/***   tj:      trajectory control information                             ***/
/***   sysUV:   energy and virial information                              ***/
/***=======================================================================***/
void VVConstraintVirial(coord *crd, prmtop *tp, trajcon *tj, Energy *sysUV)
{
  int i, j, k, atjk3;
  int atnum[24];
  double invdt, acc, jmass;
  double* oldcrd;
  double* newcrd;
  double* oldvel;
  double* newvel;

  /*** Shortcuts ***/
  invdt = (tj->dt > 1.0e-12) ? 1.0 / tj->dt : 0.0;
  newcrd = crd->loc;
  oldcrd = crd->prvloc;
  newvel = crd->vel;
  oldvel = crd->prvvel;

  /*** Loop over all residues ***/
  for (i = 0; i < tp->natom; i++) {

    /*** Does this atom control a SETTLE constraint group? ***/
    if (tp->SHL[i].exe == 1) {
      atnum[0] = i;
      atnum[1] = tp->SHL[i].blist[0];
      atnum[2] = tp->SHL[i].blist[1];

      /*** Loop over the oxygen and both hydrogen atoms ***/
      for (j = 0; j < 3; j++) {
	jmass = tp->Masses[atnum[j]];
        for (k = 0; k < 3; k++) {
	  atjk3 = 3*atnum[j]+k;

          if (tj->currstep == 0) {
            acc = (newcrd[atjk3] - oldcrd[atjk3])*invdt - oldvel[atjk3];
            sysUV->Vir[4*k] -= acc*oldcrd[atjk3]*jmass;
          }
          else {
            acc = newvel[atjk3] - oldvel[atjk3];
            sysUV->Vir[4*k] -= acc*newcrd[atjk3]*jmass;
          }
        }
      }
    }

    /*** Does this atom control other constraint groups? ***/
  }
}
