#include <math.h>
#include <stdlib.h>
#include "Matrix.h"
#include "pmeRecip.h"
#include "CrdManip.h"
#include "mdgxVector.h"

#include "pmeDirectDS.h"

/***=======================================================================***/
/*** CreateCoord: create a coordinate structure.                           ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   natom:  the number of atoms to allocate for                         ***/
/***=======================================================================***/
coord CreateCoord(int natom)
{
  coord crd;

  crd.natom = natom;
  crd.atmid = (int*)malloc(natom*sizeof(int));
  crd.loc = (double*)malloc(3*natom*sizeof(double));
  crd.prvloc = (double*)malloc(3*natom*sizeof(double));
  crd.bufloc = (double*)malloc(3*natom*sizeof(double));
  crd.vel = (double*)malloc(3*natom*sizeof(double));
  crd.prvvel = (double*)malloc(3*natom*sizeof(double));
  crd.frc = (double*)malloc(3*natom*sizeof(double));
  crd.prvfrc = (double*)malloc(3*natom*sizeof(double));
  crd.buffrc = (double*)malloc(3*natom*sizeof(double));
  crd.U = CreateDmat(3, 3, 0);
  crd.invU = CreateDmat(3, 3, 0);

  return crd;
}

/***=======================================================================***/
/*** CopyCoord: copy coord struct crd into Xcrd.                           ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:  the original coord struct                                     ***/
/***=======================================================================***/
coord CopyCoord(coord *crd)
{
  int i;
  coord Xcrd;

  Xcrd.natom  = crd->natom;
  Xcrd.isortho = crd->isortho;
  Xcrd.atmid  = CpyIVec(crd->atmid, crd->natom);
  Xcrd.loc    = CpyDVec(crd->loc, 3*crd->natom);
  Xcrd.prvloc = CpyDVec(crd->prvloc, 3*crd->natom);
  Xcrd.bufloc = CpyDVec(crd->bufloc, 3*crd->natom);
  Xcrd.vel    = CpyDVec(crd->vel, 3*crd->natom);
  Xcrd.prvvel = CpyDVec(crd->prvvel, 3*crd->natom);
  Xcrd.frc    = CpyDVec(crd->frc, 3*crd->natom);
  Xcrd.prvfrc = CpyDVec(crd->prvfrc, 3*crd->natom);
  Xcrd.buffrc = CpyDVec(crd->buffrc, 3*crd->natom);
  for (i = 0; i < 6; i++) {
    Xcrd.gdim[i] = crd->gdim[i];
  }
  for (i = 0; i < 3; i++) {
    Xcrd.hgdim[i] = crd->hgdim[i];
  }
  Xcrd.U = CreateDmat(3, 3, 0);
  Xcrd.invU = CreateDmat(3, 3, 0);
  CompXfrm(Xcrd.gdim, Xcrd.U, Xcrd.invU);

  return Xcrd;
}

/***=======================================================================***/
/*** DestroyCoord: free all memory associated with a coord struct.         ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd: the coord struct to destroy                                    ***/
/***=======================================================================***/
void DestroyCoord(coord *crd)
{
  free(crd->atmid);
  free(crd->loc);
  free(crd->prvloc);
  free(crd->bufloc);
  free(crd->vel);
  free(crd->prvvel);
  free(crd->frc);
  free(crd->prvfrc);
  free(crd->buffrc);
  DestroyDmat(&crd->U);
  DestroyDmat(&crd->invU);
}

/***=======================================================================***/
/*** TransCrd: translates a set of coordinates by step*tvec[].             ***/
/***=======================================================================***/
void TransCrd(double* crds, int natom, double* tvec, double step)
{
  int i;
  double nvec[3];

  for (i = 0; i < 3; i++) {
    nvec[i] = tvec[i]*step;
  }
  for (i = 0; i < natom; i++) {
    crds[3*i] += nvec[0];
    crds[3*i+1] += nvec[1];
    crds[3*i+2] += nvec[2];
  }
}

/***=======================================================================***/
/*** RotateCrd: rotates a set of coordinates using matrix U.               ***/
/***=======================================================================***/
void RotateCrd(double* crds, int natom, dmat U)
{
  int i;
  double tmp_crds[3];
  double* dtmp;

  dtmp = &crds[0];
  for (i = 0; i < natom; i++) {
    tmp_crds[0] = U.data[0]*dtmp[0] + U.data[1]*dtmp[1] + U.data[2]*dtmp[2];
    tmp_crds[1] = U.data[3]*dtmp[0] + U.data[4]*dtmp[1] + U.data[5]*dtmp[2];
    tmp_crds[2] = U.data[6]*dtmp[0] + U.data[7]*dtmp[1] + U.data[8]*dtmp[2];
    dtmp[0] = tmp_crds[0];
    dtmp[1] = tmp_crds[1];
    dtmp[2] = tmp_crds[2];
    dtmp = &dtmp[3];
  }
}

/***=======================================================================***/
/*** CompXfrm: compute the transformation matrices U and invU that take a  ***/
/***           set of coordinates into and out of box space.  The matrices ***/
/***           U and invU must be pre-allocated.                           ***/
/***=======================================================================***/
void CompXfrm(double* cd, dmat U, dmat invU)
{
  double dx, dy;

  dx = (cos(cd[4])*cos(cd[5]) - cos(cd[3])) / (sin(cd[4]) * sin(cd[5]));
  dy = sqrt(1.0 - dx*dx);
  invU.data[0] = cd[0];
  invU.data[1] = cd[1]*cos(cd[5]);
  invU.data[2] = cd[2]*cos(cd[4]);
  invU.data[4] = cd[1]*sin(cd[5]);
  invU.data[5] = -cd[2]*sin(cd[4])*dx;
  invU.data[8] = cd[2]*sin(cd[4])*dy;
  ttInv(invU, U);
}

/***=======================================================================***/
/*** PrimaryImage: re-image a set of coordinates such that everything lies ***/
/***               within [0, 1) in all three axes of box space.           ***/
/***               Coordinates are returned to real space before exit.     ***/
/***=======================================================================***/
void PrimaryImage(coord *crd)
{
  int i;
  double tx, ty, tz;
  double *dtmp;

  /*** Pointer to the coordinates ***/
  dtmp = crd->loc;

  /*** If we have an orthorhombic unit cell, it's easy ***/
  if (TestUnitCellOrtho(crd->gdim) == 1) {

    int i3;
    double bx, by, bz, invbx, invby, invbz;

    bx = crd->gdim[0];
    by = crd->gdim[1];
    bz = crd->gdim[2];
    invbx = 1.0/bx;
    invby = 1.0/by;
    invbz = 1.0/bz;
    for (i = 0; i < crd->natom; i++) {
      i3 = 3*i;
      tx = invbx*dtmp[i3];
      ty = invby*dtmp[i3+1];
      tz = invbz*dtmp[i3+2];
      tx -= floor(tx);
      ty -= floor(ty);
      tz -= floor(tz);
      dtmp[i3] = tx*bx;
      dtmp[i3+1] = ty*by;
      dtmp[i3+2] = tz*bz;
    }
  }

  /*** Otherwise, we need to use the box transformation matrices ***/
  else {
    RotateCrd(dtmp, crd->natom, crd->U);
    for (i = 0; i < 3*crd->natom; i++) {
      dtmp[i] -= floor(dtmp[i]);
    }
    RotateCrd(dtmp, crd->natom, crd->invU);
  }
}

/***=======================================================================***/
/*** OrthoReim: re-image a single set of three coordinates (such as a      ***/
/***            displacement) given an orthorhombic box and the knowledge  ***/
/***            that nothing is more than one box length from the primary  ***/
/***            image.                                                     ***/
/***=======================================================================***/
void OrthoReim(double *dx, double *dy, double *dz, double* hgdim)
{
  if (*dx >= hgdim[0]) {
    *dx -= 2.0*hgdim[0];
  }
  else if (*dx < -hgdim[0]) {
    *dx += 2.0*hgdim[0];
  }
  if (*dy >= hgdim[1]) {
    *dy -= 2.0*hgdim[1];
  }
  else if (*dy < -hgdim[1]) {
    *dy += 2.0*hgdim[1];
  }
  if (*dz >= hgdim[2]) {
    *dz -= 2.0*hgdim[2];
  }
  else if (*dz < -hgdim[2]) {
    *dz += 2.0*hgdim[2];
  }
}

/***=======================================================================***/
/*** NonOrthoReim: re-image a single set of three coordinates (such as a   ***/
/***               displacement) given an non-orthorhombic box and the     ***/
/***               knowledge that nothing is more than one box length from ***/
/***               the primary image.                                      ***/
/***=======================================================================***/
void NonOrthoReim(double *dx, double *dy, double *dz, dmat *U, dmat *invU)
{
  double ndx, ndy, ndz;
  double *dtmp;

  dtmp = U->data;
  ndx = dtmp[0]*(*dx) + dtmp[1]*(*dy) + dtmp[2]*(*dz);
  ndy = dtmp[3]*(*dx) + dtmp[4]*(*dy) + dtmp[5]*(*dz);
  ndz = dtmp[6]*(*dx) + dtmp[7]*(*dy) + dtmp[8]*(*dz);
  if (ndx >= 0.5) {
    ndx -= 1.0;
  }
  else if (ndx < -0.5) {
    ndx += 1.0;
  }
  if (ndy >= 0.5) {
    ndy -= 1.0;
  }
  else if (ndy < -0.5) {
    ndy += 1.0;
  }
  if (ndz >= 0.5) {
    ndz -= 1.0;
  }
  else if (ndz < -0.5) {
    ndz += 1.0;
  }
  dtmp = invU->data;
  *dx = dtmp[0]*ndx + dtmp[1]*ndy + dtmp[2]*ndz;
  *dy = dtmp[3]*ndx + dtmp[4]*ndy + dtmp[5]*ndz;
  *dz = dtmp[6]*ndx + dtmp[7]*ndy + dtmp[8]*ndz;
}
