#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "Random.h"
#include "Nonbonded.h"
#include "Bonded.h"
#include "pmeRecip.h"
#include "mleRecip.h"
#include "CellManip.h"
#include "BSpline.h"
#include "mdgxVector.h"
#include "Grid.h"
#include "ChargeMap.h"
#include "Constraints.h"
#include "Integrator.h"
#include "Trajectory.h"
#include "Topology.h"
#include "Timings.h"
#include "VirtualSites.h"
#include "Thermostats.h"
#include "Barostats.h"
#include "CrdManip.h"

#include "CompFrcDS.h"
#include "pmeDirectDS.h"

/***=======================================================================***/
/*** AtomForces: the complete force calculation routine, with branches for ***/
/***             special cases in which the energy or virial is desired.   ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:     the coordinates (for velocity information)                 ***/
/***   CG:      the cell grid                                              ***/
/***   tp:      the topology (for masses)                                  ***/
/***   dcinp:   the direct space parameters                                ***/
/***   Etab:    the direct space electrostatic interaction spline table    ***/
/***   EHtab:   the high-resolution direct space electrostatic interaction ***/
/***            spline table                                               ***/
/***   rcinp:   the reciprocal space control parameters                    ***/
/***   Q:       the reciprocal space charge / potential mesh               ***/
/***   pmmap:   the reciprocal space particle<-->mesh spline map           ***/
/***   PPk:     the reciprocal space pair potential mesh construction kit  ***/
/***   sysUV:   state information data (potential energy and virial)       ***/
/***=======================================================================***/
void AtomForces(coord *crd, cellgrid *CG, prmtop *tp, dircon *dcinp,
		FrcTab *Etab, FrcTab *EHtab, reccon *rcinp, bmap *pmmap,
		bckit *PPk, Energy *sysUV, execon *etimers)
{
  int i;

  /*** Extra points must be positioned so that forces may be calculated ***/
  PlaceExtraPoint(tp, crd);

  /*** Direct space force calculation ****/
  AtomsToCellsOrtho(crd, dcinp, CG, tp);
  ShareTriple2(CG, dcinp->Mcut, crd, tp);
  etimers->plist += mdgxStopTimer(etimers);
  for (i = 0; i < CG->ng[0]*CG->ng[1]*CG->ng[2]; i++) {
    DirectTriple2(&CG->data[i], CG->celldim, Etab, dcinp, tp, sysUV);
  }
  MergeTriple2(CG);
  MapCellsToAtoms(CG, crd);
  etimers->nbInt += mdgxStopTimer(etimers);

  /*** Bonded interaction force calculation ***/
  ListBondedIntr(crd, tp, Etab, EHtab, sysUV);
  etimers->bonds += mdgxStopTimer(etimers);

  /*** Reciprocal space force calculation ***/
  SplCoeff(crd, pmmap, rcinp);
  etimers->nbBsp += mdgxStopTimer(etimers);
  Charges2Book(pmmap, rcinp, tp, &rcinp->QL[0]);
  etimers->nbPtM += mdgxStopTimer(etimers);

  /*** Branch for SPME or MLE reciprocal space sum ***/
  if (rcinp->nlev == 1) {

    /*** The usual Smooth Particle Mesh Ewald calculation ***/
    if (sysUV->updateU == 2) {
      ConvQBCnrgvir(rcinp, crd, &rcinp->QL[0], PPk, sysUV, etimers);
    }
    else if (sysUV->updateU == 1 || sysUV->updateU == -1) {
      ConvQBCnrg(rcinp, crd, &rcinp->QL[0], PPk, sysUV, etimers);
    }
    else {
      ConvQBC(rcinp, crd, &rcinp->QL[0], PPk, etimers);
    }
  }
  else {

    /*** Multi-Level Ewald calculation ***/
    mleConvQBC(rcinp, sysUV, crd, PPk, etimers);
  }
  if (sysUV->updateU >= 0) {
    IntrpFrc(pmmap, crd, rcinp, tp, &rcinp->QL[0]);
  }
  etimers->nbMtP += mdgxStopTimer(etimers);

  /*** Forces on extra points must be transferred to atoms with mass ***/
  TransferEPForces(tp, crd);
}

/***=======================================================================***/
/*** VelVerletC: velocity Verlet algorithm for updating positions.  This   ***/
/***             puts the current coordinates of the atoms in the prvloc   ***/
/***             array of the coord struct and the current set of forces   ***/
/***             in the prvfrc array of the coord struct.                  ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:  the coordinates (also contains velocity information)          ***/
/***   dt:   the timestep                                                  ***/
/***=======================================================================***/
static void VelVerletC(coord *crd, prmtop *tp, double dt)
{
  int i, i3, i3p1, i3p2;
  double *ltmp, *ptmp, *vtmp, *ftmp, *fptmp, *mtmp;

  /*** Save the current set of coordinates ***/
  ltmp = crd->prvloc;
  ptmp = crd->loc;
  vtmp = crd->vel;
  ftmp = crd->frc;
  fptmp = crd->prvfrc;
  mtmp = tp->InvMasses;

  /*** Loop over all atoms ***/
  const double hdt = 0.5*418.4*dt;
  i3 = 0;
  i3p1 = 1;
  i3p2 = 2;
  for (i = 0; i < crd->natom; i++) {
    const double hmdt = hdt*mtmp[i];
    ltmp[i3] = ptmp[i3] + dt*(vtmp[i3] + hmdt*ftmp[i3]);
    ltmp[i3p1] = ptmp[i3p1] + dt*(vtmp[i3p1] + hmdt*ftmp[i3p1]);
    ltmp[i3p2] = ptmp[i3p2] + dt*(vtmp[i3p2] + hmdt*ftmp[i3p2]);
    fptmp[i3] = ftmp[i3];
    fptmp[i3p1] = ftmp[i3p1];
    fptmp[i3p2] = ftmp[i3p2];
    i3 += 3;
    i3p1 += 3;
    i3p2 += 3;
  }
  crd->loc = ltmp;
  crd->prvloc = ptmp;
}

/***=======================================================================***/
/*** VelVerletV: the velocity Verlet algorithm for updating velocities.    ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:   the coordinates (also contains velocity information)         ***/
/***   tp:    the topology (for masses)                                    ***/
/***   dt:    the timestep                                                 ***/
/***=======================================================================***/
static void VelVerletV(coord *crd, prmtop *tp, double dt)
{
  int i, i3, i3p1, i3p2;
  double *vtmp, *ftmp, *fptmp, *mtmp;

  vtmp = crd->vel;
  mtmp = tp->InvMasses;
  ftmp = crd->frc;
  fptmp = crd->prvfrc;

  /*** Loop over all atoms ***/
  const double hdt = 0.5*418.4*dt;
  i3 = 0;
  i3p1 = 1;
  i3p2 = 2;
  for (i = 0; i < crd->natom; i++) {
    const double hmdt = hdt*mtmp[i];
    vtmp[i3] += hmdt*(ftmp[i3] + fptmp[i3]);
    vtmp[i3p1] += hmdt*(ftmp[i3p1] + fptmp[i3p1]);
    vtmp[i3p2] += hmdt*(ftmp[i3p2] + fptmp[i3p2]);
    i3 += 3;
    i3p1 += 3;
    i3p2 += 3;
  }
}

/***=======================================================================***/
/*** RemoveSystemMomentum: this function removes the net momentum of the   ***/
/***                       system by applying a uniform correction to all  ***/
/***                       particles.                                      ***/
/***=======================================================================***/
void RemoveSystemMomentum(coord *crd, prmtop *tp)
{
  int i;
  double mvx, mvy, mvz, invn;
  double *vtmp, *mtmp;

  vtmp = crd->vel;
  mtmp = tp->Masses;
  mvx = 0.0;
  mvy = 0.0;
  mvz = 0.0;
  for (i = 0; i < crd->natom; i++) {
    mvx += vtmp[3*i]*mtmp[i];
    mvy += vtmp[3*i+1]*mtmp[i];
    mvz += vtmp[3*i+2]*mtmp[i];
  }
  invn = 1.0/tp->TotalMass;
  mvx *= invn;
  mvy *= invn;
  mvz *= invn;
  for (i = 0; i < crd->natom; i++) {
    vtmp[3*i] -= mvx;
    vtmp[3*i+1] -= mvy;
    vtmp[3*i+2] -= mvz;
  }
}

/***=======================================================================***/
/*** SumTotalEnergy: this routine computes the total potential energy as   ***/
/***                 well as the total (kinetic plus potential) energy.    ***/
/***                 The kinetic energy is not updated inside this         ***/
/***                 function, however, and so must be computed prior to   ***/
/***                 calling this function.                                ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   sysUV:  the energy / virial accumulator                             ***/
/***=======================================================================***/
void SumTotalEnergy(Energy *sysUV)
{
  sysUV->eptot = sysUV->bond + sysUV->angl + sysUV->dihe + sysUV->delec +
                 sysUV->relec + sysUV->vdw12 + sysUV->vdw6;
  sysUV->etot = sysUV->eptot + sysUV->kine;
}

/***=======================================================================***/
/*** WriteTrajFiles: write all the requested trajectory files if this step ***/
/***                 is one of interest.  The restart file is not written  ***/
/***                 at this time, but rather earlier in the file, because ***/
/***                 the checkpointing information must be sufficient to   ***/
/***                 continue the run as if it were continuous.  Restart   ***/
/***                 files are therefore written before this routine is    ***/
/***                 called, after the velocity update but before velocity ***/
/***                 corrections for any constraints.  Checkpoint data     ***/
/***                 contains position, but not velocity corrections, and  ***/
/***                 does not contain the latest thermostat or barostat    ***/
/***                 corrections as are found in the trajectory data for   ***/
/***                 the same step number.                                 ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tj:      trajectory control parameters                              ***/
/***   tp:      the topology                                               ***/
/***   dcinp:   direct space control data                                  ***/
/***   rcinp:   reciprocal space control data                              ***/
/***   Etab:    the electrostatic direct space spline interpolation table  ***/
/***   crd:     the coordinates (for velocity information)                 ***/
/***   sysUV:   the system energy and virial information                   ***/
/***=======================================================================***/
static void WriteTrajFiles(trajcon *tj, prmtop *tp, dircon *dcinp,
			   reccon *rcinp, FrcTab *Etab, coord *crd,
			   Energy *sysUV, execon *etimers, cdftrj* Acdf)
{
  /*** Output diagnostics ***/
  if (tj->ntpr != 0 && tj->currstep % tj->ntpr == 0) {
    WriteDiagnostics(tj, tp, dcinp, rcinp, Etab, crd, sysUV, etimers);
  }

  /*** Coordinates trajectory ***/
  if (tj->ntwx != 0 && tj->currstep % tj->ntwx == 0) {
    if (tj->ioutfm == 0) {
      WriteCrd(crd, 1, tj);
    }
    else if (tj->ioutfm == 1) {
      WriteCDF(crd, 1, tj, Acdf);
    }
  }

  /*** Velocities trajectory ***/
  if (tj->ntwv != 0 && tj->currstep % tj->ntwv == 0) {
    if (tj->ioutfm == 0) {
      WriteCrd(crd, 2, tj);
    }
    else if (tj->ioutfm == 1) {
      WriteCDF(crd, 2, tj, &Acdf[1]);
    }
  }

  /*** Forces trajectory ***/
  if (tj->ntwf != 0 && tj->currstep % tj->ntwf == 0) {
    if (tj->ioutfm == 0) {
      WriteCrd(crd, 3, tj);
    }
    else if (tj->ioutfm == 1) {
      WriteCDF(crd, 3, tj, &Acdf[2]);
    }
  }
}

/***=======================================================================***/
/*** InitVelocities: initialize the velocities of all atoms in the system  ***/
/***                 using random values in a Maxwell distribution, if the ***/
/***                 velocities are not yet assigned, or resumes the       ***/
/***                 dynamics if velocities have been read from a restart  ***/
/***                 file.                                                 ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:     the coordinates (for velocity information)                 ***/
/***   tp:      the topology (for masses)                                  ***/
/***   T:       the temperature                                            ***/
/***   dcinp:   the direct space control parameters                        ***/
/***   Etab:    the direct space electrostatic interaction spline table    ***/
/***   EHtab:   the high-resolution direct space electrostatic interaction ***/
/***            spline table                                               ***/
/***   rcinp:   the reciprocal space control parameters                    ***/
/***   Q:       the reciprocal space charge / potential mesh               ***/
/***   pmmap:   the reciprocal space particle<-->mesh spline map           ***/
/***   PPk:     the reciprocal space pair potential mesh construction kit  ***/
/***=======================================================================***/
void InitVelocities(coord *crd, cellgrid *CG, prmtop *tp, dircon *dcinp,
		    FrcTab *Etab, FrcTab *EHtab, reccon *rcinp, bmap *pmmap,
		    bckit *PPk, trajcon *tj, Energy *sysUV, execon *etimers,
		    cdftrj* Acdf)
{
  int i;

  /*** Random initial velocities ***/
  if (tj->irest == 0) {
    if (tj->Tinit > 0.0) {
      AndersenThermostat(crd, tp, tj, tj->Tinit);
    }
    else {
      AndersenThermostat(crd, tp, tj, tj->Ttarget);
    }
  }

  /*** Initialize coordinates arrays not present ***/
  /*** in the checkpoint information             ***/
  for (i = 0; i < 3*crd->natom; i++) {
    crd->prvloc[i] = crd->loc[i];
    crd->prvfrc[i] = 0.0;
  }

  /*** Constraints are applied to positions simply for the      ***/
  /*** purpose of overcoming imprecision in the checkpoint data ***/
  /*** if this is a restart from a simulation in progress.      ***/
  PositionConstraints(crd, tp, 1.0/tj->dt);
  InitializeEnergy(sysUV, tj, tp, 0);
  AtomForces(crd, CG, tp, dcinp, Etab, EHtab, rcinp, pmmap, PPk, sysUV,
             etimers);

  /*** Velocity Verlet, update velocities ***/
  VelVerletV(crd, tp, tj->dt);
  etimers->Integ += mdgxStopTimer(etimers);

  /*** Write restart file ***/
  if (tj->ntwr != 0 && tj->currstep % tj->ntwr == 0 &&
      tj->currstep >= tj->ntwr) {
    WriteRst(crd, tp, tj);
  }

  /*** Velocity constraints ***/
  if (tj->ntp == 1) {
    ReflectDVec(crd->prvvel, crd->vel, 3*tp->natom);
  }
  VelocityConstraints(crd, tp);
  if (tj->ntp == 1) {
    VVConstraintVirial(crd, tp, tj, sysUV);
  }

  /*** Compute kinetic energy after the force computation   ***/
  /*** in preparation for Andersen or Berendsen thermostats ***/
  if (tj->ntt == 1 || tj->ntt == 2) {
    sysUV->T = CurrentSystemTemperature(crd, tp, sysUV, 1);
  }

  /*** Write trajectory files and outputs ***/
  WriteTrajFiles(tj, tp, dcinp, rcinp, Etab, crd, sysUV, etimers, Acdf);
  etimers->Write += mdgxStopTimer(etimers);

  /*** Berendsen thermostat ***/
  if (tj->ntt == 1) {
    BerendsenThermostat(crd, tp, tj, sysUV);
  }

  /*** Andersen thermostat ***/
  if (tj->ntt == 2 && tj->currstep % tj->vrand == 0) {
    AndersenThermostat(crd, tp, tj, tj->Ttarget);
  }

  /*** Berendsen barostat ***/
  if (tj->ntp == 1) {
    BerendsenBarostat(crd, tp, tj, sysUV, dcinp, rcinp, PPk, CG);
  }

  /*** Monte-Carlo barostat ***/
  if (tj->ntp == 5 && tj->currstep % tj->MCBarostatFreq == 0) {
    MonteCarloBarostat(crd, tp, tj, sysUV, dcinp, Etab, EHtab, rcinp, pmmap,
		       PPk, CG, etimers);
  }

  /*** Remove system momentum ***/
  if (tj->RemoveMomentum > 0 && tj->currstep % tj->RemoveMomentum == 0) {
    RemoveSystemMomentum(crd, tp);
  }

  /*** If we have restarted dynamics, this counts as a step ***/
  if (tj->irest == 1) {
    UpdateStepNumber(tj, tj->currstep + 1);
  }
}

/***=======================================================================***/
/*** Dynamics: routine to manage one step of dynamics, whether that be in  ***/
/***           an ensemble of constant volume, temperature, or both.       ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:     the coordinates                                            ***/
/***   CG:      the cell grid                                              ***/
/***   tp:      the topology                                               ***/
/***   dcinp:   the direct space control parameters                        ***/
/***   Etab:    the direct space electrostatic interaction spline table    ***/
/***   EHtab:   the high-resolution direct space electrostatic interaction ***/
/***            spline table                                               ***/
/***   rcinp:   the reciprocal space control parameters                    ***/
/***   Q:       the reciprocal space charge / potential mesh               ***/
/***   pmmap:   the reciprocal space particle<-->mesh spline map           ***/
/***   PPk:     the reciprocal space pair potential mesh construction kit  ***/
/***   tj:      trajectory control parameters                              ***/
/***=======================================================================***/
void Dynamics(coord *crd, cellgrid *CG, prmtop *tp, dircon *dcinp,
	      FrcTab *Etab, FrcTab *EHtab, reccon *rcinp, bmap *pmmap,
	      bckit *PPk, trajcon *tj, Energy *sysUV, execon *etimers,
	      cdftrj* Acdf)
{
  double CHIt, CHItp1e, CHItp1q, CHItp3e, CHItp1h, CHItp5e, CHItp3q, CHItp7e;
  double ETAt, ETAtp3q, ETAtp1h, ETAtp1q;

  /*** Velocity Verlet, update positions ***/
  VelVerletC(crd, tp, tj->dt);
  etimers->Integ += mdgxStopTimer(etimers);

  /*** Position constraints ***/
  if (tj->ntp == 1) {
    ReflectDVec(crd->prvvel, crd->vel, 3*tp->natom);
  }
  PositionConstraints(crd, tp, 1.0/tj->dt);
  if (tj->ntp == 1) {
    VVConstraintVirial(crd, tp, tj, sysUV);
  }

  /*** Compute forces ***/
  InitializeEnergy(sysUV, tj, tp, 0);
  AtomForces(crd, CG, tp, dcinp, Etab, EHtab, rcinp, pmmap, PPk, sysUV,
	     etimers);

  /*** Velocity Verlet, update velocities ***/
  VelVerletV(crd, tp, tj->dt);
  etimers->Integ += mdgxStopTimer(etimers);

  /*** Write restart file ***/
  if (tj->ntwr != 0 && tj->currstep % tj->ntwr == 0 &&
      tj->currstep >= tj->ntwr) {
    WriteRst(crd, tp, tj);
  }

  /*** Velocity constraints ***/
  if (tj->ntp == 1) {
    ReflectDVec(crd->prvvel, crd->vel, 3*tp->natom);
  }
  VelocityConstraints(crd, tp);
  if (tj->ntp == 1) {
    VVConstraintVirial(crd, tp, tj, sysUV);
  }

  /*** Compute kinetic energy after the force computation   ***/
  /*** in preparation for Andersen or Berendsen thermostats ***/
  if (tj->ntt == 1 || tj->ntt == 2) {
    sysUV->T = CurrentSystemTemperature(crd, tp, sysUV, 1);
  }

  /*** Write trajectory files and outputs ***/
  WriteTrajFiles(tj, tp, dcinp, rcinp, Etab, crd, sysUV, etimers, Acdf);
  etimers->Write += mdgxStopTimer(etimers);

  /*** Berendsen thermostat ***/
  if (tj->ntt == 1) {
    BerendsenThermostat(crd, tp, tj, sysUV);
  }

  /*** Andersen thermostat ***/
  if (tj->ntt == 2 && tj->currstep % tj->vrand == 0) {
    AndersenThermostat(crd, tp, tj, tj->Ttarget);
  }

  /*** Nose-Hoover thermostat ***/
  if (tj->ntt == 4) {
    if (tj->ntp == 0) {
      NHThermostat(crd, tp, tj, sysUV, 0.0, &CHItp1h, &CHItp3q, &CHIt, 0);
    }
    tj->npth.chi = CHIt;
  }

  /*** Berendsen barostat ***/
  if (tj->ntp == 1) {
    BerendsenBarostat(crd, tp, tj, sysUV, dcinp, rcinp, PPk, CG);
  }

  /*** Monte-Carlo barostat ***/
  if (tj->ntp == 5 && tj->currstep % tj->MCBarostatFreq == 0) {
    MonteCarloBarostat(crd, tp, tj, sysUV, dcinp, Etab, EHtab, rcinp, pmmap,
		       PPk, CG, etimers);
  }

  /*** Remove system momentum ***/
  if (tj->RemoveMomentum > 0 && tj->currstep % tj->RemoveMomentum == 0) {
    RemoveSystemMomentum(crd, tp);
  }
}

/***=======================================================================***/
/*** PrintTriLine: this function just helps me to print out numbers in     ***/
/***               triplets, line by line.                                 ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   val:   the values to print                                          ***/
/***   n:     the number of triplets                                       ***/
/***   outp:  the file to print values into                                ***/
/***=======================================================================***/
static void PrintTriLine(double* val, int n, FILE *outp)
{
  int i;

  for (i = 0; i < n; i++) {
    fprintf(outp, "%20.13e %20.13e %20.13e\n", val[3*i], val[3*i+1],
	    val[3*i+2]);
  }
}

/***=======================================================================***/
/*** PrepForceReport: this function writes a complete report on forces     ***/
/***                  acting on all atoms.  The breakdown of forces is     ***/
/***                  similar to the breakdown of energies shown in the    ***/
/***                  diagnostics / state information output file.         ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:     the coordinates (for velocity information)                 ***/
/***   CG:      the cell grid                                              ***/
/***   tp:      the topology (for masses)                                  ***/
/***   dcinp:   the direct space parameters                                ***/
/***   Etab:    the direct space electrostatic interaction spline table    ***/
/***   EHtab:   the high-resolution direct space electrostatic interaction ***/
/***            spline table                                               ***/
/***   rcinp:   the reciprocal space control parameters                    ***/
/***   Q:       the reciprocal space charge / potential mesh               ***/
/***   pmmap:   the reciprocal space particle<-->mesh spline map           ***/
/***   PPk:     the reciprocal space pair potential mesh construction kit  ***/
/***   sysUV:   state information data (potential energy and virial)       ***/
/***=======================================================================***/
void PrepForceReport(coord *crd, cellgrid *CG, prmtop *tp, dircon *dcinp,
                     FrcTab *Etab, FrcTab *EHtab, reccon *rcinp, bmap *pmmap,
		     bckit *PPk, trajcon *tj, execon *etimers)
{
  int i, natm;
  double* bondVir;
  double* anglVir;
  double* diheVir;
  double* vdwVir;
  double* relecVir;
  double* delecVir;
  double *dtmp;
  double* bondfrc;
  double* anglfrc;
  double* dihefrc;
  double* relecfrc;
  double* delecfrc;
  double* vdwfrc;
  FILE *outp;
  prmtop tpX;
  Energy sysUV, sysUVt;
  time_t ct;

  /*** Begin printing the output file ***/
  outp = fopen(tj->dumpname, "w");
  ct = time(&ct);
  fprintf(outp, "%% This is an mdgx force report.  It contains information on "
	  "all types of\n%% forces acting on the system's atoms, much like a "
	  "force dump file written by\n%% SANDER.\n\n%% Compilation date: %s"
	  "\n%% Input file text:\n\n", asctime(localtime(&ct)));
  for (i = 0; i < tj->inptext.row; i++) {
    fprintf(outp, "%% %s", tj->inptext.map[i]);
  }
  fprintf(outp, "\n");
  fclose(outp);

  /*** Forces, energies, and virials are all wanted ***/
  InitializeEnergy(&sysUV, tj, tp, 0);
  InitializeEnergy(&sysUVt, tj, tp, 0);
  sysUV.updateU = 2;
  sysUVt.updateU = 2;

  /*** Make a copy of the topology ***/
  tpX = CopyTopology(tp);
  natm = tp->natom;
  dtmp = crd->frc;

  /*** Remove nonbonded interactions from consideration ***/
  SetDVec(tpX.Charges, natm, 0.0);
  SetDVec(tpX.LJftab.data, 2*tp->ntypes*tp->ntypes, 0.0);
  SetDVec(tpX.LJutab.data, 2*tp->ntypes*tp->ntypes, 0.0);

  /*** Compute bonded interactions ***/
  bondfrc = (double*)calloc(3*natm, sizeof(double));
  crd->frc = bondfrc;
  for (i = 0; i < tpX.nBAH.ndihe; i++) {
    tpX.HParam[i].K = 0.0;
  }
  for (i = 0; i < tpX.nBAH.nangl; i++) {
    tpX.AParam[i].K = 0.0;
  }
  ListBondedIntr(crd, &tpX, Etab, EHtab, &sysUVt);
  sysUV.bond = sysUVt.bond;
  bondVir = CpyDVec(sysUVt.Vir, 9);
  SetDVec(sysUVt.Vir, 9, 0.0);

  /*** Compute angle interactions ***/
  anglfrc = (double*)calloc(3*natm, sizeof(double));
  crd->frc = anglfrc;
  for (i = 0; i < tpX.nBAH.nbond; i++) {
    tpX.BParam[i].K = 0.0;
  }
  for (i = 0; i < tpX.nBAH.nangl; i++) {
    tpX.AParam[i].K = tp->AParam[i].K;
  }
  ListBondedIntr(crd, &tpX, Etab, EHtab, &sysUVt);
  sysUV.angl = sysUVt.angl;
  anglVir = CpyDVec(sysUVt.Vir, 9);
  SetDVec(sysUVt.Vir, 9, 0.0);

  /*** Compute dihedral interactions ***/
  dihefrc = (double*)calloc(3*natm, sizeof(double));
  crd->frc = dihefrc;
  for (i = 0; i < tpX.nBAH.ndihe; i++) {
    tpX.HParam[i].K = tp->HParam[i].K;
  }
  for (i = 0; i < tpX.nBAH.nangl; i++) {
    tpX.AParam[i].K = 0.0;
  }
  ListBondedIntr(crd, &tpX, Etab, EHtab, &sysUVt);
  sysUV.dihe = sysUVt.dihe;
  diheVir = CpyDVec(sysUVt.Vir, 9);
  SetDVec(sysUVt.Vir, 9, 0.0);

  /*** Reciprocal space force calculation ***/
  relecfrc = (double*)calloc(3*natm, sizeof(double));
  crd->frc = relecfrc;
  for (i = 0; i < natm; i++) {
    tpX.Charges[i] = tp->Charges[i];
  }
  SplCoeff(crd, pmmap, rcinp);
  Charges2Book(pmmap, rcinp, &tpX, &rcinp->QL[0]);
  if (rcinp->nlev == 1) {
    ConvQBCnrg(rcinp, crd, &rcinp->QL[0], PPk, &sysUVt, etimers);
  }
  else {
    mleConvQBC(rcinp, &sysUVt, crd, PPk, etimers);
  }
  IntrpFrc(pmmap, crd, rcinp, &tpX, &rcinp->QL[0]);
  sysUV.relec = sysUVt.relec;
  relecVir = CpyDVec(sysUVt.Vir, 9);
  SetDVec(sysUVt.Vir, 9, 0.0);

  /*** Direct space electrostatic calculation ***/
  delecfrc = (double*)calloc(3*natm, sizeof(double));
  crd->frc = delecfrc;
  for (i = 0; i < tpX.nBAH.ndihe; i++) {
    tpX.HParam[i].K = 0.0;
  }
  AtomsToCellsOrtho(crd, dcinp, CG, &tpX);
  ShareTriple2(CG, dcinp->Mcut, crd, &tpX);
  for (i = 0; i < CG->ng[0]*CG->ng[1]*CG->ng[2]; i++) {
    DirectTriple2(&CG->data[i], CG->celldim, Etab, dcinp, &tpX, &sysUVt);
  }
  MergeTriple2(CG);
  MapCellsToAtoms(CG, crd);
  ListBondedIntr(crd, &tpX, Etab, EHtab, &sysUVt);
  sysUV.delec = sysUVt.delec;
  delecVir = CpyDVec(sysUVt.Vir, 9);
  SetDVec(sysUVt.Vir, 9, 0.0);

  /*** Direct space vdW force calculation ***/
  vdwfrc = (double*)calloc(3*natm, sizeof(double));
  crd->frc = vdwfrc;
  SetDVec(tpX.Charges, natm, 0.0);
  for (i = 0; i < 2*tp->ntypes*tp->ntypes; i++) {
    tpX.LJftab.data[i] = tp->LJftab.data[i];
    tpX.LJutab.data[i] = tp->LJutab.data[i];
  }
  AtomsToCellsOrtho(crd, dcinp, CG, &tpX);
  ShareTriple2(CG, dcinp->Mcut, crd, &tpX);
  for (i = 0; i < CG->ng[0]*CG->ng[1]*CG->ng[2]; i++) {
    DirectTriple2(&CG->data[i], CG->celldim, Etab, dcinp, &tpX, &sysUVt);
  }
  MergeTriple2(CG);
  MapCellsToAtoms(CG, crd);
  ListBondedIntr(crd, &tpX, Etab, EHtab, &sysUVt);
  sysUV.vdw12 = sysUVt.vdw12;
  sysUV.vdw6 = sysUVt.vdw6;
  vdwVir = CpyDVec(sysUVt.Vir, 9);
  SetDVec(sysUVt.Vir, 9, 0.0);

  /*** Free the extra allocated topology, sum forces and virials ***/
  FreeTopology(&tpX);
  DVec2VecAdd(sysUV.Vir, bondVir, 9);
  DVec2VecAdd(sysUV.Vir, anglVir, 9);
  DVec2VecAdd(sysUV.Vir, diheVir, 9);
  DVec2VecAdd(sysUV.Vir, delecVir, 9);
  DVec2VecAdd(sysUV.Vir, relecVir, 9);
  DVec2VecAdd(sysUV.Vir, vdwVir, 9);
  for (i = 0; i < 3*natm; i++) {
    dtmp[i] = bondfrc[i] + anglfrc[i] + dihefrc[i] + relecfrc[i] + 
      delecfrc[i] + vdwfrc[i];
  }
  crd->frc = dtmp;

  /*** Write all forces and energies ***/
  outp = fopen(tj->dumpname, "a");
  fprintf(outp, "%s.Ubond = %20.13e; %% Bond interaction energy\n",
	  tj->DMPvar, sysUV.bond);
  fprintf(outp, "%s.Uangl = %20.13e; %% Angle interaction energy\n",
	  tj->DMPvar, sysUV.angl);
  fprintf(outp, "%s.Udihe = %20.13e; %% Dihedral interaction energy\n",
	  tj->DMPvar, sysUV.dihe);
  fprintf(outp, "%s.Udir  = %20.13e; %% Direct space sum electrostatic "
	  "energy\n", tj->DMPvar, sysUV.delec);
  fprintf(outp, "%s.Urec  = %20.13e; %% Reciprocal space sum "
	  "electrostatic energy\n", tj->DMPvar, sysUV.relec);
  fprintf(outp, "%s.Uvdw  = %20.13e; %% van-der Waals energy\n", tj->DMPvar,
	  sysUV.vdw12 + sysUV.vdw6);
  if (tj->DMPcrd == 1) {
    fprintf(outp, "\n%% Coordinates:\n%s.crd = [\n", tj->DMPvar);
    PrintTriLine(crd->loc, tp->natom, outp);
    fprintf(outp, "];\n");
    fprintf(outp, "\n%% Box dimensions (X, Y, Z, followed by a, b, g in "
	    "radians):\n%s.box = [\n", tj->DMPvar);
    PrintTriLine(crd->gdim, 2, outp);
    fprintf(outp, "];\n");
  }
  if (tj->DMPbond == 1) {
    fprintf(outp, "\n%% Virial contributions from bonds\n%s.bond_vir = [\n",
	    tj->DMPvar);
    PrintTriLine(bondVir, 3, outp);
    fprintf(outp, "];\n");
    fprintf(outp, "\n%% Forces due to bonds\n%s.bond_frc = [\n", tj->DMPvar);
    PrintTriLine(bondfrc, tp->natom, outp);
    fprintf(outp, "];\n");
  }
  if (tj->DMPangl == 1) {
    fprintf(outp, "\n%% Virial contributions from angles\n%s.angl_vir = [\n",
	    tj->DMPvar);
    PrintTriLine(anglVir, 3, outp);
    fprintf(outp, "];\n");
    fprintf(outp, "\n%% Forces due to angles\n%s.angl_frc = [\n", tj->DMPvar);
    PrintTriLine(anglfrc, tp->natom, outp);
    fprintf(outp, "];\n");
  }
  if (tj->DMPdihe == 1) {
    fprintf(outp, "\n%% Virial contributions from dihedrals\n%s.dihe_vir = "
	    "[\n", tj->DMPvar);
    PrintTriLine(diheVir, 3, outp);
    fprintf(outp, "];\n");
    fprintf(outp, "\n%% Forces due to dihedrals\n%s.dihe_frc = [\n",
	    tj->DMPvar);
    PrintTriLine(dihefrc, tp->natom, outp);
    fprintf(outp, "];\n");
  }
  if (tj->DMPdelec == 1) {
    fprintf(outp, "\n%% Virial contributions from direct space electrostatics"
	    "\n%s.delec_vir = [\n", tj->DMPvar);
    PrintTriLine(delecVir, 3, outp);
    fprintf(outp, "];\n");
    fprintf(outp, "\n%% Forces due to electrostatic direct space sum\n"
	    "%s.delec_frc = [\n", tj->DMPvar);
    PrintTriLine(delecfrc, tp->natom, outp);
    fprintf(outp, "];\n");
  }
  if (tj->DMPrelec == 1) {
    fprintf(outp, "\n%% Virial contributions from reciprocal space "
	    "electrostatics\n%s.relec_vir = [\n", tj->DMPvar);
    PrintTriLine(relecVir, 3, outp);
    fprintf(outp, "];\n");
    fprintf(outp, "\n%% Forces due to electrostatic reciprocal space sum\n"
	    "%s.relec_frc = [\n", tj->DMPvar);
    PrintTriLine(relecfrc, tp->natom, outp);
    fprintf(outp, "];\n");
  }
  if (tj->DMPvdw == 1) {
    fprintf(outp, "\n%% Virial contributions from van der Waals interactions"
	    "\n%s.vdw_vir = [\n", tj->DMPvar);
    PrintTriLine(vdwVir, 3, outp);
    fprintf(outp, "];\n");
    fprintf(outp, "\n%% Forces due to van-der Waals interactions\n"
	    "%s.vdw_frc = [\n", tj->DMPvar);
    PrintTriLine(vdwfrc, tp->natom, outp);
    fprintf(outp, "];\n");
  }
  if (tj->DMPall == 1) {
    fprintf(outp, "\n%% Total system virial\n%s.sum_vir = [\n", tj->DMPvar);
    PrintTriLine(sysUV.Vir, 3, outp);
    fprintf(outp, "];\n");
    fprintf(outp, "\n%% Total force on all atoms\n"
	    "%s.sum_frc = [\n", tj->DMPvar);
    PrintTriLine(crd->frc, tp->natom, outp);
    fprintf(outp, "];\n");
  }

  /*** Close up the output file ***/
  fclose(outp);
}
