/***=======================================================================***/
/*** CompIntrVSSE: compute all interactions for two lists of particles.    ***/
/***               In the event that the lists contain the same atoms, the ***/
/***               inner loop limits will be adjusted to avoid double-     ***/
/***               counting.  In the "VgtE" version, the van-der Waals     ***/
/***               Waals cutoff is greater than the electrostatic          ***/
/***               direct-space cutoff.  If the two cutofss are in fact    ***/
/***               identical, the more efficient "VeqE" version is used    ***/
/***               instead.  The "CR" version implements range checking    ***/
/***               van-der Waals interactions that might take place at     ***/
/***               very close range, but this check is not needed for      ***/
/***               many cases.  The "nrg" extension of the function is     ***/
/***               used when a potential energy computation is needed, the ***/
/***               "nrgvir" extension is used when virial calculations are ***/
/***               required, and the "nrgx" extension is used when energy, ***/
/***               and no other property, is needed.                       ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   atm1/2:  atom list 1 or 2                                           ***/
/***   ordr1/2: the order in which to access atoms from list 1             ***/
/***   natm1/2: the number of atoms in list 1                              ***/
/***   EFrc:    the electrostatic potential spline table (the force is     ***/
/***            obtained as the derivative of the spline)                  ***/
/***   dcinp:   the direct space input parameters                          ***/
/***=======================================================================***/
#if NEEDFORCE == 1
  #if NEEDENERGY == 0
    #if VGTE == 1 && CHKRANGE == 1
      #define COMPINTRVSSE CompIntrVgtECR
    #elif VGTE == 1 && CHKRANGE == 0
      #define COMPINTRVSSE CompIntrVgtE
    #elif VGTE == 0 && CHKRANGE == 1
      #define COMPINTRVSSE CompIntrVeqECR
    #else
      #define COMPINTRVSSE CompIntrVeqE
    #endif
  #else
    #if NEEDVIRIAL == 0
      #if VGTE == 1 && CHKRANGE == 1
        #define COMPINTRVSSE CompIntrVgtECRnrg
      #elif VGTE == 1 && CHKRANGE == 0
        #define COMPINTRVSSE CompIntrVgtEnrg
      #elif VGTE == 0 && CHKRANGE == 1
        #define COMPINTRVSSE CompIntrVeqECRnrg
      #else
        #define COMPINTRVSSE CompIntrVeqEnrg
      #endif
    #else
      #if VGTE == 1 && CHKRANGE == 1
        #define COMPINTRVSSE CompIntrVgtECRnrgvir
      #elif VGTE == 1 && CHKRANGE == 0
        #define COMPINTRVSSE CompIntrVgtEnrgvir
      #elif VGTE == 0 && CHKRANGE == 1
        #define COMPINTRVSSE CompIntrVeqECRnrgvir
      #else
        #define COMPINTRVSSE CompIntrVeqEnrgvir
      #endif
    #endif
  #endif
#else
  #if VGTE == 1 && CHKRANGE == 1
    #define COMPINTRVSSE CompIntrVgtECRnrgx
  #elif VGTE == 1 && CHKRANGE == 0
    #define COMPINTRVSSE CompIntrVgtEnrgx
  #elif VGTE == 0 && CHKRANGE == 1
    #define COMPINTRVSSE CompIntrVeqECRnrgx
  #else
    #define COMPINTRVSSE CompIntrVeqEnrgx
  #endif
#endif

#if NEEDENERGY == 0
void COMPINTRVSSE(atomc *atm1, atomc *atm2, int* ordr1, int* ordr2, int natm1,
                  int natm2, FrcTab *EFrc, dircon *dcinp, prmtop *tp)
#else
void COMPINTRVSSE(atomc *atm1, atomc *atm2, int* ordr1, int* ordr2, int natm1,
                  int natm2, FrcTab *EFrc, dircon *dcinp, prmtop *tp,
		  Energy *sysU)
#endif
{
  int i, j, ir;
  int aiid, ajid, ailjt, ajljt;
#if NEEDFORCE == 1
  double aifx, aify, aifz;
  double fx, fy, fz, fmag;
#endif
  double aiq;
  double invr2, invr4;
  double *ljftmp;
  double **ljfmap;
  CSpln *Fspl;
#if NEEDENERGY == 1
  double uelec, uvdwr6, uvdwr12;
  double *ljutmp;
  double **ljumap;
  CSpln *Uspl;
#endif

  /*** Pre-compute a bunch of constants and unpack structures ***/
  const int listsame = (natm1 == natm2 && atm1 == atm2) ? 1 : 0;
  const int imax = (listsame == 1) ? natm1-1 : natm1;
  const double Vcut2 = dcinp->Vcut*dcinp->Vcut;
  const double Ecut2 = dcinp->Ecut*dcinp->Ecut;
  const double uspc = EFrc->ivdr;
#if NEEDFORCE == 1
  ljfmap = tp->LJftab.map;
  Fspl = EFrc->dSD;
#endif
#if NEEDENERGY == 1
  ljumap = tp->LJutab.map;
  Uspl = EFrc->SD;
  uelec = 0.0;
  uvdwr6 = 0.0;
  uvdwr12 = 0.0;
#endif

  /*** The outer direct space loop ***/
  for (i = 0; i < imax; i++) {
    const int ni = ordr1[i];
    const double atmx = atm1[ni].loc[0];
    const double atmy = atm1[ni].loc[1];
    const double atmz = atm1[ni].loc[2];
    const int jmin = (listsame == 1) ? i+1 : 0;

#if NEEDFORCE == 1
    /*** Force accumulators ***/
    aifx = 0.0;
    aify = 0.0;
    aifz = 0.0;
#endif

    /*** Properties of this atom ***/
    aiq = atm1[ni].q;
    ailjt = atm1[ni].lj;
    if (ailjt < 0) {

      /*** The inner direct space loop, excluding vdW interactions ***/
      for (j = jmin; j < natm2; j++) {
        const int nj = ordr2[j];
        const double dx = atm2[nj].loc[0] - atmx;
        const double dy = atm2[nj].loc[1] - atmy;
        const double dz = atm2[nj].loc[2] - atmz;
        const double r2 = dx*dx + dy*dy + dz*dz;

        if (r2 < Ecut2) {

	  /*** Compute the electrostatic force ***/
          ir = r2*uspc;
#if NEEDFORCE
          fmag = (((Fspl[ir].A*r2 + Fspl[ir].B)*r2 + Fspl[ir].C)*r2 +
                  Fspl[ir].D)*aiq*atm2[nj].q;

          /*** Accumulate the force ***/
          fx = fmag*dx;
          fy = fmag*dy;
          fz = fmag*dz;
          aifx += fx;
          aify += fy;
          aifz += fz;
          atm2[nj].frc[0] -= fx;
          atm2[nj].frc[1] -= fy;
          atm2[nj].frc[2] -= fz;
#endif
#if NEEDENERGY == 1
	  uelec += (((Uspl[ir].A*r2 + Uspl[ir].B)*r2 + Uspl[ir].C)*r2 +
		    Uspl[ir].D)*aiq*atm2[nj].q;
#endif

#if NEEDENERGY == 1 && NEEDVIRIAL == 1
	  /*** Accumulate the virial ***/
	  sysU->Vir[0] += dx*fmag*dx;
	  sysU->Vir[1] += dx*fmag*dy;
	  sysU->Vir[2] += dx*fmag*dz;
	  sysU->Vir[3] += dy*fmag*dx;
	  sysU->Vir[4] += dy*fmag*dy;
	  sysU->Vir[5] += dy*fmag*dz;
	  sysU->Vir[6] += dz*fmag*dx;
	  sysU->Vir[7] += dz*fmag*dy;
	  sysU->Vir[8] += dz*fmag*dz;
#ifdef IDEBUG
          /*** Accumulate the virial specific to       ***/
	  /*** electrostatic direct space interactions ***/
          sysU->EdirVir[0] += dx*fmag*dx;
          sysU->EdirVir[1] += dx*fmag*dy;
          sysU->EdirVir[2] += dx*fmag*dz;
          sysU->EdirVir[3] += dy*fmag*dx;
          sysU->EdirVir[4] += dy*fmag*dy;
          sysU->EdirVir[5] += dy*fmag*dz;
          sysU->EdirVir[6] += dz*fmag*dx;
          sysU->EdirVir[7] += dz*fmag*dy;
          sysU->EdirVir[8] += dz*fmag*dz;
#endif
#endif
        }
      }
    }
    else {

      /*** The inner direct space loop, including vdW interactions ***/
#if NEEDFORCE == 1
      ljftmp = ljfmap[ailjt];
#endif
#if NEEDENERGY == 1
      ljutmp = ljumap[ailjt];
#endif
      aiid = atm1[ni].id;
      for (j = jmin; j < natm2; j++) {
        const int nj = ordr2[j];
        const double dx = atm2[nj].loc[0] - atmx;
        const double dy = atm2[nj].loc[1] - atmy;
        const double dz = atm2[nj].loc[2] - atmz;
        const double r2 = dx*dx + dy*dy + dz*dz;

        if (r2 < Vcut2) {

          /*** Always compute the electrostatic interaction by table ***/
#if NEEDFORCE == 1
	  fmag = 0.0;
#endif
#if VGTE == 1
          if (r2 < Ecut2) {
#endif
            ir = r2*uspc;
#if NEEDFORCE == 1
            fmag = (((Fspl[ir].A*r2 + Fspl[ir].B)*r2 + Fspl[ir].C)*r2 +
                    Fspl[ir].D)*aiq*atm2[nj].q;
#endif
#if NEEDENERGY == 1
	    uelec += (((Uspl[ir].A*r2 + Uspl[ir].B)*r2 + Uspl[ir].C)*r2 +
		      Uspl[ir].D)*aiq*atm2[nj].q;
#endif
#if VGTE == 1
          }
#endif

          /*** Identification number of this atom ***/
	  ajljt = atm2[nj].lj;
          ajid = atm2[nj].id;

          /*** Check for range before computing the LJ interaction ***/
#if CHKRANGE == 1
          if (ajljt >= 0 &&
	      (r2 > MINNB2 || TestBondedExclusion(aiid, ajid, tp) == 0)) {
#else
          if (ajljt >= 0) {
#endif

            /*** Compute the Lennard-Jones interaction ***/
            invr2 = 1.0/r2;
            invr4 = invr2*invr2;
#if NEEDFORCE == 1
            fmag += invr4*invr4*
              (ljftmp[2*ajljt]*invr4*invr2 + ljftmp[2*ajljt+1]);

#endif
#if NEEDENERGY == 1
	    uvdwr12 += invr4*invr2*invr4*invr2*ljutmp[2*ajljt];
	    uvdwr6 += invr4*invr2*ljutmp[2*ajljt+1];
#endif
#if CHKRANGE == 1
          }
#else
	  }
#endif
      
#if NEEDFORCE == 1
          /*** Accumulate the force ***/
          fx = fmag*dx;
          fy = fmag*dy;
          fz = fmag*dz;
          aifx += fx;
          aify += fy;
          aifz += fz;
          atm2[nj].frc[0] -= fx;
          atm2[nj].frc[1] -= fy;
          atm2[nj].frc[2] -= fz;
#endif

#if NEEDENERGY == 1 && NEEDVIRIAL == 1
	  /*** Accumulate the virial ***/
	  sysU->Vir[0] += dx*fmag*dx;
	  sysU->Vir[1] += dx*fmag*dy;
	  sysU->Vir[2] += dx*fmag*dz;
	  sysU->Vir[3] += dy*fmag*dx;
	  sysU->Vir[4] += dy*fmag*dy;
	  sysU->Vir[5] += dy*fmag*dz;
	  sysU->Vir[6] += dz*fmag*dx;
	  sysU->Vir[7] += dz*fmag*dy;
	  sysU->Vir[8] += dz*fmag*dz;
#ifdef IDEBUG
          /*** Accumulate the virial specific to       ***/
	  /*** electrostatic direct space interactions ***/
          sysU->EdirVir[0] += dx*fmag*dx;
          sysU->EdirVir[1] += dx*fmag*dy;
          sysU->EdirVir[2] += dx*fmag*dz;
          sysU->EdirVir[3] += dy*fmag*dx;
          sysU->EdirVir[4] += dy*fmag*dy;
          sysU->EdirVir[5] += dy*fmag*dz;
          sysU->EdirVir[6] += dz*fmag*dx;
          sysU->EdirVir[7] += dz*fmag*dy;
          sysU->EdirVir[8] += dz*fmag*dz;
#endif
#endif
        }
      }
    }

#if NEEDFORCE == 1
    atm1[ni].frc[0] += aifx;
    atm1[ni].frc[1] += aify;
    atm1[ni].frc[2] += aifz;
#endif
  }
#if NEEDENERGY == 1

  /*** Accumulate the energy ***/
  sysU->vdw12 += uvdwr12;
  sysU->vdw6 += uvdwr6;
  sysU->delec += uelec;
#endif
}

#undef COMPINTRVSSE
