#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "Random.h"
#include "mdgxVector.h"

#include "TrajectoryDS.h"
#include "TopologyDS.h"
#include "CrdManipDS.h"

/***=======================================================================***/
/*** CurrentKineticEnergy: compute the current kinetic energy.             ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:     the coordinates (for velocity information)                 ***/
/***   tp:      the topology (for masses)                                  ***/
/***=======================================================================***/
double CurrentKineticEnergy(coord *crd, prmtop *tp)
{
  int i, natm;
  double result, vx, vy, vz;
  double *vel, *mass;

  natm = crd->natom;
  vel = crd->vel;
  mass = tp->Masses;
  result = 0.0;
  for (i = 0; i < natm; i++) {
    vx = vel[3*i];
    vy = vel[3*i+1];
    vz = vel[3*i+2];
    result += mass[i]*(vx*vx + vy*vy + vz*vz);
  }

  return 0.5 * result / 418.4;
}

/***=======================================================================***/
/*** CurrentSystemTemperature: computes the instantaneous temperature of   ***/
/***                           the system using the particle velocities.   ***/
/***                           Assumes the system has no net velocity.     ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:       the coordinates (for velocity information)               ***/
/***   tp:        the topology (for masses)                                ***/
/***   sysUV:     the system energy (and virial) record                    ***/
/***   updateKE:  flag to recompute kinetic energy (if it may be stale)    ***/
/***=======================================================================***/
double CurrentSystemTemperature(coord *crd, prmtop *tp, Energy *sysUV,
				int updateKE)
{
  if (updateKE == 1) {
    sysUV->kine = CurrentKineticEnergy(crd, tp);
  }

  return 2.0*sysUV->kine/(tp->ndf * GASCNST);
}

/***=======================================================================***/
/*** BerendsenThermostat: the dirt-simple Berendsen thermostat.  Beware,   ***/
/***                      this does not sample the canonical ensemble!     ***/
/***                      Clamped the rescaling values were introduced to  ***/
/***                      prevent catastrophic rescaling.                  ***/
/***                                                                       ***/
/***                      Reference: Allen and Tildesley, Eq. 7.59, pg 231 ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:     the coordinates (for velocity information)                 ***/
/***   tp:      the topology (for masses)                                  ***/
/***   tj:      trajectory control information                             ***/
/***   T:       the temperature target                                     ***/
/***=======================================================================***/
void BerendsenThermostat(coord *crd, prmtop *tp, trajcon *tj, Energy *sysUV)
{
  int i, n;
  double tscale, chi, T;

  /*** Temperature was computed after the force computation; ***/
  /*** velocities have been updated since the temperature    ***/
  /*** computation but that is part of the algorithm.        ***/
  n = 3*crd->natom;
  T = sysUV->T;

  /*** Brrr!  Let's call it half the target temperature ***/
  if (T < 0.5 * tj->Ttarget) {
    tscale = 1.0;
  }

  /*** Too hot!  Let's call it eight times the target temperature ***/
  else if (T > 8.0 * tj->Ttarget) {
    tscale = -0.875;
  }

  /*** Ahh, just right... ***/
  else {
    tscale = (tj->Ttarget / T) - 1.0;
  }
  chi = sqrt(1.0 + tscale * tj->dt / tj->BerendsenTCoupl);

  for (i = 0; i < n; i++) {

    /*** Velocity scaling is not restricted to atoms with mass, ***/
    /*** as atoms without mass (virtual sites) should have zero ***/
    /*** velocity already.                                      ***/
    crd->vel[i] *= chi;
  }
}

/***=======================================================================***/
/*** AndersenThermostat: a routine for re-assigning velocities from a      ***/
/***                     constant-temperature bath.                        ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   crd:     the coordinates (for velocity information)                 ***/
/***   tp:      the topology (for masses)                                  ***/
/***   tj:      trajectory control information                             ***/
/***   T:       the temperature target                                     ***/
/***=======================================================================***/
void AndersenThermostat(coord *crd, prmtop *tp, trajcon *tj, double T)
{
  int i, i3;
  long counter;
  double efac, ebeta;
  double *vtmp;

  /*** Loop over all atoms and reassign velocities ***/
  counter = tj->rndcon;
  ebeta = sqrt(GASCNST*T*418.4);
  vtmp = crd->vel;
  for (i = 0; i < crd->natom; i++) {
    i3 = 3*i;
    efac = ebeta*sqrt(tp->InvMasses[i]);
    vtmp[i3] = efac*GaussBoxMuller(&counter);
    vtmp[i3+1] = efac*GaussBoxMuller(&counter);
    vtmp[i3+2] = efac*GaussBoxMuller(&counter);
  }
  tj->rndcon = counter;
}

/***=======================================================================***/
/*** NHThermostat: a compact function for a nasty series of expressions.   ***/
/***               This implements the repeatedly used thermostat in the   ***/
/***               Nose-Hoover thermo- (baro-)stat above.  Note that if    ***/
/***               there is no barsotat (barostat == 0), then this works   ***/
/***               on a quarter-timestep interval.  If there is, it works  ***/
/***               on an eighth-timestep interval.                         ***/
/***=======================================================================***/
void NHThermostat(coord *crd, prmtop *tp, trajcon *tj, Energy *sysUV,
		  double ETAt, double *CHIt, double *CHItp1e, double *CHItp1q,
                  int barostat)
{
  double dt, Ttarget, pmass, qmass, invqmass, sigma;

  /*** Unpack ***/
  dt = (barostat == 1) ? 0.125*tj->dt : 0.25*tj->dt;
  Ttarget = (barostat == 1) ? GASCNST*tj->Ttarget : 0.0;
  pmass = (barostat == 1) ? ETAt*ETAt*tj->npth.pmass : 0.0;
  qmass = tj->npth.qmass;
  sigma = 2.0*tj->npth.sigma;
  invqmass = 1.0/qmass;

  /*** Compute the kinetic energy ***/
  sysUV->kine = CurrentKineticEnergy(crd, tp);
  *CHItp1e = *CHIt + dt*(2.0*sysUV->kine + pmass - sigma - Ttarget)*invqmass;
  DVecMult(crd->vel, 3*crd->natom, exp(-2.0*dt*(*CHItp1e)));
  sysUV->kine = CurrentKineticEnergy(crd, tp);
  *CHItp1q = *CHItp1e +
    dt*(2.0*sysUV->kine + pmass - sigma - Ttarget)*invqmass;
}

#if 0
/***=======================================================================***/
/*** NPTHBarostat: a compact function for another nasty series of          ***/
/***               expressions in the coupled Nose-Hoover thermo-barostat  ***/
/***               encoded above.                                          ***/
/***=======================================================================***/
void NPTHBarostat(coord *crd, trajcon *tj, double CHIt, double *ETAt,
                  double *ETAtp1q, double *ETAtp1h)
{
  double CHIfac, dt, Vt, Ptarget, invpmass, invqmass;

  /*** Unpack ***/
  dt = rp->dt;
  Ptarget = rp->Ptarget;
  Vt = rp->stv.V;
  invpmass = 1.0/tj->npth.pmass;
  invqmass = 1.0/tj->npth.qmass;

  CHIfac = exp(-0.125*CHIt*dt);
  *ETAt *= CHIfac;

  rp->stv.P = InstantSystemPressure(rp);
  *ETAtp1q = *ETAt + 0.25*dt*(3.0*(rp->stv.P - Ptarget)*Vt)*invpmass;
  *ETAtp1q *= CHIfac;

  DVecMult(ma->vel, 3*ma->natm, exp(-0.5*dt*(*ETAtp1q)));
  rp->nrg.ke = KineticEnergyComp(ma);
  *ETAtp1q *= CHIfac;

  rp->stv.P = InstantSystemPressure(rp);
  *ETAtp1h = *ETAtp1q + 0.25*dt*(3.0*(rp->stv.P - Ptarget)*Vt)*invpmass;
}
#endif

/***=======================================================================***/
/*** PrepThermoBarostat: prepare constants for a thermostat and barostat,  ***/
/***                     if needed.                                        ***/
/***=======================================================================***/
void PrepThermoBarostat(prmtop *tp, trajcon *tj)
{
  tj->npth.sigma = 0.5*tp->ndf*GASCNST*tj->Ttarget;
  tj->npth.qmass = 2.0*tj->npth.sigma*tj->npth.TauT*tj->npth.TauT;
  tj->npth.pmass = (tp->ndf + 3)*GASCNST*tj->Ttarget*pow(tj->npth.TauP, 2.0);
  tj->npth.chi = 0.0;
  tj->npth.eta = 0.0;
}
