#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "Topology.h"
#include "mdgxVector.h"
#include "Matrix.h"
#include "Grid.h"
#include "CrdManip.h"
#include "Constants.h"
#include "Macros.h"
#include "CompFrc.h"
#include "Manual.h"
#include "Parse.h"
#include "VirtualSites.h"
#include "Buckingham.h"

/***=======================================================================***/
/*** RoundTotalCharge: force the total system charge to be an integer      ***/
/***                   value if it is close to an integer multiple of the  ***/
/***                   proton charge to begin with.                        ***/
/***=======================================================================***/
static void RoundTotalCharge(prmtop *tp)
{
  int i;
  double tq;

  /*** Compute the sum of all charges ***/
  tq = DSum(tp->Charges, tp->natom);
  tp->initq = tq;
  if (tq - floor(tq) > 0.1 && ceil(tq) - tq > 0.1 &&
      fabs(tq - floor(tq)) > 1.0e-8 && fabs(ceil(tq) - tq) > 1.0e-8) {
    return;
  }

  /*** If we're still here, the charge is close to an ***/
  /*** integer value and so should be neutralized.    ***/
  tq = (tq - floor(tq) < 0.1) ? floor(tq) - tq : ceil(tq) - tq;
  tq /= tp->natom;
  for (i = 0; i < tp->natom; i++) {
    tp->Charges[i] += tq;
  }
}

/***=======================================================================***/
/*** RecursiveBondSeek: recursive routine for seeking out, and labeling,   ***/
/***                    all atoms connected together by bonds.  This       ***/
/***                    routine contains an internal check to ensure that  ***/
/***                    atoms do not belong to multiple groups.            ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   oatm:   the origin atom                                             ***/
/***   labl:   the number with which to label all atoms connected to oatm  ***/
/***   glist:  the groups list, storing labels for all atoms               ***/
/***   tp:     the system topology                                         ***/
/***=======================================================================***/
void RecursiveBondSeek(int oatm, int labl, int* glist, prmtop *tp)
{
  int i, xatm;

  /*** If the atom is already counted in this group, return ***/
  if (glist[oatm] == labl) {
    return;
  }

  /*** If the atom is already in another group, we're in trouble ***/
  if (glist[oatm] >= 0) {
    printf("RecursiveBondSeek >> Error.  Atom %d is in more than one "
	   "group!\n", oatm);
    exit(1);
  }
  glist[oatm] = labl;
  for (i = 0; i < tp->BLC[oatm].nbond; i++) {
    xatm = tp->BLC[oatm].BC[i].b;
    glist[xatm] = labl;
    RecursiveBondSeek(xatm, labl, glist, tp);
  }
}

/***=======================================================================***/
/*** DefineAtomChains: loop over all atoms, recursively listing all atoms  ***/
/***                   which are linked by bonds.                          ***/
/***=======================================================================***/
static void DefineAtomChains(prmtop *tp)
{
  int i, j, k, ngrp;
  int* glist;

  /*** Set a list of atoms that have been linked ***/
  glist = (int*)malloc(tp->natom*sizeof(int));
  SetIVec(glist, tp->natom, -1);

  ngrp = 0;
  for (i = 0; i < tp->natom; i++) {

    /*** If this atom is already in a group, continue ***/
    if (glist[i] >= 0) {
      continue;
    }

    /*** Recursively find all atoms linked to this one via bonds ***/
    RecursiveBondSeek(i, ngrp, glist, tp);
    ngrp++;
  }

  /*** Write the topology key for linked groups of atoms ***/
  tp->ngrp = ngrp;
  tp->lgrps = (lgrp*)malloc(tp->ngrp*sizeof(lgrp));
  for (i = 0; i < tp->ngrp; i++) {
    tp->lgrps[i].natom = 0;
    for (j = 0; j < tp->natom; j++) {
      if (glist[j] == i) {
	tp->lgrps[i].natom += 1;
      }
    }
    tp->lgrps[i].atoms = (int*)malloc(tp->lgrps[i].natom*sizeof(int));
    k = 0;
    for (j = 0; j < tp->natom; j++) {
      if (glist[j] == i) {
	tp->lgrps[i].atoms[k] = j;
        k++;
      }
    }
  }

  /*** Free allocated memory ***/
  free(glist);
}

/***=======================================================================***/
/*** GetPrmTop: load an AMBER 7 prmtop topology file and return it as a    ***/
/***            prmtop data structure.                                     ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:         the topology structure                                  ***/
/***   adjbnd:     flag to activate number index adjustments, to make the  ***/
/***               Fortran-designed AMBER prmtop readable by a C program   ***/
/***=======================================================================***/
void GetPrmTop(prmtop *tp, int adjbnd)
{
  int i, j, natm;
  FILE *inp;

  /*** Open the source file ***/
  if ((inp = fopen(tp->source, "r")) == NULL) {
    printf("GetPrmTop >> Error.  AMBER prmtop file %s not found!\n",
	   tp->source);
    exit(1);
  }

  /*** Version stamp ***/
  fgets(tp->vstamp, 128, inp);
  
  /*** First, the pointer information ***/
  ReadToFlag("POINTERS", inp, 0);
  fscanf(inp, "%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d",
	 &tp->natom, &tp->ntypes, &tp->withH.nbond, &tp->woH.nbond,
	 &tp->withH.nangl, &tp->woH.nangl, &tp->withH.ndihe, &tp->woH.ndihe,
	 &tp->nhparm, &tp->nparm, &tp->tnexcl, &tp->nres, &tp->woHC.nbond,
	 &tp->woHC.nangl, &tp->woHC.ndihe, &tp->nBAH.nbond, &tp->nBAH.nangl,
	 &tp->nBAH.ndihe, &tp->natyp, &tp->nphb, &tp->ifpert, &tp->pert.nbond,
	 &tp->pert.nangl, &tp->pert.ndihe, &tp->wpert.nbond, &tp->wpert.nangl,
	 &tp->wpert.ndihe, &tp->ifbox, &tp->nmxrs, &tp->ifcap, &tp->blank);

  /*** Now, read each atom name, charge, mass, type, type name, and class ***/
  natm = tp->natom;
  ReadToFlag("ATOM_NAME", inp, 0);
  tp->AtomNames = LoadChar(inp, natm, 4, 20);
  ReadToFlag("CHARGE", inp, 0);
  tp->Charges = LoadDouble(inp, natm);
  DVecMult(tp->Charges, tp->natom, 1.0/18.2223);
  ReadToFlag("MASS", inp, 0);
  tp->Masses = LoadDouble(inp, natm);
  ReadToFlag("ATOM_TYPE_INDEX", inp, 0);
  tp->LJIdx = LoadInteger(inp, natm);
  if (adjbnd == 1) {
    IVecAdd(tp->LJIdx, natm, -1);
  }
  ReadToFlag("AMBER_ATOM_TYPE", inp, 0);
  tp->AtomTypes = LoadChar(inp, natm, 4, 20);
  ReadToFlag("TREE_CHAIN_CLASSIFICATION", inp, 0);
  tp->TreeSymbols = LoadChar(inp, natm, 4, 20);

  /*** Read the number of excluded atoms and make lists ***/
  ReadToFlag("NUMBER_EXCLUDED_ATOMS", inp, 0);
  tp->NExcl = LoadInteger(inp, natm);
  tp->ConExcl = (int*)calloc(natm, sizeof(int));
  j = 0;
  for (i = 0; i < natm; i++) {
    tp->ConExcl[i] = j;
    j = j + tp->NExcl[i];
  }
  ReadToFlag("EXCLUDED_ATOMS_LIST", inp, 0);
  tp->ExclList = LoadInteger(inp, tp->ConExcl[natm-1] + tp->NExcl[natm-1]);
  if (adjbnd == 1) {
    IVecAdd(tp->ExclList, tp->ConExcl[natm-1] + tp->NExcl[natm-1], -1);
  }

  /*** Now get the force constants for all sorts of interactions ***/
  ReadToFlag("NONBONDED_PARM_INDEX", inp, 0);
  tp->NBParmIdx = LoadInteger(inp, tp->ntypes*tp->ntypes);
  if (adjbnd == 1) {
    IVecAdd(tp->NBParmIdx, tp->ntypes*tp->ntypes, -1);
  }
  ReadToFlag("BOND_FORCE_CONSTANT", inp, 0);
  tp->BondK = LoadDouble(inp, tp->nBAH.nbond);
  ReadToFlag("BOND_EQUIL_VALUE", inp, 0);
  tp->BondEq = LoadDouble(inp, tp->nBAH.nbond);
  ReadToFlag("ANGLE_FORCE_CONSTANT", inp, 0);
  tp->AnglK = LoadDouble(inp, tp->nBAH.nangl);
  ReadToFlag("ANGLE_EQUIL_VALUE", inp, 0);
  tp->AnglEq = LoadDouble(inp, tp->nBAH.nangl);
  ReadToFlag("DIHEDRAL_FORCE_CONSTANT", inp, 0);
  tp->DiheK = LoadDouble(inp, tp->nBAH.ndihe);
  ReadToFlag("DIHEDRAL_PERIODICITY", inp, 0);
  tp->DiheN = LoadDouble(inp, tp->nBAH.ndihe);
  ReadToFlag("DIHEDRAL_PHASE", inp, 0);
  tp->DihePhi = LoadDouble(inp, tp->nBAH.ndihe);
  ReadToFlag("LENNARD_JONES_ACOEF", inp, 0);
  tp->LJA = LoadDouble(inp, tp->ntypes*(tp->ntypes+1)/2);
  ReadToFlag("LENNARD_JONES_BCOEF", inp, 0);
  tp->LJB = LoadDouble(inp, tp->ntypes*(tp->ntypes+1)/2);
  if (tp->ljbuck == 1) {
    ReadToFlag("LENNARD_JONES_CCOEF", inp, 0);
    tp->LJC = LoadDouble(inp, tp->ntypes*(tp->ntypes+1)/2);
  }
  ReadToFlag("HBOND_ACOEF", inp, 0);
  tp->SolA = LoadDouble(inp, tp->nphb*(tp->nphb+1)/2);
  ReadToFlag("HBOND_BCOEF", inp, 0);
  tp->SolB = LoadDouble(inp, tp->nphb*(tp->nphb+1)/2);
  ReadToFlag("HBCUT", inp, 0);
  tp->HBCut = LoadDouble(inp, tp->nphb*(tp->nphb+1)/2);
  ReadToFlag("RADIUS_SET", inp, 0);
  fgets(tp->RadSet, 128, inp);
  ReadToFlag("RADII", inp, 0);
  tp->Radii = LoadDouble(inp, tp->natom);
  ReadToFlag("SCREEN", inp, 0);
  tp->Screen = LoadDouble(inp, tp->natom);

  /*** Get residue limits and labels ***/
  ReadToFlag("RESIDUE_LABEL", inp, 0);
  tp->ResNames = LoadChar(inp, tp->nres, 4, 20);
  ReadToFlag("RESIDUE_POINTER", inp, 0);
  tp->ResLims = LoadInteger(inp, tp->nres);
  if (adjbnd == 1) {
    IVecAdd(tp->ResLims, tp->nres, -1);
  }
  tp->ResLims = (int*)realloc(tp->ResLims, (tp->nres+1)*sizeof(int));
  tp->ResLims[tp->nres] = tp->natom;

  /*** Bond, angle, and dihedral lists ***/
  ReadToFlag("BONDS_INC_HYDROGEN", inp, 0);
  tp->BIncH = (bond*)LoadInteger(inp, 3*tp->withH.nbond);
  if (adjbnd == 1) {
    AdjustBondArray((int*)tp->BIncH, tp->withH.nbond, 3);
  }
  ReadToFlag("BONDS_WITHOUT_HYDROGEN", inp, 0);
  tp->BNoH = (bond*)LoadInteger(inp, 3*tp->woH.nbond);
  if (adjbnd == 1) {
    AdjustBondArray((int*)tp->BNoH, tp->woH.nbond, 3);
  }
  ReadToFlag("ANGLES_INC_HYDROGEN", inp, 0);
  tp->AIncH = (angle*)LoadInteger(inp, 4*tp->withH.nangl);
  if (adjbnd == 1) {
    AdjustBondArray((int*)tp->AIncH, tp->withH.nangl, 4);
  }
  ReadToFlag("ANGLES_WITHOUT_HYDROGEN", inp, 0);
  tp->ANoH = (angle*)LoadInteger(inp, 4*tp->woH.nangl);
  if (adjbnd == 1) {
    AdjustBondArray((int*)tp->ANoH, tp->woH.nangl, 4);
  }
  ReadToFlag("DIHEDRALS_INC_HYDROGEN", inp, 0);
  tp->HIncH = (dihedral*)LoadInteger(inp, 5*tp->withH.ndihe);
  if (adjbnd == 1) {
    AdjustBondArray((int*)tp->HIncH, tp->withH.ndihe, 5);
  }
  ReadToFlag("DIHEDRALS_WITHOUT_HYDROGEN", inp, 0);
  tp->HNoH = (dihedral*)LoadInteger(inp, 5*tp->woH.ndihe);
  if (adjbnd == 1) {
    AdjustBondArray((int*)tp->HNoH, tp->woH.ndihe, 5);
  }

  /*** Miscellaneous ***/
  ReadToFlag("JOIN_ARRAY", inp, 0);
  tp->Join = LoadInteger(inp, natm);
  if (adjbnd == 1) {
    IVecAdd(tp->Join, natm, -1);
  }
  ReadToFlag("IROTAT", inp, 0);
  tp->Rotat = LoadInteger(inp, natm);
  if (adjbnd == 1) {
    IVecAdd(tp->Rotat, natm, -1);
  }

  /*** For periodic boxes ***/
  if (tp->ifbox > 0) {
    ReadToFlag("SOLVENT_POINTERS", inp, 0);
    fscanf(inp, "%d%d%d", &tp->iptres, &tp->nspm, &tp->nspsol);
    ReadToFlag("ATOMS_PER_MOLECULE", inp, 0);
    tp->Nsp = LoadInteger(inp, tp->nspm);
    if (adjbnd == 1) {
      IVecAdd(tp->Nsp, tp->nspm, -1);
    }
    ReadToFlag("BOX_DIMENSIONS", inp, 0);
    fscanf(inp, "%lf%lf%lf%lf", &tp->gdim[4], &tp->gdim[0], &tp->gdim[1],
	   &tp->gdim[2]);
  }

  /*** Finally, if command file directives have been given to read ***/
  /*** additional components of the topology file, do that now.    ***/
  if (tp->ReadVirtualQ == 1) {
    //    ReadVirtualCharges(tp, inp);
  }


  /*** Close the input file ***/
  fclose(inp);

  /*** Pre-computations for masses or other constants on a per-atom basis ***/
  tp->InvMasses = (double*)malloc(tp->natom*sizeof(double));
  tp->TotalMass = 0.0;
  for (i = 0; i < tp->natom; i++) {
    tp->InvMasses[i] = (tp->Masses[i] < 1.0e-12) ? 0.0 : 1.0/tp->Masses[i];
    tp->TotalMass += tp->Masses[i];
  }

  /*** Currently no support for IFCAP > 0, IFPERT > 0, and/or IPOL == 1 ***/

  /*** Re-organize dihedral information ***/
  OrderParameters(tp);
  EnumerateBonds(tp);
  EnumerateAngles(tp);
  EnumerateDihedrals(tp);

  /*** Round the total amount of charge on the system ***/
  RoundTotalCharge(tp);

  /*** Define groups of linked atoms ***/
  DefineAtomChains(tp);

  /*** Prepare for SETTLE ***/
  FindThreePointWaters(tp);
  if (tp->nwat > 0) {
    CompSettleGeom(tp);
  }

  /*** Process information regarding extra points ***/
  ReadEPRuleFile(tp);
  DetermineEPFrames(tp);

  /*** Determine the number of system degrees of freedom and particles ***/
  tp->ncnst = 0;
  tp->nprtcl = tp->natom - tp->nxtrapt;
  if (tp->rattle == 1 || tp->settle == 1) {
    for (i = 0; i < tp->natom; i++) {
      if (tp->SHL[i].exe == 1) {
	tp->ncnst += 3;
	tp->nprtcl -= 2;
      }
    }
  }
  tp->ndf = 3*(tp->natom - 1) - tp->ncnst - 3*tp->nxtrapt;
}

/***=======================================================================***/
/*** ReadToFlag: reads a topology file from the beginning until a          ***/
/***             particular flag is found, then skips one additional line  ***/
/***             and checks that this line is also a comment line before   ***/
/***             exiting, leaving the file pointer inp set to the next     ***/
/***             line after the desired flag.                              ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   fname:     the flag to search for                                   ***/
/***   inp:       the topology file                                        ***/
/***   tryrewind: flag to indicate that rewinding the file should be       ***/
/***              tried; initially set to 1, but then set to zero on any   ***/
/***              recursive calls to avoid futile recycling of a topology  ***/
/***              file that doesn't contain the information sought         ***/
/***=======================================================================***/
void ReadToFlag(char* fname, FILE *inp, int tryrewind)
{
  int collect, flen;
  char line[MAXLINE];

  collect = 0;
  flen = strlen(fname);
  if (tryrewind == 1) {
    rewind(inp);
  }
  while (collect == 0) {
    if (fgets(line, MAXLINE, inp) == NULL) {

      /*** If we have reached the end of the file and not yet found  ***/
      /*** the search term, we may be able to try rewinding the file ***/
      /*** and search it from the beginning.  However, if we have    ***/
      /*** already tried to rewind the file, an error has occured.   ***/
      if (tryrewind == 0) {
	ReadToFlag(fname, inp, 1);

	/*** If we recursively called ReadToFlag with a directive ***/
	/*** to rewind the file and search from the beginning, we ***/
	/*** assume that the search term has been found, as the   ***/
	/*** entire program would have existed in error on the    ***/
	/*** recursive call to ReadToFlag otherwise.              ***/
	collect = 1;
      }
      else {
        printf("ReadToFlag >> Error.  End of File reached seeking %s.\n",
	       fname);
	exit(1);
      }
    }
    else if (line[0] == '%' && strncmp(&line[6], fname, flen) == 0) {
      fgets(line, MAXLINE, inp);
      if (line[0] != '%') {
	printf("ReadToFlag >> Error.  Comment line expected following flag.\n"
	       "ReadToFlag >> Found \"%s\"\n", line);
	exit(1);
      }
      collect = 1;
    }
  }
}

/***=======================================================================***/
/*** LoadChar: read in N sets of P characters each, at the rate of Q sets  ***/
/***           per line from file inp.                                     ***/
/***=======================================================================***/
char* LoadChar(FILE *inp, int N, int P, int Q)
{
  int i, j, k;
  char line[MAXLINE];
  char* s;

  i = 0;
  s = (char*)malloc(N*P*sizeof(char));
  while (i < N) {
    if (fgets(line, MAXLINE, inp) == NULL) {
      printf("LoadChar >> Error.  EOF reached.  Terminating.\n");
      exit(1);
    }
    j = 0;
    while (j < Q && i < N) {
      for (k = P*j; k < P*(j+1); k++) {
	s[(i-j)*P+k] = line[k];
      }
      i++;
      j++;
    }
  }

  return s;
}

/***=======================================================================***/
/*** LoadInteger: read in N integers from the a file inp.                  ***/
/***=======================================================================***/
int* LoadInteger(FILE *inp, int N)
{
  int i;
  int* s;

  s = (int*)malloc(N*sizeof(int));
  for (i = 0; i < N; i++) {
    fscanf(inp, "%d", &s[i]);
  }

  return s;
}

/***=======================================================================***/
/*** LoadDouble: read in N double-precision reals from a file inp.         ***/
/***=======================================================================***/
double* LoadDouble(FILE *inp, int N)
{
  int i;
  double* s;

  s = (double*)malloc(N*sizeof(double));
  for (i = 0; i < N; i++) {
    fscanf(inp, "%lf", &s[i]);
  }

  return s;
}

/***=======================================================================***/
/*** CopyTopology: copy a topology structure.                              ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:   the original topology                                         ***/
/***   tpC:  the new topology                                              ***/
/***=======================================================================***/
prmtop CopyTopology(prmtop *tp)
{
  int i, j, natm;
  prmtop tpC;

  /*** General copying ***/
  tpC = *tp;

  /*** Copy integer arrays ***/
  natm = tp->natom; 
  tpC.LJIdx = CpyIVec(tp->LJIdx, natm);
  tpC.NExcl = CpyIVec(tp->NExcl, natm);
  tpC.ConExcl = CpyIVec(tp->ConExcl, natm);
  tpC.ExclList = CpyIVec(tp->ExclList, tp->ConExcl[natm-1]+tp->NExcl[natm-1]);
  tpC.NBParmIdx = CpyIVec(tp->NBParmIdx, tp->ntypes*tp->ntypes);
  tpC.ResLims = CpyIVec(tp->ResLims, tp->nres);
  tpC.Join = CpyIVec(tp->Join, natm);
  tpC.Rotat = CpyIVec(tp->Rotat, natm);
  tpC.Nsp = CpyIVec(tp->Nsp, tp->nspm);

  /*** Copy double-precision real arrays ***/
  tpC.Charges = CpyDVec(tp->Charges, natm);
  tpC.Masses = CpyDVec(tp->Masses, natm);
  tpC.InvMasses = CpyDVec(tp->InvMasses, natm);
  tpC.BondK = CpyDVec(tp->BondK, tp->nBAH.nbond);
  tpC.BondEq = CpyDVec(tp->BondEq, tp->nBAH.nbond);
  tpC.AnglK = CpyDVec(tp->AnglK, tp->nBAH.nangl);
  tpC.AnglEq = CpyDVec(tp->AnglEq, tp->nBAH.nangl);
  tpC.DiheK = CpyDVec(tp->DiheK, tp->nBAH.ndihe);
  tpC.DiheN = CpyDVec(tp->DiheN, tp->nBAH.ndihe);
  tpC.DihePhi = CpyDVec(tp->DihePhi, tp->nBAH.ndihe);
  tpC.LJA = CpyDVec(tp->LJA, tp->ntypes*(tp->ntypes+1)/2);
  tpC.LJB = CpyDVec(tp->LJB, tp->ntypes*(tp->ntypes+1)/2);
  tpC.SolA = CpyDVec(tp->SolA, tp->nphb*(tp->nphb+1)/2);
  tpC.SolB = CpyDVec(tp->SolB, tp->nphb*(tp->nphb+1)/2);
  tpC.HBCut = CpyDVec(tp->HBCut, tp->nphb*(tp->nphb+1)/2);
  tpC.Radii = CpyDVec(tp->Radii, natm);
  tpC.Screen = CpyDVec(tp->Screen, natm);

  /*** Copy character arrays ***/
  tpC.AtomNames = (char*)malloc(4*natm*sizeof(char));
  tpC.AtomTypes = (char*)malloc(4*natm*sizeof(char));
  tpC.TreeSymbols = (char*)malloc(4*natm*sizeof(char));
  tpC.ResNames = (char*)malloc(4*tp->nres*sizeof(char));
  for (i = 0; i < 4*natm; i++) {
    tpC.AtomNames[i] = tp->AtomNames[i];
    tpC.AtomTypes[i] = tp->AtomTypes[i];
    tpC.TreeSymbols[i] = tp->TreeSymbols[i];
  }
  for (i = 0; i < 4*tp->nres; i++) {
    tpC.ResNames[i] = tp->ResNames[i];
  }

  /*** Copy arrays of structures ***/
  tpC.BIncH = (bond*)malloc(tp->withH.nbond*sizeof(bond));
  for (i = 0; i < tp->withH.nbond; i++) {
    tpC.BIncH[i] = tp->BIncH[i];
  }
  tpC.BNoH = (bond*)malloc(tp->woH.nbond*sizeof(bond));
  for (i = 0; i < tp->woH.nbond; i++) {
    tpC.BNoH[i] = tp->BNoH[i];
  }
  tpC.AIncH = (angle*)malloc(tp->withH.nangl*sizeof(angle));
  for (i = 0; i < tp->withH.nangl; i++) {
    tpC.AIncH[i] = tp->AIncH[i];
  }
  tpC.ANoH = (angle*)malloc(tp->woH.nangl*sizeof(angle));
  for (i = 0; i < tp->woH.nangl; i++) {
    tpC.ANoH[i] = tp->ANoH[i];
  }
  tpC.HIncH = (dihedral*)malloc(tp->withH.ndihe*sizeof(dihedral));
  for (i = 0; i < tp->withH.ndihe; i++) {
    tpC.HIncH[i] = tp->HIncH[i];
  }
  tpC.HNoH = (dihedral*)malloc(tp->woH.ndihe*sizeof(dihedral));
  for (i = 0; i < tp->woH.ndihe; i++) {
    tpC.HNoH[i] = tp->HNoH[i];
  }
  tpC.BLC = (bondlist*)malloc(natm*sizeof(bondlist));
  for (i = 0; i < tp->natom; i++) {
    tpC.BLC[i] = tp->BLC[i];
    tpC.BLC[i].BC = (bondcomm*)malloc(tpC.BLC[i].nbond*sizeof(bondcomm));
    for (j = 0; j < tpC.BLC[i].nbond; j++) {
      tpC.BLC[i].BC[j] = tp->BLC[i].BC[j];
    }
  }
  tpC.ALC = (angllist*)malloc(natm*sizeof(angllist));
  for (i = 0; i < tp->natom; i++) {
    tpC.ALC[i] = tp->ALC[i];
    tpC.ALC[i].AC = (anglcomm*)malloc(tpC.ALC[i].nangl*sizeof(anglcomm));
    for (j = 0; j < tpC.ALC[i].nangl; j++) {
      tpC.ALC[i].AC[j] = tp->ALC[i].AC[j];
    }
  }
  tpC.HLC = (dihelist*)malloc(natm*sizeof(dihelist));
  for (i = 0; i < tp->natom; i++) {
    tpC.HLC[i] = tp->HLC[i];
    tpC.HLC[i].HC = (dihecomm*)malloc(tpC.HLC[i].ndihe*sizeof(dihecomm));
    for (j = 0; j < tpC.HLC[i].ndihe; j++) {
      tpC.HLC[i].HC[j] = tp->HLC[i].HC[j];
      tpC.HLC[i].HC[j].t = CpyIVec(tp->HLC[i].HC[j].t, tp->HLC[i].HC[j].nt);
    }
  }
  tpC.BParam = (bonddef*)malloc(tp->nBAH.nbond*sizeof(bonddef));
  for (i = 0; i < tp->nBAH.nbond; i++) {
    tpC.BParam[i] = tp->BParam[i];
  }
  tpC.AParam = (angldef*)malloc(tp->nBAH.nangl*sizeof(angldef));
  for (i = 0; i < tp->nBAH.nangl; i++) {
    tpC.AParam[i] = tp->AParam[i];
  }
  tpC.HParam = (dihedef*)malloc(tp->nBAH.ndihe*sizeof(dihedef));
  for (i = 0; i < tp->nBAH.ndihe; i++) {
    tpC.HParam[i] = tp->HParam[i];
  }
  tpC.SHL = (cnstcomm*)malloc(natm*sizeof(cnstcomm));
  for (i = 0; i < natm; i++) {
    tpC.SHL[i] = tp->SHL[i];
    if (tpC.SHL[i].exe == 1) {
      tpC.SHL[i].blist = CpyIVec(tp->SHL[i].blist, 2);
    }

    /*** FIX ME!  Copy blist if there are other types of constraints ***/
  }

  /*** Copy matrices ***/
  tpC.LJftab = CreateDmat(tp->ntypes, 2*tp->ntypes, 0);
  tpC.LJutab = CreateDmat(tp->ntypes, 2*tp->ntypes, 0);
  for (i = 0; i < 2*tp->ntypes*tp->ntypes; i++) {
    tpC.LJftab.data[i] = tp->LJftab.data[i];
    tpC.LJutab.data[i] = tp->LJutab.data[i];
  }

  /*** Copy lone pair group lists ***/
  tpC.ngrp = tp->ngrp;
  tpC.lgrps = (lgrp*)malloc(tpC.ngrp*sizeof(lgrp));
  for (i = 0; i < tpC.ngrp; i++) {
    tpC.lgrps[i].natom = tp->lgrps[i].natom;
    tpC.lgrps[i].atoms = CpyIVec(tp->lgrps[i].atoms, tpC.lgrps[i].natom);
  }

  return tpC;
}

/***=======================================================================***/
/*** FreeTopology: free a topology structure.                              ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:  the topology                                                   ***/
/***=======================================================================***/
void FreeTopology(prmtop *tp)
{
  int i;

  /*** Free arrays of elemental types ***/
  free(tp->LJIdx);
  free(tp->NExcl);
  free(tp->ConExcl);
  free(tp->ExclList);
  free(tp->NBParmIdx);
  free(tp->ResLims);
  free(tp->Join);
  free(tp->Rotat);
  free(tp->Nsp);
  free(tp->Charges);
  free(tp->Masses);
  free(tp->InvMasses);
  free(tp->BondK);
  free(tp->BondEq);
  free(tp->AnglK);
  free(tp->AnglEq);
  free(tp->DiheK);
  free(tp->DiheN);
  free(tp->DihePhi);
  free(tp->LJA);
  free(tp->LJB);
  free(tp->SolA);
  free(tp->SolB);
  free(tp->HBCut);
  free(tp->Radii);
  free(tp->Screen);
  free(tp->AtomNames);
  free(tp->ResNames);
  free(tp->AtomTypes);
  free(tp->TreeSymbols);

  /*** Free arrays of structs ***/
  free(tp->BIncH);
  free(tp->BNoH);
  free(tp->AIncH);
  free(tp->ANoH);
  free(tp->HIncH);
  free(tp->HNoH);
  for (i = 0; i < tp->natom; i++) {
    free(tp->BLC[i].BC);
    free(tp->ALC[i].AC);
    free(tp->HLC[i].HC);
  }
  free(tp->BLC);
  free(tp->ALC);
  free(tp->HLC);
  free(tp->BParam);
  free(tp->AParam);
  free(tp->HParam);
  free(tp->SHL);
  for (i = 0; i < tp->ngrp; i++) {
    free(tp->lgrps[i].atoms);
  }
  free(tp->lgrps);

  /*** Free some pre-computed parameter tables ***/
  DestroyDmat(&tp->LJftab);
  DestroyDmat(&tp->LJutab);
}

/***=======================================================================***/
/*** BondScan: computes the mean value of some bond length over all        ***/
/***           matching residues of a simulation.                          ***/
/***=======================================================================***/
static void BondScan(prmtop *tp, int* cpylist, int finst, int A, int B, int C,
		     int D, coord *tc, double *mval, double *rmsval,
		     int bondorder)
{
  int i, j, tcon, aA, aB, aC, aD;
  double ax[3], dx[3];
  double* blist;
  char *cA, *cB, *cC, *cD;

  /*** Determine which atoms "A" and "B" are ***/
  cA = &tp->AtomNames[4*A];
  cB = &tp->AtomNames[4*B];
  cC = &tp->AtomNames[4*C];
  cD = &tp->AtomNames[4*D];
  if (bondorder > 2) {
    cC = &tp->AtomNames[4*C];
  }
  if (bondorder > 3) {
    cD = &tp->AtomNames[4*D];
  }

  tcon = 0;
  blist = (double*)malloc(tp->nres*sizeof(double));
  for (i = 0; i < tp->nres; i++) {
    if (cpylist[i] != cpylist[finst]) {
      continue;
    }

    /*** Find the bond length in this residue ***/
    aA = -1;
    aB = -1;
    aC = -1;
    aD = -1;
    for (j = tp->ResLims[i]; j < tp->ResLims[i+1]; j++) {
      if (strncmp(&tp->AtomNames[4*j], cA, 4) == 0) {
        aA = j;
      }
      else if (strncmp(&tp->AtomNames[4*j], cB, 4) == 0) {
        aB = j;
      }
      else if (bondorder > 2 && strncmp(&tp->AtomNames[4*j], cC, 4) == 0) {
        aC = j;
      }
      else if (bondorder > 3 && strncmp(&tp->AtomNames[4*j], cD, 4) == 0) {
        aD = j;
      }
    }
    if (aA == -1) {
      printf("BondScan >> Error.  Atom %.4s not found in residue %d!\n", cA,
             i);
      exit(1);
    }
    if (aB == -1) {
      printf("BondScan >> Error.  Atom %.4s not found in residue %d!\n", cB,
             i);
      exit(1);
    }
    if (bondorder > 2 && aC == -1) {
      printf("BondScan >> Error.  Atom %.4s not found in residue %d!\n", cC,
             i);
      exit(1);
    }
    if (bondorder > 3 && aD == -1) {
      printf("BondScan >> Error.  Atom %.4s not found in residue %d!\n", cD,
             i);
      exit(1);
    }
    if (bondorder == 2) {
      for (j = 0; j < 3; j++) {
        dx[j] = tc->loc[3*aB+j] - tc->loc[3*aA+j];
      }
      blist[tcon] = DIST(dx[0], dx[1], dx[2]);
    }
    else if (bondorder == 3) {
      for (j = 0; j < 3; j++) {
        dx[j] = tc->loc[3*aC+j] - tc->loc[3*aB+j];
        ax[j] = tc->loc[3*aA+j] - tc->loc[3*aB+j];
      }
      blist[tcon] = acos(TrimAcos(DotP(ax, dx, 3)/(DMag(dx, 3)*DMag(ax, 3))));
      blist[tcon] *= PI/180.0;
    }
    else if (bondorder == 4) {
      blist[tcon] = Dihedral(&tc->loc[3*aA], &tc->loc[3*aB], &tc->loc[3*aC],
                             &tc->loc[3*aD]);
      blist[tcon] *= PI/180.0;
    }
    tcon++;
  }
  *mval = DAverage(blist, tcon);
  *rmsval = DStDev(blist, tcon);

  /*** Free Allocated Memory ***/
  free(blist);
}

/***=======================================================================***/
/*** PrintBondInfo: a general routine for printing bond information about  ***/
/***                a residue.                                             ***/
/***=======================================================================***/
static void PrintBondInfo(prmtop *tp, int* B, int nB, int P, int finst,
			  double* K, double* Eq, double* Pd, FILE *outp,
			  int* cpylist, coord *tc)
{
  int i, j, rmin, rmax, okgo, iC, iD;
  double mlen, rmsval;

  rmin = tp->ResLims[finst];
  rmax = tp->ResLims[finst+1];
  for (i = 0; i < nB; i++) {

    /*** Make sure this bonded term is solely within the residue ***/
    okgo = 1;
    for (j = 0; j < P-1; j++) {
      if (B[i*P+j] < rmin || B[i*P+j] >= rmax) {
        okgo = 0;
        break;
      }
    }
    if (okgo == 0) {
      continue;
    }

    /*** Print ***/
    fprintf(outp, " ");
    for (j = 0; j < P-1; j++) {
      fprintf(outp, "%.4s", &tp->AtomNames[4*B[i*P+j]]);
      if (j < P-2) {
        fprintf(outp, ":");
      }
    }
    fprintf(outp, "  %10.6lf  %10.6lf", K[B[(i+1)*P-1]], Eq[B[(i+1)*P-1]]);
    if (P == 5) {
      fprintf(outp, "  %10.6lf", Pd[B[(i+1)*P-1]]);
    }
    iC = (P > 3) ? B[i*P+2] : -1;
    iD = (P > 4) ? B[i*P+3] : -1;
    BondScan(tp, cpylist, finst, B[i*P], B[i*P+1], iC, iD, tc, &mlen, &rmsval,
	     P-1);
    fprintf(outp, "  %10.6lf  %10.6lf", mlen, rmsval);
    fprintf(outp, "\n");
  }
}

/***=======================================================================***/
/*** MapResidues: find distinct residues within a topology and map them to ***/
/***              human-readable formats.                                  ***/
/***=======================================================================***/
void MapResidues(prmtop *tp, coord *tc, char* outname)
{
  int i, j, k, nres, rlen, ai, aj, match, ninst, finst, natm;
  int* cpylist;
  double qi, qj, lja, ljb, sig, eps, bfac;
  char *ctmp;
  FILE *outp;

  /*** Some conversion factors ***/
  bfac = 1.0/sqrt(BIOQ);

  cpylist = (int*)malloc(tp->nres*sizeof(int));
  SetIVec(cpylist, tp->nres, -1);
  nres = 0;
  for (i = 0; i < tp->nres; i++) {

    /*** Make sure we haven't done this already ***/
    if (cpylist[i] > -1) {
      continue;
    }

    /*** This residue is number "nres" ***/
    cpylist[i] = nres;

    /*** Now, look for all other instances ***/
    ctmp = &tp->ResNames[4*i];
    rlen = tp->ResLims[i+1] - tp->ResLims[i];
    for (j = i+1; j < tp->nres; j++) {
      if (strncmp(ctmp, &tp->ResNames[4*j], 4) == 0 &&
	  rlen == tp->ResLims[j+1] - tp->ResLims[j]) {
	match = 1;
	for (k = 0; k < rlen; k++) {
	  qi = tp->Charges[tp->ResLims[i]+k];
	  qj = tp->Charges[tp->ResLims[j]+k];
	  ai = tp->LJIdx[tp->ResLims[i]+k];
	  aj = tp->LJIdx[tp->ResLims[j]+k];
	  if (DNEQ(qi, qj) || ai != aj) {
	    match = 0;
	  }
	}
	if (match == 1) {
	  cpylist[j] = nres;
	}
      }
    }

    /*** Increment our counter ***/
    nres++;
  }

  /*** Now, write the map for each residue ***/
  outp = fopen(outname, "w");
  for (i = 0; i < nres; i++) {

    /*** Find the first instance of the residue and store it as "finst" ***/
    finst = -1;
    ninst = 0;
    for (j = 0; j < tp->nres; j++) {
      if (cpylist[j] == i) {
	if (finst < 0) {
	  finst = j;
	}
	ninst++;
      }
    }

    /*** Print the atomic information about this residue ***/
    natm = tp->ResLims[finst+1] - tp->ResLims[finst];
    fprintf(outp, "%% RESIDUE  %.4s  [ %4d atoms, %8.4lf net charge, %6d "
	    "occurences ]\n"
	    " Atom Name      Charge   Type Name    LJ Index    LJ Sigma  "
	    "LJ Epsilon  Tree Class\n----------  ----------  ----------  "
	    "----------  ----------  ----------  ----------\n",
	    &tp->ResNames[4*finst], natm,
	    bfac*DSum(&tp->Charges[tp->ResLims[finst]], natm), ninst);
    for (j = tp->ResLims[finst]; j < tp->ResLims[finst+1]; j++) {
      lja = tp->LJA[tp->NBParmIdx[(tp->ntypes+1)*tp->LJIdx[j]]];
      ljb = tp->LJB[tp->NBParmIdx[(tp->ntypes+1)*tp->LJIdx[j]]];
      sig = (lja > 1.0e-04 && ljb > 1.0e-04) ? pow(lja/ljb, 1.0/6.0) : 0.0;
      eps = (sig > 1.0e-04) ? 0.25*ljb/pow(sig, 6.0) : 0.0;
      fprintf(outp, "      %.4s  %10.6lf        %.4s  %10d  %10.6lf  %10.6lf  "
	      "      %.4s\n", &tp->AtomNames[4*j], tp->Charges[j]*bfac,
	      &tp->AtomTypes[4*j], tp->LJIdx[j], sig, eps,
	      &tp->TreeSymbols[4*j]);
    }
    fprintf(outp, "\n");

    /*** Now seek the bond information on this residue ***/
    fprintf(outp, "  Bond A:B   Stiffness  Eq. Length  Mean Value    St. Dev."
	    "\n----------  ----------  ----------  ----------  ----------\n");
    PrintBondInfo(tp, (int*)tp->BIncH, tp->withH.nbond, 3, finst, tp->BondK,
		  tp->BondEq, tp->DihePhi, outp, cpylist, tc);
    PrintBondInfo(tp, (int*)tp->BNoH, tp->woH.nbond, 3, finst, tp->BondK,
		  tp->BondEq, tp->DihePhi, outp, cpylist, tc);
    fprintf(outp, "\n");

    /*** Now seek the angle information on this residue ***/
    fprintf(outp, "  Angle A:B:C     Stiffness  Eq. Angle  Mean Value    "
	    "St. Dev.\n---------------  ----------  ----------  ----------  "
	    "----------\n");
    PrintBondInfo(tp, (int*)tp->AIncH, tp->withH.nangl, 4, finst, tp->AnglK,
		  tp->AnglEq, tp->DihePhi, outp, cpylist, tc);
    PrintBondInfo(tp, (int*)tp->ANoH, tp->woH.nangl, 4, finst, tp->AnglK,
		  tp->AnglEq, tp->DihePhi, outp, cpylist, tc);
    fprintf(outp, "\n");

    /*** Now seek the dihedral information on this residue ***/
    fprintf(outp, "  Dihedral A:B:C:D       Penalty     Period       Phase  "
	    "Mean Value    St. Dev.\n"
	    "--------------------  ----------  ----------  ----------  "
	    "----------  ----------\n");
    PrintBondInfo(tp, (int*)tp->HIncH, tp->withH.ndihe, 5, finst, tp->DiheK,
		  tp->DiheN, tp->DihePhi, outp, cpylist, tc);
    PrintBondInfo(tp, (int*)tp->HNoH, tp->woH.ndihe, 5, finst, tp->DiheK,
		  tp->DiheN, tp->DihePhi, outp, cpylist, tc);
    fprintf(outp, "\n");
  }
  fclose(outp);
}

/***=======================================================================***/
/*** AdjustBondArray: the bonded arrays are tricky to adjust, because the  ***/
/***                  atom numbers are given as pointer indices and the    ***/
/***                  bond constants are in need of the typical Fortran-C  ***/
/***                  array indexing adjustment.                           ***/
/***                  We assume array A to have N sets of P numbers each,  ***/
/***                  of which the first P-1 are atom numbers.             ***/
/***=======================================================================***/
void AdjustBondArray(int* A, int N, int P)
{
  int h, i, j;

  h = 0;
  for (i = 0; i < N; i++) {
    for (j = 0; j < P-1; j++) {
      A[h] = A[h]/3;
      h++;
    }
    A[h] -= 1;
    h++;
  }
}

/***=======================================================================***/
/*** FindAtom: finds an atom in a list of names, between limits il and ih. ***/
/***=======================================================================***/
int FindAtom(prmtop *tp, int il, int ih, char* aname)
{
  int i;

  for (i = il; i < ih; i++) {
    if (strncmp(aname, &tp->AtomNames[4*i], 4) == 0) {
      return i;
    }
  }

  /*** Error message ***/
  printf("FindAtom >> Error.  Could not locate atom %.s\n", aname);
  exit(1);

  return -1;
}

/***=======================================================================***/
/*** Format5e16: print a chunk of topology file with this numerical        ***/
/***             format.                                                   ***/
/***=======================================================================***/
void Format5e16(char* cname, double* values, int N, FILE *outp)
{
  int h, i, j, slen;
  char nstr[MAXNAME];

  fprintf(outp, "%%FLAG %s\n%%FORMAT(5E16.8)\n", cname);
  h = 0;
  for (i = 0; i < N; i++) {
    sprintf(nstr, "%16.8e", values[i]);
    slen = strlen(nstr);
    for (j = 0; j < slen; j++) {
      if (nstr[j] == 'e') {
	nstr[j] = 'E';
      }
    }
    fprintf(outp, "%s", nstr);
    h++;
    if (h == 5) {
      h = 0;
      fprintf(outp, "\n");
    }
  }
  if (h != 0 || N == 0) {
    fprintf(outp, "\n");
  }
}

/***=======================================================================***/
/*** Format10i8: print a chunk of topology file with this numerical        ***/
/***             format.                                                   ***/
/***=======================================================================***/
void Format10i8(char* cname, int* values, int N, FILE *outp)
{
  int h, i;

  fprintf(outp, "%%FLAG %s\n%%FORMAT(10I8)\n", cname);
  h = 0;
  for (i = 0; i < N; i++) {
    fprintf(outp, "%8d", values[i]);
    h++;
    if (h == 10) {
      h = 0;
      fprintf(outp, "\n");
    }
  }
  if (h != 0 || N == 0) {
    fprintf(outp, "\n");
  }
}

/***=======================================================================***/
/*** Format20a4: print a chunk of topology file with this character        ***/
/***             format.                                                   ***/
/***=======================================================================***/
void Format20a4(char* cname, char* values, int N, FILE *outp)
{
  int h, i;

  fprintf(outp, "%%FLAG %s\n%%FORMAT(20a4)\n", cname);
  h = 0;
  for (i = 0; i < N; i++) {
    fprintf(outp, "%.4s", &values[4*i]);
    h++;
    if (h == 20) {
      h = 0;
      fprintf(outp, "\n");
    }
  }
  if (h != 0 || N == 0) {
    fprintf(outp, "\n");
  }
}

/***=======================================================================***/
/*** PutPrmTop: put a topology into a file.                                ***/
/***=======================================================================***/
void PutPrmTop(prmtop *tp, char* fname, char* title)
{
  int* itmp;
  double* solty;
  FILE *outp = fopen(fname, "w");

  fprintf(outp, "%s%%FLAG TITLE\n%%FORMAT(20a4)\n%s\n", tp->vstamp, title);
  fprintf(outp, "%%FLAG POINTERS\n%%FORMAT(10I8)\n");
  fprintf(outp, "%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d\n"
	  "%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d\n"
	  "%8d%8d%8d%8d%8d%8d%8d%8d%8d%8d\n%8d\n", tp->natom,
	  tp->ntypes, tp->withH.nbond, tp->woH.nbond, tp->withH.nangl,
	  tp->woH.nangl, tp->withH.ndihe, tp->woH.ndihe, tp->nhparm,
	  tp->nparm, tp->tnexcl, tp->nres, tp->woHC.nbond, tp->woHC.nangl,
	  tp->woHC.ndihe, tp->nBAH.nbond, tp->nBAH.nangl, tp->nBAH.ndihe,
	  tp->natyp, tp->nphb, tp->ifpert, tp->pert.nbond, tp->pert.nangl,
	  tp->pert.ndihe, tp->wpert.nbond, tp->wpert.nangl, tp->wpert.ndihe,
	  tp->ifbox, tp->nmxrs, tp->ifcap, tp->blank);
  Format20a4("ATOM_NAME", tp->AtomNames, tp->natom, outp);
  DVecMult(tp->Charges, tp->natom, 18.2223);
  Format5e16("CHARGE", tp->Charges, tp->natom, outp);
  Format5e16("MASS", tp->Masses, tp->natom, outp);
  Format10i8("ATOM_TYPE_INDEX", tp->LJIdx, tp->natom, outp);
  Format10i8("NUMBER_EXCLUDED_ATOMS", tp->NExcl, tp->natom, outp);
  Format10i8("NONBONDED_PARM_INDEX", tp->NBParmIdx, tp->ntypes*tp->ntypes,
	     outp);
  Format20a4("RESIDUE_LABEL", tp->ResNames, tp->nres, outp);
  Format10i8("RESIDUE_POINTER", tp->ResLims, tp->nres, outp);
  Format5e16("BOND_FORCE_CONSTANT", tp->BondK, tp->nBAH.nbond, outp);
  Format5e16("BOND_EQUIL_VALUE", tp->BondEq, tp->nBAH.nbond, outp);
  Format5e16("ANGLE_FORCE_CONSTANT", tp->AnglK, tp->nBAH.nangl, outp);
  Format5e16("ANGLE_EQUIL_VALUE", tp->AnglEq, tp->nBAH.nangl, outp);
  Format5e16("DIHEDRAL_FORCE_CONSTANT", tp->DiheK, tp->nBAH.ndihe, outp);
  Format5e16("DIHEDRAL_PERIODICITY", tp->DiheN, tp->nBAH.ndihe, outp);
  Format5e16("DIHEDRAL_PHASE", tp->DihePhi, tp->nBAH.ndihe, outp);
  solty = (double*)calloc(tp->natyp, sizeof(double));
  Format5e16("SOLTY", solty, tp->natyp, outp);
  free(solty);
  Format5e16("LENNARD_JONES_ACOEF", tp->LJA, tp->ntypes*(tp->ntypes+1)/2,
	     outp);
  Format5e16("LENNARD_JONES_BCOEF", tp->LJB, tp->ntypes*(tp->ntypes+1)/2,
	     outp);
  Format10i8("BONDS_INC_HYDROGEN", (int*)tp->BIncH, 3*tp->withH.nbond, outp);
  Format10i8("BONDS_WITHOUT_HYDROGEN", (int*)tp->BNoH, 3*tp->woH.nbond, outp);
  Format10i8("ANGLES_INC_HYDROGEN", (int*)tp->AIncH, 4*tp->withH.nangl, outp);
  Format10i8("ANGLES_WITHOUT_HYDROGEN", (int*)tp->ANoH, 4*tp->woH.nangl, outp);
  Format10i8("DIHEDRALS_INC_HYDROGEN", (int*)tp->HIncH, 5*tp->withH.ndihe,
	     outp);
  Format10i8("DIHEDRALS_WITHOUT_HYDROGEN", (int*)tp->HNoH, 5*tp->woH.ndihe,
	     outp);
  Format10i8("EXCLUDED_ATOMS_LIST", tp->ExclList,
	     tp->ConExcl[tp->natom-1] + tp->NExcl[tp->natom-1], outp);
  Format5e16("HBOND_ACOEF", tp->SolA, tp->nphb*(tp->nphb+1)/2, outp);
  Format5e16("HBOND_BCOEF", tp->SolB, tp->nphb*(tp->nphb+1)/2, outp);
  Format5e16("HBCUT", tp->HBCut, tp->nphb*(tp->nphb+1)/2, outp);
  Format20a4("AMBER_ATOM_TYPE", tp->AtomTypes, tp->natom, outp);
  Format20a4("TREE_CHAIN_CLASSIFICATION", tp->TreeSymbols, tp->natom, outp);
  Format10i8("JOIN_ARRAY", tp->Join, tp->natom, outp);
  Format10i8("IROTAT", tp->Rotat, tp->natom, outp);
  if (tp->ifbox > 0) {
    itmp = (int*)malloc(3*sizeof(int));
    itmp[0] = tp->iptres;
    itmp[1] = tp->nspm;
    itmp[2] = tp->nspsol;
    Format10i8("SOLVENT_POINTERS", itmp, 3, outp);
    free(itmp);
    Format10i8("ATOMS_PER_MOLECULE", tp->Nsp, tp->nspm, outp);
    solty = (double*)malloc(4*sizeof(double));
    solty[0] = tp->gdim[4];
    solty[1] = tp->gdim[0];
    solty[2] = tp->gdim[1];
    solty[3] = tp->gdim[2];
    Format5e16("BOX_DIMENSIONS", solty, 4, outp);
  }
  fprintf(outp, "%%FLAG RADIUS_SET\n%%FORMAT(1a80)\n"
	  "%s", tp->RadSet);
  Format5e16("RADII", tp->Radii, tp->natom, outp);
  Format5e16("SCREEN", tp->Screen, tp->natom, outp);
  fclose(outp);
}

/***=======================================================================***/
/*** EnumerateBonds: this function loops through all bond lists, creating  ***/
/***                 atom-specific lists of bonds based on the Ath atom in ***/
/***                 the A-B sequence.  The Ath atom "owns" the bond;      ***/
/***                 cells which control the Ath atom (cells which have    ***/
/***                 the Ath atom in their primary sector) are responsible ***/
/***                 for figuring out where other atoms in the bond are    ***/
/***                 and then computing forces.                            ***/
/***                                                                       ***/
/***                 This function should be called once, after the system ***/
/***                 topology has been read, as is done at the end of      ***/
/***                 GetPrmTop() above.                                    ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:  the topology                                                   ***/
/***=======================================================================***/
void EnumerateBonds(prmtop *tp)
{
  int i, j, k;

  /*** Allocate the array for all atoms ***/
  tp->BLC = (bondlist*)calloc(tp->natom, sizeof(bondlist));

  /*** One pass to figure out how many bonds each atom controls. ***/
  for (i = 0; i < tp->withH.nbond; i++) {
    j = tp->BIncH[i].a;
    if (j < 0) {
      printf("EnumerateBonds >> Error.  Bond anchor index value is %d.\n"
	     "EnumerateBonds >> No anchor index value should be less than "
	     "zero.\n", j);
      exit(1);
    }
    tp->BLC[j].nbond += 1;
  }
  for (i = 0; i < tp->woH.nbond; i++) {
    j = tp->BNoH[i].a;
    if (j < 0) {
      printf("EnumerateBonds >> Error.  Bond anchor index value is %d.\n"
	     "EnumerateBonds >> No anchor index value should be less than "
	     "zero.\n", j);
      exit(1);
    }
    tp->BLC[j].nbond += 1;
  }

  /*** Allocate the bond arrays for each individual atom ***/
  for (i = 0; i < tp->natom; i++) {
    tp->BLC[i].BC = (bondcomm*)malloc(tp->BLC[i].nbond*sizeof(bondcomm));
    tp->BLC[i].nbond = 0;
  }

  /*** A second pass to index the bonds by anchor atoms ***/
  for (i = 0; i < tp->withH.nbond; i++) {
    j = tp->BIncH[i].a;
    k = tp->BLC[j].nbond;
    tp->BLC[j].BC[k].a = j;
    tp->BLC[j].BC[k].b = tp->BIncH[i].b;
    tp->BLC[j].BC[k].t = tp->BIncH[i].idx;
    tp->BLC[j].BC[k].H = 1;
    tp->BLC[j].nbond = k+1;
  }
  for (i = 0; i < tp->woH.nbond; i++) {
    j = tp->BNoH[i].a;
    k = tp->BLC[j].nbond;
    tp->BLC[j].BC[k].a = j;
    tp->BLC[j].BC[k].b = tp->BNoH[i].b;
    tp->BLC[j].BC[k].t = tp->BNoH[i].idx;
    tp->BLC[j].BC[k].H = 0;
    tp->BLC[j].nbond = k+1;
  }
}

/***=======================================================================***/
/*** EnumerateAngles: this function loops through all angle lists, making  ***/
/***                  atom-specific lists of angles based on the Bth atom  ***/
/***                  in the A-B-C sequence.  The Bth atom "owns" the      ***/
/***                  angle.                                               ***/
/***                                                                       ***/
/***                  This function should be called once, after the       ***/
/***                  system topology has been read, as is done at the end ***/
/***                  of GetPrmTop() above.                                ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:  the topology                                                   ***/
/***=======================================================================***/
void EnumerateAngles(prmtop *tp)
{
  int i, j, k, atma, atmc;

  /*** Allocate the array for all atoms ***/
  tp->ALC = (angllist*)calloc(tp->natom, sizeof(angllist));

  /*** One pass to figure out how many angles each atom controls. ***/
  for (i = 0; i < tp->withH.nangl; i++) {
    j = tp->AIncH[i].b;
    if (j < 0) {
      printf("EnumerateAngles >> Error.  Angle anchor index value is %d.\n"
	     "EnumerateAngles >> No anchor index value should be less than "
	     "zero.\n", j);
      exit(1);
    }
    tp->ALC[j].nangl += 1;
  }
  for (i = 0; i < tp->woH.nangl; i++) {
    j = tp->ANoH[i].b;
    if (j < 0) {
      printf("EnumerateAngles >> Error.  Angle anchor index value is %d.\n"
	     "EnumerateAngles >> No anchor index value should be less than "
	     "zero.\n", j);
      exit(1);
    }
    tp->ALC[j].nangl += 1;
  }

  /*** Allocate the angle arrays for each individual atom ***/
  for (i = 0; i < tp->natom; i++) {
    tp->ALC[i].AC = (anglcomm*)malloc(tp->ALC[i].nangl*sizeof(anglcomm));
    tp->ALC[i].nangl = 0;
  }

  /*** A second pass to index the angles by anchor atoms ***/
  for (i = 0; i < tp->withH.nangl; i++) {
    j = tp->AIncH[i].b;
    k = tp->ALC[j].nangl;
    tp->ALC[j].AC[k].a = tp->AIncH[i].a;
    tp->ALC[j].AC[k].c = tp->AIncH[i].c;
    tp->ALC[j].AC[k].t = tp->AIncH[i].idx;
    tp->ALC[j].nangl = k+1;
  }
  for (i = 0; i < tp->woH.nangl; i++) {
    j = tp->ANoH[i].b;
    k = tp->ALC[j].nangl;
    tp->ALC[j].AC[k].a = tp->ANoH[i].a;
    tp->ALC[j].AC[k].c = tp->ANoH[i].c;
    tp->ALC[j].AC[k].t = tp->ANoH[i].idx;
    tp->ALC[j].nangl = k+1;
  }

  /*** A final pass to determine whether to exclude 1:3 interactions ***/
  for (i = 0; i < tp->natom; i++) {
    for (j = 0; j < tp->ALC[i].nangl; j++) {

      /*** By default, 1:3 interactions are excluded ***/
      tp->ALC[i].AC[j].excl = 1;
      atma = tp->ALC[i].AC[j].a;
      atmc = tp->ALC[i].AC[j].c;
      for (k = 0; k < tp->BLC[atma].nbond; k++) {
	if (tp->BLC[atma].BC[k].b == atmc) {
	  tp->ALC[i].AC[j].excl = 0;
	  break;
	}
      }
      if (tp->ALC[i].AC[j].excl == 1) {
	for (k = 0; k < tp->BLC[atmc].nbond; k++) {
	  if (tp->BLC[atmc].BC[k].b == atma) {
	    tp->ALC[i].AC[j].excl = 0;
	    break;
	  }
	}
      }
    }
  }
}

/***=======================================================================***/
/*** EnumerateDihedrals: this function loops through all dihedral lists,   ***/
/***                     creating atom-specific lists of dihedrals based   ***/
/***                     on the Bth atom in the A-B-C-D sequence.  The Bth ***/
/***                     atom "owns" the dihedral; cells which control the ***/
/***                     Bth atom for bonded interactions are responsible  ***/
/***                     for figuring out where the other atoms in the     ***/
/***                     dihedral are and then computing forces.           ***/
/***                                                                       ***/
/***                     This function should be called once, after the    ***/
/***                     system topology has been read, as is done at the  ***/
/***                     end of GetPrmTop() above.                         ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:  the topology                                                   ***/
/***=======================================================================***/
void EnumerateDihedrals(prmtop *tp)
{
  int i, j, k, maxdihe, dihefound;
  dihelist scrd;

  /*** Allocate the array for all atoms ***/
  tp->HLC = (dihelist*)calloc(tp->natom, sizeof(dihelist));

  /*** One pass to figure out how many dihedrals each atom controls. ***/
  /*** The distinction between dihedrals with and without hydrogen   ***/
  /*** is a historical artifact, so we lump them all together here.  ***/
  for (i = 0; i < tp->withH.ndihe; i++) {
    j = tp->HIncH[i].b;
    if (j < 0) {
      printf("EnumerateDihedrals >> Error.  Dihedral anchor index value is "
	     "%d.\nEnumerateDihedrals >> No anchor index value should be less "
	     "than zero.\n", j);
      exit(1);
    }
    tp->HLC[j].ndihe += 1;
  }
  for (i = 0; i < tp->woH.ndihe; i++) {
    j = tp->HNoH[i].b;
    if (j < 0) {
      printf("EnumerateDihedrals >> Error.  Dihedral anchor index value is "
	     "%d.\nEnumerateDihedrals >> No anchor index value should be less "
	     "than zero.\n", j);
      exit(1);
    }
    tp->HLC[j].ndihe += 1;
  }

  /*** Allocate the dihedral arrays for each individual atom ***/
  for (i = 0; i < tp->natom; i++) {
    tp->HLC[i].HC = (dihecomm*)malloc(tp->HLC[i].ndihe*sizeof(dihecomm));
    tp->HLC[i].ndihe = 0;
  }

  /*** Another pass to assign dihedrals to individual atoms ***/
  for (i = 0; i < tp->withH.ndihe; i++) {
    j = tp->HIncH[i].b;
    k = tp->HLC[j].ndihe;
    tp->HLC[j].HC[k].impr = (tp->HIncH[i].d < 0) ? 1 : 0;
    tp->HLC[j].HC[k].a = abs(tp->HIncH[i].a);
    tp->HLC[j].HC[k].eval14 = (tp->HIncH[i].c < 0) ? 0 : 1;
    tp->HLC[j].HC[k].c = abs(tp->HIncH[i].c);
    tp->HLC[j].HC[k].d = abs(tp->HIncH[i].d);
    tp->HLC[j].HC[k].nt = tp->HIncH[i].idx;
    tp->HLC[j].ndihe = k+1;
  }
  for (i = 0; i < tp->woH.ndihe; i++) {
    j = tp->HNoH[i].b;
    k = tp->HLC[j].ndihe;
    tp->HLC[j].HC[k].impr = (tp->HNoH[i].d < 0) ? 1 : 0;
    tp->HLC[j].HC[k].a = abs(tp->HNoH[i].a);
    tp->HLC[j].HC[k].eval14 = (tp->HNoH[i].c < 0) ? 0 : 1;
    tp->HLC[j].HC[k].c = abs(tp->HNoH[i].c);
    tp->HLC[j].HC[k].d = abs(tp->HNoH[i].d);
    tp->HLC[j].HC[k].nt = tp->HNoH[i].idx;
    tp->HLC[j].ndihe = k+1;
  }

  /*** Now a pass to condense dihedrals that share the same atoms ***/
  maxdihe = 0;
  for (i = 0; i < tp->natom; i++) {
    if (tp->HLC[i].ndihe > maxdihe) {
      maxdihe = tp->HLC[i].ndihe;
    }
  }
  scrd.HC = (dihecomm*)malloc(maxdihe*sizeof(dihecomm));
  for (i = 0; i < maxdihe; i++) {
    scrd.HC[i].t = (int*)malloc(maxdihe*sizeof(int));
  }
  for (i = 0; i < tp->natom; i++) {

    /*** Initialize the buffer dihedral command structure ***/
    scrd.ndihe = 0;
    for (j = 0; j < maxdihe; j++) {
      scrd.HC[j].nt = 0;
    }

    /*** Loop over all dihedrals controlled by this atom and ***/
    /*** contribute them to the buffer structure.            ***/
    for (j = 0; j < tp->HLC[i].ndihe; j++) {
      dihefound = 0;
      for (k = 0; k < scrd.ndihe; k++) {
	if (scrd.HC[k].a == tp->HLC[i].HC[j].a &&
	    scrd.HC[k].c == tp->HLC[i].HC[j].c &&
	    scrd.HC[k].d == tp->HLC[i].HC[j].d &&
	    scrd.HC[k].impr == tp->HLC[i].HC[j].impr) {
	  dihefound = 1;
	  scrd.HC[k].t[scrd.HC[k].nt] = tp->HLC[i].HC[j].nt;
	  scrd.HC[k].nt += 1;
	  if (tp->HLC[i].HC[j].eval14 == 1) {
	    scrd.HC[k].eval14 = 1;
	  }
	  break;
	}
      }

      /*** Make a new dihedral for this atom ***/
      if (dihefound == 0) {
	scrd.HC[scrd.ndihe].a = tp->HLC[i].HC[j].a;
	scrd.HC[scrd.ndihe].c = tp->HLC[i].HC[j].c;
	scrd.HC[scrd.ndihe].d = tp->HLC[i].HC[j].d;
	scrd.HC[scrd.ndihe].impr = tp->HLC[i].HC[j].impr;
	scrd.HC[scrd.ndihe].eval14 = tp->HLC[i].HC[j].eval14;
	scrd.HC[scrd.ndihe].t[0] = tp->HLC[i].HC[j].nt;
	scrd.HC[scrd.ndihe].nt = 1;
	scrd.ndihe += 1;
      }
    }

    /*** Copy the buffer dihedral command structure into the topology ***/
    tp->HLC[i].ndihe = scrd.ndihe;
    free(tp->HLC[i].HC);
    tp->HLC[i].HC = (dihecomm*)malloc(scrd.ndihe*sizeof(dihecomm));
    for (j = 0; j < scrd.ndihe; j++) {
      tp->HLC[i].HC[j].a = scrd.HC[j].a;
      tp->HLC[i].HC[j].c = scrd.HC[j].c;
      tp->HLC[i].HC[j].d = scrd.HC[j].d;
      tp->HLC[i].HC[j].impr = scrd.HC[j].impr;
      tp->HLC[i].HC[j].eval14 = scrd.HC[j].eval14;
      tp->HLC[i].HC[j].nt = scrd.HC[j].nt;
      tp->HLC[i].HC[j].t = CpyIVec(scrd.HC[j].t, scrd.HC[j].nt);
    }
  }

  /*** Free allocated memory ***/
  for (i = 0; i < maxdihe; i++) {
    free(scrd.HC[i].t);
  }
  free(scrd.HC);
}

/***=======================================================================***/
/*** OrderParameters: this routine re-organizes the bond, angle, dihedral, ***/
/***                  and van-der Waals parameters into things that are    ***/
/***                  much more intuitive to read from.                    ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:  the topology                                                   ***/
/***=======================================================================***/
void OrderParameters(prmtop *tp)
{
  int i, j, haslj;
  double aval, bval, sig, eps, gam, cval, dval;
  double *ftmp, *utmp;

  /*** Build the Lennard-Jones tables ***/
  if (tp->ljbuck == 0) {
    tp->LJftab = CreateDmat(tp->ntypes, 2*tp->ntypes, 0);
    tp->LJutab = CreateDmat(tp->ntypes, 2*tp->ntypes, 0);
    for (i = 0; i < tp->ntypes; i++) {
      ftmp = tp->LJftab.map[i];
      utmp = tp->LJutab.map[i];
      for (j = 0; j < tp->ntypes; j++) {

	/*** There is no support for the 10-12 hydrogen bonding potential. ***/
	/*** Atoms whose van-der Waals types are less than 0 are assumed   ***/
	/*** to have no van-der Waals properties whatsoever.               ***/
	if (tp->NBParmIdx[i*tp->ntypes+j] >= 0) {
	  aval = tp->LJA[tp->NBParmIdx[i*tp->ntypes+j]];
	  bval = tp->LJB[tp->NBParmIdx[i*tp->ntypes+j]];
	  if (aval > 1.0e-8) {
	    sig = pow(aval/bval, 1.0/6.0);
	    eps = bval/pow(sig, 6.0);
	    ftmp[2*j] = -12.0*eps*pow(sig, 12.0);
	    ftmp[2*j+1] = 6.0*eps*pow(sig, 6.0);
	    utmp[2*j] = eps*pow(sig, 12.0);
	    utmp[2*j+1] = -eps*pow(sig, 6.0);
	  }
	  else {
	    ftmp[2*j] = 0.0;
	    ftmp[2*j+1] = 0.0;
	    utmp[2*j] = 0.0;
	    utmp[2*j+1] = 0.0;
	  }
	}
	else {
	  ftmp[2*j] = 0.0;
	  ftmp[2*j+1] = 0.0;
	  utmp[2*j] = 0.0;
	  utmp[2*j+1] = 0.0;
	}
      }
    }
  }
  else if (tp->ljbuck == 1) {
    tp->LJftab = CreateDmat(tp->ntypes, 4*tp->ntypes, 0);
    tp->LJutab = CreateDmat(tp->ntypes, 4*tp->ntypes, 0);
    for (i = 0; i < tp->ntypes; i++) {
      ftmp = tp->LJftab.map[i];
      utmp = tp->LJutab.map[i];
      for (j = 0; j < tp->ntypes; j++) {

	/*** This modified Buckingham exp-6 potential uses a (1/r)^12 term ***/
	/*** to continue the positive divergence of the potential as r     ***/
	/*** goes to zero, such that there isn't even an inflection point. ***/
	/*** This (1/r)^12 terms requires its own scaling factor, which is ***/
	/*** computed here based on the exp terms (aval and bval) and the  ***/
	/*** (1/r)^6 term (cval).  The potential is, succinctly:           ***/
	/***                                                               ***/
	/***      U(r) = aval*exp(-bval*r) - cval/(r^6) + dval/(r^12)      ***/
	/***                                                               ***/
	/*** The terms aval, bval, and cval are computed based on the eps, ***/
	/*** gam, and sig parameters which are actually read in from the   ***/
	/*** topology file.  This is in contrast to the Lennard-Jones      ***/
	/*** potential, where it is aval and bval which are read from the  ***/
	/*** topology file and eps and sig which are back-calculated.      ***/
        if (tp->NBParmIdx[i*tp->ntypes+j] >= 0) {
          eps = tp->LJA[tp->NBParmIdx[i*tp->ntypes+j]];
          gam = tp->LJB[tp->NBParmIdx[i*tp->ntypes+j]];
          sig = tp->LJC[tp->NBParmIdx[i*tp->ntypes+j]];
	  aval = 6.0*eps*exp(eps)/(gam-6.0);
	  bval = gam/sig;
	  cval = eps*pow(sig, 6.0)/(1.0-6.0/gam);
	  dval = DetermineDCoef(eps, gam, sig);
          utmp[4*j] = aval;
          utmp[4*j+1] = bval;
          utmp[4*j+2] = cval;
          utmp[4*j+3] = dval;
	  ftmp[4*j] = -aval*bval;
	  ftmp[4*j+1] = bval;
	  ftmp[4*j+2] = -6.0*cval;
	  ftmp[4*j+3] = -12.0*dval;
        }
        else {
          ftmp[4*j] = 0.0;
          ftmp[4*j+1] = 0.0;
          ftmp[4*j+2] = 0.0;
          ftmp[4*j+3] = 0.0;
          utmp[4*j] = 0.0;
          utmp[4*j+1] = 0.0;
          utmp[4*j+2] = 0.0;
          utmp[4*j+3] = 0.0;
        }
      }
    }
  }
  else {
    printf("OrderParameters >> Error.  Unrecognized van-der Waals model type "
	   "%d.\n", tp->ljbuck);
    exit(1);
  }

  /*** Identify atoms that have no Lennard-Jones properties ***/
  for (i = 0; i < tp->ntypes; i++) {
    haslj = 0;
    utmp = tp->LJutab.map[i];
    for (j = 0; j < tp->ntypes; j++) {
      if (fabs(utmp[2*j]) > 1.0e-8 || fabs(utmp[2*j+1]) > 1.0e-8) {
	haslj = 1;
	break;
      }
    }
    if (haslj == 0) {
      for (j = 0; j < tp->natom; j++) {
	if (tp->LJIdx[j] == i) {
	  tp->LJIdx[j] = -i;
	}
      }
    }
  }

  /*** Organize bond parameters ***/
  tp->BParam = (bonddef*)malloc(tp->nBAH.nbond*sizeof(bonddef));
  for (i = 0; i < tp->nBAH.nbond; i++) {
    tp->BParam[i].K = tp->BondK[i];
    tp->BParam[i].l0 = tp->BondEq[i];
  }

  /*** Organize angle parameters ***/
  tp->AParam = (angldef*)malloc(tp->nBAH.nangl*sizeof(angldef));
  for (i = 0; i < tp->nBAH.nangl; i++) {
    tp->AParam[i].K = tp->AnglK[i];
    tp->AParam[i].th0 = tp->AnglEq[i];
  }

  /*** Organize dihedral parameters ***/
  tp->HParam = (dihedef*)malloc(tp->nBAH.ndihe*sizeof(dihedef));
  for (i = 0; i < tp->nBAH.ndihe; i++) {
    tp->HParam[i].K = tp->DiheK[i];
    tp->HParam[i].N = tp->DiheN[i];
    tp->HParam[i].Phi = tp->DihePhi[i];
  }

}

/***=======================================================================***/
/*** FindThreePointWaters: locate all three-point water molecules within   ***/
/***                       the topology, with the user-specified water     ***/
/***                       name as the reference.                          ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:  the topology                                                   ***/
/***=======================================================================***/
void FindThreePointWaters(prmtop *tp)
{
  int i, j, nox, nh1, nh2, nwat;

  /*** Allocate space for the array of fast water indices ***/
  nwat = 0;
  tp->SHL = (cnstcomm*)calloc(tp->natom, sizeof(cnstcomm));

  /*** If SETTLE is not requested, return ***/
  if (tp->settle == 0) {
    tp->nwat = nwat;
    return;
  }

  /*** Locate all fast waters ***/
  for (i = 0; i < tp->nres; i++) {
    if (strncmp(&tp->ResNames[4*i], tp->WaterName, 4) == 0) {
      nox = -1;
      nh1 = -1;
      nh2 = -1;
      for (j = tp->ResLims[i]; j < tp->ResLims[i+1]; j++) {
	if (tp->Masses[j] > 15.9 && tp->Masses[j] < 16.1) {	    
	  nox = j;
	}
	else if (tp->Masses[j] < 1.1 && tp->Masses[j] > 0.9 && nh1 < 0) {
	  nh1 = j;
	}
	else if (tp->Masses[j] < 1.1 && tp->Masses[j] > 0.9 && nh2 < 0) {
	  nh2 = j;
	}
      }
      if (nox < 0) {
	printf("FindThreePointWaters >> Error.  Oxygen not found for residue "
	       "%d.\n", i);
	exit(1);
      }
      if (nh1 < 0 || nh2 < 0) {
	printf("FindThreePointWaters >> Error.  Hydrogen not found for "
	       "residue %d.\n", i);
	exit(1);
      }

      /*** The oxygen atom is flagged with a +1 SHL execution value, ***/
      /*** to signify that it controls a three-point water style     ***/
      /*** constraint group.                                         ***/
      tp->SHL[nox].exe = 1;
      tp->SHL[nox].blist = (int*)malloc(2*sizeof(int));
      tp->SHL[nox].blist[0] = nh1;
      tp->SHL[nox].blist[1] = nh2;

      /*** The hydrogens are flagged with -1 SHL execution values, ***/
      /*** to signify that they are part of a constraint group     ***/
      /*** but not in control of it.                               ***/
      tp->SHL[nh1].exe = -1;
      tp->SHL[nh2].exe = -1;
      nwat++;
    }
  }
  tp->nwat = nwat;
}

/***=======================================================================***/
/*** CompSettleGeom: compute the geometric parameters for SETTLE on the    ***/
/***                 user-specified water molecules.                       ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:  the topology                                                   ***/
/***=======================================================================***/
void CompSettleGeom(prmtop *tp)
{
  int i, j, k, resfound, nox, atma, atmb, nh1, nh2;
  double dOH, dHH, massO, massH, t1;

  /*** Find an example of a rigid water molecule ***/
  resfound = 0;
  i = 0;
  while (resfound == 0 && i < tp->nres) {
    for (j = tp->ResLims[i]; j < tp->ResLims[i+1]; j++) {
      if (tp->SHL[j].exe == 1) {
	resfound = 1;
	nox = j;
	nh1 = tp->SHL[j].blist[0];
	nh2 = tp->SHL[j].blist[1];
      }
    }
    if (resfound == 0) {
      i++;
      continue;
    }

    /*** Get bond length for O-H ***/
    dOH = -1.0;
    dHH = -1.0;
    for (j = tp->ResLims[i]; j < tp->ResLims[i+1]; j++) {
      for (k = 0; k < tp->BLC[j].nbond; k++) {
	atma = tp->BLC[j].BC[k].a;
	atmb = tp->BLC[j].BC[k].b;
	if (tp->Masses[atma] > 14.9 &&
	    tp->Masses[atmb] < 3.1 && tp->Masses[atmb] > 0.5) {
	  dOH = tp->BParam[tp->BLC[j].BC[k].t].l0;
	}
	else if (tp->Masses[atma] < 3.1 && tp->Masses[atma] > 0.5 &&
		 tp->Masses[atmb] > 14.9) {
	  dOH = tp->BParam[tp->BLC[j].BC[k].t].l0;
	}
	else if (tp->Masses[atma] < 3.1 && tp->Masses[atma] > 0.5 &&
		 tp->Masses[atmb] < 3.1 && tp->Masses[atmb] > 0.5) {
	  dHH = tp->BParam[tp->BLC[j].BC[k].t].l0;
	}
      }
    }

    /*** Check that bond lengths are known ***/
    if (dOH < 0.0) {
      printf("CompSettleGeom >> Error.  Unable to determine O-H bond "
	     "length.\n");
      exit(1);
    }
    if (dHH < 0.0) {
      printf("CompSettleGeom >> Error.  Unable to determine H-H bond "
	     "length.\n");
      exit(1);
    }

    massO = tp->Masses[nox];
    massH = tp->Masses[nh1];
    if (fabs(tp->Masses[nh2] - tp->Masses[nh1]) > 1.0e-9) {
      printf("CompSettleGeom >> Error.  Masses of hydrogen atoms do not "
	     "agree.\n");
      exit(1);
    }
    t1 = 0.5 * massO / massH;
    tp->FWtab.rc = 0.5* dHH;
    tp->FWtab.ra = sqrt(dOH*dOH - tp->FWtab.rc*tp->FWtab.rc) / (t1 + 1.0);
    tp->FWtab.rb = t1 * tp->FWtab.ra;
    tp->FWtab.mO = massO;
    tp->FWtab.mH = massH;
  }
  if (resfound == 0) {
    printf("CompSettleGeom >> Error.  Unable to locate a rigid water "
	   "molecule.\n");
    exit(1);
  }
}

/***=======================================================================***/
/*** Chirality: get the chirality of a center, given the number CA of the  ***/
/***            atom at the center and four atoms around it.  We expect A  ***/
/***            to be the first atom in the list, B the second, C the      ***/
/***            third, and D the fourth atom in order of greatest to least ***/
/***            chiral dominance.                                          ***/
/***=======================================================================***/
static int Chirality(prmtop *tp, coord *tc, int CA, int A, int B, int C, int D)
{
  int i;
  double cax[3], ax[3], bx[3], cx[3], dx[3], ra[3], rb[3], rc[3], rd[3];
  double prja[3], prjb[3], prjc[3], acrb[3], bcrc[3];

  /*** Get coordinates ***/
  for (i = 0; i < 3; i++) {
    cax[i] = tc->loc[3*CA+i];
    ax[i] = tc->loc[3*A+i];
    bx[i] = tc->loc[3*B+i];
    cx[i] = tc->loc[3*C+i];
    dx[i] = tc->loc[3*D+i];
    ra[i] = ax[i] - cax[i];
    rb[i] = bx[i] - cax[i];
    rc[i] = cx[i] - cax[i];
    rd[i] = dx[i] - cax[i];
  }

  /*** Remove projection ***/
  Project(ra, rd, prja, 3);
  Project(rb, rd, prjb, 3);
  Project(rc, rd, prjc, 3);
  for (i = 0; i < 3; i++) {
    ra[i] -= prja[i];
    rb[i] -= prjb[i];
    rc[i] -= prjc[i];
  }

  /*** Find chirality via cross products ***/
  CrossP(ra, rb, acrb);
  CrossP(rb, rc, bcrc);

  /*** L chirality is when ACRB and BCRC point opposite RD ***/
  if (DotP(acrb, rd, 3) < 0.0 && DotP(bcrc, rd, 3) < 0.0) {
    return 1;
  }

  /*** Otherwise, this has D chirality ***/
  return -1;
}

/***=======================================================================***/
/*** ProteinChiralityCheck: checks the chiral centers of amino acids.  Any ***/
/***                        residues with D chirality are reported.        ***/
/***=======================================================================***/
int ProteinChiralityCheck(prmtop *tp, coord *tc, FILE *outp)
{
  int i, nca, nn, nc, ncb, nha, iwarning;
  char* ctmp;

  iwarning = 0;
  for (i = 0; i < tp->nres; i++) {
    ctmp = &tp->ResNames[4*i];
    if (!(strncmp(ctmp, "ALA", 3) == 0 || strncmp(ctmp, "ARG", 3) == 0 ||
          strncmp(ctmp, "ASN", 3) == 0 || strncmp(ctmp, "ASP", 3) == 0 ||
          strncmp(ctmp, "ASH", 3) == 0 || strncmp(ctmp, "GLH", 3) == 0 ||
          strncmp(ctmp, "CYS", 3) == 0 || strncmp(ctmp, "CYX", 3) == 0 ||
          strncmp(ctmp, "GLN", 3) == 0 || strncmp(ctmp, "GLU", 3) == 0 ||
          strncmp(ctmp, "HID", 3) == 0 || strncmp(ctmp, "HIE", 3) == 0 ||
          strncmp(ctmp, "HIP", 3) == 0 || strncmp(ctmp, "HIS", 3) == 0 ||
          strncmp(ctmp, "ILE", 3) == 0 || strncmp(ctmp, "LEU", 3) == 0 ||
          strncmp(ctmp, "LYS", 3) == 0 || strncmp(ctmp, "MET", 3) == 0 ||
          strncmp(ctmp, "PHE", 3) == 0 || strncmp(ctmp, "PRO", 3) == 0 ||
          strncmp(ctmp, "SER", 3) == 0 || strncmp(ctmp, "THR", 3) == 0 ||
          strncmp(ctmp, "TRP", 3) == 0 || strncmp(ctmp, "TYR", 3) == 0 ||
          strncmp(ctmp, "VAL", 3) == 0)) {
      continue;
    }

    /*** If we're still here, we've got some amino acid ***/
    nca = FindAtom(tp, tp->ResLims[i], tp->ResLims[i+1], "CA  ");
    nn = FindAtom(tp, tp->ResLims[i], tp->ResLims[i+1], "N   ");
    nc = FindAtom(tp, tp->ResLims[i], tp->ResLims[i+1], "C   ");
    ncb = FindAtom(tp, tp->ResLims[i], tp->ResLims[i+1], "CB  ");
    nha = FindAtom(tp, tp->ResLims[i], tp->ResLims[i+1], "HA  ");

    /*** Now, get the chirality of these four atoms ***/
    if (Chirality(tp, tc, nca, nn, nc, ncb, nha) == -1) {
      fprintf(outp, " Warning.  Residue %.4s %d has D chirality at its CA "
	      "atom.\n", &tp->ResNames[4*i], i);
      iwarning++;
    }
  }

  return iwarning;
}

/***=======================================================================***/
/*** FindDisulfides: find sulfur atoms on CYS residues that are very close ***/
/***                 together and prompt the user if they are not bonded.  ***/
/***=======================================================================***/
int FindDisulfides(prmtop *tp, coord *tc, FILE *outp)
{
  int i, j, k, iresno, jresno, bonded, ndss, iwarning;
  double dx[3];
  char wrnmsg[256];

  ndss = 0;
  iwarning = 0;
  for (i = 0; i < tp->natom; i++) {
    if ((tp->Masses[i] <= 32.0 || tp->Masses[i] >= 33.0) &&
	tp->AtomNames[4*i] != 'S') {
      continue;
    }

    /*** This looks like sulfur, but is it part of a cysteine? ***/
    for (j = 0; j < tp->nres; j++) {
      if (tp->ResLims[j] < i) {
	iresno = j;
      }
    }
    if (strncmp(&tp->ResNames[4*iresno], "CY", 2) != 0 &&
	strncmp(&tp->ResNames[4*iresno+1], "CY", 2) != 0) {
      continue;
    }
    for (j = i+1; j < tp->natom; j++) {
      if ((tp->Masses[j] <= 32.0 || tp->Masses[j] >= 33.0) &&
	  tp->AtomNames[4*j] != 'S') {
	continue;
      }

      for (k = 0; k < tp->nres; k++) {
	if (tp->ResLims[k] < j) {
	  jresno = k;
	}
      }
      if (strncmp(&tp->ResNames[4*jresno], "CY", 2) != 0 &&
	  strncmp(&tp->ResNames[4*jresno+1], "CY", 2) != 0) {
	continue;
      }

      /*** So, we've got two atoms that look interesting ***/
      for (k = 0; k < 3; k++) {
	dx[k] = tc->loc[3*i+k] - tc->loc[3*j+k];
      }
      if (SQ_DIST(dx[0], dx[1], dx[2]) < 16.0) {

	/*** They're close together... are they bonded? ***/
	bonded = 0;
	for (k = 0; k < tp->woH.nbond; k++) {
	  if ((tp->BNoH[k].a == i && tp->BNoH[k].b == j) ||
	      (tp->BNoH[k].b == i && tp->BNoH[k].a == j)) {
	    bonded = 1;
	    ndss++;
	    break;
	  }
	}
	if (bonded == 0) {
	  sprintf(wrnmsg, " Warning.  Atoms %.4s %6d %.4s and %.4s %6d %.4s "
		  "may be part of an unfulfilled\n  disulfide bridge.",
		  &tp->ResNames[4*iresno], iresno+1, &tp->AtomNames[4*i],
                  &tp->ResNames[4*jresno], jresno+1, &tp->ResNames[4*jresno]);
	  PrintParagraph(wrnmsg, 79, outp);
	  iwarning++;
	}
      }
    }
  }
  if (ndss > 0) {
    fprintf(outp, " There are %d disulfide bonds in the system.\n", ndss);
  }

  return iwarning;
}

/***=======================================================================***/
/*** LocateResID: locate the number of a residue containing atom ai within ***/
/***              topology tp.  This function uses a binary search.        ***/
/***=======================================================================***/
int LocateResID(prmtop *tp, int ai, int llim, int hlim)
{
  int rnum, cres;

  /*** Midpoint ***/
  cres = (hlim+llim)/2;

  /*** First, test the middle residue ***/
  if (ai >= tp->ResLims[cres] && ai < tp->ResLims[cres+1]) {
    return cres;
  }
  else if (ai < tp->ResLims[cres]) {
    rnum = LocateResID(tp, ai, llim, cres);
  }
  else if (ai >= tp->ResLims[cres+1]) {
    rnum = LocateResID(tp, ai, cres, hlim);
  }

  return rnum;
}
