#ifndef TopologyStructs
#define TopologyStructs

#include "Constants.h"
#include "MatrixDS.h"
#include "CompFrcDS.h"

/***=======================================================================***/
/*** BNDANGDIHE: a group of three integers, bonds, angles, and dihedrals.  ***/
/***=======================================================================***/
struct bndangdihe {
  int nbond;
  int nangl;
  int ndihe;
};
typedef struct bndangdihe bah;

/***=======================================================================***/
/*** BONDIDX: a bond index structure matching the AMBER topology file      ***/
/***          format.                                                      ***/
/***=======================================================================***/
struct bondidx {
  int a;          // The "A" atom in the A-B bond, as shown below:
                  //
                  //         A--B
                  //
                  // The "A" atom serves as the "anchor" when determining
                  // which cell or group of cells is charged with computing
                  // the bond force and energy
  int b;          // The "B" atom
  int idx;        // An index into the bond parameter arrays
};
typedef struct bondidx bond;

/***=======================================================================***/
/*** BONDCOMMAND: the bond command structure, containing information       ***/
/***              needed during program execution for evaluating a         ***/
/***              particular bond.                                         ***/
/***=======================================================================***/
struct bondcommand {
  int a;     // The "A" atom in the bond
  int b;     // The "B" atom in the bond
  int t;     // An index into the bond parameter arrays
  char H;    // Flag to tell whether a bond has hydrogen in it
};
typedef struct bondcommand bondcomm;

/***=======================================================================***/
/*** ATOMBONDLIST: a list of bond commands for a particular atom.          ***/
/***=======================================================================***/
struct atombondlist {
  int nbond;     // The number of bonds this atom controls
  bondcomm* BC;  // The bond commands
};
typedef struct atombondlist bondlist;

/***=======================================================================***/
/*** BONDDEF: a structure for defining a bond in terms that are more       ***/
/***          convenient to reference during program execution.            ***/
/***=======================================================================***/
struct BondDef {
  double K;    // Spring constant
  double l0;   // Equilibrium length
};
typedef struct BondDef bonddef;

/***=======================================================================***/
/*** ANGLIDX: an angle index structure matching the AMBER topology file    ***/
/***          format.                                                      ***/
/***=======================================================================***/
struct anglidx {
  int a;          // The "A" atom in the A-B-C angle, as shown below:
                  //
                  //              C
                  //             /
                  //         A--B
                  //
  int b;          // The "B" atom (this atom serves as the "anchor" when
                  // determining which cell or group of cells is charged with
                  // computing the angle force and energy)
  int c;          // The "C" atom
  int idx;        // An index into the angle parameter arrays
};
typedef struct anglidx angle;

/***=======================================================================***/
/*** ANGLECOMMAND: the angle command structure, containing information     ***/
/***               needed during program execution for evaluating a        ***/
/***               particular angle.                                       ***/
/***=======================================================================***/
struct anglecommand {
  int a;     // The "A" atom in the angle
  int c;     // The "C" atom in the angle
  int t;     // An index into the angle parameter arrays
  int excl;  // Flag to active exclusion of the 1:3 interaction
};
typedef struct anglecommand anglcomm;

/***=======================================================================***/
/*** ATOMANGLELIST: a list of bond commands for a particular atom.         ***/
/***=======================================================================***/
struct atomanglelist {
  int nangl;     // The number of angles this atom controls
  anglcomm* AC;  // The angle commands
};
typedef struct atomanglelist angllist;

/***=======================================================================***/
/*** ANGLEDEF: a structure for defining an angle in terms that are more    ***/
/***           convenient to reference during program execution.           ***/
/***=======================================================================***/
struct AngleDef {
  double K;    // Spring constant
  double th0;  // Equilibrium angle
};
typedef struct AngleDef angldef;

/***=======================================================================***/
/*** DIHEIDX: a dihedral index structure matching the AMBER topology file  ***/
/***          format.                                                      ***/
/***=======================================================================***/
struct diheidx {
  int a;          // The "A" atom in the A-B-C-D dihedral, as shown below:
                  //
                  //              D
                  //             /
                  //         B--C
                  //        /
                  //       A
                  //
  int b;          // The "B" atom (this atom serves as the "anchor," as
                  // discussed below)
  int c;          // The "C" atom
  int d;          // The "D" atom
  int idx;        // An index into the dihedral parameter arrays
};
typedef struct diheidx dihedral;

/***=======================================================================***/
/*** DIHEDRALCOMMAND: this structure lists multiple Fourier terms that may ***/
/***                  be associated with a dihedral, and also whether it   ***/
/***                  is an improper.  This structure is what mdgx relies  ***/
/***                  on when actually evaluating dihedrals.               ***/
/***=======================================================================***/
struct dihedralcommand {
  int a;          // The "A" atom (note that the b atom is implicit in the
                  // ordered list in which this dihedralcomm structure appears)
  int c;          // The "C" atom
  int d;          // The "D" atom
  char impr;      // Flag to tell us whether this is an improper
  char eval14;    // Flag to tell us whether to evaluate the 1:4 interactions 
  int nt;         // The number of Fourier terms
  int* t;         // A list of indices into the dihedral definitions of this
                  // topology (see the diehdraldefs structure)
};
typedef struct dihedralcommand dihecomm;

/***=======================================================================***/
/*** DIHEDRALLIST: a list of dihedrals associated with a particular atom,  ***/
/***               more suitable for my dihedral calculation schemes.      ***/
/***=======================================================================***/
struct dihedrallist {
  int ndihe;          // The number of dihedral angles this atom controls
  dihecomm* HC;       // The list of dihedral commands
};
typedef struct dihedrallist dihelist;

/***=======================================================================***/
/*** DIHEDRALDEF: a structure for defining a single Fourier series term in ***/
/***              a dihedral interaction.  This structure is used, rather  ***/
/***              than the data read directly from the topology file,      ***/
/***              during program execution.                                ***/
/***=======================================================================***/
struct DihedralDef {
  double K;     // The amplitude
  double N;     // The periodicity
  double Phi;   // The phase angle
};
typedef struct DihedralDef dihedef;

/***=======================================================================***/
/*** CONSTRAINTCOMMAND: command to implement constraints, whether they be  ***/
/***                    to use SETTLE for a rigid water molecule, to use   ***/
/***                    RATTLE to constrain a bond, or to use P-RATTLE to  ***/
/***                    constrain a complex rigid or semi-rigid structure. ***/
/***=======================================================================***/
struct ConstraintCommand {
  int exe;     // Flag to execute SETTLE or other constraints with this atom 
               //   as the base
  int* blist;  // Atom identification number of the other atoms in the group
};
typedef struct ConstraintCommand cnstcomm;

/***=======================================================================***/
/*** SETTLEPARAMETERS: three parameters that can be conveniently pre-      ***/
/***                   computed for SETTLE rigid water constraints.        ***/
/***=======================================================================***/
struct SettleParameters {
  double ra;
  double rb;
  double rc;
  double mO;
  double mH;
};
typedef struct SettleParameters settleparm;

/***=======================================================================***/
/*** EXTRAPOINT: this structure identifies a massless, "extra point"       ***/
/***             particle to distinguish it from the rest of the system.   ***/
/***             Frame atoms, frame style, and other features that show    ***/
/***             what needs to be done.                                    ***/
/***=======================================================================***/
struct ExtraPoint {
  int atomid;          // The atom number in the regular topology
  int frstyle;         // The frame style:
                       //   0: extra point on TIP4P-style water
                       //   1: extra point on a line between two frame atoms
                       //  (Other frame types have yet to be included)
  int nfratm;          // The number of atoms in the frame
  int fr1;             // The first frame atom
  int fr2;             // The second frame atom
  int fr3;             // The third frame atom
  double d1;           // Distance from frame atom 1
  double d2;           // Distance from frame atom 2
  double d3;           // Distance from frame atom 3
};
typedef struct ExtraPoint expt;

/***=======================================================================***/
/*** EXTRAPOINTRULE: stores a rule for dealing with certain classes of     ***/
/***                 extra points; rules specified by the user override    ***/
/***                 built-in rules for objects such as four-point waters. ***/
/***=======================================================================***/
struct ExtraPointRule {
  int nfratm;          // The number of frame atoms
  char epname[5];      // The name of the atom that is an extra point
  char fr1[5];         // The name of the first frame atom
  char fr2[5];         // The name of the second frame atom
  char fr3[5];         // The name of the (optional) third frame atom
  char resname[5];     // The residue name to which this rule applies
  int frstyle;         // The frame style (choose from the built-in options)
  double d1;           // The distance from the extra point to frame atom 1
  double d2;           // The distance from the extra point to frame atom 2
  double d3;           // The distance from the extra point to frame atom 3
};
typedef struct ExtraPointRule eprule;

/***=======================================================================***/
/*** ListOfGroup: a list of grouped atoms, such as atoms linked by bonds.  ***/
/***=======================================================================***/
struct ListOfGroup {
  int natom;        // The number of atoms in this group
  int* atoms;       // The topology numbers of the atoms in this group
};
typedef struct ListOfGroup lgrp;

/***=======================================================================***/
/*** AMBPRMTOP: a structure for AMBER 7 topology file data.                ***/
/***=======================================================================***/
struct ambprmtop {

  /*** Integers ***/
  int natom;    // The number of atoms
  int nres;     // The number of residues
  int ntypes;   // Total number of distinct atom types
  int nhparm;   // Currently not used
  int nparm;    // Currently not used
  int tnexcl;   // The number of excluded atoms
  int natyp;    // The number of atom types in the parameter file (see Solty)
  int nphb;     // The number of distinct 10-12 interaction types
  int ifpert;   // Set to 1 if perturbation information is to be read
  int ifbox;    // Set to 1 if standard periodic box, 2 if truncated octahedron
  int nmxrs;    // Number of atoms in the largest residue
  int ifcap;    // Set to 1 if the CAP option from edit was specified
  int iptres;   // The final residues that is considered part of the solute
  int iptatom;  // The final atom that is considered part of the solute
  int nspm;     // The total number of molecules
  int nspsol;   // The first solvent molecule
  int natcap;   // The last atom before the start of cap waters placed by edit
  int blank;    // Not sure what this is
  int rattle;   // Flag to activate RATTLE
  int settle;   // Flag to activate SETTLE
  int nwat;     // The number of rigid three point water molecules
  int ncnst;    // The number of constraints in the system
  int ndf;      // The number of degrees of freedom in the system
  int nprtcl;   // The number of particles in the system (each constrained
                //   group of atoms counts as one particle, and each atom not
                //   in a constrained group, counts as one particle)
  int ljbuck;   // Flag to signal whether a Lennard-Jones 12-6 or Buckingham
                //   exp-6 potential is to be used
  int qshape;   // Flag to signal whether point charges or spherical Gaussian-
                //   distributed charges are to be used
  int neprule;  // The number of extra point rules, obtained from an auxiliary
                //   file (the name of the file is stored in the eprulesource
                //   field)
  int nxtrapt;  // The number of extra points
  int ngrp;     // The number of bonded atom groups

  int ReadVirtualQ; // Directive to read virtual charge sites from the topology
                    //   file.

  /*** Reals ***/
  double cutcap;    // The distance from the center of the cap to the outside
  double xcap;      // X coordinate for the center of the cap
  double ycap;      // Y coordinate for the center of the cap
  double zcap;      // Z coordinate for the center of the cap
  double lj14fac;   // The van-der Waals 1:4 interaction adjustment factor
  double elec14fac; // The electrostatic 1:4 interaction adjustment factor
  double rattletol; // The RATTLE tolerance
  double TotalMass; // The total mass of the system (atomic mass units) 
  double initq;     // The initial total charge on the system (for output
                    //   record keeping--system charges are typically rounded
                    //   to the nearest integer, often zero, by adding a small
                    //   amount of charge to every atom)

  /*** In what follows, (b, a, h) means "(bonds, angles, and dihedrals)" ***/
  bah withH;    // Number of (b, a, h) with hydrogen
  bah woH;      // Number of (b, a, h) without hydrogen
  bah woHC;     // Number of (b, a, h) without hydrogen plus constraints
  bah nBAH;     // Number of unique (b, a, h) types
  bah pert;     // Number of (b, a, h) to be perturbed
  bah wpert;    // Number of (b, a, h) with all atoms in the perturbed group

  /*** Arrays of elemental types ***/
  int* LJIdx;        // The atom type index for each atom
  int* IsExpt;       // Array of flags, one per atom, to identify whether each
                     //   atom is really an extra point
  int* NExcl;        // The number of excluded atoms for each atom
  int* ConExcl;      // Start positions in the excluded atom list
  int* ExclList;     // The excluded atom list
  int* NBParmIdx;    // Nonbonded parameter index, ntypes*ntypes elements
  int* ResLims;      // The residue limits
  int* Join;         // Tree joining information, used by ancient programs
  int* Rotat;        // No longer in use
  int* Nsp;          // The total number of atoms in each molecule
  double* Charges;   // The atomic charges
  double* Masses;    // The atomic masses (atomic mass units)
  double* InvMasses; // The square root inverse atomic masses
  double* BondK;     // The bond force constants (kcal/mol-A^2)
  double* BondEq;    // The equilibrium bond lengths (Angstroms)
  double* AnglK;     // The angle force constants (kcal/mol-rad^2)
  double* AnglEq;    // The equilibrium angle values (radians)
  double* DiheK;     // The dihedral force constants (kcal/mol)
  double* DiheN;     // The dihedral periodicities
  double* DihePhi;   // The dihedral phase
  double* LJA;       // The Lennard-Jones A parameters
  double* LJB;       // The Lennard-Jones B parameters
  double* LJC;       // The "Lennard-Jones" C parameters (if these are given,
                     //   it means that the van-der Waals model is a modified
                     //   Buckingham potential, and thus the Lennard-Jones A
                     //   and B parameters are really Buckingham A and B
                     //   parameters)
  double* SolA;      // The HBond A parameters
  double* SolB;      // The HBond B parameters
  double* HBCut;     // No longer in use
  double* Radii;     // Atomic radii
  double* Screen;    // Atomic screening terms (?)
  char RadSet[128];  // Radius set name
  char vstamp[128];  // Unknown         
  char WaterName[8]; // The name of water molecules (for implementing SETTLE)
  char* AtomNames;   // The atom names
  char* ResNames;    // The residue names
  char* AtomTypes;   // The atom type names
  char* TreeSymbols; // The atom tree chain symbols
  double gdim[6];    // The box dimensions

  /*** Arrays of structures ***/
  bond* BIncH;       // Bonds including Hydrogen
  bond* BNoH;        // Bonds not including Hydrogen
  angle* AIncH;      // Angles including Hydrogen
  angle* ANoH;       // Angles not including Hydrogen
  dihedral* HIncH;   // Dihedrals including Hydrogen
  dihedral* HNoH;    // Dihedrals not including Hydrogen
  bondlist* BLC;     // The list of bonds controlled by each atom
  angllist* ALC;     // The list of angles controlled by each atom
  dihelist* HLC;     // The list of dihedral angles controlled by each atom
  bonddef* BParam;   // The re-organized, unified bond parameter array
  angldef* AParam;   // The re-organized, unified angle parameter array
  dihedef* HParam;   // The re-organized, unified dihedral parameter array
  cnstcomm* SHL;     // Constraint commands for each individual atom (SETTLE is
                     //   performed starting with the oxygens)
  eprule* eprules;   // List of rules for extra points
  expt* xtrapts;     // List of extra points (field nxtrapts holds the size)
  lgrp* lgrps;       // List of atom groups in this topology, defining
                     //   separate molecules within the simulation

  /*** Lennard-Jones pre-computed parameter tables ***/
  dmat LJftab;
  dmat LJutab;

  /*** SETTLE pre-computed parameters ***/
  settleparm FWtab;

  /*** Source information ***/
  char source[MAXNAME];
  char eprulesource[MAXNAME];
};
typedef struct ambprmtop prmtop;

#endif
