#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "Parse.h"
#include "mdgxVector.h"
#include "Constraints.h"

#include "CrdManipDS.h"
#include "TopologyDS.h"

/***=======================================================================***/
/*** ReadEPRuleFile: read a file containing user-specified extra point     ***/
/***                 rules.  Rules specified in this file will override    ***/
/***                 built-in rules within the program, such as four-point ***/
/***                 water rules.                                          ***/
/***=======================================================================***/
void ReadEPRuleFile(prmtop *tp)
{
  int i, neprule, maxrule, collect;
  char line[MAXLINE], tmpfr1[MAXNAME], tmpfr2[MAXNAME], tmpfr3[MAXNAME];
  char tmpepn[MAXNAME];
  FILE *inp;
  cmat lwords;
  eprule *tmr;

  /*** If there's no file specified, we're done. ***/
  if (tp->eprulesource[0] == '\0') {
    tp->neprule = 0;
    return;
  }

  maxrule = 10;
  tp->eprules = (eprule*)malloc(maxrule*sizeof(eprule));
  if ((inp = fopen(tp->eprulesource, "r")) == NULL) {
    printf("ReadEPRuleFile >> Error.  Extra points rule file %s not found.\n",
           tp->eprulesource);
    exit(1);
  }
  neprule = 0;
  while (AdvanceToSegment(inp, "rule") != 0) {
    collect = 1;

    /*** Initialize this rule ***/
    tmr = &tp->eprules[neprule];
    tmr->nfratm = 0;
    tmpfr1[0] = '\0';
    tmpfr2[0] = '\0';
    tmpfr3[0] = '\0';
    tmpepn[0] = '\0';
    tmr->fr1[0] = '\0';
    tmr->fr2[0] = '\0';
    tmr->fr3[0] = '\0';
    tmr->frstyle = -1;
    for (i = 0; i < 5; i++) {
      tmr->resname[i] = '\0';
    }
    while (collect == 1) {

      /*** Information about each rule is collected in much the same way   ***/
      /*** as information for any of the namelists in the command section. ***/
      fgets(line, MAXLINE, inp);
      RemoveWhiteSpace(line, MAXLINE);
      RemoveComments(line);
      collect = DetectNamelistEnd(line, "GetCntrlNamelist");
      if (collect == 0) {
        continue;
      }
      NixCommaCarriage(line);
      EqualSpace(line);
      lwords = ParseWords(line);
      SeekInt(lwords, &tmr->nfratm, "AtomCount", "natom");
      SeekString(lwords, tmpepn, "ExtraPoint", "epname");
      SeekString(lwords, tmpfr1, "FrameAtom1", "frame1");
      SeekString(lwords, tmpfr2, "FrameAtom2", "frame2");
      SeekString(lwords, tmpfr3, "FrameAtom3", "frame3");
      SeekReal(lwords, &tmr->d1, "Range1", "range1");
      SeekReal(lwords, &tmr->d2, "Range2", "range2");
      SeekReal(lwords, &tmr->d3, "Range3", "range3");
      SeekInt(lwords, &tmr->frstyle, "FrameStyle", "style");
      SeekString(lwords, tmr->resname, "ResidueName", "residue");
    }

    /*** Check this rule ***/
    if (strlen(tmpfr1) > 4 || strlen(tmpfr2) > 4 || strlen(tmpfr3) > 4) {
      printf("ReadEPRule >> Error.  Atom names [ %s %s %s ] invalid (4 "
             "characters max\nReadEPRule >> per name).\n", tmpfr1, tmpfr2,
             tmpfr3);
      exit(1);
    }
    if (tmr->nfratm < 2 || tmr->nfratm > 3) {
      printf("ReadEPRuleFile >> Error.  This frame rule has %d atoms.  "
             "Frames must have\nReadEPRuleFile >> two or three.\n",
             tmr->nfratm);
      exit(1);
    }
    if (tmr->fr1[0] == '\0' || tmr->fr2[0] == '\0' ||
        (tmr->nfratm == 3 && tmr->fr3[0] == '\0')) {
      printf("ReadEPRuleFile >> Error.  One or more frame atoms is not "
             "properly specified.\n");
      exit(1);
    }
    if (strncmp(tmr->fr1, tmr->fr2, 4) == 0 ||
        strncmp(tmr->fr1, tmr->fr3, 4) == 0 ||
        strncmp(tmr->fr2, tmr->fr3, 4)) {
      printf("ReadEPRuleFile >> Error.  One or more frame atoms is not "
             "properly specified.\n");
      exit(1);
    }
    if (tmr->frstyle < 0) {
      printf("ReadEPRuleFile >> Error.  Frame style %d is not allowed.\n",
             tmr->frstyle);
      exit(1);
    }
    if (tmr->resname[0] == '\0') {
      printf("ReadEPRuleFile >> Error.  Residue name unspecified.\n");
      exit(1);
    }
    for (i = 1; i < 4; i++) {
      if (tmr->resname[i] == '\0') {
        tmr->resname[i] = ' ';
      }
    }
    neprule++;
    if (neprule == maxrule) {
      maxrule += 10;
      tp->eprules = (eprule*)realloc(tp->eprules, maxrule*sizeof(eprule));
    }
  }
  tp->eprules = (eprule*)realloc(tp->eprules, neprule*sizeof(eprule));
}

/***=======================================================================***/
/*** CountBonds: count all bonds extending to and from atom ai according   ***/
/***             to topology tp.                                           ***/
/***=======================================================================***/
static int CountBonds(prmtop *tp, int ai)
{
  int i, j, nbnd;

  nbnd = 0;
  for (i = 0; i < tp->natom; i++) {
    if (i == ai) {
      nbnd += tp->BLC[i].nbond;
    }
    else {
      for (j = 0; j < tp->BLC[i].nbond; j++) {
	if (tp->BLC[i].BC[j].b == ai) {
	  nbnd++;
	}
      }
    }
  }

  return nbnd;
}

/***=======================================================================***/
/*** EPRuleDefined: this routine searches through a list of rules in the   ***/
/***                topology tp to see if any rules matching the residue   ***/
/***                name can be found.  Although this function will return ***/
/***                the number of a matching rule it one is found, there   ***/
/***                may be more matching rules, so the list is re-checked  ***/
/***                in the ImplementEPRule routine.                        ***/
/***=======================================================================***/
static int EPRuleDefined(prmtop *tp, int ri)
{
  int i, epid;

  epid = -1;
  for (i = 0; i < tp->neprule; i++) {
    if (strncmp(tp->eprules[i].resname, &tp->ResNames[4*ri], 4) == 0) {
      epid = i;
      break;
    }
  }

  return epid;
}

/***=======================================================================***/
/*** ImplementEPRule: implement a user-defined extra point rule.           ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:     the topology structure                                      ***/
/***   resid:  the residue number                                          ***/
/***=======================================================================***/
static void ImplementEPRule(prmtop *tp, int resid)
{
  int i, j, llim, hlim;

  llim = tp->ResLims[resid];
  hlim = tp->ResLims[resid+1];

  /*** Loop through all rules and implement those that fit ***/
  for (i = 0; i < tp->neprule; i++) {
    if (strncmp(tp->eprules[i].resname, &tp->ResNames[4*resid], 4) == 0) {

      /*** Transfer information from the rule to the particle ***/
      tp->xtrapts[tp->nxtrapt].frstyle = tp->eprules[i].frstyle;
      tp->xtrapts[tp->nxtrapt].nfratm = tp->eprules[i].nfratm;

      /*** Find the extra point and frame atoms ***/
      for (j = llim; j < hlim; j++) {
	if (strncmp(tp->eprules[i].epname, &tp->AtomNames[4*j], 4) == 0) {
	  tp->xtrapts[tp->nxtrapt].atomid = j;
	}
	if (strncmp(tp->eprules[i].fr1, &tp->AtomNames[4*j], 4) == 0) {
	  tp->xtrapts[tp->nxtrapt].fr1 = j;
	}
	if (strncmp(tp->eprules[i].fr2, &tp->AtomNames[4*j], 4) == 0) {
	  tp->xtrapts[tp->nxtrapt].fr2 = j;
	}
	if (strncmp(tp->eprules[i].fr3, &tp->AtomNames[4*j], 4) == 0) {
	  tp->xtrapts[tp->nxtrapt].fr3 = j;
	}
      }

      /*** Increment the number of extra points ***/
      tp->nxtrapt += 1;
    }
  }
}

/***=======================================================================***/
/*** Implement4PWaterRule: implement a rule for the family of four-point   ***/
/***                       rigid waters.                                   ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:     the topology structure                                      ***/
/***   resid:  the residue number                                          ***/
/***=======================================================================***/
static void Implement4PWaterRule(prmtop *tp, int resid)
{
  int i, llim, hlim, bondfound, epatomnum;
  double maxmass;

  /*** The frame style is "0" ***/
  tp->xtrapts[tp->nxtrapt].frstyle = 0;
  tp->xtrapts[tp->nxtrapt].nfratm = 3;
  tp->xtrapts[tp->nxtrapt].fr2 = -1;
  tp->xtrapts[tp->nxtrapt].fr3 = -1;

  /*** Find the extra point and frame atoms by mass ***/
  llim = tp->ResLims[resid];
  hlim = tp->ResLims[resid+1];
  maxmass = 0.0;
  for (i = llim; i < hlim; i++) {
    if (tp->Masses[i] < 1.0e-8) {
      tp->xtrapts[tp->nxtrapt].atomid = i;
    }
    if (tp->Masses[i] > maxmass) {
      maxmass = tp->Masses[i];
      tp->xtrapts[tp->nxtrapt].fr1 = i;
    }
  }
  for (i = llim; i < hlim; i++) {
    if (i != tp->xtrapts[tp->nxtrapt].fr1 &&
	i != tp->xtrapts[tp->nxtrapt].atomid) {
      if (tp->xtrapts[tp->nxtrapt].fr2 == -1) {
	tp->xtrapts[tp->nxtrapt].fr2 = i;
      }
      else {
	tp->xtrapts[tp->nxtrapt].fr3 = i;
      }
    }
  }

  /*** Determine the distance between the extra point and atom 1 ***/
  bondfound = 0;
  epatomnum = tp->xtrapts[tp->nxtrapt].atomid;
  if (tp->BLC[epatomnum].nbond == 1) {
    if (tp->BLC[epatomnum].BC[0].b != 
	tp->xtrapts[tp->nxtrapt].fr1) {
      printf("ImplementrPWaterRule >> Error.  There is a bond controlled by "
	     "the extra\nImplement4PWaterRule >> point, but it does not "
	     "connect to the oxygen atom.\n");
      exit(1);
    }
    tp->xtrapts[tp->nxtrapt].d1 = tp->BParam[tp->BLC[epatomnum].BC[0].t].l0;
    bondfound = 1;
  }
  if (bondfound == 0) {
    epatomnum = tp->xtrapts[tp->nxtrapt].fr1;
    for (i = 0; i < tp->BLC[epatomnum].nbond; i++) {
      if (tp->BLC[epatomnum].BC[i].b == tp->xtrapts[tp->nxtrapt].atomid) {
	tp->xtrapts[tp->nxtrapt].d1 =
	  tp->BParam[tp->BLC[epatomnum].BC[i].t].l0;
	bondfound = 1;
      }
    }
  }
  if (bondfound == 0) {
    printf("ImplementrPWaterRule >> Error.  No bond could be located between "
	   "the extra\nImplementrPWaterRule >> point and the oxygen atom.\n");
    exit(1);
  }

  /*** Increment the extra point counter ***/
  tp->nxtrapt += 1;
}

/***=======================================================================***/
/*** DetermineEPFrames: determine the frame atoms, and frame type, for all ***/
/***                    massless atoms in the system.  The basic rules are ***/
/***                    that a frame must be found for every extra point,  ***/
/***                    that the frame atoms must lie within the same      ***/
/***                    residue.  While extra points are initialized on a  ***/
/***                    residue-by-residue basis, the frames are stored on ***/
/***                    an atom-by-atom basis so that calculations do not  ***/
/***                    need to reference whole residues.                  ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:     the topology structure                                      ***/
/***=======================================================================***/
void DetermineEPFrames(prmtop *tp)
{
  int i, j, epid, nep, nresatm;
  int* eplist;

  tp->nxtrapt = 0;
  tp->xtrapts = (expt*)malloc(tp->natom*sizeof(expt));
  for (i = 0; i < tp->nres; i++) {

    /*** Count any extra points that this residue may have ***/
    nep = 0;
    nresatm = tp->ResLims[i+1] - tp->ResLims[i];
    eplist = (int*)calloc(nresatm, sizeof(int));
    for (j = tp->ResLims[i]; j < tp->ResLims[i+1]; j++) {
      if (tp->Masses[j] < 1.0e-8) {
	eplist[nep] = j;
	nep++;
      }
    }
    if (nep == 0) {
      free(eplist);
      continue;
    }

    /*** This residue has some extra points.  Figuring out what ***/
    /*** to do is a tedious process; not much elegance to it.   ***/
    for (j = 0; j < nep; j++) {

      /*** To begin, we check to see if there is a user-defined rule ***/
      /*** for dealing with this extra point.  If there is no such   ***/
      /*** rule, then the program's built-in rules take effect.      ***/
      epid = EPRuleDefined(tp, i);
      if (epid > -1) {
	ImplementEPRule(tp, i);
      }

      /*** If there is only a bond between the extra point and the  ***/
      /*** oxygen atom, then we will assume that the extra point    ***/
      /*** lies along the bisector between the O-H1 and O-H2 bonds. ***/
      else if (nresatm == 4 && nep == 1 && CountBonds(tp, eplist[0]) == 1 &&
	       strncmp(&tp->ResNames[4*i], tp->WaterName, 4) == 0) {
	Implement4PWaterRule(tp, i);
      }

      /*** Otherwise, there must be three bonds to the extra point ***/
      /*** or it's unclear what to do.                             ***/
      else if (CountBonds(tp, eplist[0]) == 3) {
	printf("Error.  Not yet ready to do triple-bonded extra points.\n");
	exit(1);
      }
      else {
	printf("DetermineEPFrames >> Error.  There are %d bonds to the extra "
	       "point.  No rules\nDetermineEPFrames >> are defined for this "
	       "residue.  It is not clear what to\nDetermineEPFrames >> do in "
	       "this case.\n", CountBonds(tp, eplist[0]));
	exit(1);
      }
    }

    /*** Free allocated memory ***/
    free(eplist);
  }

  /*** Trim the extra points array to fit the actual size needed ***/
  tp->xtrapts = (expt*)realloc(tp->xtrapts, tp->nxtrapt*sizeof(expt));
}

/***=======================================================================***/
/*** ReimageEPGroup: reimage an extra point and its frame to ensure that   ***/
/***                 calculations can be performed without consideration   ***/
/***                 to periodicity.                                       ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   eploc:      the location of the extra point                         ***/
/***   [a,b,c]loc: the locations of frame atoms A, B, and C                ***/
/***   tmr:        the extra point rule struct                             ***/
/***   crd:        the system coordinates                                  ***/
/***=======================================================================***/
static void ReimageEPGroup(double *eploc, double *aloc, double *bloc,
			   double *cloc, expt *tmr, coord *crd)
{
  int i;
  double xvec[12];
  
  for (i = 0; i < 3; i++) {
    xvec[i] = eploc[i];
    xvec[i+3] = aloc[i];
    xvec[i+6] = bloc[i];
    if (tmr->frstyle != 1) {
      xvec[i+9] = cloc[i];
    }
  }
  if (tmr->frstyle != 1) {
    ReimageConstraintGroup(xvec, crd, 4);
  }
  else {
    ReimageConstraintGroup(xvec, crd, 3);
  }
  for (i = 0; i < 3; i++) {
    eploc[i] = xvec[i];
    aloc[i] = xvec[i+3];
    bloc[i] = xvec[i+6];
    if (tmr->frstyle != 1) {
      cloc[i] = xvec[i+9];
    }
  }
}

/***=======================================================================***/
/*** PlaceExtraPoints: places extra points based on a class index for each ***/
/***                   point and the positions of its frame atoms.         ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:     the topology structure                                      ***/
/***   crd:    the system coordinates                                      ***/
/***=======================================================================***/
void PlaceExtraPoint(prmtop *tp, coord *crd)
{
  int i, j;
  double mae;
  double ae[3];
  double *aloc, *bloc, *cloc, *eploc;
  expt *tmr;

  /*** Loop over all extra points ***/
  for (i = 0; i < tp->nxtrapt; i++) {

    /*** Pointer to extra point structure ***/
    tmr = &tp->xtrapts[i];

    /*** Pointers to atom locations ***/
    eploc = &crd->loc[3*tmr->atomid];
    aloc = &crd->loc[3*tmr->fr1];
    bloc = &crd->loc[3*tmr->fr2];
    if (tmr->frstyle == 0) {
      cloc = &crd->loc[3*tmr->fr3];
    }

    /*** Re-image the extra point and frame atoms ***/
    ReimageEPGroup(eploc, aloc, bloc, cloc, tmr, crd);

    /*** Place extra point ***/
    if (tmr->frstyle == 0) {
      mae = 0.0;
      for (j = 0; j < 3; j++) {
	ae[j] = 0.5*(bloc[j] + cloc[j] - 2.0*aloc[j]);
	mae += ae[j]*ae[j];
      }
      mae = tmr->d1/sqrt(mae);
      for (j = 0; j < 3; j++) {
	eploc[j] = aloc[j] + ae[j]*mae;
      }
    }
    else {
      printf("PlaceExtraPoints >> Error.  Not prepared to handle frame type "
	     "%d.\n", tmr->frstyle);
      exit(1);
    }
  }
}

/***=======================================================================***/
/*** Torque: compute the torque for a residue.                             ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:     the topology structure                                      ***/
/***   crd:    the system coordinates                                      ***/
/***   resid:  the residue number                                          ***/
/***=======================================================================***/
void Torque(prmtop *tp, coord *crd, int resid)
{
  int i, j;
  double com[3], r[3], rxf[3], trq[3];

  /*** Compute center of mass ***/
  com[0] = com[1] = com[2] = 0.0;
  for (i = tp->ResLims[resid]; i < tp->ResLims[resid+1]; i++) {
    for (j = 0; j < 3; j++) {
      com[j] += tp->Masses[i]*crd->loc[3*i+j];
    }
  }

  /*** Compute torque ***/
  trq[2] = trq[1] = trq[0] = 0.0;
  for (i = tp->ResLims[resid]; i < tp->ResLims[resid+1]; i++) {
    for (j = 0; j < 3; j++) {
      r[j] = crd->loc[3*i+j] - com[j];
    }
    CrossP(r, &crd->frc[3*i], rxf);
    for (j = 0; j < 3; j++) {
      trq[j] += rxf[j];
    }
  }

  printf("Torque = [ %12.8lf %12.8lf %12.8lf ]\n", trq[0], trq[1], trq[2]);
}

/***=======================================================================***/
/*** TransferEPForces: transfer forces from the extra points to the frame  ***/
/***                   atoms according to the frame type.                  ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tp:     the topology structure                                      ***/
/***   crd:    the system coordinates                                      ***/
/***=======================================================================***/
void TransferEPForces(prmtop *tp, coord *crd)
{
  int i, j;
  double abcoef, mab, mac, angcab;
  double ab[3], ac[3];
  double *aloc, *bloc, *cloc, *eploc, *afrc, *bfrc, *cfrc, *epfrc;
  expt *tmr;

  for (i = 0; i < tp->nxtrapt; i++) {
    tmr = &tp->xtrapts[i];

    /*** Set pointers ***/
    eploc = &crd->loc[3*tmr->atomid];
    aloc = &crd->loc[3*tmr->fr1];
    bloc = &crd->loc[3*tmr->fr2];
    if (tmr->frstyle != 1) {
      cloc = &crd->loc[3*tmr->fr3];
    }

    /*** Re-image the extra point and frame atoms ***/
    ReimageEPGroup(eploc, aloc, bloc, cloc, tmr, crd);

    /*** Frame for four-point waters ***/
    if (tmr->frstyle == 0) {

      /*** Determine the coefficient based on distances between  ***/
      /*** the extra point and the oxygen and between the oxygen ***/
      /*** and each hydrogen.                                    ***/ 
      mab = 0.0;
      mac = 0.0;
      for (j = 0; j < 3; j++) {
	ab[j] = bloc[j] - aloc[j];
	ac[j] = cloc[j] - aloc[j];
	mab += ab[j]*ab[j];
	mac += ac[j]*ac[j];
      }
      mab = sqrt(mab);
      mac = sqrt(mac);

      /*** Check ***/
      if (fabs(mab - mac) > 1.0e-8) {
	printf("TransferEPForces >> Error.  The distances between the oxygen "
	       "and each hydrogen\nTransferEPForces >> atom differ by %9.6lf."
	       "\n", fabs(mab - mac));
	exit(1);
      }
      angcab = 0.5*acos(DotP(ab, ac, 3)/(mab*mac));
      abcoef = tmr->d1/(2.0*mab*cos(angcab));

      /*** Transfer forces ***/
      afrc = &crd->frc[3*tmr->fr1];
      bfrc = &crd->frc[3*tmr->fr2];
      cfrc = &crd->frc[3*tmr->fr3];
      epfrc = &crd->frc[3*tmr->atomid];
      for (j = 0; j < 3; j++) {
	afrc[j] += (1.0-2.0*abcoef)*epfrc[j];
	bfrc[j] += abcoef*epfrc[j];
	cfrc[j] += abcoef*epfrc[j];
	epfrc[j] = 0.0;
      }
    }
  }
}
