#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "CompFrc.h"
#include "Random.h"
#include "pmeRecip.h"
#include "pmeDirect.h"
#include "BSpline.h"
#include "Topology.h"
#include "ChargeMap.h"
#include "Constants.h"
#include "mdgxVector.h"
#include "CrdManip.h"
#include "Trajectory.h"
#include "Grid.h"
#include "Matrix.h"
#include "CellManip.h"
#include "SpecialMath.h"
#include "Nonbonded.h"
#include "Bonded.h"
#include "Command.h"
#include "Constraints.h"
#include "Integrator.h"
#include "Thermostats.h"
#include "Barostats.h"
#include "Manual.h"
#include "mleRecip.h"
#include "Timings.h"

/***=======================================================================***/
/*** CheckArgumentOverflow: this function is called before accessing the   ***/
/***                        next element in the argv[] list.  It checks    ***/
/***                        to make sure that such an argument exists; if  ***/
/***                        not, the program exits with an error message.  ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   i:     the number of the current element in argv[] minus 1          ***/
/***   argc:  the total number of elements in the argv[] list              ***/
/***   ev:    the current element in the argv[] list                       ***/
/***=======================================================================***/
void CheckArgumentOverflow(int i, int argc, char* ev)
{
  if (i < argc-2) {
    return;
  }
  else {
    printf("CheckArgumentOverflow >> Error.  No value specified for %s.\n",
	   ev);
    exit(1);
  }
}

/***=======================================================================***/
/*** CommandLineControl: command-line control functions for mdgx.  This    ***/
/***                     function interprets data on the command line and  ***/
/***                     then prints out help information or begins        ***/
/***                     dynamics.                                         ***/
/***=======================================================================***/
void CommandLineControl(int argc, char *argv[], trajcon *tj, dircon *dcinp,
			reccon *rcinp, prmtop *tp)
{
  int i;
  char tag[64];

  /*** Defaults for critical I/O files ***/
  sprintf(tj->ipcname, "inpcrd");
  sprintf(tj->dumpname, "forcedump.dat");
  sprintf(tj->rstbase, "restrt");
  tj->rstsuff[0] = '\0';
  sprintf(tj->trjbase, "mdcrd");
  tj->trjsuff[0] = '\0';
  sprintf(tj->outbase, "mdout");
  tj->outsuff[0] = '\0';
  sprintf(tj->enebase, "mden");
  tj->enesuff[0] = '\0';
  sprintf(tj->velbase, "mdvel");
  tj->velsuff[0] = '\0';
  sprintf(tj->frcbase, "mdfrc");
  tj->frcsuff[0] = '\0';
  sprintf(tp->source, "prmtop");
  tp->eprulesource[0] = '\0';
  tj->OverwriteOutput = 0;

  /*** If no arguments, print the manual front page ***/
  if (argc == 1) {
    PrintSplash(stdout);
    PrintUsage();
    exit(1);
  }

  /*** Print additional documentation ***/
  else if (argc == 2) {
    strcpy(tag, *++argv);
    if (strcmp(tag, "-INPUT") == 0) {
      PrintCommandLineInputOptions();
    }
    else if (strcmp(tag, "-IFILE") == 0) {
      PrintInputFormat();
    }
    else if (strcmp(tag, "-FILES") == 0) {
      PrintFilesNamelistVariables();
    }
    else if (strcmp(tag, "-CNTRL") == 0) {
      PrintCntrlNamelistVariables();
    }
    else if (strcmp(tag, "-EWALD") == 0) {
      PrintEwaldNamelistVariables();
    }
    else if (strcmp(tag, "-FORCE") == 0) {
      PrintForceNamelistVariables();
    }
    else if (strcmp(tag, "-ATTR") == 0) {
      PrintAttributions();
    }
    else {
      printf("CommandLineControl >> Error.  Unrecognized argument %s.\n", tag);
    }
    exit(1);
  }

  /*** Read command line input and execute dynamics ***/
  for (i = 0; i < argc-1; i += 2) {
    strcpy(tag, *++argv);
    if (strcmp(tag, "-i") == 0) {
      CheckArgumentOverflow(i, argc, "-i");
      strcpy(tj->inpname, *++argv);
    }
    else if (strcmp(tag, "-p") == 0) {
      CheckArgumentOverflow(i, argc, "-p");
      strcpy(tp->source, *++argv);
    }
    else if (strcmp(tag, "-c") == 0) {
      CheckArgumentOverflow(i, argc, "-c");
      strcpy(tj->ipcname, *++argv);
    }
    else if (strcmp(tag, "-d") == 0) {
      CheckArgumentOverflow(i, argc, "-d");
      strcpy(tj->dumpname, *++argv);
    }
    else if (strcmp(tag, "-o") == 0) {
      CheckArgumentOverflow(i, argc, "-o");
      strcpy(tj->outbase, *++argv);
    }
    else if (strcmp(tag, "-e") == 0) {
      CheckArgumentOverflow(i, argc, "-e");
      strcpy(tj->enebase, *++argv);
    }
    else if (strcmp(tag, "-x") == 0) {
      CheckArgumentOverflow(i, argc, "-x");
      strcpy(tj->trjbase, *++argv);
    }
    else if (strcmp(tag, "-v") == 0) {
      CheckArgumentOverflow(i, argc, "-v");
      strcpy(tj->velbase, *++argv);
    }
    else if (strcmp(tag, "-f") == 0) {
      CheckArgumentOverflow(i, argc, "-f");
      strcpy(tj->frcbase, *++argv);
    }
    else if (strcmp(tag, "-r") == 0) {
      CheckArgumentOverflow(i, argc, "-r");
      strcpy(tj->rstbase, *++argv);
    }
    else if (strcmp(tag, "-O") == 0) {
      tj->OverwriteOutput = 1;
      i--;
    }
    else {
      printf("CommandLineControl >> Error.  Unrecognized tag %s.\n", tag);
      exit(1);
    }
  }
}

/***=======================================================================***/
/*** TheGreatCaseInhibitor: this routine serves as a prphylactic against   ***/
/***                        input combinations that mdgx is not yet ready  ***/
/***                        to handle, even if certain input directives in ***/
/***                        isolation are accetable.  The routine simply   ***/
/***                        checks for a list of possible combinations, a  ***/
/***                        list which will hopefully shrink over time,    ***/
/***                        but also which may not be complete.            ***/
/***                                                                       ***/
/*** Arguments:                                                            ***/
/***   tj:    the trajectory control information                           ***/
/***   tp:    the topology                                                 ***/
/***   rcinp: the reciprocal space control information                     ***/
/***   dcinp: the direct space control information                         ***/
/***=======================================================================***/
void TheGreatCaseInhibitor(trajcon *tj, prmtop *tp, dircon *dcinp,
			   reccon *rcinp)
{
  /*** Case: Energy minimization ***/
  if (tj->mode == 1) {
    printf("TheGreatCaseInhibitor >> Energy minimization is not yet "
	   "implemented.\n");
    exit(1);
  }

  /*** Case: MLE and NPT, the MLE in mdgx is not yet ready ***/
  /*** to work with anything but a Monte-Carlo barostat    ***/
  if (rcinp->nlev > 1 && tj->ntp == 1) {
    printf("TheGreatCaseInhibitor >> Unable to perform MLE in combination "
	   "with NPT.\nTheGreatCaseInhibitor >> In principle, MLE is "
	   "compatible with constant\nTheGreatCaseInhibitor >> pressure "
	   "simulations, but not as yet in mdgx.\n");
    exit(1);
  }

  /*** Case: SETTLE / RATTLE and NPT ***/
  if ((tp->settle == 1 || tp->rattle == 1) && tj->ntp == 1) {
    printf("TheGreatCaseInhibitor >> Unable to perform bond constraints in "
	   "combination\nTheGreatCaseInhibitor >> with a Berendsen barostat "
	   "at this time.\n");
    exit(1);
  }

  /*** Case: Monte-Carlo barostat and no thermostat ***/
  if (tj->ntp == 5 && tj->ntt == 0) {
    printf("TheGreatCaseInhibitor >> Unable to perform isenthalpic ensemble "
	   "with\nTheGreatCaseInhibitor >> Monte-Carlo barostat.\n");
    exit(1);
  }
}

/***=======================================================================***/
/*** main                                                                  ***/
/***=======================================================================***/
int main(int argc, char *argv[])
{
  reccon rcinp;
  dircon dcinp;
  coord crd;
  cellgrid CG;
  bmap pmmap;
  bckit PPk;
  prmtop tp;
  trajcon tj;
  FrcTab Etab, EHtab;
  Energy sysUV;
  execon etimers;
  cdftrj Acdf[3];

  /*** Initialize timers ***/
  InitExecon(&etimers);
  mdgxStartTimer(&etimers);

  /*** Command line input ***/
  CommandLineControl(argc, argv, &tj, &dcinp, &rcinp, &tp);

  /*** Parse input command file ***/
  ReadCommFile(&dcinp, &rcinp, &tp, &tj, tj.inpname);

  /*** Read topology ***/
  GetPrmTop(&tp, 1);

  /*** Check for cases which mdgx is not yet prepared to handle ***/
  TheGreatCaseInhibitor(&tj, &tp, &dcinp, &rcinp);

  /*** Read starting coordinates ***/
  crd = InitCoords(&tp, &tj);
  InitializeEnergy(&sysUV, &tj, &tp, 1);

  /*** Compute force / energy lookup tables ***/
  Etab = DirectSpaceR2(dcinp.Ecut, dcinp.lkpspc, dcinp.ewcoeff, 0);
  EHtab = DirectSpaceR2(MINNB, 0.0625*dcinp.lkpspc, dcinp.ewcoeff, 0);

  /*** Preparation for reciprocal space calculation ***/
  if (rcinp.nlev == 1) {
    SetMeshDims(rcinp.ng, crd.gdim);
    rcinp.QL = (dbook*)malloc(sizeof(dbook));
    rcinp.QL[0] = CreateDbook(rcinp.ng[0], rcinp.ng[1], rcinp.ng[2], 1);
  }
  else {
    PrepMLE(&rcinp, &crd, &tp);
  }
  PPk = CreateBCKit(&rcinp, &rcinp.QL[0], &crd, &tp);
  pmmap = CreateBmap(&rcinp, crd.natom);

  /*** Preparation for direct space calculation ***/
  CG = CreateCellGrid(&crd, &dcinp);

  /*** Prepare thermostat ***/
  if (tj.ntt == 4) {
    PrepThermoBarostat(&tp, &tj);
  }

  /*** Prepare Monte-Carlo barostat ***/
  if (tj.ntp == 5) {
    PrepMCBarostat(&tj, &crd);
  }

  /*** Initialize velocities ***/
  if (tj.mode == 0) {
    InitVelocities(&crd, &CG, &tp, &dcinp, &Etab, &EHtab, &rcinp, &pmmap, &PPk,
		   &tj, &sysUV, &etimers, Acdf);
  }

  /*** Record initialization / startup time ***/
  etimers.Setup = mdgxStopTimer(&etimers);

  /*** Run dynamics or energy minimization ***/
  if (tj.mode == 0 || tj.mode == 1) {

    /*** Loop for dynamics ***/
    while (tj.currstep < tj.nstep) {
      if (tj.mode == 0) {
	Dynamics(&crd, &CG, &tp, &dcinp, &Etab, &EHtab, &rcinp, &pmmap, &PPk,
		 &tj, &sysUV, &etimers, Acdf);
      }
      else if (tj.mode == 1) {

	/*** Haven't yet written energy minimization scheme ***/
	printf("main >>> Error.  Energy minimization is not yet implemented."
	       "\n");
	exit(1);
      }
      UpdateStepNumber(&tj, tj.currstep + 1);
    }

    /*** Print the final restart file ***/
    WriteRst(&crd, &tp, &tj);
  }

  /*** Run a force calculation ***/
  else if (tj.mode == 2) {
    PrepForceReport(&crd, &CG, &tp, &dcinp, &Etab, &EHtab, &rcinp, &pmmap,
		    &PPk, &tj, &etimers);
  }

  /*** Free allocated memory ***/
  FreeTopology(&tp);
  DestroyCoord(&crd);
  FreeFrcTab(&Etab);
  FreeFrcTab(&EHtab);
  DestroyCellGrid(&CG);
  DestroyBCKit(&PPk);
  DestroyBmap(&pmmap);
  DestroyRecCon(&rcinp);
  DestroyTrajCon(&tj);
  DestroyEnergyTracker(&sysUV);

  return 0;
}
