########################################################################
#                                                                      #
# This is a module that contains functions responsible for parsing the #
# input file for MMPBSA.py. It must be included with MMPBSA.py to      #
# ensure proper functioning.                                           #
#                                                                      #
#          Last updated: 02/15/2011                                    #
#                                                                      #
########################################################################

########################## GPL LICENSE INFO ############################

#  Copyright (C) 2009  Dwight McGee, Billy Miller III, and Jason Swails

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.

import sys, os

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _compare(ref_str, teststr, tomatch):
   """ Tests if a test string matches at least the first *tomatch* characters of a reference string """
   if len(teststr) > len(ref_str):
      return False
   if len(teststr) < tomatch:
      return False
   return ref_str[0:len(teststr)] == teststr

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class MMPBSA_input:
   """ MMPBSA.py input file class that has all of the routines to parse a namelist-style input
       file and modify the INPUT array based on the input. """

   # Dictionaries that contain, for each namelist, the name of the variable and the number of
   # characters required to match that variable in the input file.

   general = {'startframe' : 4, 'endframe' : 4, 'interval' : 4, 'verbose' : 4, 'strip_mdcrd' : 8, 
              'receptor_mask' : 4, 'ligand_mask' : 4, 'entropy' : 4, 'keep_files' : 4, 'strip_mask' : 8,
              'netcdf' : 4, 'use_sander' : 4 }
   gb      = {'igb' : 3, 'saltcon' : 4, 'surften' : 5, 'surfoff' : 5, 'ifqnt' : 4, 'qmcut' : 5,
              'qm_residues' : 4, 'qm_theory' : 4, 'qmcharge_com' : 10, 'qmcharge_rec' : 10, 'qmcharge_lig' : 10}
   pb      = {'indi' : 4, 'exdi' : 4, 'scale' : 4, 'linit' : 4, 'prbrad' : 4, 'istrng' : 4, 'inp' : 3, 
              'fillratio' : 4,'radiopt' : 4, 'cavity_surften' : 8, 'cavity_offset' : 8, 'sander_apbs' : 4 }
   ala     = {'mutant_only' : 4 }
   nmode   = {'dielc' : 4, 'drms' : 4, 'maxcyc' : 4, 'nmstartframe' : 4, 'nmendframe' : 4, 'nminterval' : 4,
              'ala_entropy' : 4, 'nmode_igb' : 8, 'nmode_istrng' : 8 }
   decomp  = {'idecomp' : 4, 'print_res' : 4, 'dec_verbose' : 4 }
   rism    = {'closure' : 7, 'closureorder' : 8, 'polardecomp' : 4, 'buffer' : 4, 'grdspc' : 4, 'ng' : 2, 
              'solvbox' : 5, 'solvcut' : 5, 'tolerance' : 1.0e-5, 'rism_verbose' : 4, 'thermo' : 4 }

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

   def __init__(self, filename):
      """ Set up the name of the input file and test for its existence """
      self.name = filename
      self.exists = os.path.exists(filename)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

   def __str__(self):
      """ Returns the total text of the input file along with a header for inclusion in the final output file """
      total_string = ''
      if self.exists:
         file = open(self.name,'r')
         total_string = '|Input file:\n|--------------------------------------------------------------\n'
         for line in file:
            total_string += '|' + line
         total_string += '|--------------------------------------------------------------\n'
         file.close()
      else:
         total_string = 'Input file %s does not exist!' % self.name

      return total_string

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

   def Parse(self, master_dict):
      """ Master routine that parses the input file and puts entries into the master dictionary """
      # parse the input file as a Fortran namelist. Only call this function if it exists, but test anyway
      if not self.exists:
         return -1

      # open the input file and read lines into memory
      file = open(self.name, 'r')
      lines = file.readlines()
      file.close()

      # We will loop through the input file a couple times: First will be to load all of the data into 
      # an array (namelist_fields).  Next, we will loop through the fields and combine multi-element values
      # so that we can allow commas in input values.  The last loop through the fields will be to put the
      # input values into the master_dict input dictionary.

      # split up input file into separate fields according to their comma-delimitations
      namelists = []       # array with namelists in order they appear
      namelist_fields = [] # array with entries in a given namelist
      innml = False        # if we are inside a namelist right now
      
      # loop through lines in the input file
      for i in range(len(lines)):
         # skip over comments
         if (not innml and not lines[i].strip().startswith('&')) or (innml and lines[i].strip().startswith('#')):
            continue
         # prevent starting a new namelist before an old one is terminated
         elif innml and lines[i].strip().startswith('&'):
            print >> sys.stderr, 'Error: Invalid input file (%s)! Terminate each namelist before another is started.' % self.name
            return -1
         # find the end of a namelist
         elif innml and (lines[i].strip() == '/' or lines[i].strip() == '&end'):
            innml = False
         # Now if we finally find a namelist
         elif not innml and lines[i].strip().startswith('&'):
            innml = True
            namelist = lines[i].strip()[1:].lower()
            if not self._validNamelist(namelist, master_dict):
               print >> sys.stderr, 'Error: Unrecognized namelist (%s) in MMPBSA.py input file (%s)!' % (namelist, self.name)
               return -1
            if namelist in namelists:
               print >> sys.stderr, 'Error: Namelist (%s) declared multiple times!' % namelist
               return -1
            namelists.append(namelist) # add namelist to list of defined namelists
            namelist_fields.append([]) # add empty array to be filled with vars defined in this namelist
         # Finally, we are in a namelist, now fill in namelist fields
         elif innml:
            items = lines[i].strip().split(',')
            j = 0
            while j < len(items):
               items[j] = items[j].strip()
               if len(items[j]) == 0:
                  items.pop(j)
               else:
                  j += 1
            namelist_fields[len(namelist_fields)-1].extend(items)
      
      # combine any multi-element fields into the last field that has a "=" in it
      begin_field = -1
      for i in range(len(namelist_fields)):
         for j in range(len(namelist_fields[i])):
            if not '=' in namelist_fields[i][j]:
               if begin_field == -1:
                  print >> sys.stderr, 'Error: Invalid input file! Error in reading namelist %s' % namelists[i]
                  return -1
               else:
                  namelist_fields[i][begin_field] += ',%s' % namelist_fields[i][j]
            else:
               begin_field = j

      # now parse through the options and add them to master_dict
      for i in range(len(namelist_fields)):
         for j in range(len(namelist_fields[i])):
            if not '=' in namelist_fields[i][j]:
               continue
            else:
               var = namelist_fields[i][j].split('=')
               if self._change(namelists[i], var[0].strip().lower(), var[1].strip(), master_dict) == -1:
                  return -1

      return 1 # if we've gotten here, we've succeeded!

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

   def _change(self, namelist, variable, value, master_dict):
      """ Changes a given variable in a given namelist and puts the new value in the master dictionary
          so long as that variable actually belongs in the specified namelist """

      namelist = namelist.replace('_',' ')
      if _compare('general', namelist, 3):
         compare_dict = self.general
      elif _compare('gb', namelist, 2):
         compare_dict = self.gb
      elif _compare('pb', namelist, 2):
         compare_dict = self.pb
      elif _compare('alanine scanning', namelist, 3):
         compare_dict = self.ala
      elif _compare('nmode', namelist, 3):
         compare_dict = self.nmode
      elif _compare('decomp', namelist, 3):
         compare_dict = self.decomp
      elif _compare('rism', namelist, 4):
         compare_dict = self.rism

      keys = compare_dict.keys()
      valid_variable = False
      for i in range(len(keys)):
         if _compare(keys[i], variable, compare_dict[keys[i]]):
            valid_variable = True
            variable = variable.strip().strip('"').strip("'")
            master_dict[keys[i]] = value.strip('"').strip("'")
            break
      if not valid_variable:
         print >> sys.stderr, 'Error: Unrecognized variable (%s) in namelist (%s)!' % (variable, namelist)
         return -1

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

   def _validNamelist(self, namelist, master_dict):
      """ Determines if the given namelist is a recognized input namelist and turns on that calculation type """
      namelist = namelist.replace('_',' ')
      if _compare('general', namelist, 3):
         return True
      elif _compare('gb', namelist, 2):
         master_dict['gbrun'] = True
         return True
      elif _compare('pb', namelist, 2):
         master_dict['pbrun'] = True
         return True
      elif _compare('alanine scanning', namelist, 3) or _compare('alanine_scanning', namelist, 3):
         master_dict['alarun'] = True
         return True
      elif _compare('nmode', namelist, 2):
         master_dict['nmoderun'] = True
         return True
      elif _compare('decomp', namelist, 2):
         master_dict['decomprun'] = True
         return True
      elif _compare('rism', namelist, 4):
         master_dict['rismrun'] = True
         return True
      else:
         return False

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
