#########################################################################
#                                                                       #
# This is a module that contains functions generally useful for the     #
# MMPBSA.py script. A full list of functions/subroutines is shown below.#
# It must be included to insure proper functioning of MMPBSA.py         #
#                                                                       #
#          Last updated: 02/25/2011                                     #
#                                                                       #
#########################################################################

########################## GPL LICENSE INFO ############################

#  Copyright (C) 2010  Dwight McGee, Billy Miller III, and Jason Swails

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
   
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.

##########################################################################################################
#  List of functions and a brief description of their purpose
#
#  which: finds an executable in the PATH much like the unix command
#  argParser: Parses the command-line arguments and fills the corresponding elements of FILES
#  printusage: Prints the usage statement
#  remove: removes all files corresponding to a specific level of verbosity
#  GetMasks: given complex,receptor,ligand prmtops, returns the ligand/receptor masks and
#            where the ligand starts
#  CheckIncomp: Checks incompatibilities in the input files
#  createInput: Create sander input files from the INPUT dictionary
#  InterruptHandler: Exit cleanly displaying an "interrupt" message
#  PtrajNumFrames: Parses the number of frames processed from ptraj stdout stream
#  sandercalc: Runs GB and PB sander calculations
#  nmodecalc: Runs normal mode calculations with mmpbsa_py_nabnmode
#  rismcalc: Runs rism3d.snglpnt for 3D-RISM calculations
#  idecompString: Prints a string with the idecomp value and details about that choice
#  PrintFileInfo: Prints information about files to the final output file
#  separate: Prints a double-line of dashes to separate fields
#  Abort: Invokes MPI.Abort for a failed calculation and prints out a useful message
#  _gb_universal: Take an mdin object and make all changes that are necessary in all GB input files
#  _gb_qmmm: Take an mdin object and make the necessary changes for QM/MM calculations
#  _pb_universal: Take an mdin object and make all changes that are necessary in all PB input files
#  _pb_apbs: Take an mdin object and make all changes that are necessary for sander.APBS input files
#  _MakeGBin: Makes a GB input file for mmpbsa_py_energy
#  _MakePBin: Makes a PB input file for mmpbsa_py_energy
#  _getligmask: Internal - gets the ligand mask. Called by GetMasks
#  _getrecmask: Internal - gets the receptor mask. Called by GetMasks
#  _maskToDec: Internal - changes a mask to a Decomp RES string for mdin card
#  _getResStrings: Internal - the strings used in decomp input files to declare which residues are printed
#  _checktrajs: Checks that all specified trajectory files exist
#  _getQMMMmasks: Internal - gets the QM/MM masks for complex, receptor, and ligand based on input
##########################################################################################################

import os, sys
from mdin import mdin

MAX_RES = 20

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def which(program, searchpath=False):
   """ Finds an executable in AMBERHOME/bin or looks in PATH if it's allowed """
   def is_exe(fpath): # determines if a file is executable or not
      """ Returns if a file has execute permissions or not """
      return os.path.exists(fpath) and os.access(fpath, os.X_OK)

   try:
      exe_file = os.path.join(os.environ["AMBERHOME"] + '/bin/', program)
      if is_exe(exe_file):
         return exe_file
   except KeyError:
      print >> sys.stderr, 'Error: AMBERHOME is not defined!'

   if searchpath:
      fpath, fname = os.path.split(program)
      if fpath:
         if is_exe(program):
            return program # if found, return program
      else: # if above doesn't work, try looking for program in every folder in PATH
         for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
               return exe_file # if it's executable, return the file

   return 'none'  # if program can still not be found... return "none"

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def argParser(args, FILES, version):
   """ Parses the command line arguments and fills in the FILES dictionary from them """
   # answer calls for help
   if len(args) < 2 or (args[1] == '-h' or args[1] == '--help'):
      printusage()
      return -1
   # option to clean the screen
   if args[1] == '-clean' or args[1] == '--clean' or args[1] == '-clear' or args[1] == '--clear':
      return -2
   # return the version if requested
   if args[1] == '--version' or args[1] == '-V':
      print >> sys.stdout, version
      return -1

   i = 1
   try:
      while i < len(args):
         if args[i] == '-O':
            FILES['overwrite'] = True
         elif args[i] == '-i':
            i += 1
            FILES['input_file'] = args[i]
         elif args[i] == '-sp':
            i += 1
            FILES['solvated_prmtop'] = args[i]
         elif args[i] == '-cp':
            i += 1
            FILES['complex_prmtop'] = args[i]
         elif args[i] == '-rp':
            i += 1
            FILES['receptor_prmtop'] = args[i]
         elif args[i] == '-lp':
            i += 1
            FILES['ligand_prmtop'] = args[i]
         elif args[i] == '-mc':
            i += 1
            FILES['mutant_complex_prmtop'] = args[i]
         elif args[i] == '-mr':
            i += 1
            FILES['mutant_receptor_prmtop'] = args[i]
         elif args[i] == '-ml':
            i += 1
            FILES['mutant_ligand_prmtop'] = args[i]
         elif args[i] == '-srp':
            i += 1
            FILES['solvated_receptor_prmtop'] = args[i]
         elif args[i] == '-slp':
            i += 1
            FILES['solvated_ligand_prmtop'] = args[i]
         elif args[i] == '-make-mdins':
            FILES['make_mdins'] = True
         elif args[i] == '-use-mdins':
            FILES['use_mdins'] = True
         elif args[i] == '-rewrite-output':
            FILES['rewrite_output'] = True
         elif args[i] == '-o':
            i += 1
            FILES['output_file'] = args[i]
         elif args[i] == '-do':
            i += 1
            FILES['decompout'] = args[i]
         elif args[i] == '-y':
            i += 1
            while i < len(args) and not args[i].startswith('-'):
               mdcrd_holder = args[i].split(',')
               for j in range(len(mdcrd_holder)):
                  if len(mdcrd_holder[j].strip()) != 0:
                     FILES['mdcrd'].append(mdcrd_holder[j].strip())
               i += 1
            i -= 1
         elif args[i] == '-yr':
            i += 1
            while i < len(args) and not args[i].startswith('-'):
               mdcrd_holder = args[i].split(',')
               for j in range(len(mdcrd_holder)):
                  if len(mdcrd_holder[j].strip()) != 0:
                     FILES['receptor_mdcrd'].append(mdcrd_holder[j].strip())
               i += 1
            i -= 1
         elif args[i] == '-yl':
            i += 1
            while i < len(args) and not args[i].startswith('-'):
               mdcrd_holder = args[i].split(',')
               for j in range(len(mdcrd_holder)):
                  if len(mdcrd_holder[j].strip()) != 0:
                     FILES['ligand_mdcrd'].append(mdcrd_holder[j].strip())
               i += 1
            i -= 1
         else:
            print >> sys.stderr, 'Error: Unrecognized command-line argument (%s)!' % args[i]
            return -1

         i += 1 # move to the next argument

   except IndexError:
      print >> sys.stderr, 'Error: Command line error!'
      printusage()
      return -1

   # Check for incompatible choices
   if FILES['rewrite_output'] and (FILES['use_mdins'] or FILES['make_mdins']):
      print >> sys.stderr, 'Error: -rewrite-output cannot be used with -use-mdins or -make-mdins!'
      return -1
   if FILES['use_mdins'] and FILES['make_mdins']:
      print >> sys.stderr, 'Error: -use-mdins and -make-mdins are mutually exclusive!'
      return -1

   # Make sure that -make-mdins and -rewrite-output is only done in serial
   if (FILES['rewrite_output'] or FILES['make_mdins']) and os.path.split(args[0])[1] == 'MMPBSA.py.MPI':
      print >> sys.stderr, 'Error: Use the serial version of MMPBSA.py for -make-mdins or -rewrite-output!'
      return -1

   # Check that, if not overwriting, the original output files do not already exist
   if not FILES['overwrite'] and (os.path.exists(FILES['output_file']) or os.path.exists(FILES['decompout'])):
      print >> sys.stderr, 'Error: Output file(s) %s and/or %s already exist! Set -O to overwrite.' % \
                        (FILES['output_file'], FILES['decompout'])
      return -1

   # Load default mdcrd if none was specified
   if len(FILES['mdcrd']) == 0:
      FILES['mdcrd'].append('mdcrd')

   # Check that all specified mdcrd files exist and return its return code
   return _checktrajs(FILES)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def printusage():
   """ Prints the program usage to standard output """
   print >> sys.stdout, "Usage: MMPBSA.py {-O} -i   inputfile \\"
   print >> sys.stdout, "                      -o   output_file \\"
   print >> sys.stdout, "                      -sp  solvated_prmtop \\"
   print >> sys.stdout, "                      -cp  complex_prmtop \\"
   print >> sys.stdout, "                      -rp  receptor_prmtop \\"
   print >> sys.stdout, "                      -lp  ligand_prmtop \\"
   print >> sys.stdout, "                      -y   mdcrd1 mdcrd2 mdcrd3 ... mdcrdN \\"
   print >> sys.stdout, "                     {-do  decomp_output_file} \\"
   print >> sys.stdout, "                     {-yr  receptor_mdcrd1 ... receptor_mdcrdN} \\"
   print >> sys.stdout, "                     {-yl  ligand_mdcrd1 ... ligand_mdcrdN} \\ "
   print >> sys.stdout, "                     {-mc  mutant_complex_prmtop} \\"
   print >> sys.stdout, "                     {-mr  mutant_receptor_prmtop} \\"
   print >> sys.stdout, "                     {-ml  mutant_ligand_prmtop} \\"
   print >> sys.stdout, "                     {-srp solvated_receptor_prmtop} \\"
   print >> sys.stdout, "                     {-slp solvated_ligand_prmtop} \\"
   print >> sys.stdout, "                     {-xvvfile xvvfile} \\"
   print >> sys.stdout, "                     {-make-mdins} || {-use-mdins} || {-rewrite-output}"

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def remove(flag, mpi_size=0):
   """ Removes temporary files. Allows for different levels of cleanliness """
   if flag == -3: # internal -- keep all mdin files
      os.system('tar -cf tmp.tar _MMPBSA_*.mdin; rm -f _MMPBSA_* io.mc; tar -xf tmp.tar; rm tmp.tar')
   elif flag == -2: # internal -- if program exits in error, keep all files and say so
      print >> sys.stderr, 'NOTE: All files have been retained for debugging purposes. ' + \
                           'Type MMPBSA.py --clean to erase these files.'
      sys.exit(1)
   elif flag == 0: # remove all temporary files
      os.system('rm -f _MMPBSA_* io.mc')
   elif flag == 1: # keep keep mdcrds, mdouts, and other relevant output files
      os.system('rm -f _MMPBSA_*.inpcrd _MMPBSA_*.in _MMPBSA_*mdin io.mc _MMPBSA_*restrt')
      for i in range(mpi_size): # remove all temp files created by individual MPI threads
         os.system('rm -f _MMPBSA_*.%s' % i)
      
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def GetMasks(comres, recres, ligres):
   """ Guesses receptor and ligand masks based on complex, receptor, and ligand residue sequences """
   # getligmask returns array with [lig starting position, lig mask]
   liginfo = _getligmask(comres, recres, ligres)

   # if it returned an error code
   if liginfo == -1:
      return -1, -1, -1

   # getrecmask returns string with receptor mask
   recinfo = _getrecmask(len(recres), liginfo[0], len(ligres))

   # return ligand mask, rec mask, lig starting position
   return liginfo[1], recinfo, liginfo[0]

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def CheckIncomp(INPUT, stability):
   """ Checks for inconsistencies in the input parameters """
   isinerr = 0
   if INPUT['startframe'] > INPUT['endframe']:
      print >> sys.stderr, 'Input Error: STARTFRAME (%s) must be smaller than ENDFRAME (%s)' % (INPUT['startframe'],INPUT['endframe'])
      isinerr = -1
   if INPUT['startframe'] < 1 or INPUT['endframe'] < 1:
      print >> sys.stderr, 'Input Error: STARTFRAME and ENDFRAME must both be positive integers!'
      isinerr = -1
   if INPUT['interval'] < 1:
      print >> sys.stderr, 'Input Error: INTERVAL must be a positive integer!'
      isinerr = -1
   if not INPUT['igb'] in [1, 2, 5, 7, 8]:
      print >> sys.stderr, 'Input Error: Invalid value for IGB (%s)! It must be 1, 2, 5, 7, or 8.' % INPUT['igb']
      isinerr = -1
   if INPUT['saltcon'] < 0:
      print >> sys.stderr, 'Input Error: SALTCON must be non-negative!'
      isinerr = -1
   if INPUT['extdiel'] < 0:
      print >> sys.stderr, 'Input Error: EXTDIEL must be non-negative!'
      isinerr = -1
   if INPUT['intdiel'] < 0:
      print >> sys.stderr, 'Input Error: INTDIEL must be non-negative!'
      isinerr = -1
   if INPUT['surften'] < 0:
      print >> sys.stderr, 'Input Error: SURFTEN must be non-negative!'
      isinerr = -1
   if INPUT['indi'] < 0:
      print >> sys.stderr, 'Input Error: INDI must be non-negative!'
      isinerr = -1
   if INPUT['exdi'] < 0:
      print >> sys.stderr, 'Input Error: EXDI must be non-negative!'
      isinerr = -1
   if INPUT['scale'] < 0:
      print >> sys.stderr, 'Input Error: SCALE must be non-negative!'
      isinerr = -1
   if INPUT['linit'] < 0:
      print >> sys.stderr, 'Input Error: LINIT must be a positive integer!'
      isinerr = -1
   if not INPUT['prbrad'] in [1.4, 1.6]:
      print >> sys.stderr, 'Input Error: PRBRAD (%s) only has support for values 1.4 and 1.6!' % INPUT['prbrad']
      isinerr = -1
   if INPUT['istrng'] < 0:
      print >> sys.stderr, 'Input Error: ISTRNG must be non-negative!'
      isinerr = -1
   if not INPUT['inp'] in [0, 1, 2]:
      print >> sys.stderr, 'Input Error: INP/NPOPT (%s) must be either 0, 1, or 2!' % INPUT['inp']
      isinerr = -1
   if INPUT['cavity_surften'] < 0:
      print >> sys.stderr, 'Input Error: CAVITY_SURFTEN must be non-negative!'
      isinerr = -1
   if INPUT['fillratio'] <= 0:
      print >> sys.stderr, 'Input Error: FILL_RATIO must be positive!'
      isinerr = -1
   if not INPUT['radiopt'] in [0, 1]:
      print >> sys.stderr, 'Input Error: RADIOPT (%s) must be 0 or 1!' % INPUT['radiopt']
      isinerr = -1
   if INPUT['dielc'] <= 0:
      print >> sys.stderr, 'Input Error: DIELC must be positive!'
      isinerr = -1
   if INPUT['maxcyc'] < 1:
      print >> sys.stderr, 'Input Error: MAXCYC (%s) must be a positive integer!' % INPUT['maxcyc']
      isinerr = -1
   if INPUT['nmstartframe'] > INPUT['nmendframe']:
      print >> sys.stderr, 'Input Error: NMSTARTFRAME (%s) must be less than NMENDFRAME (%s) !' % (INPUT['nmstartframe'],INPUT['nmendframe'])
      isinerr = -1
   if INPUT['nmstartframe'] < 1 or INPUT['nmendframe'] < 1:
      print >> sys.stderr, 'Input Error: NMSTARTFRAME and NMENDFRAME must both be positive integers!'
      isinerr = -1
   if INPUT['nminterval'] < 1:
      print >> sys.stderr, 'Input Error: NMINTERVAL must be a positive integer!'
      isinerr = -1
   if not INPUT['idecomp'] in [0, 1, 2, 3, 4]:
      print >> sys.stderr, 'Input Error: IDECOMP (%s) must be 1, 2, 3, or 4!' % INPUT['idecomp']
      isinerr = -1
   if INPUT['idecomp'] != 0 and INPUT['gbsa'] != 2:
      print >> sys.stderr, 'Input Error: When IDECOMP is set, GBSA must be set to 2!'
      isinerr = -1
   if INPUT['idecomp'] != 0 and INPUT['sander_apbs'] == 1:
      print >> sys.stderr, 'Input Error: IDECOMP cannot be used with sander.APBS!'
      isinerr = -1
   if not INPUT['entropy'] in [0, 1]:
      print >> sys.stderr, 'Input Error: ENTROPY (%s) must be 0 or 1!' % INPUT['entropy']
      isinerr = -1
   if not INPUT['sander_apbs'] in [0, 1]:
      print >> sys.stderr, 'Input Error: SANDER_APBS (%s) must be 0 or 1!' % INPUT['sander_apbs']
      isinerr = -1
   if INPUT['alarun'] and INPUT['netcdf'] != '':
      print >> sys.stderr, 'Input Error: Alanine scanning is incompatible with NETCDF != 0!'
      isinerr = -1
   if INPUT['decomprun'] and INPUT['idecomp'] == 0:
      print >> sys.stderr, 'Input Error: IDECOMP cannot be 0 for Decomposition analysis!'
      isinerr = -1
   if not INPUT['gbrun'] and not INPUT['pbrun'] and not INPUT['nmoderun'] and not INPUT['entropy'] == 1 and not INPUT['rismrun']:
      print >> sys.stderr, 'Error: No calculation type specified. Make sure you select GB, PB, or an entropy calculation!'
      isinerr = -1
   if INPUT['use_sander'] != 0 and INPUT['use_sander'] != 1:
      print >> sys.stderr, 'Error: USE_SANDER must be set to 0 or 1!'
      isinerr = -1
   if INPUT['ifqnt'] != 0 and INPUT['ifqnt'] != 1:
      print >> sys.stderr, 'Error: QMMM must be set to 0 or 1!'
      isinerr = -1
   if INPUT['ifqnt'] == 1:
      if not INPUT['qm_theory'] in ['PM3','AM1','MNDO','PDDG-PM3','PM3PDDG','PDDG-MNDO','PDDGMNDO', 
                                    'PM3-CARB1','PM3CARB1','DFTB','SCC-DFTB','RM1','PM6','PM3-ZnB','PM3-MAIS']:
         print >> sys.stderr, 'Error: Invalid QM_THEORY (%s)! This variable must be set to allowable options.' % INPUT['qm_theory']
         print >> sys.stderr, '       See the AmberTools manual for allowable options.'
         isinerr = -1
      if INPUT['qm_residues'] == '':
         print >> sys.stderr, 'Error: QM_RESIDUES must be specified for IFQNT = 1!'
         isinerr = -1
      if INPUT['pbrun']:
         print >> sys.stderr, 'Error: PB and QM/MM are incompatible!'
         isinerr = -1
      if not INPUT['gbrun']:
         print >> sys.stderr, 'Error: QM/MM must be run with GB!'
         isinerr = -1
      if INPUT['decomprun']:
         print >> sys.stderr, 'Error: QM/MM and decomposition analysis are incompatible!'
         isinerr = -1
      if INPUT['qmcharge_lig'] + INPUT['qmcharge_rec'] != INPUT['qmcharge_com'] and not stability:
         print >> sys.stderr, 'Error: The total charge of the ligand and receptor does not equal the charge of the complex!'
         isinerr = -1
   if INPUT['rismrun']:
      if INPUT['rism_verbose'] > 2 or INPUT['rism_verbose'] < 0:
         print >> sys.stderr, 'Error: RISM_VERBOSE must be 0, 1, or 2!'
         isinerr = -1
      if INPUT['buffer'] < 0 and INPUT['solvcut'] < 0:
         print >> sys.stderr, 'Error: If BUFFER < 0, SOLVCUT must be > 0!'
         isinerr = -1
      if INPUT['tolerance'] < 0:
         print >> sys.stderr, 'Error: TOLERANCE must be positive!'
         isinerr = -1
      if INPUT['buffer'] < 0 and INPUT['ng'] == '':
         print >> sys.stderr, 'Error: You must specify NG if BUFFER < 0!'
         isinerr = -1
      if INPUT['closure'] == 'pse' and INPUT['closureorder'] < 1:
         print >> sys.stderr, 'Error: You must specify CLOSUREORDER if CLOSURE=pse!'
         isinerr = -1
      if not INPUT['polardecomp'] in [0,1]:
         print >> sys.stderr, 'Error: POLARDECOMP must be either 0 or 1!'
         isinerr = -1
      if not INPUT['thermo'] in ['std', 'gf', 'both']:
         print >> sys.stderr, 'Error: THERMO must be "std", "gf", or "both"!'
         isinerr = -1

   return isinerr

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def createInput(INPUT, ligstart, stability, comres, recres, ligres, prog):
   """ Creates sander input files for calculations """
   # get the strings we need for decomp files if necessary
   if INPUT['decomprun']:
      printstrings = _getResStrings(comres, recres, ligres, INPUT['print_res'], ligstart, stability)
      if printstrings == -1:
         return -1
   # create the gb mdin files
   if INPUT['gbrun']:
      # decomp gb mdin files
      if INPUT['decomprun']:
         com_gbmdin = mdin('sander')
         _gb_universal(INPUT, com_gbmdin)
#        com_gbmdin.change('cntrl','dec_verbose',INPUT['dec_verbose'])
         if stability:
            com_gbmdin.AddCard('Residues considered as REC','RRES 1 %s' % comres)
         else:
            com_gbmdin.AddCard('Residues considered as REC','RRES %s' % _maskToDec(INPUT['receptor_mask']))
            com_gbmdin.AddCard('Residues considered as LIG','LRES %s' % _maskToDec(INPUT['ligand_mask']))
         com_gbmdin.AddCard('Residues to print', 'RES %s' % printstrings[0])
         com_gbmdin.write('_MMPBSA_gb_decomp_com.mdin')
         # receptor and ligand mdin files if necessary
         if not stability:
            rec_gbmdin = mdin('sander')
            _gb_universal(INPUT, rec_gbmdin)
#           rec_gbmdin.change('cntrl','dec_verbose',INPUT['dec_verbose'])
            rec_gbmdin.AddCard('Residues considered as REC','RRES 1 %s' % recres)
            rec_gbmdin.AddCard('Residues to print','RES %s' % printstrings[1])
            rec_gbmdin.write('_MMPBSA_gb_decomp_rec.mdin')
            # ligand input file
            lig_gbmdin = mdin('sander')
            _gb_universal(INPUT, lig_gbmdin)
#           lig_gbmdin.change('cntrl','dec_verbose',INPUT['dec_verbose'])
            lig_gbmdin.AddCard('Residues considered as LIG','LRES 1 %s' % ligres)
            lig_gbmdin.AddCard('Residues to print','RES %s' % printstrings[2])
            lig_gbmdin.write('_MMPBSA_gb_decomp_lig.mdin')
      elif INPUT['ifqnt'] == 1: # qm/mm calculation
         qmmasks = _getQMMMmasks(INPUT['qm_residues'], comres, recres, ligres, ligstart, stability)
         com_gbmdin = mdin('sander')
         _gb_qmmm(INPUT, com_gbmdin, qmmasks[0], INPUT['qmcharge_com'])
         com_gbmdin.write('_MMPBSA_gb_qmmm_com.mdin')
         if not stability:
            rec_gbmdin = mdin('sander')
            _gb_qmmm(INPUT, rec_gbmdin, qmmasks[1], INPUT['qmcharge_rec'])
            rec_gbmdin.write('_MMPBSA_gb_qmmm_rec.mdin')
            lig_gbmdin = mdin('sander')
            _gb_qmmm(INPUT, lig_gbmdin, qmmasks[2], INPUT['qmcharge_lig'])
            lig_gbmdin.write('_MMPBSA_gb_qmmm_lig.mdin')
      else:
         if prog == 'mmpbsa_py_energy':
            _MakeGBin(INPUT)
         else:
            gbmdin = mdin('sander')
            _gb_universal(INPUT, gbmdin)
            gbmdin.write('_MMPBSA_gb.mdin')

   # create the pb mdin files
   if INPUT['pbrun']:
      # decomp pb mdin files
      if INPUT['decomprun']:
         com_pbmdin = mdin('sander')
         _pb_universal(INPUT, com_pbmdin)
         if stability:
            com_pbmdin.AddCard('Residues considered as REC','RRES 1 %s' % comres)
         else:
            com_pbmdin.AddCard('Residues considered as REC','RRES %s' % _maskToDec(INPUT['receptor_mask']))
            com_pbmdin.AddCard('Residues considered as LIG','LRES %s' % _maskToDec(INPUT['ligand_mask']))
         com_pbmdin.AddCard('Residues to print', 'RES %s' % printstrings[0])
         com_pbmdin.write('_MMPBSA_pb_decomp_com.mdin')
         # receptor and ligand mdin files if necessary
         if not stability:
            rec_pbmdin = mdin('sander')
            _pb_universal(INPUT, rec_pbmdin)
            rec_pbmdin.AddCard('Residues considered as REC','RRES 1 %s' % recres)
            rec_pbmdin.AddCard('Residues to print','RES %s' % printstrings[1])
            rec_pbmdin.write('_MMPBSA_pb_decomp_rec.mdin')
            # ligand input file
            lig_pbmdin = mdin('sander')
            _pb_universal(INPUT, lig_pbmdin)
            lig_pbmdin.AddCard('Residues considered as LIG','LRES 1 %s' % ligres)
            lig_pbmdin.AddCard('Residues to print','RES %s' % printstrings[2])
            lig_pbmdin.write('_MMPBSA_pb_decomp_lig.mdin')
      else: # not decomp
         if INPUT['sander_apbs'] == 1:
            pbmdin = mdin('sander.APBS')
            _pb_apbs(INPUT, pbmdin)
            pbmdin.write('_MMPBSA_pb.mdin')
         else:
            if prog == "mmpbsa_py_energy":
               _MakePBin(INPUT)
            else:
               pbmdin = mdin('sander')
               _pb_universal(INPUT, pbmdin)
               pbmdin.write('_MMPBSA_pb.mdin')

   return 1

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def InterruptHandler(signal, frame):
   """ Handles interrupt signals for a clean exit """
   print >> sys.stderr, ''
   print >> sys.stderr, '%s interrupted! Program terminated. All files are kept.' % os.path.split(sys.argv[0])[1]
   sys.exit(2)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def PtrajNumFrames(filename):
   """ Parses the number of files from ptraj output """
   if not os.path.exists(filename):
      return -1

   file = open(filename, 'r') # open the file

   for line in file: # parse each line
      if 'Coordinate processing' in line: # this is the line that has number of frames run
         return int(line.split()[5]) # it is the 6th word in the line

   return -1 # return error code if we couldn't find the number of threads processed

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def sandercalc(sander, suffix, prefix, trjsuffix, complex_prmtop, receptor_prmtop, ligand_prmtop, idecomp, stability, type, ifqnt):
   """ Executes a sander calculation for complex and/or receptor/ligand. """
   if idecomp > 0:
      mdin1 = '_MMPBSA_%s_decomp_com.mdin' % type
      mdin2 = '_MMPBSA_%s_decomp_rec.mdin' % type
      mdin3 = '_MMPBSA_%s_decomp_lig.mdin' % type
   elif ifqnt == 1:
      mdin1 = '_MMPBSA_%s_qmmm_com.mdin' % type
      mdin2 = '_MMPBSA_%s_qmmm_rec.mdin' % type
      mdin3 = '_MMPBSA_%s_qmmm_lig.mdin' % type
   else: # no decomp means all input files are the same
      mdin1 = mdin2 = mdin3 = '_MMPBSA_%s.mdin' % type

   if sander.endswith('mmpbsa_py_energy'):
      pdbflag1 = '-pdb %scomplex.pdb' % prefix
      pdbflag2 = '-pdb %sreceptor.pdb' % prefix
      pdbflag3 = '-pdb %sligand.pdb' % prefix
   else:
      pdbflag1 = pdbflag2 = pdbflag3 = ''

   # in case users did not disable sander.APBS output to stdout, direct it to a file.  There may be a LOT
   # overwrite so we don't use up too much space... may be bad for some hardcore debuggers, but they can work around
   if sander.endswith('APBS'):
      endcmd = '> _MMPBSA_pbsanderoutput.junk'
   else:
      endcmd = ''

   # rc is the return code -- make sure our calculation was successful before we continue on with others
   print >> sys.stdout, '  calculating complex contribution...'
   rc = os.system('%s -O -i %s -o %scomplex_%s.mdout%s -p %s -c %sdummycomplex.inpcrd.1 -y %scomplex%s%s -r %s.restrt%s %s %s' %
            (sander, mdin1, prefix, type, suffix, complex_prmtop, prefix, prefix, trjsuffix, suffix, prefix, suffix, pdbflag1, endcmd))
   if rc != 0: return -1

   if not stability:
      print >> sys.stdout, '  calculating receptor contribution...'
      rc = os.system('%s -O -i %s -o %sreceptor_%s.mdout%s -p %s -c %sdummyreceptor.inpcrd.1 -y %sreceptor%s%s -r %s.restrt%s %s %s' %
            (sander, mdin2, prefix, type, suffix, receptor_prmtop, prefix, prefix, trjsuffix, suffix, prefix, suffix, pdbflag2, endcmd))
      if rc != 0: return -1

      print >> sys.stdout, '  calculating ligand contribution...'
      rc = os.system('%s -O -i %s -o %sligand_%s.mdout%s -p %s -c %sdummyligand.inpcrd.1 -y %sligand%s%s -r %s.restrt%s %s %s' %
            (sander, mdin3, prefix, type, suffix, ligand_prmtop, prefix, prefix, trjsuffix,suffix, prefix, suffix, pdbflag3, endcmd))
      if rc != 0: return -1
      
   return 0

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def nmodecalc(prefix, istrng, maxcyc, drms, gasorvac, dielc, com_prm, rec_prm, lig_prm, suffix, exe, sfx_str, temp, wcons, mpi, stability):
   """ Executes mmpbsa_py_nabnmode for nmode calculations on complex and/or receptor/ligand """
   from math import sqrt

   kappa = sqrt(0.10806 * istrng)

   if gasorvac == 1:
      optstring = '"ntpr=10000, diel=C, kappa=%s, cut=1000, gb=1, dielc=%s, temp0=%s, wcons=%s"' % (kappa, dielc, temp, wcons)
   else:
      optstring = '"ntpr=10000, diel=R, kappa=%s, cut=1000, gb=0, dielc=%s, temp0=%s, wcons=%s"' % (kappa, dielc, temp, wcons)

   if mpi:
      print >> sys.stdout, 'Master thread is calculating normal modes for %s frames\n' % len(suffix)
   else:
      print >> sys.stdout, 'Calculating normal modes for %s frames\n' % len(suffix)

   for i in range(len(suffix)):
      print >> sys.stdout, '  calculating complex  contribution for frame %s' % i
      rc = os.system('%s %scomplex_nm.pdb.%s %s %s %s %s >> %scomplex_nm.out%s' % 
                (exe, prefix, suffix[i], com_prm, maxcyc, drms, optstring, prefix, sfx_str))
      if rc != 0: return -1
      if not stability:
         print >> sys.stdout, '  calculating receptor contribution for frame %s' % i
         rc = os.system('%s %sreceptor_nm.pdb.%s %s %s %s %s >> %sreceptor_nm.out%s' %
                   (exe, prefix, suffix[i], rec_prm, maxcyc, drms, optstring, prefix, sfx_str))
         if rc != 0: return -1
         print >> sys.stdout, '  calculating  ligand  contribution for frame %s' % i
         rc = os.system('%s %sligand_nm.pdb.%s %s %s %s %s >> %sligand_nm.out%s' %
                   (exe, prefix, suffix[i], lig_prm, maxcyc, drms, optstring, prefix, sfx_str))
         if rc != 0: return -1

   return 0

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def rismcalc(rism3d, xvvfile, INPUT, prefix, sfx_str, stability, trjsuffix, comprm, recprm, ligprm):
   """ Responsible for running 3D-RISM simulations """
   trajflag = '--traj'

   if INPUT['ng'] == "-1,-1,-1":
      ngstring = ''
   else:
      ngstring = '--ng %s' % INPUT['ng']

   if INPUT['solvbox'] == "-1,-1,-1":
      solvboxstring = ''
   else:
      solvboxstring = '--solvbox %s' % INPUT['solvbox']

   if INPUT['polardecomp'] > 0:
      polardecomp = '--polarDecomp'
   else:
      polardecomp = ''

   print >> sys.stdout, '  calculating complex contribution...'
   rc = os.system(('%s --xvv %s --closure %s --closureorder %s %s --buffer %s --grdspc %s %s %s ' +
              '--solvcut %s --tolerance %s --verbose %s %s %scomplex%s > %scomplex_rism.out%s ' + 
              '--pdb %scomplex.pdb --prmtop %s') % (rism3d, xvvfile,
              INPUT['closure'], INPUT['closureorder'], polardecomp, INPUT['buffer'], 
              INPUT['grdspc'], ngstring, solvboxstring, INPUT['solvcut'], INPUT['tolerance'], 
              INPUT['rism_verbose'], trajflag, prefix, trjsuffix + sfx_str, prefix, sfx_str, 
              prefix, comprm))
   if rc != 0: return -1

   if not stability:
      print >> sys.stdout, '  calculating receptor contribution...'
      rc = os.system(('%s --xvv %s --closure %s --closureorder %s %s --buffer %s '+
                 '--grdspc %s %s %s --solvcut %s --tolerance %s --verbose %s %s %sreceptor%s '+
                 '> %sreceptor_rism.out%s --pdb %sreceptor.pdb --prmtop %s') % (rism3d, xvvfile,
                 INPUT['closure'], INPUT['closureorder'], polardecomp, INPUT['buffer'], 
                 INPUT['grdspc'], ngstring, solvboxstring, INPUT['solvcut'], INPUT['tolerance'], 
                 INPUT['rism_verbose'], trajflag, prefix, trjsuffix + sfx_str, prefix, sfx_str, prefix, recprm))
      if rc != 0: return -1
              
      print >> sys.stdout, '  calculating ligand contribution...'
      os.system(('%s --xvv %s --closure %s --closureorder %s %s --buffer %s --grdspc %s %s %s ' +
                 '--solvcut %s --tolerance %s --verbose %s %s %sligand%s > %sligand_rism.out%s ' +
                 '--pdb %sligand.pdb --prmtop %s') % (rism3d, xvvfile,
                 INPUT['closure'], INPUT['closureorder'], polardecomp, INPUT['buffer'], 
                 INPUT['grdspc'], ngstring, solvboxstring, INPUT['solvcut'], INPUT['tolerance'], 
                 INPUT['rism_verbose'], trajflag, prefix, trjsuffix + sfx_str, prefix, sfx_str, 
                 prefix, ligprm))
      if rc != 0: return -1

   return 0

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def idecompString(idecomp):
   """ Returns the decomposition scheme description based on the value of idecomp """
   options = ['idecomp = 0: No decomposition analysis',
              'idecomp = 1: Decomposition per-residue adding 1-4 interactions to Internal.',
              'idecomp = 2: Decomposition per-residue adding 1-4 interactions to EEL and VDW.',
              'idecomp = 3: Pairwise decomposition adding 1-4 interactions to Internal.'
              'idecomp = 4: Pairwise decomposition adding 1-4 interactions to EEL and VDW.' ]

   return options[idecomp]
              
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def PrintFileInfo(outfile, FILES, INPUT, stability):
   """ Prints information about used files to the final output file """
   # Write information about prmtop files
   if INPUT['strip_mdcrd'] == 1:
      outfile.write('|Solvated complex topology file:  %s\n' % FILES['solvated_prmtop'])
   outfile.write('|Complex topology file:           %s\n' % FILES['complex_prmtop'])
   if not stability:
      if len(FILES['receptor_mdcrd']) != 0 and INPUT['strip_mdcrd'] == 1:
         outfile.write('|Solvated receptor topology file: %s\n' % FILES['solvated_receptor_prmtop'])
      outfile.write('|Receptor topology file:          %s\n' % FILES['receptor_prmtop'])
      if len(FILES['ligand_mdcrd']) != 0 and INPUT['strip_mdcrd'] == 1:
         outfile.write('|Solvated ligand topology file:   %s\n' % FILES['solvated_ligand_prmtop'])
      outfile.write('|Ligand topology file:            %s\n' % FILES['ligand_prmtop'])

   if INPUT['alarun']:
      outfile.write('|Mutant complex topology file:    %s\n' % FILES['mutant_complex_prmtop'])
      if not stability:
         outfile.write('|Mutant receptor topology file:   %s\n' % FILES['mutant_receptor_prmtop'])
         outfile.write('|Mutant ligand topology file:     %s\n' % FILES['mutant_ligand_prmtop'])

      
   # Write information about the initial trajectories:
   outfile.write('|Initial mdcrd(s):                ')
   for i in range(len(FILES['mdcrd'])):
      if i == 0: outfile.write('%s\n' % FILES['mdcrd'][i])
      else: outfile.write('|                                 %s\n' % FILES['mdcrd'][i])
      
   if len(FILES['receptor_mdcrd']) != 0:
      outfile.write('|Initial Receptor mdcrd(s):       ')
      for i in range(len(FILES['receptor_mdcrd'])):
         if i == 0: outfile.write('%s\n' % FILES['receptor_mdcrd'][i])
         else: outfile.write('|                                 %s\n' % FILES['receptor_mdcrd'][i])
   
   if len(FILES['ligand_mdcrd']) != 0:
      outfile.write('|Initial Ligand mdcrd(s):         ')
      for i in range(len(FILES['ligand_mdcrd'])):
         if i == 0: outfile.write('%s\n' % FILES['ligand_mdcrd'][i])
         else: outfile.write('|                                 %s\n' % FILES['ligand_mdcrd'][i])
   
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def separate(file):
   """ Final output field separator """
   file.write('-------------------------------------------------------------------------------\n')
   file.write('-------------------------------------------------------------------------------\n')

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def Abort(comm, calctype, rank):
   """ Prints out useful error message (i.e. which process died) in which rank, then
       calls the MPI Abort routine to kill all processes """
   # Make sure we open any thread back up to printing to stderr
   sys.stderr = os.fdopen(sys.stderr.fileno(),'a',0)
   print >> sys.stderr, "MMPBSA.py failed during the %s calculation on processor %d" % (calctype, rank)
   print >> sys.stderr, "Aborting. All files have been retained for debugging purposes"
   comm.Abort()

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _gb_universal(INPUT, infile):
   """ Set GB mdin options """
   infile.title = "File generated by MMPBSA.py"
   infile.change('cntrl','ntb', 0)
   infile.change('cntrl','cut', 999.0)
   infile.change('cntrl','nsnb', 99999)
   infile.change('cntrl','imin', 5)
   infile.change('cntrl','maxcyc', 1)
   infile.change('cntrl','ncyc', 0)
   infile.change('cntrl','igb', INPUT['igb'])
   infile.change('cntrl','saltcon', INPUT['saltcon'])
   infile.change('cntrl','intdiel', INPUT['intdiel'])
   infile.change('cntrl','extdiel', INPUT['extdiel'])
   infile.change('cntrl','gbsa', INPUT['gbsa'])
   infile.change('cntrl','surften', INPUT['surften'])
   infile.change('cntrl','offset', INPUT['surfoff'])
   infile.change('cntrl','idecomp', INPUT['idecomp'])
   if INPUT['netcdf'].strip().lower() == 'netcdf':
      infile.change('cntrl','ioutfm', 1)
   
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _gb_qmmm(INPUT, infile, qmmask, charge):
   """ Set GB QM/MM options """
   # Write a universal GB input file, then only write the QM/MM parts if there are QM/MM atoms in
   # the given mask.
   _gb_universal(INPUT, infile)
   if len(qmmask.strip()) > 0:
      infile.change('cntrl','ifqnt', 1)
      infile.change('qmmm','qmmask', "'%s'" % qmmask)
      infile.change('qmmm','qm_theory', "'%s'" % INPUT['qm_theory'])
      infile.change('qmmm','qmcharge', charge)
      infile.change('qmmm','qmgb', 2)
      infile.change('qmmm','qmcut', INPUT['qmcut'])

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _pb_universal(INPUT, infile):
   """ Set PB mdin options """
   infile.title = "File generated by MMPBSA.py"
   infile.change('cntrl','ntb', 0)
   infile.change('cntrl','cut', 999.0)
   infile.change('cntrl','nsnb', 99999)
   infile.change('cntrl','imin', 5)
   infile.change('cntrl','maxcyc', 1)
   infile.change('cntrl','igb', 10)
   infile.change('cntrl','ipb', 1)
   infile.change('cntrl','inp', INPUT['inp'])
   infile.change('cntrl','idecomp', INPUT['idecomp'])
   infile.change('pb','dbfopt', 1)
   infile.change('pb','epsin', INPUT['indi'])
   infile.change('pb','epsout', INPUT['exdi'])
   infile.change('pb','istrng', INPUT['istrng'] * 1000)
   infile.change('pb','radiopt', INPUT['radiopt'])
   infile.change('pb','sprob', INPUT['prbrad'])
   infile.change('pb','space', INPUT['scale'])
   infile.change('pb','maxitn', INPUT['linit'])
   infile.change('pb','cavity_surften', INPUT['cavity_surften'])
   infile.change('pb','cavity_offset', INPUT['cavity_offset'])
   infile.change('pb','fillratio', INPUT['fillratio'])
   if INPUT['netcdf'].strip().lower() == 'netcdf':
      infile.change('cntrl','ioutfm', 1)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _pb_apbs(INPUT, infile):
   """ Set apbs PB input options """
   infile.title = "File generated by MMPBSA.py"
   infile.change('cntrl','ntb', 0)
   infile.change('cntrl','cut', 999.0)
   infile.change('cntrl','nsnb', 99999)
   infile.change('cntrl','imin', 5)
   infile.change('cntrl','maxcyc', 1)
   infile.change('cntrl','igb', 6)
   infile.change('apbs','apbs_print', 1)
   infile.change('apbs','calc_type', 0)
   infile.change('apbs','cmeth', 1)
   infile.change('apbs','bcfl', 2)
   infile.change('apbs','srfm', 1)
   infile.change('apbs','chgm', 1)
   infile.change('apbs','pdie', INPUT['indi'])
   infile.change('apbs','sdie', INPUT['exdi'])
   infile.change('apbs','srad', INPUT['prbrad'])
   infile.change('apbs','nion', 2)
   infile.change('apbs','ionq','1.0,-1.0')
   infile.change('apbs','ionc','%s,%s' % (INPUT['istrng'], INPUT['istrng']))
   infile.change('apbs','ionrr','2.0,2.0')
   infile.change('apbs','radiopt', INPUT['radiopt'])
   infile.change('apbs','calcforce', 0)
   infile.change('apbs','calcnpenergy', 1)
   infile.change('apbs','grid','%s,%s,%s' % (INPUT['scale'], INPUT['scale'], INPUT['scale']))
   infile.change('apbs','gamma', INPUT['cavity_surften'] * 4.184)
   if INPUT['netcdf'].strip().lower() == 'netcdf':
      infile.change('cntrl','ioutfm', 1)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _MakeGBin(INPUT):
   """ Makes the input file for mmpbsa_py_energy for GB calcs """
   infile = open("_MMPBSA_gb.mdin", 'w')
   infile.write("""GB
igb = %d
gbsa = %d
extdiel = %f
saltcon = %f
surften = %f
rgbmax = 25.0
""" % (INPUT['igb'], INPUT['gbsa'], INPUT['extdiel'], INPUT['saltcon'], INPUT['surften']))

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _MakePBin(INPUT):
   """ Makes the input file for mmpbsa_py_energy for PB calcs """
   infile = open("_MMPBSA_pb.mdin", 'w')
   infile.write("""PB
inp = %d
smoothopt = 0
radiopt = %d
npbopt = 0
solvopt = 1
maxitn = %d
nfocus = 2
bcopt = 5
eneopt = 2
fscale = 8
epsin = %f
epsout = %f
istrng = %f
dprob = %f
iprob = 2.0
accept = 0.001
fillratio = %f
space = %f
cutnb = 0
sprob = %f
cavity_surften = %f
cavity_offset = %f
dbfopt = 1
""" % (INPUT['inp'], INPUT['radiopt'], INPUT['linit'], INPUT['indi'], INPUT['exdi'], INPUT['istrng'],
       INPUT['prbrad'], INPUT['fillratio'], INPUT['scale'], INPUT['prbrad'], INPUT['cavity_surften'],
       INPUT['cavity_offset']))

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _getligmask(comres, recres, ligres):
   """ Get the ligand mask from the complex, receptor, and ligand residue sequences """
   cresnum = len(comres) # this is the number of residues in the complex
   rresnum = len(recres) # this is the number of residues in the receptor
   lresnum = len(ligres) # this is the number of residues in the ligand

   cres = '' # residue from the Complex to compare
   ores = '' # residue from the Other prmtop to compare (rec or lig based on where in the complex it is)

   # The way this is supposed to work: start from the end of the receptor, and insert all ligand residues
   # at each location in the receptor residue sequence. Compare that resulting sequence to the complex
   # sequence.  If it matches, we know the ligand_mask and receptor_mask.

   for x in range(rresnum): # loop through all positions in the receptor prmtop
      z = rresnum - x - 1 # z is the location of the ligand -- start from the end rather than the beginning
      rightpos = True # we will assume it's the RIGHT POSition until proven otherwise

      for y in range(cresnum): # now loop through all complex resnames to compare with rec/lig resnames
         cres = comres[y] # the residue name from the complex...
         if y <= z: # if our complex resname counter is less than our ligand location...
            ores = recres[y] # ores is from receptor at location y
         elif y > z and y - z <= lresnum: # otherwise, if it's greater than z and less than z + num residues in ligand...
            ores = ligres[y-z-1] # residue is in ligand, but subtract out number of receptor residues that aren't in ligand
         else: # otherwise, it's still in receptor, but after the ligand appears
            ores = recres[y-lresnum] # it's in receptor, but subtract number of ligand residues as they're not in receptor

         if cres != ores: # if they're not equal...
            rightpos = False # it's not the right position

      if rightpos: # we've now finished comparison for that particular ligand location, so if it's right position...
         if lresnum == 1: # if the ligand is only one residue:
            return [z+2,':%s' % (z+2)] # ligand residue is z+2 (+1 for index starting at 0, +1 for lig starting AFTER rec ends)
         else: # if ligand is more than 1 residue, we need to specify a range
            return [z+2,':%s-%s' % (z+2,z+lresnum+1)] # same ligand start, but goes through lresnum (+1 for index at 0)

   # the previous loop will not check for ligand in position 1, do that now if we reach this far
   rightpos = True
   for x in range(cresnum):
      cres = comres[x]
      if x < lresnum:
         ores = ligres[x]
      else:
         ores = recres[x-lresnum]

      if cres != ores:
         rightpos = False

   if rightpos:
      if lresnum == 1:
         return [1,':1']
      else:
         return [1,':1-%s' % lresnum]

   else: # if none of these work, we couldn't find the correct masks:
      print >> sys.stderr, "Error: Couldn't predict mask from topology files! If your ligand residues should be sequential"
      print >> sys.stderr, "       in your complex_prmtop, there are likely problems with your prmtops. Otherwise, manually"
      print >> sys.stderr, "       specify ligand_mask and receptor_mask in your input file"
      return -1 # return error code

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _getrecmask(rresnum, ligpos, lresnum):
   """ Get the receptor mask from the ligand position and number of ligand residues """

   recmask = ':' # receptor mask always starts with : for specifying residue
   if ligpos != 1: # take the case when ligand does not start in the first location
      if ligpos == 2: # if ligand is in position 2, receptor does not have a range for first location...
         recmask += '1' # it's just 1
      else: # otherwise it does have a range
         recmask += '1-%s' % (ligpos-1) # it spans 1 through 1 position before the ligand starts
   else: # if ligpos is 1, receptor is the range after ligand ends
      recmask += '%s-%s' % (1+lresnum,rresnum+lresnum)
      return recmask # now we know this is the end of the receptor mask -- return it

   if ligpos == rresnum + 1: # if the ligand position is 1 after receptor res number, it's at the end...
      return recmask # so return the receptor mask now
   else: # otherwise, the ligand is in the middle, and we have to add the second chunk of the receptor mask
      recmask += ':%s' % (ligpos+lresnum)
      if ligpos+lresnum != lresnum+rresnum: # if there's more than just 1 residue in receptor after ligand...
         recmask += '-%s' % (lresnum+rresnum) # specify the range to the end
      return recmask # return the final receptor mask

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _maskToDec(mask):
   """ Turn a mask into a Decomp residue card string """

   # split a mask at each :
   maskcomponents = mask.split(':') # array of :-delimited fields
   totstring = ''  # total string
   for x in range(len(maskcomponents)): # loop through all fields
      maskcomponents[x] = maskcomponents[x].strip() # get rid of whitespace
      if len(maskcomponents[x]) == 0: # skip empty fields
         continue
      if '-' in maskcomponents[x]: # if multiple residues specified
         maskcomponents[x] = maskcomponents[x].replace('-',' ') # replace - with space
      else: # if single residue specified
         maskcomponents[x] += ' ' + maskcomponents[x] # duplicate residue as needed

      totstring += maskcomponents[x] + ' ' # add this to the total string

   return totstring # return the total string

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _getResStrings(comnum, recnum, lignum, selection, ligstart, stability):
   """ Get the strings used in the decomp input files to declare which residues are printed """

   # change commas to semi-colons since that's how the function was originally written
   selection = selection.replace(',',';')
   # get rid of any quotes
   selection = selection.strip("'").strip('"')

   defined_res = 0
   global MAX_RES

   def improperuse(code):
      if code == 1:
         print >> sys.stderr, "Error: Improper definition of print_res!"
      elif code == 2:
         print >> sys.stderr, "Error: You specified a residue multiple times in print_res!"
      return -1

   comsel = '' # string containing the complex selection
   recsel = '' # string containing the receptor selection
   ligsel = '' # string containing the ligand selection

   # parse the selection
   if ligstart == -1 and not stability:
      return ["EDIT","EDIT","EDIT"] # if we didn't guess the mask, we can't go through this
   if selection == "none" and not stability:
      return ["1 %s" % comnum, "1 %s" % recnum, "1 %s" % lignum] # default, all residues

   if stability and selection == 'none':
      return ["1 %s" % comnum, -1, -1]
              
   else:
      items = selection.split(';')  # otherwise, split along delimiter

   try:

      for i in range(len(items)):  # loop through all delimited fields
         items[i] = items[i].strip() # get rid of whitespace
         if len(items[i]) == 0:  # skip over any blank fields
            continue
         defined_res += 1 # we've defined another residue field
         individual = items[i].split('-') # break each field up by -
         comsel += individual[0].strip() + ' ' # the first number is always in complex
         if len(individual) > 1:
            comsel += individual[1].strip() + ' '  # put 2nd number if it's there
         else:
            comsel += individual[0].strip() + ' '  # otherwise repeat the 1st number
         # Now it's time to add the residues to the receptor/ligand selections
         if not stability:
            if len(individual) == 1: # if there is only 1 res in this field
               individual[0] = int(individual[0].strip()) # turn it into an integer
               if individual[0] < 0 or individual[0] > comnum: # catch blatantly illegal choices
                  return improperuse(1) # return error code
               if individual[0] >= ligstart and individual[0] < ligstart + lignum: # if it's in ligand
                  individual[0] = individual[0] - ligstart + 1  # adjust number to its value in ligand prmtop
                  ligsel += str(individual[0]) + ' ' + str(individual[0]) + ' ' # add it to ligand selection
                  
               elif individual[0] < ligstart: # if it comes before ligand in receptor ...
                  recsel += str(individual[0]) + ' ' + str(individual[0]) + ' '  # add it to receptor_selection as-is
               else: # otherwise... 
                  recsel += str(individual[0] - lignum) + ' ' + str(individual[0] - lignum) + ' ' # adjust for absence of ligand
            else: # if it's a field like " num-num "
               individual[0] = int(individual[0].strip()) # make both numbers integers
               individual[1] = int(individual[1].strip())
               if individual[0] > individual[1] or individual[0] <= 0 or individual[1] <= 0 or \
                   individual[0] > comnum or individual[1] > comnum: # catch blatantly illegal choices
                  return improperuse(1) # return error code
               if individual[0] < ligstart and individual[1] < ligstart: # if it starts and ends before start of ligand
                  recsel += str(individual[0]) + ' ' + str(individual[1]) + ' ' # add it to receptor selection
               elif individual[0] < ligstart and \
                   (individual[1] >= ligstart and individual[1] < ligstart + lignum): # starts before ligand ends in ligand
                  recsel += str(individual[0]) + ' ' + str(ligstart - 1) + ' '
                  ligsel += '1 ' + str(individual[1] - ligstart + 1)
               elif (individual[0] >= ligstart and individual[0] < ligstart + lignum) and \
                    (individual[1] >= ligstart and individual[1] < ligstart + lignum): # starts and ends in ligand
                  ligsel += str(individual[0] - ligstart + 1) + ' ' + \
                            str(individual[1] - ligstart + 1) + ' '
               elif (individual[0] >= ligstart and individual[0] < ligstart + lignum) and \
                    (individual[1] >= ligstart + lignum): # starts in ligand ends after ligand (in receptor)
                  recsel += str(ligstart) + ' ' + str(individual[1] - lignum) + ' '
                  ligsel += str(individual[0] - ligstart + 1) + ' ' + str(lignum) + ' '
               elif (individual[0] < ligstart and individual[0] >= ligstart + lignum):
                  # Starts in ligand, ends after ligand
                  recsel += str(individual[0]) + ' ' + str(individual[1] - lignum) + ' '
                  ligsel += '1 %d' % lignum
               else: # starts and ends after ligand
                  recsel += str(individual[0] - lignum) + ' ' + str(individual[1] - lignum) + ' '
   except ValueError: # catch any use of non-integers
      return improperuse(1)
   except IndexError: # catch any other weird errors coming from poorly-defined fields
      return improperuse(1)

   if defined_res > MAX_RES:
      print >> sys.stderr, 'Error: You defined too many residue ranges in print_res! You defined (%s)' % defined_res
      print >> sys.stderr, '       but only %s are allowed! Change MAX_RES in $AMBERHOME/src/mmpbsa_py/utils.py and ' % MAX_RES
      print >> sys.stderr, '       $AMBERHOME/src/sander/rgroup.f and recompile Amber. Or define fewer fields.'
      return -1
   if not stability:
      return [comsel, recsel, ligsel] # return the selections
   else:
      return [comsel, -1, -1]

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _checktrajs(FILES):
   """ Check to make sure that all of the trajectory files that should exist do exist """
   from os.path import exists

   isinerr = 0
   for i in range(len(FILES['mdcrd'])):
      if not exists(FILES['mdcrd'][i]):
         print >> sys.stderr, 'Error: mdcrd file (%s) does not exist!' % FILES['mdcrd'][i]
         isinerr = -1
   for i in range(len(FILES['ligand_mdcrd'])):
      if not exists(FILES['ligand_mdcrd'][i]):
         print >> sys.stderr, 'Error: ligand mdcrd file (%s) does not exist!' % FILES['ligand_mdcrd'][i]
         isinerr = -1
   for i in range(len(FILES['receptor_mdcrd'])):
      if not exists(FILES['receptor_mdcrd'][i]):
         print >> sys.stderr, 'Error: receptor mdcrd file (%s) does not exist!' % FILES['receptor_mdcrd'][i]
         isinerr = -1

   return isinerr

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _getQMMMmasks(atom_list, comres, recres, ligres, ligstart, stability):
   """ Determines the qmmask variable based on the QM_RESIDUES input variable """
   import string
   # Make sure we replace ; by , so we recognize both delimiters, then split along the delimiter
   atom_list = string.replace(atom_list, ';', ',')
   fields = atom_list.split(',')
   return_list = ['','',''] # com mask, rec mask, lig mask
   
   for i in range(len(fields)):
      # get rid of white space and skip over blank fields
      fields[i] = fields[i].strip()
      if len(fields[i]) == 0:
         continue

      # handle residue ranges
      if '-' in fields[i]:
         res_range = fields[i].split('-')
         res_range[0] = int(res_range[0].strip())
         res_range[1] = int(res_range[1].strip())
         if res_range[0] <= 0 or res_range[1] <= 0:
            print >> sys.stderr, 'Error: Improperly defined QM_RESIDUES! Residue numbers must be > 0.'
            return -1
         if res_range[0] > comres or res_range[1] > comres:
            print >> sys.stderr, 'Error: Improperly defined QM_RESIDUES! Asked for a residue that does not exist.'
            return -1
         if res_range[0] >= res_range[1]:
            print >> sys.stderr, 'Error: Improperly defined QM_RESIDUES! Illegal residue range (%d-%d).' % \
                                 (res_range[0], res_range[1])
            return -1
         return_list[0] += ',%d-%d' % (res_range[0], res_range[1])
      # also handle individual residues
      else:
         fields[i] = int(fields[i])
         if fields[i] <= 0:
            print >> sys.stderr, 'Error: Improperly defined QM_RESIDUES! Residue numbers must be > 0.'
            return -1
         if fields[i] > comres:
            print >> sys.stderr, 'Error: Improperly defined QM_RESIDUES! Residue (%d) does not exist!' % fields[i]
            return -1
         return_list[0] += ',%d' % (int(fields[i]))

   # if we're not doing a stability calculation, also determine receptor and ligand masks
   if not stability:
      if ligstart == -1:
         return_list[1] = return_list[2] = 'EDIT'
      else:
         for i in range(len(fields)):
            fields[i] = str(fields[i]) # put back to string in case previous loop turned them to ints
            # skip over blank fields
            if len(fields[i]) == 0:
               continue
            
            # deal with ranges first
            if '-' in fields[i]:
               res_range = fields[i].split('-')
               res_range[0] = int(res_range[0].strip())
               res_range[1] = int(res_range[1].strip())
               # case 1: we have a range, and it's all before the ligand
               if res_range[0] < ligstart and res_range[1] < ligstart:
                  return_list[1] += ',%d-%d' % (res_range[0], res_range[1])
               # case 2: we have a range, and it starts in receptor and ends in ligand
               elif res_range[0] < ligstart and res_range[1] >= ligstart and res_range[1] < ligstart + ligres:
                  return_list[1] += ',%d-%d' % (res_range[0], ligstart-1)
                  if res_range[1] == ligstart:
                     return_list[2] += ',1'
                  else:
                     return_list[2] += ',1-%d' % (res_range[1] - ligstart + 1)
               # case 3: we have a range, and it starts before ligand and ends after ligand
               elif res_range[0] < ligstart and res_range[1] >= ligstart + ligres:
                  return_list[1] += ',%d-%d' % (res_range[0], res_range[1] - ligres)
                  return_list[2] += ',1-%d' % ligres
               # case 4: we have a range, and it starts in ligand and ends after ligand
               elif res_range[0] >= ligstart and res_range[0] < ligstart + ligres and res_range[1] >= ligstart + ligres:
                  if res_range[1] == ligstart + ligres:
                     return_list[1] += ',%d' % ligstart
                  else:
                     return_list[1] += ',%d-%d' % (ligstart, res_range[1] - ligres)
                  if res_range[0] == ligstart + ligres - 1:
                     return_list[2] += ',%d' % ligres
                  else:
                     return_list[2] += ',%d-%d' % (res_range[0] - ligstart + 1, ligres)
               # case 5: we have a range and it starts in ligand and ends in ligand
               elif res_range[0] >= ligstart and res_range[0] < ligstart + ligres and \
                     res_range[1] >= ligstart and res_range[1] < ligstart + ligres:
                  return_list[2] += ',%d-%d' % (res_range[0] - ligstart + 1, res_range[1] - ligstart + 1)
               # case 6: we have a range and it starts and ends after the ligand
               elif res_range[0] >= ligstart + ligres:
                  return_list[1] += ',%d-%d' % (res_range[0] - ligres, res_range[1] - ligres)
            # now deal with individual residues
            else:
               fields[i] = int(fields[i])
               # case 1: it's in the receptor before the ligand
               if fields[i] < ligstart:
                  return_list[1] += ',%d' % fields[i]
               # case 2: it's in the ligand
               elif fields[i] >= ligstart and fields[i] < ligstart + ligres:
                  return_list[2] += ',%d' % (fields[i] - ligstart + 1)
               # case 3: it's in the receptor after the ligand
               elif fields[i] >= ligstart + ligres:
                  return_list[1] += ',%d' % (fields[i] - ligres)

   # Now change the first , back into : to conform to Amber Mask format
   return_list[0] = string.replace(return_list[0],',',':',1)
   if not stability and ligstart != -1:
      return_list[1] = string.replace(return_list[1],',',':',1)
      return_list[2] = string.replace(return_list[2],',',':',1)

   return return_list
