#! PYTHONEXE

from chemistry.amber import *
from MMPBSA_mods import utils

import os, sys

def printusage():
   """ Prints the usage statement and exits """
   prog = os.path.split(sys.argv[0])[1]

   print "This program is useful for creating compatible complex, receptor"
   print "and ligand prmtop files to be used with MMPBSA.py. It will read"
   print "in a complex or solvated topology file and create a complex prmtop,"
   print "receptor prmtop, and ligand prmtop based on the given residues."
   print "It can also be used to generate an frcmod and OFF library file containing"
   print "all of the information found in a given topology file. The Amber force"
   print "field files created are _AnteMMPBSA_.off, _AnteMMPBSA_.frcmod, and _AnteMMPBSA_.leaprc"
   print ""
   print "Usage: %s -p <input prmtop> \\" % prog
   print "          -c <input inpcrd> \\"
   print "         {-cp <complex prmtop>} \\"
   print "         {-rp <receptor prmtop>} \\"
   print "         {-lp <ligand prmtop>} \\"
   print "         {-strip_mask \"strip mask\"} \\"
   print "         {-receptor_mask \"receptor mask\"} \\"
   print ""
   print "-p <input prmtop>       : existing prmtop from which to create other prmtops from (required)"
   print "-c <input inpcrd>       : restart-formatted input coordinates from which to build OFF file. (required)"
   print "-cp <complex_prmtop>    : name of the complex topology file created if strip_mask is defined"
   print "-strip_mask \"strip mask\": ptraj-style mask defining what should be stripped from initial to form complex"
   print "-rp <receptor prmtop>   : name of the receptor prmtop (must also define receptor_mask and ligand prmtop)"
   print "-lp <ligand prmtop>     : name of the ligand prmtop (must also define receptor_mask and receptor prmtop)"
   print "-receptor_mask \"receptor mask\": ptraj-style mask defining what belongs to the receptor from the complex"
   print "                                or initial coordinates if -cp and -strip_mask are omitted"

   sys.exit(0)

def strip_atoms(prmtop, inpcrd, mask, ptraj, name):
   """ Strips a mask from a given prmtop and inpcrd file to create a PDB """
   file = open("_AnteMMPBSA_strip.ptraj",'w')
   file.write('trajin %s\n' % inpcrd)
   file.write('strip "%s"\n' % mask)
   file.write('trajout %s.pdb PDB nobox' % name)
   file.close()

   os.system(ptraj + ' ' + prmtop + ' _AnteMMPBSA_strip.ptraj')
   os.system('mv %s.pdb.1 %s.pdb' % (name, name))

def create_parm(prmtop_name, dummy_inpcrd, leap):
   """ Runs leap to get the desired prmtop """
   file = open("_AnteMMPBSA_createparm.leap", 'w')
   file.write('source _AnteMMPBSA_.leaprc\n')
   file.write('sys = loadPDB %s.pdb\n' % dummy_inpcrd)
   file.write('saveAmberParm sys %s %s.rst7\n' % (prmtop_name, dummy_inpcrd))
   file.write('quit\n')
   file.close()

   os.system(leap + ' -f _AnteMMPBSA_createparm.leap')

# Make sure we have enough arguments
if len(sys.argv) < 5:
   printusage()

# Now set up the default values
complex_prmtop = None
receptor_prmtop = None
ligand_prmtop = None
strip_mask = None
receptor_mask = None
prm_name = None
crd_name = None

# Parse command-line arguments
i = 1
try:
   while i < len(sys.argv):
      if sys.argv[i] == '-p':
         i += 1
         prm_name = sys.argv[i]
      elif sys.argv[i] == '-c':
         i += 1
         crd_name = sys.argv[i]
      elif sys.argv[i] == '-cp':
         i += 1
         complex_prmtop = sys.argv[i]
      elif sys.argv[i] == '-rp':
         i += 1
         receptor_prmtop = sys.argv[i]
      elif sys.argv[i] == '-lp':
         i += 1
         ligand_prmtop = sys.argv[i]
      elif sys.argv[i] == '-strip_mask':
         i += 1
         strip_mask = sys.argv[i].strip('"').strip("'")
      elif sys.argv[i] == '-receptor_mask':
         i += 1
         receptor_mask = sys.argv[i].strip("'").strip('"')
      else:
         print >> sys.stderr, 'Error: Unknown argument %s\n' % sys.argv[i]
         printusage()
      i += 1
except IndexError:
   print >> sys.stderr, 'Error: Command-line error!\n'
   printusage()

# Make sure we have the necessary programs (tleap and ptraj)
leap = utils.which('sleap', searchpath=True)
ptraj = utils.which('ptraj', searchpath=True)

# check that necessary parameters are met
if leap == 'none' or ptraj == 'none':
   print >> sys.stderr, 'Error: tleap and ptraj are needed!'
   sys.exit(1)

if (receptor_mask == None or ligand_prmtop == None or receptor_prmtop == None) and \
   (receptor_mask != None or ligand_prmtop != None or receptor_prmtop != None):
   print >> sys.stderr, 'Error: You must specify receptor_mask, ligand_prmtop, and receptor_prmtop or none of them!\n'
   printusage()

if (strip_mask == None and complex_prmtop != None) or (strip_mask != None and complex_prmtop == None):
   print >> sys.stderr, 'Error: You must specify both strip_mask and complex_prmtop or neither!\n'
   printusage()
   

# Make sure we got a prmtop or inpcrd
if prm_name == None or crd_name == None:
   print >> sys.stderr, 'Error: <input prmtop> and <input inpcrd> are required!\n'
   printusage()

# Load prmtop class and make sure it is valid. Get radius set and dump out force field files
prmtop = readparm.amberParm(prm_name, crd_name)

if not prmtop.valid:
   print >> sys.stderr, 'Error: Invalid prmtop (%s)!\n' % prm_name
   printusage()

try:
   radii = prmtop.parm_data['RADIUS_SET'][0]
except KeyError:
   print >> sys.stderr, 'Warning: RADIUS_SET not found in prmtop. Assuming default PBRadii mbondi'
   radii = 'modified Bondi radii (mbondi)'

prmtop.frcmod('_AnteMMPBSA_.frcmod')
prmtop.writeOFF('_AnteMMPBSA_.off')
leaprc.MakeLeaprc('_AnteMMPBSA_.frcmod', '_AnteMMPBSA_.off', '_AnteMMPBSA_.leaprc', radii, '_AnteMMPBSA_leap.log')

# Now strip the strip_mask if it's applied
if strip_mask != None:
   strip_atoms(prm_name, crd_name, strip_mask.replace(' ',''), ptraj,'_AnteMMPBSA_complex')
   create_parm(complex_prmtop, '_AnteMMPBSA_complex', leap)

if receptor_mask != None:
   if strip_mask != None:
      prm = complex_prmtop
      crd = '_AnteMMPBSA_complex.rst7'
   else:
      prm = prm_name
      crd = crd_name

   strip_atoms(prm, crd, '!' + receptor_mask.replace(' ',''), ptraj,'_AnteMMPBSA_receptor')
   create_parm(receptor_prmtop, '_AnteMMPBSA_receptor', leap)

   strip_atoms(prm, crd, receptor_mask.replace(' ',''), ptraj,'_AnteMMPBSA_ligand')
   create_parm(ligand_prmtop, '_AnteMMPBSA_ligand', leap)


