!<compile=optimized>

!The 3D-RISM-KH software found here is copyright (c) 2010 by Andriy Kovalenko,
!Tyler Luchko and David A. Case.
!
!This program is free software: you can redistribute it and/or modify it
!under the terms of the GNU General Public License as published by the Free
!Software Foundation, either version 3 of the License, or (at your option)
!any later version.
!
!This program is distributed in the hope that it will be useful, but
!WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
!for more details.
!
!You should have received a copy of the GNU General Public License in the
!../../LICENSE file.  If not, see <http://www.gnu.org/licenses/>.
!
!Users of the 3D-RISM capability found here are requested to acknowledge
!use of the software in reports and publications.  Such acknowledgement
!should include the following citations:
!
!1) A. Kovalenko and F. Hirata. J. Chem. Phys., 110:10095-10112  (1999); 
!ibid. 112:10391-10417 (2000).   
!
!2) A. Kovalenko,  in:  Molecular  Theory  of  Solvation,  edited  by  
!F. Hirata  (Kluwer Academic Publishers, Dordrecht, 2003), pp.169-275.  
!
!3) T. Luchko, S. Gusarov, D.R. Roe, C. Simmerling, D.A. Case, J. Tuszynski,
!and  A. Kovalenko, J. Chem. Theory Comput., 6:607-624 (2010). 

#include "../include/dprec.fh"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!General interface between 3D-RISM and SANDER/SFF in the Amber suite.  Except
!!!where noted, all data and subroutines may be called from either SFF or SANDER.
!!!
!!!To make this file interoperable with C while still maintaining the Fortran 95
!!!standard, we must not put subroutines or functions inside of modules. However,
!!!some variables must be available at the global scope either to be accessed
!!!from outside routines (e.g. print statements) or be shared between local
!!!routines (e.g. instances of derived types and parameters for the run).  This 
!!!is accomplished with two modules. RISM_INTERFACE is public to the outside 
!!!world and RISM_DATA should be treated as private to this file.  Subroutines
!!! are then used to manipulate and access the variables in the modules.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Module to hold global data and preserve namespace.  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
module amber_rism_interface
  use rism3d_c
  use rism3d_solv_c
  use rism3d_solu_c
  use fce_c
  use rism_report_c
  type rismprm_t
     sequence
     
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!PRIVATE variables
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     
     !solvcut :: cutoff for rism calculations (separate from SANDER non-bond). 
     _REAL_ :: solvcut
     !buffer   :: buffer distance to the edge of the box for the solvent
     _REAL_ :: buffer
     !grdspc   :: grid spacing for all of the grids
     _REAL_ :: grdspc(3)
     !boxlen   :: box size for 3d-rism.  for PBC calculations, these should generallly be equal
     _REAL_ :: solvbox(3)
     !tolerance :: residual tolerance for solution
     _REAL_ :: tolerance
     !mdiis_del :: 'step size' for MDIIS
     _REAL_ :: mdiis_del
     !fcecut :: FCE cutoff distance
     _REAL_ :: fcecut
     !ng3(3)    :: number of grid points in each dimension
     integer :: ng3(3)
     !irism - use 3d-rism
     integer :: irism
     !closure    :: closure type:
     !             +  1 - KH
     !             +  0 - HNC
     !             +  1 - KH
     !             +  2 - PSE-n
     integer :: closure
     !closureOrder  :: set the order for the given closure (if applicable)
     integer :: closureOrder 
     !asympCorr :: use long range asymptotic corrections for thermodynamic calculations
     logical*4 :: asympCorr
     !gauss_fluct :: use the gaussian fluctuation approximation for the chemical potential
     integer :: gauss_fluct
     !mdiis_nvec       :: number of vectors used for MDIIS (consequently, the number of copies of
     !             CUV we need to keep for MDIIS)
     integer :: mdiis_nvec
     !mdiis_method :: MDIIS implementation to use    
     integer :: mdiis_method
     !maxstep :: maximum number of rism relaxation steps
     integer :: maxstep
     !npropagate :: number of past cuv time steps saves
     integer :: npropagate
     !centering :: center the solute in the solvation box.  
     !           0 - off, 1 - center of mass, 2 - center of geometry
     !           For negative numbers the centering translation is only calculated
     !           the for the first solution and used for subsequent calculations.
     !           This allows the solute to drift in the box.
     integer :: centering
     !zerofrc ::0 - do nothing 
     !          1 - redistribute forces to get zero total force
     integer :: zerofrc
     !apply_rism_force :: if 0, the 3D-RISM solution is calculated but the resulting forces are not
     integer :: apply_rism_force
     !polarDecomp :: do a polar decomposition of the chemical potential
     integer :: polarDecomp
     !rismnrespa :: size of rism multiple timestep
     integer :: rismnrespa
     !fce_stride :: FCE MTS stride length
     integer :: fcestride
     !fcenbasis :: number of FCE basis vectors
     integer :: fcenbasis
     !fcecrd :: FCE coordinate basis type
     integer ::  fcecrd
     !saveprogress  :: save itermediate results every saveprogress interations (0 means no saves)
     integer :: saveprogress
     integer :: ntwrism
     !verbose :: 0 - no ouput
     !                1 - memory allocation and steps for convergence
     !                2 - 1 + convergence progress
     integer :: verbose
     !progress  :: print parameter for relaxation steps every progress iteration (0 means no saves)
     integer :: progress
     
     ! BPR: Make sure the number of INTEGERs is not odd,
     ! to shut up a compiler whinge about misalignment
     ! Note: this should be deleted if ever there are grounds
     ! for a new real INTEGER
     integer :: padding
  end type rismprm_t

  type(rismprm_t), save :: rismprm
  type(rism3d),save :: rism_3d
  type(fce),save :: fce_o
  type(rism3d_solv),save :: solv
  type(rism3d_solu),save :: solu

  integer :: pa_orient, rmsd_orient
  _REAL_:: ratucm(3)

  !egauss_fluct :: the 3D-RISM energy w Gaussian fluctuation correction
  _REAL_ :: egauss_fluct

  integer :: outunit

  character(len=256) :: xvvfile='', guvfile='', huvfile='', cuvfile='',&
       uuvfile='', asympfile=''

  integer :: mpirank=0,mpisize=1,mpicomm=0

  private :: rism_mpi_bcast

end module amber_rism_interface

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Performs all of the initialization required for 3D-RISM calculations.  This 
!!!includes reading input files and allocating solvent and solute objects
!!!Reads in and initializes the solvent, solute, FCE and 3D-RISM
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine rism_init(comm)
  use amber_rism_interface
  use safemem
#ifdef RISM_CRDINTERP
  use fce_c
#endif /*RISM_CRDINTERP*/
  implicit none
#ifdef MPI
  include 'mpif.h'
#endif /*MPI*/  
  integer, intent(in) :: comm

  integer :: err

#ifdef SANDER
  call rism_report_setMUnit(6)
  call rism_report_setWUnit(6)
  call rism_report_setEUnit(6)
#endif /*SANDER*/

#ifdef MPI
  mpicomm = comm
  if(mpicomm == MPI_COMM_NULL)&
       call rism_report_error("RISM3D interface: received NULL MPI communicator")
  call mpi_comm_rank(mpicomm,mpirank,err)
  if(err /=0) call rism_report_error&
       ("(a,i8)","RISM3D interface: could not get MPI rank for communicator ",mpicomm)
  call mpi_comm_size(mpicomm,mpisize,err)
  if(err /=0) call rism_report_error&
       ("(a,i8)","RISM3D interface: could not get MPI size for communicator ",mpicomm)
  call rism_mpi_bcast(mpirank,mpisize,mpicomm)
#endif /*MPI*/

#ifdef RISM_CRDINTERP
  call fce_new(fce_o, solu%natom, rismprm%fcenbasis, rismprm%fcecrd, rismprm%fcecut,mpicomm)
#endif /*RISM_CRDINTERP*/

  if(rismprm%buffer >0)then
     call rism3d_new(rism_3d,solu,solv,rismprm%centering, rismprm%npropagate,&
          rismprm%closure,rismprm%closureOrder,rismprm%solvcut,&
          rismprm%mdiis_nvec,rismprm%mdiis_del,rismprm%mdiis_method,&
          o_buffer=rismprm%buffer, o_grdspc=rismprm%grdspc, o_mpicomm=mpicomm)
  else
     call rism3d_new(rism_3d,solu,solv,rismprm%centering, rismprm%npropagate,&
          rismprm%closure,rismprm%closureOrder,rismprm%solvcut,&
          rismprm%mdiis_nvec,rismprm%mdiis_del,rismprm%mdiis_method,&
          o_boxlen=rismprm%solvbox, o_ng3=rismprm%ng3, o_mpicomm=mpicomm)
  end if
  call rism3d_setverbosity(rism_3d,rismprm%verbose)

  !Free up a bit of memory
  call rism3d_solv_destroy(solv)
  call rism3d_solu_destroy(solu)
  
  call flush(outunit)
end subroutine rism_init

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Outputs values for all 3D-RISM parameters
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine rism_setparam(&
#ifdef SANDER
     !     xvvfile, mdin,&
  mdin,&
#else /*SANDER*/
  userData,xvvlen,xvvchar,guvlen,guvchar,huvlen,huvchar,cuvlen,cuvchar,&
       uuvlen,uuvchar,asymplen,asympchar,&
#endif /*SANDER*/
  comm,&
  igb, natom, ntypes, charge, mass, cn1, cn2, iac, ico)
  use amber_rism_interface
  
  implicit none
#ifdef MPI
  include 'mpif.h'
#endif /*MPI*/
  integer, intent(inout) :: igb
  integer, intent(in) ::natom, ntypes,iac(ntypes**2),ico(natom)
  _REAL_, intent(in) :: charge(natom),mass(natom),cn1(ntypes*(ntypes+1)/2),&
       cn2(ntypes*(ntypes+1)/2)
#ifdef SANDER
  !  character(*), intent(in) :: xvvfile,mdin
  character(*), intent(in) :: mdin
  integer :: mdin_unit=55
#else /*SANDER*/
  type(rismprm_t), intent(in) :: userData;
  integer, intent(in) :: xvvlen, guvlen, huvlen, cuvlen, uuvlen, asymplen
  integer(kind=1), intent(in) :: xvvchar(xvvlen+1),guvchar(guvlen+1),&
       huvchar(huvlen+1),cuvchar(cuvlen+1),uuvchar(uuvlen+1),asympchar(asymplen+1)
  !  character(xvvlen):: xvvfile
#endif /*SANDER*/
  integer, intent(in) :: comm

  integer :: i,stat, err
  logical :: op
  integer ::  un

#ifdef MPI
  mpicomm = comm
  if(mpicomm == MPI_COMM_NULL)&
       call rism_report_error("RISM3D interface: received NULL MPI communicator")
  call mpi_comm_rank(mpicomm,mpirank,err)
  if(err /=0) call rism_report_error&
       ("(a,i8)","RISM3D interface: could not get MPI rank for communicator ",mpicomm)
  call mpi_comm_size(mpicomm,mpisize,err)
  if(err /=0) call rism_report_error&
       ("(a,i8)","RISM3D interface: could not get MPI size for communicator ",mpicomm)
#endif /*MPI*/  
 if(mpirank /= 0) return

#if !defined(SANDER)
 call cstr2fstr(xvvfile,xvvchar,xvvlen)
 call cstr2fstr(guvfile,guvchar,guvlen)
 call cstr2fstr(huvfile,huvchar,huvlen)
 call cstr2fstr(cuvfile,cuvchar,cuvlen)
 call cstr2fstr(uuvfile,uuvchar,uuvlen)
 call cstr2fstr(asympfile,asympchar,asymplen)
#endif /*!SANDER*/
 
#ifdef SANDER
 if(rismprm%irism == 0) return
#else /*SANDER*/
 if(userData%irism == 0) return
#endif /*SANDER*/
 outunit = rism_report_getMUnit()
 call defaults()
 
#ifdef SANDER
 inquire(file=mdin,opened=op, number=un)
 if(op) mdin_unit=un
 open(unit=mdin_unit,file=mdin,status='OLD', form='FORMATTED',iostat=stat)
 if(stat/=0)then
    call rism_report_error('(a,i4)', "opening "//trim(mdin)//"failed. IOSTAT=",stat)
 end if
 call read_namelist(mdin_unit)
 if(.not.op) close(unit=mdin_unit)
#else /*SANDER*/
 call update_param(userData)
#endif /*SANDER*/
 
 call sanity_check(igb)
 
 call rism3d_solv_new(solv,xvvfile)
 
 call rism3d_solu_new_sander(solu,natom,ntypes,iac,&
      ico,charge,cn1,cn2,mass,solv%temperature)

#ifdef SANDER
  if(rismprm%irism >=1)then
     write(outunit,'(a)') "3D-RISM:"
     if(rismprm%irism < 1) then
        write(outunit,'(5x,a,i10)') 'irism   =',rismprm%irism
     elseif(rismprm%irism == 1) then
        write(outunit,'(5x,3(a10,"=",i10))') 'closure        ', rismprm%closure, ', gauss_fluct', rismprm%mdiis_nvec
        write(outunit,'(5x,3(a10,"="f10.5))') 'solvcut     ', rismprm%solvcut,&
             ', buffer    ', rismprm%buffer
        write(outunit,'(5x,a10,"=",3(f10.5,1x))') 'grd_spc        ', rismprm%grdspc
        write(outunit,'(5x,a10,"=",3(i10,1x))') 'ng3              ', rismprm%ng3
        write(outunit,'(5x,a10,"=",3(f10.5,1x))') 'solvbox        ', rismprm%solvbox
        write(outunit,'(5x,a10,"=",d10.2,a10,"=",f10.5,a10,"=",i10)')  'tolerance          ', rismprm%tolerance, &
             ', mdiis_del       ', rismprm%mdiis_del,', mdiis_nvec          ',rismprm%mdiis_nvec
        write(outunit,'(5x,3(a10,"=",i10))') 'mdiis_method      ', rismprm%mdiis_method, &
             ', maxstep      ', rismprm%maxstep, ', npropagate    ',rismprm%npropagate
        write(outunit,'(5x,3(a10,"=",i10))') 'centering      ', rismprm%centering, &
             ', zerofrc       ', rismprm%zerofrc
        write(outunit,'(5x,3(a10,"=",i10))') 'apply_rism_force      ', rismprm%apply_rism_force, &
             ', pa_orient     ', pa_orient, ', rmsd_orient     ', rmsd_orient
        write(outunit,'(5x,2(a10,"=",i10),a10,"=",f10.5)') 'rismnrespa      ', rismprm%rismnrespa, &
             ', fcestride     ', rismprm%fcestride, ', fcecut      ',rismprm%fcecut
        write(outunit,'(5x,3(a10,"=",i10))') 'fcenbasis      ', rismprm%fcenbasis, &
             ', fcecrd      ', rismprm%fcecrd
        write(outunit,'(5x,3(a10,"=",i10))') 'saveprogress           ', rismprm%saveprogress, &
             ', ntwrism      ', rismprm%ntwrism, ', verbose      ', rismprm%verbose
        write(outunit,'(5x,3(a10,"=",i10))') 'progress              ', rismprm%progress
     endif
     call flush(outunit)
  end if
#endif /*SANDER*/
end subroutine rism_setparam

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Broadcasts initialization information about the system from the master to all
!!!the other processes.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#ifdef MPI
subroutine rism_mpi_bcast(rank,size,comm)
  use amber_rism_interface
  implicit none
  include 'mpif.h'
  integer, intent(in) :: rank, size, comm
  integer :: err

#ifdef MPI
  mpirank = rank
  mpisize = size
  mpicomm = comm
#endif /*MPI*/  

  call mpi_bcast(rismprm%irism,1,mpi_integer,0,mpicomm,err)
  if(err /=0) call rism_report_error&
       ("RISM3D interface: could not broadcast IRISM")
  if (rismprm%irism == 1) then
     !broadcast the entire rismprm object.  Not everything actually needs
     !to be transferred but there are not that many exceptions.  This is
     !could be done by passing creating and passing an MPI derived type.
     !However, this is only being done once so it is not really worth it.
     call mpi_bcast(rismprm%solvcut,1,mpi_double_precision,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast SOLVCUT")
     call mpi_bcast(rismprm%buffer,1,mpi_double_precision,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast BUFFER")
     call mpi_bcast(rismprm%grdspc,3,mpi_double_precision,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast GRDSPC")
     call mpi_bcast(rismprm%solvbox,3,mpi_double_precision,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast SOLVBOX")
     call mpi_bcast(rismprm%tolerance,1,mpi_double_precision,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast TOLERANCE")
     call mpi_bcast(rismprm%mdiis_del,1,mpi_double_precision,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast MDIIS_DEL")
     call mpi_bcast(rismprm%fcecut,1,mpi_double_precision,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast FCECUT")
     call mpi_bcast(rismprm%ng3,3,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast NG3")
     call mpi_bcast(rismprm%closure,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast CLOSURE")
     call mpi_bcast(rismprm%closureOrder,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast CLOSUREORDER")
     call mpi_bcast(rismprm%asympCorr,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast ASYMPCORR")
     call mpi_bcast(rismprm%gauss_fluct,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast GAUSS_FLUCT")
     call mpi_bcast(rismprm%maxstep,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast MAXSTEP")
     call mpi_bcast(rismprm%npropagate,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast NPROPAGATE")
     call mpi_bcast(rismprm%centering,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast CENTERING")
     call mpi_bcast(rismprm%zerofrc,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast ZEROFRC")
     call mpi_bcast(rismprm%apply_rism_force,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast APPLY_RISM_FORCE")
     call mpi_bcast(rismprm%rismnrespa,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast RISMNRESPA")
     call mpi_bcast(rismprm%fcestride,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast FCESTRIDE")
     call mpi_bcast(rismprm%fcenbasis,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast FCENBASIS")
     call mpi_bcast(rismprm%fcecrd,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast FCECRD")
     call mpi_bcast(rismprm%saveprogress,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast SAVEPROGRESS")
     call mpi_bcast(rismprm%ntwrism,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast NTWRISM")
     call mpi_bcast(rismprm%verbose,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast VERBOSE")
     call mpi_bcast(rismprm%progress,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast PROGRESS")

     !these are not being used currently.
     call mpi_bcast(pa_orient,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast PA_ORIENT")
     call mpi_bcast(rmsd_orient,1,mpi_integer,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast RMSD_ORIENT")

     !I/O
     !special output files that all nodes write to
     call mpi_bcast(guvfile,len(guvfile),mpi_character,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast GUVFILE")
     call mpi_bcast(huvfile,len(huvfile),mpi_character,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast HUVFILE")
     call mpi_bcast(cuvfile,len(cuvfile),mpi_character,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast CUVFILE")
     call mpi_bcast(uuvfile,len(uuvfile),mpi_character,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast UUVFILE")
     call mpi_bcast(asympfile,len(asympfile),mpi_character,0,mpicomm,err)
     if(err /=0) call rism_report_error&
          ("RISM3D interface: could not broadcast ASYMPFILE")

  end if
end subroutine rism_mpi_bcast
#endif /*MPI*/

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Driver routine for 3D-RISM.
!!!Calculates the 3D-RISM solution, energy and forces.  Determines if an interpolation 
!!!step can be used.  If so, the interpolation code is called, otherwise a full 
!!!3D-RISM calculation is performed.  Energies are only valid for full 3D-RISM 
!!!calculations
!!!IN:
!!!    ratu_md    :: atom positions for solute.
!!!    frc        :: pre-3D-RISM force.  The 3D-RISM forces are added to this
!!!    epol       :: polarization energy, excess chemical potential
!!!    irespa     :: respa iteration
!!!MODIFIED:
!!!    frc
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine rism_force (ratu_md,frc, epol, irespa)
  use amber_rism_interface
  
  use constants, only : KB
  use rism3d_c, only : rism3d_solve
  use rism_util, only: corr_drift, alignorient, translate, calc_cm, calc_rotation
  implicit none
#include "def_time.h" 

#ifdef MPI
  include 'mpif.h'
#endif /*MPI*/

  integer,intent(in) :: irespa
  _REAL_,intent(in) :: ratu_md(3,rism_3d%solu%natom)
  _REAL_,intent(inout) :: frc(3,rism_3d%solu%natom)
  _REAL_,intent(out) :: epol

  !ff :: 3D-RISM forces
  _REAL_ :: ff(3,rism_3d%solu%natom),ff_temp(3,rism_3d%solu%natom),qback(4)
  _REAL_ :: q1(4),q2(4),q3(4),v1(3),v2(3), angle, dir(3), rmsd
  
  !ratu_fce :: coordinate oriented especially for FCE
  _REAL_ :: ratu_fce(3,rism_3d%solu%natom)

  integer :: i, iatu
  integer :: err
  _REAL_ :: mpi_temp, pmv, ene(rism_3d%solv%natom), mpiv(rism_3d%solv%natom)

  epol = 0
  ff=0
  !
  !Test for interpolation, minimum samples and if this is an interpolation step
  !
  if(mod(irespa,rismprm%rismnrespa) /= 0) then !test for RESPA
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!No forces this steps. DO NOTHING!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#if defined(RISM_CRDINTERP)
  elseif(rismprm%fcestride >0 .and. fce_o%nsample >= fce_o%nbasis &
       .and. mod(irespa,rismprm%rismnrespa*rismprm%fcestride) /= 0)then
!!!!!!!!!!!!!!!!
!!!FCE forces!!!
!!!!!!!!!!!!!!!!
     if(rismprm%verbose>=2) call rism_report_message("|LINEAR PROJECTION PREDICT!!!")

     call timer_start(TIME_REORIENT)
     ratu_fce = ratu_md
     call orient(rism_3d%solu,ratu_fce,rism_3d%nsolution)
     call timer_stop(TIME_REORIENT)
     !linproj predict
     call timer_start(TIME_CRDINTERP)
     if(rismprm%apply_rism_force==1)&
          call fce_force(fce_o,ff,ratu_fce)
     call timer_stop(TIME_CRDINTERP)

!!$     call timer_start(TIME_REORIENT)
!!$     call unorient(rism_3d%solu,ratu_md)
!!$     call timer_stop(TIME_REORIENT)

     if(rismprm%zerofrc==1)then
#ifdef MPI
        call corr_drift(ff,rism_3d%solu%mass,rism_3d%solu%natom,&
             mpirank,mpisize,mpicomm)
#else
        call corr_drift(ff,rism_3d%solu%mass,rism_3d%solu%natom)
#endif /*MPI*/
     end if
     if(rismprm%fcestride >1)then
        if(rismprm%verbose>=2) call rism_report_message("|IMPULSE FORCE - INTERP!!!")
        ff=rismprm%rismnrespa*ff
     end if

#endif /*RISM_CRDINTERP*/
  else
     if(rismprm%verbose>=2) call rism_report_message("|FULL RISM!!!")
!!!!!!!!!!!!!!!!!!!!!!!!
!!!Full RISM SOLUTION!!!
!!!!!!!!!!!!!!!!!!!!!!!!
     call rism3d_setCoord(rism_3d,ratu_md)
     call rism3d_solve(rism_3d,rismprm%saveprogress,rismprm%progress,rismprm%maxstep,rismprm%tolerance)
#ifdef SANDER
     !ugly, ugly hack.  SANDER runs force on the first frame twice.  This messes up the 
     !solution propagation.  Here we set the solution counter back one to ignore one of
     !the duplicate solutions
     if(irespa == 1)then
        rism_3d%nsolution = 1
     end if
#else /*SANDER*/
#endif /*SANDER*/
     if(rismprm%apply_rism_force==1)then
        call rism3d_force(rism_3d,ff)
        if(rismprm%zerofrc==1)then
#ifdef MPI
           call corr_drift(ff,rism_3d%solu%mass,rism_3d%solu%natom,&
                mpirank,mpisize,mpicomm)
#else
           call corr_drift(ff,rism_3d%solu%mass,rism_3d%solu%natom)
#endif /*MPI*/
        end if
        !convert to [kcal/mol/A]
        ff = ff*KB*rism_3d%solv%temperature
     end if
     !get the excess chemical potential
     call timer_start(TIME_EXCHEM)
     epol = rism3d_exchem_tot(rism_3d,rismprm%asympCorr)*KB*rism_3d%solv%temperature

     !convert to [kcal/mol]
     if(rismprm%gauss_fluct /= 0)then
        egauss_fluct = rism3d_exchemGF_tot(rism_3d,rismprm%asympCorr)*KB*rism_3d%solv%temperature
#if defined(MPI)
        call mpi_reduce(egauss_fluct, mpi_temp, 1, MPI_DOUBLE_PRECISION, &
             MPI_SUM, 0, mpicomm, err)
        if(err /=0) call rism_report_error&
             ("RISM3D interface: could not reduce EGAUSS_FLUCT")
        egauss_fluct = mpi_temp
#endif /*defined(MPI)*/
     end if
     call timer_stop(TIME_EXCHEM)


#ifdef RISM_CRDINTERP
     if(rismprm%fcestride >0 .and. rismprm%apply_rism_force==1)then
        call timer_start(TIME_REORIENT)
        ratu_fce = ratu_md
        call orient(rism_3d%solu,ratu_fce,rism_3d%nsolution)
        call timer_stop(TIME_REORIENT)
        call timer_start(TIME_SAVECRDINTERP)
        call fce_update(fce_o,ff,ratu_fce)
        call timer_stop(TIME_SAVECRDINTERP)
!!$        call timer_start(TIME_REORIENT)
!!$        call unorient(rism_3d%solu,ratu_fce)
!!$        call timer_stop(TIME_REORIENT)
     endif
#endif /*RISM_CRDINTERP*/

     !         if(rismnrespa >1 .and. (.not. interpcrd>0 .or. .not. nsample >= fcenbasis))then
     if(rismprm%rismnrespa >1)then
        if(rismprm%verbose>=2) call rism_report_message("|IMPULSE FORCE!!!")
        ff=rismprm%rismnrespa*ff
     end if

  end if
  if(rismprm%apply_rism_force==1) frc = frc+ff
  call flush(outunit)
end subroutine rism_force

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Prints results of various thermodynamics calculations.  Quantities
!!!printed are: Solute internal energy, solvation free energy,
!!!solvation free energy with GF correction, excess solvation free
!!!energy, PMV, excess number of each solvent type.
!!!IN:
!!!   description : (optional) (default=.false.) Print a short description of 
!!!                  each of the output values as they appear in the output 
!!!                  instead of the actual values.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#ifdef SANDER
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!   solu_pot : (optional) the potential energy of the solute
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine rism_thermo(description, solu_pot)
  implicit none
  logical, optional, intent(in) :: description
  _REAL_,optional, intent(in) :: solu_pot

  character(len=10) :: titleFmt="(8(a10))"
  
end subroutine rism_thermo
#else /*SANDER*/
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!The format for NAB programs will consist of one line for each catgory of 
!!!data: energy, volume and excess.  Each is preceeded with "rism_[cat]:" where
!!!"cat" is a abreviation of the category.
!!!IN:
!!!   description : if .true., output a description of the table that will
!!!                  be printed but do not print out any values
!!!   polarDecomp : if .true. repeat the most recent 
!!!                 calculation with with solute charges turned off.  Report
!!!                 the usual quanities but also decompose the chemical 
!!!                 potential into polar and non-polar terms
!!!   soluPot     : total and component potential energy of the solute.
!!!                 Expected order is: total, LJ, elec, bond, angle, dihedral,
!!!                 H-bond, LJ-14, elec-14, restraints, 3D-RISM
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine rism_thermo(description,polarDecomp,soluPot)
  use amber_rism_interface
  use constants, only : kb,COULOMB_CONST_E
  implicit none
#ifdef MPI
  include "mpif.h"
#endif /*MPI*/
  logical*4, intent(in) :: description
  _REAL_, intent(in) :: soluPot(11)
  logical*4, intent(in) :: polarDecomp
  character(len=16) :: catFmt="(a12)", catbarFMT="(a25)", titleFmt="(a17)", valFmt='(1p,1x,e16.8e3)' 
  !mpi_buffer : all thermodynamic results are stored in a single array so the necessary reductions are 
  !         done in a single communication.  Order and size: exchem(solv%natom), exchemGF(solv%natom)
  !         exEne(solv%natom), excessNum(solv%natom), PMV
  !apol_mpi_buffer : apolar chemical potentials for decomposition
  _REAL_, target :: mpi_buffer(4*rism_3d%solv%natom+1), &
       apol_mpi_buffer(2*rism_3d%solv%natom)
  _REAL_, pointer :: exchem(:), exchemGF(:), exEne(:), PMV, exNum(:),&
       apol_exchem(:), apol_exchemGF(:)
  integer :: iv, il, iu, err
#if !defined( USE_MPI_IN_PLACE)
  _REAL_ :: tmpi_buffer(4*rism_3d%solv%natom+1)
#endif

  if(description)then
     if(mpirank==0)then
        write(outunit,'(a)') "|3D-RISM thermodynamic data key"
        call flush(outunit)
        write(outunit,catbarFmt,advance='no') '|'//'solute_epot [kcal/mol] :'
        write(outunit,'(12'//titleFmt//')') 'Total', 'LJ', 'Coulomb', 'Bond', &
             'Angle', 'Dihedral', 'H-Bond', 'LJ-14', 'Coulomb-14', 'Restraints', '3D-RISM'

        write(outunit,catbarFmt,advance='no') '|'//'rism_exchem [kcal/mol] :'
        write(outunit,titleFmt,advance='no') 'Total'
        do iv=1, rism_3d%solv%natom
           write(outunit,titleFmt,advance='no') 'ExChem_'//trim(rism_3d%solv%atomname(iv))
        end do
        write(outunit,'(a)')

        write(outunit,catbarFmt,advance='no') '|'//'rism_exchGF [kcal/mol] :'
        write(outunit,titleFmt,advance='no') 'Total'
        do iv=1, rism_3d%solv%natom
           write(outunit,titleFmt,advance='no') 'ExChem_GF_'//trim(rism_3d%solv%atomname(iv))
        end do
        write(outunit,'(a)')

        write(outunit,catbarFmt,advance='no') '|'//'rism_exEnUV [kcal/mol] :'
        write(outunit,titleFmt,advance='no') 'Total'
        do iv=1, rism_3d%solv%natom
           write(outunit,titleFmt,advance='no') 'Energy_'//trim(rism_3d%solv%atomname(iv))
        end do
        write(outunit,'(a)')

        write(outunit,catbarFmt,advance='no') '|'//'rism_volume [A^ 3]     :'
        write(outunit,titleFmt,advance='no') 'PMV'
        write(outunit,'(a)')

        write(outunit,catbarFmt,advance='no') '|'//'rism_exNumb [#]        :'
        write(outunit,titleFmt,advance='no') ""
        do iv=1, rism_3d%solv%natom
           write(outunit,titleFmt,advance='no') 'ExNum_'//trim(rism_3d%solv%atomname(iv))
        end do
        write(outunit,'(a)')

        write(outunit,catbarFmt,advance='no') '|'//'rism_exChrg [e]        :'
        write(outunit,titleFmt,advance='no') 'Total'
        do iv=1, rism_3d%solv%natom
           write(outunit,titleFmt,advance='no') 'ExChg_'//trim(rism_3d%solv%atomname(iv))
        end do
        write(outunit,'(a)')

        if(polarDecomp)then
           write(outunit,catbarFmt,advance='no') '|'//'rism_polar_ [kcal/mol] :'
           write(outunit,titleFmt,advance='no') 'Total'
           do iv=1, rism_3d%solv%natom
              write(outunit,titleFmt,advance='no') 'polar_'//trim(rism_3d%solv%atomname(iv))
           end do
           write(outunit,'(a)')
           
           write(outunit,catbarFmt,advance='no') '|'//'rism_apolar [kcal/mol] :'
           write(outunit,titleFmt,advance='no') 'Total'
           do iv=1, rism_3d%solv%natom
              write(outunit,titleFmt,advance='no') 'apolar_'//trim(rism_3d%solv%atomname(iv))
           end do
           write(outunit,'(a)')
           
           write(outunit,catbarFmt,advance='no') '|'//'rism_polGF_ [kcal/mol] :'
           write(outunit,titleFmt,advance='no') 'Total'
           do iv=1, rism_3d%solv%natom
              write(outunit,titleFmt,advance='no') 'polarGF_'//trim(rism_3d%solv%atomname(iv))
           end do
           write(outunit,'(a)')
           
           write(outunit,catbarFmt,advance='no') '|'//'rism_apolGF [kcal/mol] :'
           write(outunit,titleFmt,advance='no') 'Total'
           do iv=1, rism_3d%solv%natom
              write(outunit,titleFmt,advance='no') 'apolarGF_'//trim(rism_3d%solv%atomname(iv))
           end do
           write(outunit,'(a)')
        end if
     endif
  else
     !setup memory space
     exchem => mpi_buffer(1:rism_3d%solv%natom) 
     exchemGF => mpi_buffer(rism_3d%solv%natom+1:2*rism_3d%solv%natom) 
     exEne => mpi_buffer(2*rism_3d%solv%natom+1:3*rism_3d%solv%natom) 
     exNum => mpi_buffer(3*rism_3d%solv%natom+1:4*rism_3d%solv%natom) 
     pmv => mpi_buffer(4*rism_3d%solv%natom+1)

     !calculate thermodynamics
     exchem = rism3d_exchem(rism_3d,rismprm%asympCorr)*KB*rism_3d%solv%temperature
     exchemGF = rism3d_exchemGF(rism_3d,rismprm%asympCorr)*KB*rism_3d%solv%temperature
     exEne = rism3d_exEne(rism_3d)*KB*rism_3d%solv%temperature
     pmv= rism3d_pmv(rism_3d)
     exNum = rism3d_exNum(rism_3d,rismprm%asympCorr)

     !parallel communication
#ifdef MPI
#  ifdef USE_MPI_IN_PLACE
     if(mpirank==0)then
        call MPI_REDUCE(MPI_IN_PLACE, mpi_buffer, ubound(mpi_buffer,1),MPI_DOUBLE_PRECISION,&
             MPI_SUM,0,mpicomm,err)
     else
        call MPI_REDUCE(mpi_buffer, mpi_buffer, ubound(mpi_buffer,1),MPI_DOUBLE_PRECISION,&
             MPI_SUM,0,mpicomm,err)
     end if
#  else /*USE_MPI_IN_PLACE*/
     call MPI_REDUCE(mpi_buffer, tmpi_buffer, ubound(mpi_buffer,1),MPI_DOUBLE_PRECISION,&
          MPI_SUM,0,mpicomm,err)
     mpi_buffer = tmpi_buffer
#  endif /*USE_MPI_IN_PLACE*/
     if(err/=0) call rism_report_error("RISM_THERMO: MPI_REDUCE failed.")
#endif
     !By doing the the second RISM calculation here output from the
     !RISM routines does not break up the thermodynamic output below
     if(polarDecomp)then
        !redo calculation with charges off
        call rism3d_unsetCharges(rism_3d)
        call rism3d_solve(rism_3d,rismprm%saveprogress,rismprm%progress,rismprm%maxstep,rismprm%tolerance)
        call rism3d_resetCharges(rism_3d)

        !setup memory space
        apol_exchem => apol_mpi_buffer(1:rism_3d%solv%natom) 
        apol_exchemGF => apol_mpi_buffer(rism_3d%solv%natom+1:2*rism_3d%solv%natom) 
        
        !calculate thermodynamics
        apol_exchem = rism3d_exchem(rism_3d,rismprm%asympCorr)*KB*rism_3d%solv%temperature
        apol_exchemGF = rism3d_exchemGF(rism_3d,rismprm%asympCorr)*KB*rism_3d%solv%temperature
        
        !parallel communication
#ifdef MPI
#  ifdef USE_MPI_IN_PLACE
        if(mpirank==0)then
           call MPI_REDUCE(MPI_IN_PLACE, apol_mpi_buffer, ubound(apol_mpi_buffer,1),MPI_DOUBLE_PRECISION,&
                MPI_SUM,0,mpicomm,err)
        else
           call MPI_REDUCE(apol_mpi_buffer, apol_mpi_buffer, ubound(apol_mpi_buffer,1),MPI_DOUBLE_PRECISION,&
                MPI_SUM,0,mpicomm,err)
        end if
#  else /*USE_MPI_IN_PLACE*/
        call MPI_REDUCE(apol_mpi_buffer, tmpi_buffer, ubound(apol_mpi_buffer,1),MPI_DOUBLE_PRECISION,&
             MPI_SUM,0,mpicomm,err)
        apol_mpi_buffer = tmpi_buffer(:ubound(apol_mpi_buffer,1))
#  endif /*USE_MPI_IN_PLACE*/
        if(err/=0) call rism_report_error("RISM_THERMO: MPI_REDUCE failed.")
#endif
     end if

     !output
     if(mpirank==0)then
        write(outunit,catFmt,advance='no') 'solute_epot:'
        write(outunit,'(12'//trim(valFmt)//')') soluPot

        write(outunit,catFmt,advance='no') 'rism_exchem:'
        write(outunit,valFmt,advance='no') sum(exchem)
        do iv=1, rism_3d%solv%natom
           write(outunit,valFmt,advance='no') exchem(iv)
        end do
        write(outunit,'(a)')

        write(outunit,catFmt,advance='no') 'rism_exchGF:'
        write(outunit,valFmt,advance='no') sum(exchemGF)
        do iv=1, rism_3d%solv%natom
           write(outunit,valFmt,advance='no') exchemGF(iv)
        end do
        write(outunit,'(a)')

        write(outunit,catFmt,advance='no') 'rism_exEnUV:'
        write(outunit,valFmt,advance='no') sum(exEne)
        do iv=1, rism_3d%solv%natom
           write(outunit,valFmt,advance='no') exEne(iv)
        end do
        write(outunit,'(a)')

        write(outunit,catFmt,advance='no') 'rism_volume:'
        write(outunit,valFmt,advance='no') pmv
        write(outunit,'(a)')

        write(outunit,catFmt,advance='no') 'rism_exNumb:'
        write(outunit,titleFmt,advance='no') ""
        do iv=1, rism_3d%solv%natom
           write(outunit,valFmt,advance='no') exNum(iv)
        end do
        write(outunit,'(a)')

        write(outunit,catFmt,advance='no') 'rism_exChrg:'
        write(outunit,valFmt,advance='no') sum(exNum*rism_3d%solv%charge)&
             *sqrt((KB *rism_3d%solv%temperature)/COULOMB_CONST_E)
        do iv=1, rism_3d%solv%natom
           write(outunit,valFmt,advance='no') exNum(iv)*rism_3d%solv%charge(iv)&
                *sqrt((KB *rism_3d%solv%temperature)/COULOMB_CONST_E)
        end do
        write(outunit,'(a)')
        !output
        if(polarDecomp)then
           write(outunit,catFmt,advance='no') 'rism_polar_:'
           write(outunit,valFmt,advance='no') sum(exchem-apol_exchem)
           do iv=1, rism_3d%solv%natom
              write(outunit,valFmt,advance='no') exchem(iv)-apol_exchem(iv)
           end do
           write(outunit,'(a)')
           
           write(outunit,catFmt,advance='no') 'rism_apolar:'
           write(outunit,valFmt,advance='no') sum(apol_exchem)
           do iv=1, rism_3d%solv%natom
              write(outunit,valFmt,advance='no') apol_exchem(iv)
           end do
           write(outunit,'(a)')
           
           write(outunit,catFmt,advance='no') 'rism_polGF_:'
           write(outunit,valFmt,advance='no') sum(exchemGF-apol_exchemGF)
           do iv=1, rism_3d%solv%natom
              write(outunit,valFmt,advance='no') exchemGF(iv)-apol_exchemGF(iv)
           end do
           write(outunit,'(a)')
           
           write(outunit,catFmt,advance='no') 'rism_apolGF:'
           write(outunit,valFmt,advance='no') sum(apol_exchemGF)
           do iv=1, rism_3d%solv%natom
              write(outunit,valFmt,advance='no') apol_exchemGF(iv)
           end do
           write(outunit,'(a)')
        end if
     end if
  end if
  call flush(outunit)
end subroutine rism_thermo
#endif /*SANDER*/
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Finalizes all of the 3D-RISM objects and frees memory.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine rism_finalize()
  use amber_rism_interface
  
  use fce_c, only : fce_destroy
  use safemem
  implicit none
  integer :: err
  integer*8 :: memstats(10)
  if(rismprm%irism==1)then
     call rism3d_destroy(rism_3d)
     call fce_destroy(fce_o)
!!$     memstats = memStatus()
!!$     write(0,'(a)') "3D-RISM memory allocation summary"
!!$     write(0,'(a)') "Type         Current         Maximum"
!!$     write(0,'(a,i12,a,f12.5,a)') "Integer  ",memstats(1)," B ",&
!!$          dble(memstats(6))/BYTES_PER_GB," GB"
!!$     write(0,'(a,i12,a,f12.5,a)') "Real     ",memstats(2)," B ",&
!!$          dble(memstats(7))/BYTES_PER_GB," GB"
!!$     write(0,'(a,i12,a,f12.5,a)') "Logical  ",memstats(3)," B ",&
!!$          dble(memstats(8))/BYTES_PER_GB," GB"
!!$     write(0,'(a,i12,a,f12.5,a)') "Character",memstats(4)," B ",&
!!$          dble(memstats(9))/BYTES_PER_GB," GB"
!!$     write(0,'(a)') "---------------------------------------"
!!$     write(0,'(a,i12,a,f12.5,a)') "Total    ",memstats(5)," B ",&
!!$          dble(memstats(10))/BYTES_PER_GB," GB"
  end if
end subroutine rism_finalize

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!PRIVATE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Sets default values for 3D-RISM paramters
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine defaults()
  use amber_rism_interface
  
  implicit none

  rismprm%closure           = 1
  rismprm%closureOrder      = 3
  rismprm%asympCorr         = .true.
  rismprm%gauss_fluct      = 0

  !solvation box
  rismprm%solvcut         = -1
  rismprm%buffer          = 14d0
  rismprm%grdspc          = 0.5d0
  rismprm%ng3             = -1
  rismprm%solvbox         = -1d0

  !convergence
  rismprm%tolerance            = 1d-5
  rismprm%mdiis_del            = 0.7d0
  rismprm%mdiis_nvec              = 5
  rismprm%mdiis_method     = 2
  rismprm%maxstep           = 10000
  rismprm%npropagate        = 5

  !imin = 1 (minimization)
  rismprm%centering          = 1
  rismprm%zerofrc          = 1

  !imin = 5 (trajectory analysis)
  rismprm%apply_rism_force = 1
  pa_orient        = 0
  rmsd_orient      = 0

  !imin = 0 (MD)
  rismprm%rismnrespa       = 1
#ifdef RISM_CRDINTERP
  rismprm%fcestride        = 0
  rismprm%fcecut           = 9999d0
  rismprm%fcenbasis        = 10
  rismprm%fcecrd           = 0
#endif

  !output
  rismprm%saveprogress            = 0
  rismprm%ntwrism          = -1
  rismprm%verbose     = 0
  rismprm%progress            = 1

end subroutine defaults

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Transfers RISM settings from a rismprm_t type (user) to the rismprm parameter set
!!!for the 3D-RISM calculation.  For each possible setting the new value is used
!!!IFF the value > -9999
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine update_param(user)
  use amber_rism_interface
  implicit none
  type(rismprm_t), intent(in) :: user
  rismprm%irism = user%irism
  if(user%closure >-9999) rismprm%closure = user%closure
  if(user%closureOrder >-9999) rismprm%closureOrder = user%closureOrder
  !asympCorr is a logical, so we can't perform the test
  rismprm%asympCorr = user%asympCorr
  if(user%gauss_fluct > -9999) rismprm%gauss_fluct = user%gauss_fluct
  if(user%buffer > -9999) rismprm%buffer = user%buffer
  if(user%solvcut > 0) then
     rismprm%solvcut = user%solvcut
  else
     rismprm%solvcut = rismprm%buffer     
  end if
  if(user%grdspc(1) > -9999) rismprm%grdspc = user%grdspc
  if(user%ng3(1) > -9999) rismprm%ng3 = user%ng3
  if(user%solvbox(1) > -9999) rismprm%solvbox = user%solvbox
  if(user%tolerance > -9999) rismprm%tolerance = user%tolerance
  if(user%mdiis_del > -9999) rismprm%mdiis_del = user%mdiis_del
  if(user%mdiis_nvec > -9999) rismprm%mdiis_nvec = user%mdiis_nvec
  if(user%mdiis_method > -9999) rismprm%mdiis_method = user%mdiis_method
  if(user%maxstep > -9999) rismprm%maxstep = user%maxstep
  if(user%npropagate > -9999) rismprm%npropagate = user%npropagate
  if(user%centering > -9999) rismprm%centering = user%centering
  if(user%zerofrc > -9999) rismprm%zerofrc = user%zerofrc
  if(user%apply_rism_force > -9999) rismprm%apply_rism_force = user%apply_rism_force
  if(user%rismnrespa > -9999) rismprm%rismnrespa = user%rismnrespa
  if(user%fcestride > -9999) rismprm%fcestride = user%fcestride
  if(user%fcecut > -9999) rismprm%fcecut = user%fcecut
  if(user%fcenbasis > -9999) rismprm%fcenbasis = user%fcenbasis
  if(user%fcecrd > -9999) rismprm%fcecrd = user%fcecrd
  if(user%saveprogress > -9999) rismprm%saveprogress = user%saveprogress
  if(user%ntwrism > -9999) rismprm%ntwrism = user%ntwrism
  if(user%verbose > -9999) rismprm%verbose = user%verbose
  if(user%progress > -9999) rismprm%progress = user%progress
end subroutine update_param

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Reads the RISM namelist from the input file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine read_namelist(mdin_unit)
  use amber_rism_interface

  implicit none
  !mdin_unit :: unit number for mdin file
  integer, intent(in) :: mdin_unit
  integer :: closure
  integer :: closureOrder
  logical :: asympCorr
  integer :: gauss_fluct
  _REAL_ :: solvcut
  _REAL_ :: buffer
  _REAL_ :: grdspc(3)
  integer ::  ng3(3)
  _REAL_ :: solvbox(3)
  _REAL_ :: tolerance
  _REAL_ :: mdiis_del
  integer :: mdiis_nvec
  integer :: mdiis_method
  integer :: maxstep
  integer :: npropagate
  integer :: centering
  integer :: zerofrc
  integer :: apply_rism_force
  integer :: rismnrespa
  integer :: fcestride
  _REAL_ :: fcecut
  integer ::  fcenbasis
  integer ::  fcecrd
  integer :: saveprogress
  integer :: ntwrism
  integer :: verbose
  integer :: progress
  namelist /rism/ &
       !closure
       closure, closureOrder, gauss_fluct,&
       !thermodynamics
       asympCorr, &
       !solvation box
       buffer, grdspc, solvcut, &
       ng3, solvbox,&
       !convergence
       tolerance, mdiis_del, mdiis_nvec, mdiis_method, maxstep, npropagate,&
       !minimization
       centering,zerofrc,&
       !imin=5
       apply_rism_force, pa_orient, rmsd_orient,&
       !md
       rismnrespa,&
#ifdef RISM_CRDINTERP
  fcestride, fcecut, fcenbasis, fcecrd, &
#endif /*RISM_CRDINTERP*/
  !output
  saveprogress, ntwrism, verbose, progress

  closure = rismprm%closure
  closureOrder = rismprm%closureOrder
  asympCorr = rismprm%asympCorr
  gauss_fluct = rismprm%gauss_fluct
  solvcut = rismprm%solvcut
  buffer = rismprm%buffer
  grdspc= rismprm%grdspc
  ng3 = rismprm%ng3
  solvbox = rismprm%solvbox
  tolerance = rismprm%tolerance
  mdiis_del = rismprm%mdiis_del
  mdiis_nvec = rismprm%mdiis_nvec
  mdiis_method = rismprm%mdiis_method
  maxstep = rismprm%maxstep
  npropagate = rismprm%npropagate
  centering = rismprm%centering
  zerofrc = rismprm%zerofrc
  apply_rism_force = rismprm%apply_rism_force
  rismnrespa = rismprm%rismnrespa
  fcestride = rismprm%fcestride
  fcecut = rismprm%fcecut
  fcenbasis = rismprm%fcenbasis
  fcecrd = rismprm%fcecrd
  saveprogress = rismprm%saveprogress
  ntwrism = rismprm%ntwrism
  verbose= rismprm%verbose
  progress = rismprm%progress

  rewind(mdin_unit)
  read(mdin_unit,nml=rism)

  rismprm%closure=closure
  rismprm%closureOrder=closureOrder
  rismprm%asympCorr=asympCorr
  rismprm%gauss_fluct=gauss_fluct
  !solvation box
  rismprm%buffer=buffer
  rismprm%grdspc=grdspc
  rismprm%solvcut=solvcut
  rismprm%ng3=ng3
  rismprm%solvbox=solvbox
  !convergence
  rismprm%tolerance=tolerance
  rismprm%mdiis_del=mdiis_del
  rismprm%mdiis_nvec=mdiis_nvec
  rismprm%mdiis_method=mdiis_method
  rismprm%maxstep=maxstep
  rismprm%npropagate=npropagate
  !minimization
  rismprm%centering=centering
  rismprm%zerofrc=zerofrc
  !imin=5
  rismprm%apply_rism_force=apply_rism_force
  pa_orient=pa_orient
  rmsd_orient=rmsd_orient
  !md
  rismprm%rismnrespa=rismnrespa
#ifdef RISM_CRDINTERP
  rismprm%fcestride=fcestride
  rismprm%fcecut=fcecut
  rismprm%fcenbasis=fcenbasis
  rismprm%fcecrd=fcecrd
#endif /*RISM_CRDINTERP*/
  !output
  rismprm%saveprogress=saveprogress
  rismprm%ntwrism=ntwrism
  rismprm%verbose=verbose
  rismprm%progress=progress

  !set the RISM cutoff if not set by the user
  if(rismprm%solvcut < 0) then
     rismprm%solvcut = rismprm%buffer
  end if
end subroutine read_namelist

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Checks  user input to ensure that it is not completely crazy
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine sanity_check(igb)
  use amber_rism_interface
  implicit none
  !igb       :: generalized born model to use.  3D-RISM requires this to be 6 
  !             (vacuum)
  integer, intent(inout) :: igb
  if(rismprm%closure > 2 .or. rismprm%closure < 0) then
     call rism_report_error('(a,i4)','Unexpected closure for RISM:  ',rismprm%closure)
  endif
!!$  if(rismprm%asympCorr > 1 .or. rismprm%asympCorr < 0) then
!!$     call rism_report_error('(a,i2)','RISM asymptotic correction must be 0 or 1:  ',rismprm%asympCorr)
!!$  endif

  !ensure that the cutoff is set to a reasonable value.  This can
  !happen if buffer is < 0
  if(rismprm%solvcut <0) &
       call rism_report_error('solvcut must > 0.')

  !ensure that solvbox and ng3 have been set if buffer < 0
  if(rismprm%buffer <0)then
     if(minval(rismprm%ng3) < 0) &
          call rism_report_error('if buffer < 0 ng3 (grid size) must be set.')
     if(minval(rismprm%solvbox) < 0) &
          call rism_report_error('if buffer < 0 solvbox must be set.')
  end if

#ifdef SANDER
  !force igb=6 to get vacuum electrostatics
!  if(igb/=6) write(6,'(a)') "|3D-RISM Forcing igb=6"
!  igb=6
#else /*SANDER*/
  igb=0
#endif /*SANDER*/
end subroutine sanity_check

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!modifies the position of the solute using the centering method 
!!!requested.  (This used to also rotate the solute in the box but this
!!!has been disabled.)
!!!If centering is /= 0 we want to move the solute to the center of the solvent box
!!!However, if centering < 0 we only figure out the displacement required the _first_
!!!time we see the solute.  Thus, for centering <= 0, the solute's CM can move relative
!!!to the grid
!!!IN:
!!!   solu :: solute object
!!!   ratu :: the x,y,z position of each solute atom.  This is modified 
!!!           to place the solute in the center of the box
!!!   nsolution :: number of times a RISM solution has been calculated.
!!!                Methods < 0 only are used it nsolution ==0
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine orient(solute,ratu, nsolution)
  use amber_rism_interface
  use rism_util, only : calc_cm,translate
  use constants, only : PI
  implicit none
  type(rism3d_solu) :: solute
  _REAL_,intent(inout) :: ratu(3,solute%natom)
  integer, intent(in) :: nsolution

  _REAL_ :: cm(3), weight(solute%natom)
  !counters
  integer :: iu, id
  !................ setting solute site positions in box .................
#ifdef RISM_DEBUG
  write(6,*)"RISM REORIENT"
  call flush(6)
#endif /*RISM_DEBUG*/


  if(abs(rismprm%centering)==1)then
     weight=solute%mass
  elseif(abs(rismprm%centering)==2)then
     weight=1
  end if
  if(rismprm%centering > 0 .or. (rismprm%centering < 0 .and. nsolution == 0))then
     call calc_cm(ratu,ratucm,weight,solute%natom)
  end if
  if(rismprm%centering /= 0) then
     call translate(ratu,solute%natom,-1*ratucm)
  end if
end subroutine orient

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!return the solute to its original position
!!!IN:
!!!   solute :: solute object
!!!   ratu  :: the x,y,z position of each solute atom.  This is modified 
!!!            to return the solute to its original MD position and orientation
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine unorient(solute,ratu)
  use amber_rism_interface
  use rism_util, only: calc_cm, translate
  implicit none
  type(rism3d_solu) :: solute
  _REAL_,intent(inout) :: ratu(3,solute%natom)

  _REAL_ :: cm(3),weight(solute%natom)
  !counters
  integer :: iu, id

  !we don't know if additional translations have been performed on the
  !solute.  1) The first step is to calculate the current COM and
  !translate this to the origin.  2) Then the system and forces are
  !rotated according to qback.  3) Finally, using ratucm, we translate
  !the system back to its original MD COM

#ifdef RISM_DEBUG
  write(6,*)"RISM UNORIENT", ratucm
  write(6,*) qback
  call flush(6)
#endif /*RISM_DEBUG*/
  if(abs(rismprm%centering)==1)then
     weight=solute%mass
     call calc_cm(ratu,cm,weight,solute%natom)
     call translate(ratu,solute%natom,-1*cm)
     call calc_cm(ratu,cm,weight,solute%natom)
  elseif(abs(rismprm%centering)==2)then
     weight=1
     call calc_cm(ratu,cm,weight,solute%natom)
     call translate(ratu,solute%natom,-1*cm)
     call calc_cm(ratu,cm,weight,solute%natom)
  end if
  if(rismprm%centering /=0)then
     call translate(ratu,solute%natom,ratucm)
  end if

end subroutine unorient

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! I/O: performs RISM related I/O for files that only deal with RISM data
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Provides access to writeSolvDist for non-Fortran code
!!!IN:
!!!   step :: step number used as a suffix
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine rism_writeSolvDistC(step)
  use amber_rism_interface
  implicit none
  integer, intent(in) :: step
!  write(6,*) "writeSolvDistC stub"
  call rism_writeSolvDistF(rism_3d,step)
end subroutine rism_writeSolvDistC

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Outputs current solvent distributions to file.  Each distribution is written
!!!in a separate file with the step number before the suffix.  File names are taken 
!!!from guvfile, huvfile and cuvfile.
!!!IN:
!!!   this :: 3D-RISM object
!!!   step :: step number used as a suffix
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine rism_writeSolvDistF(this,step)
  use amber_rism_interface
  use rism3d_opendx
  use safemem
  implicit none
#if defined(MPI)
  include 'mpif.h'
#endif
  type(rism3d),intent(in) :: this
  integer, intent(in) :: step
  integer :: iv, ivv, nsolv, igx,igy,igz,ios, i,j,k
  character(len=16) :: cstep
  character(len=64) :: suffix
#ifdef MPI
  integer :: err
  _REAL_, pointer::work(:,:,:)=>NULL()
#endif /*MPI*/  
#ifdef RISM_DEBUG
  write(6,*) "printrism"
#endif
#ifdef MPI
  if(len_trim(guvfile) /= 0 .or. len_trim(huvfile) /= 0)then
     work => safemem_realloc(work, this%grid%ngr(1), this%grid%ngr(2), this%grid%nr(3))
  end if
#endif /*MPI*/
  !................... outputting Guv and Cuv profiles ...................
  do iv=1,this%solv%natom
     write(cstep,'(i16)') step
     cstep = adjustl(cstep)
     suffix = '.'//trim(rism_3d%solv%atomname(iv))//'.'//trim(cstep)//'.dx'

     !GUV
     if (len_trim(guvfile) /= 0)  then
#  if defined(MPI)
        do k=1,this%grid%nr(3)
           do j=1,this%grid%ngr(2)
              do i=1,this%grid%ngr(1)
                 work(i,j,k) = this%guv(i+(j-1)*(this%grid%ngr(1)+2)+(k-1)*(this%grid%ngr(1)+2)*this%grid%ngr(2),iv)
              end do
           end do
        end do
        call writeDX(trim(guvfile)//suffix,&
             work,&
             this%grid%boxlen,this%grid%nr,this%grid%ngr(3),this%ratucm,&
             mpirank,mpisize,mpicomm)
#  else
        call writeDX(trim(guvfile)//suffix,this%guv(:,iv),this%grid%boxlen,&
             this%grid%nr,this%grid%ngr(3),this%ratucm)
#  endif /*defined(MPI)*/
     endif

     !HUV
     if (len_trim(huvfile) /= 0)  then
#  if defined(MPI)
        do k=1,this%grid%nr(3)
           do j=1,this%grid%ngr(2)
              do i=1,this%grid%ngr(1)
                 work(i,j,k) =this%huv(i+(j-1)*(this%grid%ngr(1)+2)+(k-1)*(this%grid%ngr(1)+2)*this%grid%ngr(2),iv)
              end do
           end do
        end do
        call writeDX(trim(huvfile)//suffix,&
             work,&
             this%grid%boxlen,this%grid%nr,this%grid%ngr(3),this%ratucm,&
             mpirank,mpisize,mpicomm)
#  else
        call writeDX(trim(huvfile)//suffix,this%huv(:,iv),this%grid%boxlen,&
             this%grid%nr,this%grid%ngr(3),this%ratucm)
#  endif /*defined(MPI)*/
     endif

     !CUV
     if (len_trim(cuvfile) /= 0)  then
        call writeDX(trim(cuvfile)//suffix,this%cuv(:,:,:,iv),&
             this%grid%boxlen,this%grid%nr,this%grid%ngr(3),this%ratucm,&
             mpirank,mpisize,mpicomm)
     endif

     !UUV
     if (len_trim(uuvfile) /= 0)  then
        call writeDX(trim(uuvfile)//suffix,this%pot%uuv(:,:,:,iv),&
             this%grid%boxlen,this%grid%nr,this%grid%ngr(3),this%ratucm,&
             mpirank,mpisize,mpicomm)
     endif
  end do
  !asymptotics files.  There are four but it is a simple loop
  if (len_trim(asympfile) /= 0)  then
     write(cstep,'(i16)') step
     cstep = adjustl(cstep)
     suffix = '.'//trim(cstep)//'.dx'
     !h(r)
     call writeDX(trim(asympfile)//"hr"//suffix,this%pot%asymhr,this%grid%boxlen,this%grid%nr,this%grid%ngr(3),this%ratucm,&
          mpirank,mpisize,mpicomm)
     !c(r)
     call writeDX(trim(asympfile)//"cr"//suffix,this%pot%asymcr,this%grid%boxlen,this%grid%nr,this%grid%ngr(3),this%ratucm,&
          mpirank,mpisize,mpicomm)
     !h(k)
     !in (k) space the array is transposed (distributed over the y axis) so writeDX doesn't
     !know how to handle it
!     call writeDX(unit,this%pot%asymhk,this%grid%boxlen,uboundthis%grid%nr,this%grid%ngr(3),this%ratucm,&
!          mpirank,mpisize,mpicomm)
!     call writeDX(trim(asympfile)//"hk"//suffix,this%pot%asymhk,&
!          this%grid%boxlen,(/this%grid%ngr(1)+2,this%ny_local, this,this%,this%grid%ngr(2),this%ratucm)
     !c(k)
     !in (k) space the array is transposed (distributed over the y axis) so writeDX doesn't
     !know how to handle it
!     call writeDX(unit,this%pot%asymck,this%grid%boxlen,uboundthis%grid%nr,this%grid%ngr(3),this%ratucm,&
!          mpirank,mpisize,mpicomm)
!     call writeDX(trim(asympfile)//"ck"//suffix,this%pot%asymck,&
!          this%grid%boxlen,ubound(this%pot%asymck),this%grid%ngr(2),this%ratucm)
  endif
  
#ifdef MPI
     err = safemem_dealloc(work)
#endif /*MPI*/
end subroutine rism_writeSolvDistF

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!INTERPOLATION RESTART FILE I/O
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#if 0
#if defined(RISM_CRDINTERP)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!Writes the interpolation restart file
!!IN:
!!   ratu    :: the position of each solute atom for each step 
!!   frc     :: the force of each solute atom for each step
!!   this%solu%natom    :: number of solute atoms
!!   nstep   :: number of steps
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!FIX - modify this to use the NetCDF format
subroutine rismrestrtwrit(this,ratu,frc,nstep)
  use amber_rism_interface
  use rism_util, only : freeUnit
  implicit none
#include "files.h"
  type(rism3d) :: this
  _REAL_,intent(in) :: ratu(3,this%solu%natom,nstep),frc(3,this%solu%natom,nstep)
  integer,intent(in) :: nstep
  integer :: istep
  integer :: unit
  !this is performed by the master node only:
  if(mpirank == 0) then
#ifdef RISM_DEBUG
     write(6,*) 'entering rismrestrtwrit'
     call flush(6)
#endif /*RISM_DEBUG*/
     unit = freeUnit()
     if (len_trim(rismcrdfil) /= 0)  then
        open(unit=unit,file=rismcrdfil,status='new', form='FORMATTED',iostat=stat)
        write(unit,'(i8)') nstep
        do istep = 1,nstep
           call corpac(ratu(1:3,1:this%solu%natom,istep),1,this%solu%natom*3,unit,.true.)
        end do
        close(unit)
     end if
     if (len_trim(rismfrcfil) /= 0)  then
        open(unit=unit,file=rismfrcfil,status='new', form='FORMATTED',iostat=stat)
        write(unit,'(i8)') nstep
        do istep = 1,nstep
           call corpac(frc(1:3,1:this%solu%natom,istep),1,this%solu%natom*3,unit,.true.)
        end do
        close(unit)
     end if
#ifdef RISM_DEBUG
     write(6,*) 'done rismrestrtwrit'
     call flush(6)
#endif /*RISM_DEBUG*/
  end if
end subroutine rismrestrtwrit
!!endfix

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!Reads in the interpolation restart file
!!IN:
!!   ratu    :: the position of each atom for each step read in
!!   frc     :: the force of each atom for each step read in
!!   this%solu%natom    :: number of solute atoms
!!   nstep   :: maximum number of steps to read
!!   nsample :: will hold the total number of steps read in
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!FIX - modify this to use the NetCDF format
subroutine rismrestrtread(this,ratu,frc,nstep,nsample)
  use amber_rism_interface
  use rism_util, only : freeUnit
  implicit none
#include "files.h"
#if defined(MPI)
  include 'mpif.h'
#endif /*defined(MPI)*/
  type(rism3d) :: this
  _REAL_,intent(out) :: ratu(3,this%solu%natom,nstep),frc(3,this%solu%natom,nstep)
  integer,intent(in) :: nstep
  integer,intent(out) :: nsample
  integer :: istep,iatu,csteps,fsteps,err
  integer :: unit
  !this is performed by the master node only:
  if(mpirank == 0) then
#ifdef RISM_DEBUG
     write(6,*) "Reading interpolation restart files..."
     write(6,*) 'entering rismrestrtread'
     call flush(6)
#endif /*RISM_DEBUG*/
     unit=freeeUnit()
     if (len_trim(rismcrdrstfil) /= 0)  then
        open(unit=unit,file=rismcrdfil,status='old', form='FORMATTED',iostat=stat)
        read(unit,'(i12)') csteps
        !discard the extra samples at the beginning
#ifdef RISM_DEBUG
        write(6,*) csteps, nstep, csteps-nstep
        call flush(6)
#endif /*RISM_DEBUG*/
        do istep = 1,csteps-nstep
           !            read(unit,'(10f8.3)') (ratu(1:3,iatu,1),iatu=1,this%solu%natom)
           read(unit,'(10f21.16)') (ratu(1:3,iatu,1),iatu=1,this%solu%natom)
        end do
        do istep = 1,min(csteps,nstep)
           !            read(unit,'(10f8.3)') (ratu(1:3,iatu,istep),iatu=1,this%solu%natom)
           read(unit,'(10f21.16)') (ratu(1:3,iatu,istep),iatu=1,this%solu%natom)
        end do
        close(unit)
     endif
     if (len_trim(rismfrcrstfil) /= 0)  then
        open(unit=unit,file=rismfrcfil,status='old', form='FORMATTED',iostat=stat)
        read(unit,'(i12)') fsteps
        !discard the extra samples at the beginning
        do istep = 1,fsteps-nstep
           !            read(unit,'(10f8.3)') (frc(1:3,iatu,1),iatu=1,this%solu%natom)
           read(unit,'(10f21.16)') (frc(1:3,iatu,1),iatu=1,this%solu%natom)
        end do
        do istep = 1,min(fsteps,nstep)
           !            read(unit,'(10f8.3)') (frc(1:3,iatu,istep),iatu=1,this%solu%natom)
           read(unit,'(10f21.16)') (frc(1:3,iatu,istep),iatu=1,this%solu%natom)
        end do
        close(unit)
        if(csteps /= fsteps)then
           call rism_report_error('RISM interpolation restart files have different numbers of entries.')
        end if
        nsample = min(csteps,nstep)
     end if
#ifdef RISM_DEBUG
     write(6,*) nsample," samples read"
     write(6,*) 'done rismrestrtread'
     call flush(6)
#endif /*RISM_DEBUG*/
  end if
  !
  !broadcast results to the other processes: nsamples, ratu and frc
  !
!!$#if defined(MPI)
!!$      CALL MPI_BCAST(nsample,1,MPI_INTEGER,0,mpicomm,err)
!!$      CALL MPI_BCAST(ratu,3*this%solu%natom*nstep,MPI_DOUBLE_PRECISION,0,mpicomm,err)
!!$      CALL MPI_BCAST(frc,3*this%solu%natom*nstep,MPI_DOUBLE_PRECISION,0,mpicomm,err)
!!$#endif /*defined(MPI)*/

end subroutine rismrestrtread
#endif /*RISM_CRDINTERP*/
#endif  /* 0 */

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Writes the contents of a C string (array of chars) to a Fortran string.  Will
!!!not write past the end of the Fortran string
!!!IN:
!!!   fstr : Fortran string to write to.
!!!   cstr : C string to write from.
!!!   nchar  : Number of chars in cstr (not including null).
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine cstr2fstr(fstr,cstr,nchar)
  implicit none
  character(len=*), intent(out) :: fstr
  integer, intent(in) :: nchar
  integer(kind=1), intent(in) :: cstr(nchar+1)
  integer :: i
  fstr = ""
  do i = 1, min(nchar,len(fstr))
     fstr = trim(fstr)//char(cstr(i),1)
  end do
end subroutine cstr2fstr


#ifdef SANDER
function mylcm(a,b)
  use rism_util, only: lcm
  implicit none
  integer :: mylcm,a,b
  mylcm = lcm(a,b)
end function mylcm
#else /*SANDER*/
subroutine timer_start( label )
integer label
end subroutine timer_start

subroutine timer_stop( label )
integer label
end subroutine timer_stop
#endif /*SANDER*/

