! <compile=optimized>

!The 3D-RISM-KH software found here is copyright (c) 2010 by Andriy Kovalenko,
!Tyler Luchko and David A. Case.
!
!This program is free software: you can redistribute it and/or modify it
!under the terms of the GNU General Public License as published by the Free
!Software Foundation, either version 3 of the License, or (at your option)
!any later version.
!
!This program is distributed in the hope that it will be useful, but
!WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
!for more details.
!
!You should have received a copy of the GNU General Public License in the
!../../LICENSE file.  If not, see <http://www.gnu.org/licenses/>.
!
!Users of the 3D-RISM capability found here are requested to acknowledge
!use of the software in reports and publications.  Such acknowledgement
!should include the following citations:
!
!1) A. Kovalenko and F. Hirata. J. Chem. Phys., 110:10095-10112  (1999); 
!ibid. 112:10391-10417 (2000).   
!
!2) A. Kovalenko,  in:  Molecular  Theory  of  Solvation,  edited  by  
!F. Hirata  (Kluwer Academic Publishers, Dordrecht, 2003), pp.169-275.  
!
!3) T. Luchko, S. Gusarov, D.R. Roe, C. Simmerling, D.A. Case, J. Tuszynski,
!and  A. Kovalenko, J. Chem. Theory Comput., 6:607-624 (2010). 

#include "../include/dprec.fh"
  module rism1d_m
    use rism1d_c
    use rism_report_c
    implicit none
    integer, parameter :: CLEN=256
    character(len=CLEN) :: fileroot
    type(rism1d),save :: rism

!!!!!!!!!!!!!
!!!Parameters
!!!!!!!!!!!!!
    !THEORY
    !theory  :: 1D-RISM theory to use.  May be DRISM or XRISM
    !closure :: closure to use. May be KH, PSE, HNC, PY or MV0
    character(len=CLEN) :: theory, closure
    !closure_order :: order parameter for closures like PSE
    integer :: closure_order
    !closure_coeff :: coefficients for closures that can make use of them (none so far)
    _REAL_ :: closure_coeff(10)

    !GRID SIZE
    !dr :: grid spacing [A]
    _REAL_ :: dr
    !nr :: number of grid points
    integer :: nr
    
    !OUTPUT
    !outlist :: list of output files to produce. See output() for details
    character(len=CLEN) :: outlist
    !nrout    :: Max r-space index to output
    !nkout    :: Max k-space index to output
    !progress :: Display iteration progress every 'progress' iterations
    !ksave    :: write a restart file every ksave iterations
    integer :: nrout, nkout, progress, ksave
    
    !rout    :: Max r-space distance to output
    !kout    :: Max k-space wavenumber to output
    _REAL_ :: rout, kout

    !SOLUTION CONVERGENCE
    !mdiis_nvec :: number of MDIIS vectors
    !maxstep    :: maximum number of iterations
    integer :: mdiis_nvec, maxstep
    !mdiis_del  :: MDIIS step size
    !tolerance  :: maximum tolerance for solution
    _REAL_ :: mdiis_del, tolerance
    
    !SOLVENT DESCRIPTION
    !temperature :: temperature of the solvent [K]
    !dieps       :: dielectric constant of the solvent
    _REAL_ :: temperature, dieps
    
    !ELECTROSTATICS
    !smear  :: smear parameter for long range electrostatics [A]
    !adbcor :: coefficient for DRISM (see Perkyns and Pettitt. J. Chem. Phys. 97, 1992, Eq. 34, variable 'a')
    _REAL_ ::  smear, adbcor
    
    
    !DEPRICATED
    !closur :: synonym for closure
    !outlst :: synonym for outlist
    character(len=CLEN) :: closur,outlst 
    !routup :: synonym for rout
    !toutup :: synonym for kout
    !kshow  :: synonym for progress
    !maxste :: synonym for maxstep
    !nis    :: synonym for mdiis_nvec
    integer :: routup, toutup, kshow, nis, maxste
    !delvv :: synonym for mdiis_del
    !tolvv :: synonym for tolerance
    !temper :: synonym for temperature
    _REAL_ :: delvv, tolvv, temper
  contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Reads user input and initialize rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine initialize()
      use rism_util, only : freeUnit
      use constants, only : avogadro
      use solvMDL_c
      implicit none

      character(len=clen) :: inpfile
      integer :: unit,i, iostat
      type(solvMDL) :: mdl

      !nsp :: number of solvent species
      integer :: nsp
      !density :: density of the solvent species
      _REAL_ :: density
      !model :: MDL file
      character(len=CLEN) :: model,units

      namelist /parameters/ outlist,theory,closure,nr,dr,rout,kout,&
           mdiis_nvec, mdiis_del, tolerance, ksave, progress, maxstep, smear, adbcor, temperature,&
           dieps,nsp,closure_order,closure_coeff,&
           closur,outlst, routup, toutup, kshow, nis, maxste, delvv, tolvv, temper
      namelist /species/ density,model,units

      call defaults()

      call  getarg (1,fileroot)
      inpfile = trim(fileroot)//".inp"

      call rism_report_message('reading input data file: '//trim(inpfile))

      unit = freeUnit()
      open(unit=unit,file=trim(inpfile), status='old',iostat=iostat)
      if(iostat /= 0) &
         call rism_report_error('(a,i4)',"Could not open "//trim(inpfile)//":",iostat)

      read (unit,parameters)

      call sanity_check()

      call rism1d_new(rism,theory, closure, closure_order,closure_coeff, temperature, dieps, smear, &
           adbcor, nr, dr,mdiis_nvec,mdiis_del,trim(fileroot)//'.sav')

      if(rout==0) then
         nrout = nr
      else
         nrout = min(nr, int(rout/dr))
      end if
      if(kout==0) then
         nkout = nr
      else
         nkout = min(nr, int(kout/rism%pot%dk))
      end if

      !---read paramters for each species of solvent---!
      do i=1,nsp
         !default units
         units='M'
         read (unit,species)
         call solvMDL_new(mdl,model)
         select case (units)
         case ('M') 
            density = density*avogadro*1d-27
         case ('mM') 
            density = density*avogadro*1d-30
         case ('1/A^3') 
         case ('g/cm^3') 
            density = density/sum(mdl%mass*mdl%multi)*avogadro*1d-24               
         case ('kg/m^3') 
            density = density/sum(mdl%mass*mdl%multi)*avogadro*1d-27               
         case default 
            call rism_report_error("'"//trim(units)//"' are not valid density units")
         end select
         call rism1d_addSpecies(rism,mdl,density)
         call solvMDL_destroy(mdl)
      enddo
    end subroutine initialize

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Set default user parameters
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine defaults()
      implicit none
      theory='DRISM'
      closure='KH'
      outlist=''
      nr=2**14
      dr=0.025d0
      rout=0
      kout=0
      mdiis_nvec=20
      mdiis_del=0.3d0
      tolerance=1d-12
      ksave=-2
      progress=1
      maxstep=10000
      smear=1d0
      adbcor=0.5d0
      temperature=298.15d0
      dieps=-1d0             !must be set by user if theory='drism'
      closure_order=3
      closure_coeff=0d0        !place holder
      !deprecated variable names
      closur=''
      outlst=''
      routup=-1
      toutup=-1
      kshow =-1
      nis   =-1
      maxste=-1
      delvv =-1d0
      tolvv =-1d0
      temper=-1d0
    end subroutine defaults

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Check user parameters
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine sanity_check()
      use rism_util, only : caseup
      implicit none
      !set everything to upper case
      call  caseup (outlist)
      call  caseup (outlst)
      call  caseup (theory)
      call  caseup (closure)
      call  caseup (closur)

      !handle deprecated parameter names
      if(closur .ne. '')then
         call rism_report_warn("'closur' is deprecated")
         if(closure .ne. 'KH') call rism_report_error("Both 'closure' and 'closur' defined")
         closure = closur
      end if
      if(outlst .ne. '')then
         call rism_report_warn("'outlst' is deprecated")
         if(outlist .ne. '') call rism_report_error("Both 'outlist' and 'outlst' defined")
         outlist = outlst
      end if
      if(routup /=-1)then
         call rism_report_warn("'routup' is deprecated")
         if(rout /= 0) call rism_report_error("Both 'rout' and 'routup' defined")
         rout = routup
      end if
      if(toutup /=-1)then
         call rism_report_warn("'toutup' is deprecated")
         if(kout /= 0) call rism_report_error("Both 'kout' and 'toutup' defined")
         kout = toutup
      end if
      if(kshow /=-1)then
         call rism_report_warn("'kshow' is deprecated")
         if(progress /= 1) call rism_report_error("Both 'progress' and 'kshow' defined")
         progress = kshow
      end if
      if(nis /=-1)then
         call rism_report_warn("'nis' is deprecated")
         if(mdiis_nvec /= 20) call rism_report_error("Both 'mdiis_nvec' and 'nis' defined")
         mdiis_nvec = nis
      end if
      if(maxste /=-1)then
         call rism_report_warn("'maxste' is deprecated")
         if(maxstep /= 10000) call rism_report_error("Both 'maxstep' and 'maxste' defined")
         maxstep = maxste
      end if
      if(delvv /=-1d0)then
         call rism_report_warn("'delvv' is deprecated")
         if(mdiis_del /= 0.3d0) call rism_report_error("Both 'mdiis_del' and 'delvv' defined")
         mdiis_del = delvv
      end if
      if(tolvv /=-1d0)then
         call rism_report_warn("'tolvv' is deprecated")
         if(tolerance /= 1d-12) call rism_report_error("Both 'tolerance' and 'tolvv' defined")
         tolerance = tolvv
      end if
      if(temper /=-1d0)then
         call rism_report_warn("'temper' is deprecated")
         if(temperature /= 298.15d0) call rism_report_error("Both 'temperature' and 'temper' defined")
         temperature = temper
      end if

      !outlist: hmmm... leave this for later...

      if(rout < 0)&
           call rism_report_error("ROUT must be >= 0")
      if(kout < 0)&
           call rism_report_error("KOUT must be >= 0")

      !Check the dielectric constant
      if(theory .eq. "XRISM" .and. dieps < 0) &
           call rism_report_error("DIEPS must be set for XRISM")

    end subroutine sanity_check

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Invokes the rism1d solver
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine solve()
      implicit none
      logical :: converged
      converged = rism1d_solve(rism,ksave,progress,maxstep,tolerance)
      if(.not. converged)then
         call rism_report_error('(a,i5)','1D-RISM failed to converge.  Reached steps limit Maxstep=',maxstep)
      end if
    end subroutine solve

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Cleanup memory
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine finalize()
      implicit none
      integer*8 :: memstats(10)
      integer :: unit 
      unit = rism_report_getMUnit()
      call rism1d_destroy(rism)
      memstats = memStatus()
      write(unit,'(a)') "1D-RISM memory allocation summary"
      write(unit,'(a)') "Type         Current         Maximum"
      write(unit,'(a,i12,a,f12.5,a)') "Integer  ",memstats(1)," B ",&
           dble(memstats(6))/BYTES_PER_GB," GB"
      write(unit,'(a,i12,a,f12.5,a)') "Real     ",memstats(2)," B ",&
           dble(memstats(7))/BYTES_PER_GB," GB"
      write(unit,'(a,i12,a,f12.5,a)') "Logical  ",memstats(3)," B ",&
           dble(memstats(8))/BYTES_PER_GB," GB"
      write(unit,'(a,i12,a,f12.5,a)') "Character",memstats(4)," B ",&
           dble(memstats(9))/BYTES_PER_GB," GB"
      write(unit,'(a)') "---------------------------------------"
      write(unit,'(a,i12,a,f12.5,a)') "Total    ",memstats(5)," B ",&
           dble(memstats(10))/BYTES_PER_GB," GB"

    end subroutine finalize

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate and output the thermodynamic properties the user has requested
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine output()
      use constants, only : KB,COULOMB_CONST_E
      implicit none
      _REAL_, pointer :: exnvv(:,:,:)=>NULL(), nvv(:,:,:)=>NULL(), &
           bvv(:,:)=>NULL()
      _REAL_ :: qEx(rism%pot%nv,rism%pot%nv)
      integer :: iv
      !NOTE: we should be able to combine writeUVV and writeRunning Coordination
      !into something more general

      if (index(outlist,'U') /= 0)  then
         call writeVV(rism%pot%uvv,'uvv','POTENTIAL ENERGY [kT]')
      endif
      if (index(outlist,'X') /= 0)  then
         call writeXVV()
      endif
      if (index(outlist,'G') /= 0)  then
         call writeVV(rism%gvv,'gvv','PAIR DISTRIBUTION')
      endif
      if (index(outlist,'H') /= 0)  then
         call writeVV(rism%hvv,'hvv','TOTAL CORRELATION',k_space=.true.)
      endif
      if (index(outlist,'C') /= 0)  then
         call writeVV(rism%cvv,'cvv','DIRECT CORRELATION')
      endif
      if (index(outlist,'B') /= 0)  then
         bvv => rism1d_bvv(rism)
         call writeVV(bvv,'bvv','BRIDGE FUNCTION')
         if(safemem_dealloc(bvv) /=0)&
              call rism_report_error("OUTPUT: dealloc of Bvv failed")
      endif
      if (index(outlist,'T') /= 0)  then
         call writeTD()
      endif
      if (index(outlist,'E') /= 0)  then
         exnvv => rism1d_getRunExNumber(rism)
         call writeRunningCoordination(exnvv,'exnvv',"RUNNING EXCESS COORDINATION NUMBER")
         if(safemem_dealloc(exnvv)/=0) &
            call rism_report_error("Deallocating memory for running excess number")
         call writeTotalExcess(rism1d_getExNumber(rism),'n00',&
              "Total excess coordination number")
      endif
      if (index(outlist,'Q') /= 0)  then
         qEx = rism1d_getExNumber(rism)
         do iv =1, rism%pot%nv
            qEx(iv,:) = qEx(iv,:)*rism%pot%qv*rism%pot%mtv*sqrt(KB*rism%pot%temperature/COULOMB_CONST_E)
         end do
         call writeTotalExcess(qEx,'q00',&
              "Total excess coordinated charge [e]",.true.)
      end if
      if (index(outlist,'N') /= 0)  then
         nvv => rism1d_getRunNumber(rism)
         call writeRunningCoordination(nvv,'nvv',"RUNNING COORDINATION NUMBER")
         if(safemem_dealloc(nvv)/=0)&
            call rism_report_error("Deallocating memory for running number")
      endif
      if (index(outlist,'S') /= 0)  then
         call writeSVV()
      endif

    end subroutine output

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Writes the site-site suceptibility to file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine writeXVV()
      use rism_util
      use constants, only : KB, COULOMB_CONST_E, JPKC
      implicit none
      character(len=clen) :: file
      integer :: unit, iostat
      character(len=8) :: date
      character(len=10) :: time
      character(len=5) :: zone
      integer :: values(8)
      integer :: iv, ir, iv1, iv2
      _REAL_ :: delhv0(rism%pot%nv)
      _REAL_, pointer :: xvv(:,:,:)

      unit = freeUnit()
      file = trim(fileroot)//'.xvv'
      call rism_report_message('outputting Xvv(K) to file: '//trim(file))
      open(unit,file=file,status='replace',iostat=iostat)
      if(iostat/=0)&
         call rism_report_error("failed to open :"//trim(file))

      call date_and_time(date,time,zone,values)
      write (unit,'(12a)') "%VERSION  VERSION_STAMP = V0000.001  DATE = ",&
           date(5:6),":",date(7:8),":",date(3:4)," ",time(1:2),":",time(3:4),":",time(5:6)
      write (unit,100) "%COMMENT NR,NATV,NSP"
      write (unit,100) "%FLAG POINTERS"
      write (unit,100) "%FORMAT(10I8)"
      write (unit,201) rism%POT%NR,rism%POT%NV,rism%POT%NSP
      write (unit,100) "%COMMENT TEMPER,DIEPS,XAPPA,XIKT,DR,SMEAR"
      write (unit,100) "%FLAG THERMO"
      write (unit,100) "%FORMAT(5E24.16)"
      write (unit,202) rism%POT%TEMPERATURE,rism%pot%dielconst,rism1d_getInvDebyeLen(rism),&
           rism1d_getCompressibility(rism),rism%POT%DR,rism%POT%SMEAR
      write (unit,100) "%FLAG ATOM_NAME"
      write (unit,100) "%FORMAT(20a4)"
      write (unit,301) rism%pot%namev
      write (unit,100) "%FLAG MTV"
      write (unit,100) "%FORMAT(10I8)"
      write (unit,201) rism%pot%mtv
      write (unit,100) "%FLAG RHOV"
      write (unit,100) "%FORMAT(5E24.16)"
      write (unit,202) rism%pot%rhov
      write (unit,100) "%FLAG QV"
      write (unit,100) "%FORMAT(5E24.16)"
      write (unit,202) rism%pot%qv/sqrt(COULOMB_CONST_E/KB/rism%pot%temperature)
      write (unit,100) "%FLAG QSPV"
      write (unit,100) "%FORMAT(5E24.16)"
      write (unit,202) rism%pot%qspv/sqrt(COULOMB_CONST_E/KB/rism%pot%temperature)
      write (unit,100) "%FLAG EPSV"
      write (unit,100) "%FORMAT(5E24.16)"
      write (unit,202) rism%pot%epsv*KB*rism%pot%temperature*JPKC
      write (unit,100) "%FLAG SIGV"
      write (unit,100) "%FORMAT(5E24.16)"
      write (unit,202) rism%pot%rminv
      write (unit,100) "%FLAG DELHV0"
      write (unit,100) "%FORMAT(5E24.16)"
      delhv0 = rism1d_getDelHvLimit(rism)
      write (unit,202) delhv0*sqrt(COULOMB_CONST_E/KB/rism%pot%temperature)
      write (unit,100) "%COMMENT COLUMN MAJOR NR X NAT X NAT"
      write (unit,100) "%FLAG XVV"
      write (unit,100) "%FORMAT(5E24.16)"
      xvv => rism1d_getSusceptibility(rism)
      write (unit,202)  xvv

      close (unit)
100   format(A)
201   format(10i8)
202   format(1p5E24.16)
301   format(20a4)
      close(unit)
    end subroutine writeXVV

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Writes the r-space site-site correlation function to file
!!!IN:
!!!   vv    : correlation function (nr,nvv)
!!!   suffix: suffix for the file
!!!   descript : description of the column  
!!!   k_space : (optional) if true, data is in reciprocal space and the frequency
!!!            will be printed out.  if false (default), data is in real space
!!!            and the separation will be printed out    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine writeVV(vv, suffix, descript,k_space)
      use rism_util
      implicit none
      _REAL_, intent(in) :: vv(:,:)
      character(len=*),intent(in) :: suffix, descript
      logical, optional, intent(in) :: k_space
      logical :: kspace
      character(len=clen) :: file,usuffix
      integer :: unit, iostat
      character(len=64) :: fmt
      integer :: ir, ivv
      kspace = .false.
      if(present(k_space)) kspace = k_space
      unit = freeUnit()
      file = trim(fileroot)//'.'//suffix
      usuffix = suffix
      call caseup(usuffix(1:1))
      call rism_report_message('outputting '//trim(usuffix)//'(R) to file: '//trim(file))
      open(unit,file=file,status='replace',iostat=iostat)
      if(iostat/=0)&
         call rism_report_error("failed to open :"//trim(file))
      if(kspace)then
         call writeVVHeader(unit, &
              "#RISM1D ATOM-ATOM INTERACTIONS: "//trim(descript)//" VS. FREQUENCY [1/A]",&
              17,k_space=kspace)
      else
         call writeVVHeader(unit, &
              "#RISM1D ATOM-ATOM INTERACTIONS: "//trim(descript)//" VS. SEPARATION [A]",17)
      end if
      write(fmt,'(a,i8,a)') "(1p,1x,",rism%pot%nvv+1,"(1x,E16.8E3))"
!      write(fmt,'(a,i8,a)') "(1p,1x,",rism%pot%nvv+1,"(1x,E24.16E3))"
      if(kspace)then
         do ir=1,nkout
            write (unit,fmt)  (ir-1)*rism%pot%dk, rmExPrec(vv(ir,:))
         enddo
      else
         do ir=1,nrout
            write (unit,fmt)  (ir-1)*rism%pot%dr, rmExPrec(vv(ir,:))
         enddo
      endif
      close(unit)
    end subroutine writeVV

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Writes bulk thermodynamic quantities to file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine writeTD()
      use rism_util, only : freeUnit
      use constants, only : boltzmann, KB
      implicit none
      _REAL_ :: pmv(rism%pot%nsp), exchem(rism%pot%nv), exchemsp(rism%pot%nsp),&
           exchemIon(rism%pot%nv), exchemIonsp(rism%pot%nsp)
      integer :: isp,iv, iat
      character(len=clen) :: file
      integer :: unit, iostat
      !field : Fortran likes to right-justify if the text string is too short 
      !        for the specifier.  'field' is used as a work around
      character(len=40) :: field
      unit = freeUnit()
      file = trim(fileroot)//'.td'
      call rism_report_message('outputting thermodynamics to file: '//trim(file))
      open(unit,file=file,status='replace',iostat=iostat)
      if(iostat/=0)&
           call rism_report_error("failed to open :"//trim(file))

      field = "!total compressibility    "
      write(unit,'(a40,a)') field,"Xi  [10-4/MPa]"
!!$      field = "!pressure                 "
!!$      write(unit,'(a40,a)') field,"P   [MPa]"
      field = "!excess free energy       "
      write(unit,'(a40,a)') field,"xA  [kcal/mol]"
      write(unit,'(a)') "&THERMO"
      write (unit,'(A,1p,(g16.8,:,","))')  "XITOT=",rism1d_getCompressibility(rism)&
           *1d-20/(boltzmann*rism%pot%temperature)
!!$      write (unit,'(A,1p,(g16.8,:,","))')  "FE_PRESURE=",rism1d_getPressureFE(rism)&
!!$           *1.d24 * boltzmann*rism%pot%temperature 
!!$      write (unit,'(A,1p,(g16.8,:,","))')  "VIRIAL_PRESURE=",rism1d_getPressureVirial(rism)&
!!$           *1.d24 * boltzmann*rism%pot%temperature 
      write(unit,'(A,1p,(g16.8,:,","))')   "XFE=",rism1d_getFreeEnergy(rism)*KB*rism%pot%temperature
      write(unit,'(a)') "/"
      pmv = rism1d_getPMV(rism)
      exchem = rism1d_getExChem(rism)
      exchemIon = rism1d_getExChemIon(rism)
      exchemsp=0
      exchemIonsp=0
      iv = 0
      do isp=1,rism%pot%nsp
         do iat=1,rism%pot%nat(isp)
            iv = iv + 1
            exchemsp(isp) = exchemsp(isp) + exchem(iv)
            exchemIonsp(isp) = exchemIonsp(isp) + exchemIon(iv)
         enddo
      enddo
      field = "!Molecular Species Contributions"
      write(unit,'(a40)') field
      field = "!partial molar volume"
      write(unit,'(a40,a)') field,"PMV [1/A^3]"
      field = "!excess chemical potential"
      write(unit,'(a40,a)') field,"xmu [kcal/mol]"
      write(unit,'(a)') "&PMV"
      write(unit,'(A,1p,(g16.8,:,","))') "PVM=",pmv
      write(unit,'(A,1p,(g16.8,:,","))') "XMUSP=",exchemsp*KB*rism%pot%temperature
      write(unit,'(A,1p,(g16.8,:,","))') "XMUISP=",exchemIonsp*KB*rism%pot%temperature
      write(unit,'(a)') "/"
      field = "!Site Contributions"
      write(unit,'(a40)') field
      field = "!excess chemical potential"
      write(unit,'(a40,a)') field,"xmu [kcal/mol]"
      write(unit,'(a)') "&XMU"
      write(unit,'(A,1p,(g16.8,:,","))') "XMUV=",exchem*KB*rism%pot%temperature
      write(unit,'(A,1p,(g16.8,:,","))') "XMUIV=",exchemIon*KB*rism%pot%temperature
      write(unit,'(a)') "/"
      close(unit)
    end subroutine writeTD

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Writes the coordination number to file.  Use Gvv for coordination number and
!!!Hvv for excess coordination number
!!!IN:
!!!   vv    : Gvv or Hvv
!!!   suffix: suffix for the end of the file
!!!   descript : description of the column    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine writeRunningCoordination(vv, suffix,descript)
      use rism_util
      use constants, only : pi
      implicit none
      _REAL_, intent(in) :: vv(:,:,:)
      character(len=*), intent(in) :: suffix, descript
      character(len=clen) :: file
      integer :: unit, iostat
      integer :: ivv, iv1, iv2, ir
      unit = freeUnit()
      file = trim(fileroot)//'.'//suffix
      call rism_report_message('outputting running coordination number to file: '//trim(file))
      open(unit,file=file,status='replace',iostat=iostat)
      if(iostat/=0)&
         call rism_report_error("failed to open :"//trim(file))
      call writeVVHeader(unit,&
           "#RISM1D ATOM-ATOM INTERACTIONS: "//trim(descript)//" VS. SEPARATION [A]",17,.true.)

      do ir = 1,nrout
         write (unit,'(1p,5050(1x,e16.8e3))')  (ir-1)*rism%pot%dr, rmExPrec(vv(ir,:,:))
      enddo

      close(unit)
    end subroutine writeRunningCoordination

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Writes the total excess coordination number and site-site structure factor to
!!!file
!!!IN:
!!!   excess   : Excess quantity of sepcies in dimension 1 about species in 
!!!              dimension 2 (nv,nv)
!!!   suffix   : suffix for the file
!!!   descript : description of the column  
!!!   total    : (optional) write out the sum of each row
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine writeTotalExcess(excess,suffix,descript,total)
      use rism_util, only : freeUnit
      implicit none
      _REAL_, intent(in) :: excess(:,:)
      character(len=*),intent(in) :: suffix, descript
      logical, optional, intent(in) :: total
      logical ::optTotal
      character(len=clen) :: file
      integer :: unit, iostat
      integer :: ivv, iv1, iv2, ir
      character(len=CLEN) :: colTitle_fmt, colData_fmt, advance
      unit = freeUnit()

      advance='yes'
      optTotal=.false.
      if(present(total)) optTotal=total
      if(optTotal) advance = 'no'

      file = trim(fileroot)//'.'//suffix
      call rism_report_message('outputting Nvv(R=infty) to file: '//trim(file))
      open(unit,file=file,status='replace',iostat=iostat)
      if(iostat/=0)&
           call rism_report_error("failed to open :"//trim(file))

      write(colTitle_fmt,'(a,i4,a)') '(4x,', rism%pot%nv, '(7x,a4,6x))'
      write(colData_fmt,'(a,i4,a)') '(a4,1p,', rism%pot%nv, '(1x,e16.8e3))'

      !we will assume that atom names are four characters long form simplicity
      write(unit,'(a)') "#"//descript//" of column site about row site"
      write(unit,colTitle_fmt,advance=advance) rism%pot%namev
      if(optTotal)&
           write(unit,'(a)') "   Total charge  "
      do iv1=1,rism%pot%nv
         write (unit,colData_fmt,advance=advance)  rism%pot%namev(iv1),excess(iv1,:)
         if(optTotal)&
              write(unit,'(1p,2(1x,e16.8e3))') sum(excess(iv1,:))
      enddo
      close (unit)
    end subroutine writeTotalExcess

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Writes the total excess coordination number and site-site structure factor to
!!!file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine writeSVV()
      implicit none
      _REAL_, pointer :: wrkvv(:,:)=>NULL()
      !.......................... outputting Svv(k) ..........................
      wrkvv => rism1d_getStructFactor(rism)
      call writeVV(wrkvv,'svv','STRUCTURE FACTOR',k_space=.true.)
      if(safemem_dealloc(wrkvv)/=0)&
         call rism_report_error("memory deallocation failed in writeSVVN00")
    end subroutine writeSVV

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Writes out a header for VV files to describe the contents.  Columns are 
!!!labelled by species and site numbers. All columns should be the same width to
!!!acheive the correct centering of columns
!!!IN:
!!!   unit : unit to write to
!!!   head : One line description of the file
!!!   width : column width in characters  
!!!   full : (optional) If true do the full NV*NV header. If false do the 
!!!          site-site header only.  Default: .false.
!!!   k_space : (optional) if true, data is in reciprocal space and the frequency
!!!            will be printed out.  if false (default), data is in real space
!!!            and the separation will be printed out    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine writeVVHeader(unit,head,width,full,k_space)
      implicit none
      integer, intent(in) :: unit, width
      character(len=*), intent(in) :: head
      logical, optional, intent(in) :: full,k_space
      logical :: all, kspace
      integer,pointer :: interactions(:)=>NULL()
      integer :: ivv, iv1, iv2, isp, isp2, iat, iat2
      !sepbuffer : number of buffer spaces before and after 'separation', the first column
      !pairbuffer : number of buffer spaces before and after each atom pair column
      integer :: sepbuffer(2), pairbuffer(2)
      !form : we will write the format to this string
      !xlabel : SEPARATION or FREQUENCY depending on the values of kspace
      character(len=256) :: form, xlabel
      all = .false.
      if(present(full)) all = full
      kspace = .false.
      if(present(k_space)) kspace = k_space

      xlabel = "SEPARATION"
      if(kspace) xlabel = "FREQUENCY"

      write (unit,'(a)') head
      write(unit,'(a)') "#S=SPECIES, A=ATOM"

      !get the correct order for the labels.  We assume that the internal RISM1D
      !ordering is used
      ivv=0
      if(all)then
         interactions => safemem_realloc(interactions,(4*rism%pot%nv**2))
         do isp=1,rism%pot%nsp
            do iat=1,rism%pot%nat(isp)
               do isp2=1,rism%pot%nsp
                  do iat2=1,rism%pot%nat(isp2)
                     ivv = ivv+1
                     interactions(ivv)=isp
                     ivv = ivv+1
                     interactions(ivv)=iat
                     ivv = ivv+1
                     interactions(ivv)=isp2
                     ivv = ivv+1
                     interactions(ivv)=iat2
                  enddo
               enddo
            enddo
         enddo
      else
         interactions => safemem_realloc(interactions,(4*rism%pot%nvv))
         do isp=1,rism%pot%nsp
            do iat=1,rism%pot%nat(isp)
               do isp2=1,isp
                  do iat2=1,rism%pot%nat(isp2)
                     if(isp2 .eq. isp .and. iat2 .lt. iat) then
                     else                 
                        ivv = ivv+1
                        interactions(ivv)=isp
                        ivv = ivv+1
                        interactions(ivv)=iat
                        ivv = ivv+1
                        interactions(ivv)=isp2
                        ivv = ivv+1
                        interactions(ivv)=iat2
                     endif
                  enddo
               enddo
            enddo
         enddo
      end if
      !get the number of blanks needed to fill the column width
      sepbuffer(1) = width - len("#"//trim(xlabel))
      !use half minus one for the right side of the column
      sepbuffer(2) = sepbuffer(1) - sepbuffer(1)/2 -1
      !use half plus one for the left side of the column
      sepbuffer(1) = sepbuffer(1)/2 + 1
      !do the same for the atom pair columns.  Assume only one character 
      !for the species and site numbers
      pairbuffer(1) = width - len("S1A1:S1A1")
      pairbuffer(2) = pairbuffer(1) - pairbuffer(1)/2 -1 
      pairbuffer(1) = pairbuffer(1)/2 +1

      !write the format to string
      write(form,'(a,i2,a,i2,a,i2,a,i2,a)') '("#",',sepbuffer(1),&
           'x,"'//trim(xlabel)//'",',sepbuffer(2),"x,1000(:,",&
           pairbuffer(1),'x,"S",i1,"A",i1,":S",i1,"A",i1,',pairbuffer(2),'x))'
      !write column headers
      write(unit,form) interactions

      !do it again with atom names
      write(form,'(a,i2,a,i2,a,i2,a,i2,a)') '("#",',sepbuffer(1),&
           'x,"'//trim(xlabel)//'",',sepbuffer(2),"x)"
      write(unit,form,advance='no')
      ivv=0
      if(all)then
         do iv1 = 1, rism%pot%nv
            do iv2 = 1, rism%pot%nv
               pairbuffer(1) = width - len(trim(rism%pot%namev(iv1))//":"//trim(rism%pot%namev(iv2)))
               pairbuffer(2) = pairbuffer(1) - pairbuffer(1)/2 -1 
               pairbuffer(1) = pairbuffer(1)/2 +1
               write(form,'(a,i2,a,i2,a)') '(',pairbuffer(1),'x,a,',pairbuffer(2),'x)'
               write(unit,form,advance='no')trim(rism%pot%namev(iv1))//":"//trim(rism%pot%namev(iv2))
            end do
         end do
      else
         do iv1 = 1, rism%pot%nv
            do iv2 = 1, iv1
               pairbuffer(1) = width - len(trim(rism%pot%namev(iv1))//":"//trim(rism%pot%namev(iv2)))
               pairbuffer(2) = pairbuffer(1) - pairbuffer(1)/2 -1 
               pairbuffer(1) = pairbuffer(1)/2 +1
               write(form,'(a,i2,a,i2,a)') '(',pairbuffer(1),'x,a,',pairbuffer(2),'x)'
               write(unit,form,advance='no')trim(rism%pot%namev(iv1))//":"//trim(rism%pot%namev(iv2))
            end do
         end do
      end if
      write(unit,*)
      if(safemem_dealloc(interactions)/=0)&
           call rism_report_error("writeVVHeader: could not deallocate interactions")
    end subroutine writeVVHeader

  end module rism1d_m

  program rism1doo
    use rism1d_m
    implicit none
    call initialize()
    call solve()
    call output()
    call finalize()
  end program rism1doo
