! <compile=optimized>

!The 3D-RISM-KH software found here is copyright (c) 2010 by Andriy Kovalenko,
!Tyler Luchko and David A. Case.
!
!This program is free software: you can redistribute it and/or modify it
!under the terms of the GNU General Public License as published by the Free
!Software Foundation, either version 3 of the License, or (at your option)
!any later version.
!
!This program is distributed in the hope that it will be useful, but
!WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
!for more details.
!
!You should have received a copy of the GNU General Public License in the
!../../LICENSE file.  If not, see <http://www.gnu.org/licenses/>.
!
!Users of the 3D-RISM capability found here are requested to acknowledge
!use of the software in reports and publications.  Such acknowledgement
!should include the following citations:
!
!1) A. Kovalenko and F. Hirata. J. Chem. Phys., 110:10095-10112  (1999); 
!ibid. 112:10391-10417 (2000).   
!
!2) A. Kovalenko,  in:  Molecular  Theory  of  Solvation,  edited  by  
!F. Hirata  (Kluwer Academic Publishers, Dordrecht, 2003), pp.169-275.  
!
!3) T. Luchko, S. Gusarov, D.R. Roe, C. Simmerling, D.A. Case, J. Tuszynski,
!and  A. Kovalenko, J. Chem. Theory Comput., 6:607-624 (2010). 

#include "../include/dprec.fh"
!***********************************************************************
! Site-site correlations and thermodynamics of a molecular mixture
! o Site-site PDFs, TCFs, DCFs, and susceptibility
! o compressibility
! o site-site RISM equation
! o HNC, KH, VM0, PY closures
! o solving for the Direct Correlation Functions
! o using the linearly spaced 1D-FFT
! o calculating the residual as r*(G_clos(r)-1-H_rism(r))
! o compact triangle ordering of site-site vectors
! o modified direct inversion in the iterative subspace (MDIIS)
! o Coulomb + 12-6 Lennard-Jones site-site potentials
! o units:  energy       [kT]
!           distances    [A]           (Angstroms)
!           site charges [sqrt(kT A)]
!           temperature  [K]
!           density      [#/A^3]
! o To convert [e] to [sqrt(kT A)] *sqrt(COULOMB_CONST_E/KB/temperature)
!***********************************************************************

module rism1d_c
  use safemem
  use rism1d_potential_c
  use rism1d_closure_c
  use rism_report_c
  implicit none
  integer,private ,parameter :: charlen = 8, maxep0=4
  type rism1d
     !closureID : string identifier for the closure
     character(len=charlen) :: closureID=''
     !savefile  : save file to allow restarts
     character(len=256) :: savefile='rism.sav'

     !pot : potential object
     type(rism1d_potential) :: pot

     !closure : Closure object.  Current options are 'KH', 'HNC', 'PY', VM0'
     type(rism1d_closure) :: closure

     !maxste : maximum number of iterations
     !Mdiis_nvec : number of MDIIS vectors
     integer :: Mdiis_nvec=0

     !mdiis_del       : MDIIS step size
     _REAL_ :: mdiis_del

     !gvv   : pair correlation function
     !hvv   : total correlation function
     !cvv   : direct correlation function
     !cvvres: residual direct correlation function (residual between iterations)
     !cvvWRK  : last Mdiis_nvec steps of cvv
     !cvvresWRK: last Mdiis_nvec steps of cvvres
     _REAL_, pointer :: &
          gvv(:,:)=>NULL(), hvv(:,:)=>NULL(), &
          cvvWRK(:,:,:)=>NULL(), cvvresWRK(:,:,:)=>NULL()

     !convenience pointers
     !cvv   : direct correlation function.  Points to cvvWRK(:,:,1)
     !cvvres: residual direct correlation function (residual between iterations)
     !        Points to cuvrei(:,:,1)
     _REAL_,pointer :: cvv(:,:)=>NULL(), cvvres(:,:)=>NULL()

     !Computing the k-space contribution to the free energy and the pressure 
     !along the free energy route can be done at the same for the same cost
     !as a single calculation.  We store the values here.  These are reset each
     !solution to indicate that they need to be recalculated
  end type rism1d

  interface rism1d_new
     module procedure new_
  end interface rism1d_new

  public rism1d_new, rism1d_destroy, rism1d_solve, rism1d_addSpecies,&
       rism1d_getInvDebyeLen, rism1d_getCompressibility , rism1d_getDelHvLimit,& 
       rism1d_getExNumber, rism1d_getStructFactor, rism1d_getRunNumber, &
       rism1d_getRunExNumber, rism1d_getPressureVirial, rism1d_getPressureFE, rism1d_getFreeEnergy, &
       rism1d_getPMV, rism1d_getExChem
  private new_, rxrism, r1rism, sanity_check
contains

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Initializes the in rism1d object
!!!IN:
!!!   this : rism1d object
!!!   theory : 'xrism' or 'drism'
!!!   closure : 'kh','hnc', 'py', 'vm0', 'psen' or 'v*'
!!!   PSEn_ order : (optional, required for PSEN) Order for PSE-n calculation
!!!   coeff  : (optional) coefficients for the selected closure
!!!   vstarB : (optional) 'B' parameter for Verlet (V*) closure
!!!   temperature : system temperature in [K]
!!!   dielconst : dielconst for DRISM calculations
!!!   smear  : electrostatic smear parameter
!!!   adbcor : ???  
!!!   nr     : number of grid points
!!!   dr     : grid spacing
!!!   Mdiis_nvec    : number of MDIIS vectors
!!!   mdiis_del    : MDIIS step size
!!!   savefile : name of the restart file
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine new_(this, theory, closure, PSEn_order, coeff, temperature, dielconst, smear, adbcor, &
       nr, dr, Mdiis_nvec,mdiis_del,savefile)
    use constants, only : pi
    use rism_util, only : caseup
    implicit none
    type(rism1d), intent(inout) :: this
    character(len=*), intent(in) :: theory, closure
    integer, optional, intent(in) :: PSEn_order
    _REAL_, optional, intent(in) :: coeff(:)
    character(len=*), optional, intent(in) :: savefile
    integer, intent(in) :: nr, Mdiis_nvec
    _REAL_, intent(in) :: temperature, dielconst, smear, adbcor, dr, mdiis_del
    integer :: factored_nr
    integer :: l_psen_order =0
    _REAL_  :: l_vstarA=0, l_vstarB=0

    call caseup(closure)
    this%closureID = closure
    this%Mdiis_nvec = Mdiis_nvec
    this%mdiis_del = mdiis_del

    call sanity_check(this)

    if(present(PSEn_order)) l_psen_order=PSEn_order
    if(present(coeff)) then
       call rism1d_closure_new(this%closure,this%closureID,this%pot,l_PSEn_order,&
            coeff)
    else
       call rism1d_closure_new(this%closure,this%closureID,this%pot,l_PSEn_order)
    end if
    call rism1d_potential_new(this%pot,theory,temperature,dielconst,smear,adbcor,nr,dr)

    if(present(savefile))then
       this%savefile = trim(savefile)
    end if
    
  end subroutine new_

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Add solvent species.  Allocates necessary memory to accomodate the additional
!!!species
!!!IN:
!!!   this : rism1d object
!!!   mdl  : solvMDL object
!!!   density : number density (A^{-3})
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine rism1d_addSpecies(this,mdl,density)
    use solvMDL_c
    implicit none
    type(rism1d), intent(inout) :: this
    type(solvMDL), intent(in) :: mdl
    _REAL_, intent(in) :: density
    integer*8 :: memstats(10)

    call rism1d_potential_addSpecies(this%pot,mdl,density)

    !data arrays in the original 1D-RISM code had an offset of 0 in the first index (like C).
    !The offset is 1 in the current code (like most Fortran).

    this%gvv => safemem_realloc(this%gvv,this%pot%nr,this%pot%nvv)
    this%hvv => safemem_realloc(this%hvv,this%pot%nr,this%pot%nvv)
    this%cvvWRK => safemem_realloc(this%cvvWRK,this%pot%nr,this%pot%nvv,this%Mdiis_nvec)
    this%cvvresWRK => safemem_realloc(this%cvvresWRK,this%pot%nr,this%pot%nvv,this%Mdiis_nvec)

    this%cvv => this%cvvWRK(:,:,1)
    this%cvvres => this%cvvresWRK(:,:,1)
  end subroutine rism1d_addSpecies

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Solves the 1D-RISM system of equations.
!!!IN:
!!!   this : rism1d object
!!!   maxste : max number of iterations
!!!   ksave  : frequency of restart file saves
!!!   kshow  : frequency of progress updates
!!!   tolerance    : tolerance
!!!OUT:
!!!    .true. if converged.  .false. otherwise.  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_solve(this, ksave, kshow, maxste, tolerance) result(converged)
    implicit none
    type(rism1d), intent(inout) :: this
    integer, intent(in) :: maxste, ksave, kshow
    _REAL_, intent(in) :: tolerance
    logical :: converged
    _REAL_ :: charge
    integer :: err

    !................ for MV0 closure, solving RISM at Q=0 .................
    if(rism1d_closure_isCharged(this%closure))then
       charge=1d0
    else
       charge=0d0
    end if
    call rism1d_potential_calc(this%pot,charge)
    if(rism1d_closure_type(this%closure) .eq. "MV0")&
         call rism1d_mv0_calcUvv(this%closure%mv0, this%pot%nr, this%pot%dr, &
         this%pot%qv, this%pot%epsv, this%pot%rminv, this%pot%smear)

    !........................ solving RISM by MDIIS ........................
    converged =  rxrism(this,ksave,kshow,maxste,tolerance)

  end function rism1d_solve

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Returns the inverse Debye length (1/A) of the solvent
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getInvDebyeLen(this) result(kappa)
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_ :: kappa
    kappa=rism1d_potential_getInvDebyeLen(this%pot)
  end function rism1d_getInvDebyeLen

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Returns the compressibility of the solvent [A^3]
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getCompressibility(this) result(xikt)
    use constants, only : pi
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_ :: xikt
    xikt = rism1d_closure_getCompressibility(this%closure,this%cvv)
  end function rism1d_getCompressibility

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Returns the the extrapolated value for
!!!DelHv=-Lim_k->0 ( Sum_v1 Qv1*Xv1v2(k)4pi/k^2 - hlkv0 )
!!!This is used by 3D-RISM long range asymptotics
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getDelHvLimit(this) result(delhv0)
    use constants, only : pi
    use rism_util, only : poly_interp
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_ :: delhv0(this%pot%nv)
    delhv0 = rism1d_potential_getDelHvLimit(this%closure,this%hvv)
  end function rism1d_getDelHvLimit

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Returns a pointer to the site-site susceptibility.
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getSusceptibility(this) result(xvv)
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_, pointer :: xvv(:,:,:)
    call rism1d_closure_calcXvv(this%closure,this%hvv)
    xvv => this%closure%xvv
  end function rism1d_getSusceptibility

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Returns an NV X NV array of total excess coordination numbers excluding
!!!multiplicity
!!!IN:
!!!   this : rism1d object
!!!OUT:
!!!    nv*nv array (iv1,iv2) of the excess of site iv2 around iv1
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getExNumber(this) result(exvv)
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_ :: exvv(this%pot%nv,this%pot%nv)
    exvv = rism1d_closure_getExNumber(this%closure,this%hvv)
  end function rism1d_getExNumber

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Returns a pointer to an NR X NVV array of structure factors.  This memory 
!!!must be freed (preferably with safemem_dealloc) as it is not freed locally or
!!!after the object instance is destroyed.
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getStructFactor(this) result(svv)
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_, pointer :: svv(:,:)
    svv => rism1d_closure_getStructFactor(this%closure,this%hvv)
  end function rism1d_getStructFactor

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Returns a pointer to an NR X NVV array of the running site-site excess number.
!!!This is the excess number of a site within a given radius. The memory for 
!!!this pointer must be freed (preferably with safemem_dealloc) as it is not 
!!!freed locally or after the object instance is destroyed.
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getRunExNumber(this) result(exnvv)
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_, pointer :: exnvv(:,:,:)
    nullify(exnvv)
    exnvv => rism1d_closure_getRunExNumber(this%closure,this%gvv)
  end function rism1d_getRunExNumber

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Returns a pointer to an NR X NVV array of the running site-site number.
!!!This is the number of a site within a given radius. The memory for 
!!!this pointer must be freed (preferably with safemem_dealloc) as it is not 
!!!freed locally or after the object instance is destroyed.
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getRunNumber(this) result(nvv)
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_, pointer :: nvv(:,:,:)
    nullify(nvv)
    nvv => rism1d_closure_getRunNumber(this%closure,this%gvv)
  end function rism1d_getRunNumber

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculates the pressure in [kT / A^3] of the system using the virial 
!!!path.  To convert to Pacals, for example, multiply by 
!!!1.d30 * kb * temperature
!!!where kb is Boltzmann's constant [j/K] and temperature is in [K]
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getPressureVirial(this) result(pressure)
    use safemem
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_ :: pressure
    pressure = rism1d_closure_getPressureVirial(this%closure,this%gvv)
  end function rism1d_getPressureVirial

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculates the pressure in [kT / A^3] of the system using the free energy 
!!!path.  To convert to Pacals, for example, multiply by 
!!!1.d30 * kb * temperature
!!!where kb is Boltzmann's constant [j/K] and temperature is in [K]
!!!IN:
!!!   this : rism1d object
!!!   rhotrgt : (optional) The final, physical density for thermodynamics.  This
!!!             can be used as an effective correction for some closures that
!!!             over estimate the pressure (e.g. HNC and KH)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getPressureFE(this,rhotrgt) result(pressure)
    use safemem
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_, optional :: rhotrgt 
    _REAL_ :: pressure
    if(present(rhotrgt))then
       pressure = rism1d_closure_getPressureFE(this%closure,this%gvv,this%cvv,rhotrgt)
    else
       pressure = rism1d_closure_getPressureFE(this%closure,this%gvv,this%cvv)
     end if
   end function rism1d_getPressureFE

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculates the freeEnergy in kT
!!!IN:
!!!   this : rism1d object
!!!   rhotrgt : (optional) The final, physical density for thermodynamics.  This
!!!             can be used as an effective correction for some closures that
!!!             over estimate the pressure (e.g. HNC and KH)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getFreeEnergy(this,rhotrgt) result(fe)
    use safemem
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_, optional :: rhotrgt 
    _REAL_ :: fe
    if(present(rhotrgt))then
       fe = rism1d_closure_getFreeEnergy(this%closure,this%gvv,this%cvv,rhotrgt)
    else
       fe = rism1d_closure_getFreeEnergy(this%closure,this%gvv,this%cvv)
     end if
  end function rism1d_getFreeEnergy

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculates the partial molar volume of each species in A^3
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getPMV(this) result(pmv)
    use constants, only : pi
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_ :: pmv(this%pot%nsp)
    pmv = rism1d_closure_getPMV(this%closure,this%cvv)
  end function rism1d_getPMV

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculates the excess chemical potential in kT for each site.
!!!IN:
!!!   this : rism1d object
!!!   rhotrgt : (optional) The final, physical density for thermodynamics.  This
!!!             can be used as an effective correction for some closures that
!!!             over estimate the pressure (e.g. HNC and KH)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getExChem(this,rhotrgt) result(exchem)
    use safemem
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_, optional :: rhotrgt 
    _REAL_ :: exchem(this%pot%nv)
    if(present(rhotrgt))then
       exchem = rism1d_closure_getExChem(this%closure,this%gvv,this%cvv,rhotrgt)
    else
       exchem = rism1d_closure_getExChem(this%closure,this%gvv,this%cvv)
    end if
  end function rism1d_getExChem

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculates the ionic(?) excess chemical potential in kT for each site.
!!!IN:
!!!   this : rism1d object
!!!   rhotrgt : (optional) The final, physical density for thermodynamics.  This
!!!             can be used as an effective correction for some closures that
!!!             over estimate the pressure (e.g. HNC and KH)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_getExChemIon(this,rhotrgt) result(exchem)
    use safemem
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_, optional :: rhotrgt 
    _REAL_ :: exchem(this%pot%nv)
    if(present(rhotrgt))then
       exchem = rism1d_closure_getExChemIon(this%closure,this%gvv,this%cvv,rhotrgt)
    else
       exchem = rism1d_closure_getExChemIon(this%closure,this%gvv,this%cvv)
    end if
  end function rism1d_getExChemIon

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculates Bvv (bridge function) for the closure used.The memory for 
!!!this pointer must be freed (preferably with safemem_dealloc) as it is not 
!!!freed locally or after the object instance is destroyed.
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_bvv(this) result(bvv)
    implicit none
    type(rism1d), intent(in) :: this
    _REAL_, pointer :: bvv(:,:)
    bvv=>rism1d_closure_bvv(this%closure,this%gvv, this%cvv)
  end function rism1d_bvv

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Frees memory.  Makes object pristine
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine rism1d_destroy(this)
    use safemem
    implicit none
    type(rism1d), intent(inout) :: this
    call rism1d_closure_destroy(this%closure)
    call rism1d_potential_destroy(this%pot)
    this%Mdiis_nvec=0
    if(safemem_dealloc(this%gvv) /=0) call rism_report_error("RISM1D: failed to deallocate GVV")
    if(safemem_dealloc(this%hvv) /=0) call rism_report_error("RISM1D: failed to deallocate HVV")
    if(safemem_dealloc(this%cvvWRK) /=0) call rism_report_error("RISM1D: failed to deallocate CVVWRK")
    if(safemem_dealloc(this%cvvresWRK) /=0) call rism_report_error("RISM1D: failed to deallocate CVVRESWRK")
    nullify(this%cvv)
    nullify(this%cvvres)
  end subroutine rism1d_destroy

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                               PRIVATE
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Check user parameters
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine sanity_check(this)
    implicit none
    type(rism1d), intent(inout) :: this
    
    !perhaps this can be moved to MDIIS code...

    if(this%mdiis_del <= 0)&
         call rism_report_error("MDIIS_MDIIS_DEL must be > 0")
    if(this%mdiis_nvec <=0)&
         call rism_report_error("MDIIS_mdiis_nvec must be > 0")

  end subroutine sanity_check

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Relaxing the RISM equation
!!!IN:
!!!   this : rism1d object
!!!   ksave : save restart frequency
!!!   progress : write progress frequency
!!!   maxstep : max number of steps
!!!   tolerance : target tolerance
!!!OUT:
!!!    .true. if converged.  .false. otherwise.  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function  rxrism (this,ksave,progress,maxstep,tolerance) result (converged)
    use mdiis_c
    use rism_util, only : poly_interp, freeUnit
    use constants, only : pi
    implicit none
    type(rism1d), intent(inout) :: this
    integer, intent(in) ::  maxstep,ksave,progress
    _REAL_, intent(in) :: tolerance
    logical :: converged
    type(mdiis), save :: mdiis_o
    integer :: unit
    logical ::  saved, start
    integer ::  nro, nvvo, ir,ivv, istep, kis
    _REAL_ ::  residual, r, err0

    !.................... k=0 extrapolation array size .....................
    _REAL_ ::  ep0(maxep0)

    _REAL_ :: k

    integer :: mdiis_method =2

    if(ksave < -1) &
         call rism_report_error("KSAVE must be >= -1") 
    if(progress < 0) &
         call rism_report_error("PROGRESS must be >= 0") 
    if(maxstep <= 0) &
         call rism_report_error("MAXSTEP must be > 0") 
    if(tolerance <= 0) &
         call rism_report_error("TOLERANCE must be > 0") 

    converged = .false.
    call mdiis_new(mdiis_o,mdiis_method,&
         this%mdiis_del,tolerance,this%cvvWRK,this%cvvresWRK,&
         this%pot%nr*this%pot%nvv, this%Mdiis_nvec)
    unit = freeUnit()

    !.................. checking extrapolation grid size ...................
    if (maxep0 > this%pot%nr) &
       call rism_report_error('(a,i4,a,i4)','RXRISM: k->0 array size MaxEp0=',&
            maxep0,' > Nr=',this%pot%nr)

    !......................... reading saved r*Cvv .........................
    inquire (file=this%savefile,exist=saved)
    if (saved)  then
       call rism_report_message('reading saved Cvv file: '//trim(this%savefile))
       open (unit,file=this%savefile,form='unformatted',status='old')
       read (unit)  nro, nvvo
       if (nro /= this%pot%nr .OR. nvvo /= this%pot%nvv)  then
          close (unit)
          call rism_report_error('(a,i6,a,i4,a,i6,a,i4)', &
               'RXRISM: dimensions discrepancy with saved data' &
               //'               Nr         Nvv  ' &
               //'  actual    ',this%pot%nr,'       ',this%pot%nvv, &
               '   saved    ',nro,'       ',nvvo)
       endif
       read (unit)  this%cvv
       close (unit)

       !....................... initial guess of r*Cvv ........................
    else
       do ivv=1,this%pot%nvv
          do ir=2,this%pot%nr
             this%cvv(ir,ivv) = - this%pot%ulrvv(ir,ivv)
          enddo
       enddo

    endif

    !............................ relaxing RISM ............................
    call rism_report_message('relaxing RISM:')
    converged = .false.
    start = .true.


    do istep=1,maxstep

       !the first element is never set but is read by the closure and MDIIS.  This
       !ensures that is it initialized properly.  It doesn't change the result but
       !does allow Valgrind to run cleanly
       this%cvv(1,:)=0
       this%cvvres(1,:)=0
       !..................... one relaxation step of RISM .....................
       call  r1rism (this,residual,tolerance, start,converged, mdiis_o)
       !................. screen outputting relaxation steps ..................
       if (progress /= 0)  then
          if (converged .or. (progress /= 0 .AND. mod(istep,progress) == 0))  then
             call rism_report_message('(a,i4,a,1pe24.16,a,i3)',"step=",istep,&
                  "     Res=",residual,"     MDIIS=",getCurrentNVec(mdiis_o))
             call flush(6)
          endif
       endif

       !............ saving intermediate and last relaxation steps ............
       if ((converged .and. ksave/=0) .OR. (ksave > 0 .AND. mod(istep,ksave) == 0))  then
          call rism_report_message('saving Cvv to file: '//trim(this%savefile)//' ...')
          open(unit,file=this%savefile,form='unformatted',status='unknown')
          write (unit)  this%pot%nr, this%pot%nvv
          write (unit)  this%cvv
          close (unit)
          call rism_report_message('done.')
       endif

       if(converged .and. this%closureID .eq. "MV0" &
            .and. .not. rism1d_closure_isCharged(this%closure))then
          converged = .false.
          call rism1d_closure_useCharged(this%closure,.true.)
          call rism1d_potential_calc(this%pot,1d0)
          call mdiis_destroy(mdiis_o)
          this%cvv=>this%cvvWRK(:,:,1)
          this%cvvres=>this%cvvresWRK(:,:,1)
          call mdiis_new(mdiis_o,mdiis_method,&
               this%mdiis_del,tolerance,this%cvvWRK,this%cvvresWRK,&
               this%pot%nr*this%pot%nvv, this%Mdiis_nvec)
          do ivv=1,this%pot%nvv
             do ir=2,this%pot%nr
                this%cvv(ir,ivv) = - this%pot%ulrvv(ir,ivv)
             enddo
          enddo
       end if

       !............... exiting relaxation loop on convergence ................
       if (converged)  exit
    enddo
    if(.not. converged) &
       call rism_report_error('(a,i5)','RXRISM: reached steps limit Maxstep=',maxstep)

    !....................... eliminating R from Cvv ........................
    do ivv=1,this%pot%nvv
       do ir=2,this%pot%nr
          r = (ir-1)*this%pot%dr
          this%cvv(ir,ivv) = this%cvv(ir,ivv) / r
       enddo
    enddo

    !................. extrapolating Cvv(r=0) and Gvv(r=0) .................
    do ir=1,maxep0
       r = ir*this%pot%dr
       ep0(ir) = r
    enddo

    do ivv=1,this%pot%nvv
       call  poly_interp (ep0,this%cvv(2:1+maxep0,ivv),maxep0, 0.d0,this%cvv(1,ivv), err0)
    enddo

    do ivv=1,this%pot%nvv
       call  poly_interp (ep0,this%gvv(2:1+maxep0,ivv),maxep0, 0.d0,this%gvv(1,ivv), err0)
    enddo

    !....................... eliminating K from Hvv ........................
    do ivv=1,this%pot%nvv
       do ir=2,this%pot%nr
          k = (ir-1)*this%pot%dk
          this%hvv(ir,ivv) = this%hvv(ir,ivv) / k
       enddo
    enddo

    !....................... extrapolating Hvv(k=0) ........................
    do ir=1,maxep0
       k = ir*this%pot%dk
       ep0(ir) = k
    enddo

    do ivv=1,this%pot%nvv
       call  poly_interp (ep0,this%hvv(2:1+maxep0,ivv),maxep0, 0.d0,this%hvv(1,ivv), err0)
    enddo

    call mdiis_destroy(mdiis_o)
    this%cvv=>this%cvvWRK(:,:,1)
    this%cvvres=>this%cvvresWRK(:,:,1)
  end function rxrism

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! One relaxation step for the direct correlation function        *
!!! in the RISM equation with the HNC or PLHNC closure,          *
!!! k*Hvv(k) = (k-(Wvv+Zvv)^(tr)*k*Cvv*Rho_v)^(-1)            *
!!! * k*(Wvv+Zvv)^(tr)*k*Cvv*(Wvv+Zvv) + k*Zvv ,   *
!!! Gvv(r) = exp(-b*Uvv + Hvv - Cvv)  or                      *
!!! Gvv(r) = exp(X), X<0                                      *
!!! = 1 + X,  X>0                                      *
!!! X = - b*Uvv + Hvv - Cvv                                 *
!!!OUT:
!!!    .true. if converged.  .false. otherwise.  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine  r1rism (this,residual,tolerance,start,converged,mdiis_o)
    use mdiis_c
    implicit none
    type(rism1d), intent(inout) :: this
    _REAL_, intent(out) :: residual
    _REAL_, intent(in) :: tolerance
    logical, intent(out) ::  start, converged
    type(mdiis), intent(inout) :: mdiis_o

    integer :: err

    integer ::  ir, ivv, iv1,iv2,iv3
    _REAL_ ::  dind, r,t, arg, tvv, tvvr, al

    integer ::  indx(this%pot%nv)
    _REAL_ ::  ak(this%pot%nv,this%pot%nv), bk(this%pot%nv,this%pot%nv), wck(this%pot%nv,this%pot%nv)

    !............................. FFT>K r*Cvv .............................
    do ivv=1,this%pot%nvv
       do ir=2,this%pot%nr
          this%cvvres(ir,ivv) = this%cvv(ir,ivv) + this%pot%ulrvv(ir,ivv)
       enddo
    enddo

    do ivv=1,this%pot%nvv
       call  sinfti (this%cvvres(2,ivv),this%pot%nr-1, this%pot%dr, +1)
    enddo

    do ivv=1,this%pot%nvv
       do ir=2,this%pot%nr
          this%cvvres(ir,ivv) = this%cvvres(ir,ivv) - this%pot%ulkvv(ir,ivv)
       enddo
    enddo

    !................... enumerating K-space gridpoints ....................
    do ir=2,this%pot%nr
       t = (ir-1)*this%pot%dk

       !............. loading and symmetrizing k*Cvv at current K .............
       ivv = 0
       do iv2=1,this%pot%nv
          do iv1=1,iv2
             ivv = ivv + 1
             bk(iv1,iv2) = this%cvvres(ir,ivv)
             bk(iv2,iv1) = bk(iv1,iv2)
          enddo
       enddo

       !.................... getting (Wvv+Zvv)^(tr)*k*Cvv .....................
       do iv2=1,this%pot%nv
          do iv1=1,this%pot%nv
             wck(iv1,iv2) = 0.d0
             do iv3=1,this%pot%nv
                wck(iv1,iv2) = wck(iv1,iv2) + this%pot%wzvv(ir,iv3,iv1)*bk(iv3,iv2)
             enddo
          enddo
       enddo

       !............. getting Bk=k*(Wvv+Zvv)^(tr)*k*Cvv*(Wvv+Zvv) .............
       do iv2=1,this%pot%nv
          do iv1=1,this%pot%nv
             bk(iv1,iv2) = 0.d0
             do iv3=1,this%pot%nv
                bk(iv1,iv2) = bk(iv1,iv2) + wck(iv1,iv3)*this%pot%wzvv(ir,iv3,iv2)
             enddo
             bk(iv1,iv2) = t*bk(iv1,iv2)
          enddo
       enddo

       !................. getting (Wvv+Zvv)^(tr)*k*Cvv*Rho_v ..................
       do iv2=1,this%pot%nv
          do iv1=1,this%pot%nv
             wck(iv1,iv2) = wck(iv1,iv2)*this%pot%rhov(iv2)
          enddo
       enddo

       !.............. getting Ak=(k-(Wvv+Zvv)^(tr)*k*Cvv*Rho_v) ..............
       do iv2=1,this%pot%nv
          do iv1=1,this%pot%nv
             ak(iv1,iv2) = - wck(iv1,iv2)
          enddo
       enddo

       do iv1=1,this%pot%nv
          ak(iv1,iv1) = t + ak(iv1,iv1)
       enddo

       !.................... calculating k*Hvv=Ak^(-1)*Bk .....................
       call DGETRF(this%pot%nv,this%pot%nv,ak,this%pot%nv,indx,err)
       if(err > 0)then
          call rism_report_error("LU-factorization failed.  U = 0")
       elseif(err<0)then
          err = err*(-1)
          call rism_report_error("LU-factorization failed.")
       endif
       call DGETRS('N', this%pot%nv, this%pot%nv,ak,this%pot%nv,indx,bk,this%pot%nv,err)
       if(err < 0)then
          err = err*(-1)
          call rism_report_error("Linear equation solver failed.")
       endif

       !......................... getting k*Hvv+k*Zvv .........................
       do iv2=1,this%pot%nv
          do iv1=1,this%pot%nv
             bk(iv1,iv2) = bk(iv1,iv2) + this%pot%zkvv(ir,iv1,iv2)
          enddo
       enddo

       !........................ unloading k*Hvv+k*Zvv ........................
       ivv = 0
       do iv2=1,this%pot%nv
          do iv1=1,iv2
             ivv = ivv + 1
             this%hvv(ir,ivv) = bk(iv1,iv2)
          enddo
       enddo

    enddo

    !............................. FFT>R k*Hvv .............................
    do ivv=1,this%pot%nvv
       do ir=2,this%pot%nr
          this%cvvres(ir,ivv) = this%hvv(ir,ivv) - this%pot%hlkvv(ir,ivv)
       enddo
    enddo
    do ivv=1,this%pot%nvv
       call  sinfti (this%cvvres(2,ivv),this%pot%nr-1, this%pot%dr, -1)
    enddo

    do ivv=1,this%pot%nvv
       do ir=2,this%pot%nr
          this%cvvres(ir,ivv) = this%cvvres(ir,ivv) + this%pot%hlrvv(ir,ivv)
       enddo
    enddo

    do ivv=1,this%pot%nvv
       do ir=2,this%pot%nr
          r = (ir-1)*this%pot%dr
          this%cvvres(ir,ivv) = this%cvvres(ir,ivv)/r
          this%cvv(ir,ivv) = this%cvv(ir,ivv)/r
       end do
    end do
    call rism1d_closure_gvv(this%closure,this%gvv, this%cvvres, this%cvv)
    do ivv=1,this%pot%nvv
       do ir=2,this%pot%nr
          r = (ir-1)*this%pot%dr
          this%cvvres(ir,ivv) = (this%gvv(ir,ivv)-1d0 - this%cvvres(ir,ivv))*r
          this%cvv(ir,ivv) = this%cvv(ir,ivv)*r
       end do
    end do
    !........................ performing MDIIS step ........................

    call mdiis_advance(mdiis_o,residual,converged)
    this%cvv=>this%cvvWRK(:,:,mdiis_getWorkVector(mdiis_o))
    this%cvvres=>this%cvvresWRK(:,:,mdiis_getWorkVector(mdiis_o))
  end subroutine r1rism

end module rism1d_c
