! <compile=optimized>
#include "../include/dprec.fh"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Potential class for 1D-RISM.  Used to calculate/store quantities that are
!!!potential dependent and do not change during the calculation.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
module rism1d_potential_c
  use safemem
  implicit none
  integer,private ,parameter :: charlen = 8
  type rism1d_potential
     !theory  : type of RISM theory.  Current options are 'rism' (XRISM) and 
     !          drism (dielectrically conistent RISM)
     character(len=charlen) :: theory=''

     !dr : r-space grid spacing
     !dk : k-space grid spacing
     _REAL_ :: dr=0, dk=0

     !nr  : 1D grid size
     integer :: nr = 0

     !temperature : (input) solvent temperature
     !dielconst : (input) dielectric constant
     !smear     : (input) charge smearing parameter for long range asymptotics
     !adbcor    : (input) parameter use in calculating Zvv(k)...
     _REAL_ :: temperature, dielconst, smear, adbcor

     !nsp : Number of molecular species
     !nv  : Number of sites
     !nvv : Number of site-site interactions (nv*(nv+1)/2)
     integer :: nsp=0, nv=0,nvv=0

     !namev : atom site names
     character(len=4), pointer :: namev(:)=>NULL()

     !nat :: number of atom types for each species
     !mta :: multiplicity for each atom in each type
     !mtv :: multiplicity for each atom by site
     !jvv :: triangle packing pointer
     integer,pointer ::  nat(:)=>NULL(), mta(:,:)=>NULL() , mtv(:)=>NULL(), &
          jvv(:,:)=>NULL()

     !kappa     : (output) inverse Debye length [1/A]
     _REAL_ :: kappa=HUGE(1d0)

     !rhosp : number density of each species in [A^{-3}]
     !rhov  : number density of each site in [A^{-3}]
     !qsp   : charge of each species [sqrt(kT A)]
     !qat   : charge of each site in each species [sqrt(kT A)]
     !qv    : charge of each site [sqrt(kT A)]
     !qspv  : charge of the parent species for each site [sqrt(kT A)]
     !epsv  : LJ epsilon for each site in [kT]
     !rminv : LJ sigma* (rmin) for each site in [A]
     !rma   : Positions of all atoms in each species
     !wlmvv :
     !wvv   : Omega_VV. Intramolecular distance matrix.
     !wzvv  :
     !zkvv  :
     !ulrvv : r-space long range smeared Coulomb site-site potential
     !ulkvv : k-space long range smeared Coulombsite-site potential
     !hlrvv : r-space long range smeared potential contribution to (asymptotics) of hvv
     !hlkvv : k-space long range smeared potential contribution to (asymptotics) of hvv
     !duvv  : r-psace derivative of the potential
     _REAL_, pointer :: &
          rhosp(:)=>NULL(), rhov(:)=>NULL(), &
          qsp(:)=>NULL(), qat(:,:)=>NULL(), qv(:)=>NULL(), qspv(:)=>NULL(), &
          epsv(:)=>NULL(), &
          rminv(:)=>NULL(), &
          rma(:,:,:,:)=>NULL(), &
          wlmvv(:,:,:)=>NULL(), &
          wvv(:,:,:)=>NULL(), &
          wzvv(:,:,:)=>NULL(), zkvv(:,:,:)=>NULL(), &
          uvv(:,:)=>NULL(), &
          ulrvv(:,:)=>NULL(), ulkvv(:,:)=>NULL(), &
          hlrvv(:,:)=>NULL(), hlkvv(:,:)=>NULL(), &
          duvv(:,:)=>NULL()
  end type rism1d_potential
  private intramolecular_dist, sanity_check
contains
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Initializes new potential
!!!IN:
!!!   this : potential object  
!!!   dr   : r-space grid spacing
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine rism1d_potential_new(this,theory,temperature,dielconst,smear,adbcor,nr,dr)
    use constants, only : pi
    use rism_util, only : caseup
    implicit none
    type(rism1d_potential), intent(inout) ::this
    character(len=*), intent(in) :: theory
    integer, intent(in) :: nr
    _REAL_, intent(in) :: temperature,dielconst,smear,adbcor, dr
    call caseup(this%theory)
    this%theory = trim(theory)
    this%temperature = temperature
    this%dielconst = dielconst
    this%smear = smear
    this%adbcor = adbcor
    this%nr = nr
    this%dr = dr
    this%dk = pi/(nr*dr)

    !label kappa as 'undefined'
    this%kappa=HUGE(1d0)

    call sanity_check(this)

  end subroutine rism1d_potential_new

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Add solvent species.  Allocates necessary memory to accomodate the additional
!!!species
!!!IN:
!!!   this : potential object
!!!   mdl  : solvMDL object
!!!   density : number density (A^{-3})
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine rism1d_potential_addSpecies(this,mdl,density)
    use solvMDL_c
    implicit none
    type(rism1d_potential), intent(inout) :: this
    type(solvMDL), intent(in) :: mdl
    _REAL_, intent(in) :: density
    integer :: oldnv
    integer :: iv1, iv2, ivv, iv, iat, imlt

    !increment scalar variables
    oldnv = this%nv
    this%nsp = this%nsp + 1
    this%nv = this%nv + mdl%ntype
    this%nvv = this%nv*(this%nv+1)/2

    !(re)allocate arrays and fill with data
    this%namev => safemem_realloc(this%namev,len(this%namev),this%nv)
    this%namev(oldnv+1:this%nv) = mdl%atmname
    this%nat => safemem_realloc(this%nat,this%nsp)
    this%nat(this%nsp) = mdl%ntype
    this%mta => safemem_realloc(this%mta, max(ubound(this%mta,1),mdl%ntype), this%nsp)
    this%mta(1:mdl%ntype,this%nsp) = mdl%multi 

    this%mtv => safemem_realloc(this%mtv,this%nv)
    this%mtv(oldnv+1:this%nv) = mdl%multi

    this%jvv => safemem_realloc(this%jvv,this%nv,this%nv)
    !.................. setting triangle packing pointer ...................
    ivv = 0
    do iv2=1,this%nv
       do iv1=1,iv2
          ivv = ivv + 1
          this%jvv(iv1,iv2) = ivv
          this%jvv(iv2,iv1) = ivv
       enddo
    enddo

    this%rhosp => safemem_realloc(this%rhosp,this%nsp)
    this%rhosp(this%nsp) = density

    this%rhov => safemem_realloc(this%rhov,this%nv)
    this%rhov(oldnv+1:this%nv) = mdl%multi*density

    this%qsp => safemem_realloc(this%qsp,this%nsp)
    this%qsp(this%nsp) = sum(mdl%charge*mdl%multi)/sqrt(this%temperature)

    this%qat => safemem_realloc(this%qat,ubound(this%mta,1),this%nsp)
    this%qat(1:mdl%ntype,this%nsp) = mdl%charge/sqrt(this%temperature)

    this%qv => safemem_realloc(this%qv,this%nv)
    this%qv(oldnv+1:this%nv) = mdl%charge/sqrt(this%temperature)

    this%qspv => safemem_realloc(this%qspv,this%nv)
    this%qspv(oldnv+1:this%nv) = this%qsp(this%nsp)

    this%epsv => safemem_realloc(this%epsv,this%nv)
    this%epsv(oldnv+1:this%nv) = mdl%epsilon/this%temperature

    this%rminv => safemem_realloc(this%rminv,this%nv)
    this%rminv(oldnv+1:this%nv) = mdl%rmin/2d0

    this%rma => safemem_realloc(this%rma,3,ubound(this%mta,1),maxval(this%nat),this%nsp)
    iv = 0
    do iat=1,this%nat(this%nsp)
       do imlt = 1, this%mta(iat,this%nsp)
          iv = iv+1
          this%rma(:,imlt,iat,this%nsp) = mdl%coord(:,iv)
       end do
    end do
    this%wlmvv => safemem_realloc(this%wlmvv,ubound(this%mta,1),this%nv,this%nv)
    call intramolecular_dist(this)

    !data arrays in the original 1D-RISM code had an offset of 0 in the first index (like C).
    !The offset is 1 in the current code (like most Fortran).
    this%wvv => safemem_realloc(this%wvv,this%nr,this%nv,this%nv)
    this%wzvv => safemem_realloc(this%wzvv,this%nr,this%nv,this%nv)
    this%zkvv => safemem_realloc(this%zkvv,this%nr,this%nv,this%nv)
    this%uvv => safemem_realloc(this%uvv,this%nr,this%nvv)
    this%ulrvv => safemem_realloc(this%ulrvv,this%nr,this%nvv)
    this%ulkvv => safemem_realloc(this%ulkvv,this%nr,this%nvv)
    this%hlrvv => safemem_realloc(this%hlrvv,this%nr,this%nvv)
    this%hlkvv => safemem_realloc(this%hlkvv,this%nr,this%nvv)
  end subroutine rism1d_potential_addSpecies

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Destroys a potential object
!!!IN:
!!!   this : potential object  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine rism1d_potential_destroy(this)
    implicit none
    type(rism1d_potential), intent(inout) ::this
    integer :: err
    this%theory=''
    this%dielconst = 0
    this%smear = 0
    this%adbcor = 0
    this%nr = 0
    this%dr = 0
    this%dk = 0
    this%nsp = 0
    this%nv = 0
    this%nvv = 0
    this%kappa = HUGE(1d0)
    err = 0
    err = err + safemem_dealloc(this%nat)
    err = err + safemem_dealloc(this%namev)
    err = err + safemem_dealloc(this%mta)
    err = err + safemem_dealloc(this%mtv)
    err = err + safemem_dealloc(this%jvv)
    err = err + safemem_dealloc(this%rhosp)
    err = err + safemem_dealloc(this%rhov)
    err = err + safemem_dealloc(this%qsp)
    err = err + safemem_dealloc(this%qat)
    err = err + safemem_dealloc(this%qv)
    err = err + safemem_dealloc(this%qspv)
    err = err + safemem_dealloc(this%epsv)
    err = err + safemem_dealloc(this%rminv)
    err = err + safemem_dealloc(this%rma)
    err = err + safemem_dealloc(this%wlmvv)
    err = err + safemem_dealloc(this%wvv)
    err = err + safemem_dealloc(this%wzvv)
    err = err + safemem_dealloc(this%zkvv)
    err = err + safemem_dealloc(this%uvv)
    err = err + safemem_dealloc(this%ulrvv)
    err = err + safemem_dealloc(this%ulkvv)
    err = err + safemem_dealloc(this%hlrvv)
    err = err + safemem_dealloc(this%hlkvv)
    err = err + safemem_dealloc(this%duvv)
   end subroutine rism1d_potential_destroy

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Tabulates the intramolecular matrix
!!! and the site-site potential comprising
!!! the Coulomb and Lennard-Jones (LJ) terms,
!!! the latter for Composite Sphere (CS) sites
!!! as well as for usual point sites.
!!! Arguments Units:  charge[sqrt(kT A)], energy[kT], distance [A]
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine  rism1d_potential_calc(this,charge)
    use constants, only : pi
    implicit none
    type(rism1d_potential), intent(inout) :: this
    _REAL_, intent(in) :: charge

    integer ::  id, imt,imt1, iat,iat1,iat2, isp,isp1,isp2, &
         iv,iv1,iv2, ivv, ir
    _REAL_ ::  dipmom,rhodm,rhodm2, ydiel,cdiel, &
         k,kl,krx,kry,krz, hck, wk, qvv, epsvv, rminvv, &
         r,rs, rs6,ri6, usr,usra

    _REAL_ ::  d0x(this%nv), d0y(this%nv), d1z(this%nv), &
         wkvv(this%nv,this%nv)
    _REAL_, external :: erfc


    !............... shifting and orienting species dipoles ................
    call dipord(this)

    !............................ screen output ............................
    call rism_report_message('tabulating potentials and intramolecular functions ...')

    !............ calculating total Rho and Rho*D^2 of dipoles .............
    rhodm = 0.d0
    rhodm2 = 0.d0
    do isp=1,this%nsp

       do id=1,3
          dipmom = 0.d0
          do iat=1,this%nat(isp)
             do imt=1,this%mta(iat,isp)
                dipmom = dipmom + charge*this%qat(iat,isp)*this%rma(id,imt,iat,isp)
             enddo
          enddo
          if (id /= 3 .AND. abs(dipmom) > 1.d-8)  then
             call rism_report_error('POTENT: dipole moment not in z-direction')
          endif
       enddo

       if (abs(dipmom) > 1.d-8)  then
          rhodm = rhodm + this%rhosp(isp)
          rhodm2 = rhodm2 + this%rhosp(isp)*dipmom**2
       endif

    enddo
    !............ calculating dielectric asymptotics parameters ............
    ydiel = 4.d0/9.d0*pi*rhodm2
    if (ydiel == 0.d0)  then
       this%dielconst = 1.d0
       cdiel = 0.d0
    elseif (this%theory == 'DRISM')  then
       cdiel = ((this%dielconst-1.d0)/ydiel - 3.d0) / rhodm
    else
       this%dielconst = 1.d0 + 3.d0*ydiel
       cdiel = 0.d0
    endif

    !.................. calculating inverse Debye length ...................
    this%kappa = 0.d0
    do isp=1,this%nsp
       this%kappa = this%kappa + this%rhosp(isp)*charge*this%qsp(isp)**2
    enddo
    this%kappa = sqrt( 4.d0*pi*this%kappa/this%dielconst )
    !................ calculating intramolecular distances .................
    iv2 = 0
    do isp2=1,this%nsp
       do iat2=1,this%nat(isp2)
          iv2 = iv2 + 1

          iv1 = 0
          do isp1=1,this%nsp
             do iat1=1,this%nat(isp1)
                iv1 = iv1 + 1

                do imt1=1,this%mtv(iv1)

                   if (isp1 /= isp2)  then
                      this%wlmvv(imt1,iv1,iv2) = -1.d0
                   else
                      this%wlmvv(imt1,iv1,iv2) = sqrt( &
                           ( this%rma(1,imt1,iat1,isp1) - this%rma(1,1,iat2,isp2) )**2 &
                           + ( this%rma(2,imt1,iat1,isp1) - this%rma(2,1,iat2,isp2) )**2 &
                           + ( this%rma(3,imt1,iat1,isp1) - this%rma(3,1,iat2,isp2) )**2 )
                   endif
                enddo
             enddo
          enddo
       enddo
    enddo



    !...................... enumerating K-grid points ......................
    do ir=1,this%nr
       k = (ir-1)*this%dk
       hck = cdiel * exp(-(this%adbcor*k/2.d0)**2)

       !....................... calculating j0x,j0y,j1z .......................
       iv = 0
       do isp=1,this%nsp
          do iat=1,this%nat(isp)
             iv = iv + 1

             krx = k*this%rma(1,1,iat,isp)
             if (krx == 0.d0)  then
                d0x(iv) = 1.d0
             else
                d0x(iv) = sin(krx)/krx
             endif

             kry = k*this%rma(2,1,iat,isp)
             if (kry == 0.d0)  then
                d0y(iv) = 1.d0
             else
                d0y(iv) = sin(kry)/kry
             endif

             krz = k*this%rma(3,1,iat,isp)
             if (krz == 0.d0)  then
                d1z(iv) = 0.d0
             else
                d1z(iv) = sin(krz)/krz**2 - cos(krz)/krz
             endif

          enddo
       enddo

       !..................... getting and reducing Wvv(k) .....................
       do iv2=1,this%nv
          do iv1=1,this%nv
             wkvv(iv1,iv2) = 0.d0
             do imt1=1,this%mtv(iv1)
                kl = k * this%wlmvv(imt1,iv1,iv2)
                if (this%wlmvv(imt1,iv1,iv2) < 0.d0)  then
                   wk = 0.d0
                elseif (kl == 0.d0)  then
                   wk = 1.d0
                else
                   wk = sin(kl)/kl
                endif
                wkvv(iv1,iv2) = wkvv(iv1,iv2) + wk
             enddo
          enddo
       enddo

       !........................... loading Wvv(k) ............................
       do iv2=1,this%nv
          do iv1=1,this%nv
             this%wvv(ir,iv1,iv2) = wkvv(iv1,iv2)
          enddo
       enddo

       !........................... getting Zvv(k) ............................
       do iv2=1,this%nv
          do iv1=1,this%nv
             this%zkvv(ir,iv1,iv2) = d0x(iv1)*d0y(iv1)*d1z(iv1) * hck &
                  * d0x(iv2)*d0y(iv2)*d1z(iv2)
          enddo
       enddo

       !.............. getting Wvv(k)+Rho_v*Zvv(k) and k*Zvv(k) ...............
       do iv2=1,this%nv
          do iv1=1,this%nv
             this%wzvv(ir,iv1,iv2) = wkvv(iv1,iv2) + this%rhov(iv1)*this%zkvv(ir,iv1,iv2)
             this%zkvv(ir,iv1,iv2) = k*this%zkvv(ir,iv1,iv2)
          enddo
       enddo

    enddo

    ivv = 0
    do iv2=1,this%nv
       do iv1=1,iv2
          ivv = ivv + 1
          !ensure all elements are initialized
          this%uvv(1,ivv) = 0
          !.......................... Coulomb potential ..........................
          !.......................... 12-6 LJ potential ..........................
          qvv = charge*this%qv(iv1)*this%qv(iv2)
          rminvv = (this%rminv(iv1)+this%rminv(iv2))
          epsvv = sqrt(this%epsv(iv1)*this%epsv(iv2))
          do ir=2,this%nr
             r = (ir-1)*this%dr
             rs = r/rminvv
             rs = max( rs, 0.002d0)
             rs6 = rs**6
             ri6 = 1.d0/rs6
             usr = (ri6-2.d0)*ri6
             usr = usr * epsvv
             this%uvv(ir,ivv) = qvv/r + usr
          enddo
       enddo
    enddo

    !......... smeared Coulomb potential: r*Ulvv(r) and k*Ulvv(k) ..........
    ivv = 0
    do iv2=1,this%nv
       do iv1=1,iv2
          ivv = ivv + 1

          do ir=2,this%nr
             r = (ir-1)*this%dr
             k = (ir-1)*this%dk
             qvv = charge*this%qv(iv1)*this%qv(iv2)
             this%ulrvv(ir,ivv) = qvv*(1.d0-erfc(r/this%smear))
             this%ulkvv(ir,ivv) = qvv*4.d0*pi*exp(-(this%smear*k/2.d0)**2)/k
          enddo

       enddo
    enddo

    !........ smeared Coulomb term of TCF: r*Hlvv(r) and k*Hlvv(k) .........
    ivv = 0
    do iv2=1,this%nv
       do iv1=1,iv2
          ivv = ivv + 1

          do ir=2,this%nr
             r = (ir-1)*this%dr
             k = (ir-1)*this%dk
             qvv = charge*this%qspv(iv1)*this%qspv(iv2)
             this%hlrvv(ir,ivv) = -qvv/this%dielconst &
                  * 0.5d0 * exp((this%smear*this%kappa/2.d0)**2) &
                  * ( exp(-this%kappa*r)*erfc(this%kappa*this%smear/2.d0 - r/this%smear) &
                  - exp(this%kappa*r)*erfc(this%kappa*this%smear/2.d0 + r/this%smear) )
             this%hlkvv(ir,ivv) = -qvv/this%dielconst &
                  * 4.d0*pi*exp(-(0.5d0*this%smear*k)**2) * k/(k**2+this%kappa**2)
          enddo

       enddo
    enddo

    return
  end subroutine rism1d_potential_calc

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Tabulates the r derivative of the site-site potential comprising
!!!the Coulomb and Lennard-Jones (LJ) terms.  The result is stored in duvv.  
!!!This is only run once so the results may be reused.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine  rism1d_potential_derivative(this)
    implicit none
    type(rism1d_potential), intent(inout) :: this
    _REAL_ :: qvv, rminvv, epsvv, r, ri6, rs, rs6, usr
    integer :: ivv, iv1, iv2, ir
    if(.not.associated(this%duvv))then
       this%duvv => safemem_realloc(this%duvv,this%nr,this%nvv)

       ivv = 0
       do iv2=1,this%nv
          do iv1=1,iv2
             ivv = ivv + 1
             !ensure all elements are initialized
             this%uvv(1,ivv) = 0
             !.......................... Coulomb potential ..........................
             !.......................... 12-6 LJ potential ..........................
             qvv = this%qv(iv1)*this%qv(iv2)
             rminvv = (this%rminv(iv1)+this%rminv(iv2))
             epsvv = sqrt(this%epsv(iv1)*this%epsv(iv2))
             do ir=2,this%nr
                r = (ir-1)*this%dr
                rs = r/rminvv
                rs = max( rs, 0.002d0)
                rs6 = rs**6
                ri6 = 1.d0/rs6
                usr = (1d0-ri6)*ri6
                usr = usr * epsvv * 12d0
                this%duvv(ir,ivv) = (qvv/r + usr)/r
             enddo
          enddo
       enddo
    endif
  end subroutine rism1d_potential_derivative

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Returns the inverse Debye length (1/A) of the solvent. rism1d_potential_calc
!!!must be called first or huge(1d0) will be returned.
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_potential_getInvDebyeLen(this) result(kappa)
    implicit none
    type(rism1d_potential), intent(inout) :: this
    _REAL_ :: kappa
    kappa=this%kappa
  end function rism1d_potential_getInvDebyeLen

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!                               PRIVATE
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Check user parameters
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine sanity_check(this)
    implicit none
    type(rism1d_potential), intent(inout) :: this
    integer :: factored_nr
    !supported theory...
    if(.not.(this%theory.eq.'XRISM' .or. this%theory.eq.'DRISM'))then
       call rism_report_error("Only DRISM and XRISM theories are supported. Invalid theory: "//trim(this%theory))
    end if

    !check grid size
    if(this%nr <1) &
         call rism_report_error("NR must be > 1")
    !check the grid size prime factorization...
    factored_nr=this%nr
    do while(mod(factored_nr,2) == 0)
       factored_nr = factored_nr/2
    end do
    do while(mod(factored_nr,3) == 0)
       factored_nr = factored_nr/3
    end do
    do while(mod(factored_nr,5) == 0)
       factored_nr = factored_nr/5
    end do
    if(factored_nr /= 1)&
         call rism_report_error("NR must be a product of 2, 3 and 5 only")

    !grid spacing
    if(this%dr <=0) &
         call rism_report_error("DR must be > 0")

    !temperature
    if(this%temperature <=0) &
       call rism_report_error("TEMPERATURE must be > 0")

    !smear
    if(this%smear <= 0) &
         call rism_report_error("SMEAR must be > 0")

    !adbcor
    if(this%adbcor <= 0) &
         call rism_report_error("ADBCOR must be > 0")

    !dielconst
    if(this%dielconst <= 0 .and. this%theory .eq. 'DRISM') &
         call rism_report_error("dielectric constant must be > 0 for DRISM")
    
  end subroutine sanity_check

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! Ordering the species dipole moments:
!!! Shifting the molecule center of mass to the coordinate origin
!!! and orienting the dipole moment along the OZ axis.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine  dipord(this)
    implicit none
    type(rism1d_potential), intent(inout) :: this
    integer ::  imt,iat,isp, id
    _REAL_ ::  rqcm(3), rdm(3), &
         qcm, dm,dmxy, cosxy,sinxy,cosz,sinz, r1x,r1y,r1z, r2x,r2y,r2z

    !............... shifting and orienting species dipoles ................
    do isp=1,this%nsp

       qcm = 0.d0
       do iat=1,this%nat(isp)
          do imt=1,this%mta(iat,isp)
             qcm = qcm + abs(this%qat(iat,isp))
          enddo
       enddo
       if (qcm == 0.d0)  cycle

       do id=1,3
          rqcm(id) = 0.d0
       enddo
       do iat=1,this%nat(isp)
          do imt=1,this%mta(iat,isp)
             do id=1,3
                rqcm(id) = rqcm(id) + abs(this%qat(iat,isp))*this%rma(id,imt,iat,isp)
             enddo
          enddo
       enddo

       do id=1,3
          rqcm(id) = rqcm(id)/qcm
       enddo
       do iat=1,this%nat(isp)
          do imt=1,this%mta(iat,isp)
             do id=1,3
                this%rma(id,imt,iat,isp) = this%rma(id,imt,iat,isp) - rqcm(id)
             enddo
          enddo
       enddo

       do id=1,3
          rdm(id) = 0.d0
       enddo
       do iat=1,this%nat(isp)
          do imt=1,this%mta(iat,isp)
             do id=1,3
                rdm(id) = rdm(id) + this%qat(iat,isp)*this%rma(id,imt,iat,isp)
             enddo
          enddo
       enddo

       dm = 0.d0
       do id=1,3
          dm = dm + rdm(id)**2
       enddo
       dm = sqrt(dm)

       if (dm < 1.d-16)  cycle

       dmxy = 0.d0
       do id=1,2
          dmxy = dmxy + rdm(id)**2
       enddo
       dmxy = sqrt(dmxy)

       if (dmxy > 1.d-16)  then
          cosxy = rdm(1)/dmxy
          sinxy = rdm(2)/dmxy
          do iat=1,this%nat(isp)
             do imt=1,this%mta(iat,isp)
                r1x = this%rma(1,imt,iat,isp)
                r1y = this%rma(2,imt,iat,isp)
                r1z = this%rma(3,imt,iat,isp)
                r2x = r1x*cosxy + r1y*sinxy
                r2y = -r1x*sinxy + r1y*cosxy
                r2z = r1z
                this%rma(1,imt,iat,isp) = r2x
                this%rma(2,imt,iat,isp) = r2y
                this%rma(3,imt,iat,isp) = r2z
             enddo
          enddo
       endif

       cosz = rdm(3)/dm
       sinz = sqrt(1.d0-cosz**2)
       do iat=1,this%nat(isp)
          do imt=1,this%mta(iat,isp)
             r1x = this%rma(1,imt,iat,isp)
             r1y = this%rma(2,imt,iat,isp)
             r1z = this%rma(3,imt,iat,isp)
             r2z = r1z*cosz + r1x*sinz
             r2x = -r1z*sinz + r1x*cosz
             r2y = r1y
             this%rma(1,imt,iat,isp) = r2x
             this%rma(2,imt,iat,isp) = r2y
             this%rma(3,imt,iat,isp) = r2z
          enddo
       enddo
    enddo
  end subroutine dipord

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the intramolecular distance matrix
!!!IN:
!!!   this : rism1d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine intramolecular_dist(this)
    implicit none
    type(rism1d_potential), intent(inout) :: this
    integer :: iv1, iv2, iat1, iat2, imlt, isp1, isp2
    iv2 = 0
    do isp2=1,this%nsp
       do iat2=1,this%nat(isp2)
          iv2 = iv2 + 1

          iv1 = 0
          do isp1=1,this%nsp
             do iat1=1,this%nat(isp1)
                iv1 = iv1 + 1

                do imlt=1,this%mtv(iv1)

                   if (isp1 /= isp2)  then
                      this%wlmvv(imlt,iv1,iv2) = -1.d0
                   else
                      this%wlmvv(imlt,iv1,iv2) = sqrt( &
                           ( this%rma(1,imlt,iat1,isp1) - this%rma(1,1,iat2,isp2) )**2 &
                           + ( this%rma(2,imlt,iat1,isp1) - this%rma(2,1,iat2,isp2) )**2 &
                           + ( this%rma(3,imlt,iat1,isp1) - this%rma(3,1,iat2,isp2) )**2 )
                   endif
                enddo
             enddo
          enddo
       enddo
    enddo
  end subroutine intramolecular_dist

end module rism1d_potential_c
