! <compile=optimized>
#include "../include/dprec.fh"
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Partial series expansion of order-n (PSE-n) closure
!!!Stefan M. Kast and Thomas Kloss. J. Chem. Phys. 129, 236101 (2008)
!!!
!!!Interpolates between the Kovalenko-Hirata (KH) and hypernetted chain equation
!!!closures (HNC).  Order-1 gives KH and order-infinity gives HNC.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
module rism1d_psen_c
  use rism_report_c
  use safemem
  !the PSE-n type
  type rism1d_psen
     !order    : order of the series expansion.  >= 1
     !order1   : order+1
     integer :: order=0, order1
     !order1fac: (order+1)!
     _REAL_ :: order1fac
  end type rism1d_psen

  public rism1d_psen_new, rism1d_psen_destroy!, rism1d_psen_gvv
contains
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Initializes the PSEN closure
!!!IN:
!!!   this : PSEN object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine rism1d_psen_new(this, order)
    implicit none
    type(rism1d_psen), intent(inout) :: this
    integer, intent(in) :: order
    integer :: i
    double precision :: orderfac
    if(order <1)then
       call rism_report_error("PSE-n closure: order must be > 0.")
    end if
    this%order = order
    this%order1 = order+1
    this%order1fac = 1
    orderfac=1

    !calculate the factorial of (order+1) for future reference.  Check that we 
    !are not losing precision or overflowing the precision
    do i=2,this%order1
       orderfac = this%order1fac
       this%order1fac = this%order1fac*i
       !keep 16 significant digits
       if( abs((this%order1fac/i)/orderfac -1d0) >1e-16 )then
          call rism_report_error("(a,i4)",&
               "PSE-n: numerical precision exceeded.  Use a smaller order. Try ",i-1)
       end if
    end do
  end subroutine rism1d_psen_new

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculates Gvv from Uvv, Hvv, and Cvv using the PSEN closure
!!!IN:
!!!   this : the PSEN closure object
!!!   gvv  : site-site pair correlation function
!!!   uvv  : site-site potential
!!!   hvv  : site-site total correlation function
!!!   cvv  : site-site direct correlation function
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine rism1d_psen_gvv(this,gvv, uvv, hvv, cvv)
    implicit none
    type(rism1d_psen), intent(in) :: this
    _REAL_, intent(out) :: gvv(:,:)
    _REAL_, intent(in) :: uvv(:,:),hvv(:,:),cvv(:,:)
    integer :: ivv, ir, i
    _REAL_ :: tvv, orderfac
    do ivv = 1, ubound(gvv,2)
       do ir = 1, ubound(gvv,1)
          tvv = -uvv(ir,ivv) +hvv(ir,ivv)-cvv(ir,ivv)
          if(tvv>=0d0)then
             gvv(ir,ivv) = 1d0
             orderfac = 1d0
             do i=1,this%order
                orderfac = orderfac*i
                gvv(ir,ivv) = gvv(ir,ivv) + (tvv**i)/orderfac
             end do
          else
             gvv(ir,ivv) = exp(tvv)
          end if
       end do
    end do
  end subroutine rism1d_psen_gvv

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculates Bvv (bridge function) from Uvv, Gvv, and Cvv using the PSEN closure
!!!IN:
!!!   this : the PSEN closure object
!!!   uvv  : site-site potential
!!!   gvv  : site-site total correlation function
!!!   cvv  : site-site direct correlation function
!!!OUT:
!!!   bvv  : site-site bridge function
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_psen_bvv(this, uvv, gvv, cvv) result(bvv)
    implicit none
    type(rism1d_psen), intent(in) :: this
    _REAL_, pointer :: bvv(:,:)
    _REAL_, intent(in) :: uvv(:,:),gvv(:,:),cvv(:,:)
    integer :: ivv, ir, i
    _REAL_ :: tvv, orderfac
    nullify(bvv)
    bvv =>safemem_realloc(bvv,ubound(gvv,1),ubound(gvv,2))
    do ivv = 1, ubound(bvv,2)
       do ir = 1, ubound(bvv,1)
          tvv = -uvv(ir,ivv) +gvv(ir,ivv)-1d0-cvv(ir,ivv)
          if(tvv>=0d0)then
             orderfac = 1d0
             bvv(ir,ivv) = 1d0
             do i=1,this%order
                orderfac = orderfac*i
                bvv(ir,ivv) = bvv(ir,ivv) + (tvv**i)/orderfac
             end do
             bvv(ir,ivv) = log(bvv(ir,ivv))
             bvv(ir,ivv) = bvv(ir,ivv)-tvv
          else
             bvv(ir,ivv) = 0
          end if
       end do
    end do
  end function rism1d_psen_bvv

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the excess chemical potential in kT
!!!IN:
!!!   this : the closure object
!!!   gvv  : site-site pair correlation function
!!!   cvv  : site-site direct correlation function
!!!   mtv  : multiplicity of each site
!!!   rhov : number density of each site
!!!   rhotrgt : (optional) The final, physical density for thermodynamics.  This
!!!             can be used as an effective correction for some closures that
!!!             over estimate the pressure (e.g. HNC and PSEN)
!!!   dr   : r-space grid spacing
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_psen_exchem(this, gvv, uvv, cvv, mtv, jvv, rhov, rhotrgtv, dr) result(exchem)
    use constants, only : pi
    implicit none
    type(rism1d_psen), intent(in) :: this
    _REAL_, intent(in) :: gvv(:,:),uvv(:,:),cvv(:,:), rhov(:), rhotrgtv(:), dr
    integer, intent(in) :: mtv(:),jvv(:,:)
    _REAL_ :: exchem(ubound(mtv,1))
    !hnc    : HNC contribution
    !tsvv   : t^* = u +h -c
    !series : contribution from the series expansion
    _REAL_ :: r, hnc,hvv,tsvv, exchemv,series
    integer :: ivv, iv1, iv2, ir,i
    do iv2=1,ubound(mtv,1) !nv
       exchem(iv2) = 0.d0
       do ir=2,ubound(gvv,1) !nr
          r = (ir-1)*dr
          exchemv = 0.d0
          do iv1=1,ubound(mtv,1) !nv
             ivv = jvv(iv1,iv2)
             hvv = gvv(ir,ivv) - 1.d0
             tsvv = -uvv(ir,ivv) + hvv -cvv(ir,ivv)
             if (tsvv >= 0.d0)  then
                series = tsvv**(this%order1)/this%order1fac
             else
                series = 0
             endif
             hnc = 0.5d0*hvv*hvv - cvv(ir,ivv)  - 0.5d0*hvv*cvv(ir,ivv)           
             !                     exchemv = exchemv + rhov(iv1)*mtv(iv2) &
             exchemv = exchemv + rhotrgtv(iv1)*mtv(iv2) &
                  *(hnc-series)
          enddo
          exchem(iv2) = exchem(iv2) + exchemv*r**2
       enddo
    enddo
    exchem = exchem * 4.d0*pi*dr
end function rism1d_psen_exchem

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the ionic(?) excess chemical potential in kT.  This seems to take
!!!the contribution only from the last solvent type.  I'm not sure what the point is
!!!here.
!!!IN:
!!!   this : the closure object
!!!   gvv  : site-site pair correlation function
!!!   cvv  : site-site direct correlation function
!!!   mtv  : multiplicity of each site
!!!   rhov : number density of each site
!!!   rhotrgt : (optional) The final, physical density for thermodynamics.  This
!!!             can be used as an effective correction for some closures that
!!!             over estimate the pressure (e.g. HNC and PSEN)
!!!   dr   : r-space grid spacing
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_psen_exchemIon(this, gvv, uvv, cvv, mtv, jvv, rhov, rhotrgtv, dr) result(exchem)
    use constants, only : pi
    implicit none
    type(rism1d_psen), intent(in) :: this
    _REAL_, intent(in) :: gvv(:,:),uvv(:,:),cvv(:,:), rhov(:), rhotrgtv(:), dr
    integer, intent(in) :: mtv(:),jvv(:,:)
    _REAL_ :: exchem(ubound(mtv,1))
    !hnc    : HNC contribution
    !tsvv   : t^* = u +h -c
    !series : contribution from the series expansion
    _REAL_ :: r, hnc,hvv, tsvv, exchemv,series
    integer :: ivv, iv1, iv2, ir, i
    
    do iv2=1,ubound(mtv,1) !nv
       exchem(iv2) = 0.d0
       do ir=2,ubound(gvv,1) !nr
          r = (ir-1)*dr
          exchemv = 0.d0
!          do iv1=1,ubound(mtv,1) !nv
          iv1=ubound(mtv,1)
             ivv = jvv(iv1,iv2)
             hvv = gvv(ir,ivv) - 1.d0
             tsvv = -uvv(ir,ivv) + hvv -cvv(ir,ivv)
             if (tsvv >= 0.d0)  then
                series = tsvv**(this%order1)/this%order1fac
             else
                series = 0
             endif
             hnc = 0.5d0*hvv*hvv - cvv(ir,ivv)  - 0.5d0*hvv*cvv(ir,ivv)           
             !                     exchemv = exchemv + rhov(iv1)*mtv(iv2) &
             exchemv = exchemv + rhotrgtv(iv2)*mtv(iv1) &
                  *(hnc - series)
 !         enddo
          exchem(iv2) = exchem(iv2) + exchemv*r**2
       enddo
    enddo
    exchem = exchem * 4.d0*pi*dr
  end function rism1d_psen_exchemIon

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the r-space contribution to the pressure in internal units
!!!using the free energy route
!!!IN:
!!!   this : the closure object
!!!   gvv  : site-site pair correlation function
!!!   cvv  : site-site direct correlation function
!!!   mtv  : multiplicity of each site
!!!   rhov : number density of each site
!!!   dr   : r-space grid spacing
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_psen_pressureFE(this, gvv, uvv, cvv, mtv, rhov, dr) result(pr)
    use constants, only : pi
    implicit none
    type(rism1d_psen), intent(in) :: this
    _REAL_, intent(in) :: gvv(:,:),uvv(:,:),cvv(:,:),rhov(:),dr
    integer, intent(in) :: mtv(:)
    _REAL_ :: pr
    _REAL_ :: prv
    !hnc    : h^2*c/2 contribution
    !tsvv   : t^* = u +h -c
    !series : contribution from the series expansion
    _REAL_ :: r, hvv, h2c, tsvv, series
    integer :: ir,ivv,iv1,iv2, cnt, i
    !r-space
    pr = 0.d0
    do ir=2,ubound(gvv,1) !nr
       r = (ir-1)*dr
       prv = 0.d0
       ivv = 0
       do iv2=1,ubound(mtv,1) !nv
          do iv1=1,iv2
             ivv = ivv + 1
             !if the sites are different, we need to double count the contribution
             if (iv1 == iv2)  then
                cnt = 1
             else
                cnt = 2
             endif
             hvv = gvv(ir,ivv) - 1.d0
             tsvv = -uvv(ir,ivv) + hvv -cvv(ir,ivv)
             if (tsvv >= 0.d0)  then
                series = tsvv**(this%order1)/this%order1fac
             else
                series = 0
             endif
             h2c = 0.5d0*hvv*hvv - cvv(ir,ivv)
             prv = prv + cnt*mtv(iv1)*rhov(iv2)*(h2c-series)
          enddo
       enddo
       pr = pr + prv*r**2
    enddo
    pr = pr * 2.d0*pi*dr
  end function rism1d_psen_pressureFE

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the r-space contribution to the free energy in internal units
!!!IN:
!!!   this : the closure object
!!!   gvv  : site-site pair correlation function
!!!   cvv  : site-site direct correlation function
!!!   mtv  : multiplicity of each site
!!!   rhov : number density of each site
!!!   dr   : r-space grid spacing
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism1d_psen_freeEnergy(this, gvv, uvv, cvv, mtv, rhov, dr) result(fe)
    use constants, only : pi
    implicit none
    type(rism1d_psen), intent(in) :: this
    _REAL_, intent(in) :: gvv(:,:),uvv(:,:),cvv(:,:),rhov(:),dr
    integer, intent(in) :: mtv(:)
    _REAL_ :: fe
    _REAL_ :: fev
    !h2c    : h^2*c/2 contribution
    !tsvv   : t^* = u +h -c
    !series : contribution from the series expansion
    _REAL_ :: r, hvv, h2c, tsvv, series
    integer :: ir,ivv,iv1,iv2, cnt
    !r-space
    fe = 0.d0
    do ir=2,ubound(gvv,1) !nr
       r = (ir-1)*dr
       fev = 0.d0
       ivv = 0
       do iv2=1,ubound(mtv,1) !nv
          do iv1=1,iv2
             ivv = ivv + 1
             !if the sites are different, we need to double count the contribution
             if (iv1 == iv2)  then
                cnt = 1
             else
                cnt = 2
             endif
             hvv = gvv(ir,ivv) - 1.d0
             tsvv = -uvv(ir,ivv) + hvv -cvv(ir,ivv)
             if (tsvv >= 0.d0)  then
                series = tsvv**(this%order1)/this%order1fac
             else
                series = 0
             endif
             h2c = 0.5d0*hvv*hvv - cvv(ir,ivv)
             fev = fev + cnt*rhov(iv1)*rhov(iv2)*(h2c-series)
          enddo
       enddo
       fe = fe + fev*r**2
    enddo
    fe = -fe * 2.d0*pi*dr
  end function rism1d_psen_freeEnergy

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Frees memory and resets the PSEN closure
!!!IN:
!!!   this : PSEN object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine rism1d_psen_destroy(this)
    implicit none
    type(rism1d_psen), intent(inout) :: this
    this%order=0
  end subroutine rism1d_psen_destroy
end module rism1d_psen_c
