!<compile=optimized>

!The 3D-RISM-KH software found here is copyright (c) 2010 by Andriy Kovalenko,
!Tyler Luchko and David A. Case.
!
!This program is free software: you can redistribute it and/or modify it
!under the terms of the GNU General Public License as published by the Free
!Software Foundation, either version 3 of the License, or (at your option)
!any later version.
!
!This program is distributed in the hope that it will be useful, but
!WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
!for more details.
!
!You should have received a copy of the GNU General Public License in the
!../../LICENSE file.  If not, see <http://www.gnu.org/licenses/>.
!
!Users of the 3D-RISM capability found here are requested to acknowledge
!use of the software in reports and publications.  Such acknowledgement
!should include the following citations:
!
!1) A. Kovalenko and F. Hirata. J. Chem. Phys., 110:10095-10112  (1999); 
!ibid. 112:10391-10417 (2000).   
!
!2) A. Kovalenko,  in:  Molecular  Theory  of  Solvation,  edited  by  
!F. Hirata  (Kluwer Academic Publishers, Dordrecht, 2003), pp.169-275.  
!
!3) T. Luchko, S. Gusarov, D.R. Roe, C. Simmerling, D.A. Case, J. Tuszynski,
!and  A. Kovalenko, J. Chem. Theory Comput., 6:607-624 (2010). 

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!3D-RISM solver.  
!This is defines the 3D-RISM type and associated subroutines.  All type elements 
!are public.  In general, read but do not write these variables.  This provides
!an object-orientiented interface without needing a function to access every 
!variable.
!Features of this solver include:
! o Multiple closures: KH, HNC, PSE-n
! o MDIIS accelerated solutions
! o Optional cutoffs
! o Supercell method for long range asymptotics
! o Analytic forces
! o Variable grid size and dynamic memory allocation
! o MPI support
! o units:  energy       [kT]
!           distances    [A]         (Angstroms)
!           site charges [sqrt(kT A)]
!           temperature  [K]
!           density      [#/A^3]
!           mass         [au]
! o To convert [e] to [sqrt(kT A)] *sqrt(COULOMB_CONST_E/KB/temperature)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

#include "../include/dprec.fh"

module rism3d_c
    use rism3d_solu_c
    use rism3d_solv_c
    use rism3d_potential_c
    use rism3d_grid_c
    use rism3d_closure_c
    use rism_report_c
    implicit none
#include "def_time.h"
    !closure enumeration
    integer,parameter :: KH=1, HNC=0

    type rism3d
!       sequence

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Solute/solvent information
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

       !solu  :: solute object
       type(rism3d_solu) :: solu
       !solv  :: solvent object
       type(rism3d_solv) :: solv

       !pot   :: potential object
       type(rism3d_potential) :: pot
       !grid  :: grid object
       type(rism3d_grid) :: grid
       !closure  :: closure object
       type(rism3d_closure) :: closure

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!private !(should be)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!       
       !verbose :: 0 - no ouput
       !                1 - memory allocation and steps for convergence
       !                2 - 1 + convergence progress
       integer :: verbose=0

       !This is a bit ugly and there may be a better solution.  We
       !need to keep track of the number of solutions for both charged
       !and un-charged solutes.  When we change between the solutes we
       !set point the nsolutions pointer to the appropriate variable.
       !However, the 'target' attribute is not allow in type
       !definitions so these variables have to be pointers and we have
       !to allocate memory for them.
       !nsolution :: number of times full solutions have been calculated
       !nsolutionChg   :: number of times full solutions with a charged solute 
       !                   have been calculated
       !nsolutionNoChg :: number of times full solutions with an uncharged solute
       !                   have been calculated
       integer, pointer :: nsolution=>NULL(), nsolutionChg=>NULL(), nsolutionNoChg=>NULL()

       !ucenter :: center the solute in the solvation box.  
       !           0 - off, 1 - center of mass, 2 - center of geometry
       !           For negative numbers the centering translation is only calculated
       !           the for the first solution and used for subsequent calculations.
       !           This allows the solute to drift in the box.
       integer :: ucenter = 1

       !ratucm   :: center of mass (geometry) of the solute [A]
       _REAL_ :: ratucm(3)

       !ncuvsteps :: number of past cuv time steps saves
       integer :: ncuvsteps

       !buffer   :: buffer distance to the edge of the box for the solvent [A]
       !boxfixlen:: fixed box size for 3d-rism.  for PBC calculations, these
       !            should generallly be equal [A]
       _REAL_ :: buffer=12d0,boxfixlen(3)
       !nboxfix(3)    :: number of grid points in each dimension for a fixed box size
       integer ::  nboxfix(3)

       !varbox  :: variable box size
       logical :: varbox = .true.

       !mdiis_method :: MDIIS implementation to use    
       !NVec       :: number of vectors used for MDIIS (consequently, the number of copies of
       !             CUV we need to keep for MDIIS)
       integer :: NVec,mdiis_method
    
       !deloz :: 'step size' for MDIIS
       _REAL_ :: deloz=0.7d0

!!!!!!!!!!!!!!!!!!
!!! MPI Support !!
!!!!!!!!!!!!!!!!!!
       integer :: mpirank=0, mpicomm=0, mpisize=1
!!!!!!!!!!!!!!!!!!!
!!! LARGE ARRAYS !!
!!!!!!!!!!!!!!!!!!!

       !
       ! all arrays are declared as pointers to ensure we can reallocate them as necessary
       !
       
       !xvva       :: solvent chi interpolated for our grid size
       !guv        :: solvent distribution function
       !huv        :: guv - 1
       !cuv        :: solvent direct correlation function
       !oldcuv     :: previous solutions of cuv. Points to oldcuvChg or oldcuvNoChg
       !oldcuvChg  :: previous solutions for the standard charged system
       !oldcuvNoChg:: previous solutions for the chargeless system.  This is only allocated
       !           :: if _unsetCharges() is called
       !cuvres     :: residual value for cuv calculation.  This should be a local variable of r1rism but
       !              is such a large chuck of memory, we would like to allocate it as few times as possible
       _REAL_,pointer :: xvva(:)=>NULL(),&
            guv(:,:)=>NULL(),huv(:,:)=>NULL(),&
            oldcuv(:,:,:,:,:)=>NULL(),&
            oldcuvChg(:,:,:,:,:)=>NULL(),&
            oldcuvNoChg(:,:,:,:,:)=>NULL(),&
            cuv(:,:,:,:)=>NULL(), cuvWRK(:,:,:,:,:)=>NULL(),&
            cuvres(:,:)=>NULL(), cuvresWRK(:,:,:)=>NULL()


       !planfwd :: FFTW plan for forward 3D-FFT
       !planfwd :: FFTW plan for backward 3D-FFT
       integer*8 :: planfwd=0, planbwd=0
       
    end type rism3d

    interface rism3d_setbox
       module procedure rism3d_setbox_variable, rism3d_setbox_fixed
    end interface

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Public suroutine and functions
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    public :: rism3d_new, rism3d_destroy, rism3d_solve, rism3d_force, &
         rism3d_exchem_tot, rism3d_exchem,&
         rism3d_exchemGF_tot, rism3d_exchemGF,&
         rism3d_setbox,&
         rism3d_setclosure, rism3d_setverbosity, rism3d_setcut, rism3d_setmdiis

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Private suroutine and functions
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    private :: resizebox, reallocbox, &
         interpolate_xvva, center_solute, rxrism, r1rism, force_lj, force_coulomb, &
         cuv_propagate, cuv_update
    
  contains
    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
! subroutines
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Constructor - precalculates the solute solvent terms that are not 
!!!configuration dependent and sets box parameters.
!!!
!!!The solvation box may be fixed size or variable.  For fixed size,
!!!define o_boxlen and o_ng3.  For variable box size, define buffer
!!!and grdspc.  Do not mix these parameters as this will cause the
!!!program to halt.
!!!
!!!If this is an MPI run, supply the MPI communicator.  Only the rank
!!!0 parameters will be used.
!!!
!!!IN:
!!!   this :: new rism3d object
!!!   solu :: 3D-RISM solute object
!!!   solv :: 3D-RISM solvent object
!!!   ucenter :: center the solute in the solvation box.  
!!!   ncuvsteps :: number of past cuv time steps saves
!!!   closure :: closure type (see closure enumeration).
!!!   order   :: order for closure 
!!!   cut     :: distance cutoff for potential and force calculations
!!!   mdiis_nvec :: number of MDIIS vectors (previous iterations) to keep
!!!   mdiis_del :: scaling factor (step size) applied to estimated gradient (residual)
!!!   mdiis_method :: which implementation of the algorithm
!!!   o_buffer :: (optional) shortest distance between solute and solvent box boundary
!!!   o_grdspc :: (optional) linear grid spacing for the solvent box in each dimension
!!!   o_boxlen :: (optional) solvent box size in each dimension [A]
!!!   o_ng3    :: (optional) number of grid points in each dimension
!!!   o_mpicomm :: (optional) MPI communicator 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_new(this,solu,solv,ucenter, ncuvsteps,&
         closure,order,cut,mdiis_nvec, mdiis_del, mdiis_method,&
         o_buffer,o_grdspc,o_boxlen, o_ng3, o_mpicomm)
      use rism3d_solu_c
      use rism3d_solv_c
      implicit none
#ifdef MPI
      include 'mpif.h'
#endif /*MPI*/
      type(rism3d), intent(inout) :: this
      type(rism3d_solu), intent(in),target :: solu
      type(rism3d_solv), intent(in),target :: solv
      integer, intent(in) :: ucenter, ncuvsteps
      integer,intent(in) :: closure, order
      _REAL_, intent(in) :: cut
      integer, intent(in) :: mdiis_nvec,mdiis_method
      _REAL_, intent(in) :: mdiis_del
      _REAL_, optional, intent(in) :: o_buffer, o_grdspc(3)
      _REAL_, optional, intent(in) :: o_boxlen(3)
      integer, optional, intent(in) :: o_ng3(3)
      integer, optional, intent(in) :: o_mpicomm
      !temporary copies
      integer :: t_closure, t_order
      _REAL_ :: t_cut
      integer :: t_mdiis_nvec,t_mdiis_method
      _REAL_ :: t_mdiis_del
      _REAL_ :: t_buffer, t_grdspc(3)
      _REAL_ :: t_boxlen(3)
      integer :: t_ng3(3)
      integer :: t_mpicomm
      integer :: err
      
      !MPI set up starts by obtaining rank and size.  Temporary copies
      !of input parameters that do not directly set object variables
      !are made.  These are they broadcast to the rank > 0 processes.
      !Then all processes complete the intitialization proceedure
      !using the temporary copies.  This leaves the input parameters
      !untouched.

      ! GET RANK AND SIZE
      this%mpicomm = 0
      this%mpisize = 1
      this%mpirank = 0
#ifdef MPI
      if(present(o_mpicomm)) then
         this%mpicomm = o_mpicomm
         if(this%mpicomm == MPI_COMM_NULL)&
              call rism_report_error("RISM3D: received NULL MPI communicator")
         call mpi_comm_rank(this%mpicomm,this%mpirank,err)
         if(err /=0) call rism_report_error&
              ("(a,i8)","RISM3D: could not get MPI rank for communicator ",this%mpicomm)
         call mpi_comm_size(this%mpicomm,this%mpisize,err)
         if(err /=0) call rism_report_error&
              ("(a,i8)","RISM3D: could not get MPI size for communicator ",this%mpicomm)
         call rism_report_mpi(this%mpicomm)
      end if
#endif /*MPI*/
      !MAKE TEMPORARY COPIES
      if(this%mpirank == 0)then
         call rism3d_solu_clone(solu,this%solu)
         call rism3d_solv_clone(solv,this%solv)
         this%ucenter = ucenter
         this%ncuvsteps = ncuvsteps
         t_closure = closure
         t_order=order
         t_cut=cut
         t_mdiis_nvec = mdiis_nvec
         t_mdiis_method = mdiis_method
         t_mdiis_del = mdiis_del
         !check box parameters
         if(present(o_buffer) .and. present(o_grdspc))then
            t_buffer = o_buffer
            t_grdspc = o_grdspc
            if(present(o_boxlen) .or. present(o_ng3))&
                 call rism_report_error("RISM3D: do not set BOXLEN or NG3 for variable box size")
         elseif(present(o_boxlen) .and. present(o_ng3)) then
            t_boxlen = o_boxlen
            t_ng3 = o_ng3
            if(present(o_buffer) .or. present(o_grdspc)) &
                 call rism_report_error("RISM3D: do not set BUFFER or GRDSPC for fixed box size")
         else
            call rism_report_error("RISM3D: not enough parameters for fixed or variable box size")
         end if
      end if
#ifdef MPI
      !BROADCAST PARAMETERS
      !set solu on all processes
      call rism3d_solu_mpi_clone(this%solu,this%mpirank,this%mpicomm)
      !set solv on all processes
      call rism3d_solv_mpi_clone(this%solv,this%mpirank,this%mpicomm)
      !set ucenter on all processes
      call mpi_bcast(this%ucenter, 1, mpi_integer, 0, this%mpicomm,err)
      if(err /=0) call rism_report_error("RISM3D: broadcast UCENTER in constructor failed")
      !set ncuvstpes on all processes
      call mpi_bcast(this%ncuvsteps, 1, mpi_integer, 0, this%mpicomm,err)
      if(err /=0) call rism_report_error("RISM3D: broadcast NCUVSTEPS in constructor failed")
      call mpi_bcast(t_closure, 1, mpi_integer, 0, this%mpicomm,err)
      if(err /=0) call rism_report_error("RISM3D: broadcast CLOSURE in constructor failed")
      call mpi_bcast(t_order, 1, mpi_integer, 0, this%mpicomm,err)
      if(err /=0) call rism_report_error("RISM3D: broadcast ORDER in constructor failed")
      call mpi_bcast(t_cut, 1, mpi_double_precision, 0, this%mpicomm,err)
      if(err /=0) call rism_report_error("RISM3D: broadcast CUT in constructor failed")
      call mpi_bcast(t_mdiis_nvec, 1, mpi_integer, 0, this%mpicomm,err)
      if(err /=0) call rism_report_error("RISM3D: broadcast MDIIS_NVEC in constructor failed")
      call mpi_bcast(t_mdiis_del, 1, mpi_double_precision, 0, this%mpicomm,err)
      if(err /=0) call rism_report_error("RISM3D: broadcast MDIIS_DEL in constructor failed")
      call mpi_bcast(t_mdiis_method, 1, mpi_integer, 0, this%mpicomm,err)
      if(err /=0) call rism_report_error("RISM3D: broadcast MDIIS_METHOD in constructor failed")
      if(present(o_buffer) .and. present(o_grdspc))then
         call mpi_bcast(t_buffer, 1, mpi_double_precision, 0, this%mpicomm,err)
         if(err /=0) call rism_report_error("RISM3D: broadcast BUFFER in constructor failed")
         call mpi_bcast(t_grdspc, 3, mpi_double_precision, 0, this%mpicomm,err)
         if(err /=0) call rism_report_error("RISM3D: broadcast GRDSPC in constructor failed")
      else
         call mpi_bcast(t_boxlen, 3, mpi_double_precision, 0, this%mpicomm,err)
         if(err /=0) call rism_report_error("RISM3D: broadcast BOXLEN in constructor failed")
         call mpi_bcast(t_ng3, 3, mpi_integer, 0, this%mpicomm,err)
         if(err /=0) call rism_report_error("RISM3D: broadcast NG3 in constructor failed")
      end if
#endif /*MPI*/
      !INITIALIZE
      call rism3d_grid_new(this%grid,this%mpicomm)
      call rism3d_setmdiis(this,t_mdiis_nvec,t_mdiis_del,t_mdiis_method)
      call rism3d_potential_new(this%pot, this%grid, this%solv, this%solu,0d0)
      call rism3d_setcut(this,t_cut)
      call rism3d_setclosure(this,t_closure,t_order)
      if(present(o_buffer) .and. present(o_grdspc))then
         call rism3d_setbox(this,t_buffer,t_grdspc)
      else
         call rism3d_setbox(this,t_boxlen,t_ng3)
      end if
      allocate(this%nsolutionChg, this%nsolutionNoChg)
      this%nsolutionChg=0
      this%nsolutionNoChg=0
      this%nsolution=>this%nsolutionChg
    end subroutine rism3d_new

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Sets the parameters for a variable solvation box.
!!!IN:
!!!  this :: rism3d object
!!!  buffer :: shortest distance between solute and solvent box boundary
!!!  grdspc :: linear grid spacing for the solvent box in each dimension
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_setbox_variable(this,buffer,grdspc)
      implicit none
      type(rism3d),intent(inout) :: this
      _REAL_, intent(in) :: buffer, grdspc(3)
      this%varbox = .true.
      this%buffer = buffer
      call rism3d_grid_setSpacing(this%grid,grdspc)
    end subroutine rism3d_setbox_variable

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Sets the parameters for a variable solvation box.
!!!IN:
!!!  this :: rism3d object
!!!  boxlen :: solvent box size in each dimension in Angstroms
!!!  ng3    :: number of grid points in each dimension
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_setbox_fixed(this,boxlen,ng3)
      implicit none
      type(rism3d),intent(inout) :: this
      _REAL_, intent(in) :: boxlen(3)
      integer, intent(in) :: ng3(3)
      this%varbox = .false.
      this%boxfixlen = boxlen
      this%nboxfix = ng3
    end subroutine rism3d_setbox_fixed

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Sets the closure type
!!!IN:
!!!   this :: rism3d object
!!!   closure :: closure type (see closure enumeration).
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_setclosure(this,closure,order)
      implicit none
      type(rism3d),intent(inout) :: this
      integer, intent(in) :: closure,order
      character(len=4) :: name
      if(closure == 0) then
         name = "HNC"
      elseif(closure == 1)then
         name = "KH"
      elseif(closure == 2)then
         name = "PSEN"
      end if
      call rism3d_closure_new(this%closure,name,this%pot,order)
    end subroutine rism3d_setclosure

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Sets verbosity of output
!!!IN:
!!!   this :: rism3d object
!!!   verbosity :: 0 - no output
!!!                1 - memory allocation and steps for convergence
!!!                2 - 1 + convergence progress
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_setverbosity(this,verbosity)
      implicit none
      type(rism3d),intent(inout) :: this
      integer, intent(in) :: verbosity
      this%verbose = verbosity
    end subroutine rism3d_setverbosity

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Sets the cut off distance for potential and force calculations
!!!IN:
!!!   this :: rism3d object
!!!   cut     :: distance cutoff for potential and force calculations
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_setcut(this,cut)
      implicit none
      type(rism3d),intent(inout) :: this
      _REAL_, intent(in) :: cut
      call rism3d_potential_setCut(this%pot,cut)
    end subroutine rism3d_setcut

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Sets MDIIS parameters
!!!IN:
!!!   this :: rism3d object!
!!!   nvec :: number of MDIIS vectors (previous iterations) to keep
!!!   del :: scaling factor (step size) applied to estimated gradient (residual)
!!!   method :: which implementation of the algorithm
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_setmdiis(this,nvec,del,method)
      implicit none
      type(rism3d),intent(inout) :: this
      integer, intent(in) :: nvec,method
      _REAL_, intent(in) :: del
      this%NVec = nvec
      this%deloz = del
      this%mdiis_method = method
    end subroutine rism3d_setmdiis

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Sets solute coordinates.
!!!IN:
!!!   this :: rism3d object
!!!   ratu :: coordinates
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_setCoord(this,ratu)
      implicit none
      type(rism3d), intent(inout) :: this
      _REAL_, intent(in) :: ratu(:,:)
      call rism3d_solu_setCoord(this%solu,ratu)
    end subroutine rism3d_setCoord

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Sets all solute partial charges to zero
!!!IN:
!!!   this :: rism3d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_unsetCharges(this)
      implicit none
      type(rism3d), intent(inout) :: this
      integer :: i
      call rism3d_solu_unsetCharges(this%solu)
      this%nsolution=>this%nsolutionNoChg
      if(.not.associated(this%oldcuvNoChg) .and. &
           associated(this%oldcuvChg))then
         this%oldcuvNoChg => safemem_realloc(this%oldcuvNoChg,&
              this%grid%nr(1),this%grid%nr(1),this%grid%nr(1),this%solv%natom,&
              this%ncuvsteps)
         this%oldcuvNoChg=0
      end if
      this%oldcuv=> this%oldcuvNoChg
      if(ubound(this%oldcuv,5)>0)then
         this%cuvWRK=>safemem_realloc(this%cuvWRK,&
              ubound(this%oldcuv,1),ubound(this%oldcuv,2),ubound(this%oldcuv,3),&
              ubound(this%oldcuv,4),ubound(this%oldcuv,5),.false.)
         this%cuv => this%cuvWRK(:,:,:,:,1)
!         this%cuv=this%oldcuv(:,:,:,:,1)
         call dcopy(product(ubound(this%oldcuv)),this%oldcuv,1,this%cuv,1)
      elseif(associated(this%cuv))then
         this%cuv=0
      end if
    end subroutine rism3d_unsetCharges

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Sets all solute partial charges to to their original
!!!values. (Undoes rism3d_unsetCharge().)
!!!IN:
!!!   this :: rism3d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_resetCharges(this)
      implicit none
      type(rism3d), intent(inout) :: this
      integer :: i
      call rism3d_solu_resetCharges(this%solu)
      this%nsolution=>this%nsolutionChg
      if(.not.associated(this%oldcuvChg) .and. &
           associated(this%oldcuvNoChg))then
         this%oldcuvChg => safemem_realloc(this%oldcuvChg,&
              this%grid%nr(1),this%grid%nr(1),this%grid%nr(1),this%solv%natom,&
              this%ncuvsteps)
         this%oldcuvChg=0
      end if
      this%oldcuv=> this%oldcuvChg
      if(ubound(this%oldcuv,5)>0) then
         this%cuvWRK=>safemem_realloc(this%cuvWRK,&
              ubound(this%oldcuv,1),ubound(this%oldcuv,2),ubound(this%oldcuv,3),&
              ubound(this%oldcuv,4),ubound(this%oldcuv,5),.false.)
         this%cuv => this%cuvWRK(:,:,:,:,1)
         call dcopy(product(ubound(this%oldcuv)),this%oldcuv,1,this%cuv,1)
      elseif(associated(this%cuv))then
         this%cuv=0
      end if
    end subroutine rism3d_resetCharges

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculates the full 3D-RISM solvent distribution.  This is required to 
!!!calculate thermodynamic quantities
!!!IN:
!!!   this :: rism3d object
!!!   ksave  :: save itermediate results every ksave interations (0 means no saves)
!!!   kshow  :: print parameter for relaxation steps every kshow iteration (0 means no saves)
!!!   maxste :: maximum number of rism relaxation steps
!!!   tol    :: convergence tolerance
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_solve (this,ksave,kshow,maxste,tol)
      use rism_util, only : calc_cm,corr_drift, translate
      use constants, only : KB
      use quaternion, only : rotate_quat
      implicit none
#if defined(MPI)
      include 'mpif.h'
#endif /*defined(MPI)*/
      type(rism3d), intent(inout) :: this
      integer, intent(in) :: ksave,kshow,maxste
      _REAL_, intent(in) :: tol
      !huvk0  :: long-range part of Huv(k) at k=0.  calculated in asympch
      !xmu    :: not sure, related to xmu0 but only for aexchem
      _REAL_ ::  huvk0(2,this%solv%natom),xmu
      !bridge :: flag for calculating the bridge correction or not.  
      !          Currently, this is always false and not used at all.
      logical ::  bridge
      
      integer :: iatu

      _REAL_ :: mpi_temp
      integer :: ierr,id, irank
!!!!
!! 1) Reorient solute along the principal axis and resize the grids if necessary
!!!!
#ifdef RISM_DEBUG
      write(0,*) "(1)"
      write(0,*) "RATU", this%mpirank, this%solu%natom, this%grid%nr, this%grid%nga
      call flush(0)
#endif /*RISM_DEBUG*/

!!!!
!! 1b) get the minimum box size for this frame
!!!!
!      if(this%varbox .or. .not. associated(this%grid%ga))then
      if(this%varbox .or. this%nsolution ==0)then
         call timer_start(TIME_RESIZE)
         call resizebox(this)
         call timer_stop(TIME_RESIZE)
      end if
      call timer_start(TIME_REORIENT)
      call center_solute(this)
      call timer_stop(TIME_REORIENT)

!!!!
!! 3) Calculate electrostatic and Lennard-Jones potential about the solute
!!!!
#ifdef RISM_DEBUG
      write(0,*) "(3)"
      call flush(0)
#endif /*RISM_DEBUG*/
      call rism3d_potential_calc(this%pot)

!!!!
!! 4) Calculate long range asymptotpics
!!!!
#ifdef RISM_DEBUG
      write(0,*) "(4)"
      call flush(0)
#endif /*RISM_DEBUG*/
      call timer_start(TIME_ASYMP)
      !.....tabulating asymptotic function of Cuv and Huv.....
      call rism3d_potential_asympch(this%pot,huvk0)
      call timer_stop(TIME_ASYMP)

!!!!
!! 2) Propagate previously saved CUV solutions to create an initial guess for this solution
!!!!
#ifdef RISM_DEBUG
      write(0,*) "(2)", this%grid%nkTotal, this%grid%nga
      call flush(0)
#endif /*RISM_DEBUG*/
      call timer_start(TIME_CUVPROP)
      call cuv_propagate(this)
      call timer_stop(TIME_CUVPROP)

!!!!
!! 5) Calculate 3D-RISM solution
!!!!
#ifdef RISM_DEBUG
      write(0,*) "(5)"
      call flush(0)
#endif /*RISM_DEBUG*/
      !................. solving 3D-RISM/HNC(PLHNC) by MDIIS .................
      bridge = .false.
      call timer_start(TIME_RXRISM)
#ifdef RISM_DEBUG
      write(0,*)"pre-rxrism"
      call flush(0)
#endif /*RISM_DEBUG*/
      call  rxrism(this,bridge,huvk0,ksave,kshow,maxste,tol)
      call timer_stop(TIME_RXRISM)

!!!!
!! 11) Update stored variables
!!!!
#ifdef RISM_DEBUG
      write(0,*) "(11)", ubound(this%oldcuv)
      call flush(0)
#endif /*RISM_DEBUG*/
      call timer_start(TIME_CUVPROP)
      this%nsolution = this%nsolution+1
      call cuv_update(this)
      call timer_stop(TIME_CUVPROP)

#ifdef RISM_DEBUG
      write(0,*) "DONE SOLVE"
      call flush(0)
#endif /*RISM_DEBUG*/
    end subroutine rism3d_solve

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!THERMODYNAMICS
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!calculates the forces on the solute contributed by the solvent according
!!!to 3D-RISM.  In fact, this subroutine calls the appropriate subroutines
!!!to calculate this
!!!IN:
!!!   this :: rism3d object with computed solution
!!!   ff   :: 3D-RISM forces
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine rism3d_force (this,ff)
    implicit none
    type(rism3d):: this
    _REAL_,intent(out) :: ff(3,this%solu%natom)
    !................... Calculation force fields    .......................
    call timer_start(TIME_FF)
    call rism3d_closure_force(this%closure,ff,this%guv)
    call timer_stop(TIME_FF)
  end subroutine rism3d_force

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the excess chemical potential of solvation for each solvent species
!!!IN:
!!!   this :: rism3d object with computed solution
!!!   lr   :: (optional) (default=.true.) Apply asymptotic long range correction
!!!OUT:
!!!    excess chemical potential of solvation for each solvent species
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism3d_exchem (this,lr) result (exchem)
      implicit none
      type(rism3d), intent(inout) :: this
      logical, optional, intent(in) :: lr
      logical :: opt_lr
      _REAL_ :: exchem(this%solv%natom)

      opt_lr = .true.
      if(present(lr)) opt_lr = lr
      
      if(opt_lr)then
         exchem = rism3d_closure_aexchem(this%closure,this%huv,this%cuv(:,:,:,:))
      else
         exchem = rism3d_closure_exchem(this%closure,this%huv,this%cuv(:,:,:,:))
      end if
    end function rism3d_exchem

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the total excess chemical potential of solvation 
!!!IN:
!!!   this :: rism3d object with computed solution
!!!   lr   :: (optional) (default=.true.) Apply asymptotic long range correction
!!!OUT:
!!!    total excess chemical potential of solvation
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism3d_exchem_tot (this,lr) result (exchem)
      implicit none
      type(rism3d), intent(inout) :: this
      logical, optional, intent(in) :: lr
      logical :: opt_lr
      _REAL_ :: exchem

      opt_lr = .true.
      if(present(lr)) opt_lr = lr

      exchem = sum(rism3d_exchem(this,opt_lr))
    end function rism3d_exchem_tot

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the excess chemical potential of solvation for each solvent species
!!!with the Gaussian fluctuation correction
!!!IN:
!!!   this :: rism3d object with computed solution
!!!   lr   :: (optional) (default=.true.) Apply asymptotic long range correction
!!!OUT:
!!!    Gaussian fluctuation excess chemical potential of solvation for each 
!!!    solvent species
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism3d_exchemGF (this,lr) result (exchem)
      implicit none
      type(rism3d), intent(inout) :: this
      logical, optional, intent(in) :: lr
      logical :: opt_lr
      _REAL_ :: exchem(this%solv%natom)

      opt_lr = .true.
      if(present(lr)) opt_lr = lr
      if(opt_lr)then
         exchem = rism3d_closure_aexchemGF(this%closure,this%huv,this%cuv(:,:,:,:))
      else
         exchem = rism3d_closure_exchemGF(this%closure,this%huv,this%cuv(:,:,:,:))
      end if
    end function rism3d_exchemGF

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the total excess chemical potential of solvation with the Gaussian 
!!!fluctuation correction
!!!IN:
!!!   this :: rism3d object with computed solution
!!!   lr   :: (optional) (default=.true.) Apply asymptotic long range correction
!!!OUT:
!!!    total Gaussian fluctuation excess chemical potential of solvation
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism3d_exchemGF_tot (this,lr) result (exchem)
      implicit none
      type(rism3d), intent(inout) :: this
      logical, optional, intent(in) :: lr
      logical :: opt_lr
      _REAL_ :: exchem

      opt_lr = .true.
      if(present(lr)) opt_lr = lr
      
      exchem = sum(rism3d_exchemGF(this,opt_lr))
    end function rism3d_exchemGF_tot

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the excess solvation energy: de = dmu + TdS = rho sum g*u for each
!!!solvent site
!!!IN:
!!!   this :: rism3d object with computed solution
!!!OUT:
!!!    the contribution of each solvent site to the total excess solvation 
!!!    energy [kT]
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism3d_exEne (this) result (ene)
      implicit none
      type(rism3d), intent(inout) :: this
      _REAL_ :: ene(this%solv%natom)
      ene = rism3d_closure_exEne(this%closure,this%guv)
    end function rism3d_exEne

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculate the total excess solvation energy: de = dmu + TdS = rho sum g*u 
!!!IN:
!!!   this :: rism3d object with computed solution
!!!OUT:
!!!    the total excess solvation energy [kT]
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rism3d_ene_tot (this) result (ene)
      implicit none
      type(rism3d), intent(inout) :: this
      _REAL_ :: ene

      ene = sum(rism3d_exEne(this))
    end function rism3d_ene_tot

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculating PMV of solute
!!!IN:
!!!   this :: rism3d object with computed solution
!!!OUT:
!!!   partial molar volume
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    function  rism3d_pmv (this) result(pmv)
      implicit none
      type(rism3d), intent(inout) :: this
      _REAL_ :: pmv
      pmv = rism3d_closure_pmv(this%closure,this%cuv(:,:,:,:))
    end function rism3d_pmv

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Calculating excess number of each solvent type associated with the solute
!!!IN:
!!!   this :: rism3d object with computed solution
!!!   lr   :: (optional) (default=.true.) Apply asymptotic long range correction
!!!OUT:
!!!    excess number of each solvent type associated with the solute
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    function  rism3d_exNum (this,lr) result(num)
      implicit none
      type(rism3d), intent(inout) :: this
      logical, optional, intent(in) :: lr
      logical :: opt_lr
      _REAL_ :: num(this%solv%natom)

      opt_lr = .true.
      if(present(lr)) opt_lr = lr
      
      if(opt_lr)then
         num = rism3d_closure_aexNum(this%closure,this%guv)
      else
         num = rism3d_closure_exNum(this%closure,this%guv)
      end if
    end function rism3d_exNum

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!DEALLOCATE
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! deconstructor - frees all memory
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rism3d_destroy(this)
      use safemem
      implicit none
      type(rism3d) :: this

      call rism3d_solv_destroy(this%solv)
      call rism3d_solu_destroy(this%solu)
      call rism3d_potential_destroy(this%pot)
      call rism3d_grid_destroy(this%grid)
      call rism3d_closure_destroy(this%closure)

      if(safemem_dealloc(this%xvva)/=0) call rism_report_error("RISM3D: failed to deallocate XVVA")
      if(safemem_dealloc(this%guv)/=0) call rism_report_error("RISM3D: failed to deallocate GUV")
      if(safemem_dealloc(this%huv)/=0) call rism_report_error("RISM3D: failed to deallocate HUV")
      if(safemem_dealloc(this%cuvWRK)/=0) call rism_report_error("RISM3D: failed to deallocate CUVWRK")
      if(safemem_dealloc(this%oldcuvChg)/=0) call rism_report_error("RISM3D: failed to deallocate OLDCUVCHG")
      if(safemem_dealloc(this%oldcuvNoChg)/=0) call rism_report_error("RISM3D: failed to deallocate OLDCUVNOCHG")
      if(safemem_dealloc(this%cuvresWRK)/=0) call rism_report_error("RISM3D: failed to deallocate CUVRESWRK")
      deallocate(this%nsolutionChg, this%nsolutionNoChg)
      nullify(this%cuv)
      nullify(this%cuvres)
      nullify(this%oldcuv)
      nullify(this%nsolution)
#  if defined(MPI)
      call rfftwnd_f77_mpi_destroy_plan(this%planfwd)
      call rfftwnd_f77_mpi_destroy_plan(this%planbwd)
#  else
      call rfftwnd_f77_destroy_plan(this%planfwd)
      call rfftwnd_f77_destroy_plan(this%planbwd)
#  endif /*defined(MPI)*/
    end subroutine rism3d_destroy

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!PRIVATE
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!using the current orientation of the solute, define the minimum box size
!!!and resize all associated grids
!!!IN:
!!!   this :: rism3d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine resizebox(this)
      use rism_util, only : isprime, lcm
      implicit none
#if defined(MPI)
      include 'mpif.h'
      integer :: ierr
#endif /*defined(MPI)*/
      type(rism3d), intent(inout) :: this
      integer :: tempngr(3),ngr(3)
      integer :: id
      _REAL_ :: boxlen(3)
#  ifdef RISM_DEBUG
      write(0,*) "RESIZEBOX"
      call flush(0)
#  endif /*RISM_DEBUG*/

      !if this%mpisize is prime _and_ larger than 5, we can not
      !properly distribute the run.  Thus, we need to tell the user the
      !bad news and exit gracefully
      if(this%mpisize > 5 .and. isprime(this%mpisize))then
         if(this%mpirank == 0)then
         call rism_report_error("Sorry, 3D-RISM requires that if there are more than "//&
              "five MPI processes the number of processes cannot be prime.")
         end if
         call mexit(6,1)
      end if
      !if we have a fixed box size, we simply retain all of the previously calculated box size values
      if(this%varbox) then
         !get minimum box size defined by the buffer
         do id = 1,3
            boxlen(id) = maxval(this%solu%ratu(id,:)) - minval(this%solu%ratu(id,:)) + 2*this%buffer
         end do

         !round this box size using the prescribed linear density and get the
         !number of grid points required
         tempngr = nint(boxlen/this%grid%grdspc)
         boxlen = tempngr*this%grid%grdspc

         !determine if the number of grid points in each dimension product
         !only has prime factors 2, 3 or 5. If not increment the number of
         !points (in that dimension) until this is true

         !make sure that each dimension is divisable by 2 and that the
         !y- and z-dimensions are divisible by this%mpisize if this%mpisize > 1

         do id = 1,3
            tempngr(id) = tempngr(id)+mod(tempngr(id),2)
         end do
         if(this%mpisize > 2)then
            do id=2,3
               if(mod(tempngr(id),this%mpisize) /=0) then
                  tempngr(id) = tempngr(id)+lcm(this%mpisize,2)-mod(tempngr(id),lcm(this%mpisize,2))
               end if
            end do
         end if
         ngr = 1

         do id = 1,3
            do while(.true.) 
               if(mod(tempngr(id),2) == 0)then
                  tempngr(id) = tempngr(id)/2
                  ngr(id) = ngr(id)*2
               elseif(mod(tempngr(id),3) == 0)then
                  tempngr(id) = tempngr(id)/3
                  ngr(id) = ngr(id)*3
               elseif(mod(tempngr(id),5) == 0)then
                  tempngr(id) = tempngr(id)/5
                  ngr(id) = ngr(id)*5
               else
                  if(this%mpisize > 1 .and. id > 1)then
                     tempngr(id) = tempngr(id)*ngr(id)+lcm(this%mpisize,2)
                  else
                     tempngr(id) = tempngr(id)*ngr(id)+2
                  end if
                  ngr(id) = 1
               endif
               if(tempngr(id) ==1)then
                  exit
               end if
            end do
         end do
         boxlen = ngr*this%grid%grdspc

#ifndef MPI
         if(.not. associated(this%cuv) .or. ngr(1) /= size(this%cuv,1) .or.&
              size(this%cuv,2) /= ngr(2) .or. size(this%cuv,3) /= ngr(3)) then
            call reallocbox(this,ngr,this%grid%grdspc)
         end if
#else
         if(.not. associated(this%cuv) .or. ngr(1) /= size(this%cuv,1) &
              .or. size(this%cuv,2) /= ngr(2) &
              .or. size(this%cuv,3) /= ngr(3)/this%mpisize) then
            call reallocbox(this,ngr,this%grid%grdspc)
         end if
#endif /*defined(MPI)*/
      else !fixed box size
         do id =2,3
            if(this%mpirank == 0 .and. (mod(this%nboxfix(id),this%mpisize) /=0 &
                 .or. mod(this%nboxfix(id),2) /=0) )then
               call rism_report_error("Sorry, MPI 3D-RISM requires that fixed grid sizes be"//&
                "divisible by two and the number of processes in the"//&
                "y and z dimensions")
            end if
         end do
         call reallocbox(this,this%nboxfix,this%boxfixlen/this%nboxfix)
      end if
    end subroutine resizebox

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!using the current box size and resize all associated grids and variables
!!!IN:
!!!   this :: rism3d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
subroutine reallocbox(this,ngr,grdspc)
  use rism_fft
  use safemem
  implicit none
  type(rism3d) :: this
  integer, intent(in) :: ngr(3)
  _REAL_,intent(in) :: grdspc(3)
  integer :: nkTotal, nkyOff,nky, nrzOff, nrz
  integer :: i, id, irank
#ifdef RISM_DEBUG
  write(0,*) "reallocbox",this%planfwd,this%planbwd
  call flush(0)
#endif /*RISM_DEBUG*/

  !
  !1) variables
  !
  !      
  !this must be called to determine how to  distibute the memory
  !
  if(this%planfwd /= 0)then
#if defined(MPI)
     !FFTW MPI memory layout
     !-transpose in yz
     !-extra 2 X nr(2) X nr(3) located added to x index
     call rfftwnd_f77_mpi_destroy_plan(this%planfwd)
     call rfftwnd_f77_mpi_destroy_plan(this%planbwd)
#else
     call rfftwnd_f77_destroy_plan(this%planfwd)
     call rfftwnd_f77_destroy_plan(this%planbwd)
#endif /*defined(MPI)*/
  end if
  
#if defined(MPI)
  call fftw_make_plans(this%planfwd,this%planbwd,ngr,nrz,&
       nrzOff,nky,nkyOff, nkTotal,this%mpicomm)
#  ifdef RISM_DEBUG
  write(0,*) "FFTW Plans Made"
  write(0,*) "nz_local", nrz
  write(0,*) "nz_start", nrzOff
  write(0,*) "nky_local", nky
  write(0,*) "nky_start", nkyOff
  write(0,*) "total_work", nkTotal
  !      stop
#  endif /*RISM_DEBUG*/
#else
  !Numerical recipes memory layout
  !-no transpose
  !-extra 2 X nr(2) X nr(3) in k-space starts at nr(1)*nr(2)*nr(3)+1
  nrz = ngr(3)
  nrzOff=0
  nky = ngr(2)
  nkyOff=0
  nkTotal = (ngr(1)+2)*ngr(2)*ngr(3)
  call fftw_make_plans(this%planfwd,this%planbwd,ngr)
#endif /*MPI*/
  
  call rism3d_grid_resize(this%grid,grdspc,&
       ngr,&                                   !r-space global
       (/(ngr(1)+2),ngr(3),ngr(2)/),&          !k-space global
       (/ngr(1),ngr(2),nrz/),&       !r-space local
       (/(ngr(1)+2),ngr(3),nky/),&   !k-space local
       (/0,0,nrzOff/), &                !r-space offset
       (/0,0,nkyOff/))                  !k-space offset
  !
  !2) allocation
  !
  !reallocate arrays that require preservation of their contents
#if defined(MPI)
  this%cuvWRK => safemem_realloc(this%cuvWRK,this%grid%nr(1),this%grid%nr(2),this%grid%nr(3),&
       this%solv%natom,this%NVec,.true.)
  if(rism3d_solu_charged(this%solu))then
     this%oldcuvChg => safemem_realloc(this%oldcuvChg,this%grid%nr(1),this%grid%nr(2),&
          this%grid%nr(3),this%solv%natom,this%ncuvsteps,.true.)
     this%oldcuv => this%oldcuvChg
  else
     this%oldcuvNoChg => safemem_realloc(this%oldcuvNoChg,this%grid%nr(1),this%grid%nr(2),&
          this%grid%nr(3),this%solv%natom,this%ncuvsteps,.true.)
     this%oldcuv => this%oldcuvNoChg
  end if
#else
  this%cuvWRK => safemem_realloc(this%cuvWRK,this%grid%nr(1),this%grid%nr(2),this%grid%nr(3),&
       this%solv%natom,this%NVec,.true.)
  if(rism3d_solu_charged(this%solu))then
     this%oldcuvChg => safemem_realloc(this%oldcuvChg,this%grid%nr(1),this%grid%nr(2),&
          this%grid%nr(3),this%solv%natom,this%ncuvsteps,.true.)
     this%oldcuv => this%oldcuvChg
  else
     this%oldcuvNoChg => safemem_realloc(this%oldcuvNoChg,this%grid%nr(1),this%grid%nr(2),&
          this%grid%nr(3),this%solv%natom,this%ncuvsteps,.true.)
     this%oldcuv => this%oldcuvNoChg
  end if
!!$  this%oldcuv => safemem_realloc(this%oldcuv,this%grid%nr(1),this%grid%nr(2),&
!!$       this%grid%nr(3),this%solv%natom,this%ncuvsteps,.true.)
#endif /*MPI*/
  !reallocate arrays that do not require preservation of their contents
  this%guv => safemem_realloc(this%guv,this%grid%nkTotal,this%solv%natom,.false.)
  this%huv => safemem_realloc(this%huv,this%grid%nkTotal,this%solv%natom,.false.)
  this%cuvresWRK => safemem_realloc(this%cuvresWRK,this%grid%nrTotal,this%solv%natom,&
       this%NVec,.false.)
  this%xvva => safemem_realloc(this%xvva,this%grid%nga* (this%solv%natom) **2,.false.)

  !updated pointers
  this%cuv=>this%cuvWRK(:,:,:,:,1)
  this%cuvres=>this%cuvresWRK(:,:,1)

  !
  !3) the remaining variables are handled by rism_setup_wavevector,interpolate_xvva
  !
  call interpolate_xvva(this)
end subroutine reallocbox

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!interpolate the xvv array for the solvent that has be read in to construct
!!!this%xvva for the specific box size we are using
!!!IN:
!!!  this :: rism3d object
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine interpolate_xvva (this)
      use rism_util, only : poly_interp
      implicit none
      type(rism3d) :: this
      integer :: iiga, itab, itab1,iv1,iv2
      _REAL_ :: err!,y,xa(5)=(/0.1,0.2,0.3,0.4,0.5/),ya(5)=(/-1.6228,-0.8218,-0.3027,0.1048,0.4542/)
      integer :: maxip
      parameter (maxip=5)

#ifdef RISM_DEBUG
      write(0,*) "INTERPOLATE_XVVA", sum(this%solv%fourier_tbl),sum(this%grid%ga); call flush(0)
#endif /*RISM_DEBUG*/

      !........................ checking R-grid size .........................
      if (this%grid%ga(this%grid%nga) > this%solv%fourier_tbl(this%solv%nr))  then
         call rism_report_error('(a,1pe16.8,a,1pe16.8)',&
              'DISTVV: bulk solvent Kmax=',this%solv%fourier_tbl(this%solv%nr),&
              'insufficient for 3D-grid Kmax=',this%grid%ga(this%grid%nga))
      elseif (maxip > this%solv%nr)  then
         call rism_report_error('(a,i7,a,i7)', &
              'DISTVV: bulk solvent grid size Nr=',this%solv%nr,&
              'insufficient for interpolation MaxIp=',maxip)
      endif

      !................. interpolating Xvv(K) at 3D-grid |K| .................
      do iiga=1,this%grid%nga
         do itab=1,this%solv%nr-maxip+1
            itab1 = itab
            if (this%solv%fourier_tbl(itab1+maxip/2) > this%grid%ga(iiga))  then
               !itab1 is now lower bracket for interpolation
               exit
            end if
         enddo
         do iv2=1,this%solv%natom
            do iv1=1,this%solv%natom
               call  poly_interp (this%solv%fourier_tbl(itab1:itab1+maxip), &
                    this%solv%xvv(itab1:itab1+maxip,iv1,iv2),maxip, &
                    this%grid%ga(iiga), &
                    this%xvva(iiga +(iv1-1)*this%grid%nga + (iv2-1)*this%grid%nga*this%solv%natom), &
                    err)
            enddo
         enddo
      enddo
    end subroutine interpolate_xvva

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! center the solute in the solvent box
!!!IN:
!!!   this :: rism3d object    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine center_solute(this)
      use rism_util, only : calc_cm,translate
      implicit none
      type(rism3d), intent(inout) :: this
      _REAL_ :: weight(this%solu%natom)

      !if ucenter is /= 0 we want to move the solute to the center of the solvent box
      !However, if ucenter < 0 we only figure out the displacement required the _first_
      !time we see the solute.  Thus, for ucenter <= 0, the solute's CM can move relative
      !to the grid
      if(abs(this%ucenter)==1)then
         weight=this%solu%mass
      elseif(abs(this%ucenter)==2)then
         weight=1
      end if
      if(this%ucenter > 0 .or. (this%ucenter < 0 .and. this%nsolution == 0))then
         call calc_cm(this%solu%ratu,this%ratucm,weight,this%solu%natom)
      end if
      if(this%ucenter /= 0) then
         call translate(this%solu%ratu,this%solu%natom,-1*this%ratucm)
      end if
      if(this%ucenter/=0)then
         call translate(this%solu%ratu,this%solu%natom,this%grid%boxlen/2d0)
      end if
    end subroutine center_solute

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!subroutines to find the iterative 3D-RISM solution.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Main driver for the 3D-RISM solver
!!!IN:
!!!   this :: rism3d object
!!!   bridge :: apply bridge correction (currently does nothing)
!!!   huvk0  :: long-range part of Huv(k) at k=0.  calculated in asympch
!!!   ksave  :: save itermediate results every ksave interations (0 means no saves)
!!!   kshow  :: print parameter for relaxation steps every kshow iteration (0 means no saves)
!!!   maxste :: maximum number of rism relaxation steps
!!!   mdiis_method :: MDIIS implementation to use    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine rxrism (this,bridge,huvk0,ksave,kshow,maxste,tol)
      use mdiis_c
      use rism3d_csv
      implicit none
#include "def_time.h" 
#if defined(MPI)
      include 'mpif.h'
#endif /*defined(MPI)*/
      type(rism3d), intent(inout) :: this
      integer, intent(in) :: ksave,kshow,maxste
      _REAL_, intent(in) :: tol
      logical ::  bridge
      character(72) ::  cuvsav='rism.csv',guvfile
      integer :: guv_local=77
      !    character*(*)  cuvsav
      !mdiis :: MDIIS solver object
      type(mdiis),save :: mdiis_o


      integer :: iatv, igx, igy, igz
      _REAL_ ::  huvk0(2,this%solv%natom)
      logical ::  found, conver=.false.
      integer ::  ig,iv,iv2,istep
      _REAL_ ::  resoz0=0

      logical, save :: first = .true.

      !irank  :: mpi rank counter
      !stat :: iostat
      integer :: irank,stat, ientry,nentry
      !ierr :: mpi error
      logical :: ierr

#ifdef MPI
      call mdiis_new_mpi(mdiis_o,this%mdiis_method, &
           this%deloz,tol, this%cuvWRK, this%cuvresWRK, &
           this%grid%nrTotal* this%solv%natom, this%NVec,&
           this%mpirank, this%mpisize, this%mpicomm)
#else
      call mdiis_new(mdiis_o,this%mdiis_method, &
           this%deloz,tol, this%cuvWRK, this%cuvresWRK, &
           this%grid%nrTotal* this%solv%natom, this%NVec)
#endif /*MPI*/

      this%cuvres=0
      !!FIX - delete this...

#ifdef MPI
#else
      if(this%mpirank == 0) then 
         inquire (file=cuvsav,exist=found)
         if (found .and. first .and. ksave/=0)  then
#ifdef RISM_DEBUG
            write(0,*)'reading saved Cuv file:  ',cuvsav
#endif /*RISM_DEBUG*/
            call  rdufma (cuvsav, this%cuv(:,:,:,:),this%grid%nrTotal,&
                 this%solv%natom)
         else
#endif /*MPI*/
            !!ENDFIX
            !..... initial Cuv(R)
            if(first .or. this%ncuvsteps == 0) then
               !       if(.true.) then
               this%cuv=0
               !..... add long-range part,
               !..... because this long-range part is subtracted in next routine..
               do iv=1,this%solv%natom
!!$               do igz=1,this%grid%ngr(3)
                  do igz=1,this%grid%nr(3)
                     do igy=1,this%grid%nr(2)
                        do igx=1,this%grid%nr(1)
                           ig = igx + (igy-1)*this%grid%nr(1) + (igz-1)*this%grid%nr(2)*this%grid%nr(1)
                           this%cuv(igx,igy,igz,iv) = this%cuv(igx,igy,igz,iv)&
                                + this%solv%charge(iv)*this%pot%asymcr(ig)
                        enddo
                     enddo
                  enddo
               enddo
            endif
#ifdef MPI
#else
         end if
      end if
#endif /*MPI*/
      !.......................... relaxing UV RISM  ..........................
#ifdef RISM_DEBUG
      write(0,*)'relaxing 3D uv RISM:'
      call flush(0)
#endif

      do istep=1,maxste
         
         !................... one relaxation step of UV RISM ....................
         call timer_start(TIME_R1RISM)    
         call r1rism(this,resoz0,conver,bridge,huvk0,mdiis_o,tol)
         call timer_stop(TIME_R1RISM)    
         
         !............. showing selected and last relaxation steps ..............
         if (kshow /= 0 .and. this%mpirank == 0 .and. this%verbose >= 2)  then
            if (conver .OR. mod(istep,kshow) == 0 .OR. &
                 ksave > 0 .AND. mod(istep,max(ksave,1)) == 0)  then
               call rism_report_message('(a,i5,5x,a,1pg10.3,5x,a,i3)',&
                    ' Step=',istep, 'Resid=',resoz0, 'IS=',getCurrentNVec(mdiis_o))
               call rism_report_flush()
            endif
         endif

         !!FIX _ DELETE this
         !.............. saving selected and last relaxation steps ..............
#ifdef MPI
#else
         if (ksave /= 0 .and. first)  then
            if (conver .OR. ksave > 0 .AND. mod(istep,ksave) == 0)  then
               call  wrufma (cuvsav, this%cuv(:,:,:,:),this%grid%nrTotal,this%solv%natom)
            endif
         endif
#endif /*MPI*/
         !!endfix
         !............... exiting relaxation loop on convergence ................
         if (conver)  goto 30
      enddo
      call rism_report_error('(a,i5)','RXRISM: reached limit # of relaxation steps: ',maxste)
30    first = .false.
      if(this%mpirank == 0 .and. this%verbose >= 1) then
         call rism_report_message('(a,i5,a)',"|RXRISM converged in ",istep)!," steps")
      end if
      call mdiis_destroy(mdiis_o)
      this%cuv=>this%cuvWRK(:,:,:,:,1)
      this%cuvres=>this%cuvresWRK(:,:,1)
      return
    end subroutine rxrism

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!! One Relaxation Step                           *
!!! for the UV RISM equation with the HNC Closure,             *
!!! Guv(R) = exp( - This%pot%uuv(R) + Tuv(R) - DelHv0 ) + DelHv0        *
!!! Cuv(R) = Guv(R) - 1 - Tvv(R)                       *
!!! Huv(K) = Cuv(K) * (Wvv(K)+Rho*Hvv(K))                 *
!!! TuvRes(R) = Huv(R) - Guv(R) - 1                    *
!!!IN:
!!!  this :: rism3d object
!!!  resoz0 ::
!!!  conver ::
!!!  bridge :: apply bridge correction (currently does nothing)
!!!  huvk0  :: long-range part of Huv(k) at k=0.  calculated in asympch
!!!  mdiis  :: MDIIS object to accelerate convergence
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine r1rism(this,resoz0,conver,bridge,huvk0,mdiis_o,tol)
      use rism_fft
      use mdiis_c
      use rism_util, only : checksum
    implicit none
#include "def_time.h" 
#if defined(MPI)
      include 'mpif.h'
#endif /*defined(MPI)*/
      type(rism3d) :: this
    logical ::  conver, bridge

    type(mdiis)::mdiis_o
    integer :: iis
    _REAL_, intent(inout) ::  resoz0
    _REAL_, intent(in) :: tol
    _REAL_ :: earg, tuv0,tvvr
    _REAL_ ::  huvk0(2,this%solv%natom)
    integer :: istep

    integer ::  ig1,iga, iv,iv1,iv2, igx, igy, igz, igk
#ifdef FFW_THREADS
    integer :: nthreads,totthreads
    integer,external :: OMP_get_max_threads,OMP_get_num_threads
    logical,external :: OMP_get_dynamic,OMP_get_nested
#endif

#ifdef RISM_DEBUG
    write(0,*)"R1RISM"
    call flush(0)
#endif

    !.....subtract short-range part from Cuv(R)
    !.....short-range part of Cuv(R) is loaded in guv array
#if defined(MPI)
    do iv=1,this%solv%natom
       do igz=1,this%grid%nr(3)
          do igy=1,this%grid%nr(2)
             do igx=1,this%grid%nr(1)
                ig1 = igx + (igy-1)*this%grid%nr(1) + (igz-1)*this%grid%nr(2)*this%grid%nr(1)
                igk = igx + (igy-1)*(this%grid%nr(1)+2) + (igz-1)*this%grid%nr(2)*(this%grid%nr(1)+2)
                this%guv(igk,iv) = this%cuv(igx,igy,igz,iv) - this%solv%charge(iv)*this%pot%asymcr(ig1)
             enddo
             igk = this%grid%nr(1)+1 + (igy-1)*(this%grid%nr(1)+2) + (igz-1)*this%grid%nr(2)*(this%grid%nr(1)+2)
             this%guv(igk:igk +1,iv) =0
          enddo
       enddo
    enddo
#else
    do iv=1,this%solv%natom
       do igz=1,this%grid%nr(3)
          do igy=1,this%grid%nr(2)
             do igx=1,this%grid%nr(1)
                ig1 = igx + (igy-1)*this%grid%nr(1) + (igz-1)*this%grid%nr(2)*this%grid%nr(1)
                this%guv(ig1,iv) = this%cuv(igx,igy,igz,iv) - this%solv%charge(iv)*this%pot%asymcr(ig1)
             enddo
          enddo
       enddo
    enddo
#endif /*defined(MPI)*/
    !.....short-range part of Cuv(R) FFT>K
    call timer_start(TIME_RISMFFT)    
    do iv=1,this%solv%natom
#if defined(MPI)
       call  rlft3i (this%guv(:,iv), this%grid%ngr, 1, &
            this%grid%nr(3),this%planfwd)
       this%guv(2:this%grid%nkTotal:2,iv) =&
            -this%guv(2:this%grid%nkTotal:2,iv)
#else
       call  rlft3i (this%guv(:,iv), this%grid%nkTotal, this%grid%ngr, 1, this%planfwd)
#endif /*defined(MPI)*/
    enddo
    call timer_stop(TIME_RISMFFT)    
    !.....add long-range part to Cuv(K) in K-space
    do iv=1,this%solv%natom
       do ig1=1,this%grid%nkTotal
          this%guv(ig1,iv) = this%guv(ig1,iv) -  this%solv%charge(iv)*this%pot%asymck(ig1)
       enddo
    enddo
    !.....Huv(k) by RISM
    do iv1=1,this%solv%natom
       do ig1=1,this%grid%nkTotal
          this%huv(ig1,iv1) = 0d0
          iga = this%grid%indga((ig1+1)/2)
          do iv2=1,this%solv%natom
             this%huv(ig1,iv1) = this%huv(ig1,iv1) + &
                  this%guv(ig1,iv2)*this%xvva(iga + (iv2-1)*this%grid%nga + (iv1-1)*this%grid%nga*this%solv%natom)
          enddo
       enddo
    enddo
    !.....add long-range part of Huv(k) at k=0
    !.....which was estimated by long-range part of Cuv(k) at k=0
    do ig1=1,2
       do iv=1,this%solv%natom
          this%huv(ig1,iv) = this%huv(ig1,iv) + huvk0(ig1,iv)
       enddo
    enddo
    !.....subtract long-range part from huv in K-space
#if defined(MPI)
    do iv=1,this%solv%natom
       if(this%mpirank==0)then
          do ig1=3,this%grid%nkTotal
             this%huv(ig1,iv) = this%huv(ig1,iv) &
                  + 1d0/this%solv%dielconst * this%solv%charge_sp(iv)*this%pot%asymhk(ig1)
          enddo
       else
          do ig1=1,this%grid%nkTotal
             this%huv(ig1,iv) = this%huv(ig1,iv) &
                  + 1d0/this%solv%dielconst * this%solv%charge_sp(iv)*this%pot%asymhk(ig1)
          enddo
       end if
    enddo
#else
    do iv=1,this%solv%natom
       do ig1=3,this%grid%nkTotal
          this%huv(ig1,iv) = this%huv(ig1,iv) &
               + 1d0/this%solv%dielconst * this%solv%charge_sp(iv)*this%pot%asymhk(ig1)
       enddo
    enddo
#endif /*defined(MPI)*/

    !.....short-range part of Huv(K) FFT>R
    call timer_start(TIME_RISMFFT)    
    do iv=1,this%solv%natom
#if defined(MPI)
       this%huv(2:this%grid%nkTotal:2,iv)=&
            -this%huv(2:this%grid%nkTotal:2,iv)
       call  rlft3i (this%huv(:,iv), &
            this%grid%ngr, -1,this%grid%nr(3),this%planbwd)
#else
       call  rlft3i (this%huv(:,iv),this%grid%nkTotal, this%grid%ngr, -1,this%planbwd)
#endif /*defined(MPI)*/
    enddo
    call timer_stop(TIME_RISMFFT)    

    !.....add long-range part to huv in R-space
    do iv=1,this%solv%natom
       do igz=1,this%grid%nr(3)
          do igy=1,this%grid%nr(2)
             do igx=1,this%grid%nr(1)
                ig1 = igx + (igy-1)*this%grid%nr(1) + (igz-1)*this%grid%nr(2)*this%grid%nr(1)
#if defined(MPI)
                igk = igx + (igy-1)*(this%grid%nr(1)+2) + (igz-1)*this%grid%nr(2)*(this%grid%nr(1)+2)
                this%huv(igk,iv) = this%huv(igk,iv) &
                     +  this%solv%charge_sp(iv)*this%pot%asymhr(ig1)
#else
                this%huv(ig1,iv) = this%huv(ig1,iv) &
                     +  this%solv%charge_sp(iv)*this%pot%asymhr(ig1)
#endif /*defined(MPI)*/
             enddo
          enddo
       enddo
    enddo
    this%cuvres(:,:)=0
    call rism3d_closure_guv(this%closure,this%guv,this%huv,this%cuv)
    do iv=1,this%solv%natom
       do igz=1,this%grid%nr(3)
          do igy=1,this%grid%nr(2)
             do igx=1,this%grid%nr(1)
                ig1 = igx + (igy-1)*this%grid%nr(1) + (igz-1)*this%grid%nr(2)*this%grid%nr(1)
#if defined(MPI)
                igk = igx + (igy-1)*(this%grid%nr(1)+2) + (igz-1)*this%grid%nr(2)*(this%grid%nr(1)+2)
#else
                igk=ig1
#endif /*defined(MPI)*/
                this%cuvres(ig1,iv) = this%guv(igk,iv) - 1d0 - this%huv(igk,iv)
             enddo
          enddo
       enddo
    enddo

    call timer_start(TIME_MDIIS)    
    call mdiis_advance (mdiis_o, resoz0,conver)
    this%cuv=>this%cuvWRK(:,:,:,:,mdiis_getWorkVector(mdiis_o))
    this%cuvres=>this%cuvresWRK(:,:,mdiis_getWorkVector(mdiis_o))
    call timer_stop(TIME_MDIIS)    

  end subroutine r1rism


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! PROPAGATE PREVIOUS SOLUTIONS
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! calculates a new initial guess for CUV based on the final solutions 
!from previous timesteps.  The maximum number of previous time steps to
!used is provided by the user in ncuvsteps.  However, if there are not
!enough previous timesteps only nsolution previous timesteps will be used.
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine cuv_propagate(this)
    implicit none
    type(rism3d) :: this
    integer :: iv, igr,n
    n=this%grid%nrTotal*this%solv%natom
    if(this%ncuvsteps >= 5 .and. this%nsolution >= 5) then
!!$     this%cuv(:,:,:,:) = 5d0*this%cuv(:,:,:,:) - 10d0*this%oldcuv(:,:,:,:,2)&
!!$          +10d0*this%oldcuv(:,:,:,:,3) - 5d0*this%oldcuv(:,:,:,:,4) + this%oldcuv(:,:,:,:,5)
       call dscal(n,5d0,this%cuv(:,:,:,:),1)
       call daxpy(n,-10d0,this%oldcuv(:,:,:,:,2),1,this%cuv(:,:,:,:),1)
       call daxpy(n,10d0,this%oldcuv(:,:,:,:,3),1,this%cuv(:,:,:,:),1)
       call daxpy(n,-5d0,this%oldcuv(:,:,:,:,4),1,this%cuv(:,:,:,:),1)
       call daxpy(n,1d0,this%oldcuv(:,:,:,:,5),1,this%cuv(:,:,:,:),1)
    elseif(this%ncuvsteps >= 4 .and. this%nsolution >= 4) then
!!$     this%cuv(:,:,:,:) = 4d0*this%cuv(:,:,:,:) - 6d0*this%oldcuv(:,:,:,:,2)&
!!$          +4d0*this%oldcuv(:,:,:,:,3) - this%oldcuv(:,:,:,:,4)
       call dscal(n,4d0,this%cuv(:,:,:,:),1)
       call daxpy(n,-6d0,this%oldcuv(:,:,:,:,2),1,this%cuv(:,:,:,:),1)
       call daxpy(n,4d0,this%oldcuv(:,:,:,:,3),1,this%cuv(:,:,:,:),1)
       call daxpy(n,-1d0,this%oldcuv(:,:,:,:,4),1,this%cuv(:,:,:,:),1)
    elseif(this%ncuvsteps >= 3 .and. this%nsolution >= 3) then
!!$     this%cuv(:,:,:,:) = 3d0*(this%cuv(:,:,:,:) - this%oldcuv(:,:,:,:,2))&
!!$          +this%oldcuv(:,:,:,:,3)
       call dscal(n,3d0,this%cuv(:,:,:,:),1)
       call daxpy(n,-1d0,this%oldcuv(:,:,:,:,2),1,this%cuv(:,:,:,:),1)
       call daxpy(n,1d0,this%oldcuv(:,:,:,:,3),1,this%cuv(:,:,:,:),1)
    elseif(this%ncuvsteps >= 2 .and. this%nsolution >= 2) then
!!$     this%cuv(:,:,:,:) = 2*this%cuv(:,:,:,:) - this%oldcuv(:,:,:,:,2)
       call dscal(n,2d0,this%cuv(:,:,:,:),1)
       call daxpy(n,-1d0,this%oldcuv(:,:,:,:,2),1,this%cuv(:,:,:,:),1)
    elseif(this%ncuvsteps == 0) then
       this%cuv(:,:,:,:) = 0
    end if
  end subroutine cuv_propagate

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!Updates the values in the this%oldcuv queue.  The oldest value (the
!ncuvstep index) is pushed out, the remainder of the data is shifted 
!and the newest solution is placed in the first index
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine cuv_update(this)
    implicit none
    type(rism3d) :: this
    integer :: istep,iv,igr
    !update this%oldcuv
#ifdef RISM_DEBUG
    write(0,*) "CUV_UPDATE"; call flush(0)
#endif /*RISM_DEBUG*/  
    if(this%ncuvsteps == 0) return
    do istep = min(this%ncuvsteps,this%nsolution),2,-1
       call dcopy(this%grid%nrTotal*this%solv%natom,this%oldcuv(:,:,:,:,istep-1),1,&
            this%oldcuv(:,:,:,:,istep),1)
    end do
    call dcopy(this%grid%nrTotal*this%solv%natom,this%cuv(:,:,:,:),1,&
         this%oldcuv(:,:,:,:,1),1)
  end subroutine cuv_update
end module rism3d_c
