!<compile=optimized>

!The 3D-RISM-KH software found here is copyright (c) 2010 by Andriy Kovalenko,
!Tyler Luchko and David A. Case.
!
!This program is free software: you can redistribute it and/or modify it
!under the terms of the GNU General Public License as published by the Free
!Software Foundation, either version 3 of the License, or (at your option)
!any later version.
!
!This program is distributed in the hope that it will be useful, but
!WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
!or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
!for more details.
!
!You should have received a copy of the GNU General Public License in the
!../../LICENSE file.  If not, see <http://www.gnu.org/licenses/>.
!
!Users of the 3D-RISM capability found here are requested to acknowledge
!use of the software in reports and publications.  Such acknowledgement
!should include the following citations:
!
!1) A. Kovalenko and F. Hirata. J. Chem. Phys., 110:10095-10112  (1999); 
!ibid. 112:10391-10417 (2000).   
!
!2) A. Kovalenko,  in:  Molecular  Theory  of  Solvation,  edited  by  
!F. Hirata  (Kluwer Academic Publishers, Dordrecht, 2003), pp.169-275.  
!
!3) T. Luchko, S. Gusarov, D.R. Roe, C. Simmerling, D.A. Case, J. Tuszynski,
!and  A. Kovalenko, J. Chem. Theory Comput., 6:607-624 (2010). 

#include "../include/dprec.fh"

module rism_fft

contains

  !***********************************************************************
  ! create FFTW plans for 3D forward and backward transforms.
  ! planfwd - 'pointer' to forward plan.  must be integer*8
  ! planbwd - 'pointer' to backward plan.  must be integer*8
  ! ng3     - size of the array dimensions
  !***********************************************************************
#if defined(MPI)
  subroutine fftw_make_plans(planfwd,planbwd,ng3,nz_local,nz_start,ny_local,ny_start, total_work,comm)
#else
    subroutine fftw_make_plans(planfwd,planbwd,ng3)
#endif /*MPI && MPI*/
      implicit none
#include "fftw_f77.h"

#if defined(MPI)
      include 'mpif.h'
      integer, intent(out) :: nz_local,nz_start,ny_local,ny_start, total_work
      integer,intent(in) :: comm
#endif /*MPI && MPI*/
      integer*8,intent(out) :: planfwd,planbwd
      integer, intent(in) :: ng3(3)
      integer :: err
#ifdef RISM_DEBUG
      write(6,*) 'FFTW_MAKE_PLANS'
      call flush(6)
#endif /*RISM_DEBUG*/

#if defined(MPI)
      call rfftw3d_f77_mpi_create_plan(planfwd,comm,ng3(1),ng3(2),ng3(3),   &
           FFTW_REAL_TO_COMPLEX,             &
           FFTW_ESTIMATE)
      call rfftw3d_f77_mpi_create_plan(planbwd,comm,ng3(1),ng3(2),ng3(3),   &
           FFTW_COMPLEX_TO_REAL,             &
           FFTW_ESTIMATE)
      call rfftwnd_f77_mpi_local_sizes(planfwd,nz_local,nz_start,       &
           ny_local,ny_start, total_work)

#else    
      call rfftw3d_f77_create_plan(planfwd,ng3(1),ng3(2),ng3(3),&
           FFTW_REAL_TO_COMPLEX,FFTW_ESTIMATE + FFTW_IN_PLACE)
      call rfftw3d_f77_create_plan(planbwd,ng3(1),ng3(2),ng3(3),&
           FFTW_COMPLEX_TO_REAL,FFTW_ESTIMATE + FFTW_IN_PLACE)
#endif
    end subroutine fftw_make_plans



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Convert 3D FFT data from FFTW layout to Numerical Recipes layout.  Both 
!!!layouts require (nx+2)*ny*nz elements.  The difference is how the Nyquist
!!!frequency data in reciprocal space is stored.
!!!
!!!NR Layout:
!!!Data is layed out as a contiguous piece of nx*ny*nz memory.  The Nyquist 
!!!frequency data (2*ny*nz) starts at the end of this block.
!!!
!!!FFTW Layout:
!!!Nyqusit frequency data is stored as two extra elements at the end of the first
!!!dimension.
!!!
!!!The strategy employed is to travel through the array, extracting Nyquist 
!!!frequency data to a temporary array, and shift the data to fill in the gaps
!!!as we go.  Once completed, the temporary array is then copied into data.
!!!This does require allocating temporary memory for the Nyquist 
!!!frequencies but this is less than using RESHAPE which creates a full sized
!!!array and copies data back and forth.  Furthermore, we can put this memory on
!!!the heap and avoid exhausting the stack.
!!!IN: 
!!!   data :: 3D FFT array in NR format.  On return, in FFTW format
!!!   n1   :: number of x points
!!!   n2   :: number of y points
!!!   n3   :: number of z points
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine tonr( data, n1,n2,n3)
      use safemem
      implicit none
      !      *copy vartypes

      integer, intent(in) :: n1,n2,n3
      _REAL_, intent(inout) ::  data((n1+2)*n2*n3)
      _REAL_,pointer :: spec(:)=>NULL()
      integer :: i,j,k,i1,j1,k1,ngr,ngk, ispec,ondx, nndx
      integer :: err
      spec => safemem_realloc(spec,2*n2*n3,.false.)
      ngr = n1*n2*n3
      ngk = (n1+2)*n2*n3

      ispec=1
      do k = 1,n3
         do j=1,n2
            ondx = 1+ (j-1)*(n1+2) + (k-1)*(n1+2)*n2
            nndx = 1+ (j-1)*(n1)   + (k-1)*(n1)*n2
            spec(ispec:ispec+1) = data(ondx+n1:ondx+n1+1)
            data(nndx:nndx+n1-1) = data(ondx:ondx+n1-1) 
            ispec = ispec+2
        end do
      end do
      data(ngr+1:ngk) = spec
!!$      data=RESHAPE((/                                                      &
!!$           (((data(i+(j-1)*(n1+2)+(k-1)*(n1+2)*n2),                    &
!!$           i=1,n1),j=1,n2),k=1,n3),            &
!!$           (((data(n1+i1+(j1-1)*(n1+2)+(k1-1)*(n1+2)*n2),              &
!!$           i1=1,2),j1=1,n2),k1=1,n3)           &
!!$           /), SHAPE=(/ngr+2*n2*n3/))
      data(2:ngr+2*n2*n3:2) = -data(2:ngr+2*n2*n3:2)
      err = safemem_dealloc(spec)
    end subroutine tonr

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!Convert 3D FFT data from Numerical Recipes layout to FFTW layout.  Both 
!!!layouts require (nx+2)*ny*nz elements.  The difference is how the Nyquist
!!!frequency data in reciprocal space is stored.
!!!
!!!NR Layout:
!!!Data is layed out as a contiguous piece of nx*ny*nz memory.  The Nyquist 
!!!frequency data (2*ny*nz) starts at the end of this block.
!!!
!!!FFTW Layout:
!!!Nyqusit frequency data is stored as two extra elements at the end of the first
!!!dimension.
!!!
!!!The strategy employed is to move data from the end of the NR array, n1 elements 
!!!at a time, to the end of the aggregate array, filling in Nyquist frequencies 
!!!as we go.  This does require allocating temporary memory for the Nyquist 
!!!frequencies but this is less than using RESHAPE which creates a full sized
!!!array and copies data back and forth.  Furthermore, we can put this memory on
!!!the heap and avoid exhausting the stack.
!!!IN: 
!!!   data :: 3D FFT array in NR format.  On return, in FFTW format
!!!   n1   :: number of x points
!!!   n2   :: number of y points
!!!   n3   :: number of z points
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    subroutine fromnr(data, n1,n2,n3)
      use safemem
      implicit none
      !      *copy vartypes

      integer, intent(in) :: n1,n2,n3
      _REAL_,intent(inout) ::  data((n1+2)*n2*n3)
      _REAL_,pointer :: spec(:)=>NULL()
      integer :: i,j,k, ngr,ngk, ispec, ondx,nndx
      integer :: err
      spec => safemem_realloc(spec,2*n2*n3,.false.)
      ngr = n1*n2*n3
      ngk = (n1+2)*n2*n3
      data(2:ngr+2*n2*n3:2) = -data(2:ngr+2*n2*n3:2)
      spec = data(ngr+1:ngk)
      ispec = size(spec)-1
      do k = 1,n3
         do j=1,n2
            ondx = ngr +1 - (n1 + (j-1)*n1 + (k-1)*n1*n2)
            nndx =  ngk  +1 - (1 + size(spec) - ispec + n1 + (j-1)*n1 + (k-1)*n1*n2)
            data(nndx:nndx+n1-1) = data(ondx:ondx+n1-1) 
            data(nndx+n1:nndx+n1+1) = spec(ispec: ispec+1)
            ispec = ispec-2
         end do
      end do
!!$      data = RESHAPE((/                                                    &
!!$           (((data(i+(j-1)*n1+(k-1)*n1*n2),i=1, n1),                  &
!!$           data(ngr+1+(j-1)*2+(k-1)*2*n2),                         &
!!$           data(ngr+2+(j-1)*2+(k-1)*2*n2),j=1,n2),k=1,n3 )         &
!!$           /),SHAPE=(/ngr+2*n2*n3/))

      err = safemem_dealloc(spec)
    end subroutine fromnr

#if defined(MPI)
    !***********************************************************************
    ! Interface to                              *
    ! 3-Dimensional Fast Fourier Transform of Real Data.           *
    ! Normalizing while Forward FFT (KEY=1)                  *
    !***********************************************************************
    subroutine  rlft3i (data, ng3, key,n_local,plan)
      use safemem
      implicit none
      integer ::  ng3(3), key
      integer :: n_local
      integer*8,intent(in) :: plan
!      _REAL_ ::  data((ng3(1)+2)*ng3(2)*n_local)
!      _REAL_ :: work((ng3(1)+2)*ng3(2)*n_local)
      _REAL_ ::  data(:)
      _REAL_,pointer :: work(:)
      integer ::  ngr, ngr_local,ngk_local, ig, ierr
      integer, parameter :: FFTW_TRANSPOSED_ORDER=1, FFTW_NORMAL_ORDER=0
      !....................... R-space data array size .......................
#  ifdef RISM_DEBUG
      write(0,*) "rlft3i",n_local
      call flush(0)
#  endif /*RISM_DEBUG*/
      work=>safemem_realloc(work,size(data),.false.)
      ngr = ng3(1)*ng3(2)*ng3(3)
      ngr_local = ng3(1)*ng3(2)*n_local
      ngk_local = (ng3(1)+2)*ng3(2)*n_local

      if (key == 1)  then
         data(1:ngk_local) = data(1:ngk_local) / ngr
         call rfftwnd_f77_mpi( plan,1,data,work,1,                    &
              FFTW_TRANSPOSED_ORDER)
      elseif (key == -1)  then
         call rfftwnd_f77_mpi( plan,1,data,work,1,                    &
              FFTW_TRANSPOSED_ORDER)
      end if
      ierr = safemem_dealloc(work)
    end subroutine rlft3i

#else /*defined(MPI)*/
    !***********************************************************************
    ! Interface to                              *
    ! 3-Dimensional Fast Fourier Transform of Real Data.           *
    ! Normalizing while Forward FFT (KEY=1)                  *
    !***********************************************************************
    subroutine  rlft3i (data,ngk, ng3, key,plan)
      implicit none
      integer,intent(in) ::  ngk, ng3(3), key
      integer*8,intent(in) :: plan
      _REAL_, intent(inout) ::  data(ngk)
      integer ::  ngr, ig
      !....................... R-space data array size .......................
      ngr = ng3(1)*ng3(2)*ng3(3)

      if (key == 1)  then
         data(2:ngr+2*ng3(2)*ng3(3):2) = -data(2:ngr+2*ng3(2)*ng3(3):2)
         data(1:ngr) = data(1:ngr) / ngr
         call fromnr(data,ng3(1),ng3(2),ng3(3))
         call rfftwnd_f77_one_real_to_complex(plan, data, 0)
         call tonr(data,ng3(1),ng3(2),ng3(3))
      elseif (key == -1)  then
         call fromnr(data,ng3(1),ng3(2),ng3(3))
         call rfftwnd_f77_one_complex_to_real(plan, data, 0)
         call tonr(data,ng3(1),ng3(2),ng3(3))
         data(2:ngr+2*ng3(2)*ng3(3):2) = -data(2:ngr+2*ng3(2)*ng3(3):2)
      endif
      return
    end subroutine rlft3i
#endif /*defined(MPI)*/

end module rism_fft
