/*
 * Daniel R. Roe
 * 2010-12-07
 * A C implementation of routines for reading and writing the Amber Netcdf 
 * trajectory format. 
 * Original implementation of netcdf in Amber by Jon Mongan.
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h> // For title length
#include "AmberNetcdf.h"

// If MPI define a function that will return current task ID in order to avoid
// printf on non-master threads.
// Otherwise define a function that just returns 0;
int netcdfTaskid();
#ifdef MPI
#include "mpi.h"
int netcdfTaskid() {
  int worldrank;
  MPI_Comm_rank(MPI_COMM_WORLD, &worldrank);
  return worldrank;
}
#else
int netcdfTaskid() { return 0; }
#endif
    
#ifdef BINTRAJ
#include "../../include/netcdf.h"

#define NCFRAME "frame"
#define NCSPATIAL "spatial"
#define NCATOM "atom"
#define NCCELL_SPATIAL "cell_spatial"
#define NCCELL_ANGULAR "cell_angular"
#define NCCOORDS "coordinates"
#define NCTIME "time"
#define NCLABEL "label"
#define NCLABELLEN 5
// For debugging make > 0:
#define NCDEBUGVAL 0

// Macro to initialize the AmberNetcdf structure
#define INIT_AmberNetcdf( _p_ ) \
  _p_->ncid=-1; \
  _p_->frameID=-1; \
  _p_->ncframe=-1; \
  _p_->currentFrame=0; \
  _p_->atomID=-1; \
  _p_->ncatom=-1; \
  _p_->coordID=-1; \
  _p_->cellAngleID=-1; \
  _p_->cellLengthID=-1; \
  _p_->temp0 = 0.0; \
  _p_->spatialID=-1; \
  _p_->labelID=-1; \
  _p_->cell_spatialID=-1; \
  _p_->cell_angularID=-1; \
  _p_->spatialVID=-1; \
  _p_->timeVID=-1; \
  _p_->cell_spatialVID=-1; \
  _p_->cell_angularVID=-1; \
  _p_->TempVarID=-1 

// ----------------- Private Routines --------------------------------
int checkNCerr(int, const char *, ...);
int GetDimInfo(int, const char *, int *);

/*
 * checkNCerr()
 * Check error status of err - print error message if not NC_NOERR
 */
int checkNCerr(int err, const char *message, ...) {
  va_list args;

  if (err!=NC_NOERR) {
    va_start(args,message);
    fprintf(stderr,"NETCDF Error (%s): ",nc_strerror(err));
    vfprintf(stderr,message,args);
    fprintf(stderr,"\n");
    va_end(args);
    return 1;
  }
  return 0;
}

/*
 * GetDimInfo()
 * Return the dimension ID of a given attribute in netcdf file ncid.
 * Also set dimension length.
 */ 
int GetDimInfo(int ncid, const char *attribute, int *length) {
  int dimID;
  size_t slength;
    
  *length = 0;
  slength = 0;
    
  // Get dimid 
  if ( checkNCerr(nc_inq_dimid(ncid, attribute, &dimID),
       "ncGetDimInfo: Getting dimID for attribute %s",attribute)!=0 ) return -1;

  // get Dim length 
  if ( checkNCerr(nc_inq_dimlen(ncid, dimID, &slength),
       "ncGetDimInfo: Getting length for attribute %s",attribute)!=0) return -1;

  *length = (int) slength;
  return dimID;
}
// -------------------------------------------------------------------
#endif

/*  
 * netcdfDebug()
 * For use in printing various attributes of a previously opened netcdf file.
 */
int netcdfDebug(struct AmberNetcdf *A) {
#ifdef BINTRAJ
  int ndimsp, nvarsp, ngattsp,unlimdimidp;
  int err,i,mytaskid;
  char *varname;

  mytaskid = netcdfTaskid();
  /* ncid:    NetCDF ID, from a previous call to nc open or nc create.
   * ndimsp:  Pointer to location for returned number of dimensions defined for 
   *         this netCDF dataset.
   * nvarsp:  Pointer to location for returned number of variables defined for 
   *         this netCDF dataset.
   * ngattsp: Pointer to location for returned number of global attributes 
   *         defined for this netCDF dataset.
   * unlimdimidp: 
   *  Pointer to location for returned ID of the unlimited dimension, if 
   *  there is one for this netCDF dataset. If no unlimited length 
   *  dimension has been defined, -1 is returned.
   */
  varname=(char*) malloc(1024*sizeof(char));
  err=nc_inq(A->ncid,&ndimsp,&nvarsp,&ngattsp,&unlimdimidp);
  if (mytaskid==0) {
    fprintf(stdout,"========== BEG. NETCDF DEBUG ==========\n");
    fprintf(stdout,"nc_inq returned %i\n",err);
    if (err==NC_NOERR)
      fprintf(stdout,"ndimsp=%i  nvarsp=%i  ngattsp=%i  unlimdimidp=%i\n",
              ndimsp,nvarsp,ngattsp,unlimdimidp);
    else
      fprintf(stdout,"NETCDF Error occurred.\n");
    fprintf(stdout,"NC VARIABLES:\n");
  }
  // Print name of each variable defined in netcdf file
  for (i=0; i<nvarsp; i++) {
    err=nc_inq_varname(A->ncid,i,varname);
    if (mytaskid==0) fprintf(stdout,"  Var %i - ",i);
    if (err==NC_NOERR) {
      if (mytaskid==0) fprintf(stdout,"%s\n",varname);
    } else {
      if (mytaskid==0) fprintf(stdout,"NETCDF Error occured.\n");
    }
  }

  // Struct variables
  if (mytaskid==0) {
    fprintf(stdout,"  TempVarID = %i\n",A->TempVarID);

    fprintf(stdout,"==========  END NETCDF DEBUG ==========\n");
  }

  free(varname);
  return 0;
#else
  fprintf(stdout,"Error: NAB compiled without Netcdf support.\n");
  fprintf(stdout,"       Recompile with -DBINTRAJ\n");
  return 1;
#endif
}

/*
 * netcdfLoad()
 * Load the netcdf file pointed to by filename, initialize the AmberNetcdf
 * data structure. Set currentFrame to 0.
 * Return NULL on error.
 */
int netcdfLoad(struct AmberNetcdf *A, char *filename) {
#ifdef BINTRAJ
  int debug;

  debug=NCDEBUGVAL;

  // Check netcdf structure
  if (A==NULL) {
    fprintf(stderr,"Error: netcdfLoad: AmberNetcdf is NULL\n");
    return 1;
  }
  // Close if already open
  //netcdfClose(A);

  INIT_AmberNetcdf(A);

  // Do not print error message on error, just return 1
  // This allows one to gracefully check if this is a netcdf file
  if ( nc_open(filename,NC_NOWRITE,&(A->ncid))!=NC_NOERR ) return 1;

  if (debug>0) fprintf(stdout,"Successfully opened %s, ncid=%i\n",filename,A->ncid);
  if (debug>1) netcdfDebug(A);

  A->frameID=GetDimInfo(A->ncid,NCFRAME,&(A->ncframe));
  if (A->frameID==-1) return 1;
  A->atomID=GetDimInfo(A->ncid,NCATOM,&(A->ncatom));
  if (A->atomID==-1) return 1;
  if (checkNCerr(nc_inq_varid(A->ncid,NCCOORDS,&(A->coordID)),
      "Getting coordinate ID")!=0) return 1;

  // Box info 
  if ( nc_inq_varid(A->ncid,"cell_lengths",&(A->cellLengthID))==NC_NOERR ) {
    if (checkNCerr(nc_inq_varid(A->ncid,"cell_angles",&(A->cellAngleID)),
      "Getting cell angles.")!=0) return 1;
    if (debug>0) fprintf(stdout,"  Netcdf Box information found.\n"); 
  } else {
    A->cellLengthID=-1;
    A->cellAngleID=-1;
  }

  // Replica Temperatures
  if ( nc_inq_varid(A->ncid,"temp0",&(A->TempVarID)) == NC_NOERR ) {
    if (debug>0) fprintf(stdout,"    Netcdf file has replica temperatures.\n");
  } else 
    A->TempVarID=-1;

  // NOTE: TO BE ADDED
  //spatialID, labelID;
  //int cell_spatialID, cell_angularID;
  //int spatialVID, timeVID, cell_spatialVID, cell_angularVID;

  return 0;
#else
  fprintf(stdout,"Error: NAB compiled without Netcdf support.\n");
  fprintf(stdout,"       Recompile with -DBINTRAJ\n");
  return 1;
#endif
} 

/*
 * netcdfClose()
 * Close file 
 */
int netcdfClose(struct AmberNetcdf *A) {
#ifdef BINTRAJ
  if (A==NULL) return 0;
  if (A->ncid!=-1)
    checkNCerr(nc_close(A->ncid),"Closing netcdf file.");
  INIT_AmberNetcdf(A);
#else
  fprintf(stdout,"Error: NAB compiled without Netcdf support.\n");
  fprintf(stdout,"       Recompile with -DBINTRAJ\n");
  return 1;
#endif
  return 0;
}

/*
 * netcdfCreate()
 * Create an Amber netcdf trajectory for the specified #atoms with name 
 * filename. Also add box information if requested.
 */
int netcdfCreate(struct AmberNetcdf *A, char *filename, int natom, int isBox) {
#ifdef BINTRAJ
  //AmberNetcdf A;
  int dimensionID[NC_MAX_VAR_DIMS];
  size_t start[3], count[3];
  char *title;
  char xyz[3];
  char abc[15] = { 'a', 'l', 'p', 'h', 'a', 
                   'b', 'e', 't', 'a', ' ',
                   'g', 'a', 'm', 'm', 'a' };
  int debug;

  debug=NCDEBUGVAL;
  title=NULL;

  // Check netcdf structure
  if (A==NULL) {
    fprintf(stderr,"Error: netcdfCreate: AmberNetcdf is NULL\n");
    return 1;
  }
  // Close if already open
  //netcdfClose(A);
  INIT_AmberNetcdf(A); 

  // Create file
  if (checkNCerr(nc_create(filename,NC_64BIT_OFFSET,&(A->ncid)),
    "Creating Netcdf file %s",filename)) return 1;
  if (debug>0) 
    fprintf(stdout,"    Successfully created Netcdf file %s, ncid %i\n",filename,A->ncid);

  // Frame, Time
  if (checkNCerr(nc_def_dim(A->ncid,NCFRAME,NC_UNLIMITED,&(A->frameID)),
    "Defining frame dimension.")) return 1;
  dimensionID[0]=A->frameID;
  if (checkNCerr(nc_def_var(A->ncid,NCTIME,NC_FLOAT,1,dimensionID,&(A->timeVID)),
    "Defining time variable.")) return 1;
  if (checkNCerr(nc_put_att_text(A->ncid,A->timeVID,"units",10,"picosecond"),
    "Writing time VID units.")) return 1;

  // Spatial
  if (checkNCerr(nc_def_dim(A->ncid,NCSPATIAL,3,&(A->spatialID)),
    "Defining spatial dimension.")) return 1;
  dimensionID[0] = A->spatialID;
  if (checkNCerr(nc_def_var(A->ncid,NCSPATIAL,NC_CHAR,1,dimensionID,&(A->spatialVID)),
    "Defining spatial variable.")) return 1;

  // Atoms
  if (checkNCerr(nc_def_dim(A->ncid,NCATOM,natom,&(A->atomID)),
    "Defining atom dimension.")) return 1;
  A->ncatom=natom;

  // Coords
  dimensionID[0] = A->frameID;
  dimensionID[1] = A->atomID;
  dimensionID[2] = A->spatialID;
  if (checkNCerr(nc_def_var(A->ncid,NCCOORDS,NC_FLOAT,3,dimensionID,&(A->coordID)),
    "Defining coordinates variable.")) return 1;
  if (checkNCerr(nc_put_att_text(A->ncid,A->coordID,"units",8,"angstrom"),
    "Writing coordinates variable units.")) return 1;

  // Cell Spatial
  if (checkNCerr(nc_def_dim(A->ncid,NCCELL_SPATIAL,3,&(A->cell_spatialID)),
    "Defining cell spatial dimension.")) return 1;
  dimensionID[0]=A->cell_spatialID;
  if (checkNCerr(nc_def_var(A->ncid,NCCELL_SPATIAL,NC_CHAR,1,dimensionID,&(A->cell_spatialVID)),
    "Defining cell spatial variable.")) return 1;

  // Cell angular
  if (checkNCerr(nc_def_dim(A->ncid,NCLABEL,NCLABELLEN,&(A->labelID)),
    "Defining label dimension.")) return 1;
  if (checkNCerr(nc_def_dim(A->ncid,NCCELL_ANGULAR,3,&(A->cell_angularID)),
    "Defining cell angular dimension.")) return 1;
  dimensionID[0] = A->cell_angularID;
  dimensionID[1] = A->labelID;
  if (checkNCerr(nc_def_var(A->ncid,NCCELL_ANGULAR,NC_CHAR,2,dimensionID,&(A->cell_angularVID)),
    "Defining cell angular variable.")) return 1;

  // Box Info
  if (isBox>0) {
    dimensionID[0]=A->frameID;
    dimensionID[1]=A->cell_spatialID;
    if (checkNCerr(nc_def_var(A->ncid,"cell_lengths",NC_DOUBLE,2,dimensionID,&(A->cellLengthID)),
      "Defining cell length variable.")) return 1;
    if (checkNCerr(nc_put_att_text(A->ncid,A->cellLengthID,"units",8,"angstrom"),
    "Writing cell length variable units.")) return 1;
    dimensionID[1]=A->cell_angularID;
    if (checkNCerr(nc_def_var(A->ncid,"cell_angles",NC_DOUBLE,2,dimensionID,&(A->cellAngleID)),
      "Defining cell angle variable.")) return 1;
    if (checkNCerr(nc_put_att_text(A->ncid,A->cellAngleID,"units",6,"degree"),
    "Writing cell angle variable units.")) return 1;
  }

  // Set up title
  if (title==NULL) {
    title=(char*) malloc(23*sizeof(char));
    strcpy(title,"AmberNetcdf trajectory");
  }

  // Attributes
  if (checkNCerr(nc_put_att_text(A->ncid,NC_GLOBAL,"title",strlen(title),title),
    "Writing title.")) {if (title!=NULL) free(title); return 1;}
  if (title!=NULL) free(title);
  if (checkNCerr(nc_put_att_text(A->ncid,NC_GLOBAL,"application",5,"AMBER"),
    "Writing application.")) return 1;
  if (checkNCerr(nc_put_att_text(A->ncid,NC_GLOBAL,"program",11,"AmberNetcdf"),
    "Writing program.")) return 1;
  if (checkNCerr(nc_put_att_text(A->ncid,NC_GLOBAL,"programVersion",3,"1.0"),
    "Writing program version.")) return 1;
  if (checkNCerr(nc_put_att_text(A->ncid,NC_GLOBAL,"Conventions",5,"AMBER"),
    "Writing conventions.")) return 1;
  if (checkNCerr(nc_put_att_text(A->ncid,NC_GLOBAL,"ConventionVersion",3,"1.0"),
    "Writing conventions version.")) return 1;

  /* DAN ROE: Replica temperature 
  if (trajInfo->isREMDTRAJ) {
    fprintf(stdout,"NETCDF: Defining replica temperature in output trajectory.\n");
    dimensionID[0] = NCInfo->frameDID;
    netcdfDefineVariable(NCInfo->ncid, "temp0", NC_DOUBLE, 1, dimensionID, &NCInfo->TempVarID);
    netcdfPutAttributeText(NCInfo->ncid, NCInfo->TempVarID,"units","kelvin");
  }*/

  // Set fill mode
  if (checkNCerr(nc_set_fill(A->ncid, NC_NOFILL, dimensionID),
    "NetCDF setting fill value.")) return 1;

  // End netcdf definitions
  if (checkNCerr(nc_enddef(A->ncid),"NetCDF error on ending definitions."))
    return 1;

  // Specify spatial dimension labels
  start[0] = 0;
  count[0] = 3;

  xyz[0] = 'x'; xyz[1] = 'y'; xyz[2] = 'z';
  if (checkNCerr(nc_put_vara_text(A->ncid, A->spatialVID, start, count, xyz),
    "Error on NetCDF output of spatial VID 'x', 'y' and 'z'")) return 1;

  xyz[0] = 'a'; xyz[1] = 'b'; xyz[2] = 'c';
  if (checkNCerr(nc_put_vara_text(A->ncid, A->cell_spatialVID, start, count, xyz),
    "Error on NetCDF output of cell spatial VID 'a', 'b' and 'c'")) return 1;

  start[0] = 0; start[1] = 0;
  count[0] = 3; count[1] = 5;
  if (checkNCerr(nc_put_vara_text(A->ncid, A->cell_angularVID, start, count, abc),
    "Error on NetCDF output of cell angular VID 'alpha', 'beta ' and 'gamma'")) 
    return 1;

  return 0;
#else
  fprintf(stdout,"Error: NAB compiled without Netcdf support.\n");
  fprintf(stdout,"       Recompile with -DBINTRAJ\n");
  return 1;
#endif
}

/*
 * netcdfGetFrame()
 * Get the specified frame from amber netcdf file.
 * Also get the box coords if present and box is not NULL.
 * Also get temperature T if TempVarID defined and T is not NULL.
 * Coords are a 1 dimensional array of format X1,Y1,Z1,X2,Y2,Z2,...
 */
int netcdfGetFrame(struct AmberNetcdf *A, int set, double *X, double *box) {
#ifdef BINTRAJ
  size_t start[3], count[3];
  float *Coord;
  int i;

  if (A==NULL) {
    fprintf(stderr,"Error: netcdfGetFrame: AmberNetcdf structure not allocated.\n");
    return 1;
  }

  if (A->ncid==-1) {
    fprintf(stderr,"Error: netcdfGetFrame: AmberNetcdf file not open.\n");
    return 1;
  }

  if (X==NULL) {
    fprintf(stderr,"Error: netcdfGetFrame: Memory for coords not allocated.\n");
    return 1;
  }

  // Frame bounds check
  if (set>=A->ncframe || set<0) return 1;

  // Get temperature
  if (A->TempVarID!=-1) {
    start[0]=set;
    count[0]=1;
    if ( checkNCerr(nc_get_vara_double(A->ncid, A->TempVarID, start,count,&(A->temp0)),
                    "Getting replica temperature.")!=0 ) return 1;
    //fprintf(stderr,"DEBUG: Replica Temperature %lf\n",A->temp0);
  }

  // Allocate memory for reading single precision coordinates.
  Coord=(float*) malloc(A->ncatom * 3 * sizeof(float));
  if (Coord==NULL) {
    fprintf(stderr,"Error: netcdfGetFrame: Could not allocate memory for coords.\n");
    return 1;
  }  

  // Read Coords 
  start[0]=set;
  start[1]=0;
  start[2]=0;
  count[0]=1;
  count[1]=A->ncatom;
  count[2]=3;
  if ( checkNCerr(nc_get_vara_float(A->ncid, A->coordID, start, count, Coord),
                  "Getting frame %i",set)!=0 ) {free(Coord); return 1;}

  // Read box info 
  if (box!=NULL && A->cellLengthID!=-1) {
    count [1]=3;
    count [2]=0;
    if ( checkNCerr(nc_get_vara_double(A->ncid, A->cellLengthID, start, count, box),
                    "Getting cell lengths.")!=0 ) {free(Coord); return 1;}
    if ( checkNCerr(nc_get_vara_double(A->ncid, A->cellAngleID, start, count, box+3),
                    "Getting cell angles.")!=0 ) {free(Coord); return 1;}
  }

  // Convert from single precision to double precision
  for (i=0; i < A->ncatom * 3; i++)
    X[i]=(double) Coord[i];

  // Set the currentFrame
  A->currentFrame=set; 

  // Free temporary memory
  free(Coord);

  return 0;
#else
  fprintf(stdout,"Error: NAB compiled without Netcdf support.\n");
  fprintf(stdout,"       Recompile with -DBINTRAJ\n");
  return 1;
#endif
}

/*
 * netcdfGetNextFrame()
 * Get the frame at currentFrame and increment. Return 0 if no more frames.
 */
int netcdfGetNextFrame(struct AmberNetcdf *A, double *X, double *box) {
#ifdef BINTRAJ
  if ( netcdfGetFrame(A, A->currentFrame, X, box) ) return 0;
  A->currentFrame++;
  return 1;
#else
  fprintf(stdout,"Error: NAB compiled without Netcdf support.\n");
  fprintf(stdout,"       Recompile with -DBINTRAJ\n");
  return 0;
#endif
}

/*
 * netcdfWriteFrame()
 * Write coords (and box if specified) to netcdf file. Return 1 on error. 
 */
int netcdfWriteFrame(struct AmberNetcdf *A, int set, double *X, double *box) {
#ifdef BINTRAJ
  size_t start[3], count[3];
  float *Coord;
  int i;

  if (A==NULL) {
    fprintf(stderr,"Error: netcdfWriteFrame: AmberNetcdf structure not allocated.\n");
    return 1;
  } 

  if (A->ncid==-1) {
    fprintf(stderr,"Error: netcdfWriteFrame: AmberNetcdf file not open.\n");
    return 1;
  }
    
  if (X==NULL) {
    fprintf(stderr,"Error: netcdfWriteFrame: Memory for coords not allocated.\n");
    return 1;
  }

  // Bounds check
  if (set<0) return 1;

  // Allocate memory for reading single precision coordinates.
  Coord=(float*) malloc(A->ncatom * 3 * sizeof(float));
  if (Coord==NULL) {
    fprintf(stderr,"Error: netcdfCreate: Could not allocate memory for coords.\n");
    return 1;
  }  

  // Convert to single precision
  for (i=0; i < A->ncatom * 3; i++)
    Coord[i]=(float) X[i];

  // write coords
  start[0]=set;
  start[1]=0;
  start[2]=0;
  count[0]=1;
  count[1]=A->ncatom;
  count[2]=3;
  if (checkNCerr(nc_put_vara_float(A->ncid,A->coordID,start,count,Coord),
    "Netcdf Writing frame %i",set)) {free(Coord); return 1;}

  // write box
  if (box!=NULL && A->cellLengthID!=-1) {
    count[1]=3;
    count[2]=0;
    if (checkNCerr(nc_put_vara_double(A->ncid,A->cellLengthID,start,count,box),
      "Writing cell lengths.")) {free(Coord); return 1;}
    if (checkNCerr(nc_put_vara_double(A->ncid,A->cellAngleID,start,count,box+3),
      "Writing cell angles.")) {free(Coord); return 1;}
  }
  
  nc_sync(A->ncid); // Necessary after every write??

  // Set the current frame
  A->currentFrame=set;

  // Free temporary memory 
  free(Coord);

  return 0;
#else
  fprintf(stdout,"Error: NAB compiled without Netcdf support.\n");
  fprintf(stdout,"       Recompile with -DBINTRAJ\n");
  return 1;
#endif
}  

/*
 * netcdfWriteNextFrame()
 * Write coords to currentFrame. Increment currentFrame.
 * Return 1 on error.
 */
int netcdfWriteNextFrame(struct AmberNetcdf *A, double *X, double *box) {
#ifdef BINTRAJ
  if ( netcdfWriteFrame(A, A->currentFrame, X, box) ) return 1;
  A->currentFrame++;
  return 0;
#else
  fprintf(stdout,"Error: NAB compiled without Netcdf support.\n");
  fprintf(stdout,"       Recompile with -DBINTRAJ\n");
  return 1;
#endif
}
  

/*
 * netcdfInfo()
 * Print general information about the netcdf file.
 */
int netcdfInfo(struct AmberNetcdf *A) {
#ifdef BINTRAJ
  if (netcdfTaskid()!=0) return 0;
  if (A==NULL) {
    fprintf(stdout,"Error: netcdfInfo: AmberNetcdf is NULL.\n");
    return 1;
  }
  fprintf(stdout,"  File is a NetCDF AMBER trajectory"
            //(p->isVelocity ? " and velocities" : "")
         );
  if (A->cellLengthID!=-1) fprintf(stdout," with Box info"); 
  if (A->TempVarID!=-1) fprintf(stdout,", with replica temperatures");
  fprintf(stdout,"\n");
  fprintf(stdout,"    %i atoms.\n",A->ncatom);
  fprintf(stdout,"    %i frames.\n",A->ncframe);

  /*if (debug > 2) {
      if (title != NULL)
        printfone("    title:        \"%s\"\n", p->title);
      if (application != NULL)  
        printfone("    application:  \"%s\"\n", p->application);
      if (program != NULL) 
        printfone("    program:      \"%s\"\n", p->program);
      if (version != NULL) 
        printfone("    version:      \"%s\"\n", p->version);
  }*/
#else
  fprintf(stdout,"Error: NAB compiled without Netcdf support.\n");
  fprintf(stdout,"       Recompile with -DBINTRAJ\n");
  return 1;
#endif
  return 0;
}
