########################################################################
#                                                                      #
# This is a module that will load amber output objects. The decomp     #
# output is parsed by functions since they can get very large. The     #
# others are small enough to load completely into memory.              #
#                                                                      #
#          Last updated: 01/27/2011                                    #
#                                                                      #
########################################################################

########################## GPL LICENSE INFO ############################

#  Copyright (C) 2010  Dwight McGee, Billy Miller III, and Jason Swails

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.

#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
   
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330,
#  Boston, MA 02111-1307, USA.

##########################################################################################################
#  List of functions and classes with a brief description of their purpose
#
#  FUNCTIONS
#  
#  CombineOutputs: Creates 'global' output files from the data stored in thread-specific output files
#                  made during the execution of MMPBSA.py.MPI. This should only be imported in the MPI
#                  version.
#
#  PrintDecompResults: Parses decomposition output files and prints results to decomp output -- wrapper for
#                      internal functions _decoutput12 and _decoutput34
#
#  _nmodeCombine: Internal -- combines nmode output files
#  _noDecSander: Internal -- Combines the output files from thread-specific sander output files into a master
#  _DecSander: Internal -- combines decomp sander output files
#  _rismCombine: Internal -- combines 3D-RISM output files
#  _nmodefile: Internal -- Searches out the frame results from an nmode output file to a new file
#  _sanderfile: Internal -- write the frame results from a sander output to a new file
#  _rismfile: Internal -- write the frame results from a rism3d.snglpnt output to a new file
#  _avg: Internal -- returns an average of an array
#  _stdev: Internal -- returns a standard deviation of an array
#  _decoutput12: Internal -- parses/prints decomposition 
#
#  CLASSES
#
#  QHout: An object that has all of the arrays/data associated with a ptraj Quasi-harmonic output file
#  NMODEout: An object that has all of the arrays/data associated with an NMODE output file
#  QMMMout: An object that has all of the arrays/data associated with a QM/MM GB mdout file
#  GBout: An object that has all of the arrays/data associated with a GB mdout file
#  PBout: An object that has all of the arrays/data associated with a PB mdout file
#  RISMout: An object that has all of the arrays/data associated with a RISM output file
#
#  The __str__ function of each class is the output associated with that object. Thus, all you have to do
#  is print the object as a string (e.g. "%s" % object or str(object)) to get its output.
##########################################################################################################

import sys, os, math
from chemistry.amber.readparm import amberParm

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

TOLERANCE = 0.001 # how large a difference can be to still be considered "0"

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def CombineOutputs(INPUT, numthreads, stability):
   """ Combine all output files from individual MPI threads if we ran MMPBSA.MPI """
   # First combine all GB mdout files if necessary: not Decomp yet
   if INPUT['gbrun'] and not INPUT['decomprun']:
      _noDecSander(INPUT, numthreads, stability, 'gb')

   # Next combine all PB mdout files if necessary: not Decomp yet
   if INPUT['pbrun'] and not INPUT['decomprun']:
      _noDecSander(INPUT, numthreads, stability, 'pb')

   # Next combine all nmode output files
   if INPUT['nmoderun']:
      _nmodeCombine(INPUT, numthreads, stability)

   # Next combine all rism output files
   if INPUT['rismrun']:
      _rismCombine(INPUT, numthreads, stability)

   # Next combine all GB decomp mdout files
   if INPUT['gbrun'] and INPUT['decomprun']:
      _DecSander(INPUT, numthreads, stability, 'gb')

   # Next combine all PB decomp mdout files
   if INPUT['pbrun'] and INPUT['decomprun']:
      _DecSander(INPUT, numthreads, stability, 'pb')
               
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def PrintDecompResults(outname, idecomp, debug, verbose, type, frames, comprm, \
                       recprm, ligprm, ligstart, surften, header, prefix):
   """ Prints the decomposition results to the decomp output file """
   comres = comprm.parm_data['RESIDUE_LABEL']

   # Are we doing a stability calculation or not?
   if not recprm.exists and not ligprm.exists: 
      stability = True
      recres = []
      ligres = []
   else:
      stability = False
      recres = recprm.parm_data['RESIDUE_LABEL']
      ligres = ligprm.parm_data['RESIDUE_LABEL']

   # open the output file
   outfile = open(outname, 'a')
   outfile.write(header)

   # per-residue parsing
   if idecomp == 1 or idecomp == 2: 
      return _decoutput12(outfile,debug,verbose,frames,comres,recres,ligres,ligstart,type,idecomp,surften,stability,prefix)

   # pairwise parsing
   elif idecomp == 3 or idecomp == 4:
      return _decoutput34(outfile,debug,verbose,frames,comres,recres,ligres,ligstart,type,idecomp,surften,outname,stability,prefix)

   # close the output file
   outfile.close()

   return -1

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _decoutput12(output, debug, verbose, frames, com_resnames, rec_resnames, lig_resnames, ligstart, type, \
                 decomp, surften, stability, prefix):
   """ Prints the output for per-residue decomp """
   frames = float(frames)

   mdout = open("%scomplex_%s.mdout" % (prefix,type),'r')
   cress = []      # Complex RESidueS -- array that will contain all printed decomp residues in complex
   numres = 0      # the NUMber RESidue that we're currently on -- used as an index for data arrays

# All arrays below are arrays of 2-element arrays. First element is a running sum for the average, second
# element is a running sum of squares for the standard deviation
   ctint = []      # Complex Total INTernal decomposition
   ctvdw = []      # Complex Total VDW decomposition
   cteel = []      # Complex Total EEL (electrostatic) decomposition
   ctpol = []      # Complex Total POLar solvation decomposition
   ctsas = []      # Complex Total non-polar solvation decomposition
   csint = []      # Complex Sidechain INTernal decomposition
   csvdw = []      # Complex Sidechain VDW decomposition
   cseel = []      # Complex Sidechain EEL (elecrostatic) decomposition
   cspol = []      # Complex Sidechain POLar solvation decomposition
   cssas = []      # Complex Sidechain non-polar solvation decomposition
   cbint = []      # Complex Backbone INTernal decomposition
   cbvdw = []      # Complex Backbone VDW decomposition
   cbeel = []      # Complex Backbone EEL decomposition
   cbpol = []      # Complex Backbone POLar solvation decomposition
   cbsas = []      # Complex Backbone non-polar solvation decomposition

   line = mdout.readline()
   # Find the first frame in output
   while line != "":
      if line[0:49] == "                    PRINT DECOMP - TOTAL ENERGIES":
         if len(ctint) > 0: # we've already found the first block
            break
         # discard the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline() # data starts here
         while line[0:3] == 'TDC':
            cress.append(int(line[4:10]))
            eint = float(line[11:20])
            vdw = float(line[21:30])
            eel = float(line[31:40])
            pol = float(line[41:50])
            sas = float(line[51:60]) * surften
            ctint.append([eint, eint * eint])
            ctvdw.append([vdw, vdw * vdw])
            cteel.append([eel, eel * eel])
            ctpol.append([pol, pol * pol])
            ctsas.append([sas, sas * sas])
            line = mdout.readline()
      if line[0:53] == "                    PRINT DECOMP - SIDECHAIN ENERGIES":
         # discard the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline() # data starts here
         while line[0:3] == 'SDC':
            eint = float(line[11:20])
            vdw = float(line[21:30])
            eel = float(line[31:40])
            pol = float(line[41:50])
            sas = float(line[51:60]) * surften
            csint.append([eint, eint * eint])
            csvdw.append([vdw, vdw * vdw])
            cseel.append([eel, eel * eel])
            cspol.append([pol, pol * pol])
            cssas.append([sas, sas * sas])
            line = mdout.readline() # get the next line
      if line[0:52] == "                    PRINT DECOMP - BACKBONE ENERGIES":
         # discard the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline() # data starts here
         while line[0:3] == 'BDC':
            eint = float(line[11:20])
            vdw = float(line[21:30])
            eel = float(line[31:40])
            pol = float(line[41:50])
            sas = float(line[51:60]) * surften
            cbint.append([eint, eint * eint])
            cbvdw.append([vdw, vdw * vdw])
            cbeel.append([eel, eel * eel])
            cbpol.append([pol, pol * pol])
            cbsas.append([sas, sas * sas])
            line = mdout.readline() # get the next line
         break # we're now done with the first frame
      line = mdout.readline()

   # now quit if we don't have enough data
   if len(csint) == 0 and (verbose == 1 or verbose == 3):
      print >> sys.stderr, 'Error: dec_verbose = %s requires sidechain and backbone data in the mdout files!' % verbose
      print >> sys.stderr, '       Re-run your calculations with the appropriate value for dec_verbose'
      return -1

   # otherwise go on: do the same as above, but no longer append
   while line != "":
      if line[0:49] == "                    PRINT DECOMP - TOTAL ENERGIES":
         # discard the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline() # data starts here
         numres = 0
         while line[0:3] == 'TDC':
            eint = float(line[11:20])
            vdw = float(line[21:30])
            eel = float(line[31:40])
            pol = float(line[41:50])
            sas = float(line[51:60]) * surften
            ctint[numres][0], ctint[numres][1] = ctint[numres][0] + eint, ctint[numres][1] + eint * eint
            ctvdw[numres][0], ctvdw[numres][1] = ctvdw[numres][0] + vdw, ctvdw[numres][1] + vdw * vdw
            cteel[numres][0], cteel[numres][1] = cteel[numres][0] + eel, cteel[numres][1] + eel * eel
            ctpol[numres][0], ctpol[numres][1] = ctpol[numres][0] + pol, ctpol[numres][1] + pol * pol
            ctsas[numres][0], ctsas[numres][1] = ctsas[numres][0] + sas, ctsas[numres][1] + sas * sas
            numres += 1
            line = mdout.readline()
      if line[0:53] == "                    PRINT DECOMP - SIDECHAIN ENERGIES":
         # discard the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline() # data starts here
         numres = 0
         while line[0:3] == 'SDC':
            eint = float(line[11:20])
            vdw = float(line[21:30])
            eel = float(line[31:40])
            pol = float(line[41:50])
            sas = float(line[51:60]) * surften
            csint[numres][0], csint[numres][1] = csint[numres][0] + eint, csint[numres][1] + eint * eint
            csvdw[numres][0], csvdw[numres][1] = csvdw[numres][0] + vdw, csvdw[numres][1] + vdw * vdw
            cseel[numres][0], cseel[numres][1] = cseel[numres][0] + eel, cseel[numres][1] + eel * eel
            cspol[numres][0], cspol[numres][1] = cspol[numres][0] + pol, cspol[numres][1] + pol * pol
            cssas[numres][0], cssas[numres][1] = cssas[numres][0] + sas, cssas[numres][1] + sas * sas
            numres += 1
            line = mdout.readline() # get the next line
      if line[0:52] == "                    PRINT DECOMP - BACKBONE ENERGIES":
         # discard the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline() # data starts here
         numres = 0
         while line[0:3] == 'BDC':
            eint = float(line[11:20])
            vdw = float(line[21:30])
            eel = float(line[31:40])
            pol = float(line[41:50])
            sas = float(line[51:60]) * surften
            cbint[numres][0], cbint[numres][1] = cbint[numres][0] + eint, cbint[numres][1] + eint * eint
            cbvdw[numres][0], cbvdw[numres][1] = cbvdw[numres][0] + vdw, cbvdw[numres][1] + vdw * vdw
            cbeel[numres][0], cbeel[numres][1] = cbeel[numres][0] + eel, cbeel[numres][1] + eel * eel
            cbpol[numres][0], cbpol[numres][1] = cbpol[numres][0] + pol, cbpol[numres][1] + pol * pol
            cbsas[numres][0], cbsas[numres][1] = cbsas[numres][0] + sas, cbsas[numres][1] + sas * sas
            numres += 1
            line = mdout.readline() 
      line = mdout.readline()

   mdout.close()

   for i in range(len(cress)): # we now compute averages and standard deviations
      ctint[i][0] /= frames
      ctint[i][1] = math.sqrt(abs(ctint[i][1]/frames - ctint[i][0] * ctint[i][0]))
      ctvdw[i][0] /= frames
      ctvdw[i][1] = math.sqrt(abs(ctvdw[i][1]/frames - ctvdw[i][0] * ctvdw[i][0]))
      cteel[i][0] /= frames
      cteel[i][1] = math.sqrt(abs(cteel[i][1]/frames - cteel[i][0] * cteel[i][0]))
      ctpol[i][0] /= frames
      ctpol[i][1] = math.sqrt(abs(ctpol[i][1]/frames - ctpol[i][0] * ctpol[i][0]))
      ctsas[i][0] /= frames
      ctsas[i][1] = math.sqrt(abs(ctsas[i][1]/frames - ctsas[i][0] * ctsas[i][0]))
      if verbose == 1 or verbose == 3:
         csint[i][0] /= frames
         csint[i][1] = math.sqrt(abs(csint[i][1]/frames - csint[i][0] * csint[i][0]))
         csvdw[i][0] /= frames
         csvdw[i][1] = math.sqrt(abs(csvdw[i][1]/frames - csvdw[i][0] * csvdw[i][0]))
         cseel[i][0] /= frames
         cseel[i][1] = math.sqrt(abs(cseel[i][1]/frames - cseel[i][0] * cseel[i][0]))
         cspol[i][0] /= frames
         cspol[i][1] = math.sqrt(abs(cspol[i][1]/frames - cspol[i][0] * cspol[i][0]))
         cssas[i][0] /= frames
         cssas[i][1] = math.sqrt(abs(cssas[i][1]/frames - cssas[i][0] * cssas[i][0]))
         cbint[i][0] /= frames
         cbint[i][1] = math.sqrt(abs(cbint[i][1]/frames - cbint[i][0] * cbint[i][0]))
         cbvdw[i][0] /= frames
         cbvdw[i][1] = math.sqrt(abs(cbvdw[i][1]/frames - cbvdw[i][0] * cbvdw[i][0]))
         cbeel[i][0] /= frames
         cbeel[i][1] = math.sqrt(abs(cbeel[i][1]/frames - cbeel[i][0] * cbeel[i][0]))
         cbpol[i][0] /= frames
         cbpol[i][1] = math.sqrt(abs(cbpol[i][1]/frames - cbpol[i][0] * cbpol[i][0]))
         cbsas[i][0] /= frames
         cbsas[i][1] = math.sqrt(abs(cbsas[i][1]/frames - cbsas[i][0] * cbsas[i][0]))

# Done with complex.  Do the same thing for the receptor, only name arrays starting with 'r'
# instead of 'c' for Receptor.
   if not stability:
      mdout = open("%sreceptor_%s.mdout" % (prefix,type),'r') 
      rress = []
      rtint = []
      rtvdw = []
      rteel = []
      rtpol = []
      rtsas = []
      rsint = []
      rsvdw = []
      rseel = []
      rspol = []
      rssas = []
      rbint = []
      rbvdw = []
      rbeel = []
      rbpol = []
      rbsas = []
   
      line = mdout.readline()
      # Find the first frame in output
      while line != "":
         if line[0:49] == "                    PRINT DECOMP - TOTAL ENERGIES":
            if len(rtint) > 0: # we've already found the first block
               break
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            while line[0:3] == 'TDC':
               rress.append(int(line[4:10]))
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               rtint.append([eint, eint * eint])
               rtvdw.append([vdw, vdw * vdw])
               rteel.append([eel, eel * eel])
               rtpol.append([pol, pol * pol])
               rtsas.append([sas, sas * sas])
               line = mdout.readline()
         if line[0:53] == "                    PRINT DECOMP - SIDECHAIN ENERGIES":
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            while line[0:3] == 'SDC':
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               rsint.append([eint, eint * eint])
               rsvdw.append([vdw, vdw * vdw])
               rseel.append([eel, eel * eel])
               rspol.append([pol, pol * pol])
               rssas.append([sas, sas * sas])
               line = mdout.readline() # get the next line
         if line[0:52] == "                    PRINT DECOMP - BACKBONE ENERGIES":
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            while line[0:3] == 'BDC':
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               rbint.append([eint, eint * eint])
               rbvdw.append([vdw, vdw * vdw])
               rbeel.append([eel, eel * eel])
               rbpol.append([pol, pol * pol])
               rbsas.append([sas, sas * sas])
               line = mdout.readline() # get the next line
            break # we're now done with the first frame
         line = mdout.readline()
   
      # otherwise go on: do the same as above, but no longer append
      while line != "":
         if line[0:49] == "                    PRINT DECOMP - TOTAL ENERGIES":
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            numres = 0
            while line[0:3] == 'TDC':
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               rtint[numres][0], rtint[numres][1] = rtint[numres][0] + eint, rtint[numres][1] + eint * eint
               rtvdw[numres][0], rtvdw[numres][1] = rtvdw[numres][0] + vdw, rtvdw[numres][1] + vdw * vdw
               rteel[numres][0], rteel[numres][1] = rteel[numres][0] + eel, rteel[numres][1] + eel * eel
               rtpol[numres][0], rtpol[numres][1] = rtpol[numres][0] + pol, rtpol[numres][1] + pol * pol
               rtsas[numres][0], rtsas[numres][1] = rtsas[numres][0] + sas, rtsas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline()
         if line[0:53] == "                    PRINT DECOMP - SIDECHAIN ENERGIES":
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            numres = 0
            while line[0:3] == 'SDC':
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               rsint[numres][0], rsint[numres][1] = rsint[numres][0] + eint, rsint[numres][1] + eint * eint
               rsvdw[numres][0], rsvdw[numres][1] = rsvdw[numres][0] + vdw, rsvdw[numres][1] + vdw * vdw
               rseel[numres][0], rseel[numres][1] = rseel[numres][0] + eel, rseel[numres][1] + eel * eel
               rspol[numres][0], rspol[numres][1] = rspol[numres][0] + pol, rspol[numres][1] + pol * pol
               rssas[numres][0], rssas[numres][1] = rssas[numres][0] + sas, rssas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline() # get the next line
         if line[0:52] == "                    PRINT DECOMP - BACKBONE ENERGIES":
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            numres = 0
            while line[0:3] == 'BDC':
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               rbint[numres][0], rbint[numres][1] = rbint[numres][0] + eint, rbint[numres][1] + eint * eint
               rbvdw[numres][0], rbvdw[numres][1] = rbvdw[numres][0] + vdw, rbvdw[numres][1] + vdw * vdw
               rbeel[numres][0], rbeel[numres][1] = rbeel[numres][0] + eel, rbeel[numres][1] + eel * eel
               rbpol[numres][0], rbpol[numres][1] = rbpol[numres][0] + pol, rbpol[numres][1] + pol * pol
               rbsas[numres][0], rbsas[numres][1] = rbsas[numres][0] + sas, rbsas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline()
         line = mdout.readline()
   
      mdout.close()

      for i in range(len(rress)):
         rtint[i][0] /= frames
         rtint[i][1] = math.sqrt(abs(rtint[i][1]/frames - rtint[i][0] * rtint[i][0]))
         rtvdw[i][0] /= frames
         rtvdw[i][1] = math.sqrt(abs(rtvdw[i][1]/frames - rtvdw[i][0] * rtvdw[i][0]))
         rteel[i][0] /= frames
         rteel[i][1] = math.sqrt(abs(rteel[i][1]/frames - rteel[i][0] * rteel[i][0]))
         rtpol[i][0] /= frames
         rtpol[i][1] = math.sqrt(abs(rtpol[i][1]/frames - rtpol[i][0] * rtpol[i][0]))
         rtsas[i][0] /= frames
         rtsas[i][1] = math.sqrt(abs(rtsas[i][1]/frames - rtsas[i][0] * rtsas[i][0]))
         if verbose == 1 or verbose == 3:
            rsint[i][0] /= frames
            rsint[i][1] = math.sqrt(abs(rsint[i][1]/frames - rsint[i][0] * rsint[i][0]))
            rsvdw[i][0] /= frames
            rsvdw[i][1] = math.sqrt(abs(rsvdw[i][1]/frames - rsvdw[i][0] * rsvdw[i][0]))
            rseel[i][0] /= frames
            rseel[i][1] = math.sqrt(abs(rseel[i][1]/frames - rseel[i][0] * rseel[i][0]))
            rspol[i][0] /= frames
            rspol[i][1] = math.sqrt(abs(rspol[i][1]/frames - rspol[i][0] * rspol[i][0]))
            rssas[i][0] /= frames
            rssas[i][1] = math.sqrt(abs(rssas[i][1]/frames - rssas[i][0] * rssas[i][0]))
            rbint[i][0] /= frames
            rbint[i][1] = math.sqrt(abs(rbint[i][1]/frames - rbint[i][0] * rbint[i][0]))
            rbvdw[i][0] /= frames
            rbvdw[i][1] = math.sqrt(abs(rbvdw[i][1]/frames - rbvdw[i][0] * rbvdw[i][0]))
            rbeel[i][0] /= frames
            rbeel[i][1] = math.sqrt(abs(rbeel[i][1]/frames - rbeel[i][0] * rbeel[i][0]))
            rbpol[i][0] /= frames
            rbpol[i][1] = math.sqrt(abs(rbpol[i][1]/frames - rbpol[i][0] * rbpol[i][0]))
            rbsas[i][0] /= frames
            rbsas[i][1] = math.sqrt(abs(rbsas[i][1]/frames - rbsas[i][0] * rbsas[i][0]))
   
   # Same as receptor above
      mdout = open("%sligand_%s.mdout" % (prefix,type),'r')
      lress = []
      ltint = []
      ltvdw = []
      lteel = []
      ltpol = []
      ltsas = []
      lsint = []
      lsvdw = []
      lseel = []
      lspol = []
      lssas = []
      lbint = []
      lbvdw = []
      lbeel = []
      lbpol = []
      lbsas = []

      line = mdout.readline()
      # Find the first frame in output
      while line != "":
         if line[0:49] == "                    PRINT DECOMP - TOTAL ENERGIES":
            if len(ltint) > 0: # we've already found the first block
               break
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            while line[0:3] == 'TDC':
               lress.append(int(line[4:10]))
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               ltint.append([eint, eint * eint])
               ltvdw.append([vdw, vdw * vdw])
               lteel.append([eel, eel * eel])
               ltpol.append([pol, pol * pol])
               ltsas.append([sas, sas * sas])
               line = mdout.readline()
         if line[0:53] == "                    PRINT DECOMP - SIDECHAIN ENERGIES":
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            while line[0:3] == 'SDC':
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               lsint.append([eint, eint * eint])
               lsvdw.append([vdw, vdw * vdw])
               lseel.append([eel, eel * eel])
               lspol.append([pol, pol * pol])
               lssas.append([sas, sas * sas])
               line = mdout.readline() # get the next line
         if line[0:52] == "                    PRINT DECOMP - BACKBONE ENERGIES":
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            while line[0:3] == 'BDC':
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               lbint.append([eint, eint * eint])
               lbvdw.append([vdw, vdw * vdw])
               lbeel.append([eel, eel * eel])
               lbpol.append([pol, pol * pol])
               lbsas.append([sas, sas * sas])
               line = mdout.readline() # get the next line
            break # we're now done with the first frame
         line = mdout.readline()
   
      # otherwise go on: do the same as above, but no longer append
      while line != "":
         if line[0:49] == "                    PRINT DECOMP - TOTAL ENERGIES":
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            numres = 0
            while line[0:3] == 'TDC':
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               ltint[numres][0], ltint[numres][1] = ltint[numres][0] + eint, ltint[numres][1] + eint * eint
               ltvdw[numres][0], ltvdw[numres][1] = ltvdw[numres][0] + vdw, ltvdw[numres][1] + vdw * vdw
               lteel[numres][0], lteel[numres][1] = lteel[numres][0] + eel, lteel[numres][1] + eel * eel
               ltpol[numres][0], ltpol[numres][1] = ltpol[numres][0] + pol, ltpol[numres][1] + pol * pol
               ltsas[numres][0], ltsas[numres][1] = ltsas[numres][0] + sas, ltsas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline()
         if line[0:53] == "                    PRINT DECOMP - SIDECHAIN ENERGIES":
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            numres = 0
            while line[0:3] == 'SDC':
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               lsint[numres][0], lsint[numres][1] = lsint[numres][0] + eint, lsint[numres][1] + eint * eint
               lsvdw[numres][0], lsvdw[numres][1] = lsvdw[numres][0] + vdw, lsvdw[numres][1] + vdw * vdw
               lseel[numres][0], lseel[numres][1] = lseel[numres][0] + eel, lseel[numres][1] + eel * eel
               lspol[numres][0], lspol[numres][1] = lspol[numres][0] + pol, lspol[numres][1] + pol * pol
               lssas[numres][0], lssas[numres][1] = lssas[numres][0] + sas, lssas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline() # get the next line
         if line[0:52] == "                    PRINT DECOMP - BACKBONE ENERGIES":
            # discard the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            numres = 0
            while line[0:3] == 'BDC':
               eint = float(line[11:20])
               vdw = float(line[21:30])
               eel = float(line[31:40])
               pol = float(line[41:50])
               sas = float(line[51:60]) * surften
               lbint[numres][0], lbint[numres][1] = lbint[numres][0] + eint, lbint[numres][1] + eint * eint
               lbvdw[numres][0], lbvdw[numres][1] = lbvdw[numres][0] + vdw, lbvdw[numres][1] + vdw * vdw
               lbeel[numres][0], lbeel[numres][1] = lbeel[numres][0] + eel, lbeel[numres][1] + eel * eel
               lbpol[numres][0], lbpol[numres][1] = lbpol[numres][0] + pol, lbpol[numres][1] + pol * pol
               lbsas[numres][0], lbsas[numres][1] = lbsas[numres][0] + sas, lbsas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline() # get the next line
         line = mdout.readline()
   
      mdout.close()
   
      for i in range(len(lress)):
         ltint[i][0] /= frames
         ltint[i][1] = math.sqrt(abs(ltint[i][1]/frames - ltint[i][0] * ltint[i][0]))
         ltvdw[i][0] /= frames
         ltvdw[i][1] = math.sqrt(abs(ltvdw[i][1]/frames - ltvdw[i][0] * ltvdw[i][0]))
         lteel[i][0] /= frames
         lteel[i][1] = math.sqrt(abs(lteel[i][1]/frames - lteel[i][0] * lteel[i][0]))
         ltpol[i][0] /= frames
         ltpol[i][1] = math.sqrt(abs(ltpol[i][1]/frames - ltpol[i][0] * ltpol[i][0]))
         ltsas[i][0] /= frames
         ltsas[i][1] = math.sqrt(abs(ltsas[i][1]/frames - ltsas[i][0] * ltsas[i][0]))
         if verbose == 1 or verbose == 3:
            lsint[i][0] /= frames
            lsint[i][1] = math.sqrt(abs(lsint[i][1]/frames - lsint[i][0] * lsint[i][0]))
            lsvdw[i][0] /= frames
            lsvdw[i][1] = math.sqrt(abs(lsvdw[i][1]/frames - lsvdw[i][0] * lsvdw[i][0]))
            lseel[i][0] /= frames
            lseel[i][1] = math.sqrt(abs(lseel[i][1]/frames - lseel[i][0] * lseel[i][0]))
            lspol[i][0] /= frames
            lspol[i][1] = math.sqrt(abs(lspol[i][1]/frames - lspol[i][0] * lspol[i][0]))
            lssas[i][0] /= frames
            lssas[i][1] = math.sqrt(abs(lssas[i][1]/frames - lssas[i][0] * lssas[i][0]))
            lbint[i][0] /= frames
            lbint[i][1] = math.sqrt(abs(lbint[i][1]/frames - lbint[i][0] * lbint[i][0]))
            lbvdw[i][0] /= frames
            lbvdw[i][1] = math.sqrt(abs(lbvdw[i][1]/frames - lbvdw[i][0] * lbvdw[i][0]))
            lbeel[i][0] /= frames
            lbeel[i][1] = math.sqrt(abs(lbeel[i][1]/frames - lbeel[i][0] * lbeel[i][0]))
            lbpol[i][0] /= frames
            lbpol[i][1] = math.sqrt(abs(lbpol[i][1]/frames - lbpol[i][0] * lbpol[i][0]))
            lbsas[i][0] /= frames
            lbsas[i][1] = math.sqrt(abs(lbsas[i][1]/frames - lbsas[i][0] * lbsas[i][0]))

# print output
   if decomp == 1:
      output.write("idecomp = 1: Decomposition per-residue adding 1-4 interactions added to Internal.\n")
   if decomp == 2:
      output.write("idecomp = 2: Decomposition per-residue adding 1-4 interactions added to EEL and VDW.\n")
   if ligstart == -1 and verbose <= 2:
      output.write("DELTAs can only be printed with the default mask assignments\n\n")
      verbose += 2
   if type == 'gb':
      output.write("Energy Decomposition Analysis (All units kcal/mol): Generalized Born solvent\n\n")
   elif type == 'pb':
      output.write("Energy Decomposition Analysis (All units kcal/mol): Poisson Boltzmann solvent\n\n")
   if verbose == 2 or verbose == 3 or stability:
      output.write("Complex:" + """
Total Energy Decomposition:
Residue |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------
""")
      for x in range(len(cress)):
         total_avg = ctint[x][0] + ctvdw[x][0] + cteel[x][0] + ctpol[x][0] + ctsas[x][0]
         total_stdev = math.sqrt(ctint[x][1]**2 + ctvdw[x][1]**2 + cteel[x][1]**2 + ctpol[x][1]**2 + ctsas[x][1]**2)
         output.write('%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), ctint[x][0], ctint[x][1], ctvdw[x][0], ctvdw[x][1], cteel[x][0],
                 cteel[x][1], ctpol[x][0], ctpol[x][1], ctsas[x][0], ctsas[x][1], total_avg, total_stdev))

   if verbose == 3 or (stability and verbose == 1):
      output.write("""
Sidechain Energy Decomposition:
Residue |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------
""")

      for x in range(len(cress)):
         total_avg = csint[x][0] + csvdw[x][0] + cseel[x][0] + cspol[x][0] + cssas[x][0]
         total_stdev = math.sqrt(csint[x][1]**2 + csvdw[x][1]**2 + cseel[x][1]**2 + cspol[x][1]**2 + cssas[x][1]**2)
         output.write('%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), csint[x][0], csint[x][1], csvdw[x][0], csvdw[x][1], cseel[x][0],
                 cseel[x][1], cspol[x][0], cspol[x][1], cssas[x][0], cssas[x][1], total_avg, total_stdev))

      output.write("""
Backbone Energy Decomposition:
Residue |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------
""")
      for x in range(len(cress)):
         total_avg = cbint[x][0] + cbvdw[x][0] + cbeel[x][0] + cbpol[x][0] + cbsas[x][0]
         total_stdev = math.sqrt(cbint[x][1]**2 + cbvdw[x][1]**2 + cbeel[x][1]**2 + cbpol[x][1]**2 + cbsas[x][1]**2)
         output.write('%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), cbint[x][0], cbint[x][1], cbvdw[x][0], cbvdw[x][1], cbeel[x][0],
                 cbeel[x][1], cbpol[x][0], cbpol[x][1], cbsas[x][0], cbsas[x][1], total_avg, total_stdev))

   if (verbose == 2 or verbose == 3) and not stability:
      output.write("\nReceptor:" + """
Total Energy Decomposition:
Residue |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------
""")
      for x in range(len(rress)):
         total_avg = rtint[x][0] + rtvdw[x][0] + rteel[x][0] + rtpol[x][0] + rtsas[x][0]
         total_stdev = math.sqrt(rtint[x][1]**2 + rtvdw[x][1]**2 + rteel[x][1]**2 + rtpol[x][1]**2 + rtsas[x][1]**2)
         output.write('%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                (rec_resnames[rress[x]-1] + str(rress[x]).rjust(4), rtint[x][0], rtint[x][1], rtvdw[x][0], rtvdw[x][1], rteel[x][0],
                 rteel[x][1], rtpol[x][0], rtpol[x][1], rtsas[x][0], rtsas[x][1], total_avg, total_stdev))

   if verbose == 3 and not stability:
      output.write("""
Sidechain Energy Decomposition:
Residue |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------
""")

      for x in range(len(rress)):
         total_avg = rsint[x][0] + rsvdw[x][0] + rseel[x][0] + rspol[x][0] + rssas[x][0]
         total_stdev = math.sqrt(rsint[x][1]**2 + rsvdw[x][1]**2 + rseel[x][1]**2 + rspol[x][1]**2 + rssas[x][1]**2)
         output.write('%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                (rec_resnames[rress[x]-1] + str(rress[x]).rjust(4), rsint[x][0], rsint[x][1], rsvdw[x][0], rsvdw[x][1], rseel[x][0],
                 rseel[x][1], rspol[x][0], rspol[x][1], rssas[x][0], rssas[x][1], total_avg, total_stdev))

      output.write("""
Backbone Energy Decomposition:
Residue |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------
""")

      for x in range(len(rress)):
         total_avg = rbint[x][0] + rbvdw[x][0] + rbeel[x][0] + rbpol[x][0] + rbsas[x][0]
         total_stdev = math.sqrt(rbint[x][1]**2 + rbvdw[x][1]**2 + rbeel[x][1]**2 + rbpol[x][1]**2 + rbsas[x][1]**2)
         output.write('%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                (rec_resnames[rress[x]-1] + str(rress[x]).rjust(4), rbint[x][0], rbint[x][1], rbvdw[x][0], rbvdw[x][1], rbeel[x][0],
                 rbeel[x][1], rbpol[x][0], rbpol[x][1], rbsas[x][0], rbsas[x][1], total_avg, total_stdev))

   if (verbose == 2 or verbose == 3) and not stability:
      output.write("\nLigand:" + """
Total Energy Decomposition:
Residue |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------
""")
      for x in range(len(lress)):
         total_avg = ltint[x][0] + ltvdw[x][0] + lteel[x][0] + ltpol[x][0] + ltsas[x][0]
         total_stdev = math.sqrt(ltint[x][1]**2 + ltvdw[x][1]**2 + lteel[x][1]**2 + ltpol[x][1]**2 + ltsas[x][1]**2)
         output.write('%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                (lig_resnames[lress[x]-1] + str(lress[x]).rjust(4), ltint[x][0], ltint[x][1], ltvdw[x][0], ltvdw[x][1], lteel[x][0],
                 lteel[x][1], ltpol[x][0], ltpol[x][1], ltsas[x][0], ltsas[x][1], total_avg, total_stdev))

   if verbose == 3 and not stability:
      output.write("""
Sidechain Energy Decomposition:
Residue |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------
""")

      for x in range(len(lress)):
         total_avg = lsint[x][0] + lsvdw[x][0] + lseel[x][0] + lspol[x][0] + lssas[x][0]
         total_stdev = math.sqrt(lsint[x][1]**2 + lsvdw[x][1]**2 + lseel[x][1]**2 + lspol[x][1]**2 + lssas[x][1]**2)
         output.write('%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                (lig_resnames[lress[x]-1] + str(lress[x]).rjust(4), lsint[x][0], lsint[x][1], lsvdw[x][0], lsvdw[x][1], lseel[x][0],
                 lseel[x][1], lspol[x][0], lspol[x][1], lssas[x][0], lssas[x][1], total_avg, total_stdev))

      output.write("""
Backbone Energy Decomposition:
Residue |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------
""")

      for x in range(len(lress)):
         total_avg = lbint[x][0] + lbvdw[x][0] + lbeel[x][0] + lbpol[x][0] + lbsas[x][0]
         total_stdev = math.sqrt(lbint[x][1]**2 + lbvdw[x][1]**2 + lbeel[x][1]**2 + lbpol[x][1]**2 + lbsas[x][1]**2)
         output.write('%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                (lig_resnames[lress[x]-1] + str(lress[x]).rjust(4), lbint[x][0], lbint[x][1], lbvdw[x][0], lbvdw[x][1], lbeel[x][0],
                 lbeel[x][1], lbpol[x][0], lbpol[x][1], lbsas[x][0], lbsas[x][1], total_avg, total_stdev))

# Time for the fun part: DELTAs.
   if ligstart != -1 and not stability:

      dtint = [] # Delta Total INTernal decomp
      dtvdw = [] # Delta Total VDW decomp
      dteel = [] # Delta Total EEL decomp
      dtpol = [] # Delta Total POLar solvation decomp
      dtsas = [] # Delta Total non-polar solvation decomp
      dsint = [] # Delta Sidechain INTernal decomp
      dsvdw = [] # Delta Sidechain VDW decomp
      dseel = [] # Delta Sidechain EEL decomp
      dspol = [] # Delta Sidechain POLar solvation decomp
      dssas = [] # Delta Sidechain non-polar solvation decomp
      dbint = [] # Delta Backbone INTernal decomp
      dbvdw = [] # Delta Backbone VDW decomp
      dbeel = [] # Delta Backbone EEL decomp
      dbpol = [] # Delta Backbone POLar solvation decomp
      dbsas = [] # Delta Backbone non-polar solvation decomp
      loc = []   # Array that holds strings for locations of residues in ligand/receptor
      recdone = 0 # how many receptor residues have been used in differences -- used as an index
      ligdone = 0 # same as above, but for ligand

      for x in range(len(cress)): # accumulate the deltas

         # Figure out location of the complex residue
         if cress[x] < ligstart or cress[x] >= ligstart + len(lig_resnames):
            loc.append(' R ' + rec_resnames[rress[recdone]-1] + ' ' + str(rress[recdone]).rjust(3))
            dtint.append([ctint[x][0] - rtint[recdone][0], math.sqrt(ctint[x][1]**2+rtint[recdone][1]**2)])
            dtvdw.append([ctvdw[x][0] - rtvdw[recdone][0], math.sqrt(ctvdw[x][1]**2+rtvdw[recdone][1]**2)])
            dteel.append([cteel[x][0] - rteel[recdone][0], math.sqrt(cteel[x][1]**2+rteel[recdone][1]**2)])
            dtpol.append([ctpol[x][0] - rtpol[recdone][0], math.sqrt(ctpol[x][1]**2+rtpol[recdone][1]**2)])
            dtsas.append([ctsas[x][0] - rtsas[recdone][0], math.sqrt(ctsas[x][1]**2+rtsas[recdone][1]**2)])
            if verbose == 1 or verbose == 3:
               dsint.append([csint[x][0] - rsint[recdone][0], math.sqrt(csint[x][1]**2+rsint[recdone][1]**2)])
               dsvdw.append([csvdw[x][0] - rsvdw[recdone][0], math.sqrt(csvdw[x][1]**2+rsvdw[recdone][1]**2)])
               dseel.append([cseel[x][0] - rseel[recdone][0], math.sqrt(cseel[x][1]**2+rseel[recdone][1]**2)])
               dspol.append([cspol[x][0] - rspol[recdone][0], math.sqrt(cspol[x][1]**2+rspol[recdone][1]**2)])
               dssas.append([cssas[x][0] - rssas[recdone][0], math.sqrt(cssas[x][1]**2+rssas[recdone][1]**2)])
               dbint.append([cbint[x][0] - rbint[recdone][0], math.sqrt(cbint[x][1]**2+rbint[recdone][1]**2)])
               dbvdw.append([cbvdw[x][0] - rbvdw[recdone][0], math.sqrt(cbvdw[x][1]**2+rbvdw[recdone][1]**2)])
               dbeel.append([cbeel[x][0] - rbeel[recdone][0], math.sqrt(cbeel[x][1]**2+rbeel[recdone][1]**2)])
               dbpol.append([cbpol[x][0] - rbpol[recdone][0], math.sqrt(cbpol[x][1]**2+rbpol[recdone][1]**2)])
               dbsas.append([cbsas[x][0] - rbsas[recdone][0], math.sqrt(cbsas[x][1]**2+rbsas[recdone][1]**2)])
            recdone += 1
#        elif cress[x] >= ligstart and cress[x] < ligstart + len(lig_resnames): # if it's in ligand
         else:
            loc.append(' L ' + lig_resnames[lress[ligdone]-1] + ' ' + str(lress[ligdone]).rjust(3))
            dtint.append([ctint[x][0] - ltint[ligdone][0], math.sqrt(ctint[x][1]**2+ltint[ligdone][1]**2)])
            dtvdw.append([ctvdw[x][0] - ltvdw[ligdone][0], math.sqrt(ctvdw[x][1]**2+ltvdw[ligdone][1]**2)])
            dteel.append([cteel[x][0] - lteel[ligdone][0], math.sqrt(cteel[x][1]**2+lteel[ligdone][1]**2)])
            dtpol.append([ctpol[x][0] - ltpol[ligdone][0], math.sqrt(ctpol[x][1]**2+ltpol[ligdone][1]**2)])
            dtsas.append([ctsas[x][0] - ltsas[ligdone][0], math.sqrt(ctsas[x][1]**2+ltsas[ligdone][1]**2)])
            if verbose == 1 or verbose == 3:
               dsint.append([csint[x][0] - lsint[ligdone][0], math.sqrt(csint[x][1]**2+lsint[ligdone][1]**2)])
               dsvdw.append([csvdw[x][0] - lsvdw[ligdone][0], math.sqrt(csvdw[x][1]**2+lsvdw[ligdone][1]**2)])
               dseel.append([cseel[x][0] - lseel[ligdone][0], math.sqrt(cseel[x][1]**2+lseel[ligdone][1]**2)])
               dspol.append([cspol[x][0] - lspol[ligdone][0], math.sqrt(cspol[x][1]**2+lspol[ligdone][1]**2)])
               dssas.append([cssas[x][0] - lssas[ligdone][0], math.sqrt(cssas[x][1]**2+lssas[ligdone][1]**2)])
               dbint.append([cbint[x][0] - lbint[ligdone][0], math.sqrt(cbint[x][1]**2+lbint[ligdone][1]**2)])
               dbvdw.append([cbvdw[x][0] - lbvdw[ligdone][0], math.sqrt(cbvdw[x][1]**2+lbvdw[ligdone][1]**2)])
               dbeel.append([cbeel[x][0] - lbeel[ligdone][0], math.sqrt(cbeel[x][1]**2+lbeel[ligdone][1]**2)])
               dbpol.append([cbpol[x][0] - lbpol[ligdone][0], math.sqrt(cbpol[x][1]**2+lbpol[ligdone][1]**2)])
               dbsas.append([cbsas[x][0] - lbsas[ligdone][0], math.sqrt(cbsas[x][1]**2+lbsas[ligdone][1]**2)])
            ligdone += 1

      if verbose == 0 or verbose == 1 or verbose == 2 or verbose == 3:
         output.write("""
DELTAS:
Total Energy Decomposition:
Residue |  Location |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------------------
""")
         for x in range(len(cress)):
            total_avg = dtint[x][0] + dtvdw[x][0] + dteel[x][0] + dtpol[x][0] + dtsas[x][0]
            total_stdev = math.sqrt(dtint[x][1]**2 + dtvdw[x][1]**2 + dteel[x][1]**2 + dtpol[x][1]**2 + dtsas[x][1]**2)
            output.write('%s |%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                   (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), loc[x], dtint[x][0], dtint[x][1], dtvdw[x][0], dtvdw[x][1], dteel[x][0],
                    dteel[x][1], dtpol[x][0], dtpol[x][1], dtsas[x][0], dtsas[x][1], total_avg, total_stdev))

      if verbose == 1 or verbose == 3:
         output.write("""
Sidechain Energy Decomposition:
Residue |  Location |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------------------
""")
         for x in range(len(cress)):
            total_avg = dsint[x][0] + dsvdw[x][0] + dseel[x][0] + dspol[x][0] + dssas[x][0]
            total_stdev = math.sqrt(dsint[x][1]**2 + dsvdw[x][1]**2 + dseel[x][1]**2 + dspol[x][1]**2 + dssas[x][1]**2)
            output.write('%s |%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                   (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), loc[x], dsint[x][0], dsint[x][1], dsvdw[x][0], dsvdw[x][1], dseel[x][0],
                    dseel[x][1], dspol[x][0], dspol[x][1], dssas[x][0], dssas[x][1], total_avg, total_stdev))

         output.write("""
Backbone Energy Decomposition:
Residue |  Location |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-------------------------------------------------------------------------------------------------------------------------------------------------------
""")
         for x in range(len(cress)):
            total_avg = dbint[x][0] + dbvdw[x][0] + dbeel[x][0] + dbpol[x][0] + dbsas[x][0]
            total_stdev = math.sqrt(dbint[x][1]**2 + dbvdw[x][1]**2 + dbeel[x][1]**2 + dbpol[x][1]**2 + dbsas[x][1]**2)
            output.write('%s |%s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' % 
                   (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), loc[x], dbint[x][0], dbint[x][1], dbvdw[x][0], dbvdw[x][1], dbeel[x][0],
                    dbeel[x][1], dbpol[x][0], dbpol[x][1], dbsas[x][0], dbsas[x][1], total_avg, total_stdev))

   output.write('\n\n')
   output.close()
   return 0

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _decoutput34(output, debug, verbose, frames, com_resnames, rec_resnames, lig_resnames, 
                 ligstart, type, decomp, surften, outname, stability, prefix):
   """ Prints the output for pairwise decomp """
   frames = float(frames)

   mdout = open("%scomplex_%s.mdout" % (prefix,type),'r')
   numres = 0       # which NUMber RESidue we're currently analyzing -- pointer into data arrays
   cress = []       # Complex RESidueS -- array that will contain all printed decomp residues in complex
   cress2 = []      # Complex RESidueS 2 -- array that will contain all printed decomp residues in complex
                    #                       that Complex RESidueS interacts with

# set up arrays and variables for the deltas, but only if we know where the ligand is
   if ligstart != -1 and not stability:

      dtint = [] # Delta Total INTernal decomp
      dtvdw = [] # Delta Total VDW decomp
      dteel = [] # Delta Total EEL decomp
      dtpol = [] # Delta Total POLar solvation decomp
      dtsas = [] # Delta Total non-polar solvation decomp
      dsint = [] # Delta Sidechain INTernal decomp
      dsvdw = [] # Delta Sidechain VDW decomp
      dseel = [] # Delta Sidechain EEL decomp
      dspol = [] # Delta Sidechain POLar solvation decomp
      dssas = [] # Delta Sidechain non-polar solvation decomp
      dbint = [] # Delta Backbone INTernal decomp
      dbvdw = [] # Delta Backbone VDW decomp
      dbeel = [] # Delta Backbone EEL decomp
      dbpol = [] # Delta Backbone POLar solvation decomp
      dbsas = [] # Delta Backbone non-polar solvation decomp
      recdone = 0 # how many receptor residues have been used in differences -- used as an index
      ligdone = 0 # same as above, but for ligand

# All arrays below are arrays of 2-element arrays. First element is a running sum for the average, second
# element is a running sum of squares for the standard deviation
   ctint = []      # Complex Total INTernal decomposition
   ctvdw = []      # Complex Total VDW decomposition
   cteel = []      # Complex Total EEL (electrostatic) decomposition
   ctpol = []      # Complex Total POLar solvation decomposition
   ctsas = []      # Complex Total non-polar solvation decomposition
   csint = []      # Complex Sidechain INTernal decomposition
   csvdw = []      # Complex Sidechain VDW decomposition
   cseel = []      # Complex Sidechain EEL (elecrostatic) decomposition
   cspol = []      # Complex Sidechain POLar solvation decomposition
   cssas = []      # Complex Sidechain non-polar solvation decomposition
   cbint = []      # Complex Backbone INTernal decomposition
   cbvdw = []      # Complex Backbone VDW decomposition
   cbeel = []      # Complex Backbone EEL decomposition
   cbpol = []      # Complex Backbone POLar solvation decomposition
   cbsas = []      # Complex Backbone non-polar solvation decomposition

   line = mdout.readline()

   while line != '':
      if line[0:54] == "                    PRINT PAIR DECOMP - TOTAL ENERGIES":
         if len(ctint) > 0: # if we've found the first frame
            break
         # eat the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline() # data starts here
         while line[0:3] == 'TDC':
            cress.append(int(line[4:11]))
            cress2.append(int(line[13:20]))
            eint = float(line[21:30])
            vdw = float(line[31:40])
            eel = float(line[41:50])
            pol = float(line[51:60])
            sas = float(line[61:70]) * surften
            ctint.append([eint, eint * eint])
            ctvdw.append([vdw, vdw * vdw])
            cteel.append([eel, eel * eel])
            ctpol.append([pol, pol * pol])
            ctsas.append([sas, sas * sas])
            line = mdout.readline()
      if line[0:58] == "                    PRINT PAIR DECOMP - SIDECHAIN ENERGIES":
         # eat the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline() # data starts here
         while line[0:3] == 'SDC':
            eint = float(line[21:30])
            vdw = float(line[31:40])
            eel = float(line[41:50])
            pol = float(line[51:60])
            sas = float(line[61:70]) * surften
            csint.append([eint, eint * eint])
            csvdw.append([vdw, vdw * vdw])
            cseel.append([eel, eel * eel])
            cspol.append([pol, pol * pol])
            cssas.append([sas, sas * sas])
            line = mdout.readline()
      if line[0:57] == "                    PRINT PAIR DECOMP - BACKBONE ENERGIES":
         # eat the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline()
         while line[0:3] == 'BDC':
            eint = float(line[21:30])
            vdw = float(line[31:40])
            eel = float(line[41:50])
            pol = float(line[51:60])
            sas = float(line[61:70]) * surften
            cbint.append([eint, eint * eint])
            cbvdw.append([vdw, vdw * vdw])
            cbeel.append([eel, eel * eel])
            cbpol.append([pol, pol * pol])
            cbsas.append([sas, sas * sas])
            line = mdout.readline()
         break # we are now at the end of the first frame
      line = mdout.readline()
   
   # now quit if we don't have enough data
   if len(csint) == 0 and (verbose == 1 or verbose == 3):
      print >> sys.stderr, 'Error: dec_verbose = %s requires sidechain and backbone data in the mdout files!' % verbose
      print >> sys.stderr, '       Re-run your calculations with the appropriate value for dec_verbose'
      return -1

   # otherwise we go on, but don't append anymore, add on
   while line != '':
      if line[0:54] == "                    PRINT PAIR DECOMP - TOTAL ENERGIES":
         # eat the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline() # data starts here
         numres = 0
         while line[0:3] == 'TDC':
            eint = float(line[21:30])
            vdw = float(line[31:40])
            eel = float(line[41:50])
            pol = float(line[51:60])
            sas = float(line[61:70]) * surften
            ctint[numres][0], ctint[numres][1] = ctint[numres][0] + eint, ctint[numres][1] + eint * eint
            ctvdw[numres][0], ctvdw[numres][1] = ctvdw[numres][0] + vdw, ctvdw[numres][1] + vdw * vdw
            cteel[numres][0], cteel[numres][1] = cteel[numres][0] + eel, cteel[numres][1] + eel * eel
            ctpol[numres][0], ctpol[numres][1] = ctpol[numres][0] + pol, ctpol[numres][1] + pol * pol
            ctsas[numres][0], ctsas[numres][1] = ctsas[numres][0] + sas, ctsas[numres][1] + sas * sas
            numres += 1
            line = mdout.readline()
      if line[0:58] == "                    PRINT PAIR DECOMP - SIDECHAIN ENERGIES":
         # eat the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline() # data starts here
         numres = 0
         while line[0:3] == 'SDC':
            eint = float(line[21:30])
            vdw = float(line[31:40])
            eel = float(line[41:50])
            pol = float(line[51:60])
            sas = float(line[61:70]) * surften
            csint[numres][0], csint[numres][1] = csint[numres][0] + eint, csint[numres][1] + eint * eint
            csvdw[numres][0], csvdw[numres][1] = csvdw[numres][0] + vdw, csvdw[numres][1] + vdw * vdw
            cseel[numres][0], cseel[numres][1] = cseel[numres][0] + eel, cseel[numres][1] + eel * eel
            cspol[numres][0], cspol[numres][1] = cspol[numres][0] + pol, cspol[numres][1] + pol * pol
            cssas[numres][0], cssas[numres][1] = cssas[numres][0] + sas, cssas[numres][1] + sas * sas
            numres += 1
            line = mdout.readline()
      if line[0:57] == "                    PRINT PAIR DECOMP - BACKBONE ENERGIES":
         # eat the next 3 lines
         mdout.readline()
         mdout.readline()
         mdout.readline()
         line = mdout.readline()
         numres = 0
         while line[0:3] == 'BDC':
            eint = float(line[21:30])
            vdw = float(line[31:40])
            eel = float(line[41:50])
            pol = float(line[51:60])
            sas = float(line[61:70]) * surften
            cbint[numres][0], cbint[numres][1] = cbint[numres][0] + eint, cbint[numres][1] + eint * eint
            cbvdw[numres][0], cbvdw[numres][1] = cbvdw[numres][0] + vdw, cbvdw[numres][1] + vdw * vdw
            cbeel[numres][0], cbeel[numres][1] = cbeel[numres][0] + eel, cbeel[numres][1] + eel * eel
            cbpol[numres][0], cbpol[numres][1] = cbpol[numres][0] + pol, cbpol[numres][1] + pol * pol
            cbsas[numres][0], cbsas[numres][1] = cbsas[numres][0] + sas, cbsas[numres][1] + sas * sas
            numres += 1
            line = mdout.readline()
      line = mdout.readline() 
            
   for i in range(len(cress)): # we now compute averages and standard deviations
      ctint[i][0] /= frames
      ctint[i][1] = math.sqrt(abs(ctint[i][1]/frames - ctint[i][0] * ctint[i][0]))
      ctvdw[i][0] /= frames
      ctvdw[i][1] = math.sqrt(abs(ctvdw[i][1]/frames - ctvdw[i][0] * ctvdw[i][0]))
      cteel[i][0] /= frames
      cteel[i][1] = math.sqrt(abs(cteel[i][1]/frames - cteel[i][0] * cteel[i][0]))
      ctpol[i][0] /= frames
      ctpol[i][1] = math.sqrt(abs(ctpol[i][1]/frames - ctpol[i][0] * ctpol[i][0]))
      ctsas[i][0] /= frames
      ctsas[i][1] = math.sqrt(abs(ctsas[i][1]/frames - ctsas[i][0] * ctsas[i][0]))
      if verbose == 1 or verbose == 3:
         csint[i][0] /= frames
         csint[i][1] = math.sqrt(abs(csint[i][1]/frames - csint[i][0] * csint[i][0]))
         csvdw[i][0] /= frames
         csvdw[i][1] = math.sqrt(abs(csvdw[i][1]/frames - csvdw[i][0] * csvdw[i][0]))
         cseel[i][0] /= frames
         cseel[i][1] = math.sqrt(abs(cseel[i][1]/frames - cseel[i][0] * cseel[i][0]))
         cspol[i][0] /= frames
         cspol[i][1] = math.sqrt(abs(cspol[i][1]/frames - cspol[i][0] * cspol[i][0]))
         cssas[i][0] /= frames
         cssas[i][1] = math.sqrt(abs(cssas[i][1]/frames - cssas[i][0] * cssas[i][0]))
         cbint[i][0] /= frames
         cbint[i][1] = math.sqrt(abs(cbint[i][1]/frames - cbint[i][0] * cbint[i][0]))
         cbvdw[i][0] /= frames
         cbvdw[i][1] = math.sqrt(abs(cbvdw[i][1]/frames - cbvdw[i][0] * cbvdw[i][0]))
         cbeel[i][0] /= frames
         cbeel[i][1] = math.sqrt(abs(cbeel[i][1]/frames - cbeel[i][0] * cbeel[i][0]))
         cbpol[i][0] /= frames
         cbpol[i][1] = math.sqrt(abs(cbpol[i][1]/frames - cbpol[i][0] * cbpol[i][0]))
         cbsas[i][0] /= frames
         cbsas[i][1] = math.sqrt(abs(cbsas[i][1]/frames - cbsas[i][0] * cbsas[i][0]))

# Done with complex.  Do the same thing for the receptor, only name arrays starting with 'r'
# instead of 'c' for Receptor.
   if not stability:
      mdout = open("%sreceptor_%s.mdout" % (prefix,type),'r') 
      rress = []
      rress2 = []
      rtint = []
      rtvdw = []
      rteel = []
      rtpol = []
      rtsas = []
      rsint = []
      rsvdw = []
      rseel = []
      rspol = []
      rssas = []
      rbint = []
      rbvdw = []
      rbeel = []
      rbpol = []
      rbsas = []

      line = mdout.readline()
   
      while line != '':
         if line[0:54] == "                    PRINT PAIR DECOMP - TOTAL ENERGIES":
            if len(rtint) > 0: # if we've found the first frame
               break
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            while line[0:3] == 'TDC':
               rress.append(int(line[4:11]))
               rress2.append(int(line[13:20]))
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               rtint.append([eint, eint * eint])
               rtvdw.append([vdw, vdw * vdw])
               rteel.append([eel, eel * eel])
               rtpol.append([pol, pol * pol])
               rtsas.append([sas, sas * sas])
               line = mdout.readline()
         if line[0:58] == "                    PRINT PAIR DECOMP - SIDECHAIN ENERGIES":
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            while line[0:3] == 'SDC':
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               rsint.append([eint, eint * eint])
               rsvdw.append([vdw, vdw * vdw])
               rseel.append([eel, eel * eel])
               rspol.append([pol, pol * pol])
               rssas.append([sas, sas * sas])
               line = mdout.readline()
         if line[0:57] == "                    PRINT PAIR DECOMP - BACKBONE ENERGIES":
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline()
            while line[0:3] == 'BDC':
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               rbint.append([eint, eint * eint])
               rbvdw.append([vdw, vdw * vdw])
               rbeel.append([eel, eel * eel])
               rbpol.append([pol, pol * pol])
               rbsas.append([sas, sas * sas])
               line = mdout.readline()
            break # we are now at the end of the first frame
         line = mdout.readline()
      
      # Go on, but don't append anymore, add on
      while line != '':
         if line[0:54] == "                    PRINT PAIR DECOMP - TOTAL ENERGIES":
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            numres = 0
            while line[0:3] == 'TDC':
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               rtint[numres][0], rtint[numres][1] = rtint[numres][0] + eint, rtint[numres][1] + eint * eint
               rtvdw[numres][0], rtvdw[numres][1] = rtvdw[numres][0] + vdw, rtvdw[numres][1] + vdw * vdw
               rteel[numres][0], rteel[numres][1] = rteel[numres][0] + eel, rteel[numres][1] + eel * eel
               rtpol[numres][0], rtpol[numres][1] = rtpol[numres][0] + pol, rtpol[numres][1] + pol * pol
               rtsas[numres][0], rtsas[numres][1] = rtsas[numres][0] + sas, rtsas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline()
         if line[0:58] == "                    PRINT PAIR DECOMP - SIDECHAIN ENERGIES":
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            numres = 0
            while line[0:3] == 'SDC':
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               rsint[numres][0], rsint[numres][1] = rsint[numres][0] + eint, rsint[numres][1] + eint * eint
               rsvdw[numres][0], rsvdw[numres][1] = rsvdw[numres][0] + vdw, rsvdw[numres][1] + vdw * vdw
               rseel[numres][0], rseel[numres][1] = rseel[numres][0] + eel, rseel[numres][1] + eel * eel
               rspol[numres][0], rspol[numres][1] = rspol[numres][0] + pol, rspol[numres][1] + pol * pol
               rssas[numres][0], rssas[numres][1] = rssas[numres][0] + sas, rssas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline()
         if line[0:57] == "                    PRINT PAIR DECOMP - BACKBONE ENERGIES":
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline()
            numres = 0
            while line[0:3] == 'BDC':
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               rbint[numres][0], rbint[numres][1] = rbint[numres][0] + eint, rbint[numres][1] + eint * eint
               rbvdw[numres][0], rbvdw[numres][1] = rbvdw[numres][0] + vdw, rbvdw[numres][1] + vdw * vdw
               rbeel[numres][0], rbeel[numres][1] = rbeel[numres][0] + eel, rbeel[numres][1] + eel * eel
               rbpol[numres][0], rbpol[numres][1] = rbpol[numres][0] + pol, rbpol[numres][1] + pol * pol
               rbsas[numres][0], rbsas[numres][1] = rbsas[numres][0] + sas, rbsas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline()
         line = mdout.readline() 

      mdout.close()
   
      for i in range(len(rress)):
         rtint[i][0] /= frames
         rtint[i][1] = math.sqrt(abs(rtint[i][1]/frames - rtint[i][0] * rtint[i][0]))
         rtvdw[i][0] /= frames
         rtvdw[i][1] = math.sqrt(abs(rtvdw[i][1]/frames - rtvdw[i][0] * rtvdw[i][0]))
         rteel[i][0] /= frames
         rteel[i][1] = math.sqrt(abs(rteel[i][1]/frames - rteel[i][0] * rteel[i][0]))
         rtpol[i][0] /= frames
         rtpol[i][1] = math.sqrt(abs(rtpol[i][1]/frames - rtpol[i][0] * rtpol[i][0]))
         rtsas[i][0] /= frames
         rtsas[i][1] = math.sqrt(abs(rtsas[i][1]/frames - rtsas[i][0] * rtsas[i][0]))
         if verbose == 1 or verbose == 3:
            rsint[i][0] /= frames
            rsint[i][1] = math.sqrt(abs(rsint[i][1]/frames - rsint[i][0] * rsint[i][0]))
            rsvdw[i][0] /= frames
            rsvdw[i][1] = math.sqrt(abs(rsvdw[i][1]/frames - rsvdw[i][0] * rsvdw[i][0]))
            rseel[i][0] /= frames
            rseel[i][1] = math.sqrt(abs(rseel[i][1]/frames - rseel[i][0] * rseel[i][0]))
            rspol[i][0] /= frames
            rspol[i][1] = math.sqrt(abs(rspol[i][1]/frames - rspol[i][0] * rspol[i][0]))
            rssas[i][0] /= frames
            rssas[i][1] = math.sqrt(abs(rssas[i][1]/frames - rssas[i][0] * rssas[i][0]))
            rbint[i][0] /= frames
            rbint[i][1] = math.sqrt(abs(rbint[i][1]/frames - rbint[i][0] * rbint[i][0]))
            rbvdw[i][0] /= frames
            rbvdw[i][1] = math.sqrt(abs(rbvdw[i][1]/frames - rbvdw[i][0] * rbvdw[i][0]))
            rbeel[i][0] /= frames
            rbeel[i][1] = math.sqrt(abs(rbeel[i][1]/frames - rbeel[i][0] * rbeel[i][0]))
            rbpol[i][0] /= frames
            rbpol[i][1] = math.sqrt(abs(rbpol[i][1]/frames - rbpol[i][0] * rbpol[i][0]))
            rbsas[i][0] /= frames
            rbsas[i][1] = math.sqrt(abs(rbsas[i][1]/frames - rbsas[i][0] * rbsas[i][0]))
   
   # Same as receptor above
      mdout = open("%sligand_%s.mdout" % (prefix,type),'r')
      lress = []
      lress2 = []
      ltint = []
      ltvdw = []
      lteel = []
      ltpol = []
      ltsas = []
      lsint = []
      lsvdw = []
      lseel = []
      lspol = []
      lssas = []
      lbint = []
      lbvdw = []
      lbeel = []
      lbpol = []
      lbsas = []

      line = mdout.readline()
   
      while line != '':
         if line[0:54] == "                    PRINT PAIR DECOMP - TOTAL ENERGIES":
            if len(ltint) > 0: # if we've found the first frame
               break
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            while line[0:3] == 'TDC':
               lress.append(int(line[4:11]))
               lress2.append(int(line[13:20]))
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               ltint.append([eint, eint * eint])
               ltvdw.append([vdw, vdw * vdw])
               lteel.append([eel, eel * eel])
               ltpol.append([pol, pol * pol])
               ltsas.append([sas, sas * sas])
               line = mdout.readline()
         if line[0:58] == "                    PRINT PAIR DECOMP - SIDECHAIN ENERGIES":
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            while line[0:3] == 'SDC':
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               lsint.append([eint, eint * eint])
               lsvdw.append([vdw, vdw * vdw])
               lseel.append([eel, eel * eel])
               lspol.append([pol, pol * pol])
               lssas.append([sas, sas * sas])
               line = mdout.readline()
         if line[0:57] == "                    PRINT PAIR DECOMP - BACKBONE ENERGIES":
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline()
            while line[0:3] == 'BDC':
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               lbint.append([eint, eint * eint])
               lbvdw.append([vdw, vdw * vdw])
               lbeel.append([eel, eel * eel])
               lbpol.append([pol, pol * pol])
               lbsas.append([sas, sas * sas])
               line = mdout.readline()
            break # we are now at the end of the first frame
         line = mdout.readline()
      
      # Go on, but don't append anymore, add on
      while line != '':
         if line[0:54] == "                    PRINT PAIR DECOMP - TOTAL ENERGIES":
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            numres = 0
            while line[0:3] == 'TDC':
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               ltint[numres][0], ltint[numres][1] = ltint[numres][0] + eint, ltint[numres][1] + eint * eint
               ltvdw[numres][0], ltvdw[numres][1] = ltvdw[numres][0] + vdw, ltvdw[numres][1] + vdw * vdw
               lteel[numres][0], lteel[numres][1] = lteel[numres][0] + eel, lteel[numres][1] + eel * eel
               ltpol[numres][0], ltpol[numres][1] = ltpol[numres][0] + pol, ltpol[numres][1] + pol * pol
               ltsas[numres][0], ltsas[numres][1] = ltsas[numres][0] + sas, ltsas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline()
         if line[0:58] == "                    PRINT PAIR DECOMP - SIDECHAIN ENERGIES":
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline() # data starts here
            numres = 0
            while line[0:3] == 'SDC':
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               lsint[numres][0], lsint[numres][1] = lsint[numres][0] + eint, lsint[numres][1] + eint * eint
               lsvdw[numres][0], lsvdw[numres][1] = lsvdw[numres][0] + vdw, lsvdw[numres][1] + vdw * vdw
               lseel[numres][0], lseel[numres][1] = lseel[numres][0] + eel, lseel[numres][1] + eel * eel
               lspol[numres][0], lspol[numres][1] = lspol[numres][0] + pol, lspol[numres][1] + pol * pol
               lssas[numres][0], lssas[numres][1] = lssas[numres][0] + sas, lssas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline()
         if line[0:57] == "                    PRINT PAIR DECOMP - BACKBONE ENERGIES":
            # eat the next 3 lines
            mdout.readline()
            mdout.readline()
            mdout.readline()
            line = mdout.readline()
            numres = 0
            while line[0:3] == 'BDC':
               eint = float(line[21:30])
               vdw = float(line[31:40])
               eel = float(line[41:50])
               pol = float(line[51:60])
               sas = float(line[61:70]) * surften
               lbint[numres][0], lbint[numres][1] = lbint[numres][0] + eint, lbint[numres][1] + eint * eint
               lbvdw[numres][0], lbvdw[numres][1] = lbvdw[numres][0] + vdw, lbvdw[numres][1] + vdw * vdw
               lbeel[numres][0], lbeel[numres][1] = lbeel[numres][0] + eel, lbeel[numres][1] + eel * eel
               lbpol[numres][0], lbpol[numres][1] = lbpol[numres][0] + pol, lbpol[numres][1] + pol * pol
               lbsas[numres][0], lbsas[numres][1] = lbsas[numres][0] + sas, lbsas[numres][1] + sas * sas
               numres += 1
               line = mdout.readline()
         line = mdout.readline() 

      mdout.close()
   
      for i in range(len(lress)):
         ltint[i][0] /= frames
         ltint[i][1] = math.sqrt(abs(ltint[i][1]/frames - ltint[i][0] ** 2))
         ltvdw[i][0] /= frames
         ltvdw[i][1] = math.sqrt(abs(ltvdw[i][1]/frames - ltvdw[i][0] ** 2))
         lteel[i][0] /= frames
         lteel[i][1] = math.sqrt(abs(lteel[i][1]/frames - lteel[i][0] ** 2))
         ltpol[i][0] /= frames
         ltpol[i][1] = math.sqrt(abs(ltpol[i][1]/frames - ltpol[i][0] ** 2))
         ltsas[i][0] /= frames
         ltsas[i][1] = math.sqrt(abs(ltsas[i][1]/frames - ltsas[i][0] ** 2))
         if verbose == 1 or verbose == 3:
            lsint[i][0] /= frames
            lsint[i][1] = math.sqrt(abs(lsint[i][1]/frames - lsint[i][0] ** 2))
            lsvdw[i][0] /= frames
            lsvdw[i][1] = math.sqrt(abs(lsvdw[i][1]/frames - lsvdw[i][0] ** 2))
            lseel[i][0] /= frames
            lseel[i][1] = math.sqrt(abs(lseel[i][1]/frames - lseel[i][0] ** 2))
            lspol[i][0] /= frames
            lspol[i][1] = math.sqrt(abs(lspol[i][1]/frames - lspol[i][0] ** 2))
            lssas[i][0] /= frames
            lssas[i][1] = math.sqrt(abs(lssas[i][1]/frames - lssas[i][0] ** 2))
            lbint[i][0] /= frames
            lbint[i][1] = math.sqrt(abs(lbint[i][1]/frames - lbint[i][0] ** 2))
            lbvdw[i][0] /= frames
            lbvdw[i][1] = math.sqrt(abs(lbvdw[i][1]/frames - lbvdw[i][0] ** 2))
            lbeel[i][0] /= frames
            lbeel[i][1] = math.sqrt(abs(lbeel[i][1]/frames - lbeel[i][0] ** 2))
            lbpol[i][0] /= frames
            lbpol[i][1] = math.sqrt(abs(lbpol[i][1]/frames - lbpol[i][0] ** 2))
            lbsas[i][0] /= frames
            lbsas[i][1] = math.sqrt(abs(lbsas[i][1]/frames - lbsas[i][0] ** 2))

# print output
   if decomp == 3:
      output.write("idecomp = 3: Pairwise decomposition adding 1-4 interactions added to Internal.\n")
   if decomp == 4:
      output.write("idecomp = 4: Pairwise decomposition adding 1-4 interactions added to EEL and VDW.\n")
   if ligstart == -1 and verbose <= 2 and not stability:
      output.write("DELTAs can only be printed with the default mask assignments\n\n")
      verbose += 2
   if type == 'gb':
      output.write("Pairwise Energy Decomposition Analysis (All units kcal/mol): Generalized Born solvent\n\n")
   elif type == 'pb':
      output.write("Pairwise Energy Decomposition Analysis (All units kcal/mol): Poisson Boltzmann solvent\n\n")

# Everything else will hereforth be added to the final output file using command-line >>. Close it here.
   output.close()

# To avoid having to loop through several times, we use several temporary files.
# tcom : file that holds   Total   COMplex  : Opened for DEC_VERBOSE 2,3
# scom : file that holds Sidechain COMplex  : Opened for DEC_VERBOSE 3
# bcom : file that holds Backbone  COMplex  : Opened for DEC_VERBOSE 3
# trec : file that holds   Total   RECeptor : Opened for DEC_VERBOSE 2,3
# srec : file that holds Sidechain RECeptor : Opened for DEC_VERBOSE 3
# brec : file that holds Backbone  RECeptor : Opened for DEC_VERBOSE 3
# tlig : file that holds   Total   LIGand   : Opened for DEC_VERBOSE 2,3
# slig : file that holds Sidechain LIGand   : Opened for DEC_VERBOSE 3
# blig : file that holds Backbone  LIGand   : Opened for DEC_VERBOSE 3
# tdel : file that holds   Total   DELta    : Opened for DEC_VERBOSE 0,1,2,3
# sdel : file that holds Sidechain DELta    : Opened for DEC_VERBOSE 1,3
# bdel : file that holds Backbone  DELta    : Opened for DEC_VERBOSE 1,3

   if not stability and ligstart != -1:
      tdel = open('_MMPBSA_tdel.txt','w')
   if verbose == 2 or verbose == 3 or stability:
      tcom = open('_MMPBSA_tcom.txt','w')
   if (verbose == 2 or verbose == 3) and not stability:
      trec = open('_MMPBSA_trec.txt','w')
      tlig = open('_MMPBSA_tlig.txt','w')
   if verbose == 3 and not stability:
      srec = open('_MMPBSA_srec.txt','w')
      brec = open('_MMPBSA_brec.txt','w')
      slig = open('_MMPBSA_slig.txt','w')
      blig = open('_MMPBSA_blig.txt','w')
   if (verbose == 3 and not stability) or ((verbose == 1 or verbose == 3) and stability):
      scom = open('_MMPBSA_scom.txt','w')
      bcom = open('_MMPBSA_bcom.txt','w')
   if (verbose == 1 or verbose == 3) and not stability and ligstart != -1:
      sdel = open('_MMPBSA_sdel.txt','w')
      bdel = open('_MMPBSA_bdel.txt','w')

   if not stability and ligstart != -1:
      tdel.write("""
DELTAS:
Total Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")

   if verbose == 2 or verbose == 3 or stability:
      tcom.write("Complex:" + """
Total Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")
   if (verbose == 2 or verbose == 3) and not stability:
      trec.write("\nReceptor:" + """
Total Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")
      tlig.write("\nLigand:" + """
Total Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")

   if (verbose == 3 and not stability) or ((verbose == 3 or verbose == 1) and stability):
      scom.write("""
Sidechain Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")
      bcom.write("""
Backbone Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")
   if verbose == 3 and not stability:
      srec.write("""
Sidechain Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")
      brec.write("""
Backbone Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")
      slig.write("""
Sidechain Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")
      blig.write("""
Backbone Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")

   if (verbose == 1 or verbose == 3) and ligstart != -1 and not stability:
      sdel.write("""
Sidechain Energy Decomposition:
Residue | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")
      bdel.write("""
Backbone Energy Decomposition:
Resid 1 | Resid 2 |       Internal      |    van der Waals    |    Electrostatic    |   Polar Solvation   |   Non-Polar Solv.   |       TOTAL       
-----------------------------------------------------------------------------------------------------------------------------------------------------
""")

   for x in range(len(cress)): # master loop -- loop through once 
      if verbose == 2 or verbose == 3 or stability: # total complex
         total_avg = ctint[x][0] + ctvdw[x][0] + cteel[x][0] + ctpol[x][0] + ctsas[x][0]
         total_stdev = math.sqrt(ctint[x][1]**2 + ctvdw[x][1]**2 + cteel[x][1]**2 + ctpol[x][1]**2 + ctsas[x][1]**2)
         tcom.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
              (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), com_resnames[cress2[x]-1] + str(cress2[x]).rjust(4), ctint[x][0], ctint[x][1], 
               ctvdw[x][0], ctvdw[x][1], cteel[x][0], cteel[x][1], ctpol[x][0], ctpol[x][1], ctsas[x][0], ctsas[x][1], total_avg, total_stdev))

      if (verbose == 3 and not stability) or ((verbose == 3 or verbose == 1) and stability): # sidechain complex
         total_avg = csint[x][0] + csvdw[x][0] + cseel[x][0] + cspol[x][0] + cssas[x][0]
         total_stdev = math.sqrt(csint[x][1]**2 + csvdw[x][1]**2 + cseel[x][1]**2 + cspol[x][1]**2 + cssas[x][1]**2)
         scom.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
              (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), com_resnames[cress2[x]-1] + str(cress2[x]).rjust(4), csint[x][0], csint[x][1],
               csvdw[x][0], csvdw[x][1], cseel[x][0], cseel[x][1], cspol[x][0], cspol[x][1], cssas[x][0], cssas[x][1], total_avg, total_stdev))
         # backbone complex
         total_avg = cbint[x][0] + cbvdw[x][0] + cbeel[x][0] + cbpol[x][0] + cbsas[x][0]
         total_stdev = math.sqrt(cbint[x][1]**2 + cbvdw[x][1]**2 + cbeel[x][1]**2 + cbpol[x][1]**2 + cbsas[x][1]**2)
         bcom.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
              (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), com_resnames[cress2[x]-1] + str(cress2[x]).rjust(4), cbint[x][0], cbint[x][1],
               cbvdw[x][0], cbvdw[x][1], cbeel[x][0], cbeel[x][1], cbpol[x][0], cbpol[x][1], cbsas[x][0], cbsas[x][1], total_avg, total_stdev))

      if not stability and (verbose == 2 or verbose == 3) and x < len(rress): # total receptor
         total_avg = rtint[x][0] + rtvdw[x][0] + rteel[x][0] + rtpol[x][0] + rtsas[x][0]
         total_stdev = math.sqrt(rtint[x][1]**2 + rtvdw[x][1]**2 + rteel[x][1]**2 + rtpol[x][1]**2 + rtsas[x][1]**2)
         trec.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
              (rec_resnames[rress[x]-1] + str(rress[x]).rjust(4), rec_resnames[rress2[x]-1] + str(rress2[x]).rjust(4), rtint[x][0], rtint[x][1],
               rtvdw[x][0], rtvdw[x][1], rteel[x][0], rteel[x][1], rtpol[x][0], rtpol[x][1], rtsas[x][0], rtsas[x][1], total_avg, total_stdev))

      if not stability and verbose == 3 and x < len(rress): # sidechain receptor
         total_avg = rsint[x][0] + rsvdw[x][0] + rseel[x][0] + rspol[x][0] + rssas[x][0]
         total_stdev = math.sqrt(rsint[x][1]**2 + rsvdw[x][1]**2 + rseel[x][1]**2 + rspol[x][1]**2 + rssas[x][1]**2)
         srec.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
              (rec_resnames[rress[x]-1] + str(rress[x]).rjust(4), rec_resnames[rress2[x]-1] + str(rress2[x]).rjust(4), rsint[x][0], rsint[x][1],
               rsvdw[x][0], rsvdw[x][1], rseel[x][0], rseel[x][1], rspol[x][0], rspol[x][1], rssas[x][0], rssas[x][1], total_avg, total_stdev))
         # backbone receptor
         total_avg = rbint[x][0] + rbvdw[x][0] + rbeel[x][0] + rbpol[x][0] + rbsas[x][0]
         total_stdev = math.sqrt(rbint[x][1]**2 + rbvdw[x][1]**2 + rbeel[x][1]**2 + rbpol[x][1]**2 + rbsas[x][1]**2)
         brec.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
              (rec_resnames[rress[x]-1] + str(rress[x]).rjust(4), rec_resnames[rress2[x]-1] + str(rress2[x]).rjust(4), rbint[x][0], rbint[x][1],
               rbvdw[x][0], rbvdw[x][1], rbeel[x][0], rbeel[x][1], rbpol[x][0], rbpol[x][1], rbsas[x][0], rbsas[x][1], total_avg, total_stdev))

      if not stability and (verbose == 2 or verbose == 3) and x < len(lress): # total ligand
         total_avg = ltint[x][0] + ltvdw[x][0] + lteel[x][0] + ltpol[x][0] + ltsas[x][0]
         total_stdev = math.sqrt(ltint[x][1]**2 + ltvdw[x][1]**2 + lteel[x][1]**2 + ltpol[x][1]**2 + ltsas[x][1]**2)
         tlig.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
              (lig_resnames[lress[x]-1] + str(lress[x]).rjust(4), lig_resnames[lress2[x]-1] + str(lress2[x]).rjust(4), ltint[x][0], ltint[x][1],
               ltvdw[x][0], ltvdw[x][1], lteel[x][0], lteel[x][1], ltpol[x][0], ltpol[x][1], ltsas[x][0], ltsas[x][1], total_avg, total_stdev))

      if not stability and verbose == 3 and x < len(lress): # sidechain ligand
         total_avg = lsint[x][0] + lsvdw[x][0] + lseel[x][0] + lspol[x][0] + lssas[x][0]
         total_stdev = math.sqrt(lsint[x][1]**2 + lsvdw[x][1]**2 + lseel[x][1]**2 + lspol[x][1]**2 + lssas[x][1]**2)
         slig.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
              (lig_resnames[lress[x]-1] + str(lress[x]).rjust(4), lig_resnames[lress2[x]-1] + str(lress2[x]).rjust(4), lsint[x][0], lsint[x][1],
               lsvdw[x][0], lsvdw[x][1], lseel[x][0], lseel[x][1], lspol[x][0], lspol[x][1], lssas[x][0], lssas[x][1], total_avg, total_stdev))
         # backbone ligand
         total_avg = lbint[x][0] + lbvdw[x][0] + lbeel[x][0] + lbpol[x][0] + lbsas[x][0]
         total_stdev = math.sqrt(lbint[x][1]**2 + lbvdw[x][1]**2 + lbeel[x][1]**2 + lbpol[x][1]**2 + lbsas[x][1]**2)
         blig.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
              (lig_resnames[lress[x]-1] + str(lress[x]).rjust(4), lig_resnames[lress2[x]-1] + str(lress2[x]).rjust(4), lbint[x][0], lbint[x][1],
               lbvdw[x][0], lbvdw[x][1], lbeel[x][0], lbeel[x][1], lbpol[x][0], lbpol[x][1], lbsas[x][0], lbsas[x][1], total_avg, total_stdev))

# Time for the fun part: DELTAs.

      if not stability and ligstart != -1:
         # Figure out location of the complex residue
         if (cress[x] < ligstart or cress[x] >= ligstart + len(lig_resnames)) and \
               (cress2[x] < ligstart or cress2[x] >= ligstart + len(lig_resnames)): # if both are in receptor
            dtint = [ctint[x][0] - rtint[recdone][0], math.sqrt(ctint[x][1]**2+rtint[recdone][1]**2)]
            dtvdw = [ctvdw[x][0] - rtvdw[recdone][0], math.sqrt(ctvdw[x][1]**2+rtvdw[recdone][1]**2)]
            dteel = [cteel[x][0] - rteel[recdone][0], math.sqrt(cteel[x][1]**2+rteel[recdone][1]**2)]
            dtpol = [ctpol[x][0] - rtpol[recdone][0], math.sqrt(ctpol[x][1]**2+rtpol[recdone][1]**2)]
            dtsas = [ctsas[x][0] - rtsas[recdone][0], math.sqrt(ctsas[x][1]**2+rtsas[recdone][1]**2)]
            if verbose == 1 or verbose == 3:
               dsint = [csint[x][0] - rsint[recdone][0], math.sqrt(csint[x][1]**2+rsint[recdone][1]**2)]
               dsvdw = [csvdw[x][0] - rsvdw[recdone][0], math.sqrt(csvdw[x][1]**2+rsvdw[recdone][1]**2)]
               dseel = [cseel[x][0] - rseel[recdone][0], math.sqrt(cseel[x][1]**2+rseel[recdone][1]**2)]
               dspol = [cspol[x][0] - rspol[recdone][0], math.sqrt(cspol[x][1]**2+rspol[recdone][1]**2)]
               dssas = [cssas[x][0] - rssas[recdone][0], math.sqrt(cssas[x][1]**2+rssas[recdone][1]**2)]
               dbint = [cbint[x][0] - rbint[recdone][0], math.sqrt(cbint[x][1]**2+rbint[recdone][1]**2)]
               dbvdw = [cbvdw[x][0] - rbvdw[recdone][0], math.sqrt(cbvdw[x][1]**2+rbvdw[recdone][1]**2)]
               dbeel = [cbeel[x][0] - rbeel[recdone][0], math.sqrt(cbeel[x][1]**2+rbeel[recdone][1]**2)]
               dbpol = [cbpol[x][0] - rbpol[recdone][0], math.sqrt(cbpol[x][1]**2+rbpol[recdone][1]**2)]
               dbsas = [cbsas[x][0] - rbsas[recdone][0], math.sqrt(cbsas[x][1]**2+rbsas[recdone][1]**2)]
            recdone += 1
         elif (cress[x] >= ligstart and cress[x] < ligstart + len(lig_resnames)) and \
                 (cress2[x] >= ligstart and cress2[x] < ligstart + len(lig_resnames)): # if both are in ligand
            dtint = [ctint[x][0] - ltint[ligdone][0], math.sqrt(ctint[x][1]**2+ltint[ligdone][1]**2)]
            dtvdw = [ctvdw[x][0] - ltvdw[ligdone][0], math.sqrt(ctvdw[x][1]**2+ltvdw[ligdone][1]**2)]
            dteel = [cteel[x][0] - lteel[ligdone][0], math.sqrt(cteel[x][1]**2+lteel[ligdone][1]**2)]
            dtpol = [ctpol[x][0] - ltpol[ligdone][0], math.sqrt(ctpol[x][1]**2+ltpol[ligdone][1]**2)]
            dtsas = [ctsas[x][0] - ltsas[ligdone][0], math.sqrt(ctsas[x][1]**2+ltsas[ligdone][1]**2)]
            if verbose == 1 or verbose == 3:
               dsint = [csint[x][0] - lsint[ligdone][0], math.sqrt(csint[x][1]**2+lsint[ligdone][1]**2)]
               dsvdw = [csvdw[x][0] - lsvdw[ligdone][0], math.sqrt(csvdw[x][1]**2+lsvdw[ligdone][1]**2)]
               dseel = [cseel[x][0] - lseel[ligdone][0], math.sqrt(cseel[x][1]**2+lseel[ligdone][1]**2)]
               dspol = [cspol[x][0] - lspol[ligdone][0], math.sqrt(cspol[x][1]**2+lspol[ligdone][1]**2)]
               dssas = [cssas[x][0] - lssas[ligdone][0], math.sqrt(cssas[x][1]**2+lssas[ligdone][1]**2)]
               dbint = [cbint[x][0] - lbint[ligdone][0], math.sqrt(cbint[x][1]**2+lbint[ligdone][1]**2)]
               dbvdw = [cbvdw[x][0] - lbvdw[ligdone][0], math.sqrt(cbvdw[x][1]**2+lbvdw[ligdone][1]**2)]
               dbeel = [cbeel[x][0] - lbeel[ligdone][0], math.sqrt(cbeel[x][1]**2+lbeel[ligdone][1]**2)]
               dbpol = [cbpol[x][0] - lbpol[ligdone][0], math.sqrt(cbpol[x][1]**2+lbpol[ligdone][1]**2)]
               dbsas = [cbsas[x][0] - lbsas[ligdone][0], math.sqrt(cbsas[x][1]**2+lbsas[ligdone][1]**2)]
            ligdone += 1
         else: # if one's in each, it's just in complex
            dtint = [ctint[x][0], ctint[x][1]]
            dtvdw = [ctvdw[x][0], ctvdw[x][1]]
            dteel = [cteel[x][0], cteel[x][1]]
            dtpol = [ctpol[x][0], ctpol[x][1]]
            dtsas = [ctsas[x][0], ctsas[x][1]]
            if verbose == 1 or verbose == 3:
               dsint = [csint[x][0], csint[x][1]]
               dsvdw = [csvdw[x][0], csvdw[x][1]]
               dseel = [cseel[x][0], cseel[x][1]]
               dspol = [cspol[x][0], cspol[x][1]]
               dssas = [cssas[x][0], cssas[x][1]]
               dbint = [cbint[x][0], cbint[x][1]]
               dbvdw = [cbvdw[x][0], cbvdw[x][1]]
               dbeel = [cbeel[x][0], cbeel[x][1]]
               dbpol = [cbpol[x][0], cbpol[x][1]]
               dbsas = [cbsas[x][0], cbsas[x][1]]

         total_avg = dtint[0] + dtvdw[0] + dteel[0] + dtpol[0] + dtsas[0]
         total_stdev = math.sqrt(dtint[1]**2 + dtvdw[1]**2 + dteel[1]**2 + dtpol[1]**2 + dtsas[1]**2)
         tdel.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
              (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), com_resnames[cress2[x]-1] + str(cress2[x]).rjust(4), dtint[0], dtint[1],
               dtvdw[0], dtvdw[1], dteel[0], dteel[1], dtpol[0], dtpol[1], dtsas[0], dtsas[1], total_avg, total_stdev))

         if verbose == 1 or verbose == 3:
            total_avg = dsint[0] + dsvdw[0] + dseel[0] + dspol[0] + dssas[0]
            total_stdev = math.sqrt(dsint[1]**2 + dsvdw[1]**2 + dseel[1]**2 + dspol[1]**2 + dssas[1]**2)
            sdel.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
                 (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), com_resnames[cress2[x]-1] + str(cress2[x]).rjust(4), dsint[0], dsint[1],
                  dsvdw[0], dsvdw[1], dseel[0], dseel[1], dspol[0], dspol[1], dssas[0], dssas[1], total_avg, total_stdev))

            total_avg = dbint[0] + dbvdw[0] + dbeel[0] + dbpol[0] + dbsas[0]
            total_stdev = math.sqrt(dbint[1]**2 + dbvdw[1]**2 + dbeel[1]**2 + dbpol[1]**2 + dbsas[1]**2)
            bdel.write('%s | %s |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f |%9.3f +/- %6.3f\n' %
                 (com_resnames[cress[x]-1] + str(cress[x]).rjust(4), com_resnames[cress2[x]-1] + str(cress2[x]).rjust(4), dbint[0], dbint[1],
                  dbvdw[0], dbvdw[1], dbeel[0], dbeel[1], dbpol[0], dbpol[1], dbsas[0], dbsas[1], total_avg, total_stdev))

# close all files
   if not stability and ligstart != -1:
      tdel.close()
   if verbose == 2 or verbose == 3 or stability:
      tcom.close()
   if (verbose == 2 or verbose == 3) and not stability:
      trec.close()
      tlig.close()
   if verbose == 3 and not stability:
      srec.close()
      brec.close()
      slig.close()
      blig.close()
   if (verbose == 3 and not stability) or ((verbose == 3 or verbose == 1) and stability):
      scom.close()
      bcom.close()
   if (verbose == 1 or verbose == 3) and ligstart != -1 and not stability:
      sdel.close()
      bdel.close()

   if verbose == 2 or verbose == 3 or stability:
      os.system('cat _MMPBSA_tcom.txt >> %s; rm _MMPBSA_tcom.txt' % outname)
   if (verbose == 3 and not stability) or ((verbose == 1 or verbose == 3) and stability):
      os.system('cat _MMPBSA_scom.txt >> %s; rm _MMPBSA_scom.txt' % outname)
      os.system('cat _MMPBSA_bcom.txt >> %s; rm _MMPBSA_bcom.txt' % outname)
   if verbose == 2 or verbose == 3 and not stability:
      os.system('cat _MMPBSA_trec.txt >> %s; rm _MMPBSA_trec.txt' % outname)
   if verbose == 3 and not stability:
      os.system('cat _MMPBSA_srec.txt >> %s; rm _MMPBSA_srec.txt' % outname)
      os.system('cat _MMPBSA_brec.txt >> %s; rm _MMPBSA_brec.txt' % outname)
   if verbose == 2 or verbose == 3 and not stability:
      os.system('cat _MMPBSA_tlig.txt >> %s; rm _MMPBSA_tlig.txt' % outname)
   if verbose == 3 and not stability:
      os.system('cat _MMPBSA_slig.txt >> %s; rm _MMPBSA_slig.txt' % outname)
      os.system('cat _MMPBSA_blig.txt >> %s; rm _MMPBSA_blig.txt' % outname)
   if not stability and ligstart != -1:
      os.system('cat _MMPBSA_tdel.txt >> %s; rm _MMPBSA_tdel.txt' % outname)
   if (verbose == 1 or verbose == 3) and not stability and ligstart != -1:
      os.system('cat _MMPBSA_sdel.txt >> %s; rm _MMPBSA_sdel.txt' % outname)
      os.system('cat _MMPBSA_bdel.txt >> %s; rm _MMPBSA_bdel.txt' % outname)
   os.system('echo "" >> %s' % outname)
   return 0


#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _nmodeCombine(INPUT, numthreads, stability):
   """ Combine nmode output files """
   if not INPUT['mutant_only'] == 1:
      combinedcom = open('_MMPBSA_complex_nm.out','w')
      if not stability:
         combinedrec = open('_MMPBSA_receptor_nm.out' ,'w')
         combinedlig = open('_MMPBSA_ligand_nm.out', 'w')

      for i in range(numthreads):
         tempout = open('_MMPBSA_complex_nm.out.%s' % i, 'r')
         _nmodefile(tempout, combinedcom)
         tempout.close()

         if not stability:
            tempout = open('_MMPBSA_receptor_nm.out.%s' % i, 'r')
            _nmodefile(tempout, combinedrec)
            tempout.close()
            
            tempout = open('_MMPBSA_ligand_nm.out.%s' % i, 'r')
            _nmodefile(tempout, combinedlig)
            tempout.close()
      combinedcom.close()
      if not stability:
         combinedrec.close()
         combinedlig.close()

   if INPUT['alarun'] == 1:
      combinedcom = open('_MMPBSA_mutant_complex_nm.out','w')
      if not stability:
         combinedrec = open('_MMPBSA_mutant_receptor_nm.out' ,'w')
         combinedlig = open('_MMPBSA_mutant_ligand_nm.out', 'w')

      for i in range(numthreads):
         tempout = open('_MMPBSA_mutant_complex_nm.out.%s' % i, 'r')
         _nmodefile(tempout, combinedcom)
         tempout.close() 
         if not stability:
            tempout = open('_MMPBSA_mutant_receptor_nm.out.%s' % i, 'r')
            _nmodefile(tempout, combinedrec)
            tempout.close()
            
            tempout = open('_MMPBSA_mutant_ligand_nm.out.%s' % i, 'r')
            _nmodefile(tempout, combinedlig)
            tempout.close()
      combinedcom.close()
      if not stability:
         combinedrec.close()
         combinedlig.close()

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _rismCombine(INPUT, numthreads, stability):
   """ Combines all 3D-RISM output files into a single output file """
   if not INPUT['mutant_only'] == 1:
      combinedcom = open('_MMPBSA_complex_rism.out','w')
      if not stability:
         combinedrec = open('_MMPBSA_receptor_rism.out','w')
         combinedlig = open('_MMPBSA_ligand_rism.out','w')
      
      for i in range(numthreads): # go through all RISM output files...
         tempout = open('_MMPBSA_complex_rism.out.%s' % i, 'r')
         _rismfile(tempout, combinedcom)
         tempout.close()

         if not stability:
            tempout = open('_MMPBSA_receptor_rism.out.%s' % i, 'r')
            _rismfile(tempout, combinedrec)
            tempout.close()

            tempout = open('_MMPBSA_ligand_rism.out.%s' % i, 'r')
            _rismfile(tempout, combinedlig)
            tempout.close()
      combinedcom.close()
      if not stability:
         combinedrec.close()
         combinedlig.close()

   # mutants now
   if INPUT['alarun']:
      combinedcom = open('_MMPBSA_mutant_complex_rism.out','w')
      if not stability:
         combinedrec = open('_MMPBSA_mutant_receptor_rism.out','w')
         combinedlig = open('_MMPBSA_mutant_ligand_rism.out','w')
      
      for i in range(numthreads): # go through all RISM output files...
         tempout = open('_MMPBSA_mutant_complex_rism.out.%s' % i, 'r')
         _rismfile(tempout, combinedcom)
         tempout.close()

         if not stability:
            tempout = open('_MMPBSA_mutant_receptor_rism.out.%s' % i, 'w')
            _rismfile(tempout, combinedrec)
            tempout.close()

            tempout = open('_MMPBSA_mutant_ligand_rism.out.%s' % i, 'w')
            _rismfile(tempout, combinedlig)
            tempout.close()
      combinedcom.close()
      if not stability:
         combinedrec.close()
         combinedlig.close()

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _noDecSander(INPUT, numthreads, stability, type):
   """ Combines all thread-specific sander output files into a single, combined one """
   if not INPUT['mutant_only'] == 1:
      combinedcom = open('_MMPBSA_complex_%s.mdout' % type,'w')
      if not stability:
         combinedrec = open('_MMPBSA_receptor_%s.mdout' % type,'w')
         combinedlig = open('_MMPBSA_ligand_%s.mdout' % type,'w')

      for i in range(numthreads): # go through all GB output files...

         tempout = open('_MMPBSA_complex_%s.mdout.%s' % (type,i), 'r')
         _sanderfile(tempout, combinedcom)
         tempout.close()

         if not stability:
            tempout = open('_MMPBSA_receptor_%s.mdout.%s' % (type, i), 'r')
            _sanderfile(tempout, combinedrec)
            tempout.close()

            tempout = open('_MMPBSA_ligand_%s.mdout.%s' % (type, i), 'r')
            _sanderfile(tempout, combinedlig)
            tempout.close()
      combinedcom.close()
      if not stability:
         combinedrec.close()
         combinedlig.close()

   # mutants now
   if INPUT['alarun']:
      combinedcom = open('_MMPBSA_mutant_complex_%s.mdout' % type,'w')
      if not stability:
         combinedrec = open('_MMPBSA_mutant_receptor_%s.mdout' % type,'w')
         combinedlig = open('_MMPBSA_mutant_ligand_%s.mdout' % type,'w')

      for i in range(numthreads):

         tempout = open('_MMPBSA_mutant_complex_%s.mdout.%s' % (type, i), 'r')
         _sanderfile(tempout, combinedcom)
         tempout.close()

         if not stability:
            tempout = open('_MMPBSA_mutant_receptor_%s.mdout.%s' % (type, i), 'r')
            _sanderfile(tempout, combinedrec)
            tempout.close()

            tempout = open('_MMPBSA_mutant_ligand_%s.mdout.%s' % (type, i), 'r')
            _sanderfile(tempout, combinedlig)
      combinedcom.close()
      if not stability:
         combinedrec.close()
         combinedlig.close()

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _DecSander(INPUT, numthreads, stability, type):
   """ Same as _noDecSander, but for sander decomposition files instead """
   if not INPUT['mutant_only'] == 1:
      combinedcom = open('_MMPBSA_complex_%s.mdout' % type, 'w')
      if not stability:
         combinedrec = open('_MMPBSA_receptor_%s.mdout' % type, 'w')
         combinedlig = open('_MMPBSA_ligand_%s.mdout' % type, 'w')

      for i in range(numthreads):

         tempout = open('_MMPBSA_complex_%s.mdout.%s' % (type,i), 'r')
         _sanderdecfile(tempout, combinedcom)
         tempout.close()

         if not stability:
            tempout = open('_MMPBSA_receptor_%s.mdout.%s' % (type,i), 'r')
            _sanderdecfile(tempout, combinedrec)
            tempout.close()

            tempout = open('_MMPBSA_ligand_%s.mdout.%s' % (type,i), 'r')
            _sanderdecfile(tempout, combinedlig)
            tempout.close()
      combinedcom.close()
      if not stability:
         combinedrec.close()
         combinedlig.close()

   if INPUT['alarun']:
      combinedcom = open('_MMPBSA_mutant_complex_%s.mdout' % type, 'w')
      if not stability:
         combinedrec = open('_MMPBSA_mutant_receptor_%s.mdout' % type, 'w')
         combinedlig = open('_MMPBSA_mutant_ligand_%s.mdout' % type, 'w')

      for i in range(numthreads):

         tempout = open('_MMPBSA_mutant_complex_%s.mdout.%s' % (type,i), 'r')
         _sanderdecfile(tempout, combinedcom)
         tempout.close()

         if not stability:
            tempout = open('_MMPBSA_mutant_receptor_%s.mdout.%s' % (type,i), 'r')
            _sanderdecfile(tempout, combinedrec)
            tempout.close()

            tempout = open('_MMPBSA_mutant_ligand_%s.mdout.%s' % (type,i), 'r')
            _sanderdecfile(tempout, combinedlig)
            tempout.close()
      combinedcom.close()
      if not stability:
         combinedrec.close()
         combinedlig.close()

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _nmodefile(infile, outfile):
   """ Pulls out important data from nmode output file *infile* and dumps it to outfile """
   rawline = infile.readline()

   while len(rawline) > 0:
      if rawline[:18] == "             freq.":
         outfile.write(rawline)
         outfile.write(infile.readline())
         outfile.write(infile.readline())
         outfile.write(infile.readline())
         outfile.write(infile.readline())
         outfile.write(infile.readline())
         outfile.write(infile.readline())
         outfile.write('\n')
      elif rawline[0:35] == '   |---- Entropy not Calculated---|':
         outfile.write(rawline)
         outfile.write('\n')

      rawline = infile.readline()

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _sanderfile(infile, outfile):
   """ Pulls out important data from nmode output file *infile* and dumps it to outfile """
   rawline = infile.readline()

   while len(rawline) > 0:
      if 'error' in rawline.lower(): # catch sander errors
         print >> sys.stderr, 'Error: sander error in GB! See error line in output file below:'
         print >> sys.stderr, rawline
         raise ValueError

      if rawline[1:5] == 'BOND': # we've found the start of a frame
         outfile.write(rawline)
         outfile.write(infile.readline())
         outfile.write(infile.readline())
         outfile.write(infile.readline())
         outfile.write(infile.readline())
         outfile.write('\n')

      rawline = infile.readline()

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _sanderdecfile(infile, outfile):
   """ Same as _sanderfile, but for decomposition sander outputs instead """
   rawline = infile.readline()

   while len(rawline) > 0:
      if rawline[1:5] == 'BOND': # we've found the start of a frame
         outfile.write(rawline)
         outfile.write(infile.readline())
         outfile.write(infile.readline())
         outfile.write(infile.readline())
         outfile.write('\n')

      elif rawline[:25] == '                    PRINT':
         while rawline[:12] != 'minimization' and len(rawline) > 0:
            outfile.write(rawline)
            rawline = infile.readline()

      rawline = infile.readline()

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _rismfile(infile, outfile):
   """ Takes 3D-RISM data from *infile* and dumps it to *outfile* """
   rawline = infile.readline()

   while len(rawline) > 0:
      if rawline[0:6] == 'Frame:': # we've found the start of a frame
         outfile.write(rawline)
         rawline = infile.readline()
         while rawline[0:6] != 'Frame:' and not "processing complete." in rawline:
            outfile.write(rawline)
            rawline = infile.readline()
      rawline = infile.readline()

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _avg(array):
   """ Returns the average of a passed array """
   return sum(array) / len(array)

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

def _stdev(array):
   """ Returns the standard deviation of a passed array """
   sum = sum2 = 0
   for i in range(len(array)):
      sum += array[i]
      sum2 += array[i] * array[i]

   avg = sum / len(array)
   return math.sqrt(abs((sum2 / len(array) - avg * avg))) # use abs for round-offs if it should be 0

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class QHout:
   """ Quasi-harmonic output file class """

#==============================

   def __init__(self, filename, stab, temp, desc=''):
      """ Initializes a QHout object. Declares whether it's stability, the filename, etc. """
      self.filename = filename
      self.stability = stab
      self.temperature = temp
      self.exists = os.path.exists(filename)
      self.valid = self.exists and self._read()
      self.description = desc

#==============================

   def __str__(self):
      """ String representation of Quasi-harmonic results -- how it appears in the output file """
      if not self.stability and self.valid:
         return '%sENTROPY RESULTS (QUASI-HARMONIC APPROXIMATION) CALCULATED WITH PTRAJ:' % self.description + \
         '\n\n           Translational      Rotational      Vibrational           Total' + \
         '\nComplex:   %13.4f %15.4f %16.4f %15.4f' % (self.com[1], self.com[2], self.com[3], self.com[0]) + \
         '\nReceptor:  %13.4f %15.4f %16.4f %15.4f' % (self.rec[1], self.rec[2], self.rec[3], self.rec[0]) + \
         '\nLigand:    %13.4f %15.4f %16.4f %15.4f' % (self.lig[1], self.lig[2], self.lig[3], self.lig[0]) + \
         '\n\nDELTA S:   %13.4f %15.4f %16.4f %15.4f\n' % (self.com[1] - self.rec[1] - self.lig[1],
                                                         self.com[2] - self.rec[2] - self.lig[2],
                                                         self.com[3] - self.rec[3] - self.lig[3],
                                                         self.com[0] - self.rec[0] - self.lig[0] )
      elif self.valid:
         return '%sENTROPY RESULTS (QUASI-HARMONIC APPROXIMATION) CALCULATED WITH PTRAJ:' % self.description + \
                '\n\n           Translational      Rotational      Vibrational           Total' + \
                '\nComplex:   %13.4f %15.4f %16.4f %15.4f\n' % (self.com[1], self.com[2], self.com[3], self.com[0])

      else:
         return 'Error in parsing Quasi-harmonic entropy! Check error messages printed to STDERR\n'

#==============================

   def TotalAvg(self):
      """ Returns the average of the total """
      return self.com[0] - self.rec[0] - self.lig[0]

#==============================

   def _read(self):
      """ Parses the output files and fills the data arrays """
      file = open(self.filename, 'r')
      rawline = file.readline()
      self.com = [0,0,0,0] # complex values
      self.rec = [0,0,0,0] # receptor values
      self.lig = [0,0,0,0] # ligand values
      comdone = False # if we've done the complex yet (filled in self.com)
      recdone = False # if we've done the receptor yet (filled in self.rec)

      # Try to fill in all found entropy values. If we can only find 1 set, we're
      # doing stability calculations.
      try:
         while len(rawline) > 0:
            if rawline[0:6] == " Total":
               if not comdone:
                  comdone = True
                  self.com[0] = float(rawline.split()[3])*self.temperature/1000
                  self.com[1] = float(file.readline().split()[3])*self.temperature/1000
                  self.com[2] = float(file.readline().split()[3])*self.temperature/1000
                  self.com[3] = float(file.readline().split()[3])*self.temperature/1000
               elif not recdone:
                  recdone = True
                  self.rec[0] = float(rawline.split()[3])*self.temperature/1000
                  self.rec[1] = float(file.readline().split()[3])*self.temperature/1000
                  self.rec[2] = float(file.readline().split()[3])*self.temperature/1000
                  self.rec[3] = float(file.readline().split()[3])*self.temperature/1000
               else:
                  self.lig[0] = float(rawline.split()[3])*self.temperature/1000
                  self.lig[1] = float(file.readline().split()[3])*self.temperature/1000
                  self.lig[2] = float(file.readline().split()[3])*self.temperature/1000
                  self.lig[3] = float(file.readline().split()[3])*self.temperature/1000
                  break
   
            rawline = file.readline()

         if not recdone:
            self.stability = True

         file.close()
         return True

      except:
         file.close()
         return False

#==============================

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class NMODEout:
   """ Nmode output file class """

#==============================

   def __init__(self, one_traj, temp, file1='', file2='', file3='', desc=''):
      """ Initializes data about and for the nmode output """
      self.stability = False              # if this is a stability calculation
      self.comfile = file1                # complex nmode output data file
      self.recfile = file2                # receptor nmode output data file
      self.ligfile = file3                # ligand nmode output data file
      self.exists = os.path.exists(file1) # logical whether all files exist
      self.missing_frames = False         # logical if we are missing frames in any of the nmode data files
      self.one_trajectory = one_traj
      self.temperature = temp             # set the temperature

      if self.recfile == '' and self.ligfile == '':
         self.stability = True # see if we do stability (only if rec, lig don't exist)
      else:
         self.exists = os.path.exists(file1) and os.path.exists(file2) and os.path.exists(file3)

      self.valid = self.exists and self._read()  # read the output files
      self.description = desc  # load a description to be printed in __str__

#==============================

   def __str__(self):
      """ String representation of nmode output results -- how it appears in final output """
      if not self.valid:
         return 'Error in computing entropy! Check nmode output files!\n'

      totavg, totstdev = self.AvgStdev()

      printed_string = '%sENTROPY RESULTS (HARMONIC APPROXIMATION) CALCULATED WITH NMODE:\n' % self.description + \
      '\nComplex:\nEntropy Term          Average        Std. Dev.' + \
      '\n-----------------------------------------------------------' + \
      '\nTranslational: %14.4f %16.4f' % (_avg(self.com_data[1]), _stdev(self.com_data[1])) + \
      '\nRotational:    %14.4f %16.4f' % (_avg(self.com_data[2]), _stdev(self.com_data[2])) + \
      '\nVibrational:   %14.4f %16.4f' % (_avg(self.com_data[3]), _stdev(self.com_data[3])) + \
      '\nTotal:         %14.4f %16.4f\n' % (_avg(self.com_data[0]), _stdev(self.com_data[0]))
      if not self.stability:
         printed_string += '\n\nReceptor:\nEntropy Term          Average        Std. Dev.' + \
         '\n-----------------------------------------------------------' + \
         '\nTranslational: %14.4f %16.4f' % (_avg(self.rec_data[1]), _stdev(self.rec_data[1])) + \
         '\nRotational:    %14.4f %16.4f' % (_avg(self.rec_data[2]), _stdev(self.rec_data[2])) + \
         '\nVibrational:   %14.4f %16.4f' % (_avg(self.rec_data[3]), _stdev(self.rec_data[3])) + \
         '\nTotal:         %14.4f %16.4f\n\n' % (_avg(self.rec_data[0]), _stdev(self.rec_data[0])) + \
         '\nLigand:\nEntropy Term          Average        Std. Dev.' + \
         '\n-----------------------------------------------------------' + \
         '\nTranslational: %14.4f %16.4f' % (_avg(self.lig_data[1]), _stdev(self.lig_data[1])) + \
         '\nRotational:    %14.4f %16.4f' % (_avg(self.lig_data[2]), _stdev(self.lig_data[2])) + \
         '\nVibrational:   %14.4f %16.4f' % (_avg(self.lig_data[3]), _stdev(self.lig_data[3])) + \
         '\nTotal:         %14.4f %16.4f\n\n' % (_avg(self.lig_data[0]), _stdev(self.lig_data[0])) + \
         '\nDELTA S total= %14.4f +/- %12.4f\n' % (totavg, totstdev)

      return printed_string

#==============================

   def AvgStdev(self):
      """ Returns the average and stdev of the total binding entropy contribution """
      if not self.valid:
         return -1, -1
      # return average, stdev for the data here
      if not self.stability:
         avg = _avg(self.com_data[0]) - _avg(self.rec_data[0]) - _avg(self.lig_data[0])
         if not self.one_trajectory or self.missing_frames:
            stdev = math.sqrt(abs(_stdev(self.com_data[0]) ** 2 + _stdev(self.rec_data[0]) ** 2 + 
                                  _stdev(self.lig_data[0]) ** 2))
         else:
            stdev = self._tot_stdevs()
      else:
         avg = _avg(self.com_data[0])
         stdev = _stdev(self.com_data[0])

      return avg, stdev

#==============================

   def _tot_stdevs(self):
      """ Returns the standard deviations of the total based on whether or not we were able to
          calculate all frames or not. Use a sample stdev if we were, sum-of-squares if we weren't """
      if len(self.com_data) != len(self.rec_data) or len(self.com_data) != len(self.lig_data):
         return math.sqrt(abs(_stdev(self.com_data[0]) ** 2 + _stdev(self.rec_data[0]) ** 2 + _stdev(self.lig_data[0]) ** 2))

      sum = sum2 = 0
      for i in range(len(self.com_data)):
         sum += self.com_data[0][i] - self.rec_data[0][i] - self.lig_data[0][i]
         sum2 += (self.com_data[0][i] - self.rec_data[0][i] - self.lig_data[0][i]) ** 2

      avg = sum / len(self.com_data)
      return math.sqrt(abs(sum2 / len(self.com_data) - avg * avg))

#==============================

   def _read(self):
      """ Parse nmode output file and fill data arrays """
      com = open(self.comfile, 'r')
      if not self.stability:
         rec = open(self.recfile, 'r')
         lig = open(self.ligfile, 'r')

      self.com_data = self._getdata(com)
      com.close()
      if not self.stability:
         self.rec_data = self._getdata(rec)
         self.lig_data = self._getdata(lig)
         rec.close()
         lig.close()

      if self.com_data == -1:
         return False
      if not self.stability and (self.rec_data == -1 or self.lig_data == -1):
         return False

      return True

#==============================

   def _getdata(self, file):
      """ Get the data from specific nmode output files """
      rawline = file.readline() # read the first line
      data = [[], [], [], []] # entropy data in the file

      while len(rawline) > 0:
         if rawline[0:35] == '   |---- Entropy not Calculated---|':
            self.missing_frames = True
         if rawline[0:6] == 'Total:':
            data[0].append(float(rawline.split()[3])*self.temperature/1000)         # Total
            data[1].append(float(file.readline().split()[3])*self.temperature/1000) # translational
            data[2].append(float(file.readline().split()[3])*self.temperature/1000) # rotational
            data[3].append(float(file.readline().split()[3])*self.temperature/1000) # vibrational

         rawline = file.readline()

      return data

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class QMMMout:
   """ QM/MM GB output file class """

#==============================

   def __init__(self, one_traj=True, verb=0, file1='', file2='', file3='', desc=''):
      """ Initializes QMMMout object with data from passed files, etc."""
      self.one_trajectory = one_traj
      self.verbose = verb
      self.description = desc
      self.comfile = file1
      if file2 == '' and file3 == '':
         self.stability = True
      else:
         self.stability = False
         self.recfile = file2
         self.ligfile = file3

      if self.stability:
         self.exists = os.path.exists(file1)
      else:
         self.exists = os.path.exists(file1) and os.path.exists(file2) and os.path.exists(file3)

      self.valid = self.exists and self._read()

#==============================

   def __str__(self):
      """ String representation of QM/MM data -- as it appears in final output """
      if not self.valid:  return "Problem parsing QM/MM GB output!\n"

      printstring = '%s' % self.description

      consistent_1_4 = self.one_trajectory and self._consistent(['1-4vdw','1-4eel'])
      consistent_int = self.one_trajectory and self._consistent(['bond','angle','dihedral'])

      print_1_4 = self.verbose == 2 or not consistent_1_4
      print_int = self.verbose == 2 or not consistent_int

      if not print_1_4:  self._removeFromTotal(['1-4vdw','1-4eel'])
      if not print_int:  self._removeFromTotal(['bond','angle','dihedral'])

      if not consistent_int and self.one_trajectory and not self.stability:
         printstring += '\n\nWARNING: INCONSISTENCIES EXIST WITHIN INTERNAL POTENTIAL TERMS (BOND, ANGLE, AND/OR DIHED).' + \
         '\n         CHECK YOUR INPUT FILES AND SYSTEM SETUP. THESE RESULTS MAY NOT BE RELIABLE (check differences)!\n'

      if not consistent_1_4 and self.one_trajectory and not self.stability:
         printstring += '\n\nWARNING: INCONSISTENCIES EXIST WITHIN 1-4 NON-BONDED TERMS.\n' + \
         '         CHECK YOUR INPUT FILES AND SYSTEM SETUP. THESE RESULTS MAY NOT BE RELIABLE (check differences)!\n\n' + \
         'The most common cause of this is inconsistent charge definitions across topology files.\n'

      # add the Complex contribution if we asked to print it
      if self.verbose >= 1 or self.stability:
         sqrtframes = math.sqrt(len(self.energy_data['bond'][0]))
         printstring += 'Complex:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][0]),
                           _stdev(self.energy_data['bond'][0]), _stdev(self.energy_data['bond'][0])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][0]),
                           _stdev(self.energy_data['angle'][0]), _stdev(self.energy_data['angle'][0])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][0]),
                           _stdev(self.energy_data['dihedral'][0]), _stdev(self.energy_data['dihedral'][0])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][0]),
                        _stdev(self.energy_data['vdwaals'][0]), _stdev(self.energy_data['vdwaals'][0])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][0]),
                        _stdev(self.energy_data['eel'][0]), _stdev(self.energy_data['eel'][0])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][0]),
                           _stdev(self.energy_data['1-4vdw'][0]), _stdev(self.energy_data['1-4vdw'][0])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][0]),
                           _stdev(self.energy_data['1-4eel'][0]), _stdev(self.energy_data['1-4eel'][0])/sqrtframes)

         printstring += 'EGB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['egb'][0]),
                        _stdev(self.energy_data['egb'][0]), _stdev(self.energy_data['egb'][0])/sqrtframes) + \
         'ESURF          %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['esurf'][0]),
                        _stdev(self.energy_data['esurf'][0]), _stdev(self.energy_data['esurf'][0])/sqrtframes) + \
         'ESCF           %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['escf'][0]),
                        _stdev(self.energy_data['escf'][0]), _stdev(self.energy_data['escf'][0])/sqrtframes) + \
         'G gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][0]),
                        _stdev(self.energy_data['gas'][0]), _stdev(self.energy_data['gas'][0])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][0]),
                        _stdev(self.energy_data['solv'][0]), _stdev(self.energy_data['solv'][0])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][0]),
                        _stdev(self.energy_data['total'][0]), _stdev(self.energy_data['total'][0])/sqrtframes)
      # add the Receptor contribution if we asked to print it
      if self.verbose >= 1 and not self.stability:
         sqrtframes = math.sqrt(len(self.energy_data['bond'][1]))
         printstring += 'Receptor:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][1]),
                           _stdev(self.energy_data['bond'][1]), _stdev(self.energy_data['bond'][1])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][1]),
                           _stdev(self.energy_data['angle'][1]), _stdev(self.energy_data['angle'][1])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][1]),
                           _stdev(self.energy_data['dihedral'][1]), _stdev(self.energy_data['dihedral'][1])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][1]),
                        _stdev(self.energy_data['vdwaals'][1]), _stdev(self.energy_data['vdwaals'][1])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][1]),
                        _stdev(self.energy_data['eel'][1]), _stdev(self.energy_data['eel'][1])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][1]),
                           _stdev(self.energy_data['1-4vdw'][1]), _stdev(self.energy_data['1-4vdw'][1])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][1]),
                           _stdev(self.energy_data['1-4eel'][1]), _stdev(self.energy_data['1-4eel'][1])/sqrtframes)

         printstring += 'EGB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['egb'][1]),
                        _stdev(self.energy_data['egb'][1]), _stdev(self.energy_data['egb'][1])/sqrtframes) + \
         'ESURF          %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['esurf'][1]),
                        _stdev(self.energy_data['esurf'][1]), _stdev(self.energy_data['esurf'][1])/sqrtframes) + \
         'ESCF           %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['escf'][1]),
                        _stdev(self.energy_data['escf'][1]), _stdev(self.energy_data['escf'][1])/sqrtframes) + \
         'G gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][1]),
                        _stdev(self.energy_data['gas'][1]), _stdev(self.energy_data['gas'][1])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][1]),
                        _stdev(self.energy_data['solv'][1]), _stdev(self.energy_data['solv'][1])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][1]),
                        _stdev(self.energy_data['total'][1]), _stdev(self.energy_data['total'][1])/sqrtframes)
      # add the ligand contribution if we asked to print it
         sqrtframes = math.sqrt(len(self.energy_data['bond'][2]))
         printstring += 'Ligand:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][2]),
                           _stdev(self.energy_data['bond'][2]), _stdev(self.energy_data['bond'][2])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][2]),
                           _stdev(self.energy_data['angle'][2]), _stdev(self.energy_data['angle'][2])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][2]),
                           _stdev(self.energy_data['dihedral'][2]), _stdev(self.energy_data['dihedral'][2])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][2]),
                        _stdev(self.energy_data['vdwaals'][2]), _stdev(self.energy_data['vdwaals'][2])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][2]),
                        _stdev(self.energy_data['eel'][2]), _stdev(self.energy_data['eel'][2])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][2]),
                           _stdev(self.energy_data['1-4vdw'][2]), _stdev(self.energy_data['1-4vdw'][2])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][2]),
                           _stdev(self.energy_data['1-4eel'][2]), _stdev(self.energy_data['1-4eel'][2])/sqrtframes)

         printstring += 'EGB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['egb'][2]),
                        _stdev(self.energy_data['egb'][2]), _stdev(self.energy_data['egb'][2])/sqrtframes) + \
         'ESURF          %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['esurf'][2]),
                        _stdev(self.energy_data['esurf'][2]), _stdev(self.energy_data['esurf'][2])/sqrtframes) + \
         'ESCF           %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['escf'][2]),
                        _stdev(self.energy_data['escf'][2]), _stdev(self.energy_data['escf'][2])/sqrtframes) + \
         'G gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][2]),
                        _stdev(self.energy_data['gas'][2]), _stdev(self.energy_data['gas'][2])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][2]),
                        _stdev(self.energy_data['solv'][2]), _stdev(self.energy_data['solv'][2])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][2]),
                        _stdev(self.energy_data['total'][2]), _stdev(self.energy_data['total'][2])/sqrtframes)
      # now do the differences. If we're using multiple trajectories, we'll have to propagate stdevs
      if not self.stability and self.one_trajectory:
         sqrtframes = math.sqrt(len(self.energy_data['bond'][3]))
         printstring += 'Differences (Complex - Receptor - Ligand):' + \
         '\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][3]), 
                           _stdev(self.energy_data['bond'][3]), _stdev(self.energy_data['bond'][3])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][3]), 
                           _stdev(self.energy_data['angle'][3]), _stdev(self.energy_data['angle'][3])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][3]), 
                           _stdev(self.energy_data['dihedral'][3]), _stdev(self.energy_data['dihedral'][3])/sqrtframes)
         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][3]),
                        _stdev(self.energy_data['vdwaals'][3]), _stdev(self.energy_data['vdwaals'][3])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][3]),
                        _stdev(self.energy_data['eel'][3]), _stdev(self.energy_data['eel'][3])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][3]),
                           _stdev(self.energy_data['1-4vdw'][3]), _stdev(self.energy_data['1-4vdw'][3])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][3]),
                           _stdev(self.energy_data['1-4eel'][3]), _stdev(self.energy_data['1-4eel'][3])/sqrtframes)

         printstring += 'EGB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['egb'][3]),
                        _stdev(self.energy_data['egb'][3]), _stdev(self.energy_data['egb'][3])/sqrtframes) + \
         'ESURF          %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['esurf'][3]),
                        _stdev(self.energy_data['esurf'][3]), _stdev(self.energy_data['esurf'][3])/sqrtframes) + \
         'ESCF           %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['escf'][3]),
                        _stdev(self.energy_data['escf'][3]), _stdev(self.energy_data['escf'][3])/sqrtframes) + \
         'DELTA G gas    %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][3]),
                        _stdev(self.energy_data['gas'][3]), _stdev(self.energy_data['gas'][3])/sqrtframes) + \
         'DELTA G solv   %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][3]),
                        _stdev(self.energy_data['solv'][3]), _stdev(self.energy_data['solv'][3])/sqrtframes) + \
         '\n DELTA G binding = %19.4f +/- %11.4f %22.4f\n\n\n' % (_avg(self.energy_data['total'][3]),
                        _stdev(self.energy_data['total'][3]), _stdev(self.energy_data['total'][3])/sqrtframes)
      elif not self.stability:
         # maybe not the best. sqrtframes is the sqrt of the minimum number of frames for com/rec/lig
         sqrtframes = min([len(self.energy_data['bond'][0]), len(self.energy_data['bond'][1]), \
                       len(self.energy_data['bond'][2])])
         bondstdev = math.sqrt(_stdev(self.energy_data['bond'][0]) ** 2 +
            _stdev(self.energy_data['bond'][1]) ** 2 + _stdev(self.energy_data['bond'][2]) ** 2)
         anglestdev = math.sqrt(_stdev(self.energy_data['angle'][0]) ** 2 +
            _stdev(self.energy_data['angle'][1]) ** 2 + _stdev(self.energy_data['angle'][2]) ** 2)
         dihedralstdev = math.sqrt(_stdev(self.energy_data['dihedral'][0]) ** 2 +
            _stdev(self.energy_data['dihedral'][1]) ** 2 + _stdev(self.energy_data['dihedral'][2]) ** 2)
         vdwaalsstdev = math.sqrt(_stdev(self.energy_data['vdwaals'][0]) ** 2 +
            _stdev(self.energy_data['vdwaals'][1]) ** 2 + _stdev(self.energy_data['vdwaals'][2]) ** 2)
         eelstdev = math.sqrt(_stdev(self.energy_data['eel'][0]) ** 2 +
            _stdev(self.energy_data['eel'][1]) ** 2 + _stdev(self.energy_data['eel'][2]) ** 2)
         vdwofstdev = math.sqrt(_stdev(self.energy_data['vdwaals'][0]) ** 2 +
            _stdev(self.energy_data['1-4vdw'][1]) ** 2 + _stdev(self.energy_data['vdwaals'][2]) ** 2)
         eelofstdev = math.sqrt(_stdev(self.energy_data['1-4eel'][0]) ** 2 +
            _stdev(self.energy_data['1-4eel'][1]) ** 2 + _stdev(self.energy_data['1-4eel'][2]) ** 2)
         egbstdev = math.sqrt(_stdev(self.energy_data['egb'][0]) ** 2 +
            _stdev(self.energy_data['egb'][1]) ** 2 + _stdev(self.energy_data['egb'][2]) ** 2)
         esurfstdev = math.sqrt(_stdev(self.energy_data['esurf'][0]) ** 2 +
            _stdev(self.energy_data['esurf'][1]) ** 2 + _stdev(self.energy_data['esurf'][2]) ** 2)
         escfstdev = math.sqrt(_stdev(self.energy_data['escf'][0]) ** 2 +
            _stdev(self.energy_data['escf'][1]) ** 2 + _stdev(self.energy_data['escf'][2]) ** 2)
         gasstdev = math.sqrt(_stdev(self.energy_data['gas'][0]) ** 2 +
            _stdev(self.energy_data['gas'][1]) ** 2 + _stdev(self.energy_data['gas'][2]) ** 2)
         solvstdev = math.sqrt(_stdev(self.energy_data['solv'][0]) ** 2 +
            _stdev(self.energy_data['solv'][1]) ** 2 + _stdev(self.energy_data['solv'][2]) ** 2)
         totalstdev = math.sqrt(_stdev(self.energy_data['total'][0]) ** 2 +
            _stdev(self.energy_data['total'][1]) ** 2 + _stdev(self.energy_data['total'][2]) ** 2)
         printstring += 'Differences (Complex - Receptor - Ligand):' + \
         '\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n' + \
         'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][0]) -
            _avg(self.energy_data['bond'][1])- _avg(self.energy_data['bond'][2]), bondstdev, 
            bondstdev/sqrtframes) + \
         'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][0]) -
            _avg(self.energy_data['angle'][1]) - _avg(self.energy_data['angle'][2]), anglestdev, 
            anglestdev / sqrtframes) + \
         'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][0]) -
            _avg(self.energy_data['dihedral'][1])- _avg(self.energy_data['dihedral'][2]), dihedralstdev, 
            dihedralstdev / sqrtframes) + \
         'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][0]) -
            _avg(self.energy_data['vdwaals'][1])- _avg(self.energy_data['vdwaals'][2]), vdwaalsstdev, 
            vdwaalsstdev / sqrtframes) + \
         '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][0]) -
            _avg(self.energy_data['1-4vdw'][1])- _avg(self.energy_data['1-4vdw'][2]), vdwofstdev, 
            vdwofstdev / sqrtframes) + \
         '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][0]) -
            _avg(self.energy_data['1-4eel'][1])- _avg(self.energy_data['1-4eel'][2]), eelofstdev, 
            eelofstdev / sqrtframes) + \
         'EGB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['egb'][0]) -
            _avg(self.energy_data['egb'][1])- _avg(self.energy_data['egb'][2]), egbstdev, 
            egbstdev / sqrtframes) + \
         'ESURF          %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['esurf'][0]) -
            _avg(self.energy_data['esurf'][1])- _avg(self.energy_data['esurf'][2]), esurfstdev, 
            esurfstdev / sqrtframes) + \
         'ESCF           %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['escf'][0]) -
            _avg(self.energy_data['escf'][1])- _avg(self.energy_data['escf'][2]), escfstdev, 
            escfstdev / sqrtframes) + \
         'DELTA G gas    %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][0]) -
            _avg(self.energy_data['gas'][1])- _avg(self.energy_data['gas'][2]), gasstdev, 
            gasstdev / sqrtframes) + \
         'DELTA G solv   %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][0]) -
            _avg(self.energy_data['solv'][1])- _avg(self.energy_data['solv'][2]), solvstdev, 
            solvstdev / sqrtframes) + \
         '\n DELTA G binding = %19.4f +/- %11.4f %22.4f\n' % (_avg(self.energy_data['total'][0]) - 
            _avg(self.energy_data['total'][1]) - _avg(self.energy_data['total'][2]), totalstdev,
            totalstdev / sqrtframes)

      return printstring

#==============================

   def AvgStdev(self):
      """ Returns the average and standard deviation of the total binding free energy """
      if self.stability:
         return _avg(self.energy_data['total'][0]), _stdev(self.energy_data['total'][0])

      if self.one_trajectory:
         return _avg(self.energy_data['total'][3]), _stdev(self.energy_data['total'][3])

      avg = _avg(self.energy_data['total'][0]) - _avg(self.energy_data['total'][1]) - _avg(self.energy_data['total'][2])
      stdev = math.sqrt(_stdev(self.energy_data['total'][0]) ** 2 + _stdev(self.energy_data['total'][1]) ** 2 + 
                        _stdev(self.energy_data['total'][2]) ** 2)

      return avg, stdev

#==============================

   def _read(self):
      """ Parses the output files and fills the data arrays """
      # dictionary of empty arrays to be filled
      self.energy_data = {'bond' : [[],[],[],[]], 'angle' : [[],[],[],[]], 'dihedral' : [[],[],[],[]],
                          'vdwaals' : [[],[],[],[]], 'eel' : [[],[],[],[]], '1-4vdw' : [[],[],[],[]], 
                          '1-4eel' : [[],[],[],[]], 'gas' : [[],[],[],[]], 'egb' : [[],[],[],[]], 
                          'esurf' : [[],[],[],[]], 'solv' : [[],[],[],[]], 'escf' : [[],[],[],[]],
                          'total' : [[],[],[],[]]}
      com = open(self.comfile, 'r')
      if not self.stability:
         rec = open(self.recfile, 'r')
         lig = open(self.ligfile, 'r')

      try:
         self._getenergies(com)
         com.close()
         if not self.stability:
            self._getenergies(rec, index=1)
            self._getenergies(lig, index=2)
            rec.close()
            lig.close()
         if self.one_trajectory:
            self._assemblediffs()

         return True
      except:
         print >> sys.stderr, 'Error: Could not parse QM/MM GB output files! Check for missing values (*********).'
         return False

#==============================

   def _getenergies(self, file, index=0):
      """ Actually parses output files and extracts energy terms """
      rawline = file.readline()  # get first line

      while len(rawline) != 0:
         
         if rawline[0:5] == ' BOND':
            # split the line into individual components
            words = rawline.split()
            # bond, angle, and dihedral parameters are on this line
            bond    = float(words[2])
            angle   = float(words[5])
            dihed   = float(words[8])
            # split the next line into components
            words = file.readline().split()
            # vdwaals, eel, and egb are on this line
            vdwaals = float(words[2])
            eel     = float(words[5])
            egb     = float(words[8])
            # split the next line into components
            words = file.readline().split()
            # 1-4 vdw and 1-4 eel are on this line
            vdw1_4  = float(words[3])
            eel1_4  = float(words[7])
            # split the next line into components
            words = file.readline().split()
            # escf is on this line if it's qm/mm
            if words[0] != 'ESURF':
               escf    = float(words[2])
               # split the next line into components
               words = file.readline().split()
            else:
               escf = 0
            # esurf is on this line
            esurf   = float(words[2])
            # now load these values into the data arrays
            self.energy_data['bond'][index].append(bond)
            self.energy_data['angle'][index].append(angle)
            self.energy_data['dihedral'][index].append(dihed)
            self.energy_data['vdwaals'][index].append(vdwaals)
            self.energy_data['eel'][index].append(eel)
            self.energy_data['egb'][index].append(egb)
            self.energy_data['1-4vdw'][index].append(vdw1_4)
            self.energy_data['1-4eel'][index].append(eel1_4)
            self.energy_data['esurf'][index].append(esurf)
            self.energy_data['escf'][index].append(escf)
            self.energy_data['gas'][index].append(bond + angle + dihed + vdwaals + eel + vdw1_4 + eel1_4)
            self.energy_data['solv'][index].append(egb + esurf)
            self.energy_data['total'][index].append(bond + angle + dihed + vdwaals + eel + vdw1_4 + eel1_4 + egb + esurf + escf)

         rawline = file.readline() # read the next line

#==============================

   def _assemblediffs(self):
      """ Build the difference arrays """
      self.energy_data['bond'][3]     = []
      self.energy_data['angle'][3]    = []
      self.energy_data['dihedral'][3] = []
      self.energy_data['vdwaals'][3]  = []
      self.energy_data['eel'][3]      = []
      self.energy_data['egb'][3]      = []
      self.energy_data['1-4vdw'][3]   = []
      self.energy_data['1-4eel'][3]   = []
      self.energy_data['esurf'][3]    = []
      self.energy_data['gas'][3]      = []
      self.energy_data['solv'][3]     = []
      self.energy_data['escf'][3]     = []
      self.energy_data['total'][3]    = []
      if self.stability:
         for i in range(len(self.energy_data['bond'][0])):
            self.energy_data['bond'][3].append(self.energy_data['bond'][0][i])
            self.energy_data['angle'][3].append(self.energy_data['angle'][0][i])
            self.energy_data['dihedral'][3].append(self.energy_data['dihedral'][0][i])
            self.energy_data['vdwaals'][3].append(self.energy_data['vdwaals'][0][i])
            self.energy_data['eel'][3].append(self.energy_data['eel'][0][i])
            self.energy_data['egb'][3].append(self.energy_data['egb'][0][i])
            self.energy_data['1-4vdw'][3].append(self.energy_data['1-4vdw'][0][i])
            self.energy_data['1-4eel'][3].append(self.energy_data['1-4eel'][0][i])
            self.energy_data['esurf'][3].append(self.energy_data['esurf'][0][i])
            self.energy_data['gas'][3].append(self.energy_data['gas'][0][i])
            self.energy_data['solv'][3].append(self.energy_data['solv'][0][i])
            self.energy_data['escf'][3].append(self.energy_data['escf'][0][i])
            self.energy_data['total'][3].append(self.energy_data['total'][0][i])
      else:
         for i in range(len(self.energy_data['bond'][0])):
            self.energy_data['bond'][3].append(self.energy_data['bond'][0][i] - self.energy_data['bond'][1][i] -
                                               self.energy_data['bond'][2][i])
            self.energy_data['angle'][3].append(self.energy_data['angle'][0][i] - self.energy_data['angle'][1][i] -
                                                self.energy_data['angle'][2][i])
            self.energy_data['dihedral'][3].append(self.energy_data['dihedral'][0][i] - self.energy_data['dihedral'][1][i] -
                                                self.energy_data['dihedral'][2][i])
            self.energy_data['vdwaals'][3].append(self.energy_data['vdwaals'][0][i] - self.energy_data['vdwaals'][1][i] -
                                                  self.energy_data['vdwaals'][2][i])
            self.energy_data['eel'][3].append(self.energy_data['eel'][0][i] - self.energy_data['eel'][1][i] -
                                              self.energy_data['eel'][2][i])
            self.energy_data['egb'][3].append(self.energy_data['egb'][0][i] - self.energy_data['egb'][1][i] -
                                              self.energy_data['egb'][2][i])
            self.energy_data['1-4vdw'][3].append(self.energy_data['1-4vdw'][0][i] - self.energy_data['1-4vdw'][1][i] -
                                                 self.energy_data['1-4vdw'][2][i])
            self.energy_data['1-4eel'][3].append(self.energy_data['1-4eel'][0][i] - self.energy_data['1-4eel'][1][i] -
                                                 self.energy_data['1-4eel'][2][i])
            self.energy_data['esurf'][3].append(self.energy_data['esurf'][0][i] - self.energy_data['esurf'][1][i] -
                                                self.energy_data['esurf'][2][i])
            self.energy_data['gas'][3].append(self.energy_data['gas'][0][i] - self.energy_data['gas'][1][i] -
                                              self.energy_data['gas'][2][i])
            self.energy_data['solv'][3].append(self.energy_data['solv'][0][i] - self.energy_data['solv'][1][i] -
                                               self.energy_data['solv'][2][i])
            self.energy_data['escf'][3].append(self.energy_data['escf'][0][i] - self.energy_data['escf'][1][i] -
                                               self.energy_data['escf'][2][i])
            self.energy_data['total'][3].append(self.energy_data['total'][0][i] - self.energy_data['total'][1][i] -
                                                self.energy_data['total'][2][i])

#==============================

   def _consistent(self, terms):
      """ Determine if the internal potential terms are consistent """
      global TOLERANCE

      if not self.valid:
         return False

      for i in range(len(self.energy_data['bond'][0])):
         for item in terms:
            if abs(self.energy_data[item][3][i]) > TOLERANCE:
               return False

      return True

#==============================

   def _removeFromTotal(self, terms):
      """ Removes unprinted energy contributions from the gas and total energy printouts """
      for i in range(len(self.energy_data['bond'][0])):
         for item in terms:
            self.energy_data['gas'][0][i] -= self.energy_data[item][0][i]
            self.energy_data['gas'][1][i] -= self.energy_data[item][1][i]
            self.energy_data['gas'][2][i] -= self.energy_data[item][2][i]
            self.energy_data['total'][0][i] -= self.energy_data[item][0][i]
            self.energy_data['total'][1][i] -= self.energy_data[item][1][i]
            self.energy_data['total'][2][i] -= self.energy_data[item][2][i]

      self._assemblediffs() # re-remove these from the differences

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class GBout:
   """ Class for GB output """

#==============================

   def __init__(self, one_traj=True, verb=0, file1='', file2='', file3='', desc=''):
      """ Initializes GBout object with data from passed files, etc."""
      self.one_trajectory = one_traj
      self.verbose = verb
      self.description = desc
      self.comfile = file1
      if file2 == '' and file3 == '':
         self.stability = True
      else:
         self.stability = False
         self.recfile = file2
         self.ligfile = file3

      if self.stability:
         self.exists = os.path.exists(file1)
      else:
         self.exists = os.path.exists(file1) and os.path.exists(file2) and os.path.exists(file3)

      self.valid = self.exists and self._read()

#==============================

   def __str__(self):
      """ String representation of GB data -- final output file format """
      if not self.valid:  return "Problem parsing GB output!\n"

      printstring = '%s' % self.description

      consistent_1_4 = self.one_trajectory and self._consistent(['1-4vdw','1-4eel'])
      consistent_int = self.one_trajectory and self._consistent(['bond','angle','dihedral'])

      print_1_4 = self.verbose == 2 or not consistent_1_4
      print_int = self.verbose == 2 or not consistent_int

      if not print_1_4:  self._removeFromTotal(['1-4vdw','1-4eel'])
      if not print_int:  self._removeFromTotal(['bond','angle','dihedral'])

      if not consistent_int and self.one_trajectory and not self.stability:
         printstring += '\n\nWARNING: INCONSISTENCIES EXIST WITHIN INTERNAL POTENTIAL TERMS (BOND, ANGLE, AND/OR DIHED).' + \
         '\n         CHECK YOUR INPUT FILES AND SYSTEM SETUP. THESE RESULTS MAY NOT BE RELIABLE (check differences)!\n'

      if not consistent_1_4 and self.one_trajectory and not self.stability:
         printstring += '\n\nWARNING: INCONSISTENCIES EXIST WITHIN 1-4 NON-BONDED TERMS.\n' + \
         '         CHECK YOUR INPUT FILES AND SYSTEM SETUP. THESE RESULTS MAY NOT BE RELIABLE (check differences)!\n\n' + \
         'The most common cause of this is inconsistent charge definitions across topology files.\n'

      # add the Complex contribution if we asked to print it
      if self.verbose >= 1 or self.stability:
         sqrtframes = math.sqrt(len(self.energy_data['bond'][0]))
         printstring += 'Complex:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][0]),
                           _stdev(self.energy_data['bond'][0]), _stdev(self.energy_data['bond'][0])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][0]),
                           _stdev(self.energy_data['angle'][0]), _stdev(self.energy_data['angle'][0])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][0]),
                           _stdev(self.energy_data['dihedral'][0]), _stdev(self.energy_data['dihedral'][0])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][0]),
                        _stdev(self.energy_data['vdwaals'][0]), _stdev(self.energy_data['vdwaals'][0])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][0]),
                        _stdev(self.energy_data['eel'][0]), _stdev(self.energy_data['eel'][0])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][0]),
                           _stdev(self.energy_data['1-4vdw'][0]), _stdev(self.energy_data['1-4vdw'][0])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][0]),
                           _stdev(self.energy_data['1-4eel'][0]), _stdev(self.energy_data['1-4eel'][0])/sqrtframes)

         printstring += 'EGB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['egb'][0]),
                        _stdev(self.energy_data['egb'][0]), _stdev(self.energy_data['egb'][0])/sqrtframes) + \
         'ESURF          %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['esurf'][0]),
                        _stdev(self.energy_data['esurf'][0]), _stdev(self.energy_data['esurf'][0])/sqrtframes) + \
         'G gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][0]),
                        _stdev(self.energy_data['gas'][0]), _stdev(self.energy_data['gas'][0])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][0]),
                        _stdev(self.energy_data['solv'][0]), _stdev(self.energy_data['solv'][0])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][0]),
                        _stdev(self.energy_data['total'][0]), _stdev(self.energy_data['total'][0])/sqrtframes)
      # add the Receptor contribution if we asked to print it
      if self.verbose >= 1 and not self.stability:
         sqrtframes = math.sqrt(len(self.energy_data['bond'][1]))
         printstring += 'Receptor:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][1]),
                           _stdev(self.energy_data['bond'][1]), _stdev(self.energy_data['bond'][1])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][1]),
                           _stdev(self.energy_data['angle'][1]), _stdev(self.energy_data['angle'][1])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][1]),
                           _stdev(self.energy_data['dihedral'][1]), _stdev(self.energy_data['dihedral'][1])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][1]),
                        _stdev(self.energy_data['vdwaals'][1]), _stdev(self.energy_data['vdwaals'][1])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][1]),
                        _stdev(self.energy_data['eel'][1]), _stdev(self.energy_data['eel'][1])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][1]),
                           _stdev(self.energy_data['1-4vdw'][1]), _stdev(self.energy_data['1-4vdw'][1])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][1]),
                           _stdev(self.energy_data['1-4eel'][1]), _stdev(self.energy_data['1-4eel'][1])/sqrtframes)

         printstring += 'EGB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['egb'][1]),
                        _stdev(self.energy_data['egb'][1]), _stdev(self.energy_data['egb'][1])/sqrtframes) + \
         'ESURF          %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['esurf'][1]),
                        _stdev(self.energy_data['esurf'][1]), _stdev(self.energy_data['esurf'][1])/sqrtframes) + \
         'G gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][1]),
                        _stdev(self.energy_data['gas'][1]), _stdev(self.energy_data['gas'][1])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][1]),
                        _stdev(self.energy_data['solv'][1]), _stdev(self.energy_data['solv'][1])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][1]),
                        _stdev(self.energy_data['total'][1]), _stdev(self.energy_data['total'][1])/sqrtframes)
      # add the ligand contribution if we asked to print it
         sqrtframes = math.sqrt(len(self.energy_data['bond'][2]))
         printstring += 'Ligand:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][2]),
                           _stdev(self.energy_data['bond'][2]), _stdev(self.energy_data['bond'][2])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][2]),
                           _stdev(self.energy_data['angle'][2]), _stdev(self.energy_data['angle'][2])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][2]),
                           _stdev(self.energy_data['dihedral'][2]), _stdev(self.energy_data['dihedral'][2])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][2]),
                        _stdev(self.energy_data['vdwaals'][2]), _stdev(self.energy_data['vdwaals'][2])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][2]),
                        _stdev(self.energy_data['eel'][2]), _stdev(self.energy_data['eel'][2])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][2]),
                           _stdev(self.energy_data['1-4vdw'][2]), _stdev(self.energy_data['1-4vdw'][2])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][2]),
                           _stdev(self.energy_data['1-4eel'][2]), _stdev(self.energy_data['1-4eel'][2])/sqrtframes)

         printstring += 'EGB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['egb'][2]),
                        _stdev(self.energy_data['egb'][2]), _stdev(self.energy_data['egb'][2])/sqrtframes) + \
         'ESURF          %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['esurf'][2]),
                        _stdev(self.energy_data['esurf'][2]), _stdev(self.energy_data['esurf'][2])/sqrtframes) + \
         'G gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][2]),
                        _stdev(self.energy_data['gas'][2]), _stdev(self.energy_data['gas'][2])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][2]),
                        _stdev(self.energy_data['solv'][2]), _stdev(self.energy_data['solv'][2])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][2]),
                        _stdev(self.energy_data['total'][2]), _stdev(self.energy_data['total'][2])/sqrtframes)
      # now do the differences. If we're using multiple trajectories, we'll have to propagate stdevs
      if not self.stability and self.one_trajectory:
         sqrtframes = math.sqrt(len(self.energy_data['bond'][3]))
         printstring += 'Differences (Complex - Receptor - Ligand):' + \
         '\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][3]), 
                           _stdev(self.energy_data['bond'][3]), _stdev(self.energy_data['bond'][3])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][3]), 
                           _stdev(self.energy_data['angle'][3]), _stdev(self.energy_data['angle'][3])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][3]), 
                           _stdev(self.energy_data['dihedral'][3]), _stdev(self.energy_data['dihedral'][3])/sqrtframes)
         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][3]),
                        _stdev(self.energy_data['vdwaals'][3]), _stdev(self.energy_data['vdwaals'][3])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][3]),
                        _stdev(self.energy_data['eel'][3]), _stdev(self.energy_data['eel'][3])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][3]),
                           _stdev(self.energy_data['1-4vdw'][3]), _stdev(self.energy_data['1-4vdw'][3])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][3]),
                           _stdev(self.energy_data['1-4eel'][3]), _stdev(self.energy_data['1-4eel'][3])/sqrtframes)

         printstring += 'EGB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['egb'][3]),
                        _stdev(self.energy_data['egb'][3]), _stdev(self.energy_data['egb'][3])/sqrtframes) + \
         'ESURF          %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['esurf'][3]),
                        _stdev(self.energy_data['esurf'][3]), _stdev(self.energy_data['esurf'][3])/sqrtframes) + \
         'DELTA G gas    %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][3]),
                        _stdev(self.energy_data['gas'][3]), _stdev(self.energy_data['gas'][3])/sqrtframes) + \
         'DELTA G solv   %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][3]),
                        _stdev(self.energy_data['solv'][3]), _stdev(self.energy_data['solv'][3])/sqrtframes) + \
         '\n DELTA G binding = %19.4f +/- %11.4f %22.4f\n\n\n' % (_avg(self.energy_data['total'][3]),
                        _stdev(self.energy_data['total'][3]), _stdev(self.energy_data['total'][3])/sqrtframes)
      elif not self.stability:
         # maybe not the best. sqrtframes is the sqrt of the minimum number of frames for com/rec/lig
         sqrtframes = min([len(self.energy_data['bond'][0]), len(self.energy_data['bond'][1]), \
                       len(self.energy_data['bond'][2])])
         bondstdev = math.sqrt(_stdev(self.energy_data['bond'][0]) ** 2 +
            _stdev(self.energy_data['bond'][1]) ** 2 + _stdev(self.energy_data['bond'][2]) ** 2)
         anglestdev = math.sqrt(_stdev(self.energy_data['angle'][0]) ** 2 +
            _stdev(self.energy_data['angle'][1]) ** 2 + _stdev(self.energy_data['angle'][2]) ** 2)
         dihedralstdev = math.sqrt(_stdev(self.energy_data['dihedral'][0]) ** 2 +
            _stdev(self.energy_data['dihedral'][1]) ** 2 + _stdev(self.energy_data['dihedral'][2]) ** 2)
         vdwaalsstdev = math.sqrt(_stdev(self.energy_data['vdwaals'][0]) ** 2 +
            _stdev(self.energy_data['vdwaals'][1]) ** 2 + _stdev(self.energy_data['vdwaals'][2]) ** 2)
         eelstdev = math.sqrt(_stdev(self.energy_data['eel'][0]) ** 2 +
            _stdev(self.energy_data['eel'][1]) ** 2 + _stdev(self.energy_data['eel'][2]) ** 2)
         vdwofstdev = math.sqrt(_stdev(self.energy_data['vdwaals'][0]) ** 2 +
            _stdev(self.energy_data['1-4vdw'][1]) ** 2 + _stdev(self.energy_data['vdwaals'][2]) ** 2)
         eelofstdev = math.sqrt(_stdev(self.energy_data['1-4eel'][0]) ** 2 +
            _stdev(self.energy_data['1-4eel'][1]) ** 2 + _stdev(self.energy_data['1-4eel'][2]) ** 2)
         egbstdev = math.sqrt(_stdev(self.energy_data['egb'][0]) ** 2 +
            _stdev(self.energy_data['egb'][1]) ** 2 + _stdev(self.energy_data['egb'][2]) ** 2)
         esurfstdev = math.sqrt(_stdev(self.energy_data['esurf'][0]) ** 2 +
            _stdev(self.energy_data['esurf'][1]) ** 2 + _stdev(self.energy_data['esurf'][2]) ** 2)
         gasstdev = math.sqrt(_stdev(self.energy_data['gas'][0]) ** 2 +
            _stdev(self.energy_data['gas'][1]) ** 2 + _stdev(self.energy_data['gas'][2]) ** 2)
         solvstdev = math.sqrt(_stdev(self.energy_data['solv'][0]) ** 2 +
            _stdev(self.energy_data['solv'][1]) ** 2 + _stdev(self.energy_data['solv'][2]) ** 2)
         totalstdev = math.sqrt(_stdev(self.energy_data['total'][0]) ** 2 +
            _stdev(self.energy_data['total'][1]) ** 2 + _stdev(self.energy_data['total'][2]) ** 2)
         printstring += 'Differences (Complex - Receptor - Ligand):' + \
         '\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n' + \
         'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][0]) -
            _avg(self.energy_data['bond'][1])- _avg(self.energy_data['bond'][2]), bondstdev, 
            bondstdev/sqrtframes) + \
         'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][0]) -
            _avg(self.energy_data['angle'][1]) - _avg(self.energy_data['angle'][2]), anglestdev, 
            anglestdev / sqrtframes) + \
         'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][0]) -
            _avg(self.energy_data['dihedral'][1])- _avg(self.energy_data['dihedral'][2]), dihedralstdev, 
            dihedralstdev / sqrtframes) + \
         'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][0]) -
            _avg(self.energy_data['vdwaals'][1])- _avg(self.energy_data['vdwaals'][2]), vdwaalsstdev, 
            vdwaalsstdev / sqrtframes) + \
         '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][0]) -
            _avg(self.energy_data['1-4vdw'][1])- _avg(self.energy_data['1-4vdw'][2]), vdwofstdev, 
            vdwofstdev / sqrtframes) + \
         '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][0]) -
            _avg(self.energy_data['1-4eel'][1])- _avg(self.energy_data['1-4eel'][2]), eelofstdev, 
            eelofstdev / sqrtframes) + \
         'EGB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['egb'][0]) -
            _avg(self.energy_data['egb'][1])- _avg(self.energy_data['egb'][2]), egbstdev, 
            egbstdev / sqrtframes) + \
         'ESURF          %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['esurf'][0]) -
            _avg(self.energy_data['esurf'][1])- _avg(self.energy_data['esurf'][2]), esurfstdev, 
            esurfstdev / sqrtframes) + \
         'DELTA G gas    %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][0]) -
            _avg(self.energy_data['gas'][1])- _avg(self.energy_data['gas'][2]), gasstdev, 
            gasstdev / sqrtframes) + \
         'DELTA G solv   %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][0]) -
            _avg(self.energy_data['solv'][1])- _avg(self.energy_data['solv'][2]), solvstdev, 
            solvstdev / sqrtframes) + \
         '\n DELTA G binding = %19.4f +/- %11.4f %22.4f\n' % (_avg(self.energy_data['total'][0]) - 
            _avg(self.energy_data['total'][1]) - _avg(self.energy_data['total'][2]), totalstdev,
            totalstdev / sqrtframes)

      return printstring

#==============================

   def AvgStdev(self):
      """ Returns the average and standard deviations of the total GB binding data """
      if self.stability:
         return _avg(self.energy_data['total'][0]), _stdev(self.energy_data['total'][0])

      if self.one_trajectory:
         return _avg(self.energy_data['total'][3]), _stdev(self.energy_data['total'][3])

      avg = _avg(self.energy_data['total'][0]) - _avg(self.energy_data['total'][1]) - _avg(self.energy_data['total'][2])
      stdev = math.sqrt(_stdev(self.energy_data['total'][0]) ** 2 + _stdev(self.energy_data['total'][1]) ** 2 + 
                        _stdev(self.energy_data['total'][2]) ** 2)

      return avg, stdev

#==============================

   def _read(self):
      """ Parses the GB output files and fills arrays """
      # dictionary of empty arrays to be filled
      self.energy_data = {'bond' : [[],[],[],[]], 'angle' : [[],[],[],[]], 'dihedral' : [[],[],[],[]],
                          'vdwaals' : [[],[],[],[]], 'eel' : [[],[],[],[]], '1-4vdw' : [[],[],[],[]], 
                          '1-4eel' : [[],[],[],[]], 'gas' : [[],[],[],[]], 'egb' : [[],[],[],[]], 
                          'esurf' : [[],[],[],[]], 'solv' : [[],[],[],[]], 'total' : [[],[],[],[]]}
      com = open(self.comfile, 'r')
      if not self.stability:
         rec = open(self.recfile, 'r')
         lig = open(self.ligfile, 'r')

      try:
         self._getenergies(com)
         com.close()
         if not self.stability:
            self._getenergies(rec, index=1)
            self._getenergies(lig, index=2)
            rec.close()
            lig.close()
         if self.one_trajectory:
            self._assemblediffs()

         return True
      except:
         print >> sys.stderr, 'Error: Could not parse GB output files! Check for missing values (*********).'
         return False

#==============================

   def _getenergies(self, file, index=0):
      """ Actually parses the output files and extracts energy terms """
      rawline = file.readline()  # get first line

      while len(rawline) != 0:
         
         if rawline[0:5] == ' BOND':
            # split the line into individual components
            words = rawline.split()
            # bond, angle, and dihedral parameters are on this line
            bond    = float(words[2])
            angle   = float(words[5])
            dihed   = float(words[8])
            # split the next line into components
            words = file.readline().split()
            # vdwaals, eel, and egb are on this line
            vdwaals = float(words[2])
            eel     = float(words[5])
            egb     = float(words[8])
            # split the next line into components
            words = file.readline().split()
            # 1-4 vdw and 1-4 eel are on this line
            vdw1_4  = float(words[3])
            eel1_4  = float(words[7])
            # split the next line into components
            words = file.readline().split()
            # esurf is on this line
            esurf   = float(words[2])
            # now load these values into the data arrays
            self.energy_data['bond'][index].append(bond)
            self.energy_data['angle'][index].append(angle)
            self.energy_data['dihedral'][index].append(dihed)
            self.energy_data['vdwaals'][index].append(vdwaals)
            self.energy_data['eel'][index].append(eel)
            self.energy_data['egb'][index].append(egb)
            self.energy_data['1-4vdw'][index].append(vdw1_4)
            self.energy_data['1-4eel'][index].append(eel1_4)
            self.energy_data['esurf'][index].append(esurf)
            self.energy_data['gas'][index].append(bond + angle + dihed + vdwaals + eel + vdw1_4 + eel1_4)
            self.energy_data['solv'][index].append(egb + esurf)
            self.energy_data['total'][index].append(bond + angle + dihed + vdwaals + eel + vdw1_4 + eel1_4 + egb + esurf)

         rawline = file.readline() # read the next line

#==============================

   def _assemblediffs(self):
      """ Builds the difference arrays from the data """
      self.energy_data['bond'][3]     = []
      self.energy_data['angle'][3]    = []
      self.energy_data['dihedral'][3] = []
      self.energy_data['vdwaals'][3]  = []
      self.energy_data['eel'][3]      = []
      self.energy_data['egb'][3]      = []
      self.energy_data['1-4vdw'][3]   = []
      self.energy_data['1-4eel'][3]   = []
      self.energy_data['esurf'][3]    = []
      self.energy_data['gas'][3]      = []
      self.energy_data['solv'][3]     = []
      self.energy_data['total'][3]    = []
      if self.stability:
         for i in range(len(self.energy_data['bond'][0])):
            self.energy_data['bond'][3].append(self.energy_data['bond'][0][i])
            self.energy_data['angle'][3].append(self.energy_data['angle'][0][i])
            self.energy_data['dihedral'][3].append(self.energy_data['dihedral'][0][i])
            self.energy_data['vdwaals'][3].append(self.energy_data['vdwaals'][0][i])
            self.energy_data['eel'][3].append(self.energy_data['eel'][0][i])
            self.energy_data['egb'][3].append(self.energy_data['egb'][0][i])
            self.energy_data['1-4vdw'][3].append(self.energy_data['1-4vdw'][0][i])
            self.energy_data['1-4eel'][3].append(self.energy_data['1-4eel'][0][i])
            self.energy_data['esurf'][3].append(self.energy_data['esurf'][0][i])
            self.energy_data['gas'][3].append(self.energy_data['gas'][0][i])
            self.energy_data['solv'][3].append(self.energy_data['solv'][0][i])
            self.energy_data['total'][3].append(self.energy_data['total'][0][i])
      else:
         for i in range(len(self.energy_data['bond'][0])):
            self.energy_data['bond'][3].append(self.energy_data['bond'][0][i] - self.energy_data['bond'][1][i] -
                                               self.energy_data['bond'][2][i])
            self.energy_data['angle'][3].append(self.energy_data['angle'][0][i] - self.energy_data['angle'][1][i] -
                                                self.energy_data['angle'][2][i])
            self.energy_data['dihedral'][3].append(self.energy_data['dihedral'][0][i] - self.energy_data['dihedral'][1][i] -
                                                self.energy_data['dihedral'][2][i])
            self.energy_data['vdwaals'][3].append(self.energy_data['vdwaals'][0][i] - self.energy_data['vdwaals'][1][i] -
                                                  self.energy_data['vdwaals'][2][i])
            self.energy_data['eel'][3].append(self.energy_data['eel'][0][i] - self.energy_data['eel'][1][i] -
                                              self.energy_data['eel'][2][i])
            self.energy_data['egb'][3].append(self.energy_data['egb'][0][i] - self.energy_data['egb'][1][i] -
                                              self.energy_data['egb'][2][i])
            self.energy_data['1-4vdw'][3].append(self.energy_data['1-4vdw'][0][i] - self.energy_data['1-4vdw'][1][i] -
                                                 self.energy_data['1-4vdw'][2][i])
            self.energy_data['1-4eel'][3].append(self.energy_data['1-4eel'][0][i] - self.energy_data['1-4eel'][1][i] -
                                                 self.energy_data['1-4eel'][2][i])
            self.energy_data['esurf'][3].append(self.energy_data['esurf'][0][i] - self.energy_data['esurf'][1][i] -
                                                self.energy_data['esurf'][2][i])
            self.energy_data['gas'][3].append(self.energy_data['gas'][0][i] - self.energy_data['gas'][1][i] -
                                              self.energy_data['gas'][2][i])
            self.energy_data['solv'][3].append(self.energy_data['solv'][0][i] - self.energy_data['solv'][1][i] -
                                               self.energy_data['solv'][2][i])
            self.energy_data['total'][3].append(self.energy_data['total'][0][i] - self.energy_data['total'][1][i] -
                                                self.energy_data['total'][2][i])

#==============================

   def _consistent(self, terms):
      """ Determine if the internal potential terms are consistent """
      global TOLERANCE

      if not self.valid:
         return False

      for i in range(len(self.energy_data['bond'][0])):
         for item in terms:
            if abs(self.energy_data[item][3][i]) > TOLERANCE:
               return False

      return True

#==============================

   def _removeFromTotal(self, terms):
      """ Remove unprinted energy terms from gas and total arrays """
      for i in range(len(self.energy_data['bond'][0])):
         for item in terms:
            self.energy_data['gas'][0][i] -= self.energy_data[item][0][i]
            self.energy_data['gas'][1][i] -= self.energy_data[item][1][i]
            self.energy_data['gas'][2][i] -= self.energy_data[item][2][i]
            self.energy_data['total'][0][i] -= self.energy_data[item][0][i]
            self.energy_data['total'][1][i] -= self.energy_data[item][1][i]
            self.energy_data['total'][2][i] -= self.energy_data[item][2][i]

      self._assemblediffs() # re-remove these from the differences

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class PBout:
   """ Class for PB output """

#==============================

   def __init__(self, one_traj=True, verb=0, apbs=0, file1='', file2='', file3='', desc=''):
      """ Initializes PBout object with data from passed files, etc."""
      self.one_trajectory = one_traj
      self.verbose = verb
      self.sander_apbs = apbs == 1
      self.description = desc
      self.comfile = file1
      if file2 == '' and file3 == '':
         self.stability = True
      else:
         self.stability = False
         self.recfile = file2
         self.ligfile = file3

      if self.stability:
         self.exists = os.path.exists(file1)
      else:
         self.exists = os.path.exists(file1) and os.path.exists(file2) and os.path.exists(file3)

      self.valid = self.exists and self._read()
      self.printdisper = not self.sander_apbs and not self._disperiszero()

#==============================

   def __str__(self):
      """ String representation of PB binding data -- in final output format """
      if not self.valid:  return 'Problem parsing PB output!\n'
      printstring = '%s' % self.description

      consistent_1_4 = self.one_trajectory and self._consistent(['1-4vdw','1-4eel'])
      consistent_int = self.one_trajectory and self._consistent(['bond','angle','dihedral'])

      print_1_4 = self.verbose == 2 or not consistent_1_4
      print_int = self.verbose == 2 or not consistent_int

      if not print_1_4:  self._removeFromTotal(['1-4vdw','1-4eel'])
      if not print_int:  self._removeFromTotal(['bond','angle','dihedral'])

      if not consistent_int and self.one_trajectory and not self.stability:
         printstring += '\n\nWARNING: INCONSISTENCIES EXIST WITHIN INTERNAL POTENTIAL TERMS (BOND, ANGLE, AND/OR DIHED).' + \
         '\n         CHECK YOUR INPUT FILES AND SYSTEM SETUP. THESE RESULTS MAY NOT BE RELIABLE (check differences)!\n'

      if not consistent_1_4 and self.one_trajectory and not self.stability:
         printstring += '\n\nWARNING: INCONSISTENCIES EXIST WITHIN 1-4 NON-BONDED TERMS.\n' + \
         '         CHECK YOUR INPUT FILES AND SYSTEM SETUP. THESE RESULTS MAY NOT BE RELIABLE (check differences)!\n\n' + \
         'The most common cause of this is inconsistent charge definitions across topology files.\n'

      # add the Complex contribution if we asked to print it
      if self.verbose >= 1 or self.stability:
         sqrtframes = math.sqrt(len(self.energy_data['bond'][0]))
         printstring += 'Complex:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][0]),
                           _stdev(self.energy_data['bond'][0]), _stdev(self.energy_data['bond'][0])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][0]),
                           _stdev(self.energy_data['angle'][0]), _stdev(self.energy_data['angle'][0])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][0]),
                           _stdev(self.energy_data['dihedral'][0]), _stdev(self.energy_data['dihedral'][0])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][0]),
                        _stdev(self.energy_data['vdwaals'][0]), _stdev(self.energy_data['vdwaals'][0])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][0]),
                        _stdev(self.energy_data['eel'][0]), _stdev(self.energy_data['eel'][0])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][0]),
                           _stdev(self.energy_data['1-4vdw'][0]), _stdev(self.energy_data['1-4vdw'][0])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][0]),
                           _stdev(self.energy_data['1-4eel'][0]), _stdev(self.energy_data['1-4eel'][0])/sqrtframes)

         printstring += 'EPB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['epb'][0]),
                        _stdev(self.energy_data['epb'][0]), _stdev(self.energy_data['epb'][0])/sqrtframes) + \
         'ECAVITY        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['esurf'][0]),
                        _stdev(self.energy_data['esurf'][0]), _stdev(self.energy_data['esurf'][0])/sqrtframes)
         if self.printdisper:
            printstring += 'EDISPER        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['edisper'][0]),
                        _stdev(self.energy_data['edisper'][0]), _stdev(self.energy_data['edisper'][0])/sqrtframes)
         printstring += '\nG gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][0]),
                        _stdev(self.energy_data['gas'][0]), _stdev(self.energy_data['gas'][0])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][0]),
                        _stdev(self.energy_data['solv'][0]), _stdev(self.energy_data['solv'][0])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][0]),
                        _stdev(self.energy_data['total'][0]), _stdev(self.energy_data['total'][0])/sqrtframes)
      # add the Receptor contribution if we asked to print it
      if self.verbose >= 1 and not self.stability:
         sqrtframes = math.sqrt(len(self.energy_data['bond'][1]))
         printstring += 'Receptor:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][1]),
                           _stdev(self.energy_data['bond'][1]), _stdev(self.energy_data['bond'][1])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][1]),
                           _stdev(self.energy_data['angle'][1]), _stdev(self.energy_data['angle'][1])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][1]),
                           _stdev(self.energy_data['dihedral'][1]), _stdev(self.energy_data['dihedral'][1])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][1]),
                        _stdev(self.energy_data['vdwaals'][1]), _stdev(self.energy_data['vdwaals'][1])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][1]),
                        _stdev(self.energy_data['eel'][1]), _stdev(self.energy_data['eel'][1])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][1]),
                           _stdev(self.energy_data['1-4vdw'][1]), _stdev(self.energy_data['1-4vdw'][1])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][1]),
                           _stdev(self.energy_data['1-4eel'][1]), _stdev(self.energy_data['1-4eel'][1])/sqrtframes)

         printstring += 'EPB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['epb'][1]),
                        _stdev(self.energy_data['epb'][1]), _stdev(self.energy_data['epb'][1])/sqrtframes) + \
         'ECAVITY        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['esurf'][1]),
                        _stdev(self.energy_data['esurf'][1]), _stdev(self.energy_data['esurf'][1])/sqrtframes)
         if self.printdisper:
            printstring += 'EDISPER        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['edisper'][1]),
                        _stdev(self.energy_data['edisper'][1]), _stdev(self.energy_data['edisper'][1])/sqrtframes)
         printstring += '\nG gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][1]),
                        _stdev(self.energy_data['gas'][1]), _stdev(self.energy_data['gas'][1])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][1]),
                        _stdev(self.energy_data['solv'][1]), _stdev(self.energy_data['solv'][1])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][1]),
                        _stdev(self.energy_data['total'][1]), _stdev(self.energy_data['total'][1])/sqrtframes)
      # add the ligand contribution if we asked to print it
         sqrtframes = math.sqrt(len(self.energy_data['bond'][2]))
         printstring += 'Ligand:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][2]),
                           _stdev(self.energy_data['bond'][2]), _stdev(self.energy_data['bond'][2])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][2]),
                           _stdev(self.energy_data['angle'][2]), _stdev(self.energy_data['angle'][2])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][2]),
                           _stdev(self.energy_data['dihedral'][2]), _stdev(self.energy_data['dihedral'][2])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][2]),
                        _stdev(self.energy_data['vdwaals'][2]), _stdev(self.energy_data['vdwaals'][2])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][2]),
                        _stdev(self.energy_data['eel'][2]), _stdev(self.energy_data['eel'][2])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][2]),
                           _stdev(self.energy_data['1-4vdw'][2]), _stdev(self.energy_data['1-4vdw'][2])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][2]),
                           _stdev(self.energy_data['1-4eel'][2]), _stdev(self.energy_data['1-4eel'][2])/sqrtframes)

         printstring += 'EPB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['epb'][2]),
                        _stdev(self.energy_data['epb'][2]), _stdev(self.energy_data['epb'][2])/sqrtframes) + \
         'ECAVITY        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['esurf'][2]),
                        _stdev(self.energy_data['esurf'][2]), _stdev(self.energy_data['esurf'][2])/sqrtframes)
         if self.printdisper:
            printstring += 'EDISPER        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['edisper'][2]),
                        _stdev(self.energy_data['edisper'][2]), _stdev(self.energy_data['edisper'][2])/sqrtframes)
         printstring += '\nG gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][2]),
                        _stdev(self.energy_data['gas'][2]), _stdev(self.energy_data['gas'][2])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][2]),
                        _stdev(self.energy_data['solv'][2]), _stdev(self.energy_data['solv'][2])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][2]),
                        _stdev(self.energy_data['total'][2]), _stdev(self.energy_data['total'][2])/sqrtframes)
      # now do the differences. If we're using multiple trajectories, we'll have to propagate stdevs
      if not self.stability and self.one_trajectory:
         sqrtframes = math.sqrt(len(self.energy_data['bond'][3]))
         printstring += 'Differences (Complex - Receptor - Ligand):' + \
         '\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][3]), 
                           _stdev(self.energy_data['bond'][3]), _stdev(self.energy_data['bond'][3])/sqrtframes) + \
            'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][3]), 
                           _stdev(self.energy_data['angle'][3]), _stdev(self.energy_data['angle'][3])/sqrtframes) + \
            'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][3]), 
                           _stdev(self.energy_data['dihedral'][3]), _stdev(self.energy_data['dihedral'][3])/sqrtframes)
         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][3]),
                        _stdev(self.energy_data['vdwaals'][3]), _stdev(self.energy_data['vdwaals'][3])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][3]),
                        _stdev(self.energy_data['eel'][3]), _stdev(self.energy_data['eel'][3])/sqrtframes)
         if print_1_4:
            printstring += '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][3]),
                           _stdev(self.energy_data['1-4vdw'][3]), _stdev(self.energy_data['1-4vdw'][3])/sqrtframes) + \
            '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][3]),
                           _stdev(self.energy_data['1-4eel'][3]), _stdev(self.energy_data['1-4eel'][3])/sqrtframes)

         printstring += 'EPB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['epb'][3]),
                        _stdev(self.energy_data['epb'][3]), _stdev(self.energy_data['epb'][3])/sqrtframes) + \
         'ECAVITY        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['esurf'][3]),
                        _stdev(self.energy_data['esurf'][3]), _stdev(self.energy_data['esurf'][3])/sqrtframes)
         if self.printdisper:
            printstring += 'ECAVITY        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['edisper'][3]),
                        _stdev(self.energy_data['edisper'][3]), _stdev(self.energy_data['edisper'][3])/sqrtframes)
         printstring += '\nDELTA G gas    %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][3]),
                        _stdev(self.energy_data['gas'][3]), _stdev(self.energy_data['gas'][3])/sqrtframes) + \
         'DELTA G solv   %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][3]),
                        _stdev(self.energy_data['solv'][3]), _stdev(self.energy_data['solv'][3])/sqrtframes) + \
         '\n DELTA G binding = %19.4f +/- %11.4f %22.4f\n\n\n' % (_avg(self.energy_data['total'][3]),
                        _stdev(self.energy_data['total'][3]), _stdev(self.energy_data['total'][3])/sqrtframes)
      elif not self.stability:
         # maybe not the best. sqrtframes is the sqrt of the minimum number of frames for com/rec/lig
         sqrtframes = min([len(self.energy_data['bond'][0]), len(self.energy_data['bond'][1]), \
                       len(self.energy_data['bond'][2])])
         bondstdev = math.sqrt(_stdev(self.energy_data['bond'][0]) ** 2 +
            _stdev(self.energy_data['bond'][1]) ** 2 + _stdev(self.energy_data['bond'][2]) ** 2)
         anglestdev = math.sqrt(_stdev(self.energy_data['angle'][0]) ** 2 +
            _stdev(self.energy_data['angle'][1]) ** 2 + _stdev(self.energy_data['angle'][2]) ** 2)
         dihedralstdev = math.sqrt(_stdev(self.energy_data['dihedral'][0]) ** 2 +
            _stdev(self.energy_data['dihedral'][1]) ** 2 + _stdev(self.energy_data['dihedral'][2]) ** 2)
         vdwaalsstdev = math.sqrt(_stdev(self.energy_data['vdwaals'][0]) ** 2 +
            _stdev(self.energy_data['vdwaals'][1]) ** 2 + _stdev(self.energy_data['vdwaals'][2]) ** 2)
         eelstdev = math.sqrt(_stdev(self.energy_data['eel'][0]) ** 2 +
            _stdev(self.energy_data['eel'][1]) ** 2 + _stdev(self.energy_data['eel'][2]) ** 2)
         vdwofstdev = math.sqrt(_stdev(self.energy_data['vdwaals'][0]) ** 2 +
            _stdev(self.energy_data['1-4vdw'][1]) ** 2 + _stdev(self.energy_data['vdwaals'][2]) ** 2)
         eelofstdev = math.sqrt(_stdev(self.energy_data['1-4eel'][0]) ** 2 +
            _stdev(self.energy_data['1-4eel'][1]) ** 2 + _stdev(self.energy_data['1-4eel'][2]) ** 2)
         epbstdev = math.sqrt(_stdev(self.energy_data['epb'][0]) ** 2 +
            _stdev(self.energy_data['epb'][1]) ** 2 + _stdev(self.energy_data['epb'][2]) ** 2)
         esurfstdev = math.sqrt(_stdev(self.energy_data['esurf'][0]) ** 2 +
            _stdev(self.energy_data['esurf'][1]) ** 2 + _stdev(self.energy_data['esurf'][2]) ** 2)
         if not self.sander_apbs:
            edisperstdev = math.sqrt(_stdev(self.energy_data['edisper'][0]) ** 2 +
               _stdev(self.energy_data['edisper'][1]) ** 2 + _stdev(self.energy_data['edisper'][2]) ** 2)
         gasstdev = math.sqrt(_stdev(self.energy_data['gas'][0]) ** 2 +
            _stdev(self.energy_data['gas'][1]) ** 2 + _stdev(self.energy_data['gas'][2]) ** 2)
         solvstdev = math.sqrt(_stdev(self.energy_data['solv'][0]) ** 2 +
            _stdev(self.energy_data['solv'][1]) ** 2 + _stdev(self.energy_data['solv'][2]) ** 2)
         totalstdev = math.sqrt(_stdev(self.energy_data['total'][0]) ** 2 +
            _stdev(self.energy_data['total'][1]) ** 2 + _stdev(self.energy_data['total'][2]) ** 2)
         printstring += 'Differences (Complex - Receptor - Ligand):' + \
         '\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n' + \
         'BOND           %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['bond'][0]) -
            _avg(self.energy_data['bond'][1])- _avg(self.energy_data['bond'][2]), bondstdev, 
            bondstdev/sqrtframes) + \
         'ANGLE          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['angle'][0]) -
            _avg(self.energy_data['angle'][1]) - _avg(self.energy_data['angle'][2]), anglestdev, 
            anglestdev / sqrtframes) + \
         'DIHED          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['dihedral'][0]) -
            _avg(self.energy_data['dihedral'][1])- _avg(self.energy_data['dihedral'][2]), dihedralstdev, 
            dihedralstdev / sqrtframes) + \
         'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][0]) -
            _avg(self.energy_data['vdwaals'][1])- _avg(self.energy_data['vdwaals'][2]), vdwaalsstdev, 
            vdwaalsstdev / sqrtframes) + \
         '1-4 VDW        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4vdw'][0]) -
            _avg(self.energy_data['1-4vdw'][1])- _avg(self.energy_data['1-4vdw'][2]), vdwofstdev, 
            vdwofstdev / sqrtframes) + \
         '1-4 EEL        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['1-4eel'][0]) -
            _avg(self.energy_data['1-4eel'][1])- _avg(self.energy_data['1-4eel'][2]), eelofstdev, 
            eelofstdev / sqrtframes) + \
         'EPB            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['epb'][0]) -
            _avg(self.energy_data['epb'][1])- _avg(self.energy_data['epb'][2]), epbstdev, 
            epbstdev / sqrtframes) + \
         'ECAVITY        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['esurf'][0]) -
            _avg(self.energy_data['esurf'][1])- _avg(self.energy_data['esurf'][2]), esurfstdev, 
            esurfstdev / sqrtframes)
         if self.printdisper:
            printstring += 'ECAVITY        %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['edisper'][0]) -
               _avg(self.energy_data['edisper'][1])- _avg(self.energy_data['edisper'][2]), edisperstdev, 
               edisperstdev / sqrtframes)
         printstring += '\nDELTA G gas    %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][0]) -
            _avg(self.energy_data['gas'][1])- _avg(self.energy_data['gas'][2]), gasstdev, 
            gasstdev / sqrtframes) + \
         'DELTA G solv   %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][0]) -
            _avg(self.energy_data['solv'][1])- _avg(self.energy_data['solv'][2]), solvstdev, 
            solvstdev / sqrtframes) + \
         '\n DELTA G binding = %19.4f +/- %11.4f %22.4f\n' % (_avg(self.energy_data['total'][0]) - 
            _avg(self.energy_data['total'][1]) - _avg(self.energy_data['total'][2]), totalstdev,
            totalstdev / sqrtframes)

      return printstring

#==============================

   def AvgStdev(self):
      """ Returns the average and standard deviations of the total PB binding data """
      if self.stability:
         return _avg(self.energy_data['total'][0]), _stdev(self.energy_data['total'][0])

      if self.one_trajectory:
         return _avg(self.energy_data['total'][3]), _stdev(self.energy_data['total'][3])

      avg = _avg(self.energy_data['total'][0]) - _avg(self.energy_data['total'][1]) - _avg(self.energy_data['total'][2])
      stdev = math.sqrt(_stdev(self.energy_data['total'][0]) ** 2 + _stdev(self.energy_data['total'][1]) ** 2 + 
                        _stdev(self.energy_data['total'][2]) ** 2)

      return avg, stdev

#==============================

   def _read(self):
      """ Parses output files and fills data arrays """
      # dictionary of empty arrays to be filled
      self.energy_data = {'bond' : [[],[],[],[]], 'angle' : [[],[],[],[]], 'dihedral' : [[],[],[],[]],
                          'vdwaals' : [[],[],[],[]], 'eel' : [[],[],[],[]], '1-4vdw' : [[],[],[],[]], 
                          '1-4eel' : [[],[],[],[]], 'gas' : [[],[],[],[]], 'epb' : [[],[],[],[]], 
                          'esurf' : [[],[],[],[]], 'solv' : [[],[],[],[]], 'edisper' : [[],[],[],[]],
                          'total' : [[],[],[],[]]}
      com = open(self.comfile, 'r')
      if not self.stability:
         rec = open(self.recfile, 'r')
         lig = open(self.ligfile, 'r')

      try:
         self._getenergies(com)
         com.close()
         if not self.stability:
            self._getenergies(rec, index=1)
            self._getenergies(lig, index=2)
            rec.close()
            lig.close()
         if self.one_trajectory:
            self._assemblediffs()

         return True
      except:
         print >> sys.stderr, 'Error: Could not parse PB output files! Check for missing values (*********).'
         return False

#==============================

   def _getenergies(self, file, index=0):
      """ Actually parses energy terms from output files and dumps them into the arrays """
      rawline = file.readline()  # get first line

      while len(rawline) != 0:
         
         if rawline[0:5] == ' BOND':
            # split the line into individual components
            words = rawline.split()
            # bond, angle, and dihedral parameters are on this line
            bond    = float(words[2])
            angle   = float(words[5])
            dihed   = float(words[8])
            # split the next line into components
            words = file.readline().split()
            # vdwaals, eel, and epb are on this line
            vdwaals = float(words[2])
            eel     = float(words[5])
            epb     = float(words[8])
            # split the next line into components
            words = file.readline().split()
            # 1-4 vdw and 1-4 eel are on this line
            vdw1_4  = float(words[3])
            eel1_4  = float(words[7])
            # split the next line into components
            words = file.readline().split()
            # ecavity/enpolar and edisper is on this line
            esurf   = float(words[2])
            if not self.sander_apbs:
               edisper = float(words[5])
            # now load these values into the data arrays
            self.energy_data['bond'][index].append(bond)
            self.energy_data['angle'][index].append(angle)
            self.energy_data['dihedral'][index].append(dihed)
            self.energy_data['vdwaals'][index].append(vdwaals)
            self.energy_data['eel'][index].append(eel)
            self.energy_data['epb'][index].append(epb)
            self.energy_data['1-4vdw'][index].append(vdw1_4)
            self.energy_data['1-4eel'][index].append(eel1_4)
            self.energy_data['esurf'][index].append(esurf)
            if not self.sander_apbs:
               self.energy_data['edisper'][index].append(edisper)
            self.energy_data['gas'][index].append(bond + angle + dihed + vdwaals + eel + vdw1_4 + eel1_4)
            if not self.sander_apbs:
               self.energy_data['solv'][index].append(epb + esurf + edisper)
               self.energy_data['total'][index].append(bond + angle + dihed + vdwaals + eel + vdw1_4 + 
                                                       eel1_4 + epb + esurf + edisper)
            else:
               self.energy_data['solv'][index].append(epb + esurf)
               self.energy_data['total'][index].append(bond + angle + dihed + vdwaals + eel + vdw1_4 + 
                                                       eel1_4 + epb + esurf)

         rawline = file.readline() # read the next line

#==============================

   def _assemblediffs(self):
      """ Fills the difference arrays from the extracted data """
      self.energy_data['bond'][3]     = []
      self.energy_data['angle'][3]    = []
      self.energy_data['dihedral'][3] = []
      self.energy_data['vdwaals'][3]  = []
      self.energy_data['eel'][3]      = []
      self.energy_data['epb'][3]      = []
      self.energy_data['1-4vdw'][3]   = []
      self.energy_data['1-4eel'][3]   = []
      self.energy_data['esurf'][3]    = []
      if not self.sander_apbs:
         self.energy_data['edisper'][3]  = []
      self.energy_data['gas'][3]      = []
      self.energy_data['solv'][3]     = []
      self.energy_data['total'][3]    = []
      if self.stability:
         for i in range(len(self.energy_data['bond'][0])):
            self.energy_data['bond'][3].append(self.energy_data['bond'][0][i])
            self.energy_data['angle'][3].append(self.energy_data['angle'][0][i])
            self.energy_data['dihedral'][3].append(self.energy_data['dihedral'][0][i])
            self.energy_data['vdwaals'][3].append(self.energy_data['vdwaals'][0][i])
            self.energy_data['eel'][3].append(self.energy_data['eel'][0][i])
            self.energy_data['epb'][3].append(self.energy_data['epb'][0][i])
            self.energy_data['1-4vdw'][3].append(self.energy_data['1-4vdw'][0][i])
            self.energy_data['1-4eel'][3].append(self.energy_data['1-4eel'][0][i])
            self.energy_data['esurf'][3].append(self.energy_data['esurf'][0][i])
            if not self.sander_apbs:
               self.energy_data['edisper'][3].append(self.energy_data['edisper'][0][i])
            self.energy_data['gas'][3].append(self.energy_data['gas'][0][i])
            self.energy_data['solv'][3].append(self.energy_data['solv'][0][i])
            self.energy_data['total'][3].append(self.energy_data['total'][0][i])
      else:
         for i in range(len(self.energy_data['bond'][0])):
            self.energy_data['bond'][3].append(self.energy_data['bond'][0][i] - self.energy_data['bond'][1][i] -
                                               self.energy_data['bond'][2][i])
            self.energy_data['angle'][3].append(self.energy_data['angle'][0][i] - self.energy_data['angle'][1][i] -
                                                self.energy_data['angle'][2][i])
            self.energy_data['dihedral'][3].append(self.energy_data['dihedral'][0][i] - self.energy_data['dihedral'][1][i] -
                                                self.energy_data['dihedral'][2][i])
            self.energy_data['vdwaals'][3].append(self.energy_data['vdwaals'][0][i] - self.energy_data['vdwaals'][1][i] -
                                                  self.energy_data['vdwaals'][2][i])
            self.energy_data['eel'][3].append(self.energy_data['eel'][0][i] - self.energy_data['eel'][1][i] -
                                              self.energy_data['eel'][2][i])
            self.energy_data['epb'][3].append(self.energy_data['epb'][0][i] - self.energy_data['epb'][1][i] -
                                              self.energy_data['epb'][2][i])
            self.energy_data['1-4vdw'][3].append(self.energy_data['1-4vdw'][0][i] - self.energy_data['1-4vdw'][1][i] -
                                                 self.energy_data['1-4vdw'][2][i])
            self.energy_data['1-4eel'][3].append(self.energy_data['1-4eel'][0][i] - self.energy_data['1-4eel'][1][i] -
                                                 self.energy_data['1-4eel'][2][i])
            self.energy_data['esurf'][3].append(self.energy_data['esurf'][0][i] - self.energy_data['esurf'][1][i] -
                                                self.energy_data['esurf'][2][i])
            if not self.sander_apbs:
               self.energy_data['edisper'][3].append(self.energy_data['edisper'][0][i] - self.energy_data['edisper'][1][i] -
                                                self.energy_data['edisper'][2][i])
            self.energy_data['gas'][3].append(self.energy_data['gas'][0][i] - self.energy_data['gas'][1][i] -
                                              self.energy_data['gas'][2][i])
            self.energy_data['solv'][3].append(self.energy_data['solv'][0][i] - self.energy_data['solv'][1][i] -
                                               self.energy_data['solv'][2][i])
            self.energy_data['total'][3].append(self.energy_data['total'][0][i] - self.energy_data['total'][1][i] -
                                                self.energy_data['total'][2][i])

#==============================

   def _consistent(self, terms):
      """ Determine if the internal potential terms are consistent """
      global TOLERANCE

      if not self.valid:
         return False

      for i in range(len(self.energy_data['bond'][0])):
         for item in terms:
            if abs(self.energy_data[item][3][i]) > TOLERANCE:
               return False

      return True

#==============================

   def _disperiszero(self):
      """ Determines if the dispersion term is 0, which it often is """
      iszero = True
      global TOLERANCE

      if not self.stability:
         if abs(_avg(self.energy_data['edisper'][0])) > TOLERANCE or \
            abs(_avg(self.energy_data['edisper'][1])) > TOLERANCE or \
            abs(_avg(self.energy_data['edisper'][2])) > TOLERANCE or \
            abs(_stdev(self.energy_data['edisper'][0])) > TOLERANCE or \
            abs(_stdev(self.energy_data['edisper'][1])) > TOLERANCE or \
            abs(_stdev(self.energy_data['edisper'][2])) > TOLERANCE:

            iszero = False
      else:
         if abs(_avg(self.energy_data['edisper'][0])) > TOLERANCE or \
            abs(_stdev(self.energy_data['edisper'][0])) > TOLERANCE:

            iszero = False

      return iszero

#==============================

   def _removeFromTotal(self, terms):
      """ Remove a specified energy term from the gas and total contributions if it's unprinted """
      for i in range(len(self.energy_data['bond'][0])):
         for item in terms:
            self.energy_data['gas'][0][i] -= self.energy_data[item][0][i]
            self.energy_data['gas'][1][i] -= self.energy_data[item][1][i]
            self.energy_data['gas'][2][i] -= self.energy_data[item][2][i]
            self.energy_data['total'][0][i] -= self.energy_data[item][0][i]
            self.energy_data['total'][1][i] -= self.energy_data[item][1][i]
            self.energy_data['total'][2][i] -= self.energy_data[item][2][i]

      self._assemblediffs() # re-remove these from the differences

#==============================

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

class RISMout:
   """ 3D-RISM output class """

#==============================

   def __init__(self, one_traj=True, verb=0, file1='', file2='', file3='', desc='', solvtype=0):
      """ Initializes RISMout object with data from passed files, etc."""
      self.one_trajectory = one_traj
      self.verbose = verb
      self.description = desc
      self.comfile = file1
      if file2 == '' and file3 == '':
         self.stability = True
      else:
         self.stability = False
         self.recfile = file2
         self.ligfile = file3

      if self.stability:
         self.exists = os.path.exists(file1)
      else:
         self.exists = os.path.exists(file1) and os.path.exists(file2) and os.path.exists(file3)

      self.valid = self.exists and self._read(solvtype)

#==============================

   def __str__(self):
      """ String representation of RISM output data -- in final output format """
      if not self.valid:  return 'Problem parsing 3D-RISM output!\n'
      printstring = '%s' % self.description

      consistent = self.one_trajectory and self._consistent(['eint'])

      print_int = self.verbose == 2 or not consistent

      if not print_int:  self._removeFromTotal(['eint'])

      if not consistent and self.one_trajectory and not self.stability:
         printstring += '\n\nWARNING: INCONSISTENCIES EXIST WITHIN INTERNAL POTENTIAL TERMS.' + \
         '\n         CHECK YOUR INPUT FILES AND SYSTEM SETUP. THESE RESULTS MAY NOT BE RELIABLE (check differences)!\n'

      # add the Complex contribution if we asked to print it
      if self.verbose >= 1 or self.stability:
         sqrtframes = math.sqrt(len(self.energy_data['eint'][0]))
         printstring += 'Complex:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'INTERNAL       %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eint'][0]),
                           _stdev(self.energy_data['eint'][0]), _stdev(self.energy_data['eint'][0])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][0]),
                        _stdev(self.energy_data['vdwaals'][0]), _stdev(self.energy_data['vdwaals'][0])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][0]),
                        _stdev(self.energy_data['eel'][0]), _stdev(self.energy_data['eel'][0])/sqrtframes) + \
         'ERISM          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['erism'][0]),
                        _stdev(self.energy_data['erism'][0]), _stdev(self.energy_data['erism'][0])/sqrtframes)
         if len(self.energy_data['pol'][0]) > 0:
            printstring += '   POLAR       %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['pol'][0]),
                        _stdev(self.energy_data['pol'][0]), _stdev(self.energy_data['pol'][0])/sqrtframes) + \
            '   APOLAR      %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['apol'][0]),
                        _stdev(self.energy_data['apol'][0]), _stdev(self.energy_data['apol'][0])/sqrtframes)
         printstring += '\nG gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][0]),
                        _stdev(self.energy_data['gas'][0]), _stdev(self.energy_data['gas'][0])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][0]),
                        _stdev(self.energy_data['solv'][0]), _stdev(self.energy_data['solv'][0])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][0]),
                        _stdev(self.energy_data['total'][0]), _stdev(self.energy_data['total'][0])/sqrtframes)
      # add the Receptor contribution if we asked to print it
      if self.verbose >= 1 and not self.stability:
         sqrtframes = math.sqrt(len(self.energy_data['eint'][1]))
         printstring += 'Receptor:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'INTERNAL       %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eint'][1]),
                           _stdev(self.energy_data['eint'][1]), _stdev(self.energy_data['eint'][1])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][1]),
                        _stdev(self.energy_data['vdwaals'][1]), _stdev(self.energy_data['vdwaals'][1])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][1]),
                        _stdev(self.energy_data['eel'][1]), _stdev(self.energy_data['eel'][1])/sqrtframes) + \
         'ERISM          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['erism'][1]),
                        _stdev(self.energy_data['erism'][1]), _stdev(self.energy_data['erism'][1])/sqrtframes)
         if len(self.energy_data['pol'][1]) > 0:
            printstring += '   POLAR       %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['pol'][1]),
                        _stdev(self.energy_data['pol'][1]), _stdev(self.energy_data['pol'][1])/sqrtframes) + \
            '   APOLAR      %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['apol'][1]),
                        _stdev(self.energy_data['apol'][1]), _stdev(self.energy_data['apol'][1])/sqrtframes)
         printstring += '\nG gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][1]),
                        _stdev(self.energy_data['gas'][1]), _stdev(self.energy_data['gas'][1])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][1]),
                        _stdev(self.energy_data['solv'][1]), _stdev(self.energy_data['solv'][1])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][1]),
                        _stdev(self.energy_data['total'][1]), _stdev(self.energy_data['total'][1])/sqrtframes)
      # add the ligand contribution if we asked to print it
         sqrtframes = math.sqrt(len(self.energy_data['eint'][2]))
         printstring += 'Ligand:\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'INTERNAL       %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eint'][2]),
                           _stdev(self.energy_data['eint'][2]), _stdev(self.energy_data['eint'][2])/sqrtframes)

         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][2]),
                        _stdev(self.energy_data['vdwaals'][2]), _stdev(self.energy_data['vdwaals'][2])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][2]),
                        _stdev(self.energy_data['eel'][2]), _stdev(self.energy_data['eel'][2])/sqrtframes) + \
         'ERISM          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['erism'][2]),
                        _stdev(self.energy_data['erism'][2]), _stdev(self.energy_data['erism'][2])/sqrtframes)
         if len(self.energy_data['pol'][2]) > 0:
            printstring += '   POLAR       %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['pol'][2]),
                        _stdev(self.energy_data['pol'][2]), _stdev(self.energy_data['pol'][2])/sqrtframes) + \
            '   APOLAR      %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['apol'][2]),
                        _stdev(self.energy_data['apol'][2]), _stdev(self.energy_data['apol'][2])/sqrtframes)
         printstring += '\nG gas          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][2]),
                        _stdev(self.energy_data['gas'][2]), _stdev(self.energy_data['gas'][2])/sqrtframes) + \
         'G solv         %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][2]),
                        _stdev(self.energy_data['solv'][2]), _stdev(self.energy_data['solv'][2])/sqrtframes) + \
         'TOTAL          %20.4f %21.4f %19.4f\n\n\n' % (_avg(self.energy_data['total'][2]),
                        _stdev(self.energy_data['total'][2]), _stdev(self.energy_data['total'][2])/sqrtframes)
      # now do the differences. If we're using multiple trajectories, we'll have to propagate stdevs
      if not self.stability and self.one_trajectory:
         sqrtframes = math.sqrt(len(self.energy_data['eint'][3]))
         printstring += 'Differences (Complex - Receptor - Ligand):' + \
         '\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n'
         if print_int:
            printstring += 'INTERNAL       %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eint'][3]), 
                           _stdev(self.energy_data['eint'][3]), _stdev(self.energy_data['eint'][3])/sqrtframes)
         printstring += 'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][3]),
                        _stdev(self.energy_data['vdwaals'][3]), _stdev(self.energy_data['vdwaals'][3])/sqrtframes) + \
         'EEL            %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eel'][3]),
                        _stdev(self.energy_data['eel'][3]), _stdev(self.energy_data['eel'][3])/sqrtframes) + \
         'ERISM          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['erism'][3]),
                        _stdev(self.energy_data['erism'][3]), _stdev(self.energy_data['erism'][3])/sqrtframes)
         if len(self.energy_data['pol'][3]) > 0:
            printstring += '   POLAR       %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['pol'][3]),
                        _stdev(self.energy_data['pol'][3]), _stdev(self.energy_data['pol'][3])/sqrtframes) + \
            '   APOLAR      %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['apol'][3]),
                        _stdev(self.energy_data['apol'][3]), _stdev(self.energy_data['apol'][3])/sqrtframes)
         printstring += '\nDELTA G gas    %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][3]),
                        _stdev(self.energy_data['gas'][3]), _stdev(self.energy_data['gas'][3])/sqrtframes) + \
         'DELTA G solv   %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][3]),
                        _stdev(self.energy_data['solv'][3]), _stdev(self.energy_data['solv'][3])/sqrtframes) + \
         '\n DELTA G binding = %19.4f +/- %11.4f %22.4f\n\n\n' % (_avg(self.energy_data['total'][3]),
                        _stdev(self.energy_data['total'][3]), _stdev(self.energy_data['total'][3])/sqrtframes)
      elif not self.stability:
         # maybe not the best. sqrtframes is the sqrt of the minimum number of frames for com/rec/lig
         sqrtframes = min([len(self.energy_data['eint'][0]), len(self.energy_data['eint'][1]), \
                       len(self.energy_data['eint'][2])])
         eintstdev = math.sqrt(_stdev(self.energy_data['eint'][0]) ** 2 +
            _stdev(self.energy_data['eint'][1]) ** 2 + _stdev(self.energy_data['eint'][2]) ** 2)
         vdwaalsstdev = math.sqrt(_stdev(self.energy_data['vdwaals'][0]) ** 2 +
            _stdev(self.energy_data['vdwaals'][1]) ** 2 + _stdev(self.energy_data['vdwaals'][2]) ** 2)
         eelstdev = math.sqrt(_stdev(self.energy_data['eel'][0]) ** 2 +
            _stdev(self.energy_data['eel'][1]) ** 2 + _stdev(self.energy_data['eel'][2]) ** 2)
         erismstdev = math.sqrt(_stdev(self.energy_data['erism'][0]) ** 2 +
            _stdev(self.energy_data['erism'][1]) ** 2 + _stdev(self.energy_data['erism'][2]) ** 2)
         if len(self.energy_data['pol'][3]) > 0:
            polstdev = math.sqrt(_stdev(self.energy_data['pol'][0]) ** 2 +
               _stdev(self.energy_data['pol'][1]) ** 2 + _stdev(self.energy_data['pol'][2]) ** 2)
            apolstdev = math.sqrt(_stdev(self.energy_data['apol'][0]) ** 2 +
               _stdev(self.energy_data['apol'][1]) ** 2 + _stdev(self.energy_data['apol'][2]) ** 2)
         gasstdev = math.sqrt(_stdev(self.energy_data['gas'][0]) ** 2 +
            _stdev(self.energy_data['gas'][1]) ** 2 + _stdev(self.energy_data['gas'][2]) ** 2)
         solvstdev = math.sqrt(_stdev(self.energy_data['solv'][0]) ** 2 +
            _stdev(self.energy_data['solv'][1]) ** 2 + _stdev(self.energy_data['solv'][2]) ** 2)
         totalstdev = math.sqrt(_stdev(self.energy_data['total'][0]) ** 2 +
            _stdev(self.energy_data['total'][1]) ** 2 + _stdev(self.energy_data['total'][2]) ** 2)
         printstring += 'Differences (Complex - Receptor - Ligand):' + \
         '\nEnergy Component            Average              Std. Dev.   Std. Err. of Mean\n' + \
         '-------------------------------------------------------------------------------\n' + \
         'INTERNAL       %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['eint'][0]) -
            _avg(self.energy_data['eint'][1])- _avg(self.energy_data['eint'][2]), eintstdev, 
            eintstdev/sqrtframes) + \
         'VDWAALS        %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['vdwaals'][0]) -
            _avg(self.energy_data['vdwaals'][1])- _avg(self.energy_data['vdwaals'][2]), vdwaalsstdev, 
            vdwaalsstdev / sqrtframes) + \
         'ERISM          %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['erism'][0]) -
            _avg(self.energy_data['erism'][1])- _avg(self.energy_data['erism'][2]), erismstdev, 
            erismstdev / sqrtframes)
         if len(self.energy_data['pol'][0]) > 0:
            printstring += '   POLAR       %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['pol'][0]) -
            _avg(self.energy_data['pol'][1])- _avg(self.energy_data['pol'][2]), polstdev, 
            polstdev / sqrtframes) + \
            '   APOLAR      %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['erism'][0]) -
            _avg(self.energy_data['erism'][1])- _avg(self.energy_data['erism'][2]), apolstdev, 
            apolstdev / sqrtframes)
         printstring += '\nDELTA G gas    %20.4f %21.4f %19.4f\n' % (_avg(self.energy_data['gas'][0]) -
            _avg(self.energy_data['gas'][1])- _avg(self.energy_data['gas'][2]), gasstdev, 
            gasstdev / sqrtframes) + \
         'DELTA G solv   %20.4f %21.4f %19.4f\n\n' % (_avg(self.energy_data['solv'][0]) -
            _avg(self.energy_data['solv'][1])- _avg(self.energy_data['solv'][2]), solvstdev, 
            solvstdev / sqrtframes) + \
         '\n DELTA G binding = %19.4f +/- %11.4f %22.4f\n' % (_avg(self.energy_data['total'][0]) - 
            _avg(self.energy_data['total'][1]) - _avg(self.energy_data['total'][2]), totalstdev,
            totalstdev / sqrtframes)

      return printstring

#==============================

   def AvgStdev(self):
      """ Returns the average and standard deviations of the total binding free energy """
      if self.stability:
         return _avg(self.energy_data['total'][0]), _stdev(self.energy_data['total'][0])

      if self.one_trajectory:
         return _avg(self.energy_data['total'][3]), _stdev(self.energy_data['total'][3])

      avg = _avg(self.energy_data['total'][0]) - _avg(self.energy_data['total'][1]) - _avg(self.energy_data['total'][2])
      stdev = math.sqrt(_stdev(self.energy_data['total'][0]) ** 2 + _stdev(self.energy_data['total'][1]) ** 2 + 
                        _stdev(self.energy_data['total'][2]) ** 2)

      return avg, stdev

#==============================

   def _read(self, solvtype=0):
      """ Parses the RISM output and fills the data arrays """
      # dictionary of empty arrays to be filled
      self.energy_data = {'eint' : [[],[],[],[]], 'vdwaals' : [[],[],[],[]], 'eel' : [[],[],[],[]], 
                          'gas' : [[],[],[],[]], 'pol' : [[],[],[],[]], 'apol' : [[],[],[],[]], 
                          'erism' : [[],[],[],[]], 'solv' : [[],[],[],[]], 'total' : [[],[],[],[]]}
      com = open(self.comfile, 'r')
      if not self.stability:
         rec = open(self.recfile, 'r')
         lig = open(self.ligfile, 'r')

      try:
         self._getenergies(com, solvtype)
         com.close()
         if not self.stability:
            self._getenergies(rec, solvtype, index=1)
            self._getenergies(lig, solvtype, index=2)
            rec.close()
            lig.close()
         if self.one_trajectory:
            self._assemblediffs()

         return True
      except:
         print >> sys.stderr, 'Error: Could not parse RISM output files! Check for missing values (*********).'
         return False

#==============================

   def _getenergies(self, file, solvtype, index=0):
      """ Actually parses the potential terms from the output file and puts the values in the arrays """
      rawline = file.readline()  # get first line

      while len(rawline) != 0:
         
         if rawline[0:12] == 'solute_epot:':
            pol = 0
            apol = 0
            # split the line into individual components
            words = rawline.split()
            # All energy values are on this line, according to rism src comments
            vdwaals = float(words[2])
            eel     = float(words[3])
            # Next terms are BOND, ANGLE, DIHED, 1-4 VDW, 1-4 EEL (all added into eint)
            eint    = float(words[4]) + \
                      float(words[5]) + \
                      float(words[6]) + \
                      float(words[7]) + \
                      float(words[8])
            if solvtype == 0:
               words = file.readline().split()
               erism = float(words[1])
               file.readline()
               file.readline()
               file.readline()
               file.readline()
               file.readline()
               words = file.readline().split()
               try:
                  if words[0] == 'rism_polar_:':
                     pol = float(words[1])
                     words = file.readline().split()
                     apol = float(words[1])
               except IndexError:
                  pass
            elif solvtype == 1:
               file.readline()
               words = file.readline().split()
               erism = float(words[1])
               file.readline()
               file.readline()
               file.readline()
               file.readline()
               file.readline()
               file.readline()
               words = file.readline().split()
               try:
                  if words[0] == 'rism_polGF_:':
                     pol = float(words[1])
                     words = file.readline().split()
                     apol = float(words[1])
               except IndexError:
                  pass
            etotal  = vdwaals + eel + eint + erism

            # now load these values into the data arrays
            self.energy_data['eint'][index].append(eint)
            self.energy_data['vdwaals'][index].append(vdwaals)
            self.energy_data['eel'][index].append(eel)
            self.energy_data['erism'][index].append(erism)
            if not (pol == 0 and apol == 0):
               self.energy_data['pol'][index].append(pol)
               self.energy_data['apol'][index].append(apol)
            self.energy_data['gas'][index].append(eint + vdwaals + eel)
            self.energy_data['solv'][index].append(erism)
            self.energy_data['total'][index].append(etotal)

         rawline = file.readline() # read the next line

#==============================

   def _assemblediffs(self):
      """ Generate difference array from parsed data for complex, receptor, and ligand """
      self.energy_data['eint'][3]     = []
      self.energy_data['vdwaals'][3]  = []
      self.energy_data['eel'][3]      = []
      self.energy_data['erism'][3]    = []
      self.energy_data['pol'][3]      = []
      self.energy_data['apol'][3]     = []
      self.energy_data['gas'][3]      = []
      self.energy_data['solv'][3]     = []
      self.energy_data['total'][3]    = []
      if self.stability:
         for i in range(len(self.energy_data['eint'][0])):
            self.energy_data['eint'][3].append(self.energy_data['eint'][0][i])
            self.energy_data['vdwaals'][3].append(self.energy_data['vdwaals'][0][i])
            self.energy_data['eel'][3].append(self.energy_data['eel'][0][i])
            self.energy_data['erism'][3].append(self.energy_data['erism'][0][i])
            self.energy_data['pol'][3].append(self.energy_data['pol'][0][i])
            self.energy_data['apol'][3].append(self.energy_data['apol'][0][i])
            self.energy_data['gas'][3].append(self.energy_data['gas'][0][i])
            self.energy_data['solv'][3].append(self.energy_data['solv'][0][i])
            self.energy_data['total'][3].append(self.energy_data['total'][0][i])
      else:
         for i in range(len(self.energy_data['eint'][0])):
            self.energy_data['eint'][3].append(self.energy_data['eint'][0][i] - self.energy_data['eint'][1][i] -
                                               self.energy_data['eint'][2][i])
            self.energy_data['vdwaals'][3].append(self.energy_data['vdwaals'][0][i] - self.energy_data['vdwaals'][1][i] -
                                                  self.energy_data['vdwaals'][2][i])
            self.energy_data['eel'][3].append(self.energy_data['eel'][0][i] - self.energy_data['eel'][1][i] -
                                              self.energy_data['eel'][2][i])
            self.energy_data['erism'][3].append(self.energy_data['erism'][0][i] - self.energy_data['erism'][1][i] -
                                              self.energy_data['erism'][2][i])
            if len(self.energy_data['pol'][0]) > 0:
               self.energy_data['pol'][3].append(self.energy_data['pol'][0][i] - self.energy_data['pol'][1][i] -
                                                 self.energy_data['pol'][2][i])
               self.energy_data['apol'][3].append(self.energy_data['apol'][0][i] - self.energy_data['apol'][1][i] -
                                                 self.energy_data['apol'][2][i])
            self.energy_data['gas'][3].append(self.energy_data['gas'][0][i] - self.energy_data['gas'][1][i] -
                                              self.energy_data['gas'][2][i])
            self.energy_data['solv'][3].append(self.energy_data['solv'][0][i] - self.energy_data['solv'][1][i] -
                                               self.energy_data['solv'][2][i])
            self.energy_data['total'][3].append(self.energy_data['total'][0][i] - self.energy_data['total'][1][i] -
                                                self.energy_data['total'][2][i])

#==============================

   def _consistent(self, terms):
      """ Determine if the internal potential terms are consistent """
      global TOLERANCE

      if not self.valid:
         return False

      for i in range(len(self.energy_data['eint'][0])):
         for item in terms:
            if abs(self.energy_data[item][3][i]) > TOLERANCE:
               return False

      return True

#==============================

   def _removeFromTotal(self, terms):
      """ Remove contributions of unprinted potential terms from the gas and total energy contributions """
      for i in range(len(self.energy_data['eint'][0])):
         for item in terms:
            self.energy_data['gas'][0][i] -= self.energy_data[item][0][i]
            self.energy_data['gas'][1][i] -= self.energy_data[item][1][i]
            self.energy_data['gas'][2][i] -= self.energy_data[item][2][i]
            self.energy_data['total'][0][i] -= self.energy_data[item][0][i]
            self.energy_data['total'][1][i] -= self.energy_data[item][1][i]
            self.energy_data['total'][2][i] -= self.energy_data[item][2][i]

      self._assemblediffs() # re-remove these from the differences

#==============================

#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
